/* Generated by Cython 3.0.12 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "define_macros": [
            [
                "NPY_NO_DEPRECATED_API",
                "NPY_1_7_API_VERSION"
            ]
        ],
        "depends": [
            "/tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/_core/include/numpy/arrayobject.h",
            "/tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/_core/include/numpy/arrayscalars.h",
            "/tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/_core/include/numpy/ndarrayobject.h",
            "/tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/_core/include/numpy/ndarraytypes.h",
            "/tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/_core/include/numpy/ufuncobject.h"
        ],
        "include_dirs": [
            "/tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/_core/include"
        ],
        "name": "pyart.io._sigmetfile",
        "sources": [
            "pyart/io/_sigmetfile.pyx"
        ]
    },
    "module_name": "pyart.io._sigmetfile"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_12" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x03000CF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(Py_GIL_DISABLED) || defined(Py_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 0
  #endif
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && PY_VERSION_HEX < 0x030d0000 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__;  (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pyart__io___sigmetfile
#define __PYX_HAVE_API__pyart__io___sigmetfile
/* Early includes */
#include <string.h>
#include <stdio.h>

    /* Using NumPy API declarations from "numpy/__init__.cython-30.pxd" */
    
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif (defined(_Complex_I) && !defined(_MSC_VER)) || ((defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) && !defined(__STDC_NO_COMPLEX__) && !defined(_MSC_VER))
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "pyart/io/_sigmetfile.pyx",
  "<stringsource>",
  "__init__.cython-30.pxd",
  "type.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* BufferFormatStructs.proto */
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;

/* #### Code section: numeric_typedefs ### */

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":787
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":788
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":789
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":790
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":794
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":795
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":796
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":797
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":801
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":802
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":809
 * ctypedef double complex complex128_t
 * 
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":810
 * 
 * ctypedef npy_longlong   longlong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":812
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":813
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":815
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":816
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":817
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef float complex       cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* #### Code section: complex_type_declarations ### */
/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< long double > __pyx_t_long_double_complex;
  #else
    typedef long double _Complex __pyx_t_long_double_complex;
  #endif
#else
    typedef struct { long double real, imag; } __pyx_t_long_double_complex;
#endif
static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double, long double);

/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile;

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1113
 * 
 * # Iterator API added in v1.6
 * ctypedef int (*NpyIter_IterNextFunc)(NpyIter* it) noexcept nogil             # <<<<<<<<<<<<<<
 * ctypedef void (*NpyIter_GetMultiIndexFunc)(NpyIter* it, npy_intp* outcoords) noexcept nogil
 * 
 */
typedef int (*__pyx_t_5numpy_NpyIter_IterNextFunc)(NpyIter *);

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1114
 * # Iterator API added in v1.6
 * ctypedef int (*NpyIter_IterNextFunc)(NpyIter* it) noexcept nogil
 * ctypedef void (*NpyIter_GetMultiIndexFunc)(NpyIter* it, npy_intp* outcoords) noexcept nogil             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "numpy/arrayobject.h":
 */
typedef void (*__pyx_t_5numpy_NpyIter_GetMultiIndexFunc)(NpyIter *, npy_intp *);
struct __pyx_opt_args_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos;

/* "pyart/io/_sigmetfile.pyx":380
 *         return 0
 * 
 *     cdef int _incr_rbuf_pos(self, int incr=1):             # <<<<<<<<<<<<<<
 *         """
 *         Increment the record buffer position, load a new record if needed.
 */
struct __pyx_opt_args_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos {
  int __pyx_n;
  int incr;
};

/* "pyart/io/_sigmetfile.pyx":17
 * 
 * 
 * cdef class SigmetFile:             # <<<<<<<<<<<<<<
 *     """
 *     A class for accessing data from Sigmet (IRIS) product files.
 */
struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_vtab;
  PyObject *debug;
  PyObject *product_hdr;
  PyObject *ingest_header;
  PyObject *ingest_data_headers;
  PyObject *data_types;
  PyObject *data_type_names;
  PyObject *ndata_types;
  PyObject *_fh;
  PyObject *_raw_product_bhdrs;
  PyArrayObject *_rbuf;
  __pyx_t_5numpy_int16_t *_rbuf_p;
  int _rbuf_pos;
  int _record_number;
};



struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile {
  int (*_get_ray)(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *, int, PyArrayObject *);
  int (*_incr_rbuf_pos)(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *, struct __pyx_opt_args_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos *__pyx_optional_args);
  int (*_load_record)(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *);
};
static struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_vtabptr_5pyart_2io_11_sigmetfile_SigmetFile;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg
    #define __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg  /* no-op, __Pyx_Arg_FASTCALL is direct and this needs
                                                   to have the same reference counting */
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_TrueDivideCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_TrueDivideCObj(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceTrueDivide(op1, op2) : PyNumber_TrueDivide(op1, op2))
#endif

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_TrueDivideObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_TrueDivideObjC(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceTrueDivide(op1, op2) : PyNumber_TrueDivide(op1, op2))
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_TrueDivideObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_TrueDivideObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceTrueDivide(op1, op2) : PyNumber_TrueDivide(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddCObj(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_MultiplyCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_MultiplyCObj(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceMultiply(op1, op2) : PyNumber_Multiply(op1, op2))
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE int __Pyx_PyInt_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/* BufferIndexError.proto */
static void __Pyx_RaiseBufferIndexError(int axis);

#define __Pyx_BufPtrStrided1d(type, buf, i0, s0) (type)((char*)buf + i0 * s0)
/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AndObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AndObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAnd(op1, op2) : PyNumber_And(op1, op2))
#endif

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_SubtractObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_SubtractObjC(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* SliceObject.proto */
#define __Pyx_PyObject_DelSlice(obj, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)\
    __Pyx_PyObject_SetSlice(obj, (PyObject*)NULL, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(
        PyObject* obj, PyObject* value, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_SubtractCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_SubtractCObj(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

#define __Pyx_BufPtrStrided2d(type, buf, i0, s0, i1, s1) (type)((char*)buf + i0 * s0 + i1 * s1)
/* PyIntCompare.proto */
static CYTHON_INLINE int __Pyx_PyInt_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_12
#define __PYX_HAVE_RT_ImportType_proto_3_0_12
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_12(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_12(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_12 {
   __Pyx_ImportType_CheckSize_Error_3_0_12 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_12 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_12 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_12(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_12 check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *typesModule=NULL, *methodType=NULL, *result=NULL;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    typesModule = PyImport_ImportModule("types");
    if (!typesModule) return NULL;
    methodType = PyObject_GetAttrString(typesModule, "MethodType");
    Py_DECREF(typesModule);
    if (!methodType) return NULL;
    result = PyObject_CallFunctionObjArgs(methodType, func, self, NULL);
    Py_DECREF(methodType);
    return result;
}
#elif PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc);
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_long__double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_long__double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_long__double(a, b) ((a)-(b))
    #define __Pyx_c_prod_long__double(a, b) ((a)*(b))
    #define __Pyx_c_quot_long__double(a, b) ((a)/(b))
    #define __Pyx_c_neg_long__double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_long__double(z) ((z)==(long double)0)
    #define __Pyx_c_conj_long__double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_long__double(z)     (::std::abs(z))
        #define __Pyx_c_pow_long__double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_long__double(z) ((z)==0)
    #define __Pyx_c_conj_long__double(z)    (conjl(z))
    #if 1
        #define __Pyx_c_abs_long__double(z)     (cabsl(z))
        #define __Pyx_c_pow_long__double(a, b)  (cpowl(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_sum_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_diff_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_prod_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_neg_long__double(__pyx_t_long_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_long__double(__pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_conj_long__double(__pyx_t_long_double_complex);
    #if 1
        static CYTHON_INLINE long double __Pyx_c_abs_long__double(__pyx_t_long_double_complex);
        static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_pow_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    #endif
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_short(short value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_8itemsize_itemsize(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_9alignment_alignment(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_6fields_fields(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_5names_names(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyArray_ArrayDescr *__pyx_f_5numpy_5dtype_8subarray_subarray(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_uint64 __pyx_f_5numpy_5dtype_5flags_flags(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_7numiter_numiter(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_4size_size(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_5index_index(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_2nd_nd(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_9broadcast_10dimensions_dimensions(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE void **__pyx_f_5numpy_9broadcast_5iters_iters(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self); /* proto*/
static int __pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__get_ray(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, int __pyx_v_nbins, PyArrayObject *__pyx_v_out); /* proto*/
static int __pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, struct __pyx_opt_args_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos *__pyx_optional_args); /* proto*/
static int __pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__load_record(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto*/

/* Module declarations from "cython" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "numpy" */

/* Module declarations from "numpy" */

/* Module declarations from "pyart.io._sigmetfile" */
static PyObject *__pyx_f_5pyart_2io_11_sigmetfile__mask_gates_not_collected(PyArrayObject *, PyArrayObject *); /*proto*/
/* #### Code section: typeinfo ### */
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int16_t = { "int16_t", NULL, sizeof(__pyx_t_5numpy_int16_t), { 0 }, 0, __PYX_IS_UNSIGNED(__pyx_t_5numpy_int16_t) ? 'U' : 'I', __PYX_IS_UNSIGNED(__pyx_t_5numpy_int16_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t = { "uint8_t", NULL, sizeof(__pyx_t_5numpy_uint8_t), { 0 }, 0, __PYX_IS_UNSIGNED(__pyx_t_5numpy_uint8_t) ? 'U' : 'I', __PYX_IS_UNSIGNED(__pyx_t_5numpy_uint8_t), 0 };
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "pyart.io._sigmetfile"
extern int __pyx_module_is_main_pyart__io___sigmetfile;
int __pyx_module_is_main_pyart__io___sigmetfile = 0;

/* Implementation of "pyart.io._sigmetfile" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_open;
static PyObject *__pyx_builtin_xrange;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_sum;
static PyObject *__pyx_builtin_print;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_ImportError;
/* #### Code section: string_decls ### */
static const char __pyx_k_2[] = "2";
static const char __pyx_k_B[] = "B";
static const char __pyx_k_H[] = "H";
static const char __pyx_k_I[] = "I";
static const char __pyx_k_b[] = "b";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_d[] = "d";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_h[] = "h";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_j[] = "j";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_l[] = "l";
static const char __pyx_k_1s[] = "1s";
static const char __pyx_k_2s[] = "2s";
static const char __pyx_k_4s[] = "4s";
static const char __pyx_k_6s[] = "6s";
static const char __pyx_k_8s[] = "8s";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_i4[] = "i4";
static const char __pyx_k_ma[] = "ma";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_rb[] = "rb";
static const char __pyx_k_12s[] = "12s";
static const char __pyx_k_16s[] = "16s";
static const char __pyx_k_18s[] = "18s";
static const char __pyx_k_24s[] = "24s";
static const char __pyx_k_28s[] = "28s";
static const char __pyx_k_32s[] = "32s";
static const char __pyx_k_36s[] = "36s";
static const char __pyx_k_40s[] = "40s";
static const char __pyx_k_48s[] = "48s";
static const char __pyx_k_52s[] = "52s";
static const char __pyx_k_80s[] = "80s";
static const char __pyx_k_94s[] = "94s";
static const char __pyx_k_DBT[] = "DBT";
static const char __pyx_k_DBZ[] = "DBZ";
static const char __pyx_k_KDP[] = "KDP";
static const char __pyx_k_RAW[] = "RAW";
static const char __pyx_k_SQI[] = "SQI";
static const char __pyx_k_VEL[] = "VEL";
static const char __pyx_k_ZDR[] = "ZDR";
static const char __pyx_k__15[] = "";
static const char __pyx_k__19[] = "*";
static const char __pyx_k_az0[] = "az0";
static const char __pyx_k_az1[] = "az1";
static const char __pyx_k_bit[] = "bit";
static const char __pyx_k_day[] = "day";
static const char __pyx_k_dic[] = "dic";
static const char __pyx_k_el0[] = "el0";
static const char __pyx_k_el1[] = "el1";
static const char __pyx_k_exp[] = "exp";
static const char __pyx_k_fmt[] = "fmt";
static const char __pyx_k_idh[] = "idh";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_mod[] = "mod";
static const char __pyx_k_out[] = "out";
static const char __pyx_k_prf[] = "prf";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_112s[] = "112s";
static const char __pyx_k_115s[] = "115s";
static const char __pyx_k_120s[] = "120s";
static const char __pyx_k_128s[] = "128s";
static const char __pyx_k_136s[] = "136s";
static const char __pyx_k_160s[] = "160s";
static const char __pyx_k_184s[] = "184s";
static const char __pyx_k_198s[] = "198s";
static const char __pyx_k_200s[] = "200s";
static const char __pyx_k_204s[] = "204s";
static const char __pyx_k_208s[] = "208s";
static const char __pyx_k_228s[] = "228s";
static const char __pyx_k_256s[] = "256s";
static const char __pyx_k_308s[] = "308s";
static const char __pyx_k_320s[] = "320s";
static const char __pyx_k_480s[] = "480s";
static const char __pyx_k_720s[] = "720s";
static const char __pyx_k_732s[] = "732s";
static const char __pyx_k_920s[] = "920s";
static const char __pyx_k_BIN1[] = "BIN1";
static const char __pyx_k_BIN2[] = "BIN2";
static const char __pyx_k_BIN4[] = "BIN4";
static const char __pyx_k_CSP8[] = "CSP8";
static const char __pyx_k_DBT2[] = "DBT2";
static const char __pyx_k_DBZ2[] = "DBZ2";
static const char __pyx_k_DBZC[] = "DBZC";
static const char __pyx_k_FLT4[] = "FLT4";
static const char __pyx_k_FLT8[] = "FLT8";
static const char __pyx_k_KDP2[] = "KDP2";
static const char __pyx_k_LDRH[] = "LDRH";
static const char __pyx_k_LDRV[] = "LDRV";
static const char __pyx_k_LOG8[] = "LOG8";
static const char __pyx_k_PHIH[] = "PHIH";
static const char __pyx_k_PHIV[] = "PHIV";
static const char __pyx_k_PMI8[] = "PMI8";
static const char __pyx_k_RHOH[] = "RHOH";
static const char __pyx_k_RHOV[] = "RHOV";
static const char __pyx_k_SNR8[] = "SNR8";
static const char __pyx_k_SQI2[] = "SQI2";
static const char __pyx_k_USER[] = "USER";
static const char __pyx_k_VEL2[] = "VEL2";
static const char __pyx_k_VELC[] = "VELC";
static const char __pyx_k_VIL2[] = "VIL2";
static const char __pyx_k_XHDR[] = "XHDR";
static const char __pyx_k_ZDR2[] = "ZDR2";
static const char __pyx_k_ZDRC[] = "ZDRC";
static const char __pyx_k__149[] = "?";
static const char __pyx_k_bin2[] = "bin2";
static const char __pyx_k_bin4[] = "bin4";
static const char __pyx_k_bool[] = "bool";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_flag[] = "flag";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mask[] = "mask";
static const char __pyx_k_mess[] = "mess";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_nbin[] = "nbin";
static const char __pyx_k_ones[] = "ones";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_read[] = "read";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_sqrt[] = "sqrt";
static const char __pyx_k_temp[] = "temp";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_view[] = "view";
static const char __pyx_k_warn[] = "warn";
static const char __pyx_k_year[] = "year";
static const char __pyx_k_2612s[] = "2612s";
static const char __pyx_k_CSP16[] = "CSP16";
static const char __pyx_k_DBTE8[] = "DBTE8";
static const char __pyx_k_DBTV8[] = "DBTV8";
static const char __pyx_k_DBZC2[] = "DBZC2";
static const char __pyx_k_DBZE8[] = "DBZE8";
static const char __pyx_k_DBZV8[] = "DBZV8";
static const char __pyx_k_HDIR2[] = "HDIR2";
static const char __pyx_k_HVEL2[] = "HVEL2";
static const char __pyx_k_LDRH2[] = "LDRH2";
static const char __pyx_k_LDRV2[] = "LDRV2";
static const char __pyx_k_LOG16[] = "LOG16";
static const char __pyx_k_OTHER[] = "OTHER";
static const char __pyx_k_PHIDP[] = "PHIDP";
static const char __pyx_k_PHIH2[] = "PHIH2";
static const char __pyx_k_PHIV2[] = "PHIV2";
static const char __pyx_k_PMI16[] = "PMI16";
static const char __pyx_k_RHOH2[] = "RHOH2";
static const char __pyx_k_RHOHV[] = "RHOHV";
static const char __pyx_k_RHOV2[] = "RHOV2";
static const char __pyx_k_SHEAR[] = "SHEAR";
static const char __pyx_k_SINT1[] = "SINT1";
static const char __pyx_k_SINT2[] = "SINT2";
static const char __pyx_k_SINT4[] = "SINT4";
static const char __pyx_k_SNR16[] = "SNR16";
static const char __pyx_k_TIME2[] = "TIME2";
static const char __pyx_k_UINT1[] = "UINT1";
static const char __pyx_k_UINT2[] = "UINT2";
static const char __pyx_k_UINT4[] = "UINT4";
static const char __pyx_k_USER2[] = "USER2";
static const char __pyx_k_VELC2[] = "VELC2";
static const char __pyx_k_VIR16[] = "VIR16";
static const char __pyx_k_VVEL2[] = "VVEL2";
static const char __pyx_k_WIDTH[] = "WIDTH";
static const char __pyx_k_ZDRC2[] = "ZDRC2";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_debug[] = "debug";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_gparm[] = "gparm";
static const char __pyx_k_int16[] = "int16";
static const char __pyx_k_int32[] = "int32";
static const char __pyx_k_istep[] = "istep";
static const char __pyx_k_month[] = "month";
static const char __pyx_k_nbins[] = "nbins";
static const char __pyx_k_ndata[] = "ndata";
static const char __pyx_k_nrays[] = "nrays";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_power[] = "power";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_ray_i[] = "ray_i";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_uint8[] = "uint8";
static const char __pyx_k_word0[] = "word0";
static const char __pyx_k_word1[] = "word1";
static const char __pyx_k_word2[] = "word2";
static const char __pyx_k_word3[] = "word3";
static const char __pyx_k_AXDIL2[] = "AXDIL2";
static const char __pyx_k_DBTE16[] = "DBTE16";
static const char __pyx_k_DBTV16[] = "DBTV16";
static const char __pyx_k_DBZE16[] = "DBZE16";
static const char __pyx_k_DBZV16[] = "DBZV16";
static const char __pyx_k_HCLASS[] = "HCLASS";
static const char __pyx_k_HEIGHT[] = "HEIGHT";
static const char __pyx_k_PHIDP2[] = "PHIDP2";
static const char __pyx_k_RHOHV2[] = "RHOHV2";
static const char __pyx_k_TURB16[] = "TURB16";
static const char __pyx_k_VILD16[] = "VILD16";
static const char __pyx_k_WIDTH2[] = "WIDTH2";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_flags2[] = "flags2";
static const char __pyx_k_iflags[] = "iflags";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_istart[] = "istart";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_number[] = "number";
static const char __pyx_k_offset[] = "offset";
static const char __pyx_k_record[] = "record";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_shrink[] = "shrink";
static const char __pyx_k_string[] = "string";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_uint16[] = "uint16";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_x_size[] = "x_size";
static const char __pyx_k_xrange[] = "xrange";
static const char __pyx_k_y_size[] = "y_size";
static const char __pyx_k_z_size[] = "z_size";
static const char __pyx_k_2_uint8[] = "(2,) uint8";
static const char __pyx_k_ALBEDO8[] = "ALBEDO8";
static const char __pyx_k_DEFORM2[] = "DEFORM2";
static const char __pyx_k_HCLASS2[] = "HCLASS2";
static const char __pyx_k_MESSAGE[] = "MESSAGE";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_float32[] = "float32";
static const char __pyx_k_icolcnt[] = "icolcnt";
static const char __pyx_k_nsweeps[] = "nsweeps";
static const char __pyx_k_nyquist[] = "nyquist";
static const char __pyx_k_reshape[] = "reshape";
static const char __pyx_k_seconds[] = "seconds";
static const char __pyx_k_spare_0[] = "spare_0";
static const char __pyx_k_spare_1[] = "spare_1";
static const char __pyx_k_spare_2[] = "spare_2";
static const char __pyx_k_spare_3[] = "spare_3";
static const char __pyx_k_spare_4[] = "spare_4";
static const char __pyx_k_tz_name[] = "tz_name";
static const char __pyx_k_x_scale[] = "x_scale";
static const char __pyx_k_y_scale[] = "y_scale";
static const char __pyx_k_z_scale[] = "z_scale";
static const char __pyx_k_ALBEDO16[] = "ALBEDO16";
static const char __pyx_k_DIVERGE2[] = "DIVERGE2";
static const char __pyx_k_FLIQUID2[] = "FLIQUID2";
static const char __pyx_k_UINT16_T[] = "UINT16_T";
static const char __pyx_k_comments[] = "comments";
static const char __pyx_k_datetime[] = "datetime";
static const char __pyx_k_dsp_type[] = "dsp_type";
static const char __pyx_k_filename[] = "filename";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_latitude[] = "latitude";
static const char __pyx_k_ldr_bias[] = "ldr_bias";
static const char __pyx_k_like_dbt[] = "like_dbt";
static const char __pyx_k_like_sqi[] = "like_sqi";
static const char __pyx_k_metadata[] = "metadata";
static const char __pyx_k_prf_flag[] = "prf_flag";
static const char __pyx_k_raw_data[] = "raw_data";
static const char __pyx_k_ray_time[] = "ray_time";
static const char __pyx_k_reserved[] = "reserved";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_zdr_bias[] = "zdr_bias";
static const char __pyx_k_RAINRATE2[] = "RAINRATE2";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_UNKNOWN_6[] = "UNKNOWN_6";
static const char __pyx_k_YMDS_TIME[] = "YMDS_TIME";
static const char __pyx_k_azimuth_0[] = "azimuth_0";
static const char __pyx_k_azimuth_1[] = "azimuth_1";
static const char __pyx_k_data_type[] = "data_type";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_full_xhdr[] = "full_xhdr";
static const char __pyx_k_get_sweep[] = "_get_sweep";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_like_dbt2[] = "like_dbt2";
static const char __pyx_k_like_sqi2[] = "like_sqi2";
static const char __pyx_k_longitude[] = "longitude";
static const char __pyx_k_prt_value[] = "prt_value";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_ray_nbins[] = "ray_nbins";
static const char __pyx_k_read_data[] = "read_data";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_site_name[] = "site_name";
static const char __pyx_k_skip_time[] = "skip_time";
static const char __pyx_k_stop_time[] = "stop_time";
static const char __pyx_k_structure[] = "structure";
static const char __pyx_k_task_name[] = "task_name";
static const char __pyx_k_times_run[] = "times_run";
static const char __pyx_k_SigmetFile[] = "SigmetFile";
static const char __pyx_k_UNKNOWN_29[] = "UNKNOWN_29";
static const char __pyx_k_UNKNOWN_30[] = "UNKNOWN_30";
static const char __pyx_k_UNKNOWN_31[] = "UNKNOWN_31";
static const char __pyx_k_UNKNOWN_81[] = "UNKNOWN_81";
static const char __pyx_k_UNKNOWN_82[] = "UNKNOWN_82";
static const char __pyx_k_UNKNOWN_83[] = "UNKNOWN_83";
static const char __pyx_k_UNKNOWN_84[] = "UNKNOWN_84";
static const char __pyx_k_UNKNOWN_85[] = "UNKNOWN_85";
static const char __pyx_k_UNKNOWN_86[] = "UNKNOWN_86";
static const char __pyx_k_UNKNOWN_87[] = "UNKNOWN_87";
static const char __pyx_k_UNKNOWN_88[] = "UNKNOWN_88";
static const char __pyx_k_UNKNOWN_89[] = "UNKNOWN_89";
static const char __pyx_k_UNKNOWN_90[] = "UNKNOWN_90";
static const char __pyx_k_UNKNOWN_91[] = "UNKNOWN_91";
static const char __pyx_k_UNKNOWN_92[] = "UNKNOWN_92";
static const char __pyx_k_UNKNOWN_93[] = "UNKNOWN_93";
static const char __pyx_k_UNKNOWN_94[] = "UNKNOWN_94";
static const char __pyx_k_UNKNOWN_95[] = "UNKNOWN_95";
static const char __pyx_k_UNKNOWN_96[] = "UNKNOWN_96";
static const char __pyx_k_UNKNOWN_97[] = "UNKNOWN_97";
static const char __pyx_k_UNKNOWN_98[] = "UNKNOWN_98";
static const char __pyx_k_UNKNOWN_99[] = "UNKNOWN_99";
static const char __pyx_k_cluttermap[] = "cluttermap";
static const char __pyx_k_data_types[] = "data_types";
static const char __pyx_k_empty_like[] = "empty_like";
static const char __pyx_k_fill_value[] = "fill_value";
static const char __pyx_k_frombuffer[] = "frombuffer";
static const char __pyx_k_header_dic[] = "header_dic";
static const char __pyx_k_input_mask[] = "input_mask";
static const char __pyx_k_is_bit_set[] = "_is_bit_set";
static const char __pyx_k_ldr_offset[] = "ldr_offset";
static const char __pyx_k_low_prf_hz[] = "low_prf_hz";
static const char __pyx_k_major_mode[] = "major_mode";
static const char __pyx_k_multiplier[] = "multiplier";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_ray_number[] = "ray_number";
static const char __pyx_k_start_time[] = "start_time";
static const char __pyx_k_sweep_data[] = "sweep_data";
static const char __pyx_k_task_state[] = "task_state";
static const char __pyx_k_total_size[] = "total_size";
static const char __pyx_k_unpack_key[] = "_unpack_key";
static const char __pyx_k_wavelength[] = "wavelength";
static const char __pyx_k_x_location[] = "x_location";
static const char __pyx_k_x_smoother[] = "x_smoother";
static const char __pyx_k_y_location[] = "y_location";
static const char __pyx_k_y_smoother[] = "y_smoother";
static const char __pyx_k_z_location[] = "z_location";
static const char __pyx_k_zdr_offset[] = "zdr_offset";
static const char __pyx_k_zeros_like[] = "zeros_like";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_PRODUCT_END[] = "PRODUCT_END";
static const char __pyx_k_PRODUCT_HDR[] = "PRODUCT_HDR";
static const char __pyx_k_RECORD_SIZE[] = "RECORD_SIZE";
static const char __pyx_k_UNKNOWN_100[] = "UNKNOWN_100";
static const char __pyx_k_UNKNOWN_101[] = "UNKNOWN_101";
static const char __pyx_k_UNKNOWN_102[] = "UNKNOWN_102";
static const char __pyx_k_UNKNOWN_103[] = "UNKNOWN_103";
static const char __pyx_k_UNKNOWN_104[] = "UNKNOWN_104";
static const char __pyx_k_UNKNOWN_105[] = "UNKNOWN_105";
static const char __pyx_k_UNKNOWN_106[] = "UNKNOWN_106";
static const char __pyx_k_UNKNOWN_107[] = "UNKNOWN_107";
static const char __pyx_k_UNKNOWN_108[] = "UNKNOWN_108";
static const char __pyx_k_UNKNOWN_109[] = "UNKNOWN_109";
static const char __pyx_k_UNKNOWN_110[] = "UNKNOWN_110";
static const char __pyx_k_UNKNOWN_112[] = "UNKNOWN_112";
static const char __pyx_k_UNKNOWN_113[] = "UNKNOWN_113";
static const char __pyx_k_UNKNOWN_114[] = "UNKNOWN_114";
static const char __pyx_k_UNKNOWN_115[] = "UNKNOWN_115";
static const char __pyx_k_UNKNOWN_116[] = "UNKNOWN_116";
static const char __pyx_k_UNKNOWN_117[] = "UNKNOWN_117";
static const char __pyx_k_UNKNOWN_118[] = "UNKNOWN_118";
static const char __pyx_k_UNKNOWN_119[] = "UNKNOWN_119";
static const char __pyx_k_UNKNOWN_120[] = "UNKNOWN_120";
static const char __pyx_k_UNKNOWN_121[] = "UNKNOWN_121";
static const char __pyx_k_UNKNOWN_122[] = "UNKNOWN_122";
static const char __pyx_k_UNKNOWN_123[] = "UNKNOWN_123";
static const char __pyx_k_UNKNOWN_124[] = "UNKNOWN_124";
static const char __pyx_k_UNKNOWN_125[] = "UNKNOWN_125";
static const char __pyx_k_UNKNOWN_126[] = "UNKNOWN_126";
static const char __pyx_k_UNKNOWN_127[] = "UNKNOWN_127";
static const char __pyx_k_attenuation[] = "attenuation";
static const char __pyx_k_bit_per_bin[] = "bit_per_bin";
static const char __pyx_k_elevation_0[] = "elevation_0";
static const char __pyx_k_elevation_1[] = "elevation_1";
static const char __pyx_k_fixed_angle[] = "fixed_angle";
static const char __pyx_k_height_site[] = "height_site";
static const char __pyx_k_ingest_time[] = "ingest_time";
static const char __pyx_k_lead_record[] = "lead_record";
static const char __pyx_k_mask_word_0[] = "mask_word_0";
static const char __pyx_k_mask_word_1[] = "mask_word_1";
static const char __pyx_k_mask_word_2[] = "mask_word_2";
static const char __pyx_k_mask_word_3[] = "mask_word_3";
static const char __pyx_k_mask_word_4[] = "mask_word_4";
static const char __pyx_k_ndata_types[] = "ndata_types";
static const char __pyx_k_number_bins[] = "number_bins";
static const char __pyx_k_product_end[] = "product_end";
static const char __pyx_k_product_hdr[] = "product_hdr";
static const char __pyx_k_pulse_width[] = "pulse_width";
static const char __pyx_k_ray_headers[] = "ray_headers";
static const char __pyx_k_sample_size[] = "sample_size";
static const char __pyx_k_task_config[] = "task_config";
static const char __pyx_k_velocity_up[] = "velocity_up";
static const char __pyx_k_zr_constant[] = "zr_constant";
static const char __pyx_k_zr_exponent[] = "zr_exponent";
static const char __pyx_k_azimuth_list[] = "azimuth_list";
static const char __pyx_k_earth_radius[] = "earth_radius";
static const char __pyx_k_fault_status[] = "fault_status";
static const char __pyx_k_gparam_bytes[] = "gparam_bytes";
static const char __pyx_k_height_radar[] = "height_radar";
static const char __pyx_k_ilevel_seams[] = "ilevel_seams";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_iris_version[] = "iris_version";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_last_run_day[] = "last_run_day";
static const char __pyx_k_masked_array[] = "masked_array";
static const char __pyx_k_milliseconds[] = "milliseconds";
static const char __pyx_k_number_files[] = "number_files";
static const char __pyx_k_number_tasks[] = "number_tasks";
static const char __pyx_k_polarization[] = "polarization";
static const char __pyx_k_product_name[] = "product_name";
static const char __pyx_k_radar_height[] = "radar_height";
static const char __pyx_k_samples_used[] = "samples_used";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_sweep_number[] = "sweep_number";
static const char __pyx_k_trigger_rate[] = "trigger_rate";
static const char __pyx_k_DSP_DATA_MASK[] = "DSP_DATA_MASK";
static const char __pyx_k_INGEST_HEADER[] = "INGEST_HEADER";
static const char __pyx_k_RAW_PROD_BHDR[] = "RAW_PROD_BHDR";
static const char __pyx_k_TASK_DSP_INFO[] = "TASK_DSP_INFO";
static const char __pyx_k_TASK_END_INFO[] = "TASK_END_INFO";
static const char __pyx_k_TEMPERATURE16[] = "TEMPERATURE16";
static const char __pyx_k_bin2_to_angle[] = "bin2_to_angle";
static const char __pyx_k_bin4_to_angle[] = "bin4_to_angle";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_comment_bytes[] = "comment_bytes";
static const char __pyx_k_first_azimuth[] = "first_azimuth";
static const char __pyx_k_ground_height[] = "ground_height";
static const char __pyx_k_hardware_site[] = "hardware_site";
static const char __pyx_k_ingest_header[] = "ingest_header";
static const char __pyx_k_last_run_time[] = "last_run_time";
static const char __pyx_k_maximum_range[] = "maximum_range";
static const char __pyx_k_melting_layer[] = "melting_layer";
static const char __pyx_k_melting_level[] = "melting_level";
static const char __pyx_k_number_ingest[] = "number_ingest";
static const char __pyx_k_number_sweeps[] = "number_sweeps";
static const char __pyx_k_playback_flag[] = "playback_flag";
static const char __pyx_k_raw_prod_bhdr[] = "raw_prod_bhdr";
static const char __pyx_k_record_number[] = "record_number";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_sqi_threshold[] = "sqi_threshold";
static const char __pyx_k_task_dsp_info[] = "task_dsp_info";
static const char __pyx_k_task_dsp_mode[] = "task_dsp_mode";
static const char __pyx_k_task_end_info[] = "task_end_info";
static const char __pyx_k_velocity_east[] = "velocity_east";
static const char __pyx_k_wavelength_cm[] = "wavelength_cm";
static const char __pyx_k_TASK_MISC_INFO[] = "TASK_MISC_INFO";
static const char __pyx_k_TASK_SCAN_INFO[] = "TASK_SCAN_INFO";
static const char __pyx_k_altitude_radar[] = "altitude_radar";
static const char __pyx_k_clutter_filter[] = "clutter_filter";
static const char __pyx_k_data_type_name[] = "data_type_name";
static const char __pyx_k_dual_prf_delay[] = "dual_prf_delay";
static const char __pyx_k_elevation_list[] = "elevation_list";
static const char __pyx_k_format_version[] = "format_version";
static const char __pyx_k_iset_and_scale[] = "iset_and_scale";
static const char __pyx_k_last_bin_range[] = "last_bin_range";
static const char __pyx_k_latitude_radar[] = "latitude_radar";
static const char __pyx_k_local_timezone[] = "local_timezone";
static const char __pyx_k_multi_prf_flag[] = "multi_prf_flag";
static const char __pyx_k_product_config[] = "product_config";
static const char __pyx_k_raw_sweep_data[] = "raw_sweep_data";
static const char __pyx_k_sweep_metadata[] = "sweep_metadata";
static const char __pyx_k_task_data_time[] = "task_data_time";
static const char __pyx_k_task_misc_info[] = "task_misc_info";
static const char __pyx_k_task_scan_info[] = "task_scan_info";
static const char __pyx_k_transmit_power[] = "transmit_power";
static const char __pyx_k_velocity_north[] = "velocity_north";
static const char __pyx_k_COLOR_SCALE_DEF[] = "COLOR_SCALE_DEF";
static const char __pyx_k_TASK_CALIB_INFO[] = "TASK_CALIB_INFO";
static const char __pyx_k_TASK_RANGE_INFO[] = "TASK_RANGE_INFO";
static const char __pyx_k_TASK_SCHED_INFO[] = "TASK_SCHED_INFO";
static const char __pyx_k_cluttermap_flag[] = "cluttermap_flag";
static const char __pyx_k_color_scale_def[] = "color_scale_def";
static const char __pyx_k_first_bin_range[] = "first_bin_range";
static const char __pyx_k_first_elevation[] = "first_elevation";
static const char __pyx_k_first_ray_index[] = "first_ray_index";
static const char __pyx_k_gas_attenuation[] = "gas_attenuation";
static const char __pyx_k_generation_time[] = "generation_time";
static const char __pyx_k_input_data_type[] = "input_data_type";
static const char __pyx_k_longitude_radar[] = "longitude_radar";
static const char __pyx_k_mean_wind_speed[] = "mean_wind_speed";
static const char __pyx_k_nray_data_types[] = "nray_data_types";
static const char __pyx_k_number_elements[] = "number_elements";
static const char __pyx_k_power_threshold[] = "power_threshold";
static const char __pyx_k_projection_name[] = "projection_name";
static const char __pyx_k_projection_type[] = "projection_type";
static const char __pyx_k_radial_smoother[] = "radial_smoother";
static const char __pyx_k_ray_header_mask[] = "ray_header_mask";
static const char __pyx_k_resolution_rays[] = "resolution_rays";
static const char __pyx_k_scheduling_code[] = "scheduling_code";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_step_input_bins[] = "step_input_bins";
static const char __pyx_k_task_calib_info[] = "task_calib_info";
static const char __pyx_k_task_range_info[] = "task_range_info";
static const char __pyx_k_task_sched_info[] = "task_sched_info";
static const char __pyx_k_tcf_cal_flags_1[] = "tcf_cal_flags_1";
static const char __pyx_k_tcf_cal_flags_2[] = "tcf_cal_flags_2";
static const char __pyx_k_STRUCTURE_HEADER[] = "STRUCTURE_HEADER";
static const char __pyx_k_SigmetFile_close[] = "SigmetFile.close";
static const char __pyx_k_customer_storage[] = "customer_storage";
static const char __pyx_k_file_ingest_time[] = "file_ingest_time";
static const char __pyx_k_first_ray_offset[] = "first_ray_offset";
static const char __pyx_k_ingest_data_hdrs[] = "ingest_data_hdrs";
static const char __pyx_k_ingest_site_name[] = "ingest_site_name_";
static const char __pyx_k_inverse_flatting[] = "inverse_flatting";
static const char __pyx_k_playback_version[] = "playback_version";
static const char __pyx_k_ray_header_bytes[] = "ray_header_bytes";
static const char __pyx_k_step_output_bins[] = "step_output_bins";
static const char __pyx_k_structure_header[] = "structure_header";
static const char __pyx_k_sweep_start_time[] = "sweep_start_time";
static const char __pyx_k_task_description[] = "task_description";
static const char __pyx_k_tr_serial_number[] = "tr_serial_number";
static const char __pyx_k_trucation_height[] = "trucation_height";
static const char __pyx_k_unpack_structure[] = "_unpack_structure";
static const char __pyx_k_SIGMET_DATA_TYPES[] = "SIGMET_DATA_TYPES";
static const char __pyx_k_agc_feedback_code[] = "agc_feedback_code";
static const char __pyx_k_antenna_offset_up[] = "antenna_offset_up";
static const char __pyx_k_antenna_scan_mode[] = "antenna_scan_mode";
static const char __pyx_k_gain_control_flag[] = "gain_control_flag";
static const char __pyx_k_low_prf_factional[] = "low_prf_factional";
static const char __pyx_k_minor_task_suffix[] = "minor_task_suffix";
static const char __pyx_k_number_input_bins[] = "number_input_bins";
static const char __pyx_k_number_log_filter[] = "number_log_filter";
static const char __pyx_k_number_rays_sweep[] = "number_rays_sweep";
static const char __pyx_k_parse_ray_headers[] = "_parse_ray_headers";
static const char __pyx_k_polarization_type[] = "polarization_type";
static const char __pyx_k_product_type_code[] = "product_type_code";
static const char __pyx_k_self__rbuf_pos_is[] = "self._rbuf_pos is";
static const char __pyx_k_sweep_ingest_time[] = "sweep_ingest_time";
static const char __pyx_k_task_major_number[] = "task_major_number";
static const char __pyx_k_task_minor_number[] = "task_minor_number";
static const char __pyx_k_truncation_height[] = "truncation_height";
static const char __pyx_k_INGEST_DATA_HEADER[] = "INGEST_DATA_HEADER";
static const char __pyx_k_Reading_ray_i_of_i[] = "Reading ray: %i of %i";
static const char __pyx_k_TASK_CONFIGURATION[] = "TASK_CONFIGURATION";
static const char __pyx_k_TASK_PPI_SCAN_INFO[] = "TASK_PPI_SCAN_INFO";
static const char __pyx_k_TASK_RHI_SCAN_INFO[] = "TASK_RHI_SCAN_INFO";
static const char __pyx_k_antenna_offset_bow[] = "antenna_offset_bow";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_bytes_in_structure[] = "bytes_in_structure";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_configuration_name[] = "configuration_name";
static const char __pyx_k_left_azimuth_limit[] = "left_azimuth_limit";
static const char __pyx_k_number_output_bins[] = "number_output_bins";
static const char __pyx_k_reciever_bandwidth[] = "reciever_bandwidth";
static const char __pyx_k_reflectivity_slope[] = "reflectivity_slope";
static const char __pyx_k_task_configuration[] = "task_configuration";
static const char __pyx_k_time_used_last_run[] = "time_used_last_run";
static const char __pyx_k_unpack_product_hdr[] = "_unpack_product_hdr";
static const char __pyx_k_vertical_beamwidth[] = "vertical_beamwidth";
static const char __pyx_k_TASK_DSP_MODE_BATCH[] = "TASK_DSP_MODE_BATCH";
static const char __pyx_k_TASK_FILE_SCAN_INFO[] = "TASK_FILE_SCAN_INFO";
static const char __pyx_k_clutter_filter_name[] = "clutter_filter_name";
static const char __pyx_k_convert_sigmet_data[] = "convert_sigmet_data";
static const char __pyx_k_ingest_data_headers[] = "ingest_data_headers";
static const char __pyx_k_ingest_iris_version[] = "ingest_iris_version";
static const char __pyx_k_latitude_projection[] = "latitude_projection";
static const char __pyx_k_low_prf_sample_size[] = "low_prf_sample_size";
static const char __pyx_k_mean_wind_direction[] = "mean_wind_direction";
static const char __pyx_k_nexrad_clutter_skip[] = "nexrad_clutter_skip";
static const char __pyx_k_right_azimuth_limit[] = "right_azimuth_limit";
static const char __pyx_k_standard_parallel_1[] = "standard_parallel_1";
static const char __pyx_k_standard_parallel_2[] = "standard_parallel_2";
static const char __pyx_k_zdr_threshold_flags[] = "zdr_threshold_flags";
static const char __pyx_k_INGEST_CONFIGURATION[] = "INGEST_CONFIGURATION";
static const char __pyx_k_SigmetFile_read_data[] = "SigmetFile.read_data";
static const char __pyx_k_data_types_from_mask[] = "_data_types_from_mask";
static const char __pyx_k_determine_data_types[] = "_determine_data_types";
static const char __pyx_k_extended_header_type[] = "extended_header_type";
static const char __pyx_k_horizontal_beamwidth[] = "horizontal_beamwidth";
static const char __pyx_k_ingest_configuration[] = "ingest_configuration";
static const char __pyx_k_ingest_hardware_name[] = "ingest_hardware_name_";
static const char __pyx_k_iris_version_created[] = "iris_version_created";
static const char __pyx_k_log_filter_first_bin[] = "log_filter_first_bin";
static const char __pyx_k_longitude_projection[] = "longitude_projection";
static const char __pyx_k_number_linear_filter[] = "number_linear_filter";
static const char __pyx_k_pyart_io__sigmetfile[] = "pyart.io._sigmetfile";
static const char __pyx_k_seconds_between_runs[] = "seconds_between_runs";
static const char __pyx_k_structure_identifier[] = "structure_identifier";
static const char __pyx_k_unpack_ingest_header[] = "_unpack_ingest_header";
static const char __pyx_k_unpack_raw_prod_bhdr[] = "_unpack_raw_prod_bhdr";
static const char __pyx_k_PRODUCT_CONFIGURATION[] = "PRODUCT_CONFIGURATION";
static const char __pyx_k_SigmetFile__get_sweep[] = "SigmetFile._get_sweep";
static const char __pyx_k_TASK_MANUAL_SCAN_INFO[] = "TASK_MANUAL_SCAN_INFO";
static const char __pyx_k_lower_elevation_limit[] = "lower_elevation_limit";
static const char __pyx_k_product_configuration[] = "product_configuration";
static const char __pyx_k_signal_processor_type[] = "signal_processor_type";
static const char __pyx_k_upper_elevation_limit[] = "upper_elevation_limit";
static const char __pyx_k_width_threshold_flags[] = "width_threshold_flags";
static const char __pyx_k_current_data_type_mask[] = "current_data_type_mask";
static const char __pyx_k_custom_ray_header_name[] = "custom_ray_header_name";
static const char __pyx_k_product_specific_bytes[] = "product_specific_bytes";
static const char __pyx_k_vertical_current_noise[] = "vertical_current_noise";
static const char __pyx_k_volume_scan_start_time[] = "volume_scan_start_time";
static const char __pyx_k_Finished_loading_record[] = "Finished loading record:";
static const char __pyx_k_GMT_minute_offset_local[] = "GMT_minute_offset_local";
static const char __pyx_k_linear_filter_first_bin[] = "linear_filter_first_bin";
static const char __pyx_k_low_prf_range_averaging[] = "low_prf_range_averaging";
static const char __pyx_k_number_rays_file_actual[] = "number_rays_file_actual";
static const char __pyx_k_number_sweeps_completed[] = "number_sweeps_completed";
static const char __pyx_k_original_data_type_mask[] = "original_data_type_mask";
static const char __pyx_k_product_sequence_number[] = "product_sequence_number";
static const char __pyx_k_start_first_sector_flag[] = "start_first_sector_flag";
static const char __pyx_k_variable_range_bin_flag[] = "variable_range_bin_flag";
static const char __pyx_k_vertical_i0_calibration[] = "vertical_i0_calibration";
static const char __pyx_k_vertical_radar_constant[] = "vertical_radar_constant";
static const char __pyx_k_antenna_offset_starboard[] = "antenna_offset_starboard";
static const char __pyx_k_gmt_offset_minutes_local[] = "gmt_offset_minutes_local";
static const char __pyx_k_horizontal_current_noise[] = "horizontal_current_noise";
static const char __pyx_k_nexrad_clutter_threshold[] = "nexrad_clutter_threshold";
static const char __pyx_k_number_task_config_table[] = "number_task_config_table";
static const char __pyx_k_pyart_io__sigmetfile_pyx[] = "pyart/io/_sigmetfile.pyx";
static const char __pyx_k_range_bin_averaging_flag[] = "range_bin_averaging_flag";
static const char __pyx_k_reflectivity_calibration[] = "reflectivity_calibration";
static const char __pyx_k_start_first_section_flag[] = "start_first_section_flag";
static const char __pyx_k_task_scan_type_scan_info[] = "task_scan_type_scan_info";
static const char __pyx_k_velocity_threshold_flags[] = "velocity_threshold_flags";
static const char __pyx_k_extended_ray_header_bytes[] = "extended_ray_header_bytes";
static const char __pyx_k_horizontal_calibration_i0[] = "horizontal_calibration_i0";
static const char __pyx_k_horizontal_i0_calibration[] = "horizontal_i0_calibration";
static const char __pyx_k_horizontal_radar_constant[] = "horizontal_radar_constant";
static const char __pyx_k_number_rays_file_expected[] = "number_rays_file_expected";
static const char __pyx_k_tranmitter_phase_sequence[] = "tranmitter_phase_sequence";
static const char __pyx_k_unpack_ingest_data_header[] = "_unpack_ingest_data_header";
static const char __pyx_k_width_unfolding_threshold[] = "width_unfolding_threshold";
static const char __pyx_k_GMT_minute_offset_standard[] = "GMT_minute_offset_standard";
static const char __pyx_k_SigmetFile___reduce_cython[] = "SigmetFile.__reduce_cython__";
static const char __pyx_k_angular_resolution_desired[] = "angular_resolution_desired";
static const char __pyx_k_unpack_ingest_data_headers[] = "_unpack_ingest_data_headers";
static const char __pyx_k_vertical_noise_calibration[] = "vertical_noise_calibration";
static const char __pyx_k_gmt_offset_minutes_standard[] = "gmt_offset_minutes_standard";
static const char __pyx_k_SigmetFile___setstate_cython[] = "SigmetFile.__setstate_cython__";
static const char __pyx_k_clutter_correction_threshold[] = "clutter_correction_threshold";
static const char __pyx_k_horizontal_calibration_noise[] = "horizontal_calibration_noise";
static const char __pyx_k_horizontal_noise_calibration[] = "horizontal_noise_calibration";
static const char __pyx_k_radar_height_above_reference[] = "radar_height_above_reference";
static const char __pyx_k_reflectivity_noise_threshold[] = "reflectivity_noise_threshold";
static const char __pyx_k_task_configuration_file_name[] = "task_configuration_file_name";
static const char __pyx_k_velocity_unfolding_threshold[] = "velocity_unfolding_threshold";
static const char __pyx_k_extended_product_header_offset[] = "extended_product_header_offset";
static const char __pyx_k_A_class_and_supporting_function[] = "\nA class and supporting functions for reading Sigmet (raw format) files.\n\n";
static const char __pyx_k_File_truncated_or_corrupt_i_of_i[] = "File truncated or corrupt, %i of %i sweeps read";
static const char __pyx_k_SigmetFile__determine_data_types[] = "SigmetFile._determine_data_types";
static const char __pyx_k_Unknown_type_s_returning_raw_dat[] = "Unknown type: %s, returning raw data";
static const char __pyx_k_corrected_reflectivity_threshold[] = "corrected_reflectivity_threshold_flags";
static const char __pyx_k_numpy__core_multiarray_failed_to[] = "numpy._core.multiarray failed to import";
static const char __pyx_k_numpy__core_umath_failed_to_impo[] = "numpy._core.umath failed to import";
static const char __pyx_k_reflectivity_unfolding_threshold[] = "reflectivity_unfolding_threshold";
static const char __pyx_k_self__rbuf_p_cannot_be_converted[] = "self._rbuf_p cannot be converted to a Python object for pickling";
static const char __pyx_k_uncorrected_reflectivity_thresho[] = "uncorrected_reflectivity_threshold_flags";
/* #### Code section: decls ### */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile___init__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_filename, PyObject *__pyx_v_debug); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_2_determine_data_types(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_4close(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_6read_data(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_full_xhdr); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_8_get_sweep(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_full_xhdr, PyObject *__pyx_v_raw_data); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile__data_types_from_mask(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_word0, PyObject *__pyx_v_word1, PyObject *__pyx_v_word2, PyObject *__pyx_v_word3); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_2_is_bit_set(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_number, PyObject *__pyx_v_bit); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_4_parse_ray_headers(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ray_headers); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_6convert_sigmet_data(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data_type, PyObject *__pyx_v_data, PyObject *__pyx_v_nbins); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_8bin2_to_angle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bin2); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10bin4_to_angle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bin4); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_12_unpack_structure(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_string, PyObject *__pyx_v_structure); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_14_unpack_key(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_dic, PyObject *__pyx_v_key, PyObject *__pyx_v_structure); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_16_unpack_ingest_data_headers(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record, PyObject *__pyx_v_ndata_types); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_18_unpack_ingest_data_header(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record, PyObject *__pyx_v_number); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_20_unpack_raw_prod_bhdr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_22_unpack_product_hdr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_24_unpack_ingest_header(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record); /* proto */
static PyObject *__pyx_tp_new_5pyart_2io_11_sigmetfile_SigmetFile(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5numpy_dtype;
  PyTypeObject *__pyx_ptype_5numpy_flatiter;
  PyTypeObject *__pyx_ptype_5numpy_broadcast;
  PyTypeObject *__pyx_ptype_5numpy_ndarray;
  PyTypeObject *__pyx_ptype_5numpy_generic;
  PyTypeObject *__pyx_ptype_5numpy_number;
  PyTypeObject *__pyx_ptype_5numpy_integer;
  PyTypeObject *__pyx_ptype_5numpy_signedinteger;
  PyTypeObject *__pyx_ptype_5numpy_unsignedinteger;
  PyTypeObject *__pyx_ptype_5numpy_inexact;
  PyTypeObject *__pyx_ptype_5numpy_floating;
  PyTypeObject *__pyx_ptype_5numpy_complexfloating;
  PyTypeObject *__pyx_ptype_5numpy_flexible;
  PyTypeObject *__pyx_ptype_5numpy_character;
  PyTypeObject *__pyx_ptype_5numpy_ufunc;
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type_5pyart_2io_11_sigmetfile_SigmetFile;
  #endif
  PyTypeObject *__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile;
  PyObject *__pyx_kp_u_112s;
  PyObject *__pyx_kp_u_115s;
  PyObject *__pyx_kp_u_120s;
  PyObject *__pyx_kp_u_128s;
  PyObject *__pyx_kp_u_12s;
  PyObject *__pyx_kp_u_136s;
  PyObject *__pyx_kp_u_160s;
  PyObject *__pyx_kp_u_16s;
  PyObject *__pyx_kp_u_184s;
  PyObject *__pyx_kp_u_18s;
  PyObject *__pyx_kp_u_198s;
  PyObject *__pyx_kp_u_1s;
  PyObject *__pyx_kp_u_2;
  PyObject *__pyx_kp_u_200s;
  PyObject *__pyx_kp_u_204s;
  PyObject *__pyx_kp_u_208s;
  PyObject *__pyx_kp_u_228s;
  PyObject *__pyx_kp_u_24s;
  PyObject *__pyx_kp_u_256s;
  PyObject *__pyx_kp_u_2612s;
  PyObject *__pyx_kp_u_28s;
  PyObject *__pyx_kp_u_2_uint8;
  PyObject *__pyx_kp_u_2s;
  PyObject *__pyx_kp_u_308s;
  PyObject *__pyx_kp_u_320s;
  PyObject *__pyx_kp_u_32s;
  PyObject *__pyx_kp_u_36s;
  PyObject *__pyx_kp_u_40s;
  PyObject *__pyx_kp_u_480s;
  PyObject *__pyx_kp_u_48s;
  PyObject *__pyx_kp_u_4s;
  PyObject *__pyx_kp_u_52s;
  PyObject *__pyx_kp_u_6s;
  PyObject *__pyx_kp_u_720s;
  PyObject *__pyx_kp_u_732s;
  PyObject *__pyx_kp_u_80s;
  PyObject *__pyx_kp_u_8s;
  PyObject *__pyx_kp_u_920s;
  PyObject *__pyx_kp_u_94s;
  PyObject *__pyx_n_u_ALBEDO16;
  PyObject *__pyx_n_u_ALBEDO8;
  PyObject *__pyx_n_u_AXDIL2;
  PyObject *__pyx_n_u_B;
  PyObject *__pyx_n_s_BIN1;
  PyObject *__pyx_n_s_BIN2;
  PyObject *__pyx_n_s_BIN4;
  PyObject *__pyx_n_s_COLOR_SCALE_DEF;
  PyObject *__pyx_n_u_CSP16;
  PyObject *__pyx_n_u_CSP8;
  PyObject *__pyx_n_u_DBT;
  PyObject *__pyx_n_u_DBT2;
  PyObject *__pyx_n_u_DBTE16;
  PyObject *__pyx_n_u_DBTE8;
  PyObject *__pyx_n_u_DBTV16;
  PyObject *__pyx_n_u_DBTV8;
  PyObject *__pyx_n_u_DBZ;
  PyObject *__pyx_n_u_DBZ2;
  PyObject *__pyx_n_u_DBZC;
  PyObject *__pyx_n_u_DBZC2;
  PyObject *__pyx_n_u_DBZE16;
  PyObject *__pyx_n_u_DBZE8;
  PyObject *__pyx_n_u_DBZV16;
  PyObject *__pyx_n_u_DBZV8;
  PyObject *__pyx_n_u_DEFORM2;
  PyObject *__pyx_n_u_DIVERGE2;
  PyObject *__pyx_n_s_DSP_DATA_MASK;
  PyObject *__pyx_n_u_FLIQUID2;
  PyObject *__pyx_n_s_FLT4;
  PyObject *__pyx_n_s_FLT8;
  PyObject *__pyx_kp_u_File_truncated_or_corrupt_i_of_i;
  PyObject *__pyx_kp_u_Finished_loading_record;
  PyObject *__pyx_n_u_GMT_minute_offset_local;
  PyObject *__pyx_n_u_GMT_minute_offset_standard;
  PyObject *__pyx_n_u_H;
  PyObject *__pyx_n_u_HCLASS;
  PyObject *__pyx_n_u_HCLASS2;
  PyObject *__pyx_n_u_HDIR2;
  PyObject *__pyx_n_u_HEIGHT;
  PyObject *__pyx_n_u_HVEL2;
  PyObject *__pyx_n_u_I;
  PyObject *__pyx_n_s_INGEST_CONFIGURATION;
  PyObject *__pyx_n_s_INGEST_DATA_HEADER;
  PyObject *__pyx_n_s_INGEST_HEADER;
  PyObject *__pyx_n_s_ImportError;
  PyObject *__pyx_n_u_KDP;
  PyObject *__pyx_n_u_KDP2;
  PyObject *__pyx_n_u_LDRH;
  PyObject *__pyx_n_u_LDRH2;
  PyObject *__pyx_n_u_LDRV;
  PyObject *__pyx_n_u_LDRV2;
  PyObject *__pyx_n_u_LOG16;
  PyObject *__pyx_n_u_LOG8;
  PyObject *__pyx_n_s_MESSAGE;
  PyObject *__pyx_n_u_OTHER;
  PyObject *__pyx_n_u_PHIDP;
  PyObject *__pyx_n_u_PHIDP2;
  PyObject *__pyx_n_u_PHIH;
  PyObject *__pyx_n_u_PHIH2;
  PyObject *__pyx_n_u_PHIV;
  PyObject *__pyx_n_u_PHIV2;
  PyObject *__pyx_n_u_PMI16;
  PyObject *__pyx_n_u_PMI8;
  PyObject *__pyx_n_s_PRODUCT_CONFIGURATION;
  PyObject *__pyx_n_s_PRODUCT_END;
  PyObject *__pyx_n_s_PRODUCT_HDR;
  PyObject *__pyx_n_u_RAINRATE2;
  PyObject *__pyx_n_u_RAW;
  PyObject *__pyx_n_s_RAW_PROD_BHDR;
  PyObject *__pyx_n_s_RECORD_SIZE;
  PyObject *__pyx_n_u_RHOH;
  PyObject *__pyx_n_u_RHOH2;
  PyObject *__pyx_n_u_RHOHV;
  PyObject *__pyx_n_u_RHOHV2;
  PyObject *__pyx_n_u_RHOV;
  PyObject *__pyx_n_u_RHOV2;
  PyObject *__pyx_kp_u_Reading_ray_i_of_i;
  PyObject *__pyx_n_u_SHEAR;
  PyObject *__pyx_n_s_SIGMET_DATA_TYPES;
  PyObject *__pyx_n_s_SINT1;
  PyObject *__pyx_n_s_SINT2;
  PyObject *__pyx_n_s_SINT4;
  PyObject *__pyx_n_u_SNR16;
  PyObject *__pyx_n_u_SNR8;
  PyObject *__pyx_n_u_SQI;
  PyObject *__pyx_n_u_SQI2;
  PyObject *__pyx_n_s_STRUCTURE_HEADER;
  PyObject *__pyx_n_s_SigmetFile;
  PyObject *__pyx_n_s_SigmetFile___reduce_cython;
  PyObject *__pyx_n_s_SigmetFile___setstate_cython;
  PyObject *__pyx_n_s_SigmetFile__determine_data_types;
  PyObject *__pyx_n_s_SigmetFile__get_sweep;
  PyObject *__pyx_n_s_SigmetFile_close;
  PyObject *__pyx_n_s_SigmetFile_read_data;
  PyObject *__pyx_n_s_TASK_CALIB_INFO;
  PyObject *__pyx_n_s_TASK_CONFIGURATION;
  PyObject *__pyx_n_s_TASK_DSP_INFO;
  PyObject *__pyx_n_s_TASK_DSP_MODE_BATCH;
  PyObject *__pyx_n_s_TASK_END_INFO;
  PyObject *__pyx_n_s_TASK_FILE_SCAN_INFO;
  PyObject *__pyx_n_s_TASK_MANUAL_SCAN_INFO;
  PyObject *__pyx_n_s_TASK_MISC_INFO;
  PyObject *__pyx_n_s_TASK_PPI_SCAN_INFO;
  PyObject *__pyx_n_s_TASK_RANGE_INFO;
  PyObject *__pyx_n_s_TASK_RHI_SCAN_INFO;
  PyObject *__pyx_n_s_TASK_SCAN_INFO;
  PyObject *__pyx_n_s_TASK_SCHED_INFO;
  PyObject *__pyx_n_u_TEMPERATURE16;
  PyObject *__pyx_n_u_TIME2;
  PyObject *__pyx_n_u_TURB16;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_n_s_UINT1;
  PyObject *__pyx_n_s_UINT16_T;
  PyObject *__pyx_n_s_UINT2;
  PyObject *__pyx_n_s_UINT4;
  PyObject *__pyx_n_u_UNKNOWN_100;
  PyObject *__pyx_n_u_UNKNOWN_101;
  PyObject *__pyx_n_u_UNKNOWN_102;
  PyObject *__pyx_n_u_UNKNOWN_103;
  PyObject *__pyx_n_u_UNKNOWN_104;
  PyObject *__pyx_n_u_UNKNOWN_105;
  PyObject *__pyx_n_u_UNKNOWN_106;
  PyObject *__pyx_n_u_UNKNOWN_107;
  PyObject *__pyx_n_u_UNKNOWN_108;
  PyObject *__pyx_n_u_UNKNOWN_109;
  PyObject *__pyx_n_u_UNKNOWN_110;
  PyObject *__pyx_n_u_UNKNOWN_112;
  PyObject *__pyx_n_u_UNKNOWN_113;
  PyObject *__pyx_n_u_UNKNOWN_114;
  PyObject *__pyx_n_u_UNKNOWN_115;
  PyObject *__pyx_n_u_UNKNOWN_116;
  PyObject *__pyx_n_u_UNKNOWN_117;
  PyObject *__pyx_n_u_UNKNOWN_118;
  PyObject *__pyx_n_u_UNKNOWN_119;
  PyObject *__pyx_n_u_UNKNOWN_120;
  PyObject *__pyx_n_u_UNKNOWN_121;
  PyObject *__pyx_n_u_UNKNOWN_122;
  PyObject *__pyx_n_u_UNKNOWN_123;
  PyObject *__pyx_n_u_UNKNOWN_124;
  PyObject *__pyx_n_u_UNKNOWN_125;
  PyObject *__pyx_n_u_UNKNOWN_126;
  PyObject *__pyx_n_u_UNKNOWN_127;
  PyObject *__pyx_n_u_UNKNOWN_29;
  PyObject *__pyx_n_u_UNKNOWN_30;
  PyObject *__pyx_n_u_UNKNOWN_31;
  PyObject *__pyx_n_u_UNKNOWN_6;
  PyObject *__pyx_n_u_UNKNOWN_81;
  PyObject *__pyx_n_u_UNKNOWN_82;
  PyObject *__pyx_n_u_UNKNOWN_83;
  PyObject *__pyx_n_u_UNKNOWN_84;
  PyObject *__pyx_n_u_UNKNOWN_85;
  PyObject *__pyx_n_u_UNKNOWN_86;
  PyObject *__pyx_n_u_UNKNOWN_87;
  PyObject *__pyx_n_u_UNKNOWN_88;
  PyObject *__pyx_n_u_UNKNOWN_89;
  PyObject *__pyx_n_u_UNKNOWN_90;
  PyObject *__pyx_n_u_UNKNOWN_91;
  PyObject *__pyx_n_u_UNKNOWN_92;
  PyObject *__pyx_n_u_UNKNOWN_93;
  PyObject *__pyx_n_u_UNKNOWN_94;
  PyObject *__pyx_n_u_UNKNOWN_95;
  PyObject *__pyx_n_u_UNKNOWN_96;
  PyObject *__pyx_n_u_UNKNOWN_97;
  PyObject *__pyx_n_u_UNKNOWN_98;
  PyObject *__pyx_n_u_UNKNOWN_99;
  PyObject *__pyx_n_u_USER;
  PyObject *__pyx_n_u_USER2;
  PyObject *__pyx_kp_u_Unknown_type_s_returning_raw_dat;
  PyObject *__pyx_n_u_VEL;
  PyObject *__pyx_n_u_VEL2;
  PyObject *__pyx_n_u_VELC;
  PyObject *__pyx_n_u_VELC2;
  PyObject *__pyx_n_u_VIL2;
  PyObject *__pyx_n_u_VILD16;
  PyObject *__pyx_n_u_VIR16;
  PyObject *__pyx_n_u_VVEL2;
  PyObject *__pyx_n_u_WIDTH;
  PyObject *__pyx_n_u_WIDTH2;
  PyObject *__pyx_n_u_XHDR;
  PyObject *__pyx_n_s_YMDS_TIME;
  PyObject *__pyx_n_u_ZDR;
  PyObject *__pyx_n_u_ZDR2;
  PyObject *__pyx_n_u_ZDRC;
  PyObject *__pyx_n_u_ZDRC2;
  PyObject *__pyx_n_s__149;
  PyObject *__pyx_kp_u__15;
  PyObject *__pyx_n_s__19;
  PyObject *__pyx_n_u_agc_feedback_code;
  PyObject *__pyx_n_u_altitude_radar;
  PyObject *__pyx_n_u_angular_resolution_desired;
  PyObject *__pyx_n_u_antenna_offset_bow;
  PyObject *__pyx_n_u_antenna_offset_starboard;
  PyObject *__pyx_n_u_antenna_offset_up;
  PyObject *__pyx_n_u_antenna_scan_mode;
  PyObject *__pyx_n_s_append;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_u_attenuation;
  PyObject *__pyx_n_s_az0;
  PyObject *__pyx_n_s_az1;
  PyObject *__pyx_n_u_azimuth_0;
  PyObject *__pyx_n_u_azimuth_1;
  PyObject *__pyx_n_u_azimuth_list;
  PyObject *__pyx_n_u_b;
  PyObject *__pyx_n_s_bin2;
  PyObject *__pyx_n_s_bin2_to_angle;
  PyObject *__pyx_n_s_bin4;
  PyObject *__pyx_n_s_bin4_to_angle;
  PyObject *__pyx_n_s_bit;
  PyObject *__pyx_n_u_bit_per_bin;
  PyObject *__pyx_n_u_bool;
  PyObject *__pyx_n_u_bytes_in_structure;
  PyObject *__pyx_n_u_c;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_close;
  PyObject *__pyx_n_u_clutter_correction_threshold;
  PyObject *__pyx_n_u_clutter_filter;
  PyObject *__pyx_n_u_clutter_filter_name;
  PyObject *__pyx_n_u_cluttermap;
  PyObject *__pyx_n_u_cluttermap_flag;
  PyObject *__pyx_n_u_color_scale_def;
  PyObject *__pyx_n_u_comment_bytes;
  PyObject *__pyx_n_u_comments;
  PyObject *__pyx_n_u_configuration_name;
  PyObject *__pyx_n_s_convert_sigmet_data;
  PyObject *__pyx_n_s_copy;
  PyObject *__pyx_n_u_corrected_reflectivity_threshold;
  PyObject *__pyx_n_u_current_data_type_mask;
  PyObject *__pyx_n_u_custom_ray_header_name;
  PyObject *__pyx_n_u_customer_storage;
  PyObject *__pyx_n_s_d;
  PyObject *__pyx_n_u_d;
  PyObject *__pyx_n_s_data;
  PyObject *__pyx_n_s_data_type;
  PyObject *__pyx_n_u_data_type;
  PyObject *__pyx_n_s_data_type_name;
  PyObject *__pyx_n_s_data_types;
  PyObject *__pyx_n_s_data_types_from_mask;
  PyObject *__pyx_n_s_datetime;
  PyObject *__pyx_n_u_day;
  PyObject *__pyx_n_s_debug;
  PyObject *__pyx_n_s_determine_data_types;
  PyObject *__pyx_n_s_dic;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_u_dsp_type;
  PyObject *__pyx_n_s_dtype;
  PyObject *__pyx_n_u_dual_prf_delay;
  PyObject *__pyx_n_u_earth_radius;
  PyObject *__pyx_n_s_el0;
  PyObject *__pyx_n_s_el1;
  PyObject *__pyx_n_u_elevation_0;
  PyObject *__pyx_n_u_elevation_1;
  PyObject *__pyx_n_u_elevation_list;
  PyObject *__pyx_n_s_empty;
  PyObject *__pyx_n_s_empty_like;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_enumerate;
  PyObject *__pyx_n_s_exp;
  PyObject *__pyx_n_u_extended_header_type;
  PyObject *__pyx_n_u_extended_product_header_offset;
  PyObject *__pyx_n_u_extended_ray_header_bytes;
  PyObject *__pyx_n_u_f;
  PyObject *__pyx_n_u_fault_status;
  PyObject *__pyx_n_u_file_ingest_time;
  PyObject *__pyx_n_s_filename;
  PyObject *__pyx_n_u_filename;
  PyObject *__pyx_n_s_fill_value;
  PyObject *__pyx_n_u_first_azimuth;
  PyObject *__pyx_n_u_first_bin_range;
  PyObject *__pyx_n_u_first_elevation;
  PyObject *__pyx_n_u_first_ray_index;
  PyObject *__pyx_n_u_first_ray_offset;
  PyObject *__pyx_n_u_fixed_angle;
  PyObject *__pyx_n_u_flag;
  PyObject *__pyx_n_u_flags;
  PyObject *__pyx_n_u_flags2;
  PyObject *__pyx_n_u_float32;
  PyObject *__pyx_n_s_fmt;
  PyObject *__pyx_n_u_format_version;
  PyObject *__pyx_n_s_frombuffer;
  PyObject *__pyx_n_s_full_xhdr;
  PyObject *__pyx_n_u_gain_control_flag;
  PyObject *__pyx_n_u_gas_attenuation;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_u_generation_time;
  PyObject *__pyx_n_s_get_sweep;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_n_u_gmt_offset_minutes_local;
  PyObject *__pyx_n_u_gmt_offset_minutes_standard;
  PyObject *__pyx_n_u_gparam_bytes;
  PyObject *__pyx_n_u_gparm;
  PyObject *__pyx_n_u_ground_height;
  PyObject *__pyx_n_u_h;
  PyObject *__pyx_n_u_hardware_site;
  PyObject *__pyx_n_s_header_dic;
  PyObject *__pyx_n_u_height_radar;
  PyObject *__pyx_n_u_height_site;
  PyObject *__pyx_n_u_horizontal_beamwidth;
  PyObject *__pyx_n_u_horizontal_calibration_i0;
  PyObject *__pyx_n_u_horizontal_calibration_noise;
  PyObject *__pyx_n_u_horizontal_current_noise;
  PyObject *__pyx_n_u_horizontal_i0_calibration;
  PyObject *__pyx_n_u_horizontal_noise_calibration;
  PyObject *__pyx_n_u_horizontal_radar_constant;
  PyObject *__pyx_n_s_i;
  PyObject *__pyx_n_u_i;
  PyObject *__pyx_n_u_i4;
  PyObject *__pyx_n_u_icolcnt;
  PyObject *__pyx_n_s_idh;
  PyObject *__pyx_n_u_iflags;
  PyObject *__pyx_n_u_ilevel_seams;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_ingest_configuration;
  PyObject *__pyx_n_u_ingest_configuration;
  PyObject *__pyx_n_s_ingest_data_hdrs;
  PyObject *__pyx_n_s_ingest_data_headers;
  PyObject *__pyx_n_u_ingest_hardware_name;
  PyObject *__pyx_n_s_ingest_header;
  PyObject *__pyx_n_u_ingest_iris_version;
  PyObject *__pyx_n_u_ingest_site_name;
  PyObject *__pyx_n_u_ingest_time;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_u_input_data_type;
  PyObject *__pyx_n_u_input_mask;
  PyObject *__pyx_n_u_int16;
  PyObject *__pyx_n_u_int32;
  PyObject *__pyx_n_u_inverse_flatting;
  PyObject *__pyx_n_u_iris_version;
  PyObject *__pyx_n_u_iris_version_created;
  PyObject *__pyx_n_s_is_bit_set;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_u_iset_and_scale;
  PyObject *__pyx_n_u_istart;
  PyObject *__pyx_n_u_istep;
  PyObject *__pyx_n_s_j;
  PyObject *__pyx_n_s_k;
  PyObject *__pyx_n_s_key;
  PyObject *__pyx_n_s_l;
  PyObject *__pyx_n_u_last_bin_range;
  PyObject *__pyx_n_u_last_run_day;
  PyObject *__pyx_n_u_last_run_time;
  PyObject *__pyx_n_u_latitude;
  PyObject *__pyx_n_u_latitude_projection;
  PyObject *__pyx_n_u_latitude_radar;
  PyObject *__pyx_n_u_ldr_bias;
  PyObject *__pyx_n_u_ldr_offset;
  PyObject *__pyx_n_s_lead_record;
  PyObject *__pyx_n_u_left_azimuth_limit;
  PyObject *__pyx_n_s_like_dbt;
  PyObject *__pyx_n_s_like_dbt2;
  PyObject *__pyx_n_s_like_sqi;
  PyObject *__pyx_n_s_like_sqi2;
  PyObject *__pyx_n_u_linear_filter_first_bin;
  PyObject *__pyx_n_u_local_timezone;
  PyObject *__pyx_n_u_log_filter_first_bin;
  PyObject *__pyx_n_u_longitude;
  PyObject *__pyx_n_u_longitude_projection;
  PyObject *__pyx_n_u_longitude_radar;
  PyObject *__pyx_n_u_low_prf_factional;
  PyObject *__pyx_n_u_low_prf_hz;
  PyObject *__pyx_n_u_low_prf_range_averaging;
  PyObject *__pyx_n_u_low_prf_sample_size;
  PyObject *__pyx_n_u_lower_elevation_limit;
  PyObject *__pyx_n_s_ma;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_u_major_mode;
  PyObject *__pyx_n_s_mask;
  PyObject *__pyx_n_u_mask_word_0;
  PyObject *__pyx_n_u_mask_word_1;
  PyObject *__pyx_n_u_mask_word_2;
  PyObject *__pyx_n_u_mask_word_3;
  PyObject *__pyx_n_u_mask_word_4;
  PyObject *__pyx_n_s_masked_array;
  PyObject *__pyx_n_u_maximum_range;
  PyObject *__pyx_n_u_mean_wind_direction;
  PyObject *__pyx_n_u_mean_wind_speed;
  PyObject *__pyx_n_u_melting_layer;
  PyObject *__pyx_n_u_melting_level;
  PyObject *__pyx_n_s_mess;
  PyObject *__pyx_n_s_metadata;
  PyObject *__pyx_n_u_milliseconds;
  PyObject *__pyx_n_u_minor_task_suffix;
  PyObject *__pyx_n_s_mod;
  PyObject *__pyx_n_u_month;
  PyObject *__pyx_n_s_multi_prf_flag;
  PyObject *__pyx_n_u_multi_prf_flag;
  PyObject *__pyx_n_s_multiplier;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_name_2;
  PyObject *__pyx_n_s_nbin;
  PyObject *__pyx_n_s_nbins;
  PyObject *__pyx_n_u_nbins;
  PyObject *__pyx_n_s_ndata;
  PyObject *__pyx_n_s_ndata_types;
  PyObject *__pyx_n_u_nexrad_clutter_skip;
  PyObject *__pyx_n_u_nexrad_clutter_threshold;
  PyObject *__pyx_n_s_np;
  PyObject *__pyx_n_s_nray_data_types;
  PyObject *__pyx_n_s_nrays;
  PyObject *__pyx_n_s_nsweeps;
  PyObject *__pyx_n_s_number;
  PyObject *__pyx_n_u_number_bins;
  PyObject *__pyx_n_u_number_elements;
  PyObject *__pyx_n_u_number_files;
  PyObject *__pyx_n_u_number_ingest;
  PyObject *__pyx_n_u_number_input_bins;
  PyObject *__pyx_n_u_number_linear_filter;
  PyObject *__pyx_n_u_number_log_filter;
  PyObject *__pyx_n_u_number_output_bins;
  PyObject *__pyx_n_u_number_rays_file_actual;
  PyObject *__pyx_n_u_number_rays_file_expected;
  PyObject *__pyx_n_u_number_rays_sweep;
  PyObject *__pyx_n_u_number_sweeps;
  PyObject *__pyx_n_u_number_sweeps_completed;
  PyObject *__pyx_n_u_number_task_config_table;
  PyObject *__pyx_n_u_number_tasks;
  PyObject *__pyx_n_s_numpy;
  PyObject *__pyx_kp_u_numpy__core_multiarray_failed_to;
  PyObject *__pyx_kp_u_numpy__core_umath_failed_to_impo;
  PyObject *__pyx_n_s_nyquist;
  PyObject *__pyx_n_s_offset;
  PyObject *__pyx_n_s_ones;
  PyObject *__pyx_n_s_open;
  PyObject *__pyx_n_u_original_data_type_mask;
  PyObject *__pyx_n_s_out;
  PyObject *__pyx_n_s_parse_ray_headers;
  PyObject *__pyx_n_u_playback_flag;
  PyObject *__pyx_n_u_playback_version;
  PyObject *__pyx_n_u_polarization;
  PyObject *__pyx_n_u_polarization_type;
  PyObject *__pyx_n_s_power;
  PyObject *__pyx_n_u_power_threshold;
  PyObject *__pyx_n_u_prf;
  PyObject *__pyx_n_s_prf_flag;
  PyObject *__pyx_n_u_prf_flag;
  PyObject *__pyx_n_s_print;
  PyObject *__pyx_n_s_product_config;
  PyObject *__pyx_n_u_product_configuration;
  PyObject *__pyx_n_s_product_end;
  PyObject *__pyx_n_u_product_end;
  PyObject *__pyx_n_s_product_hdr;
  PyObject *__pyx_n_u_product_name;
  PyObject *__pyx_n_u_product_sequence_number;
  PyObject *__pyx_n_u_product_specific_bytes;
  PyObject *__pyx_n_u_product_type_code;
  PyObject *__pyx_n_u_projection_name;
  PyObject *__pyx_n_u_projection_type;
  PyObject *__pyx_n_s_prt_value;
  PyObject *__pyx_n_u_pulse_width;
  PyObject *__pyx_n_s_pyart_io__sigmetfile;
  PyObject *__pyx_kp_s_pyart_io__sigmetfile_pyx;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_u_radar_height;
  PyObject *__pyx_n_u_radar_height_above_reference;
  PyObject *__pyx_n_u_radial_smoother;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_u_range_bin_averaging_flag;
  PyObject *__pyx_n_s_raw_data;
  PyObject *__pyx_n_s_raw_prod_bhdr;
  PyObject *__pyx_n_s_raw_sweep_data;
  PyObject *__pyx_n_u_ray_header_bytes;
  PyObject *__pyx_n_u_ray_header_mask;
  PyObject *__pyx_n_s_ray_headers;
  PyObject *__pyx_n_s_ray_i;
  PyObject *__pyx_n_s_ray_nbins;
  PyObject *__pyx_n_u_ray_number;
  PyObject *__pyx_n_s_ray_time;
  PyObject *__pyx_n_u_rb;
  PyObject *__pyx_n_s_read;
  PyObject *__pyx_n_u_read;
  PyObject *__pyx_n_s_read_data;
  PyObject *__pyx_n_u_reciever_bandwidth;
  PyObject *__pyx_n_s_record;
  PyObject *__pyx_n_u_record_number;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_u_reflectivity_calibration;
  PyObject *__pyx_n_u_reflectivity_noise_threshold;
  PyObject *__pyx_n_u_reflectivity_slope;
  PyObject *__pyx_n_u_reflectivity_unfolding_threshold;
  PyObject *__pyx_n_u_reserved;
  PyObject *__pyx_n_s_reshape;
  PyObject *__pyx_n_u_resolution_rays;
  PyObject *__pyx_n_u_right_azimuth_limit;
  PyObject *__pyx_n_u_sample_size;
  PyObject *__pyx_n_u_samples_used;
  PyObject *__pyx_n_u_scheduling_code;
  PyObject *__pyx_n_u_seconds;
  PyObject *__pyx_n_u_seconds_between_runs;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_kp_s_self__rbuf_p_cannot_be_converted;
  PyObject *__pyx_kp_u_self__rbuf_pos_is;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_shape;
  PyObject *__pyx_n_s_shrink;
  PyObject *__pyx_n_u_signal_processor_type;
  PyObject *__pyx_n_u_site_name;
  PyObject *__pyx_n_u_skip_time;
  PyObject *__pyx_n_u_spare_0;
  PyObject *__pyx_n_u_spare_1;
  PyObject *__pyx_n_u_spare_2;
  PyObject *__pyx_n_u_spare_3;
  PyObject *__pyx_n_u_spare_4;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_n_u_sqi_threshold;
  PyObject *__pyx_n_s_sqrt;
  PyObject *__pyx_n_u_standard_parallel_1;
  PyObject *__pyx_n_u_standard_parallel_2;
  PyObject *__pyx_n_u_start_first_section_flag;
  PyObject *__pyx_n_u_start_first_sector_flag;
  PyObject *__pyx_n_u_start_time;
  PyObject *__pyx_n_u_step_input_bins;
  PyObject *__pyx_n_u_step_output_bins;
  PyObject *__pyx_n_u_stop_time;
  PyObject *__pyx_n_s_string;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_struct;
  PyObject *__pyx_n_s_structure;
  PyObject *__pyx_n_u_structure_header;
  PyObject *__pyx_n_u_structure_identifier;
  PyObject *__pyx_n_s_sum;
  PyObject *__pyx_n_s_sweep_data;
  PyObject *__pyx_n_u_sweep_ingest_time;
  PyObject *__pyx_n_s_sweep_metadata;
  PyObject *__pyx_n_u_sweep_number;
  PyObject *__pyx_n_u_sweep_start_time;
  PyObject *__pyx_n_u_task_calib_info;
  PyObject *__pyx_n_s_task_config;
  PyObject *__pyx_n_s_task_configuration;
  PyObject *__pyx_n_u_task_configuration;
  PyObject *__pyx_n_u_task_configuration_file_name;
  PyObject *__pyx_n_u_task_data_time;
  PyObject *__pyx_n_u_task_description;
  PyObject *__pyx_n_s_task_dsp_info;
  PyObject *__pyx_n_u_task_dsp_info;
  PyObject *__pyx_n_u_task_dsp_mode;
  PyObject *__pyx_n_s_task_end_info;
  PyObject *__pyx_n_u_task_end_info;
  PyObject *__pyx_n_u_task_major_number;
  PyObject *__pyx_n_u_task_minor_number;
  PyObject *__pyx_n_u_task_misc_info;
  PyObject *__pyx_n_u_task_name;
  PyObject *__pyx_n_u_task_range_info;
  PyObject *__pyx_n_u_task_scan_info;
  PyObject *__pyx_n_u_task_scan_type_scan_info;
  PyObject *__pyx_n_u_task_sched_info;
  PyObject *__pyx_n_u_task_state;
  PyObject *__pyx_n_u_tcf_cal_flags_1;
  PyObject *__pyx_n_u_tcf_cal_flags_2;
  PyObject *__pyx_n_s_temp;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_time;
  PyObject *__pyx_n_u_time;
  PyObject *__pyx_n_u_time_used_last_run;
  PyObject *__pyx_n_u_times_run;
  PyObject *__pyx_n_u_total_size;
  PyObject *__pyx_n_u_tr_serial_number;
  PyObject *__pyx_n_u_tranmitter_phase_sequence;
  PyObject *__pyx_n_u_transmit_power;
  PyObject *__pyx_n_u_trigger_rate;
  PyObject *__pyx_n_u_trucation_height;
  PyObject *__pyx_n_u_truncation_height;
  PyObject *__pyx_n_u_tz_name;
  PyObject *__pyx_n_u_uint16;
  PyObject *__pyx_n_s_uint8;
  PyObject *__pyx_n_u_uncorrected_reflectivity_thresho;
  PyObject *__pyx_n_s_unpack;
  PyObject *__pyx_n_s_unpack_ingest_data_header;
  PyObject *__pyx_n_s_unpack_ingest_data_headers;
  PyObject *__pyx_n_s_unpack_ingest_header;
  PyObject *__pyx_n_s_unpack_key;
  PyObject *__pyx_n_s_unpack_product_hdr;
  PyObject *__pyx_n_s_unpack_raw_prod_bhdr;
  PyObject *__pyx_n_s_unpack_structure;
  PyObject *__pyx_n_u_upper_elevation_limit;
  PyObject *__pyx_n_u_variable_range_bin_flag;
  PyObject *__pyx_n_u_velocity_east;
  PyObject *__pyx_n_u_velocity_north;
  PyObject *__pyx_n_u_velocity_threshold_flags;
  PyObject *__pyx_n_u_velocity_unfolding_threshold;
  PyObject *__pyx_n_u_velocity_up;
  PyObject *__pyx_n_u_vertical_beamwidth;
  PyObject *__pyx_n_u_vertical_current_noise;
  PyObject *__pyx_n_u_vertical_i0_calibration;
  PyObject *__pyx_n_u_vertical_noise_calibration;
  PyObject *__pyx_n_u_vertical_radar_constant;
  PyObject *__pyx_n_s_view;
  PyObject *__pyx_n_u_volume_scan_start_time;
  PyObject *__pyx_n_s_warn;
  PyObject *__pyx_n_s_warnings;
  PyObject *__pyx_n_u_wavelength;
  PyObject *__pyx_n_s_wavelength_cm;
  PyObject *__pyx_n_u_width_threshold_flags;
  PyObject *__pyx_n_u_width_unfolding_threshold;
  PyObject *__pyx_n_s_word0;
  PyObject *__pyx_n_s_word1;
  PyObject *__pyx_n_s_word2;
  PyObject *__pyx_n_s_word3;
  PyObject *__pyx_n_u_x_location;
  PyObject *__pyx_n_u_x_scale;
  PyObject *__pyx_n_u_x_size;
  PyObject *__pyx_n_u_x_smoother;
  PyObject *__pyx_n_s_xrange;
  PyObject *__pyx_n_u_y_location;
  PyObject *__pyx_n_u_y_scale;
  PyObject *__pyx_n_u_y_size;
  PyObject *__pyx_n_u_y_smoother;
  PyObject *__pyx_n_u_year;
  PyObject *__pyx_n_u_z_location;
  PyObject *__pyx_n_u_z_scale;
  PyObject *__pyx_n_u_z_size;
  PyObject *__pyx_n_u_zdr_bias;
  PyObject *__pyx_n_u_zdr_offset;
  PyObject *__pyx_n_u_zdr_threshold_flags;
  PyObject *__pyx_n_s_zeros_like;
  PyObject *__pyx_n_s_zip;
  PyObject *__pyx_n_u_zr_constant;
  PyObject *__pyx_n_u_zr_exponent;
  PyObject *__pyx_float_1_;
  PyObject *__pyx_float_2_;
  PyObject *__pyx_float_16_;
  PyObject *__pyx_float_64_;
  PyObject *__pyx_float_75_;
  PyObject *__pyx_float_0_25;
  PyObject *__pyx_float_100_;
  PyObject *__pyx_float_126_;
  PyObject *__pyx_float_127_;
  PyObject *__pyx_float_128_;
  PyObject *__pyx_float_129_;
  PyObject *__pyx_float_180_;
  PyObject *__pyx_float_253_;
  PyObject *__pyx_float_254_;
  PyObject *__pyx_float_256_;
  PyObject *__pyx_float_360_;
  PyObject *__pyx_float_600_;
  PyObject *__pyx_float_100_0;
  PyObject *__pyx_float_32768_;
  PyObject *__pyx_float_65533_;
  PyObject *__pyx_float_65534_;
  PyObject *__pyx_float_neg_0_25;
  PyObject *__pyx_float_neg_9999_0;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_2;
  PyObject *__pyx_int_3;
  PyObject *__pyx_int_4;
  PyObject *__pyx_int_5;
  PyObject *__pyx_int_6;
  PyObject *__pyx_int_7;
  PyObject *__pyx_int_8;
  PyObject *__pyx_int_9;
  PyObject *__pyx_int_10;
  PyObject *__pyx_int_11;
  PyObject *__pyx_int_12;
  PyObject *__pyx_int_13;
  PyObject *__pyx_int_14;
  PyObject *__pyx_int_15;
  PyObject *__pyx_int_16;
  PyObject *__pyx_int_17;
  PyObject *__pyx_int_18;
  PyObject *__pyx_int_19;
  PyObject *__pyx_int_20;
  PyObject *__pyx_int_21;
  PyObject *__pyx_int_22;
  PyObject *__pyx_int_23;
  PyObject *__pyx_int_24;
  PyObject *__pyx_int_25;
  PyObject *__pyx_int_26;
  PyObject *__pyx_int_27;
  PyObject *__pyx_int_28;
  PyObject *__pyx_int_29;
  PyObject *__pyx_int_30;
  PyObject *__pyx_int_31;
  PyObject *__pyx_int_32;
  PyObject *__pyx_int_33;
  PyObject *__pyx_int_34;
  PyObject *__pyx_int_35;
  PyObject *__pyx_int_36;
  PyObject *__pyx_int_37;
  PyObject *__pyx_int_38;
  PyObject *__pyx_int_39;
  PyObject *__pyx_int_40;
  PyObject *__pyx_int_41;
  PyObject *__pyx_int_42;
  PyObject *__pyx_int_43;
  PyObject *__pyx_int_44;
  PyObject *__pyx_int_45;
  PyObject *__pyx_int_46;
  PyObject *__pyx_int_47;
  PyObject *__pyx_int_48;
  PyObject *__pyx_int_49;
  PyObject *__pyx_int_50;
  PyObject *__pyx_int_51;
  PyObject *__pyx_int_52;
  PyObject *__pyx_int_53;
  PyObject *__pyx_int_54;
  PyObject *__pyx_int_55;
  PyObject *__pyx_int_56;
  PyObject *__pyx_int_57;
  PyObject *__pyx_int_58;
  PyObject *__pyx_int_59;
  PyObject *__pyx_int_60;
  PyObject *__pyx_int_61;
  PyObject *__pyx_int_62;
  PyObject *__pyx_int_63;
  PyObject *__pyx_int_64;
  PyObject *__pyx_int_65;
  PyObject *__pyx_int_66;
  PyObject *__pyx_int_67;
  PyObject *__pyx_int_68;
  PyObject *__pyx_int_69;
  PyObject *__pyx_int_70;
  PyObject *__pyx_int_71;
  PyObject *__pyx_int_72;
  PyObject *__pyx_int_73;
  PyObject *__pyx_int_74;
  PyObject *__pyx_int_75;
  PyObject *__pyx_int_76;
  PyObject *__pyx_int_77;
  PyObject *__pyx_int_78;
  PyObject *__pyx_int_79;
  PyObject *__pyx_int_80;
  PyObject *__pyx_int_81;
  PyObject *__pyx_int_82;
  PyObject *__pyx_int_83;
  PyObject *__pyx_int_84;
  PyObject *__pyx_int_85;
  PyObject *__pyx_int_86;
  PyObject *__pyx_int_87;
  PyObject *__pyx_int_88;
  PyObject *__pyx_int_89;
  PyObject *__pyx_int_90;
  PyObject *__pyx_int_91;
  PyObject *__pyx_int_92;
  PyObject *__pyx_int_93;
  PyObject *__pyx_int_94;
  PyObject *__pyx_int_95;
  PyObject *__pyx_int_96;
  PyObject *__pyx_int_97;
  PyObject *__pyx_int_98;
  PyObject *__pyx_int_99;
  PyObject *__pyx_int_100;
  PyObject *__pyx_int_101;
  PyObject *__pyx_int_102;
  PyObject *__pyx_int_103;
  PyObject *__pyx_int_104;
  PyObject *__pyx_int_105;
  PyObject *__pyx_int_106;
  PyObject *__pyx_int_107;
  PyObject *__pyx_int_108;
  PyObject *__pyx_int_109;
  PyObject *__pyx_int_110;
  PyObject *__pyx_int_111;
  PyObject *__pyx_int_112;
  PyObject *__pyx_int_113;
  PyObject *__pyx_int_114;
  PyObject *__pyx_int_115;
  PyObject *__pyx_int_116;
  PyObject *__pyx_int_117;
  PyObject *__pyx_int_118;
  PyObject *__pyx_int_119;
  PyObject *__pyx_int_120;
  PyObject *__pyx_int_121;
  PyObject *__pyx_int_122;
  PyObject *__pyx_int_123;
  PyObject *__pyx_int_124;
  PyObject *__pyx_int_125;
  PyObject *__pyx_int_126;
  PyObject *__pyx_int_127;
  PyObject *__pyx_int_128;
  PyObject *__pyx_int_255;
  PyObject *__pyx_int_640;
  PyObject *__pyx_int_4884;
  PyObject *__pyx_int_6144;
  PyObject *__pyx_int_65536;
  PyObject *__pyx_int_4294967296;
  PyObject *__pyx_int_neg_1;
  PyObject *__pyx_tuple_;
  PyObject *__pyx_slice__4;
  PyObject *__pyx_slice__5;
  PyObject *__pyx_tuple__2;
  PyObject *__pyx_tuple__3;
  PyObject *__pyx_tuple__6;
  PyObject *__pyx_tuple__7;
  PyObject *__pyx_tuple__8;
  PyObject *__pyx_tuple__9;
  PyObject *__pyx_slice__12;
  PyObject *__pyx_slice__13;
  PyObject *__pyx_slice__16;
  PyObject *__pyx_slice__17;
  PyObject *__pyx_slice__18;
  PyObject *__pyx_tuple__10;
  PyObject *__pyx_tuple__11;
  PyObject *__pyx_tuple__14;
  PyObject *__pyx_tuple__20;
  PyObject *__pyx_tuple__22;
  PyObject *__pyx_tuple__24;
  PyObject *__pyx_tuple__26;
  PyObject *__pyx_tuple__27;
  PyObject *__pyx_tuple__29;
  PyObject *__pyx_tuple__31;
  PyObject *__pyx_tuple__33;
  PyObject *__pyx_tuple__35;
  PyObject *__pyx_tuple__37;
  PyObject *__pyx_tuple__39;
  PyObject *__pyx_tuple__41;
  PyObject *__pyx_tuple__43;
  PyObject *__pyx_tuple__45;
  PyObject *__pyx_tuple__47;
  PyObject *__pyx_tuple__49;
  PyObject *__pyx_tuple__51;
  PyObject *__pyx_tuple__53;
  PyObject *__pyx_tuple__55;
  PyObject *__pyx_tuple__57;
  PyObject *__pyx_tuple__59;
  PyObject *__pyx_tuple__60;
  PyObject *__pyx_tuple__61;
  PyObject *__pyx_tuple__62;
  PyObject *__pyx_tuple__63;
  PyObject *__pyx_tuple__64;
  PyObject *__pyx_tuple__65;
  PyObject *__pyx_tuple__66;
  PyObject *__pyx_tuple__67;
  PyObject *__pyx_tuple__68;
  PyObject *__pyx_tuple__69;
  PyObject *__pyx_tuple__70;
  PyObject *__pyx_tuple__71;
  PyObject *__pyx_tuple__72;
  PyObject *__pyx_tuple__73;
  PyObject *__pyx_tuple__74;
  PyObject *__pyx_tuple__75;
  PyObject *__pyx_tuple__76;
  PyObject *__pyx_tuple__77;
  PyObject *__pyx_tuple__78;
  PyObject *__pyx_tuple__79;
  PyObject *__pyx_tuple__80;
  PyObject *__pyx_tuple__81;
  PyObject *__pyx_tuple__82;
  PyObject *__pyx_tuple__83;
  PyObject *__pyx_tuple__84;
  PyObject *__pyx_tuple__85;
  PyObject *__pyx_tuple__86;
  PyObject *__pyx_tuple__87;
  PyObject *__pyx_tuple__88;
  PyObject *__pyx_tuple__89;
  PyObject *__pyx_tuple__90;
  PyObject *__pyx_tuple__91;
  PyObject *__pyx_tuple__92;
  PyObject *__pyx_tuple__93;
  PyObject *__pyx_tuple__94;
  PyObject *__pyx_tuple__95;
  PyObject *__pyx_tuple__96;
  PyObject *__pyx_tuple__97;
  PyObject *__pyx_tuple__98;
  PyObject *__pyx_tuple__99;
  PyObject *__pyx_tuple__100;
  PyObject *__pyx_tuple__101;
  PyObject *__pyx_tuple__102;
  PyObject *__pyx_tuple__103;
  PyObject *__pyx_tuple__104;
  PyObject *__pyx_tuple__105;
  PyObject *__pyx_tuple__106;
  PyObject *__pyx_tuple__107;
  PyObject *__pyx_tuple__108;
  PyObject *__pyx_tuple__109;
  PyObject *__pyx_tuple__110;
  PyObject *__pyx_tuple__111;
  PyObject *__pyx_tuple__112;
  PyObject *__pyx_tuple__113;
  PyObject *__pyx_tuple__114;
  PyObject *__pyx_tuple__115;
  PyObject *__pyx_tuple__116;
  PyObject *__pyx_tuple__117;
  PyObject *__pyx_tuple__118;
  PyObject *__pyx_tuple__119;
  PyObject *__pyx_tuple__120;
  PyObject *__pyx_tuple__121;
  PyObject *__pyx_tuple__122;
  PyObject *__pyx_tuple__123;
  PyObject *__pyx_tuple__124;
  PyObject *__pyx_tuple__125;
  PyObject *__pyx_tuple__126;
  PyObject *__pyx_tuple__127;
  PyObject *__pyx_tuple__128;
  PyObject *__pyx_tuple__129;
  PyObject *__pyx_tuple__130;
  PyObject *__pyx_tuple__131;
  PyObject *__pyx_tuple__132;
  PyObject *__pyx_tuple__133;
  PyObject *__pyx_tuple__134;
  PyObject *__pyx_tuple__135;
  PyObject *__pyx_tuple__136;
  PyObject *__pyx_tuple__137;
  PyObject *__pyx_tuple__138;
  PyObject *__pyx_tuple__139;
  PyObject *__pyx_tuple__140;
  PyObject *__pyx_tuple__141;
  PyObject *__pyx_tuple__142;
  PyObject *__pyx_tuple__143;
  PyObject *__pyx_tuple__144;
  PyObject *__pyx_tuple__145;
  PyObject *__pyx_tuple__146;
  PyObject *__pyx_tuple__147;
  PyObject *__pyx_tuple__148;
  PyObject *__pyx_codeobj__21;
  PyObject *__pyx_codeobj__23;
  PyObject *__pyx_codeobj__25;
  PyObject *__pyx_codeobj__28;
  PyObject *__pyx_codeobj__30;
  PyObject *__pyx_codeobj__32;
  PyObject *__pyx_codeobj__34;
  PyObject *__pyx_codeobj__36;
  PyObject *__pyx_codeobj__38;
  PyObject *__pyx_codeobj__40;
  PyObject *__pyx_codeobj__42;
  PyObject *__pyx_codeobj__44;
  PyObject *__pyx_codeobj__46;
  PyObject *__pyx_codeobj__48;
  PyObject *__pyx_codeobj__50;
  PyObject *__pyx_codeobj__52;
  PyObject *__pyx_codeobj__54;
  PyObject *__pyx_codeobj__56;
  PyObject *__pyx_codeobj__58;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_dtype);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flatiter);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_broadcast);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ndarray);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_generic);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_number);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_integer);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_inexact);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_floating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flexible);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_character);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ufunc);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile);
  Py_CLEAR(clear_module_state->__pyx_type_5pyart_2io_11_sigmetfile_SigmetFile);
  Py_CLEAR(clear_module_state->__pyx_kp_u_112s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_115s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_120s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_128s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_12s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_136s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_160s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_16s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_184s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_18s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_198s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_1s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_200s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_204s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_208s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_228s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_24s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_256s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_2612s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_28s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_2_uint8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_2s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_308s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_320s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_32s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_36s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_40s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_480s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_48s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_4s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_52s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_6s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_720s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_732s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_80s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_8s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_920s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_94s);
  Py_CLEAR(clear_module_state->__pyx_n_u_ALBEDO16);
  Py_CLEAR(clear_module_state->__pyx_n_u_ALBEDO8);
  Py_CLEAR(clear_module_state->__pyx_n_u_AXDIL2);
  Py_CLEAR(clear_module_state->__pyx_n_u_B);
  Py_CLEAR(clear_module_state->__pyx_n_s_BIN1);
  Py_CLEAR(clear_module_state->__pyx_n_s_BIN2);
  Py_CLEAR(clear_module_state->__pyx_n_s_BIN4);
  Py_CLEAR(clear_module_state->__pyx_n_s_COLOR_SCALE_DEF);
  Py_CLEAR(clear_module_state->__pyx_n_u_CSP16);
  Py_CLEAR(clear_module_state->__pyx_n_u_CSP8);
  Py_CLEAR(clear_module_state->__pyx_n_u_DBT);
  Py_CLEAR(clear_module_state->__pyx_n_u_DBT2);
  Py_CLEAR(clear_module_state->__pyx_n_u_DBTE16);
  Py_CLEAR(clear_module_state->__pyx_n_u_DBTE8);
  Py_CLEAR(clear_module_state->__pyx_n_u_DBTV16);
  Py_CLEAR(clear_module_state->__pyx_n_u_DBTV8);
  Py_CLEAR(clear_module_state->__pyx_n_u_DBZ);
  Py_CLEAR(clear_module_state->__pyx_n_u_DBZ2);
  Py_CLEAR(clear_module_state->__pyx_n_u_DBZC);
  Py_CLEAR(clear_module_state->__pyx_n_u_DBZC2);
  Py_CLEAR(clear_module_state->__pyx_n_u_DBZE16);
  Py_CLEAR(clear_module_state->__pyx_n_u_DBZE8);
  Py_CLEAR(clear_module_state->__pyx_n_u_DBZV16);
  Py_CLEAR(clear_module_state->__pyx_n_u_DBZV8);
  Py_CLEAR(clear_module_state->__pyx_n_u_DEFORM2);
  Py_CLEAR(clear_module_state->__pyx_n_u_DIVERGE2);
  Py_CLEAR(clear_module_state->__pyx_n_s_DSP_DATA_MASK);
  Py_CLEAR(clear_module_state->__pyx_n_u_FLIQUID2);
  Py_CLEAR(clear_module_state->__pyx_n_s_FLT4);
  Py_CLEAR(clear_module_state->__pyx_n_s_FLT8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_File_truncated_or_corrupt_i_of_i);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Finished_loading_record);
  Py_CLEAR(clear_module_state->__pyx_n_u_GMT_minute_offset_local);
  Py_CLEAR(clear_module_state->__pyx_n_u_GMT_minute_offset_standard);
  Py_CLEAR(clear_module_state->__pyx_n_u_H);
  Py_CLEAR(clear_module_state->__pyx_n_u_HCLASS);
  Py_CLEAR(clear_module_state->__pyx_n_u_HCLASS2);
  Py_CLEAR(clear_module_state->__pyx_n_u_HDIR2);
  Py_CLEAR(clear_module_state->__pyx_n_u_HEIGHT);
  Py_CLEAR(clear_module_state->__pyx_n_u_HVEL2);
  Py_CLEAR(clear_module_state->__pyx_n_u_I);
  Py_CLEAR(clear_module_state->__pyx_n_s_INGEST_CONFIGURATION);
  Py_CLEAR(clear_module_state->__pyx_n_s_INGEST_DATA_HEADER);
  Py_CLEAR(clear_module_state->__pyx_n_s_INGEST_HEADER);
  Py_CLEAR(clear_module_state->__pyx_n_s_ImportError);
  Py_CLEAR(clear_module_state->__pyx_n_u_KDP);
  Py_CLEAR(clear_module_state->__pyx_n_u_KDP2);
  Py_CLEAR(clear_module_state->__pyx_n_u_LDRH);
  Py_CLEAR(clear_module_state->__pyx_n_u_LDRH2);
  Py_CLEAR(clear_module_state->__pyx_n_u_LDRV);
  Py_CLEAR(clear_module_state->__pyx_n_u_LDRV2);
  Py_CLEAR(clear_module_state->__pyx_n_u_LOG16);
  Py_CLEAR(clear_module_state->__pyx_n_u_LOG8);
  Py_CLEAR(clear_module_state->__pyx_n_s_MESSAGE);
  Py_CLEAR(clear_module_state->__pyx_n_u_OTHER);
  Py_CLEAR(clear_module_state->__pyx_n_u_PHIDP);
  Py_CLEAR(clear_module_state->__pyx_n_u_PHIDP2);
  Py_CLEAR(clear_module_state->__pyx_n_u_PHIH);
  Py_CLEAR(clear_module_state->__pyx_n_u_PHIH2);
  Py_CLEAR(clear_module_state->__pyx_n_u_PHIV);
  Py_CLEAR(clear_module_state->__pyx_n_u_PHIV2);
  Py_CLEAR(clear_module_state->__pyx_n_u_PMI16);
  Py_CLEAR(clear_module_state->__pyx_n_u_PMI8);
  Py_CLEAR(clear_module_state->__pyx_n_s_PRODUCT_CONFIGURATION);
  Py_CLEAR(clear_module_state->__pyx_n_s_PRODUCT_END);
  Py_CLEAR(clear_module_state->__pyx_n_s_PRODUCT_HDR);
  Py_CLEAR(clear_module_state->__pyx_n_u_RAINRATE2);
  Py_CLEAR(clear_module_state->__pyx_n_u_RAW);
  Py_CLEAR(clear_module_state->__pyx_n_s_RAW_PROD_BHDR);
  Py_CLEAR(clear_module_state->__pyx_n_s_RECORD_SIZE);
  Py_CLEAR(clear_module_state->__pyx_n_u_RHOH);
  Py_CLEAR(clear_module_state->__pyx_n_u_RHOH2);
  Py_CLEAR(clear_module_state->__pyx_n_u_RHOHV);
  Py_CLEAR(clear_module_state->__pyx_n_u_RHOHV2);
  Py_CLEAR(clear_module_state->__pyx_n_u_RHOV);
  Py_CLEAR(clear_module_state->__pyx_n_u_RHOV2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Reading_ray_i_of_i);
  Py_CLEAR(clear_module_state->__pyx_n_u_SHEAR);
  Py_CLEAR(clear_module_state->__pyx_n_s_SIGMET_DATA_TYPES);
  Py_CLEAR(clear_module_state->__pyx_n_s_SINT1);
  Py_CLEAR(clear_module_state->__pyx_n_s_SINT2);
  Py_CLEAR(clear_module_state->__pyx_n_s_SINT4);
  Py_CLEAR(clear_module_state->__pyx_n_u_SNR16);
  Py_CLEAR(clear_module_state->__pyx_n_u_SNR8);
  Py_CLEAR(clear_module_state->__pyx_n_u_SQI);
  Py_CLEAR(clear_module_state->__pyx_n_u_SQI2);
  Py_CLEAR(clear_module_state->__pyx_n_s_STRUCTURE_HEADER);
  Py_CLEAR(clear_module_state->__pyx_n_s_SigmetFile);
  Py_CLEAR(clear_module_state->__pyx_n_s_SigmetFile___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_SigmetFile___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_SigmetFile__determine_data_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_SigmetFile__get_sweep);
  Py_CLEAR(clear_module_state->__pyx_n_s_SigmetFile_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_SigmetFile_read_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_TASK_CALIB_INFO);
  Py_CLEAR(clear_module_state->__pyx_n_s_TASK_CONFIGURATION);
  Py_CLEAR(clear_module_state->__pyx_n_s_TASK_DSP_INFO);
  Py_CLEAR(clear_module_state->__pyx_n_s_TASK_DSP_MODE_BATCH);
  Py_CLEAR(clear_module_state->__pyx_n_s_TASK_END_INFO);
  Py_CLEAR(clear_module_state->__pyx_n_s_TASK_FILE_SCAN_INFO);
  Py_CLEAR(clear_module_state->__pyx_n_s_TASK_MANUAL_SCAN_INFO);
  Py_CLEAR(clear_module_state->__pyx_n_s_TASK_MISC_INFO);
  Py_CLEAR(clear_module_state->__pyx_n_s_TASK_PPI_SCAN_INFO);
  Py_CLEAR(clear_module_state->__pyx_n_s_TASK_RANGE_INFO);
  Py_CLEAR(clear_module_state->__pyx_n_s_TASK_RHI_SCAN_INFO);
  Py_CLEAR(clear_module_state->__pyx_n_s_TASK_SCAN_INFO);
  Py_CLEAR(clear_module_state->__pyx_n_s_TASK_SCHED_INFO);
  Py_CLEAR(clear_module_state->__pyx_n_u_TEMPERATURE16);
  Py_CLEAR(clear_module_state->__pyx_n_u_TIME2);
  Py_CLEAR(clear_module_state->__pyx_n_u_TURB16);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_UINT1);
  Py_CLEAR(clear_module_state->__pyx_n_s_UINT16_T);
  Py_CLEAR(clear_module_state->__pyx_n_s_UINT2);
  Py_CLEAR(clear_module_state->__pyx_n_s_UINT4);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_100);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_101);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_102);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_103);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_104);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_105);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_106);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_107);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_108);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_109);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_110);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_112);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_113);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_114);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_115);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_116);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_117);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_118);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_119);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_120);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_121);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_122);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_123);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_124);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_125);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_126);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_127);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_29);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_30);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_31);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_6);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_81);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_82);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_83);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_84);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_85);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_86);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_87);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_88);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_89);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_90);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_91);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_92);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_93);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_94);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_95);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_96);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_97);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_98);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN_99);
  Py_CLEAR(clear_module_state->__pyx_n_u_USER);
  Py_CLEAR(clear_module_state->__pyx_n_u_USER2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unknown_type_s_returning_raw_dat);
  Py_CLEAR(clear_module_state->__pyx_n_u_VEL);
  Py_CLEAR(clear_module_state->__pyx_n_u_VEL2);
  Py_CLEAR(clear_module_state->__pyx_n_u_VELC);
  Py_CLEAR(clear_module_state->__pyx_n_u_VELC2);
  Py_CLEAR(clear_module_state->__pyx_n_u_VIL2);
  Py_CLEAR(clear_module_state->__pyx_n_u_VILD16);
  Py_CLEAR(clear_module_state->__pyx_n_u_VIR16);
  Py_CLEAR(clear_module_state->__pyx_n_u_VVEL2);
  Py_CLEAR(clear_module_state->__pyx_n_u_WIDTH);
  Py_CLEAR(clear_module_state->__pyx_n_u_WIDTH2);
  Py_CLEAR(clear_module_state->__pyx_n_u_XHDR);
  Py_CLEAR(clear_module_state->__pyx_n_s_YMDS_TIME);
  Py_CLEAR(clear_module_state->__pyx_n_u_ZDR);
  Py_CLEAR(clear_module_state->__pyx_n_u_ZDR2);
  Py_CLEAR(clear_module_state->__pyx_n_u_ZDRC);
  Py_CLEAR(clear_module_state->__pyx_n_u_ZDRC2);
  Py_CLEAR(clear_module_state->__pyx_n_s__149);
  Py_CLEAR(clear_module_state->__pyx_kp_u__15);
  Py_CLEAR(clear_module_state->__pyx_n_s__19);
  Py_CLEAR(clear_module_state->__pyx_n_u_agc_feedback_code);
  Py_CLEAR(clear_module_state->__pyx_n_u_altitude_radar);
  Py_CLEAR(clear_module_state->__pyx_n_u_angular_resolution_desired);
  Py_CLEAR(clear_module_state->__pyx_n_u_antenna_offset_bow);
  Py_CLEAR(clear_module_state->__pyx_n_u_antenna_offset_starboard);
  Py_CLEAR(clear_module_state->__pyx_n_u_antenna_offset_up);
  Py_CLEAR(clear_module_state->__pyx_n_u_antenna_scan_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_append);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_u_attenuation);
  Py_CLEAR(clear_module_state->__pyx_n_s_az0);
  Py_CLEAR(clear_module_state->__pyx_n_s_az1);
  Py_CLEAR(clear_module_state->__pyx_n_u_azimuth_0);
  Py_CLEAR(clear_module_state->__pyx_n_u_azimuth_1);
  Py_CLEAR(clear_module_state->__pyx_n_u_azimuth_list);
  Py_CLEAR(clear_module_state->__pyx_n_u_b);
  Py_CLEAR(clear_module_state->__pyx_n_s_bin2);
  Py_CLEAR(clear_module_state->__pyx_n_s_bin2_to_angle);
  Py_CLEAR(clear_module_state->__pyx_n_s_bin4);
  Py_CLEAR(clear_module_state->__pyx_n_s_bin4_to_angle);
  Py_CLEAR(clear_module_state->__pyx_n_s_bit);
  Py_CLEAR(clear_module_state->__pyx_n_u_bit_per_bin);
  Py_CLEAR(clear_module_state->__pyx_n_u_bool);
  Py_CLEAR(clear_module_state->__pyx_n_u_bytes_in_structure);
  Py_CLEAR(clear_module_state->__pyx_n_u_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_close);
  Py_CLEAR(clear_module_state->__pyx_n_u_clutter_correction_threshold);
  Py_CLEAR(clear_module_state->__pyx_n_u_clutter_filter);
  Py_CLEAR(clear_module_state->__pyx_n_u_clutter_filter_name);
  Py_CLEAR(clear_module_state->__pyx_n_u_cluttermap);
  Py_CLEAR(clear_module_state->__pyx_n_u_cluttermap_flag);
  Py_CLEAR(clear_module_state->__pyx_n_u_color_scale_def);
  Py_CLEAR(clear_module_state->__pyx_n_u_comment_bytes);
  Py_CLEAR(clear_module_state->__pyx_n_u_comments);
  Py_CLEAR(clear_module_state->__pyx_n_u_configuration_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_convert_sigmet_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_copy);
  Py_CLEAR(clear_module_state->__pyx_n_u_corrected_reflectivity_threshold);
  Py_CLEAR(clear_module_state->__pyx_n_u_current_data_type_mask);
  Py_CLEAR(clear_module_state->__pyx_n_u_custom_ray_header_name);
  Py_CLEAR(clear_module_state->__pyx_n_u_customer_storage);
  Py_CLEAR(clear_module_state->__pyx_n_s_d);
  Py_CLEAR(clear_module_state->__pyx_n_u_d);
  Py_CLEAR(clear_module_state->__pyx_n_s_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_data_type);
  Py_CLEAR(clear_module_state->__pyx_n_u_data_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_data_type_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_data_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_data_types_from_mask);
  Py_CLEAR(clear_module_state->__pyx_n_s_datetime);
  Py_CLEAR(clear_module_state->__pyx_n_u_day);
  Py_CLEAR(clear_module_state->__pyx_n_s_debug);
  Py_CLEAR(clear_module_state->__pyx_n_s_determine_data_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_dic);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_u_dsp_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_dtype);
  Py_CLEAR(clear_module_state->__pyx_n_u_dual_prf_delay);
  Py_CLEAR(clear_module_state->__pyx_n_u_earth_radius);
  Py_CLEAR(clear_module_state->__pyx_n_s_el0);
  Py_CLEAR(clear_module_state->__pyx_n_s_el1);
  Py_CLEAR(clear_module_state->__pyx_n_u_elevation_0);
  Py_CLEAR(clear_module_state->__pyx_n_u_elevation_1);
  Py_CLEAR(clear_module_state->__pyx_n_u_elevation_list);
  Py_CLEAR(clear_module_state->__pyx_n_s_empty);
  Py_CLEAR(clear_module_state->__pyx_n_s_empty_like);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_enumerate);
  Py_CLEAR(clear_module_state->__pyx_n_s_exp);
  Py_CLEAR(clear_module_state->__pyx_n_u_extended_header_type);
  Py_CLEAR(clear_module_state->__pyx_n_u_extended_product_header_offset);
  Py_CLEAR(clear_module_state->__pyx_n_u_extended_ray_header_bytes);
  Py_CLEAR(clear_module_state->__pyx_n_u_f);
  Py_CLEAR(clear_module_state->__pyx_n_u_fault_status);
  Py_CLEAR(clear_module_state->__pyx_n_u_file_ingest_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_filename);
  Py_CLEAR(clear_module_state->__pyx_n_u_filename);
  Py_CLEAR(clear_module_state->__pyx_n_s_fill_value);
  Py_CLEAR(clear_module_state->__pyx_n_u_first_azimuth);
  Py_CLEAR(clear_module_state->__pyx_n_u_first_bin_range);
  Py_CLEAR(clear_module_state->__pyx_n_u_first_elevation);
  Py_CLEAR(clear_module_state->__pyx_n_u_first_ray_index);
  Py_CLEAR(clear_module_state->__pyx_n_u_first_ray_offset);
  Py_CLEAR(clear_module_state->__pyx_n_u_fixed_angle);
  Py_CLEAR(clear_module_state->__pyx_n_u_flag);
  Py_CLEAR(clear_module_state->__pyx_n_u_flags);
  Py_CLEAR(clear_module_state->__pyx_n_u_flags2);
  Py_CLEAR(clear_module_state->__pyx_n_u_float32);
  Py_CLEAR(clear_module_state->__pyx_n_s_fmt);
  Py_CLEAR(clear_module_state->__pyx_n_u_format_version);
  Py_CLEAR(clear_module_state->__pyx_n_s_frombuffer);
  Py_CLEAR(clear_module_state->__pyx_n_s_full_xhdr);
  Py_CLEAR(clear_module_state->__pyx_n_u_gain_control_flag);
  Py_CLEAR(clear_module_state->__pyx_n_u_gas_attenuation);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_u_generation_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_sweep);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_n_u_gmt_offset_minutes_local);
  Py_CLEAR(clear_module_state->__pyx_n_u_gmt_offset_minutes_standard);
  Py_CLEAR(clear_module_state->__pyx_n_u_gparam_bytes);
  Py_CLEAR(clear_module_state->__pyx_n_u_gparm);
  Py_CLEAR(clear_module_state->__pyx_n_u_ground_height);
  Py_CLEAR(clear_module_state->__pyx_n_u_h);
  Py_CLEAR(clear_module_state->__pyx_n_u_hardware_site);
  Py_CLEAR(clear_module_state->__pyx_n_s_header_dic);
  Py_CLEAR(clear_module_state->__pyx_n_u_height_radar);
  Py_CLEAR(clear_module_state->__pyx_n_u_height_site);
  Py_CLEAR(clear_module_state->__pyx_n_u_horizontal_beamwidth);
  Py_CLEAR(clear_module_state->__pyx_n_u_horizontal_calibration_i0);
  Py_CLEAR(clear_module_state->__pyx_n_u_horizontal_calibration_noise);
  Py_CLEAR(clear_module_state->__pyx_n_u_horizontal_current_noise);
  Py_CLEAR(clear_module_state->__pyx_n_u_horizontal_i0_calibration);
  Py_CLEAR(clear_module_state->__pyx_n_u_horizontal_noise_calibration);
  Py_CLEAR(clear_module_state->__pyx_n_u_horizontal_radar_constant);
  Py_CLEAR(clear_module_state->__pyx_n_s_i);
  Py_CLEAR(clear_module_state->__pyx_n_u_i);
  Py_CLEAR(clear_module_state->__pyx_n_u_i4);
  Py_CLEAR(clear_module_state->__pyx_n_u_icolcnt);
  Py_CLEAR(clear_module_state->__pyx_n_s_idh);
  Py_CLEAR(clear_module_state->__pyx_n_u_iflags);
  Py_CLEAR(clear_module_state->__pyx_n_u_ilevel_seams);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_ingest_configuration);
  Py_CLEAR(clear_module_state->__pyx_n_u_ingest_configuration);
  Py_CLEAR(clear_module_state->__pyx_n_s_ingest_data_hdrs);
  Py_CLEAR(clear_module_state->__pyx_n_s_ingest_data_headers);
  Py_CLEAR(clear_module_state->__pyx_n_u_ingest_hardware_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_ingest_header);
  Py_CLEAR(clear_module_state->__pyx_n_u_ingest_iris_version);
  Py_CLEAR(clear_module_state->__pyx_n_u_ingest_site_name);
  Py_CLEAR(clear_module_state->__pyx_n_u_ingest_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_u_input_data_type);
  Py_CLEAR(clear_module_state->__pyx_n_u_input_mask);
  Py_CLEAR(clear_module_state->__pyx_n_u_int16);
  Py_CLEAR(clear_module_state->__pyx_n_u_int32);
  Py_CLEAR(clear_module_state->__pyx_n_u_inverse_flatting);
  Py_CLEAR(clear_module_state->__pyx_n_u_iris_version);
  Py_CLEAR(clear_module_state->__pyx_n_u_iris_version_created);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_bit_set);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_u_iset_and_scale);
  Py_CLEAR(clear_module_state->__pyx_n_u_istart);
  Py_CLEAR(clear_module_state->__pyx_n_u_istep);
  Py_CLEAR(clear_module_state->__pyx_n_s_j);
  Py_CLEAR(clear_module_state->__pyx_n_s_k);
  Py_CLEAR(clear_module_state->__pyx_n_s_key);
  Py_CLEAR(clear_module_state->__pyx_n_s_l);
  Py_CLEAR(clear_module_state->__pyx_n_u_last_bin_range);
  Py_CLEAR(clear_module_state->__pyx_n_u_last_run_day);
  Py_CLEAR(clear_module_state->__pyx_n_u_last_run_time);
  Py_CLEAR(clear_module_state->__pyx_n_u_latitude);
  Py_CLEAR(clear_module_state->__pyx_n_u_latitude_projection);
  Py_CLEAR(clear_module_state->__pyx_n_u_latitude_radar);
  Py_CLEAR(clear_module_state->__pyx_n_u_ldr_bias);
  Py_CLEAR(clear_module_state->__pyx_n_u_ldr_offset);
  Py_CLEAR(clear_module_state->__pyx_n_s_lead_record);
  Py_CLEAR(clear_module_state->__pyx_n_u_left_azimuth_limit);
  Py_CLEAR(clear_module_state->__pyx_n_s_like_dbt);
  Py_CLEAR(clear_module_state->__pyx_n_s_like_dbt2);
  Py_CLEAR(clear_module_state->__pyx_n_s_like_sqi);
  Py_CLEAR(clear_module_state->__pyx_n_s_like_sqi2);
  Py_CLEAR(clear_module_state->__pyx_n_u_linear_filter_first_bin);
  Py_CLEAR(clear_module_state->__pyx_n_u_local_timezone);
  Py_CLEAR(clear_module_state->__pyx_n_u_log_filter_first_bin);
  Py_CLEAR(clear_module_state->__pyx_n_u_longitude);
  Py_CLEAR(clear_module_state->__pyx_n_u_longitude_projection);
  Py_CLEAR(clear_module_state->__pyx_n_u_longitude_radar);
  Py_CLEAR(clear_module_state->__pyx_n_u_low_prf_factional);
  Py_CLEAR(clear_module_state->__pyx_n_u_low_prf_hz);
  Py_CLEAR(clear_module_state->__pyx_n_u_low_prf_range_averaging);
  Py_CLEAR(clear_module_state->__pyx_n_u_low_prf_sample_size);
  Py_CLEAR(clear_module_state->__pyx_n_u_lower_elevation_limit);
  Py_CLEAR(clear_module_state->__pyx_n_s_ma);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_u_major_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_mask);
  Py_CLEAR(clear_module_state->__pyx_n_u_mask_word_0);
  Py_CLEAR(clear_module_state->__pyx_n_u_mask_word_1);
  Py_CLEAR(clear_module_state->__pyx_n_u_mask_word_2);
  Py_CLEAR(clear_module_state->__pyx_n_u_mask_word_3);
  Py_CLEAR(clear_module_state->__pyx_n_u_mask_word_4);
  Py_CLEAR(clear_module_state->__pyx_n_s_masked_array);
  Py_CLEAR(clear_module_state->__pyx_n_u_maximum_range);
  Py_CLEAR(clear_module_state->__pyx_n_u_mean_wind_direction);
  Py_CLEAR(clear_module_state->__pyx_n_u_mean_wind_speed);
  Py_CLEAR(clear_module_state->__pyx_n_u_melting_layer);
  Py_CLEAR(clear_module_state->__pyx_n_u_melting_level);
  Py_CLEAR(clear_module_state->__pyx_n_s_mess);
  Py_CLEAR(clear_module_state->__pyx_n_s_metadata);
  Py_CLEAR(clear_module_state->__pyx_n_u_milliseconds);
  Py_CLEAR(clear_module_state->__pyx_n_u_minor_task_suffix);
  Py_CLEAR(clear_module_state->__pyx_n_s_mod);
  Py_CLEAR(clear_module_state->__pyx_n_u_month);
  Py_CLEAR(clear_module_state->__pyx_n_s_multi_prf_flag);
  Py_CLEAR(clear_module_state->__pyx_n_u_multi_prf_flag);
  Py_CLEAR(clear_module_state->__pyx_n_s_multiplier);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_nbin);
  Py_CLEAR(clear_module_state->__pyx_n_s_nbins);
  Py_CLEAR(clear_module_state->__pyx_n_u_nbins);
  Py_CLEAR(clear_module_state->__pyx_n_s_ndata);
  Py_CLEAR(clear_module_state->__pyx_n_s_ndata_types);
  Py_CLEAR(clear_module_state->__pyx_n_u_nexrad_clutter_skip);
  Py_CLEAR(clear_module_state->__pyx_n_u_nexrad_clutter_threshold);
  Py_CLEAR(clear_module_state->__pyx_n_s_np);
  Py_CLEAR(clear_module_state->__pyx_n_s_nray_data_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_nrays);
  Py_CLEAR(clear_module_state->__pyx_n_s_nsweeps);
  Py_CLEAR(clear_module_state->__pyx_n_s_number);
  Py_CLEAR(clear_module_state->__pyx_n_u_number_bins);
  Py_CLEAR(clear_module_state->__pyx_n_u_number_elements);
  Py_CLEAR(clear_module_state->__pyx_n_u_number_files);
  Py_CLEAR(clear_module_state->__pyx_n_u_number_ingest);
  Py_CLEAR(clear_module_state->__pyx_n_u_number_input_bins);
  Py_CLEAR(clear_module_state->__pyx_n_u_number_linear_filter);
  Py_CLEAR(clear_module_state->__pyx_n_u_number_log_filter);
  Py_CLEAR(clear_module_state->__pyx_n_u_number_output_bins);
  Py_CLEAR(clear_module_state->__pyx_n_u_number_rays_file_actual);
  Py_CLEAR(clear_module_state->__pyx_n_u_number_rays_file_expected);
  Py_CLEAR(clear_module_state->__pyx_n_u_number_rays_sweep);
  Py_CLEAR(clear_module_state->__pyx_n_u_number_sweeps);
  Py_CLEAR(clear_module_state->__pyx_n_u_number_sweeps_completed);
  Py_CLEAR(clear_module_state->__pyx_n_u_number_task_config_table);
  Py_CLEAR(clear_module_state->__pyx_n_u_number_tasks);
  Py_CLEAR(clear_module_state->__pyx_n_s_numpy);
  Py_CLEAR(clear_module_state->__pyx_kp_u_numpy__core_multiarray_failed_to);
  Py_CLEAR(clear_module_state->__pyx_kp_u_numpy__core_umath_failed_to_impo);
  Py_CLEAR(clear_module_state->__pyx_n_s_nyquist);
  Py_CLEAR(clear_module_state->__pyx_n_s_offset);
  Py_CLEAR(clear_module_state->__pyx_n_s_ones);
  Py_CLEAR(clear_module_state->__pyx_n_s_open);
  Py_CLEAR(clear_module_state->__pyx_n_u_original_data_type_mask);
  Py_CLEAR(clear_module_state->__pyx_n_s_out);
  Py_CLEAR(clear_module_state->__pyx_n_s_parse_ray_headers);
  Py_CLEAR(clear_module_state->__pyx_n_u_playback_flag);
  Py_CLEAR(clear_module_state->__pyx_n_u_playback_version);
  Py_CLEAR(clear_module_state->__pyx_n_u_polarization);
  Py_CLEAR(clear_module_state->__pyx_n_u_polarization_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_power);
  Py_CLEAR(clear_module_state->__pyx_n_u_power_threshold);
  Py_CLEAR(clear_module_state->__pyx_n_u_prf);
  Py_CLEAR(clear_module_state->__pyx_n_s_prf_flag);
  Py_CLEAR(clear_module_state->__pyx_n_u_prf_flag);
  Py_CLEAR(clear_module_state->__pyx_n_s_print);
  Py_CLEAR(clear_module_state->__pyx_n_s_product_config);
  Py_CLEAR(clear_module_state->__pyx_n_u_product_configuration);
  Py_CLEAR(clear_module_state->__pyx_n_s_product_end);
  Py_CLEAR(clear_module_state->__pyx_n_u_product_end);
  Py_CLEAR(clear_module_state->__pyx_n_s_product_hdr);
  Py_CLEAR(clear_module_state->__pyx_n_u_product_name);
  Py_CLEAR(clear_module_state->__pyx_n_u_product_sequence_number);
  Py_CLEAR(clear_module_state->__pyx_n_u_product_specific_bytes);
  Py_CLEAR(clear_module_state->__pyx_n_u_product_type_code);
  Py_CLEAR(clear_module_state->__pyx_n_u_projection_name);
  Py_CLEAR(clear_module_state->__pyx_n_u_projection_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_prt_value);
  Py_CLEAR(clear_module_state->__pyx_n_u_pulse_width);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyart_io__sigmetfile);
  Py_CLEAR(clear_module_state->__pyx_kp_s_pyart_io__sigmetfile_pyx);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_u_radar_height);
  Py_CLEAR(clear_module_state->__pyx_n_u_radar_height_above_reference);
  Py_CLEAR(clear_module_state->__pyx_n_u_radial_smoother);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_u_range_bin_averaging_flag);
  Py_CLEAR(clear_module_state->__pyx_n_s_raw_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_raw_prod_bhdr);
  Py_CLEAR(clear_module_state->__pyx_n_s_raw_sweep_data);
  Py_CLEAR(clear_module_state->__pyx_n_u_ray_header_bytes);
  Py_CLEAR(clear_module_state->__pyx_n_u_ray_header_mask);
  Py_CLEAR(clear_module_state->__pyx_n_s_ray_headers);
  Py_CLEAR(clear_module_state->__pyx_n_s_ray_i);
  Py_CLEAR(clear_module_state->__pyx_n_s_ray_nbins);
  Py_CLEAR(clear_module_state->__pyx_n_u_ray_number);
  Py_CLEAR(clear_module_state->__pyx_n_s_ray_time);
  Py_CLEAR(clear_module_state->__pyx_n_u_rb);
  Py_CLEAR(clear_module_state->__pyx_n_s_read);
  Py_CLEAR(clear_module_state->__pyx_n_u_read);
  Py_CLEAR(clear_module_state->__pyx_n_s_read_data);
  Py_CLEAR(clear_module_state->__pyx_n_u_reciever_bandwidth);
  Py_CLEAR(clear_module_state->__pyx_n_s_record);
  Py_CLEAR(clear_module_state->__pyx_n_u_record_number);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_u_reflectivity_calibration);
  Py_CLEAR(clear_module_state->__pyx_n_u_reflectivity_noise_threshold);
  Py_CLEAR(clear_module_state->__pyx_n_u_reflectivity_slope);
  Py_CLEAR(clear_module_state->__pyx_n_u_reflectivity_unfolding_threshold);
  Py_CLEAR(clear_module_state->__pyx_n_u_reserved);
  Py_CLEAR(clear_module_state->__pyx_n_s_reshape);
  Py_CLEAR(clear_module_state->__pyx_n_u_resolution_rays);
  Py_CLEAR(clear_module_state->__pyx_n_u_right_azimuth_limit);
  Py_CLEAR(clear_module_state->__pyx_n_u_sample_size);
  Py_CLEAR(clear_module_state->__pyx_n_u_samples_used);
  Py_CLEAR(clear_module_state->__pyx_n_u_scheduling_code);
  Py_CLEAR(clear_module_state->__pyx_n_u_seconds);
  Py_CLEAR(clear_module_state->__pyx_n_u_seconds_between_runs);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_kp_s_self__rbuf_p_cannot_be_converted);
  Py_CLEAR(clear_module_state->__pyx_kp_u_self__rbuf_pos_is);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_shape);
  Py_CLEAR(clear_module_state->__pyx_n_s_shrink);
  Py_CLEAR(clear_module_state->__pyx_n_u_signal_processor_type);
  Py_CLEAR(clear_module_state->__pyx_n_u_site_name);
  Py_CLEAR(clear_module_state->__pyx_n_u_skip_time);
  Py_CLEAR(clear_module_state->__pyx_n_u_spare_0);
  Py_CLEAR(clear_module_state->__pyx_n_u_spare_1);
  Py_CLEAR(clear_module_state->__pyx_n_u_spare_2);
  Py_CLEAR(clear_module_state->__pyx_n_u_spare_3);
  Py_CLEAR(clear_module_state->__pyx_n_u_spare_4);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_n_u_sqi_threshold);
  Py_CLEAR(clear_module_state->__pyx_n_s_sqrt);
  Py_CLEAR(clear_module_state->__pyx_n_u_standard_parallel_1);
  Py_CLEAR(clear_module_state->__pyx_n_u_standard_parallel_2);
  Py_CLEAR(clear_module_state->__pyx_n_u_start_first_section_flag);
  Py_CLEAR(clear_module_state->__pyx_n_u_start_first_sector_flag);
  Py_CLEAR(clear_module_state->__pyx_n_u_start_time);
  Py_CLEAR(clear_module_state->__pyx_n_u_step_input_bins);
  Py_CLEAR(clear_module_state->__pyx_n_u_step_output_bins);
  Py_CLEAR(clear_module_state->__pyx_n_u_stop_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_string);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_struct);
  Py_CLEAR(clear_module_state->__pyx_n_s_structure);
  Py_CLEAR(clear_module_state->__pyx_n_u_structure_header);
  Py_CLEAR(clear_module_state->__pyx_n_u_structure_identifier);
  Py_CLEAR(clear_module_state->__pyx_n_s_sum);
  Py_CLEAR(clear_module_state->__pyx_n_s_sweep_data);
  Py_CLEAR(clear_module_state->__pyx_n_u_sweep_ingest_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_sweep_metadata);
  Py_CLEAR(clear_module_state->__pyx_n_u_sweep_number);
  Py_CLEAR(clear_module_state->__pyx_n_u_sweep_start_time);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_calib_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_task_config);
  Py_CLEAR(clear_module_state->__pyx_n_s_task_configuration);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_configuration);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_configuration_file_name);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_data_time);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_description);
  Py_CLEAR(clear_module_state->__pyx_n_s_task_dsp_info);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_dsp_info);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_dsp_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_task_end_info);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_end_info);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_major_number);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_minor_number);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_misc_info);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_name);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_range_info);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_scan_info);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_scan_type_scan_info);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_sched_info);
  Py_CLEAR(clear_module_state->__pyx_n_u_task_state);
  Py_CLEAR(clear_module_state->__pyx_n_u_tcf_cal_flags_1);
  Py_CLEAR(clear_module_state->__pyx_n_u_tcf_cal_flags_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_temp);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_time);
  Py_CLEAR(clear_module_state->__pyx_n_u_time);
  Py_CLEAR(clear_module_state->__pyx_n_u_time_used_last_run);
  Py_CLEAR(clear_module_state->__pyx_n_u_times_run);
  Py_CLEAR(clear_module_state->__pyx_n_u_total_size);
  Py_CLEAR(clear_module_state->__pyx_n_u_tr_serial_number);
  Py_CLEAR(clear_module_state->__pyx_n_u_tranmitter_phase_sequence);
  Py_CLEAR(clear_module_state->__pyx_n_u_transmit_power);
  Py_CLEAR(clear_module_state->__pyx_n_u_trigger_rate);
  Py_CLEAR(clear_module_state->__pyx_n_u_trucation_height);
  Py_CLEAR(clear_module_state->__pyx_n_u_truncation_height);
  Py_CLEAR(clear_module_state->__pyx_n_u_tz_name);
  Py_CLEAR(clear_module_state->__pyx_n_u_uint16);
  Py_CLEAR(clear_module_state->__pyx_n_s_uint8);
  Py_CLEAR(clear_module_state->__pyx_n_u_uncorrected_reflectivity_thresho);
  Py_CLEAR(clear_module_state->__pyx_n_s_unpack);
  Py_CLEAR(clear_module_state->__pyx_n_s_unpack_ingest_data_header);
  Py_CLEAR(clear_module_state->__pyx_n_s_unpack_ingest_data_headers);
  Py_CLEAR(clear_module_state->__pyx_n_s_unpack_ingest_header);
  Py_CLEAR(clear_module_state->__pyx_n_s_unpack_key);
  Py_CLEAR(clear_module_state->__pyx_n_s_unpack_product_hdr);
  Py_CLEAR(clear_module_state->__pyx_n_s_unpack_raw_prod_bhdr);
  Py_CLEAR(clear_module_state->__pyx_n_s_unpack_structure);
  Py_CLEAR(clear_module_state->__pyx_n_u_upper_elevation_limit);
  Py_CLEAR(clear_module_state->__pyx_n_u_variable_range_bin_flag);
  Py_CLEAR(clear_module_state->__pyx_n_u_velocity_east);
  Py_CLEAR(clear_module_state->__pyx_n_u_velocity_north);
  Py_CLEAR(clear_module_state->__pyx_n_u_velocity_threshold_flags);
  Py_CLEAR(clear_module_state->__pyx_n_u_velocity_unfolding_threshold);
  Py_CLEAR(clear_module_state->__pyx_n_u_velocity_up);
  Py_CLEAR(clear_module_state->__pyx_n_u_vertical_beamwidth);
  Py_CLEAR(clear_module_state->__pyx_n_u_vertical_current_noise);
  Py_CLEAR(clear_module_state->__pyx_n_u_vertical_i0_calibration);
  Py_CLEAR(clear_module_state->__pyx_n_u_vertical_noise_calibration);
  Py_CLEAR(clear_module_state->__pyx_n_u_vertical_radar_constant);
  Py_CLEAR(clear_module_state->__pyx_n_s_view);
  Py_CLEAR(clear_module_state->__pyx_n_u_volume_scan_start_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_warn);
  Py_CLEAR(clear_module_state->__pyx_n_s_warnings);
  Py_CLEAR(clear_module_state->__pyx_n_u_wavelength);
  Py_CLEAR(clear_module_state->__pyx_n_s_wavelength_cm);
  Py_CLEAR(clear_module_state->__pyx_n_u_width_threshold_flags);
  Py_CLEAR(clear_module_state->__pyx_n_u_width_unfolding_threshold);
  Py_CLEAR(clear_module_state->__pyx_n_s_word0);
  Py_CLEAR(clear_module_state->__pyx_n_s_word1);
  Py_CLEAR(clear_module_state->__pyx_n_s_word2);
  Py_CLEAR(clear_module_state->__pyx_n_s_word3);
  Py_CLEAR(clear_module_state->__pyx_n_u_x_location);
  Py_CLEAR(clear_module_state->__pyx_n_u_x_scale);
  Py_CLEAR(clear_module_state->__pyx_n_u_x_size);
  Py_CLEAR(clear_module_state->__pyx_n_u_x_smoother);
  Py_CLEAR(clear_module_state->__pyx_n_s_xrange);
  Py_CLEAR(clear_module_state->__pyx_n_u_y_location);
  Py_CLEAR(clear_module_state->__pyx_n_u_y_scale);
  Py_CLEAR(clear_module_state->__pyx_n_u_y_size);
  Py_CLEAR(clear_module_state->__pyx_n_u_y_smoother);
  Py_CLEAR(clear_module_state->__pyx_n_u_year);
  Py_CLEAR(clear_module_state->__pyx_n_u_z_location);
  Py_CLEAR(clear_module_state->__pyx_n_u_z_scale);
  Py_CLEAR(clear_module_state->__pyx_n_u_z_size);
  Py_CLEAR(clear_module_state->__pyx_n_u_zdr_bias);
  Py_CLEAR(clear_module_state->__pyx_n_u_zdr_offset);
  Py_CLEAR(clear_module_state->__pyx_n_u_zdr_threshold_flags);
  Py_CLEAR(clear_module_state->__pyx_n_s_zeros_like);
  Py_CLEAR(clear_module_state->__pyx_n_s_zip);
  Py_CLEAR(clear_module_state->__pyx_n_u_zr_constant);
  Py_CLEAR(clear_module_state->__pyx_n_u_zr_exponent);
  Py_CLEAR(clear_module_state->__pyx_float_1_);
  Py_CLEAR(clear_module_state->__pyx_float_2_);
  Py_CLEAR(clear_module_state->__pyx_float_16_);
  Py_CLEAR(clear_module_state->__pyx_float_64_);
  Py_CLEAR(clear_module_state->__pyx_float_75_);
  Py_CLEAR(clear_module_state->__pyx_float_0_25);
  Py_CLEAR(clear_module_state->__pyx_float_100_);
  Py_CLEAR(clear_module_state->__pyx_float_126_);
  Py_CLEAR(clear_module_state->__pyx_float_127_);
  Py_CLEAR(clear_module_state->__pyx_float_128_);
  Py_CLEAR(clear_module_state->__pyx_float_129_);
  Py_CLEAR(clear_module_state->__pyx_float_180_);
  Py_CLEAR(clear_module_state->__pyx_float_253_);
  Py_CLEAR(clear_module_state->__pyx_float_254_);
  Py_CLEAR(clear_module_state->__pyx_float_256_);
  Py_CLEAR(clear_module_state->__pyx_float_360_);
  Py_CLEAR(clear_module_state->__pyx_float_600_);
  Py_CLEAR(clear_module_state->__pyx_float_100_0);
  Py_CLEAR(clear_module_state->__pyx_float_32768_);
  Py_CLEAR(clear_module_state->__pyx_float_65533_);
  Py_CLEAR(clear_module_state->__pyx_float_65534_);
  Py_CLEAR(clear_module_state->__pyx_float_neg_0_25);
  Py_CLEAR(clear_module_state->__pyx_float_neg_9999_0);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_2);
  Py_CLEAR(clear_module_state->__pyx_int_3);
  Py_CLEAR(clear_module_state->__pyx_int_4);
  Py_CLEAR(clear_module_state->__pyx_int_5);
  Py_CLEAR(clear_module_state->__pyx_int_6);
  Py_CLEAR(clear_module_state->__pyx_int_7);
  Py_CLEAR(clear_module_state->__pyx_int_8);
  Py_CLEAR(clear_module_state->__pyx_int_9);
  Py_CLEAR(clear_module_state->__pyx_int_10);
  Py_CLEAR(clear_module_state->__pyx_int_11);
  Py_CLEAR(clear_module_state->__pyx_int_12);
  Py_CLEAR(clear_module_state->__pyx_int_13);
  Py_CLEAR(clear_module_state->__pyx_int_14);
  Py_CLEAR(clear_module_state->__pyx_int_15);
  Py_CLEAR(clear_module_state->__pyx_int_16);
  Py_CLEAR(clear_module_state->__pyx_int_17);
  Py_CLEAR(clear_module_state->__pyx_int_18);
  Py_CLEAR(clear_module_state->__pyx_int_19);
  Py_CLEAR(clear_module_state->__pyx_int_20);
  Py_CLEAR(clear_module_state->__pyx_int_21);
  Py_CLEAR(clear_module_state->__pyx_int_22);
  Py_CLEAR(clear_module_state->__pyx_int_23);
  Py_CLEAR(clear_module_state->__pyx_int_24);
  Py_CLEAR(clear_module_state->__pyx_int_25);
  Py_CLEAR(clear_module_state->__pyx_int_26);
  Py_CLEAR(clear_module_state->__pyx_int_27);
  Py_CLEAR(clear_module_state->__pyx_int_28);
  Py_CLEAR(clear_module_state->__pyx_int_29);
  Py_CLEAR(clear_module_state->__pyx_int_30);
  Py_CLEAR(clear_module_state->__pyx_int_31);
  Py_CLEAR(clear_module_state->__pyx_int_32);
  Py_CLEAR(clear_module_state->__pyx_int_33);
  Py_CLEAR(clear_module_state->__pyx_int_34);
  Py_CLEAR(clear_module_state->__pyx_int_35);
  Py_CLEAR(clear_module_state->__pyx_int_36);
  Py_CLEAR(clear_module_state->__pyx_int_37);
  Py_CLEAR(clear_module_state->__pyx_int_38);
  Py_CLEAR(clear_module_state->__pyx_int_39);
  Py_CLEAR(clear_module_state->__pyx_int_40);
  Py_CLEAR(clear_module_state->__pyx_int_41);
  Py_CLEAR(clear_module_state->__pyx_int_42);
  Py_CLEAR(clear_module_state->__pyx_int_43);
  Py_CLEAR(clear_module_state->__pyx_int_44);
  Py_CLEAR(clear_module_state->__pyx_int_45);
  Py_CLEAR(clear_module_state->__pyx_int_46);
  Py_CLEAR(clear_module_state->__pyx_int_47);
  Py_CLEAR(clear_module_state->__pyx_int_48);
  Py_CLEAR(clear_module_state->__pyx_int_49);
  Py_CLEAR(clear_module_state->__pyx_int_50);
  Py_CLEAR(clear_module_state->__pyx_int_51);
  Py_CLEAR(clear_module_state->__pyx_int_52);
  Py_CLEAR(clear_module_state->__pyx_int_53);
  Py_CLEAR(clear_module_state->__pyx_int_54);
  Py_CLEAR(clear_module_state->__pyx_int_55);
  Py_CLEAR(clear_module_state->__pyx_int_56);
  Py_CLEAR(clear_module_state->__pyx_int_57);
  Py_CLEAR(clear_module_state->__pyx_int_58);
  Py_CLEAR(clear_module_state->__pyx_int_59);
  Py_CLEAR(clear_module_state->__pyx_int_60);
  Py_CLEAR(clear_module_state->__pyx_int_61);
  Py_CLEAR(clear_module_state->__pyx_int_62);
  Py_CLEAR(clear_module_state->__pyx_int_63);
  Py_CLEAR(clear_module_state->__pyx_int_64);
  Py_CLEAR(clear_module_state->__pyx_int_65);
  Py_CLEAR(clear_module_state->__pyx_int_66);
  Py_CLEAR(clear_module_state->__pyx_int_67);
  Py_CLEAR(clear_module_state->__pyx_int_68);
  Py_CLEAR(clear_module_state->__pyx_int_69);
  Py_CLEAR(clear_module_state->__pyx_int_70);
  Py_CLEAR(clear_module_state->__pyx_int_71);
  Py_CLEAR(clear_module_state->__pyx_int_72);
  Py_CLEAR(clear_module_state->__pyx_int_73);
  Py_CLEAR(clear_module_state->__pyx_int_74);
  Py_CLEAR(clear_module_state->__pyx_int_75);
  Py_CLEAR(clear_module_state->__pyx_int_76);
  Py_CLEAR(clear_module_state->__pyx_int_77);
  Py_CLEAR(clear_module_state->__pyx_int_78);
  Py_CLEAR(clear_module_state->__pyx_int_79);
  Py_CLEAR(clear_module_state->__pyx_int_80);
  Py_CLEAR(clear_module_state->__pyx_int_81);
  Py_CLEAR(clear_module_state->__pyx_int_82);
  Py_CLEAR(clear_module_state->__pyx_int_83);
  Py_CLEAR(clear_module_state->__pyx_int_84);
  Py_CLEAR(clear_module_state->__pyx_int_85);
  Py_CLEAR(clear_module_state->__pyx_int_86);
  Py_CLEAR(clear_module_state->__pyx_int_87);
  Py_CLEAR(clear_module_state->__pyx_int_88);
  Py_CLEAR(clear_module_state->__pyx_int_89);
  Py_CLEAR(clear_module_state->__pyx_int_90);
  Py_CLEAR(clear_module_state->__pyx_int_91);
  Py_CLEAR(clear_module_state->__pyx_int_92);
  Py_CLEAR(clear_module_state->__pyx_int_93);
  Py_CLEAR(clear_module_state->__pyx_int_94);
  Py_CLEAR(clear_module_state->__pyx_int_95);
  Py_CLEAR(clear_module_state->__pyx_int_96);
  Py_CLEAR(clear_module_state->__pyx_int_97);
  Py_CLEAR(clear_module_state->__pyx_int_98);
  Py_CLEAR(clear_module_state->__pyx_int_99);
  Py_CLEAR(clear_module_state->__pyx_int_100);
  Py_CLEAR(clear_module_state->__pyx_int_101);
  Py_CLEAR(clear_module_state->__pyx_int_102);
  Py_CLEAR(clear_module_state->__pyx_int_103);
  Py_CLEAR(clear_module_state->__pyx_int_104);
  Py_CLEAR(clear_module_state->__pyx_int_105);
  Py_CLEAR(clear_module_state->__pyx_int_106);
  Py_CLEAR(clear_module_state->__pyx_int_107);
  Py_CLEAR(clear_module_state->__pyx_int_108);
  Py_CLEAR(clear_module_state->__pyx_int_109);
  Py_CLEAR(clear_module_state->__pyx_int_110);
  Py_CLEAR(clear_module_state->__pyx_int_111);
  Py_CLEAR(clear_module_state->__pyx_int_112);
  Py_CLEAR(clear_module_state->__pyx_int_113);
  Py_CLEAR(clear_module_state->__pyx_int_114);
  Py_CLEAR(clear_module_state->__pyx_int_115);
  Py_CLEAR(clear_module_state->__pyx_int_116);
  Py_CLEAR(clear_module_state->__pyx_int_117);
  Py_CLEAR(clear_module_state->__pyx_int_118);
  Py_CLEAR(clear_module_state->__pyx_int_119);
  Py_CLEAR(clear_module_state->__pyx_int_120);
  Py_CLEAR(clear_module_state->__pyx_int_121);
  Py_CLEAR(clear_module_state->__pyx_int_122);
  Py_CLEAR(clear_module_state->__pyx_int_123);
  Py_CLEAR(clear_module_state->__pyx_int_124);
  Py_CLEAR(clear_module_state->__pyx_int_125);
  Py_CLEAR(clear_module_state->__pyx_int_126);
  Py_CLEAR(clear_module_state->__pyx_int_127);
  Py_CLEAR(clear_module_state->__pyx_int_128);
  Py_CLEAR(clear_module_state->__pyx_int_255);
  Py_CLEAR(clear_module_state->__pyx_int_640);
  Py_CLEAR(clear_module_state->__pyx_int_4884);
  Py_CLEAR(clear_module_state->__pyx_int_6144);
  Py_CLEAR(clear_module_state->__pyx_int_65536);
  Py_CLEAR(clear_module_state->__pyx_int_4294967296);
  Py_CLEAR(clear_module_state->__pyx_int_neg_1);
  Py_CLEAR(clear_module_state->__pyx_tuple_);
  Py_CLEAR(clear_module_state->__pyx_slice__4);
  Py_CLEAR(clear_module_state->__pyx_slice__5);
  Py_CLEAR(clear_module_state->__pyx_tuple__2);
  Py_CLEAR(clear_module_state->__pyx_tuple__3);
  Py_CLEAR(clear_module_state->__pyx_tuple__6);
  Py_CLEAR(clear_module_state->__pyx_tuple__7);
  Py_CLEAR(clear_module_state->__pyx_tuple__8);
  Py_CLEAR(clear_module_state->__pyx_tuple__9);
  Py_CLEAR(clear_module_state->__pyx_slice__12);
  Py_CLEAR(clear_module_state->__pyx_slice__13);
  Py_CLEAR(clear_module_state->__pyx_slice__16);
  Py_CLEAR(clear_module_state->__pyx_slice__17);
  Py_CLEAR(clear_module_state->__pyx_slice__18);
  Py_CLEAR(clear_module_state->__pyx_tuple__10);
  Py_CLEAR(clear_module_state->__pyx_tuple__11);
  Py_CLEAR(clear_module_state->__pyx_tuple__14);
  Py_CLEAR(clear_module_state->__pyx_tuple__20);
  Py_CLEAR(clear_module_state->__pyx_tuple__22);
  Py_CLEAR(clear_module_state->__pyx_tuple__24);
  Py_CLEAR(clear_module_state->__pyx_tuple__26);
  Py_CLEAR(clear_module_state->__pyx_tuple__27);
  Py_CLEAR(clear_module_state->__pyx_tuple__29);
  Py_CLEAR(clear_module_state->__pyx_tuple__31);
  Py_CLEAR(clear_module_state->__pyx_tuple__33);
  Py_CLEAR(clear_module_state->__pyx_tuple__35);
  Py_CLEAR(clear_module_state->__pyx_tuple__37);
  Py_CLEAR(clear_module_state->__pyx_tuple__39);
  Py_CLEAR(clear_module_state->__pyx_tuple__41);
  Py_CLEAR(clear_module_state->__pyx_tuple__43);
  Py_CLEAR(clear_module_state->__pyx_tuple__45);
  Py_CLEAR(clear_module_state->__pyx_tuple__47);
  Py_CLEAR(clear_module_state->__pyx_tuple__49);
  Py_CLEAR(clear_module_state->__pyx_tuple__51);
  Py_CLEAR(clear_module_state->__pyx_tuple__53);
  Py_CLEAR(clear_module_state->__pyx_tuple__55);
  Py_CLEAR(clear_module_state->__pyx_tuple__57);
  Py_CLEAR(clear_module_state->__pyx_tuple__59);
  Py_CLEAR(clear_module_state->__pyx_tuple__60);
  Py_CLEAR(clear_module_state->__pyx_tuple__61);
  Py_CLEAR(clear_module_state->__pyx_tuple__62);
  Py_CLEAR(clear_module_state->__pyx_tuple__63);
  Py_CLEAR(clear_module_state->__pyx_tuple__64);
  Py_CLEAR(clear_module_state->__pyx_tuple__65);
  Py_CLEAR(clear_module_state->__pyx_tuple__66);
  Py_CLEAR(clear_module_state->__pyx_tuple__67);
  Py_CLEAR(clear_module_state->__pyx_tuple__68);
  Py_CLEAR(clear_module_state->__pyx_tuple__69);
  Py_CLEAR(clear_module_state->__pyx_tuple__70);
  Py_CLEAR(clear_module_state->__pyx_tuple__71);
  Py_CLEAR(clear_module_state->__pyx_tuple__72);
  Py_CLEAR(clear_module_state->__pyx_tuple__73);
  Py_CLEAR(clear_module_state->__pyx_tuple__74);
  Py_CLEAR(clear_module_state->__pyx_tuple__75);
  Py_CLEAR(clear_module_state->__pyx_tuple__76);
  Py_CLEAR(clear_module_state->__pyx_tuple__77);
  Py_CLEAR(clear_module_state->__pyx_tuple__78);
  Py_CLEAR(clear_module_state->__pyx_tuple__79);
  Py_CLEAR(clear_module_state->__pyx_tuple__80);
  Py_CLEAR(clear_module_state->__pyx_tuple__81);
  Py_CLEAR(clear_module_state->__pyx_tuple__82);
  Py_CLEAR(clear_module_state->__pyx_tuple__83);
  Py_CLEAR(clear_module_state->__pyx_tuple__84);
  Py_CLEAR(clear_module_state->__pyx_tuple__85);
  Py_CLEAR(clear_module_state->__pyx_tuple__86);
  Py_CLEAR(clear_module_state->__pyx_tuple__87);
  Py_CLEAR(clear_module_state->__pyx_tuple__88);
  Py_CLEAR(clear_module_state->__pyx_tuple__89);
  Py_CLEAR(clear_module_state->__pyx_tuple__90);
  Py_CLEAR(clear_module_state->__pyx_tuple__91);
  Py_CLEAR(clear_module_state->__pyx_tuple__92);
  Py_CLEAR(clear_module_state->__pyx_tuple__93);
  Py_CLEAR(clear_module_state->__pyx_tuple__94);
  Py_CLEAR(clear_module_state->__pyx_tuple__95);
  Py_CLEAR(clear_module_state->__pyx_tuple__96);
  Py_CLEAR(clear_module_state->__pyx_tuple__97);
  Py_CLEAR(clear_module_state->__pyx_tuple__98);
  Py_CLEAR(clear_module_state->__pyx_tuple__99);
  Py_CLEAR(clear_module_state->__pyx_tuple__100);
  Py_CLEAR(clear_module_state->__pyx_tuple__101);
  Py_CLEAR(clear_module_state->__pyx_tuple__102);
  Py_CLEAR(clear_module_state->__pyx_tuple__103);
  Py_CLEAR(clear_module_state->__pyx_tuple__104);
  Py_CLEAR(clear_module_state->__pyx_tuple__105);
  Py_CLEAR(clear_module_state->__pyx_tuple__106);
  Py_CLEAR(clear_module_state->__pyx_tuple__107);
  Py_CLEAR(clear_module_state->__pyx_tuple__108);
  Py_CLEAR(clear_module_state->__pyx_tuple__109);
  Py_CLEAR(clear_module_state->__pyx_tuple__110);
  Py_CLEAR(clear_module_state->__pyx_tuple__111);
  Py_CLEAR(clear_module_state->__pyx_tuple__112);
  Py_CLEAR(clear_module_state->__pyx_tuple__113);
  Py_CLEAR(clear_module_state->__pyx_tuple__114);
  Py_CLEAR(clear_module_state->__pyx_tuple__115);
  Py_CLEAR(clear_module_state->__pyx_tuple__116);
  Py_CLEAR(clear_module_state->__pyx_tuple__117);
  Py_CLEAR(clear_module_state->__pyx_tuple__118);
  Py_CLEAR(clear_module_state->__pyx_tuple__119);
  Py_CLEAR(clear_module_state->__pyx_tuple__120);
  Py_CLEAR(clear_module_state->__pyx_tuple__121);
  Py_CLEAR(clear_module_state->__pyx_tuple__122);
  Py_CLEAR(clear_module_state->__pyx_tuple__123);
  Py_CLEAR(clear_module_state->__pyx_tuple__124);
  Py_CLEAR(clear_module_state->__pyx_tuple__125);
  Py_CLEAR(clear_module_state->__pyx_tuple__126);
  Py_CLEAR(clear_module_state->__pyx_tuple__127);
  Py_CLEAR(clear_module_state->__pyx_tuple__128);
  Py_CLEAR(clear_module_state->__pyx_tuple__129);
  Py_CLEAR(clear_module_state->__pyx_tuple__130);
  Py_CLEAR(clear_module_state->__pyx_tuple__131);
  Py_CLEAR(clear_module_state->__pyx_tuple__132);
  Py_CLEAR(clear_module_state->__pyx_tuple__133);
  Py_CLEAR(clear_module_state->__pyx_tuple__134);
  Py_CLEAR(clear_module_state->__pyx_tuple__135);
  Py_CLEAR(clear_module_state->__pyx_tuple__136);
  Py_CLEAR(clear_module_state->__pyx_tuple__137);
  Py_CLEAR(clear_module_state->__pyx_tuple__138);
  Py_CLEAR(clear_module_state->__pyx_tuple__139);
  Py_CLEAR(clear_module_state->__pyx_tuple__140);
  Py_CLEAR(clear_module_state->__pyx_tuple__141);
  Py_CLEAR(clear_module_state->__pyx_tuple__142);
  Py_CLEAR(clear_module_state->__pyx_tuple__143);
  Py_CLEAR(clear_module_state->__pyx_tuple__144);
  Py_CLEAR(clear_module_state->__pyx_tuple__145);
  Py_CLEAR(clear_module_state->__pyx_tuple__146);
  Py_CLEAR(clear_module_state->__pyx_tuple__147);
  Py_CLEAR(clear_module_state->__pyx_tuple__148);
  Py_CLEAR(clear_module_state->__pyx_codeobj__21);
  Py_CLEAR(clear_module_state->__pyx_codeobj__23);
  Py_CLEAR(clear_module_state->__pyx_codeobj__25);
  Py_CLEAR(clear_module_state->__pyx_codeobj__28);
  Py_CLEAR(clear_module_state->__pyx_codeobj__30);
  Py_CLEAR(clear_module_state->__pyx_codeobj__32);
  Py_CLEAR(clear_module_state->__pyx_codeobj__34);
  Py_CLEAR(clear_module_state->__pyx_codeobj__36);
  Py_CLEAR(clear_module_state->__pyx_codeobj__38);
  Py_CLEAR(clear_module_state->__pyx_codeobj__40);
  Py_CLEAR(clear_module_state->__pyx_codeobj__42);
  Py_CLEAR(clear_module_state->__pyx_codeobj__44);
  Py_CLEAR(clear_module_state->__pyx_codeobj__46);
  Py_CLEAR(clear_module_state->__pyx_codeobj__48);
  Py_CLEAR(clear_module_state->__pyx_codeobj__50);
  Py_CLEAR(clear_module_state->__pyx_codeobj__52);
  Py_CLEAR(clear_module_state->__pyx_codeobj__54);
  Py_CLEAR(clear_module_state->__pyx_codeobj__56);
  Py_CLEAR(clear_module_state->__pyx_codeobj__58);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_dtype);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flatiter);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_broadcast);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ndarray);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_generic);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_number);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_integer);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_inexact);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_floating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flexible);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_character);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ufunc);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile);
  Py_VISIT(traverse_module_state->__pyx_type_5pyart_2io_11_sigmetfile_SigmetFile);
  Py_VISIT(traverse_module_state->__pyx_kp_u_112s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_115s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_120s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_128s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_12s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_136s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_160s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_16s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_184s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_18s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_198s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_1s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_200s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_204s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_208s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_228s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_24s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_256s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_2612s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_28s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_2_uint8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_2s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_308s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_320s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_32s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_36s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_40s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_480s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_48s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_4s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_52s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_6s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_720s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_732s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_80s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_8s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_920s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_94s);
  Py_VISIT(traverse_module_state->__pyx_n_u_ALBEDO16);
  Py_VISIT(traverse_module_state->__pyx_n_u_ALBEDO8);
  Py_VISIT(traverse_module_state->__pyx_n_u_AXDIL2);
  Py_VISIT(traverse_module_state->__pyx_n_u_B);
  Py_VISIT(traverse_module_state->__pyx_n_s_BIN1);
  Py_VISIT(traverse_module_state->__pyx_n_s_BIN2);
  Py_VISIT(traverse_module_state->__pyx_n_s_BIN4);
  Py_VISIT(traverse_module_state->__pyx_n_s_COLOR_SCALE_DEF);
  Py_VISIT(traverse_module_state->__pyx_n_u_CSP16);
  Py_VISIT(traverse_module_state->__pyx_n_u_CSP8);
  Py_VISIT(traverse_module_state->__pyx_n_u_DBT);
  Py_VISIT(traverse_module_state->__pyx_n_u_DBT2);
  Py_VISIT(traverse_module_state->__pyx_n_u_DBTE16);
  Py_VISIT(traverse_module_state->__pyx_n_u_DBTE8);
  Py_VISIT(traverse_module_state->__pyx_n_u_DBTV16);
  Py_VISIT(traverse_module_state->__pyx_n_u_DBTV8);
  Py_VISIT(traverse_module_state->__pyx_n_u_DBZ);
  Py_VISIT(traverse_module_state->__pyx_n_u_DBZ2);
  Py_VISIT(traverse_module_state->__pyx_n_u_DBZC);
  Py_VISIT(traverse_module_state->__pyx_n_u_DBZC2);
  Py_VISIT(traverse_module_state->__pyx_n_u_DBZE16);
  Py_VISIT(traverse_module_state->__pyx_n_u_DBZE8);
  Py_VISIT(traverse_module_state->__pyx_n_u_DBZV16);
  Py_VISIT(traverse_module_state->__pyx_n_u_DBZV8);
  Py_VISIT(traverse_module_state->__pyx_n_u_DEFORM2);
  Py_VISIT(traverse_module_state->__pyx_n_u_DIVERGE2);
  Py_VISIT(traverse_module_state->__pyx_n_s_DSP_DATA_MASK);
  Py_VISIT(traverse_module_state->__pyx_n_u_FLIQUID2);
  Py_VISIT(traverse_module_state->__pyx_n_s_FLT4);
  Py_VISIT(traverse_module_state->__pyx_n_s_FLT8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_File_truncated_or_corrupt_i_of_i);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Finished_loading_record);
  Py_VISIT(traverse_module_state->__pyx_n_u_GMT_minute_offset_local);
  Py_VISIT(traverse_module_state->__pyx_n_u_GMT_minute_offset_standard);
  Py_VISIT(traverse_module_state->__pyx_n_u_H);
  Py_VISIT(traverse_module_state->__pyx_n_u_HCLASS);
  Py_VISIT(traverse_module_state->__pyx_n_u_HCLASS2);
  Py_VISIT(traverse_module_state->__pyx_n_u_HDIR2);
  Py_VISIT(traverse_module_state->__pyx_n_u_HEIGHT);
  Py_VISIT(traverse_module_state->__pyx_n_u_HVEL2);
  Py_VISIT(traverse_module_state->__pyx_n_u_I);
  Py_VISIT(traverse_module_state->__pyx_n_s_INGEST_CONFIGURATION);
  Py_VISIT(traverse_module_state->__pyx_n_s_INGEST_DATA_HEADER);
  Py_VISIT(traverse_module_state->__pyx_n_s_INGEST_HEADER);
  Py_VISIT(traverse_module_state->__pyx_n_s_ImportError);
  Py_VISIT(traverse_module_state->__pyx_n_u_KDP);
  Py_VISIT(traverse_module_state->__pyx_n_u_KDP2);
  Py_VISIT(traverse_module_state->__pyx_n_u_LDRH);
  Py_VISIT(traverse_module_state->__pyx_n_u_LDRH2);
  Py_VISIT(traverse_module_state->__pyx_n_u_LDRV);
  Py_VISIT(traverse_module_state->__pyx_n_u_LDRV2);
  Py_VISIT(traverse_module_state->__pyx_n_u_LOG16);
  Py_VISIT(traverse_module_state->__pyx_n_u_LOG8);
  Py_VISIT(traverse_module_state->__pyx_n_s_MESSAGE);
  Py_VISIT(traverse_module_state->__pyx_n_u_OTHER);
  Py_VISIT(traverse_module_state->__pyx_n_u_PHIDP);
  Py_VISIT(traverse_module_state->__pyx_n_u_PHIDP2);
  Py_VISIT(traverse_module_state->__pyx_n_u_PHIH);
  Py_VISIT(traverse_module_state->__pyx_n_u_PHIH2);
  Py_VISIT(traverse_module_state->__pyx_n_u_PHIV);
  Py_VISIT(traverse_module_state->__pyx_n_u_PHIV2);
  Py_VISIT(traverse_module_state->__pyx_n_u_PMI16);
  Py_VISIT(traverse_module_state->__pyx_n_u_PMI8);
  Py_VISIT(traverse_module_state->__pyx_n_s_PRODUCT_CONFIGURATION);
  Py_VISIT(traverse_module_state->__pyx_n_s_PRODUCT_END);
  Py_VISIT(traverse_module_state->__pyx_n_s_PRODUCT_HDR);
  Py_VISIT(traverse_module_state->__pyx_n_u_RAINRATE2);
  Py_VISIT(traverse_module_state->__pyx_n_u_RAW);
  Py_VISIT(traverse_module_state->__pyx_n_s_RAW_PROD_BHDR);
  Py_VISIT(traverse_module_state->__pyx_n_s_RECORD_SIZE);
  Py_VISIT(traverse_module_state->__pyx_n_u_RHOH);
  Py_VISIT(traverse_module_state->__pyx_n_u_RHOH2);
  Py_VISIT(traverse_module_state->__pyx_n_u_RHOHV);
  Py_VISIT(traverse_module_state->__pyx_n_u_RHOHV2);
  Py_VISIT(traverse_module_state->__pyx_n_u_RHOV);
  Py_VISIT(traverse_module_state->__pyx_n_u_RHOV2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Reading_ray_i_of_i);
  Py_VISIT(traverse_module_state->__pyx_n_u_SHEAR);
  Py_VISIT(traverse_module_state->__pyx_n_s_SIGMET_DATA_TYPES);
  Py_VISIT(traverse_module_state->__pyx_n_s_SINT1);
  Py_VISIT(traverse_module_state->__pyx_n_s_SINT2);
  Py_VISIT(traverse_module_state->__pyx_n_s_SINT4);
  Py_VISIT(traverse_module_state->__pyx_n_u_SNR16);
  Py_VISIT(traverse_module_state->__pyx_n_u_SNR8);
  Py_VISIT(traverse_module_state->__pyx_n_u_SQI);
  Py_VISIT(traverse_module_state->__pyx_n_u_SQI2);
  Py_VISIT(traverse_module_state->__pyx_n_s_STRUCTURE_HEADER);
  Py_VISIT(traverse_module_state->__pyx_n_s_SigmetFile);
  Py_VISIT(traverse_module_state->__pyx_n_s_SigmetFile___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_SigmetFile___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_SigmetFile__determine_data_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_SigmetFile__get_sweep);
  Py_VISIT(traverse_module_state->__pyx_n_s_SigmetFile_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_SigmetFile_read_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_TASK_CALIB_INFO);
  Py_VISIT(traverse_module_state->__pyx_n_s_TASK_CONFIGURATION);
  Py_VISIT(traverse_module_state->__pyx_n_s_TASK_DSP_INFO);
  Py_VISIT(traverse_module_state->__pyx_n_s_TASK_DSP_MODE_BATCH);
  Py_VISIT(traverse_module_state->__pyx_n_s_TASK_END_INFO);
  Py_VISIT(traverse_module_state->__pyx_n_s_TASK_FILE_SCAN_INFO);
  Py_VISIT(traverse_module_state->__pyx_n_s_TASK_MANUAL_SCAN_INFO);
  Py_VISIT(traverse_module_state->__pyx_n_s_TASK_MISC_INFO);
  Py_VISIT(traverse_module_state->__pyx_n_s_TASK_PPI_SCAN_INFO);
  Py_VISIT(traverse_module_state->__pyx_n_s_TASK_RANGE_INFO);
  Py_VISIT(traverse_module_state->__pyx_n_s_TASK_RHI_SCAN_INFO);
  Py_VISIT(traverse_module_state->__pyx_n_s_TASK_SCAN_INFO);
  Py_VISIT(traverse_module_state->__pyx_n_s_TASK_SCHED_INFO);
  Py_VISIT(traverse_module_state->__pyx_n_u_TEMPERATURE16);
  Py_VISIT(traverse_module_state->__pyx_n_u_TIME2);
  Py_VISIT(traverse_module_state->__pyx_n_u_TURB16);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_UINT1);
  Py_VISIT(traverse_module_state->__pyx_n_s_UINT16_T);
  Py_VISIT(traverse_module_state->__pyx_n_s_UINT2);
  Py_VISIT(traverse_module_state->__pyx_n_s_UINT4);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_100);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_101);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_102);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_103);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_104);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_105);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_106);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_107);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_108);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_109);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_110);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_112);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_113);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_114);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_115);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_116);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_117);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_118);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_119);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_120);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_121);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_122);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_123);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_124);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_125);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_126);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_127);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_29);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_30);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_31);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_6);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_81);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_82);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_83);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_84);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_85);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_86);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_87);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_88);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_89);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_90);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_91);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_92);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_93);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_94);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_95);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_96);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_97);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_98);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN_99);
  Py_VISIT(traverse_module_state->__pyx_n_u_USER);
  Py_VISIT(traverse_module_state->__pyx_n_u_USER2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unknown_type_s_returning_raw_dat);
  Py_VISIT(traverse_module_state->__pyx_n_u_VEL);
  Py_VISIT(traverse_module_state->__pyx_n_u_VEL2);
  Py_VISIT(traverse_module_state->__pyx_n_u_VELC);
  Py_VISIT(traverse_module_state->__pyx_n_u_VELC2);
  Py_VISIT(traverse_module_state->__pyx_n_u_VIL2);
  Py_VISIT(traverse_module_state->__pyx_n_u_VILD16);
  Py_VISIT(traverse_module_state->__pyx_n_u_VIR16);
  Py_VISIT(traverse_module_state->__pyx_n_u_VVEL2);
  Py_VISIT(traverse_module_state->__pyx_n_u_WIDTH);
  Py_VISIT(traverse_module_state->__pyx_n_u_WIDTH2);
  Py_VISIT(traverse_module_state->__pyx_n_u_XHDR);
  Py_VISIT(traverse_module_state->__pyx_n_s_YMDS_TIME);
  Py_VISIT(traverse_module_state->__pyx_n_u_ZDR);
  Py_VISIT(traverse_module_state->__pyx_n_u_ZDR2);
  Py_VISIT(traverse_module_state->__pyx_n_u_ZDRC);
  Py_VISIT(traverse_module_state->__pyx_n_u_ZDRC2);
  Py_VISIT(traverse_module_state->__pyx_n_s__149);
  Py_VISIT(traverse_module_state->__pyx_kp_u__15);
  Py_VISIT(traverse_module_state->__pyx_n_s__19);
  Py_VISIT(traverse_module_state->__pyx_n_u_agc_feedback_code);
  Py_VISIT(traverse_module_state->__pyx_n_u_altitude_radar);
  Py_VISIT(traverse_module_state->__pyx_n_u_angular_resolution_desired);
  Py_VISIT(traverse_module_state->__pyx_n_u_antenna_offset_bow);
  Py_VISIT(traverse_module_state->__pyx_n_u_antenna_offset_starboard);
  Py_VISIT(traverse_module_state->__pyx_n_u_antenna_offset_up);
  Py_VISIT(traverse_module_state->__pyx_n_u_antenna_scan_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_append);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_u_attenuation);
  Py_VISIT(traverse_module_state->__pyx_n_s_az0);
  Py_VISIT(traverse_module_state->__pyx_n_s_az1);
  Py_VISIT(traverse_module_state->__pyx_n_u_azimuth_0);
  Py_VISIT(traverse_module_state->__pyx_n_u_azimuth_1);
  Py_VISIT(traverse_module_state->__pyx_n_u_azimuth_list);
  Py_VISIT(traverse_module_state->__pyx_n_u_b);
  Py_VISIT(traverse_module_state->__pyx_n_s_bin2);
  Py_VISIT(traverse_module_state->__pyx_n_s_bin2_to_angle);
  Py_VISIT(traverse_module_state->__pyx_n_s_bin4);
  Py_VISIT(traverse_module_state->__pyx_n_s_bin4_to_angle);
  Py_VISIT(traverse_module_state->__pyx_n_s_bit);
  Py_VISIT(traverse_module_state->__pyx_n_u_bit_per_bin);
  Py_VISIT(traverse_module_state->__pyx_n_u_bool);
  Py_VISIT(traverse_module_state->__pyx_n_u_bytes_in_structure);
  Py_VISIT(traverse_module_state->__pyx_n_u_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_close);
  Py_VISIT(traverse_module_state->__pyx_n_u_clutter_correction_threshold);
  Py_VISIT(traverse_module_state->__pyx_n_u_clutter_filter);
  Py_VISIT(traverse_module_state->__pyx_n_u_clutter_filter_name);
  Py_VISIT(traverse_module_state->__pyx_n_u_cluttermap);
  Py_VISIT(traverse_module_state->__pyx_n_u_cluttermap_flag);
  Py_VISIT(traverse_module_state->__pyx_n_u_color_scale_def);
  Py_VISIT(traverse_module_state->__pyx_n_u_comment_bytes);
  Py_VISIT(traverse_module_state->__pyx_n_u_comments);
  Py_VISIT(traverse_module_state->__pyx_n_u_configuration_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_convert_sigmet_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_copy);
  Py_VISIT(traverse_module_state->__pyx_n_u_corrected_reflectivity_threshold);
  Py_VISIT(traverse_module_state->__pyx_n_u_current_data_type_mask);
  Py_VISIT(traverse_module_state->__pyx_n_u_custom_ray_header_name);
  Py_VISIT(traverse_module_state->__pyx_n_u_customer_storage);
  Py_VISIT(traverse_module_state->__pyx_n_s_d);
  Py_VISIT(traverse_module_state->__pyx_n_u_d);
  Py_VISIT(traverse_module_state->__pyx_n_s_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_data_type);
  Py_VISIT(traverse_module_state->__pyx_n_u_data_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_data_type_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_data_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_data_types_from_mask);
  Py_VISIT(traverse_module_state->__pyx_n_s_datetime);
  Py_VISIT(traverse_module_state->__pyx_n_u_day);
  Py_VISIT(traverse_module_state->__pyx_n_s_debug);
  Py_VISIT(traverse_module_state->__pyx_n_s_determine_data_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_dic);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_u_dsp_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_dtype);
  Py_VISIT(traverse_module_state->__pyx_n_u_dual_prf_delay);
  Py_VISIT(traverse_module_state->__pyx_n_u_earth_radius);
  Py_VISIT(traverse_module_state->__pyx_n_s_el0);
  Py_VISIT(traverse_module_state->__pyx_n_s_el1);
  Py_VISIT(traverse_module_state->__pyx_n_u_elevation_0);
  Py_VISIT(traverse_module_state->__pyx_n_u_elevation_1);
  Py_VISIT(traverse_module_state->__pyx_n_u_elevation_list);
  Py_VISIT(traverse_module_state->__pyx_n_s_empty);
  Py_VISIT(traverse_module_state->__pyx_n_s_empty_like);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_enumerate);
  Py_VISIT(traverse_module_state->__pyx_n_s_exp);
  Py_VISIT(traverse_module_state->__pyx_n_u_extended_header_type);
  Py_VISIT(traverse_module_state->__pyx_n_u_extended_product_header_offset);
  Py_VISIT(traverse_module_state->__pyx_n_u_extended_ray_header_bytes);
  Py_VISIT(traverse_module_state->__pyx_n_u_f);
  Py_VISIT(traverse_module_state->__pyx_n_u_fault_status);
  Py_VISIT(traverse_module_state->__pyx_n_u_file_ingest_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_filename);
  Py_VISIT(traverse_module_state->__pyx_n_u_filename);
  Py_VISIT(traverse_module_state->__pyx_n_s_fill_value);
  Py_VISIT(traverse_module_state->__pyx_n_u_first_azimuth);
  Py_VISIT(traverse_module_state->__pyx_n_u_first_bin_range);
  Py_VISIT(traverse_module_state->__pyx_n_u_first_elevation);
  Py_VISIT(traverse_module_state->__pyx_n_u_first_ray_index);
  Py_VISIT(traverse_module_state->__pyx_n_u_first_ray_offset);
  Py_VISIT(traverse_module_state->__pyx_n_u_fixed_angle);
  Py_VISIT(traverse_module_state->__pyx_n_u_flag);
  Py_VISIT(traverse_module_state->__pyx_n_u_flags);
  Py_VISIT(traverse_module_state->__pyx_n_u_flags2);
  Py_VISIT(traverse_module_state->__pyx_n_u_float32);
  Py_VISIT(traverse_module_state->__pyx_n_s_fmt);
  Py_VISIT(traverse_module_state->__pyx_n_u_format_version);
  Py_VISIT(traverse_module_state->__pyx_n_s_frombuffer);
  Py_VISIT(traverse_module_state->__pyx_n_s_full_xhdr);
  Py_VISIT(traverse_module_state->__pyx_n_u_gain_control_flag);
  Py_VISIT(traverse_module_state->__pyx_n_u_gas_attenuation);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_u_generation_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_sweep);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_n_u_gmt_offset_minutes_local);
  Py_VISIT(traverse_module_state->__pyx_n_u_gmt_offset_minutes_standard);
  Py_VISIT(traverse_module_state->__pyx_n_u_gparam_bytes);
  Py_VISIT(traverse_module_state->__pyx_n_u_gparm);
  Py_VISIT(traverse_module_state->__pyx_n_u_ground_height);
  Py_VISIT(traverse_module_state->__pyx_n_u_h);
  Py_VISIT(traverse_module_state->__pyx_n_u_hardware_site);
  Py_VISIT(traverse_module_state->__pyx_n_s_header_dic);
  Py_VISIT(traverse_module_state->__pyx_n_u_height_radar);
  Py_VISIT(traverse_module_state->__pyx_n_u_height_site);
  Py_VISIT(traverse_module_state->__pyx_n_u_horizontal_beamwidth);
  Py_VISIT(traverse_module_state->__pyx_n_u_horizontal_calibration_i0);
  Py_VISIT(traverse_module_state->__pyx_n_u_horizontal_calibration_noise);
  Py_VISIT(traverse_module_state->__pyx_n_u_horizontal_current_noise);
  Py_VISIT(traverse_module_state->__pyx_n_u_horizontal_i0_calibration);
  Py_VISIT(traverse_module_state->__pyx_n_u_horizontal_noise_calibration);
  Py_VISIT(traverse_module_state->__pyx_n_u_horizontal_radar_constant);
  Py_VISIT(traverse_module_state->__pyx_n_s_i);
  Py_VISIT(traverse_module_state->__pyx_n_u_i);
  Py_VISIT(traverse_module_state->__pyx_n_u_i4);
  Py_VISIT(traverse_module_state->__pyx_n_u_icolcnt);
  Py_VISIT(traverse_module_state->__pyx_n_s_idh);
  Py_VISIT(traverse_module_state->__pyx_n_u_iflags);
  Py_VISIT(traverse_module_state->__pyx_n_u_ilevel_seams);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_ingest_configuration);
  Py_VISIT(traverse_module_state->__pyx_n_u_ingest_configuration);
  Py_VISIT(traverse_module_state->__pyx_n_s_ingest_data_hdrs);
  Py_VISIT(traverse_module_state->__pyx_n_s_ingest_data_headers);
  Py_VISIT(traverse_module_state->__pyx_n_u_ingest_hardware_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_ingest_header);
  Py_VISIT(traverse_module_state->__pyx_n_u_ingest_iris_version);
  Py_VISIT(traverse_module_state->__pyx_n_u_ingest_site_name);
  Py_VISIT(traverse_module_state->__pyx_n_u_ingest_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_u_input_data_type);
  Py_VISIT(traverse_module_state->__pyx_n_u_input_mask);
  Py_VISIT(traverse_module_state->__pyx_n_u_int16);
  Py_VISIT(traverse_module_state->__pyx_n_u_int32);
  Py_VISIT(traverse_module_state->__pyx_n_u_inverse_flatting);
  Py_VISIT(traverse_module_state->__pyx_n_u_iris_version);
  Py_VISIT(traverse_module_state->__pyx_n_u_iris_version_created);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_bit_set);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_u_iset_and_scale);
  Py_VISIT(traverse_module_state->__pyx_n_u_istart);
  Py_VISIT(traverse_module_state->__pyx_n_u_istep);
  Py_VISIT(traverse_module_state->__pyx_n_s_j);
  Py_VISIT(traverse_module_state->__pyx_n_s_k);
  Py_VISIT(traverse_module_state->__pyx_n_s_key);
  Py_VISIT(traverse_module_state->__pyx_n_s_l);
  Py_VISIT(traverse_module_state->__pyx_n_u_last_bin_range);
  Py_VISIT(traverse_module_state->__pyx_n_u_last_run_day);
  Py_VISIT(traverse_module_state->__pyx_n_u_last_run_time);
  Py_VISIT(traverse_module_state->__pyx_n_u_latitude);
  Py_VISIT(traverse_module_state->__pyx_n_u_latitude_projection);
  Py_VISIT(traverse_module_state->__pyx_n_u_latitude_radar);
  Py_VISIT(traverse_module_state->__pyx_n_u_ldr_bias);
  Py_VISIT(traverse_module_state->__pyx_n_u_ldr_offset);
  Py_VISIT(traverse_module_state->__pyx_n_s_lead_record);
  Py_VISIT(traverse_module_state->__pyx_n_u_left_azimuth_limit);
  Py_VISIT(traverse_module_state->__pyx_n_s_like_dbt);
  Py_VISIT(traverse_module_state->__pyx_n_s_like_dbt2);
  Py_VISIT(traverse_module_state->__pyx_n_s_like_sqi);
  Py_VISIT(traverse_module_state->__pyx_n_s_like_sqi2);
  Py_VISIT(traverse_module_state->__pyx_n_u_linear_filter_first_bin);
  Py_VISIT(traverse_module_state->__pyx_n_u_local_timezone);
  Py_VISIT(traverse_module_state->__pyx_n_u_log_filter_first_bin);
  Py_VISIT(traverse_module_state->__pyx_n_u_longitude);
  Py_VISIT(traverse_module_state->__pyx_n_u_longitude_projection);
  Py_VISIT(traverse_module_state->__pyx_n_u_longitude_radar);
  Py_VISIT(traverse_module_state->__pyx_n_u_low_prf_factional);
  Py_VISIT(traverse_module_state->__pyx_n_u_low_prf_hz);
  Py_VISIT(traverse_module_state->__pyx_n_u_low_prf_range_averaging);
  Py_VISIT(traverse_module_state->__pyx_n_u_low_prf_sample_size);
  Py_VISIT(traverse_module_state->__pyx_n_u_lower_elevation_limit);
  Py_VISIT(traverse_module_state->__pyx_n_s_ma);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_u_major_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_mask);
  Py_VISIT(traverse_module_state->__pyx_n_u_mask_word_0);
  Py_VISIT(traverse_module_state->__pyx_n_u_mask_word_1);
  Py_VISIT(traverse_module_state->__pyx_n_u_mask_word_2);
  Py_VISIT(traverse_module_state->__pyx_n_u_mask_word_3);
  Py_VISIT(traverse_module_state->__pyx_n_u_mask_word_4);
  Py_VISIT(traverse_module_state->__pyx_n_s_masked_array);
  Py_VISIT(traverse_module_state->__pyx_n_u_maximum_range);
  Py_VISIT(traverse_module_state->__pyx_n_u_mean_wind_direction);
  Py_VISIT(traverse_module_state->__pyx_n_u_mean_wind_speed);
  Py_VISIT(traverse_module_state->__pyx_n_u_melting_layer);
  Py_VISIT(traverse_module_state->__pyx_n_u_melting_level);
  Py_VISIT(traverse_module_state->__pyx_n_s_mess);
  Py_VISIT(traverse_module_state->__pyx_n_s_metadata);
  Py_VISIT(traverse_module_state->__pyx_n_u_milliseconds);
  Py_VISIT(traverse_module_state->__pyx_n_u_minor_task_suffix);
  Py_VISIT(traverse_module_state->__pyx_n_s_mod);
  Py_VISIT(traverse_module_state->__pyx_n_u_month);
  Py_VISIT(traverse_module_state->__pyx_n_s_multi_prf_flag);
  Py_VISIT(traverse_module_state->__pyx_n_u_multi_prf_flag);
  Py_VISIT(traverse_module_state->__pyx_n_s_multiplier);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_nbin);
  Py_VISIT(traverse_module_state->__pyx_n_s_nbins);
  Py_VISIT(traverse_module_state->__pyx_n_u_nbins);
  Py_VISIT(traverse_module_state->__pyx_n_s_ndata);
  Py_VISIT(traverse_module_state->__pyx_n_s_ndata_types);
  Py_VISIT(traverse_module_state->__pyx_n_u_nexrad_clutter_skip);
  Py_VISIT(traverse_module_state->__pyx_n_u_nexrad_clutter_threshold);
  Py_VISIT(traverse_module_state->__pyx_n_s_np);
  Py_VISIT(traverse_module_state->__pyx_n_s_nray_data_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_nrays);
  Py_VISIT(traverse_module_state->__pyx_n_s_nsweeps);
  Py_VISIT(traverse_module_state->__pyx_n_s_number);
  Py_VISIT(traverse_module_state->__pyx_n_u_number_bins);
  Py_VISIT(traverse_module_state->__pyx_n_u_number_elements);
  Py_VISIT(traverse_module_state->__pyx_n_u_number_files);
  Py_VISIT(traverse_module_state->__pyx_n_u_number_ingest);
  Py_VISIT(traverse_module_state->__pyx_n_u_number_input_bins);
  Py_VISIT(traverse_module_state->__pyx_n_u_number_linear_filter);
  Py_VISIT(traverse_module_state->__pyx_n_u_number_log_filter);
  Py_VISIT(traverse_module_state->__pyx_n_u_number_output_bins);
  Py_VISIT(traverse_module_state->__pyx_n_u_number_rays_file_actual);
  Py_VISIT(traverse_module_state->__pyx_n_u_number_rays_file_expected);
  Py_VISIT(traverse_module_state->__pyx_n_u_number_rays_sweep);
  Py_VISIT(traverse_module_state->__pyx_n_u_number_sweeps);
  Py_VISIT(traverse_module_state->__pyx_n_u_number_sweeps_completed);
  Py_VISIT(traverse_module_state->__pyx_n_u_number_task_config_table);
  Py_VISIT(traverse_module_state->__pyx_n_u_number_tasks);
  Py_VISIT(traverse_module_state->__pyx_n_s_numpy);
  Py_VISIT(traverse_module_state->__pyx_kp_u_numpy__core_multiarray_failed_to);
  Py_VISIT(traverse_module_state->__pyx_kp_u_numpy__core_umath_failed_to_impo);
  Py_VISIT(traverse_module_state->__pyx_n_s_nyquist);
  Py_VISIT(traverse_module_state->__pyx_n_s_offset);
  Py_VISIT(traverse_module_state->__pyx_n_s_ones);
  Py_VISIT(traverse_module_state->__pyx_n_s_open);
  Py_VISIT(traverse_module_state->__pyx_n_u_original_data_type_mask);
  Py_VISIT(traverse_module_state->__pyx_n_s_out);
  Py_VISIT(traverse_module_state->__pyx_n_s_parse_ray_headers);
  Py_VISIT(traverse_module_state->__pyx_n_u_playback_flag);
  Py_VISIT(traverse_module_state->__pyx_n_u_playback_version);
  Py_VISIT(traverse_module_state->__pyx_n_u_polarization);
  Py_VISIT(traverse_module_state->__pyx_n_u_polarization_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_power);
  Py_VISIT(traverse_module_state->__pyx_n_u_power_threshold);
  Py_VISIT(traverse_module_state->__pyx_n_u_prf);
  Py_VISIT(traverse_module_state->__pyx_n_s_prf_flag);
  Py_VISIT(traverse_module_state->__pyx_n_u_prf_flag);
  Py_VISIT(traverse_module_state->__pyx_n_s_print);
  Py_VISIT(traverse_module_state->__pyx_n_s_product_config);
  Py_VISIT(traverse_module_state->__pyx_n_u_product_configuration);
  Py_VISIT(traverse_module_state->__pyx_n_s_product_end);
  Py_VISIT(traverse_module_state->__pyx_n_u_product_end);
  Py_VISIT(traverse_module_state->__pyx_n_s_product_hdr);
  Py_VISIT(traverse_module_state->__pyx_n_u_product_name);
  Py_VISIT(traverse_module_state->__pyx_n_u_product_sequence_number);
  Py_VISIT(traverse_module_state->__pyx_n_u_product_specific_bytes);
  Py_VISIT(traverse_module_state->__pyx_n_u_product_type_code);
  Py_VISIT(traverse_module_state->__pyx_n_u_projection_name);
  Py_VISIT(traverse_module_state->__pyx_n_u_projection_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_prt_value);
  Py_VISIT(traverse_module_state->__pyx_n_u_pulse_width);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyart_io__sigmetfile);
  Py_VISIT(traverse_module_state->__pyx_kp_s_pyart_io__sigmetfile_pyx);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_u_radar_height);
  Py_VISIT(traverse_module_state->__pyx_n_u_radar_height_above_reference);
  Py_VISIT(traverse_module_state->__pyx_n_u_radial_smoother);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_u_range_bin_averaging_flag);
  Py_VISIT(traverse_module_state->__pyx_n_s_raw_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_raw_prod_bhdr);
  Py_VISIT(traverse_module_state->__pyx_n_s_raw_sweep_data);
  Py_VISIT(traverse_module_state->__pyx_n_u_ray_header_bytes);
  Py_VISIT(traverse_module_state->__pyx_n_u_ray_header_mask);
  Py_VISIT(traverse_module_state->__pyx_n_s_ray_headers);
  Py_VISIT(traverse_module_state->__pyx_n_s_ray_i);
  Py_VISIT(traverse_module_state->__pyx_n_s_ray_nbins);
  Py_VISIT(traverse_module_state->__pyx_n_u_ray_number);
  Py_VISIT(traverse_module_state->__pyx_n_s_ray_time);
  Py_VISIT(traverse_module_state->__pyx_n_u_rb);
  Py_VISIT(traverse_module_state->__pyx_n_s_read);
  Py_VISIT(traverse_module_state->__pyx_n_u_read);
  Py_VISIT(traverse_module_state->__pyx_n_s_read_data);
  Py_VISIT(traverse_module_state->__pyx_n_u_reciever_bandwidth);
  Py_VISIT(traverse_module_state->__pyx_n_s_record);
  Py_VISIT(traverse_module_state->__pyx_n_u_record_number);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_u_reflectivity_calibration);
  Py_VISIT(traverse_module_state->__pyx_n_u_reflectivity_noise_threshold);
  Py_VISIT(traverse_module_state->__pyx_n_u_reflectivity_slope);
  Py_VISIT(traverse_module_state->__pyx_n_u_reflectivity_unfolding_threshold);
  Py_VISIT(traverse_module_state->__pyx_n_u_reserved);
  Py_VISIT(traverse_module_state->__pyx_n_s_reshape);
  Py_VISIT(traverse_module_state->__pyx_n_u_resolution_rays);
  Py_VISIT(traverse_module_state->__pyx_n_u_right_azimuth_limit);
  Py_VISIT(traverse_module_state->__pyx_n_u_sample_size);
  Py_VISIT(traverse_module_state->__pyx_n_u_samples_used);
  Py_VISIT(traverse_module_state->__pyx_n_u_scheduling_code);
  Py_VISIT(traverse_module_state->__pyx_n_u_seconds);
  Py_VISIT(traverse_module_state->__pyx_n_u_seconds_between_runs);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_kp_s_self__rbuf_p_cannot_be_converted);
  Py_VISIT(traverse_module_state->__pyx_kp_u_self__rbuf_pos_is);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_shape);
  Py_VISIT(traverse_module_state->__pyx_n_s_shrink);
  Py_VISIT(traverse_module_state->__pyx_n_u_signal_processor_type);
  Py_VISIT(traverse_module_state->__pyx_n_u_site_name);
  Py_VISIT(traverse_module_state->__pyx_n_u_skip_time);
  Py_VISIT(traverse_module_state->__pyx_n_u_spare_0);
  Py_VISIT(traverse_module_state->__pyx_n_u_spare_1);
  Py_VISIT(traverse_module_state->__pyx_n_u_spare_2);
  Py_VISIT(traverse_module_state->__pyx_n_u_spare_3);
  Py_VISIT(traverse_module_state->__pyx_n_u_spare_4);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_n_u_sqi_threshold);
  Py_VISIT(traverse_module_state->__pyx_n_s_sqrt);
  Py_VISIT(traverse_module_state->__pyx_n_u_standard_parallel_1);
  Py_VISIT(traverse_module_state->__pyx_n_u_standard_parallel_2);
  Py_VISIT(traverse_module_state->__pyx_n_u_start_first_section_flag);
  Py_VISIT(traverse_module_state->__pyx_n_u_start_first_sector_flag);
  Py_VISIT(traverse_module_state->__pyx_n_u_start_time);
  Py_VISIT(traverse_module_state->__pyx_n_u_step_input_bins);
  Py_VISIT(traverse_module_state->__pyx_n_u_step_output_bins);
  Py_VISIT(traverse_module_state->__pyx_n_u_stop_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_string);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_struct);
  Py_VISIT(traverse_module_state->__pyx_n_s_structure);
  Py_VISIT(traverse_module_state->__pyx_n_u_structure_header);
  Py_VISIT(traverse_module_state->__pyx_n_u_structure_identifier);
  Py_VISIT(traverse_module_state->__pyx_n_s_sum);
  Py_VISIT(traverse_module_state->__pyx_n_s_sweep_data);
  Py_VISIT(traverse_module_state->__pyx_n_u_sweep_ingest_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_sweep_metadata);
  Py_VISIT(traverse_module_state->__pyx_n_u_sweep_number);
  Py_VISIT(traverse_module_state->__pyx_n_u_sweep_start_time);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_calib_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_task_config);
  Py_VISIT(traverse_module_state->__pyx_n_s_task_configuration);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_configuration);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_configuration_file_name);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_data_time);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_description);
  Py_VISIT(traverse_module_state->__pyx_n_s_task_dsp_info);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_dsp_info);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_dsp_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_task_end_info);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_end_info);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_major_number);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_minor_number);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_misc_info);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_name);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_range_info);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_scan_info);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_scan_type_scan_info);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_sched_info);
  Py_VISIT(traverse_module_state->__pyx_n_u_task_state);
  Py_VISIT(traverse_module_state->__pyx_n_u_tcf_cal_flags_1);
  Py_VISIT(traverse_module_state->__pyx_n_u_tcf_cal_flags_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_temp);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_time);
  Py_VISIT(traverse_module_state->__pyx_n_u_time);
  Py_VISIT(traverse_module_state->__pyx_n_u_time_used_last_run);
  Py_VISIT(traverse_module_state->__pyx_n_u_times_run);
  Py_VISIT(traverse_module_state->__pyx_n_u_total_size);
  Py_VISIT(traverse_module_state->__pyx_n_u_tr_serial_number);
  Py_VISIT(traverse_module_state->__pyx_n_u_tranmitter_phase_sequence);
  Py_VISIT(traverse_module_state->__pyx_n_u_transmit_power);
  Py_VISIT(traverse_module_state->__pyx_n_u_trigger_rate);
  Py_VISIT(traverse_module_state->__pyx_n_u_trucation_height);
  Py_VISIT(traverse_module_state->__pyx_n_u_truncation_height);
  Py_VISIT(traverse_module_state->__pyx_n_u_tz_name);
  Py_VISIT(traverse_module_state->__pyx_n_u_uint16);
  Py_VISIT(traverse_module_state->__pyx_n_s_uint8);
  Py_VISIT(traverse_module_state->__pyx_n_u_uncorrected_reflectivity_thresho);
  Py_VISIT(traverse_module_state->__pyx_n_s_unpack);
  Py_VISIT(traverse_module_state->__pyx_n_s_unpack_ingest_data_header);
  Py_VISIT(traverse_module_state->__pyx_n_s_unpack_ingest_data_headers);
  Py_VISIT(traverse_module_state->__pyx_n_s_unpack_ingest_header);
  Py_VISIT(traverse_module_state->__pyx_n_s_unpack_key);
  Py_VISIT(traverse_module_state->__pyx_n_s_unpack_product_hdr);
  Py_VISIT(traverse_module_state->__pyx_n_s_unpack_raw_prod_bhdr);
  Py_VISIT(traverse_module_state->__pyx_n_s_unpack_structure);
  Py_VISIT(traverse_module_state->__pyx_n_u_upper_elevation_limit);
  Py_VISIT(traverse_module_state->__pyx_n_u_variable_range_bin_flag);
  Py_VISIT(traverse_module_state->__pyx_n_u_velocity_east);
  Py_VISIT(traverse_module_state->__pyx_n_u_velocity_north);
  Py_VISIT(traverse_module_state->__pyx_n_u_velocity_threshold_flags);
  Py_VISIT(traverse_module_state->__pyx_n_u_velocity_unfolding_threshold);
  Py_VISIT(traverse_module_state->__pyx_n_u_velocity_up);
  Py_VISIT(traverse_module_state->__pyx_n_u_vertical_beamwidth);
  Py_VISIT(traverse_module_state->__pyx_n_u_vertical_current_noise);
  Py_VISIT(traverse_module_state->__pyx_n_u_vertical_i0_calibration);
  Py_VISIT(traverse_module_state->__pyx_n_u_vertical_noise_calibration);
  Py_VISIT(traverse_module_state->__pyx_n_u_vertical_radar_constant);
  Py_VISIT(traverse_module_state->__pyx_n_s_view);
  Py_VISIT(traverse_module_state->__pyx_n_u_volume_scan_start_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_warn);
  Py_VISIT(traverse_module_state->__pyx_n_s_warnings);
  Py_VISIT(traverse_module_state->__pyx_n_u_wavelength);
  Py_VISIT(traverse_module_state->__pyx_n_s_wavelength_cm);
  Py_VISIT(traverse_module_state->__pyx_n_u_width_threshold_flags);
  Py_VISIT(traverse_module_state->__pyx_n_u_width_unfolding_threshold);
  Py_VISIT(traverse_module_state->__pyx_n_s_word0);
  Py_VISIT(traverse_module_state->__pyx_n_s_word1);
  Py_VISIT(traverse_module_state->__pyx_n_s_word2);
  Py_VISIT(traverse_module_state->__pyx_n_s_word3);
  Py_VISIT(traverse_module_state->__pyx_n_u_x_location);
  Py_VISIT(traverse_module_state->__pyx_n_u_x_scale);
  Py_VISIT(traverse_module_state->__pyx_n_u_x_size);
  Py_VISIT(traverse_module_state->__pyx_n_u_x_smoother);
  Py_VISIT(traverse_module_state->__pyx_n_s_xrange);
  Py_VISIT(traverse_module_state->__pyx_n_u_y_location);
  Py_VISIT(traverse_module_state->__pyx_n_u_y_scale);
  Py_VISIT(traverse_module_state->__pyx_n_u_y_size);
  Py_VISIT(traverse_module_state->__pyx_n_u_y_smoother);
  Py_VISIT(traverse_module_state->__pyx_n_u_year);
  Py_VISIT(traverse_module_state->__pyx_n_u_z_location);
  Py_VISIT(traverse_module_state->__pyx_n_u_z_scale);
  Py_VISIT(traverse_module_state->__pyx_n_u_z_size);
  Py_VISIT(traverse_module_state->__pyx_n_u_zdr_bias);
  Py_VISIT(traverse_module_state->__pyx_n_u_zdr_offset);
  Py_VISIT(traverse_module_state->__pyx_n_u_zdr_threshold_flags);
  Py_VISIT(traverse_module_state->__pyx_n_s_zeros_like);
  Py_VISIT(traverse_module_state->__pyx_n_s_zip);
  Py_VISIT(traverse_module_state->__pyx_n_u_zr_constant);
  Py_VISIT(traverse_module_state->__pyx_n_u_zr_exponent);
  Py_VISIT(traverse_module_state->__pyx_float_1_);
  Py_VISIT(traverse_module_state->__pyx_float_2_);
  Py_VISIT(traverse_module_state->__pyx_float_16_);
  Py_VISIT(traverse_module_state->__pyx_float_64_);
  Py_VISIT(traverse_module_state->__pyx_float_75_);
  Py_VISIT(traverse_module_state->__pyx_float_0_25);
  Py_VISIT(traverse_module_state->__pyx_float_100_);
  Py_VISIT(traverse_module_state->__pyx_float_126_);
  Py_VISIT(traverse_module_state->__pyx_float_127_);
  Py_VISIT(traverse_module_state->__pyx_float_128_);
  Py_VISIT(traverse_module_state->__pyx_float_129_);
  Py_VISIT(traverse_module_state->__pyx_float_180_);
  Py_VISIT(traverse_module_state->__pyx_float_253_);
  Py_VISIT(traverse_module_state->__pyx_float_254_);
  Py_VISIT(traverse_module_state->__pyx_float_256_);
  Py_VISIT(traverse_module_state->__pyx_float_360_);
  Py_VISIT(traverse_module_state->__pyx_float_600_);
  Py_VISIT(traverse_module_state->__pyx_float_100_0);
  Py_VISIT(traverse_module_state->__pyx_float_32768_);
  Py_VISIT(traverse_module_state->__pyx_float_65533_);
  Py_VISIT(traverse_module_state->__pyx_float_65534_);
  Py_VISIT(traverse_module_state->__pyx_float_neg_0_25);
  Py_VISIT(traverse_module_state->__pyx_float_neg_9999_0);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_int_2);
  Py_VISIT(traverse_module_state->__pyx_int_3);
  Py_VISIT(traverse_module_state->__pyx_int_4);
  Py_VISIT(traverse_module_state->__pyx_int_5);
  Py_VISIT(traverse_module_state->__pyx_int_6);
  Py_VISIT(traverse_module_state->__pyx_int_7);
  Py_VISIT(traverse_module_state->__pyx_int_8);
  Py_VISIT(traverse_module_state->__pyx_int_9);
  Py_VISIT(traverse_module_state->__pyx_int_10);
  Py_VISIT(traverse_module_state->__pyx_int_11);
  Py_VISIT(traverse_module_state->__pyx_int_12);
  Py_VISIT(traverse_module_state->__pyx_int_13);
  Py_VISIT(traverse_module_state->__pyx_int_14);
  Py_VISIT(traverse_module_state->__pyx_int_15);
  Py_VISIT(traverse_module_state->__pyx_int_16);
  Py_VISIT(traverse_module_state->__pyx_int_17);
  Py_VISIT(traverse_module_state->__pyx_int_18);
  Py_VISIT(traverse_module_state->__pyx_int_19);
  Py_VISIT(traverse_module_state->__pyx_int_20);
  Py_VISIT(traverse_module_state->__pyx_int_21);
  Py_VISIT(traverse_module_state->__pyx_int_22);
  Py_VISIT(traverse_module_state->__pyx_int_23);
  Py_VISIT(traverse_module_state->__pyx_int_24);
  Py_VISIT(traverse_module_state->__pyx_int_25);
  Py_VISIT(traverse_module_state->__pyx_int_26);
  Py_VISIT(traverse_module_state->__pyx_int_27);
  Py_VISIT(traverse_module_state->__pyx_int_28);
  Py_VISIT(traverse_module_state->__pyx_int_29);
  Py_VISIT(traverse_module_state->__pyx_int_30);
  Py_VISIT(traverse_module_state->__pyx_int_31);
  Py_VISIT(traverse_module_state->__pyx_int_32);
  Py_VISIT(traverse_module_state->__pyx_int_33);
  Py_VISIT(traverse_module_state->__pyx_int_34);
  Py_VISIT(traverse_module_state->__pyx_int_35);
  Py_VISIT(traverse_module_state->__pyx_int_36);
  Py_VISIT(traverse_module_state->__pyx_int_37);
  Py_VISIT(traverse_module_state->__pyx_int_38);
  Py_VISIT(traverse_module_state->__pyx_int_39);
  Py_VISIT(traverse_module_state->__pyx_int_40);
  Py_VISIT(traverse_module_state->__pyx_int_41);
  Py_VISIT(traverse_module_state->__pyx_int_42);
  Py_VISIT(traverse_module_state->__pyx_int_43);
  Py_VISIT(traverse_module_state->__pyx_int_44);
  Py_VISIT(traverse_module_state->__pyx_int_45);
  Py_VISIT(traverse_module_state->__pyx_int_46);
  Py_VISIT(traverse_module_state->__pyx_int_47);
  Py_VISIT(traverse_module_state->__pyx_int_48);
  Py_VISIT(traverse_module_state->__pyx_int_49);
  Py_VISIT(traverse_module_state->__pyx_int_50);
  Py_VISIT(traverse_module_state->__pyx_int_51);
  Py_VISIT(traverse_module_state->__pyx_int_52);
  Py_VISIT(traverse_module_state->__pyx_int_53);
  Py_VISIT(traverse_module_state->__pyx_int_54);
  Py_VISIT(traverse_module_state->__pyx_int_55);
  Py_VISIT(traverse_module_state->__pyx_int_56);
  Py_VISIT(traverse_module_state->__pyx_int_57);
  Py_VISIT(traverse_module_state->__pyx_int_58);
  Py_VISIT(traverse_module_state->__pyx_int_59);
  Py_VISIT(traverse_module_state->__pyx_int_60);
  Py_VISIT(traverse_module_state->__pyx_int_61);
  Py_VISIT(traverse_module_state->__pyx_int_62);
  Py_VISIT(traverse_module_state->__pyx_int_63);
  Py_VISIT(traverse_module_state->__pyx_int_64);
  Py_VISIT(traverse_module_state->__pyx_int_65);
  Py_VISIT(traverse_module_state->__pyx_int_66);
  Py_VISIT(traverse_module_state->__pyx_int_67);
  Py_VISIT(traverse_module_state->__pyx_int_68);
  Py_VISIT(traverse_module_state->__pyx_int_69);
  Py_VISIT(traverse_module_state->__pyx_int_70);
  Py_VISIT(traverse_module_state->__pyx_int_71);
  Py_VISIT(traverse_module_state->__pyx_int_72);
  Py_VISIT(traverse_module_state->__pyx_int_73);
  Py_VISIT(traverse_module_state->__pyx_int_74);
  Py_VISIT(traverse_module_state->__pyx_int_75);
  Py_VISIT(traverse_module_state->__pyx_int_76);
  Py_VISIT(traverse_module_state->__pyx_int_77);
  Py_VISIT(traverse_module_state->__pyx_int_78);
  Py_VISIT(traverse_module_state->__pyx_int_79);
  Py_VISIT(traverse_module_state->__pyx_int_80);
  Py_VISIT(traverse_module_state->__pyx_int_81);
  Py_VISIT(traverse_module_state->__pyx_int_82);
  Py_VISIT(traverse_module_state->__pyx_int_83);
  Py_VISIT(traverse_module_state->__pyx_int_84);
  Py_VISIT(traverse_module_state->__pyx_int_85);
  Py_VISIT(traverse_module_state->__pyx_int_86);
  Py_VISIT(traverse_module_state->__pyx_int_87);
  Py_VISIT(traverse_module_state->__pyx_int_88);
  Py_VISIT(traverse_module_state->__pyx_int_89);
  Py_VISIT(traverse_module_state->__pyx_int_90);
  Py_VISIT(traverse_module_state->__pyx_int_91);
  Py_VISIT(traverse_module_state->__pyx_int_92);
  Py_VISIT(traverse_module_state->__pyx_int_93);
  Py_VISIT(traverse_module_state->__pyx_int_94);
  Py_VISIT(traverse_module_state->__pyx_int_95);
  Py_VISIT(traverse_module_state->__pyx_int_96);
  Py_VISIT(traverse_module_state->__pyx_int_97);
  Py_VISIT(traverse_module_state->__pyx_int_98);
  Py_VISIT(traverse_module_state->__pyx_int_99);
  Py_VISIT(traverse_module_state->__pyx_int_100);
  Py_VISIT(traverse_module_state->__pyx_int_101);
  Py_VISIT(traverse_module_state->__pyx_int_102);
  Py_VISIT(traverse_module_state->__pyx_int_103);
  Py_VISIT(traverse_module_state->__pyx_int_104);
  Py_VISIT(traverse_module_state->__pyx_int_105);
  Py_VISIT(traverse_module_state->__pyx_int_106);
  Py_VISIT(traverse_module_state->__pyx_int_107);
  Py_VISIT(traverse_module_state->__pyx_int_108);
  Py_VISIT(traverse_module_state->__pyx_int_109);
  Py_VISIT(traverse_module_state->__pyx_int_110);
  Py_VISIT(traverse_module_state->__pyx_int_111);
  Py_VISIT(traverse_module_state->__pyx_int_112);
  Py_VISIT(traverse_module_state->__pyx_int_113);
  Py_VISIT(traverse_module_state->__pyx_int_114);
  Py_VISIT(traverse_module_state->__pyx_int_115);
  Py_VISIT(traverse_module_state->__pyx_int_116);
  Py_VISIT(traverse_module_state->__pyx_int_117);
  Py_VISIT(traverse_module_state->__pyx_int_118);
  Py_VISIT(traverse_module_state->__pyx_int_119);
  Py_VISIT(traverse_module_state->__pyx_int_120);
  Py_VISIT(traverse_module_state->__pyx_int_121);
  Py_VISIT(traverse_module_state->__pyx_int_122);
  Py_VISIT(traverse_module_state->__pyx_int_123);
  Py_VISIT(traverse_module_state->__pyx_int_124);
  Py_VISIT(traverse_module_state->__pyx_int_125);
  Py_VISIT(traverse_module_state->__pyx_int_126);
  Py_VISIT(traverse_module_state->__pyx_int_127);
  Py_VISIT(traverse_module_state->__pyx_int_128);
  Py_VISIT(traverse_module_state->__pyx_int_255);
  Py_VISIT(traverse_module_state->__pyx_int_640);
  Py_VISIT(traverse_module_state->__pyx_int_4884);
  Py_VISIT(traverse_module_state->__pyx_int_6144);
  Py_VISIT(traverse_module_state->__pyx_int_65536);
  Py_VISIT(traverse_module_state->__pyx_int_4294967296);
  Py_VISIT(traverse_module_state->__pyx_int_neg_1);
  Py_VISIT(traverse_module_state->__pyx_tuple_);
  Py_VISIT(traverse_module_state->__pyx_slice__4);
  Py_VISIT(traverse_module_state->__pyx_slice__5);
  Py_VISIT(traverse_module_state->__pyx_tuple__2);
  Py_VISIT(traverse_module_state->__pyx_tuple__3);
  Py_VISIT(traverse_module_state->__pyx_tuple__6);
  Py_VISIT(traverse_module_state->__pyx_tuple__7);
  Py_VISIT(traverse_module_state->__pyx_tuple__8);
  Py_VISIT(traverse_module_state->__pyx_tuple__9);
  Py_VISIT(traverse_module_state->__pyx_slice__12);
  Py_VISIT(traverse_module_state->__pyx_slice__13);
  Py_VISIT(traverse_module_state->__pyx_slice__16);
  Py_VISIT(traverse_module_state->__pyx_slice__17);
  Py_VISIT(traverse_module_state->__pyx_slice__18);
  Py_VISIT(traverse_module_state->__pyx_tuple__10);
  Py_VISIT(traverse_module_state->__pyx_tuple__11);
  Py_VISIT(traverse_module_state->__pyx_tuple__14);
  Py_VISIT(traverse_module_state->__pyx_tuple__20);
  Py_VISIT(traverse_module_state->__pyx_tuple__22);
  Py_VISIT(traverse_module_state->__pyx_tuple__24);
  Py_VISIT(traverse_module_state->__pyx_tuple__26);
  Py_VISIT(traverse_module_state->__pyx_tuple__27);
  Py_VISIT(traverse_module_state->__pyx_tuple__29);
  Py_VISIT(traverse_module_state->__pyx_tuple__31);
  Py_VISIT(traverse_module_state->__pyx_tuple__33);
  Py_VISIT(traverse_module_state->__pyx_tuple__35);
  Py_VISIT(traverse_module_state->__pyx_tuple__37);
  Py_VISIT(traverse_module_state->__pyx_tuple__39);
  Py_VISIT(traverse_module_state->__pyx_tuple__41);
  Py_VISIT(traverse_module_state->__pyx_tuple__43);
  Py_VISIT(traverse_module_state->__pyx_tuple__45);
  Py_VISIT(traverse_module_state->__pyx_tuple__47);
  Py_VISIT(traverse_module_state->__pyx_tuple__49);
  Py_VISIT(traverse_module_state->__pyx_tuple__51);
  Py_VISIT(traverse_module_state->__pyx_tuple__53);
  Py_VISIT(traverse_module_state->__pyx_tuple__55);
  Py_VISIT(traverse_module_state->__pyx_tuple__57);
  Py_VISIT(traverse_module_state->__pyx_tuple__59);
  Py_VISIT(traverse_module_state->__pyx_tuple__60);
  Py_VISIT(traverse_module_state->__pyx_tuple__61);
  Py_VISIT(traverse_module_state->__pyx_tuple__62);
  Py_VISIT(traverse_module_state->__pyx_tuple__63);
  Py_VISIT(traverse_module_state->__pyx_tuple__64);
  Py_VISIT(traverse_module_state->__pyx_tuple__65);
  Py_VISIT(traverse_module_state->__pyx_tuple__66);
  Py_VISIT(traverse_module_state->__pyx_tuple__67);
  Py_VISIT(traverse_module_state->__pyx_tuple__68);
  Py_VISIT(traverse_module_state->__pyx_tuple__69);
  Py_VISIT(traverse_module_state->__pyx_tuple__70);
  Py_VISIT(traverse_module_state->__pyx_tuple__71);
  Py_VISIT(traverse_module_state->__pyx_tuple__72);
  Py_VISIT(traverse_module_state->__pyx_tuple__73);
  Py_VISIT(traverse_module_state->__pyx_tuple__74);
  Py_VISIT(traverse_module_state->__pyx_tuple__75);
  Py_VISIT(traverse_module_state->__pyx_tuple__76);
  Py_VISIT(traverse_module_state->__pyx_tuple__77);
  Py_VISIT(traverse_module_state->__pyx_tuple__78);
  Py_VISIT(traverse_module_state->__pyx_tuple__79);
  Py_VISIT(traverse_module_state->__pyx_tuple__80);
  Py_VISIT(traverse_module_state->__pyx_tuple__81);
  Py_VISIT(traverse_module_state->__pyx_tuple__82);
  Py_VISIT(traverse_module_state->__pyx_tuple__83);
  Py_VISIT(traverse_module_state->__pyx_tuple__84);
  Py_VISIT(traverse_module_state->__pyx_tuple__85);
  Py_VISIT(traverse_module_state->__pyx_tuple__86);
  Py_VISIT(traverse_module_state->__pyx_tuple__87);
  Py_VISIT(traverse_module_state->__pyx_tuple__88);
  Py_VISIT(traverse_module_state->__pyx_tuple__89);
  Py_VISIT(traverse_module_state->__pyx_tuple__90);
  Py_VISIT(traverse_module_state->__pyx_tuple__91);
  Py_VISIT(traverse_module_state->__pyx_tuple__92);
  Py_VISIT(traverse_module_state->__pyx_tuple__93);
  Py_VISIT(traverse_module_state->__pyx_tuple__94);
  Py_VISIT(traverse_module_state->__pyx_tuple__95);
  Py_VISIT(traverse_module_state->__pyx_tuple__96);
  Py_VISIT(traverse_module_state->__pyx_tuple__97);
  Py_VISIT(traverse_module_state->__pyx_tuple__98);
  Py_VISIT(traverse_module_state->__pyx_tuple__99);
  Py_VISIT(traverse_module_state->__pyx_tuple__100);
  Py_VISIT(traverse_module_state->__pyx_tuple__101);
  Py_VISIT(traverse_module_state->__pyx_tuple__102);
  Py_VISIT(traverse_module_state->__pyx_tuple__103);
  Py_VISIT(traverse_module_state->__pyx_tuple__104);
  Py_VISIT(traverse_module_state->__pyx_tuple__105);
  Py_VISIT(traverse_module_state->__pyx_tuple__106);
  Py_VISIT(traverse_module_state->__pyx_tuple__107);
  Py_VISIT(traverse_module_state->__pyx_tuple__108);
  Py_VISIT(traverse_module_state->__pyx_tuple__109);
  Py_VISIT(traverse_module_state->__pyx_tuple__110);
  Py_VISIT(traverse_module_state->__pyx_tuple__111);
  Py_VISIT(traverse_module_state->__pyx_tuple__112);
  Py_VISIT(traverse_module_state->__pyx_tuple__113);
  Py_VISIT(traverse_module_state->__pyx_tuple__114);
  Py_VISIT(traverse_module_state->__pyx_tuple__115);
  Py_VISIT(traverse_module_state->__pyx_tuple__116);
  Py_VISIT(traverse_module_state->__pyx_tuple__117);
  Py_VISIT(traverse_module_state->__pyx_tuple__118);
  Py_VISIT(traverse_module_state->__pyx_tuple__119);
  Py_VISIT(traverse_module_state->__pyx_tuple__120);
  Py_VISIT(traverse_module_state->__pyx_tuple__121);
  Py_VISIT(traverse_module_state->__pyx_tuple__122);
  Py_VISIT(traverse_module_state->__pyx_tuple__123);
  Py_VISIT(traverse_module_state->__pyx_tuple__124);
  Py_VISIT(traverse_module_state->__pyx_tuple__125);
  Py_VISIT(traverse_module_state->__pyx_tuple__126);
  Py_VISIT(traverse_module_state->__pyx_tuple__127);
  Py_VISIT(traverse_module_state->__pyx_tuple__128);
  Py_VISIT(traverse_module_state->__pyx_tuple__129);
  Py_VISIT(traverse_module_state->__pyx_tuple__130);
  Py_VISIT(traverse_module_state->__pyx_tuple__131);
  Py_VISIT(traverse_module_state->__pyx_tuple__132);
  Py_VISIT(traverse_module_state->__pyx_tuple__133);
  Py_VISIT(traverse_module_state->__pyx_tuple__134);
  Py_VISIT(traverse_module_state->__pyx_tuple__135);
  Py_VISIT(traverse_module_state->__pyx_tuple__136);
  Py_VISIT(traverse_module_state->__pyx_tuple__137);
  Py_VISIT(traverse_module_state->__pyx_tuple__138);
  Py_VISIT(traverse_module_state->__pyx_tuple__139);
  Py_VISIT(traverse_module_state->__pyx_tuple__140);
  Py_VISIT(traverse_module_state->__pyx_tuple__141);
  Py_VISIT(traverse_module_state->__pyx_tuple__142);
  Py_VISIT(traverse_module_state->__pyx_tuple__143);
  Py_VISIT(traverse_module_state->__pyx_tuple__144);
  Py_VISIT(traverse_module_state->__pyx_tuple__145);
  Py_VISIT(traverse_module_state->__pyx_tuple__146);
  Py_VISIT(traverse_module_state->__pyx_tuple__147);
  Py_VISIT(traverse_module_state->__pyx_tuple__148);
  Py_VISIT(traverse_module_state->__pyx_codeobj__21);
  Py_VISIT(traverse_module_state->__pyx_codeobj__23);
  Py_VISIT(traverse_module_state->__pyx_codeobj__25);
  Py_VISIT(traverse_module_state->__pyx_codeobj__28);
  Py_VISIT(traverse_module_state->__pyx_codeobj__30);
  Py_VISIT(traverse_module_state->__pyx_codeobj__32);
  Py_VISIT(traverse_module_state->__pyx_codeobj__34);
  Py_VISIT(traverse_module_state->__pyx_codeobj__36);
  Py_VISIT(traverse_module_state->__pyx_codeobj__38);
  Py_VISIT(traverse_module_state->__pyx_codeobj__40);
  Py_VISIT(traverse_module_state->__pyx_codeobj__42);
  Py_VISIT(traverse_module_state->__pyx_codeobj__44);
  Py_VISIT(traverse_module_state->__pyx_codeobj__46);
  Py_VISIT(traverse_module_state->__pyx_codeobj__48);
  Py_VISIT(traverse_module_state->__pyx_codeobj__50);
  Py_VISIT(traverse_module_state->__pyx_codeobj__52);
  Py_VISIT(traverse_module_state->__pyx_codeobj__54);
  Py_VISIT(traverse_module_state->__pyx_codeobj__56);
  Py_VISIT(traverse_module_state->__pyx_codeobj__58);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5numpy_dtype __pyx_mstate_global->__pyx_ptype_5numpy_dtype
#define __pyx_ptype_5numpy_flatiter __pyx_mstate_global->__pyx_ptype_5numpy_flatiter
#define __pyx_ptype_5numpy_broadcast __pyx_mstate_global->__pyx_ptype_5numpy_broadcast
#define __pyx_ptype_5numpy_ndarray __pyx_mstate_global->__pyx_ptype_5numpy_ndarray
#define __pyx_ptype_5numpy_generic __pyx_mstate_global->__pyx_ptype_5numpy_generic
#define __pyx_ptype_5numpy_number __pyx_mstate_global->__pyx_ptype_5numpy_number
#define __pyx_ptype_5numpy_integer __pyx_mstate_global->__pyx_ptype_5numpy_integer
#define __pyx_ptype_5numpy_signedinteger __pyx_mstate_global->__pyx_ptype_5numpy_signedinteger
#define __pyx_ptype_5numpy_unsignedinteger __pyx_mstate_global->__pyx_ptype_5numpy_unsignedinteger
#define __pyx_ptype_5numpy_inexact __pyx_mstate_global->__pyx_ptype_5numpy_inexact
#define __pyx_ptype_5numpy_floating __pyx_mstate_global->__pyx_ptype_5numpy_floating
#define __pyx_ptype_5numpy_complexfloating __pyx_mstate_global->__pyx_ptype_5numpy_complexfloating
#define __pyx_ptype_5numpy_flexible __pyx_mstate_global->__pyx_ptype_5numpy_flexible
#define __pyx_ptype_5numpy_character __pyx_mstate_global->__pyx_ptype_5numpy_character
#define __pyx_ptype_5numpy_ufunc __pyx_mstate_global->__pyx_ptype_5numpy_ufunc
#if CYTHON_USE_MODULE_STATE
#define __pyx_type_5pyart_2io_11_sigmetfile_SigmetFile __pyx_mstate_global->__pyx_type_5pyart_2io_11_sigmetfile_SigmetFile
#endif
#define __pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile __pyx_mstate_global->__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile
#define __pyx_kp_u_112s __pyx_mstate_global->__pyx_kp_u_112s
#define __pyx_kp_u_115s __pyx_mstate_global->__pyx_kp_u_115s
#define __pyx_kp_u_120s __pyx_mstate_global->__pyx_kp_u_120s
#define __pyx_kp_u_128s __pyx_mstate_global->__pyx_kp_u_128s
#define __pyx_kp_u_12s __pyx_mstate_global->__pyx_kp_u_12s
#define __pyx_kp_u_136s __pyx_mstate_global->__pyx_kp_u_136s
#define __pyx_kp_u_160s __pyx_mstate_global->__pyx_kp_u_160s
#define __pyx_kp_u_16s __pyx_mstate_global->__pyx_kp_u_16s
#define __pyx_kp_u_184s __pyx_mstate_global->__pyx_kp_u_184s
#define __pyx_kp_u_18s __pyx_mstate_global->__pyx_kp_u_18s
#define __pyx_kp_u_198s __pyx_mstate_global->__pyx_kp_u_198s
#define __pyx_kp_u_1s __pyx_mstate_global->__pyx_kp_u_1s
#define __pyx_kp_u_2 __pyx_mstate_global->__pyx_kp_u_2
#define __pyx_kp_u_200s __pyx_mstate_global->__pyx_kp_u_200s
#define __pyx_kp_u_204s __pyx_mstate_global->__pyx_kp_u_204s
#define __pyx_kp_u_208s __pyx_mstate_global->__pyx_kp_u_208s
#define __pyx_kp_u_228s __pyx_mstate_global->__pyx_kp_u_228s
#define __pyx_kp_u_24s __pyx_mstate_global->__pyx_kp_u_24s
#define __pyx_kp_u_256s __pyx_mstate_global->__pyx_kp_u_256s
#define __pyx_kp_u_2612s __pyx_mstate_global->__pyx_kp_u_2612s
#define __pyx_kp_u_28s __pyx_mstate_global->__pyx_kp_u_28s
#define __pyx_kp_u_2_uint8 __pyx_mstate_global->__pyx_kp_u_2_uint8
#define __pyx_kp_u_2s __pyx_mstate_global->__pyx_kp_u_2s
#define __pyx_kp_u_308s __pyx_mstate_global->__pyx_kp_u_308s
#define __pyx_kp_u_320s __pyx_mstate_global->__pyx_kp_u_320s
#define __pyx_kp_u_32s __pyx_mstate_global->__pyx_kp_u_32s
#define __pyx_kp_u_36s __pyx_mstate_global->__pyx_kp_u_36s
#define __pyx_kp_u_40s __pyx_mstate_global->__pyx_kp_u_40s
#define __pyx_kp_u_480s __pyx_mstate_global->__pyx_kp_u_480s
#define __pyx_kp_u_48s __pyx_mstate_global->__pyx_kp_u_48s
#define __pyx_kp_u_4s __pyx_mstate_global->__pyx_kp_u_4s
#define __pyx_kp_u_52s __pyx_mstate_global->__pyx_kp_u_52s
#define __pyx_kp_u_6s __pyx_mstate_global->__pyx_kp_u_6s
#define __pyx_kp_u_720s __pyx_mstate_global->__pyx_kp_u_720s
#define __pyx_kp_u_732s __pyx_mstate_global->__pyx_kp_u_732s
#define __pyx_kp_u_80s __pyx_mstate_global->__pyx_kp_u_80s
#define __pyx_kp_u_8s __pyx_mstate_global->__pyx_kp_u_8s
#define __pyx_kp_u_920s __pyx_mstate_global->__pyx_kp_u_920s
#define __pyx_kp_u_94s __pyx_mstate_global->__pyx_kp_u_94s
#define __pyx_n_u_ALBEDO16 __pyx_mstate_global->__pyx_n_u_ALBEDO16
#define __pyx_n_u_ALBEDO8 __pyx_mstate_global->__pyx_n_u_ALBEDO8
#define __pyx_n_u_AXDIL2 __pyx_mstate_global->__pyx_n_u_AXDIL2
#define __pyx_n_u_B __pyx_mstate_global->__pyx_n_u_B
#define __pyx_n_s_BIN1 __pyx_mstate_global->__pyx_n_s_BIN1
#define __pyx_n_s_BIN2 __pyx_mstate_global->__pyx_n_s_BIN2
#define __pyx_n_s_BIN4 __pyx_mstate_global->__pyx_n_s_BIN4
#define __pyx_n_s_COLOR_SCALE_DEF __pyx_mstate_global->__pyx_n_s_COLOR_SCALE_DEF
#define __pyx_n_u_CSP16 __pyx_mstate_global->__pyx_n_u_CSP16
#define __pyx_n_u_CSP8 __pyx_mstate_global->__pyx_n_u_CSP8
#define __pyx_n_u_DBT __pyx_mstate_global->__pyx_n_u_DBT
#define __pyx_n_u_DBT2 __pyx_mstate_global->__pyx_n_u_DBT2
#define __pyx_n_u_DBTE16 __pyx_mstate_global->__pyx_n_u_DBTE16
#define __pyx_n_u_DBTE8 __pyx_mstate_global->__pyx_n_u_DBTE8
#define __pyx_n_u_DBTV16 __pyx_mstate_global->__pyx_n_u_DBTV16
#define __pyx_n_u_DBTV8 __pyx_mstate_global->__pyx_n_u_DBTV8
#define __pyx_n_u_DBZ __pyx_mstate_global->__pyx_n_u_DBZ
#define __pyx_n_u_DBZ2 __pyx_mstate_global->__pyx_n_u_DBZ2
#define __pyx_n_u_DBZC __pyx_mstate_global->__pyx_n_u_DBZC
#define __pyx_n_u_DBZC2 __pyx_mstate_global->__pyx_n_u_DBZC2
#define __pyx_n_u_DBZE16 __pyx_mstate_global->__pyx_n_u_DBZE16
#define __pyx_n_u_DBZE8 __pyx_mstate_global->__pyx_n_u_DBZE8
#define __pyx_n_u_DBZV16 __pyx_mstate_global->__pyx_n_u_DBZV16
#define __pyx_n_u_DBZV8 __pyx_mstate_global->__pyx_n_u_DBZV8
#define __pyx_n_u_DEFORM2 __pyx_mstate_global->__pyx_n_u_DEFORM2
#define __pyx_n_u_DIVERGE2 __pyx_mstate_global->__pyx_n_u_DIVERGE2
#define __pyx_n_s_DSP_DATA_MASK __pyx_mstate_global->__pyx_n_s_DSP_DATA_MASK
#define __pyx_n_u_FLIQUID2 __pyx_mstate_global->__pyx_n_u_FLIQUID2
#define __pyx_n_s_FLT4 __pyx_mstate_global->__pyx_n_s_FLT4
#define __pyx_n_s_FLT8 __pyx_mstate_global->__pyx_n_s_FLT8
#define __pyx_kp_u_File_truncated_or_corrupt_i_of_i __pyx_mstate_global->__pyx_kp_u_File_truncated_or_corrupt_i_of_i
#define __pyx_kp_u_Finished_loading_record __pyx_mstate_global->__pyx_kp_u_Finished_loading_record
#define __pyx_n_u_GMT_minute_offset_local __pyx_mstate_global->__pyx_n_u_GMT_minute_offset_local
#define __pyx_n_u_GMT_minute_offset_standard __pyx_mstate_global->__pyx_n_u_GMT_minute_offset_standard
#define __pyx_n_u_H __pyx_mstate_global->__pyx_n_u_H
#define __pyx_n_u_HCLASS __pyx_mstate_global->__pyx_n_u_HCLASS
#define __pyx_n_u_HCLASS2 __pyx_mstate_global->__pyx_n_u_HCLASS2
#define __pyx_n_u_HDIR2 __pyx_mstate_global->__pyx_n_u_HDIR2
#define __pyx_n_u_HEIGHT __pyx_mstate_global->__pyx_n_u_HEIGHT
#define __pyx_n_u_HVEL2 __pyx_mstate_global->__pyx_n_u_HVEL2
#define __pyx_n_u_I __pyx_mstate_global->__pyx_n_u_I
#define __pyx_n_s_INGEST_CONFIGURATION __pyx_mstate_global->__pyx_n_s_INGEST_CONFIGURATION
#define __pyx_n_s_INGEST_DATA_HEADER __pyx_mstate_global->__pyx_n_s_INGEST_DATA_HEADER
#define __pyx_n_s_INGEST_HEADER __pyx_mstate_global->__pyx_n_s_INGEST_HEADER
#define __pyx_n_s_ImportError __pyx_mstate_global->__pyx_n_s_ImportError
#define __pyx_n_u_KDP __pyx_mstate_global->__pyx_n_u_KDP
#define __pyx_n_u_KDP2 __pyx_mstate_global->__pyx_n_u_KDP2
#define __pyx_n_u_LDRH __pyx_mstate_global->__pyx_n_u_LDRH
#define __pyx_n_u_LDRH2 __pyx_mstate_global->__pyx_n_u_LDRH2
#define __pyx_n_u_LDRV __pyx_mstate_global->__pyx_n_u_LDRV
#define __pyx_n_u_LDRV2 __pyx_mstate_global->__pyx_n_u_LDRV2
#define __pyx_n_u_LOG16 __pyx_mstate_global->__pyx_n_u_LOG16
#define __pyx_n_u_LOG8 __pyx_mstate_global->__pyx_n_u_LOG8
#define __pyx_n_s_MESSAGE __pyx_mstate_global->__pyx_n_s_MESSAGE
#define __pyx_n_u_OTHER __pyx_mstate_global->__pyx_n_u_OTHER
#define __pyx_n_u_PHIDP __pyx_mstate_global->__pyx_n_u_PHIDP
#define __pyx_n_u_PHIDP2 __pyx_mstate_global->__pyx_n_u_PHIDP2
#define __pyx_n_u_PHIH __pyx_mstate_global->__pyx_n_u_PHIH
#define __pyx_n_u_PHIH2 __pyx_mstate_global->__pyx_n_u_PHIH2
#define __pyx_n_u_PHIV __pyx_mstate_global->__pyx_n_u_PHIV
#define __pyx_n_u_PHIV2 __pyx_mstate_global->__pyx_n_u_PHIV2
#define __pyx_n_u_PMI16 __pyx_mstate_global->__pyx_n_u_PMI16
#define __pyx_n_u_PMI8 __pyx_mstate_global->__pyx_n_u_PMI8
#define __pyx_n_s_PRODUCT_CONFIGURATION __pyx_mstate_global->__pyx_n_s_PRODUCT_CONFIGURATION
#define __pyx_n_s_PRODUCT_END __pyx_mstate_global->__pyx_n_s_PRODUCT_END
#define __pyx_n_s_PRODUCT_HDR __pyx_mstate_global->__pyx_n_s_PRODUCT_HDR
#define __pyx_n_u_RAINRATE2 __pyx_mstate_global->__pyx_n_u_RAINRATE2
#define __pyx_n_u_RAW __pyx_mstate_global->__pyx_n_u_RAW
#define __pyx_n_s_RAW_PROD_BHDR __pyx_mstate_global->__pyx_n_s_RAW_PROD_BHDR
#define __pyx_n_s_RECORD_SIZE __pyx_mstate_global->__pyx_n_s_RECORD_SIZE
#define __pyx_n_u_RHOH __pyx_mstate_global->__pyx_n_u_RHOH
#define __pyx_n_u_RHOH2 __pyx_mstate_global->__pyx_n_u_RHOH2
#define __pyx_n_u_RHOHV __pyx_mstate_global->__pyx_n_u_RHOHV
#define __pyx_n_u_RHOHV2 __pyx_mstate_global->__pyx_n_u_RHOHV2
#define __pyx_n_u_RHOV __pyx_mstate_global->__pyx_n_u_RHOV
#define __pyx_n_u_RHOV2 __pyx_mstate_global->__pyx_n_u_RHOV2
#define __pyx_kp_u_Reading_ray_i_of_i __pyx_mstate_global->__pyx_kp_u_Reading_ray_i_of_i
#define __pyx_n_u_SHEAR __pyx_mstate_global->__pyx_n_u_SHEAR
#define __pyx_n_s_SIGMET_DATA_TYPES __pyx_mstate_global->__pyx_n_s_SIGMET_DATA_TYPES
#define __pyx_n_s_SINT1 __pyx_mstate_global->__pyx_n_s_SINT1
#define __pyx_n_s_SINT2 __pyx_mstate_global->__pyx_n_s_SINT2
#define __pyx_n_s_SINT4 __pyx_mstate_global->__pyx_n_s_SINT4
#define __pyx_n_u_SNR16 __pyx_mstate_global->__pyx_n_u_SNR16
#define __pyx_n_u_SNR8 __pyx_mstate_global->__pyx_n_u_SNR8
#define __pyx_n_u_SQI __pyx_mstate_global->__pyx_n_u_SQI
#define __pyx_n_u_SQI2 __pyx_mstate_global->__pyx_n_u_SQI2
#define __pyx_n_s_STRUCTURE_HEADER __pyx_mstate_global->__pyx_n_s_STRUCTURE_HEADER
#define __pyx_n_s_SigmetFile __pyx_mstate_global->__pyx_n_s_SigmetFile
#define __pyx_n_s_SigmetFile___reduce_cython __pyx_mstate_global->__pyx_n_s_SigmetFile___reduce_cython
#define __pyx_n_s_SigmetFile___setstate_cython __pyx_mstate_global->__pyx_n_s_SigmetFile___setstate_cython
#define __pyx_n_s_SigmetFile__determine_data_types __pyx_mstate_global->__pyx_n_s_SigmetFile__determine_data_types
#define __pyx_n_s_SigmetFile__get_sweep __pyx_mstate_global->__pyx_n_s_SigmetFile__get_sweep
#define __pyx_n_s_SigmetFile_close __pyx_mstate_global->__pyx_n_s_SigmetFile_close
#define __pyx_n_s_SigmetFile_read_data __pyx_mstate_global->__pyx_n_s_SigmetFile_read_data
#define __pyx_n_s_TASK_CALIB_INFO __pyx_mstate_global->__pyx_n_s_TASK_CALIB_INFO
#define __pyx_n_s_TASK_CONFIGURATION __pyx_mstate_global->__pyx_n_s_TASK_CONFIGURATION
#define __pyx_n_s_TASK_DSP_INFO __pyx_mstate_global->__pyx_n_s_TASK_DSP_INFO
#define __pyx_n_s_TASK_DSP_MODE_BATCH __pyx_mstate_global->__pyx_n_s_TASK_DSP_MODE_BATCH
#define __pyx_n_s_TASK_END_INFO __pyx_mstate_global->__pyx_n_s_TASK_END_INFO
#define __pyx_n_s_TASK_FILE_SCAN_INFO __pyx_mstate_global->__pyx_n_s_TASK_FILE_SCAN_INFO
#define __pyx_n_s_TASK_MANUAL_SCAN_INFO __pyx_mstate_global->__pyx_n_s_TASK_MANUAL_SCAN_INFO
#define __pyx_n_s_TASK_MISC_INFO __pyx_mstate_global->__pyx_n_s_TASK_MISC_INFO
#define __pyx_n_s_TASK_PPI_SCAN_INFO __pyx_mstate_global->__pyx_n_s_TASK_PPI_SCAN_INFO
#define __pyx_n_s_TASK_RANGE_INFO __pyx_mstate_global->__pyx_n_s_TASK_RANGE_INFO
#define __pyx_n_s_TASK_RHI_SCAN_INFO __pyx_mstate_global->__pyx_n_s_TASK_RHI_SCAN_INFO
#define __pyx_n_s_TASK_SCAN_INFO __pyx_mstate_global->__pyx_n_s_TASK_SCAN_INFO
#define __pyx_n_s_TASK_SCHED_INFO __pyx_mstate_global->__pyx_n_s_TASK_SCHED_INFO
#define __pyx_n_u_TEMPERATURE16 __pyx_mstate_global->__pyx_n_u_TEMPERATURE16
#define __pyx_n_u_TIME2 __pyx_mstate_global->__pyx_n_u_TIME2
#define __pyx_n_u_TURB16 __pyx_mstate_global->__pyx_n_u_TURB16
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_n_s_UINT1 __pyx_mstate_global->__pyx_n_s_UINT1
#define __pyx_n_s_UINT16_T __pyx_mstate_global->__pyx_n_s_UINT16_T
#define __pyx_n_s_UINT2 __pyx_mstate_global->__pyx_n_s_UINT2
#define __pyx_n_s_UINT4 __pyx_mstate_global->__pyx_n_s_UINT4
#define __pyx_n_u_UNKNOWN_100 __pyx_mstate_global->__pyx_n_u_UNKNOWN_100
#define __pyx_n_u_UNKNOWN_101 __pyx_mstate_global->__pyx_n_u_UNKNOWN_101
#define __pyx_n_u_UNKNOWN_102 __pyx_mstate_global->__pyx_n_u_UNKNOWN_102
#define __pyx_n_u_UNKNOWN_103 __pyx_mstate_global->__pyx_n_u_UNKNOWN_103
#define __pyx_n_u_UNKNOWN_104 __pyx_mstate_global->__pyx_n_u_UNKNOWN_104
#define __pyx_n_u_UNKNOWN_105 __pyx_mstate_global->__pyx_n_u_UNKNOWN_105
#define __pyx_n_u_UNKNOWN_106 __pyx_mstate_global->__pyx_n_u_UNKNOWN_106
#define __pyx_n_u_UNKNOWN_107 __pyx_mstate_global->__pyx_n_u_UNKNOWN_107
#define __pyx_n_u_UNKNOWN_108 __pyx_mstate_global->__pyx_n_u_UNKNOWN_108
#define __pyx_n_u_UNKNOWN_109 __pyx_mstate_global->__pyx_n_u_UNKNOWN_109
#define __pyx_n_u_UNKNOWN_110 __pyx_mstate_global->__pyx_n_u_UNKNOWN_110
#define __pyx_n_u_UNKNOWN_112 __pyx_mstate_global->__pyx_n_u_UNKNOWN_112
#define __pyx_n_u_UNKNOWN_113 __pyx_mstate_global->__pyx_n_u_UNKNOWN_113
#define __pyx_n_u_UNKNOWN_114 __pyx_mstate_global->__pyx_n_u_UNKNOWN_114
#define __pyx_n_u_UNKNOWN_115 __pyx_mstate_global->__pyx_n_u_UNKNOWN_115
#define __pyx_n_u_UNKNOWN_116 __pyx_mstate_global->__pyx_n_u_UNKNOWN_116
#define __pyx_n_u_UNKNOWN_117 __pyx_mstate_global->__pyx_n_u_UNKNOWN_117
#define __pyx_n_u_UNKNOWN_118 __pyx_mstate_global->__pyx_n_u_UNKNOWN_118
#define __pyx_n_u_UNKNOWN_119 __pyx_mstate_global->__pyx_n_u_UNKNOWN_119
#define __pyx_n_u_UNKNOWN_120 __pyx_mstate_global->__pyx_n_u_UNKNOWN_120
#define __pyx_n_u_UNKNOWN_121 __pyx_mstate_global->__pyx_n_u_UNKNOWN_121
#define __pyx_n_u_UNKNOWN_122 __pyx_mstate_global->__pyx_n_u_UNKNOWN_122
#define __pyx_n_u_UNKNOWN_123 __pyx_mstate_global->__pyx_n_u_UNKNOWN_123
#define __pyx_n_u_UNKNOWN_124 __pyx_mstate_global->__pyx_n_u_UNKNOWN_124
#define __pyx_n_u_UNKNOWN_125 __pyx_mstate_global->__pyx_n_u_UNKNOWN_125
#define __pyx_n_u_UNKNOWN_126 __pyx_mstate_global->__pyx_n_u_UNKNOWN_126
#define __pyx_n_u_UNKNOWN_127 __pyx_mstate_global->__pyx_n_u_UNKNOWN_127
#define __pyx_n_u_UNKNOWN_29 __pyx_mstate_global->__pyx_n_u_UNKNOWN_29
#define __pyx_n_u_UNKNOWN_30 __pyx_mstate_global->__pyx_n_u_UNKNOWN_30
#define __pyx_n_u_UNKNOWN_31 __pyx_mstate_global->__pyx_n_u_UNKNOWN_31
#define __pyx_n_u_UNKNOWN_6 __pyx_mstate_global->__pyx_n_u_UNKNOWN_6
#define __pyx_n_u_UNKNOWN_81 __pyx_mstate_global->__pyx_n_u_UNKNOWN_81
#define __pyx_n_u_UNKNOWN_82 __pyx_mstate_global->__pyx_n_u_UNKNOWN_82
#define __pyx_n_u_UNKNOWN_83 __pyx_mstate_global->__pyx_n_u_UNKNOWN_83
#define __pyx_n_u_UNKNOWN_84 __pyx_mstate_global->__pyx_n_u_UNKNOWN_84
#define __pyx_n_u_UNKNOWN_85 __pyx_mstate_global->__pyx_n_u_UNKNOWN_85
#define __pyx_n_u_UNKNOWN_86 __pyx_mstate_global->__pyx_n_u_UNKNOWN_86
#define __pyx_n_u_UNKNOWN_87 __pyx_mstate_global->__pyx_n_u_UNKNOWN_87
#define __pyx_n_u_UNKNOWN_88 __pyx_mstate_global->__pyx_n_u_UNKNOWN_88
#define __pyx_n_u_UNKNOWN_89 __pyx_mstate_global->__pyx_n_u_UNKNOWN_89
#define __pyx_n_u_UNKNOWN_90 __pyx_mstate_global->__pyx_n_u_UNKNOWN_90
#define __pyx_n_u_UNKNOWN_91 __pyx_mstate_global->__pyx_n_u_UNKNOWN_91
#define __pyx_n_u_UNKNOWN_92 __pyx_mstate_global->__pyx_n_u_UNKNOWN_92
#define __pyx_n_u_UNKNOWN_93 __pyx_mstate_global->__pyx_n_u_UNKNOWN_93
#define __pyx_n_u_UNKNOWN_94 __pyx_mstate_global->__pyx_n_u_UNKNOWN_94
#define __pyx_n_u_UNKNOWN_95 __pyx_mstate_global->__pyx_n_u_UNKNOWN_95
#define __pyx_n_u_UNKNOWN_96 __pyx_mstate_global->__pyx_n_u_UNKNOWN_96
#define __pyx_n_u_UNKNOWN_97 __pyx_mstate_global->__pyx_n_u_UNKNOWN_97
#define __pyx_n_u_UNKNOWN_98 __pyx_mstate_global->__pyx_n_u_UNKNOWN_98
#define __pyx_n_u_UNKNOWN_99 __pyx_mstate_global->__pyx_n_u_UNKNOWN_99
#define __pyx_n_u_USER __pyx_mstate_global->__pyx_n_u_USER
#define __pyx_n_u_USER2 __pyx_mstate_global->__pyx_n_u_USER2
#define __pyx_kp_u_Unknown_type_s_returning_raw_dat __pyx_mstate_global->__pyx_kp_u_Unknown_type_s_returning_raw_dat
#define __pyx_n_u_VEL __pyx_mstate_global->__pyx_n_u_VEL
#define __pyx_n_u_VEL2 __pyx_mstate_global->__pyx_n_u_VEL2
#define __pyx_n_u_VELC __pyx_mstate_global->__pyx_n_u_VELC
#define __pyx_n_u_VELC2 __pyx_mstate_global->__pyx_n_u_VELC2
#define __pyx_n_u_VIL2 __pyx_mstate_global->__pyx_n_u_VIL2
#define __pyx_n_u_VILD16 __pyx_mstate_global->__pyx_n_u_VILD16
#define __pyx_n_u_VIR16 __pyx_mstate_global->__pyx_n_u_VIR16
#define __pyx_n_u_VVEL2 __pyx_mstate_global->__pyx_n_u_VVEL2
#define __pyx_n_u_WIDTH __pyx_mstate_global->__pyx_n_u_WIDTH
#define __pyx_n_u_WIDTH2 __pyx_mstate_global->__pyx_n_u_WIDTH2
#define __pyx_n_u_XHDR __pyx_mstate_global->__pyx_n_u_XHDR
#define __pyx_n_s_YMDS_TIME __pyx_mstate_global->__pyx_n_s_YMDS_TIME
#define __pyx_n_u_ZDR __pyx_mstate_global->__pyx_n_u_ZDR
#define __pyx_n_u_ZDR2 __pyx_mstate_global->__pyx_n_u_ZDR2
#define __pyx_n_u_ZDRC __pyx_mstate_global->__pyx_n_u_ZDRC
#define __pyx_n_u_ZDRC2 __pyx_mstate_global->__pyx_n_u_ZDRC2
#define __pyx_n_s__149 __pyx_mstate_global->__pyx_n_s__149
#define __pyx_kp_u__15 __pyx_mstate_global->__pyx_kp_u__15
#define __pyx_n_s__19 __pyx_mstate_global->__pyx_n_s__19
#define __pyx_n_u_agc_feedback_code __pyx_mstate_global->__pyx_n_u_agc_feedback_code
#define __pyx_n_u_altitude_radar __pyx_mstate_global->__pyx_n_u_altitude_radar
#define __pyx_n_u_angular_resolution_desired __pyx_mstate_global->__pyx_n_u_angular_resolution_desired
#define __pyx_n_u_antenna_offset_bow __pyx_mstate_global->__pyx_n_u_antenna_offset_bow
#define __pyx_n_u_antenna_offset_starboard __pyx_mstate_global->__pyx_n_u_antenna_offset_starboard
#define __pyx_n_u_antenna_offset_up __pyx_mstate_global->__pyx_n_u_antenna_offset_up
#define __pyx_n_u_antenna_scan_mode __pyx_mstate_global->__pyx_n_u_antenna_scan_mode
#define __pyx_n_s_append __pyx_mstate_global->__pyx_n_s_append
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_u_attenuation __pyx_mstate_global->__pyx_n_u_attenuation
#define __pyx_n_s_az0 __pyx_mstate_global->__pyx_n_s_az0
#define __pyx_n_s_az1 __pyx_mstate_global->__pyx_n_s_az1
#define __pyx_n_u_azimuth_0 __pyx_mstate_global->__pyx_n_u_azimuth_0
#define __pyx_n_u_azimuth_1 __pyx_mstate_global->__pyx_n_u_azimuth_1
#define __pyx_n_u_azimuth_list __pyx_mstate_global->__pyx_n_u_azimuth_list
#define __pyx_n_u_b __pyx_mstate_global->__pyx_n_u_b
#define __pyx_n_s_bin2 __pyx_mstate_global->__pyx_n_s_bin2
#define __pyx_n_s_bin2_to_angle __pyx_mstate_global->__pyx_n_s_bin2_to_angle
#define __pyx_n_s_bin4 __pyx_mstate_global->__pyx_n_s_bin4
#define __pyx_n_s_bin4_to_angle __pyx_mstate_global->__pyx_n_s_bin4_to_angle
#define __pyx_n_s_bit __pyx_mstate_global->__pyx_n_s_bit
#define __pyx_n_u_bit_per_bin __pyx_mstate_global->__pyx_n_u_bit_per_bin
#define __pyx_n_u_bool __pyx_mstate_global->__pyx_n_u_bool
#define __pyx_n_u_bytes_in_structure __pyx_mstate_global->__pyx_n_u_bytes_in_structure
#define __pyx_n_u_c __pyx_mstate_global->__pyx_n_u_c
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_close __pyx_mstate_global->__pyx_n_s_close
#define __pyx_n_u_clutter_correction_threshold __pyx_mstate_global->__pyx_n_u_clutter_correction_threshold
#define __pyx_n_u_clutter_filter __pyx_mstate_global->__pyx_n_u_clutter_filter
#define __pyx_n_u_clutter_filter_name __pyx_mstate_global->__pyx_n_u_clutter_filter_name
#define __pyx_n_u_cluttermap __pyx_mstate_global->__pyx_n_u_cluttermap
#define __pyx_n_u_cluttermap_flag __pyx_mstate_global->__pyx_n_u_cluttermap_flag
#define __pyx_n_u_color_scale_def __pyx_mstate_global->__pyx_n_u_color_scale_def
#define __pyx_n_u_comment_bytes __pyx_mstate_global->__pyx_n_u_comment_bytes
#define __pyx_n_u_comments __pyx_mstate_global->__pyx_n_u_comments
#define __pyx_n_u_configuration_name __pyx_mstate_global->__pyx_n_u_configuration_name
#define __pyx_n_s_convert_sigmet_data __pyx_mstate_global->__pyx_n_s_convert_sigmet_data
#define __pyx_n_s_copy __pyx_mstate_global->__pyx_n_s_copy
#define __pyx_n_u_corrected_reflectivity_threshold __pyx_mstate_global->__pyx_n_u_corrected_reflectivity_threshold
#define __pyx_n_u_current_data_type_mask __pyx_mstate_global->__pyx_n_u_current_data_type_mask
#define __pyx_n_u_custom_ray_header_name __pyx_mstate_global->__pyx_n_u_custom_ray_header_name
#define __pyx_n_u_customer_storage __pyx_mstate_global->__pyx_n_u_customer_storage
#define __pyx_n_s_d __pyx_mstate_global->__pyx_n_s_d
#define __pyx_n_u_d __pyx_mstate_global->__pyx_n_u_d
#define __pyx_n_s_data __pyx_mstate_global->__pyx_n_s_data
#define __pyx_n_s_data_type __pyx_mstate_global->__pyx_n_s_data_type
#define __pyx_n_u_data_type __pyx_mstate_global->__pyx_n_u_data_type
#define __pyx_n_s_data_type_name __pyx_mstate_global->__pyx_n_s_data_type_name
#define __pyx_n_s_data_types __pyx_mstate_global->__pyx_n_s_data_types
#define __pyx_n_s_data_types_from_mask __pyx_mstate_global->__pyx_n_s_data_types_from_mask
#define __pyx_n_s_datetime __pyx_mstate_global->__pyx_n_s_datetime
#define __pyx_n_u_day __pyx_mstate_global->__pyx_n_u_day
#define __pyx_n_s_debug __pyx_mstate_global->__pyx_n_s_debug
#define __pyx_n_s_determine_data_types __pyx_mstate_global->__pyx_n_s_determine_data_types
#define __pyx_n_s_dic __pyx_mstate_global->__pyx_n_s_dic
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_u_dsp_type __pyx_mstate_global->__pyx_n_u_dsp_type
#define __pyx_n_s_dtype __pyx_mstate_global->__pyx_n_s_dtype
#define __pyx_n_u_dual_prf_delay __pyx_mstate_global->__pyx_n_u_dual_prf_delay
#define __pyx_n_u_earth_radius __pyx_mstate_global->__pyx_n_u_earth_radius
#define __pyx_n_s_el0 __pyx_mstate_global->__pyx_n_s_el0
#define __pyx_n_s_el1 __pyx_mstate_global->__pyx_n_s_el1
#define __pyx_n_u_elevation_0 __pyx_mstate_global->__pyx_n_u_elevation_0
#define __pyx_n_u_elevation_1 __pyx_mstate_global->__pyx_n_u_elevation_1
#define __pyx_n_u_elevation_list __pyx_mstate_global->__pyx_n_u_elevation_list
#define __pyx_n_s_empty __pyx_mstate_global->__pyx_n_s_empty
#define __pyx_n_s_empty_like __pyx_mstate_global->__pyx_n_s_empty_like
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_enumerate __pyx_mstate_global->__pyx_n_s_enumerate
#define __pyx_n_s_exp __pyx_mstate_global->__pyx_n_s_exp
#define __pyx_n_u_extended_header_type __pyx_mstate_global->__pyx_n_u_extended_header_type
#define __pyx_n_u_extended_product_header_offset __pyx_mstate_global->__pyx_n_u_extended_product_header_offset
#define __pyx_n_u_extended_ray_header_bytes __pyx_mstate_global->__pyx_n_u_extended_ray_header_bytes
#define __pyx_n_u_f __pyx_mstate_global->__pyx_n_u_f
#define __pyx_n_u_fault_status __pyx_mstate_global->__pyx_n_u_fault_status
#define __pyx_n_u_file_ingest_time __pyx_mstate_global->__pyx_n_u_file_ingest_time
#define __pyx_n_s_filename __pyx_mstate_global->__pyx_n_s_filename
#define __pyx_n_u_filename __pyx_mstate_global->__pyx_n_u_filename
#define __pyx_n_s_fill_value __pyx_mstate_global->__pyx_n_s_fill_value
#define __pyx_n_u_first_azimuth __pyx_mstate_global->__pyx_n_u_first_azimuth
#define __pyx_n_u_first_bin_range __pyx_mstate_global->__pyx_n_u_first_bin_range
#define __pyx_n_u_first_elevation __pyx_mstate_global->__pyx_n_u_first_elevation
#define __pyx_n_u_first_ray_index __pyx_mstate_global->__pyx_n_u_first_ray_index
#define __pyx_n_u_first_ray_offset __pyx_mstate_global->__pyx_n_u_first_ray_offset
#define __pyx_n_u_fixed_angle __pyx_mstate_global->__pyx_n_u_fixed_angle
#define __pyx_n_u_flag __pyx_mstate_global->__pyx_n_u_flag
#define __pyx_n_u_flags __pyx_mstate_global->__pyx_n_u_flags
#define __pyx_n_u_flags2 __pyx_mstate_global->__pyx_n_u_flags2
#define __pyx_n_u_float32 __pyx_mstate_global->__pyx_n_u_float32
#define __pyx_n_s_fmt __pyx_mstate_global->__pyx_n_s_fmt
#define __pyx_n_u_format_version __pyx_mstate_global->__pyx_n_u_format_version
#define __pyx_n_s_frombuffer __pyx_mstate_global->__pyx_n_s_frombuffer
#define __pyx_n_s_full_xhdr __pyx_mstate_global->__pyx_n_s_full_xhdr
#define __pyx_n_u_gain_control_flag __pyx_mstate_global->__pyx_n_u_gain_control_flag
#define __pyx_n_u_gas_attenuation __pyx_mstate_global->__pyx_n_u_gas_attenuation
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_u_generation_time __pyx_mstate_global->__pyx_n_u_generation_time
#define __pyx_n_s_get_sweep __pyx_mstate_global->__pyx_n_s_get_sweep
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_n_u_gmt_offset_minutes_local __pyx_mstate_global->__pyx_n_u_gmt_offset_minutes_local
#define __pyx_n_u_gmt_offset_minutes_standard __pyx_mstate_global->__pyx_n_u_gmt_offset_minutes_standard
#define __pyx_n_u_gparam_bytes __pyx_mstate_global->__pyx_n_u_gparam_bytes
#define __pyx_n_u_gparm __pyx_mstate_global->__pyx_n_u_gparm
#define __pyx_n_u_ground_height __pyx_mstate_global->__pyx_n_u_ground_height
#define __pyx_n_u_h __pyx_mstate_global->__pyx_n_u_h
#define __pyx_n_u_hardware_site __pyx_mstate_global->__pyx_n_u_hardware_site
#define __pyx_n_s_header_dic __pyx_mstate_global->__pyx_n_s_header_dic
#define __pyx_n_u_height_radar __pyx_mstate_global->__pyx_n_u_height_radar
#define __pyx_n_u_height_site __pyx_mstate_global->__pyx_n_u_height_site
#define __pyx_n_u_horizontal_beamwidth __pyx_mstate_global->__pyx_n_u_horizontal_beamwidth
#define __pyx_n_u_horizontal_calibration_i0 __pyx_mstate_global->__pyx_n_u_horizontal_calibration_i0
#define __pyx_n_u_horizontal_calibration_noise __pyx_mstate_global->__pyx_n_u_horizontal_calibration_noise
#define __pyx_n_u_horizontal_current_noise __pyx_mstate_global->__pyx_n_u_horizontal_current_noise
#define __pyx_n_u_horizontal_i0_calibration __pyx_mstate_global->__pyx_n_u_horizontal_i0_calibration
#define __pyx_n_u_horizontal_noise_calibration __pyx_mstate_global->__pyx_n_u_horizontal_noise_calibration
#define __pyx_n_u_horizontal_radar_constant __pyx_mstate_global->__pyx_n_u_horizontal_radar_constant
#define __pyx_n_s_i __pyx_mstate_global->__pyx_n_s_i
#define __pyx_n_u_i __pyx_mstate_global->__pyx_n_u_i
#define __pyx_n_u_i4 __pyx_mstate_global->__pyx_n_u_i4
#define __pyx_n_u_icolcnt __pyx_mstate_global->__pyx_n_u_icolcnt
#define __pyx_n_s_idh __pyx_mstate_global->__pyx_n_s_idh
#define __pyx_n_u_iflags __pyx_mstate_global->__pyx_n_u_iflags
#define __pyx_n_u_ilevel_seams __pyx_mstate_global->__pyx_n_u_ilevel_seams
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_ingest_configuration __pyx_mstate_global->__pyx_n_s_ingest_configuration
#define __pyx_n_u_ingest_configuration __pyx_mstate_global->__pyx_n_u_ingest_configuration
#define __pyx_n_s_ingest_data_hdrs __pyx_mstate_global->__pyx_n_s_ingest_data_hdrs
#define __pyx_n_s_ingest_data_headers __pyx_mstate_global->__pyx_n_s_ingest_data_headers
#define __pyx_n_u_ingest_hardware_name __pyx_mstate_global->__pyx_n_u_ingest_hardware_name
#define __pyx_n_s_ingest_header __pyx_mstate_global->__pyx_n_s_ingest_header
#define __pyx_n_u_ingest_iris_version __pyx_mstate_global->__pyx_n_u_ingest_iris_version
#define __pyx_n_u_ingest_site_name __pyx_mstate_global->__pyx_n_u_ingest_site_name
#define __pyx_n_u_ingest_time __pyx_mstate_global->__pyx_n_u_ingest_time
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_u_input_data_type __pyx_mstate_global->__pyx_n_u_input_data_type
#define __pyx_n_u_input_mask __pyx_mstate_global->__pyx_n_u_input_mask
#define __pyx_n_u_int16 __pyx_mstate_global->__pyx_n_u_int16
#define __pyx_n_u_int32 __pyx_mstate_global->__pyx_n_u_int32
#define __pyx_n_u_inverse_flatting __pyx_mstate_global->__pyx_n_u_inverse_flatting
#define __pyx_n_u_iris_version __pyx_mstate_global->__pyx_n_u_iris_version
#define __pyx_n_u_iris_version_created __pyx_mstate_global->__pyx_n_u_iris_version_created
#define __pyx_n_s_is_bit_set __pyx_mstate_global->__pyx_n_s_is_bit_set
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_u_iset_and_scale __pyx_mstate_global->__pyx_n_u_iset_and_scale
#define __pyx_n_u_istart __pyx_mstate_global->__pyx_n_u_istart
#define __pyx_n_u_istep __pyx_mstate_global->__pyx_n_u_istep
#define __pyx_n_s_j __pyx_mstate_global->__pyx_n_s_j
#define __pyx_n_s_k __pyx_mstate_global->__pyx_n_s_k
#define __pyx_n_s_key __pyx_mstate_global->__pyx_n_s_key
#define __pyx_n_s_l __pyx_mstate_global->__pyx_n_s_l
#define __pyx_n_u_last_bin_range __pyx_mstate_global->__pyx_n_u_last_bin_range
#define __pyx_n_u_last_run_day __pyx_mstate_global->__pyx_n_u_last_run_day
#define __pyx_n_u_last_run_time __pyx_mstate_global->__pyx_n_u_last_run_time
#define __pyx_n_u_latitude __pyx_mstate_global->__pyx_n_u_latitude
#define __pyx_n_u_latitude_projection __pyx_mstate_global->__pyx_n_u_latitude_projection
#define __pyx_n_u_latitude_radar __pyx_mstate_global->__pyx_n_u_latitude_radar
#define __pyx_n_u_ldr_bias __pyx_mstate_global->__pyx_n_u_ldr_bias
#define __pyx_n_u_ldr_offset __pyx_mstate_global->__pyx_n_u_ldr_offset
#define __pyx_n_s_lead_record __pyx_mstate_global->__pyx_n_s_lead_record
#define __pyx_n_u_left_azimuth_limit __pyx_mstate_global->__pyx_n_u_left_azimuth_limit
#define __pyx_n_s_like_dbt __pyx_mstate_global->__pyx_n_s_like_dbt
#define __pyx_n_s_like_dbt2 __pyx_mstate_global->__pyx_n_s_like_dbt2
#define __pyx_n_s_like_sqi __pyx_mstate_global->__pyx_n_s_like_sqi
#define __pyx_n_s_like_sqi2 __pyx_mstate_global->__pyx_n_s_like_sqi2
#define __pyx_n_u_linear_filter_first_bin __pyx_mstate_global->__pyx_n_u_linear_filter_first_bin
#define __pyx_n_u_local_timezone __pyx_mstate_global->__pyx_n_u_local_timezone
#define __pyx_n_u_log_filter_first_bin __pyx_mstate_global->__pyx_n_u_log_filter_first_bin
#define __pyx_n_u_longitude __pyx_mstate_global->__pyx_n_u_longitude
#define __pyx_n_u_longitude_projection __pyx_mstate_global->__pyx_n_u_longitude_projection
#define __pyx_n_u_longitude_radar __pyx_mstate_global->__pyx_n_u_longitude_radar
#define __pyx_n_u_low_prf_factional __pyx_mstate_global->__pyx_n_u_low_prf_factional
#define __pyx_n_u_low_prf_hz __pyx_mstate_global->__pyx_n_u_low_prf_hz
#define __pyx_n_u_low_prf_range_averaging __pyx_mstate_global->__pyx_n_u_low_prf_range_averaging
#define __pyx_n_u_low_prf_sample_size __pyx_mstate_global->__pyx_n_u_low_prf_sample_size
#define __pyx_n_u_lower_elevation_limit __pyx_mstate_global->__pyx_n_u_lower_elevation_limit
#define __pyx_n_s_ma __pyx_mstate_global->__pyx_n_s_ma
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_u_major_mode __pyx_mstate_global->__pyx_n_u_major_mode
#define __pyx_n_s_mask __pyx_mstate_global->__pyx_n_s_mask
#define __pyx_n_u_mask_word_0 __pyx_mstate_global->__pyx_n_u_mask_word_0
#define __pyx_n_u_mask_word_1 __pyx_mstate_global->__pyx_n_u_mask_word_1
#define __pyx_n_u_mask_word_2 __pyx_mstate_global->__pyx_n_u_mask_word_2
#define __pyx_n_u_mask_word_3 __pyx_mstate_global->__pyx_n_u_mask_word_3
#define __pyx_n_u_mask_word_4 __pyx_mstate_global->__pyx_n_u_mask_word_4
#define __pyx_n_s_masked_array __pyx_mstate_global->__pyx_n_s_masked_array
#define __pyx_n_u_maximum_range __pyx_mstate_global->__pyx_n_u_maximum_range
#define __pyx_n_u_mean_wind_direction __pyx_mstate_global->__pyx_n_u_mean_wind_direction
#define __pyx_n_u_mean_wind_speed __pyx_mstate_global->__pyx_n_u_mean_wind_speed
#define __pyx_n_u_melting_layer __pyx_mstate_global->__pyx_n_u_melting_layer
#define __pyx_n_u_melting_level __pyx_mstate_global->__pyx_n_u_melting_level
#define __pyx_n_s_mess __pyx_mstate_global->__pyx_n_s_mess
#define __pyx_n_s_metadata __pyx_mstate_global->__pyx_n_s_metadata
#define __pyx_n_u_milliseconds __pyx_mstate_global->__pyx_n_u_milliseconds
#define __pyx_n_u_minor_task_suffix __pyx_mstate_global->__pyx_n_u_minor_task_suffix
#define __pyx_n_s_mod __pyx_mstate_global->__pyx_n_s_mod
#define __pyx_n_u_month __pyx_mstate_global->__pyx_n_u_month
#define __pyx_n_s_multi_prf_flag __pyx_mstate_global->__pyx_n_s_multi_prf_flag
#define __pyx_n_u_multi_prf_flag __pyx_mstate_global->__pyx_n_u_multi_prf_flag
#define __pyx_n_s_multiplier __pyx_mstate_global->__pyx_n_s_multiplier
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_name_2 __pyx_mstate_global->__pyx_n_s_name_2
#define __pyx_n_s_nbin __pyx_mstate_global->__pyx_n_s_nbin
#define __pyx_n_s_nbins __pyx_mstate_global->__pyx_n_s_nbins
#define __pyx_n_u_nbins __pyx_mstate_global->__pyx_n_u_nbins
#define __pyx_n_s_ndata __pyx_mstate_global->__pyx_n_s_ndata
#define __pyx_n_s_ndata_types __pyx_mstate_global->__pyx_n_s_ndata_types
#define __pyx_n_u_nexrad_clutter_skip __pyx_mstate_global->__pyx_n_u_nexrad_clutter_skip
#define __pyx_n_u_nexrad_clutter_threshold __pyx_mstate_global->__pyx_n_u_nexrad_clutter_threshold
#define __pyx_n_s_np __pyx_mstate_global->__pyx_n_s_np
#define __pyx_n_s_nray_data_types __pyx_mstate_global->__pyx_n_s_nray_data_types
#define __pyx_n_s_nrays __pyx_mstate_global->__pyx_n_s_nrays
#define __pyx_n_s_nsweeps __pyx_mstate_global->__pyx_n_s_nsweeps
#define __pyx_n_s_number __pyx_mstate_global->__pyx_n_s_number
#define __pyx_n_u_number_bins __pyx_mstate_global->__pyx_n_u_number_bins
#define __pyx_n_u_number_elements __pyx_mstate_global->__pyx_n_u_number_elements
#define __pyx_n_u_number_files __pyx_mstate_global->__pyx_n_u_number_files
#define __pyx_n_u_number_ingest __pyx_mstate_global->__pyx_n_u_number_ingest
#define __pyx_n_u_number_input_bins __pyx_mstate_global->__pyx_n_u_number_input_bins
#define __pyx_n_u_number_linear_filter __pyx_mstate_global->__pyx_n_u_number_linear_filter
#define __pyx_n_u_number_log_filter __pyx_mstate_global->__pyx_n_u_number_log_filter
#define __pyx_n_u_number_output_bins __pyx_mstate_global->__pyx_n_u_number_output_bins
#define __pyx_n_u_number_rays_file_actual __pyx_mstate_global->__pyx_n_u_number_rays_file_actual
#define __pyx_n_u_number_rays_file_expected __pyx_mstate_global->__pyx_n_u_number_rays_file_expected
#define __pyx_n_u_number_rays_sweep __pyx_mstate_global->__pyx_n_u_number_rays_sweep
#define __pyx_n_u_number_sweeps __pyx_mstate_global->__pyx_n_u_number_sweeps
#define __pyx_n_u_number_sweeps_completed __pyx_mstate_global->__pyx_n_u_number_sweeps_completed
#define __pyx_n_u_number_task_config_table __pyx_mstate_global->__pyx_n_u_number_task_config_table
#define __pyx_n_u_number_tasks __pyx_mstate_global->__pyx_n_u_number_tasks
#define __pyx_n_s_numpy __pyx_mstate_global->__pyx_n_s_numpy
#define __pyx_kp_u_numpy__core_multiarray_failed_to __pyx_mstate_global->__pyx_kp_u_numpy__core_multiarray_failed_to
#define __pyx_kp_u_numpy__core_umath_failed_to_impo __pyx_mstate_global->__pyx_kp_u_numpy__core_umath_failed_to_impo
#define __pyx_n_s_nyquist __pyx_mstate_global->__pyx_n_s_nyquist
#define __pyx_n_s_offset __pyx_mstate_global->__pyx_n_s_offset
#define __pyx_n_s_ones __pyx_mstate_global->__pyx_n_s_ones
#define __pyx_n_s_open __pyx_mstate_global->__pyx_n_s_open
#define __pyx_n_u_original_data_type_mask __pyx_mstate_global->__pyx_n_u_original_data_type_mask
#define __pyx_n_s_out __pyx_mstate_global->__pyx_n_s_out
#define __pyx_n_s_parse_ray_headers __pyx_mstate_global->__pyx_n_s_parse_ray_headers
#define __pyx_n_u_playback_flag __pyx_mstate_global->__pyx_n_u_playback_flag
#define __pyx_n_u_playback_version __pyx_mstate_global->__pyx_n_u_playback_version
#define __pyx_n_u_polarization __pyx_mstate_global->__pyx_n_u_polarization
#define __pyx_n_u_polarization_type __pyx_mstate_global->__pyx_n_u_polarization_type
#define __pyx_n_s_power __pyx_mstate_global->__pyx_n_s_power
#define __pyx_n_u_power_threshold __pyx_mstate_global->__pyx_n_u_power_threshold
#define __pyx_n_u_prf __pyx_mstate_global->__pyx_n_u_prf
#define __pyx_n_s_prf_flag __pyx_mstate_global->__pyx_n_s_prf_flag
#define __pyx_n_u_prf_flag __pyx_mstate_global->__pyx_n_u_prf_flag
#define __pyx_n_s_print __pyx_mstate_global->__pyx_n_s_print
#define __pyx_n_s_product_config __pyx_mstate_global->__pyx_n_s_product_config
#define __pyx_n_u_product_configuration __pyx_mstate_global->__pyx_n_u_product_configuration
#define __pyx_n_s_product_end __pyx_mstate_global->__pyx_n_s_product_end
#define __pyx_n_u_product_end __pyx_mstate_global->__pyx_n_u_product_end
#define __pyx_n_s_product_hdr __pyx_mstate_global->__pyx_n_s_product_hdr
#define __pyx_n_u_product_name __pyx_mstate_global->__pyx_n_u_product_name
#define __pyx_n_u_product_sequence_number __pyx_mstate_global->__pyx_n_u_product_sequence_number
#define __pyx_n_u_product_specific_bytes __pyx_mstate_global->__pyx_n_u_product_specific_bytes
#define __pyx_n_u_product_type_code __pyx_mstate_global->__pyx_n_u_product_type_code
#define __pyx_n_u_projection_name __pyx_mstate_global->__pyx_n_u_projection_name
#define __pyx_n_u_projection_type __pyx_mstate_global->__pyx_n_u_projection_type
#define __pyx_n_s_prt_value __pyx_mstate_global->__pyx_n_s_prt_value
#define __pyx_n_u_pulse_width __pyx_mstate_global->__pyx_n_u_pulse_width
#define __pyx_n_s_pyart_io__sigmetfile __pyx_mstate_global->__pyx_n_s_pyart_io__sigmetfile
#define __pyx_kp_s_pyart_io__sigmetfile_pyx __pyx_mstate_global->__pyx_kp_s_pyart_io__sigmetfile_pyx
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_u_radar_height __pyx_mstate_global->__pyx_n_u_radar_height
#define __pyx_n_u_radar_height_above_reference __pyx_mstate_global->__pyx_n_u_radar_height_above_reference
#define __pyx_n_u_radial_smoother __pyx_mstate_global->__pyx_n_u_radial_smoother
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_u_range_bin_averaging_flag __pyx_mstate_global->__pyx_n_u_range_bin_averaging_flag
#define __pyx_n_s_raw_data __pyx_mstate_global->__pyx_n_s_raw_data
#define __pyx_n_s_raw_prod_bhdr __pyx_mstate_global->__pyx_n_s_raw_prod_bhdr
#define __pyx_n_s_raw_sweep_data __pyx_mstate_global->__pyx_n_s_raw_sweep_data
#define __pyx_n_u_ray_header_bytes __pyx_mstate_global->__pyx_n_u_ray_header_bytes
#define __pyx_n_u_ray_header_mask __pyx_mstate_global->__pyx_n_u_ray_header_mask
#define __pyx_n_s_ray_headers __pyx_mstate_global->__pyx_n_s_ray_headers
#define __pyx_n_s_ray_i __pyx_mstate_global->__pyx_n_s_ray_i
#define __pyx_n_s_ray_nbins __pyx_mstate_global->__pyx_n_s_ray_nbins
#define __pyx_n_u_ray_number __pyx_mstate_global->__pyx_n_u_ray_number
#define __pyx_n_s_ray_time __pyx_mstate_global->__pyx_n_s_ray_time
#define __pyx_n_u_rb __pyx_mstate_global->__pyx_n_u_rb
#define __pyx_n_s_read __pyx_mstate_global->__pyx_n_s_read
#define __pyx_n_u_read __pyx_mstate_global->__pyx_n_u_read
#define __pyx_n_s_read_data __pyx_mstate_global->__pyx_n_s_read_data
#define __pyx_n_u_reciever_bandwidth __pyx_mstate_global->__pyx_n_u_reciever_bandwidth
#define __pyx_n_s_record __pyx_mstate_global->__pyx_n_s_record
#define __pyx_n_u_record_number __pyx_mstate_global->__pyx_n_u_record_number
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_u_reflectivity_calibration __pyx_mstate_global->__pyx_n_u_reflectivity_calibration
#define __pyx_n_u_reflectivity_noise_threshold __pyx_mstate_global->__pyx_n_u_reflectivity_noise_threshold
#define __pyx_n_u_reflectivity_slope __pyx_mstate_global->__pyx_n_u_reflectivity_slope
#define __pyx_n_u_reflectivity_unfolding_threshold __pyx_mstate_global->__pyx_n_u_reflectivity_unfolding_threshold
#define __pyx_n_u_reserved __pyx_mstate_global->__pyx_n_u_reserved
#define __pyx_n_s_reshape __pyx_mstate_global->__pyx_n_s_reshape
#define __pyx_n_u_resolution_rays __pyx_mstate_global->__pyx_n_u_resolution_rays
#define __pyx_n_u_right_azimuth_limit __pyx_mstate_global->__pyx_n_u_right_azimuth_limit
#define __pyx_n_u_sample_size __pyx_mstate_global->__pyx_n_u_sample_size
#define __pyx_n_u_samples_used __pyx_mstate_global->__pyx_n_u_samples_used
#define __pyx_n_u_scheduling_code __pyx_mstate_global->__pyx_n_u_scheduling_code
#define __pyx_n_u_seconds __pyx_mstate_global->__pyx_n_u_seconds
#define __pyx_n_u_seconds_between_runs __pyx_mstate_global->__pyx_n_u_seconds_between_runs
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_kp_s_self__rbuf_p_cannot_be_converted __pyx_mstate_global->__pyx_kp_s_self__rbuf_p_cannot_be_converted
#define __pyx_kp_u_self__rbuf_pos_is __pyx_mstate_global->__pyx_kp_u_self__rbuf_pos_is
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_shape __pyx_mstate_global->__pyx_n_s_shape
#define __pyx_n_s_shrink __pyx_mstate_global->__pyx_n_s_shrink
#define __pyx_n_u_signal_processor_type __pyx_mstate_global->__pyx_n_u_signal_processor_type
#define __pyx_n_u_site_name __pyx_mstate_global->__pyx_n_u_site_name
#define __pyx_n_u_skip_time __pyx_mstate_global->__pyx_n_u_skip_time
#define __pyx_n_u_spare_0 __pyx_mstate_global->__pyx_n_u_spare_0
#define __pyx_n_u_spare_1 __pyx_mstate_global->__pyx_n_u_spare_1
#define __pyx_n_u_spare_2 __pyx_mstate_global->__pyx_n_u_spare_2
#define __pyx_n_u_spare_3 __pyx_mstate_global->__pyx_n_u_spare_3
#define __pyx_n_u_spare_4 __pyx_mstate_global->__pyx_n_u_spare_4
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_n_u_sqi_threshold __pyx_mstate_global->__pyx_n_u_sqi_threshold
#define __pyx_n_s_sqrt __pyx_mstate_global->__pyx_n_s_sqrt
#define __pyx_n_u_standard_parallel_1 __pyx_mstate_global->__pyx_n_u_standard_parallel_1
#define __pyx_n_u_standard_parallel_2 __pyx_mstate_global->__pyx_n_u_standard_parallel_2
#define __pyx_n_u_start_first_section_flag __pyx_mstate_global->__pyx_n_u_start_first_section_flag
#define __pyx_n_u_start_first_sector_flag __pyx_mstate_global->__pyx_n_u_start_first_sector_flag
#define __pyx_n_u_start_time __pyx_mstate_global->__pyx_n_u_start_time
#define __pyx_n_u_step_input_bins __pyx_mstate_global->__pyx_n_u_step_input_bins
#define __pyx_n_u_step_output_bins __pyx_mstate_global->__pyx_n_u_step_output_bins
#define __pyx_n_u_stop_time __pyx_mstate_global->__pyx_n_u_stop_time
#define __pyx_n_s_string __pyx_mstate_global->__pyx_n_s_string
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_struct __pyx_mstate_global->__pyx_n_s_struct
#define __pyx_n_s_structure __pyx_mstate_global->__pyx_n_s_structure
#define __pyx_n_u_structure_header __pyx_mstate_global->__pyx_n_u_structure_header
#define __pyx_n_u_structure_identifier __pyx_mstate_global->__pyx_n_u_structure_identifier
#define __pyx_n_s_sum __pyx_mstate_global->__pyx_n_s_sum
#define __pyx_n_s_sweep_data __pyx_mstate_global->__pyx_n_s_sweep_data
#define __pyx_n_u_sweep_ingest_time __pyx_mstate_global->__pyx_n_u_sweep_ingest_time
#define __pyx_n_s_sweep_metadata __pyx_mstate_global->__pyx_n_s_sweep_metadata
#define __pyx_n_u_sweep_number __pyx_mstate_global->__pyx_n_u_sweep_number
#define __pyx_n_u_sweep_start_time __pyx_mstate_global->__pyx_n_u_sweep_start_time
#define __pyx_n_u_task_calib_info __pyx_mstate_global->__pyx_n_u_task_calib_info
#define __pyx_n_s_task_config __pyx_mstate_global->__pyx_n_s_task_config
#define __pyx_n_s_task_configuration __pyx_mstate_global->__pyx_n_s_task_configuration
#define __pyx_n_u_task_configuration __pyx_mstate_global->__pyx_n_u_task_configuration
#define __pyx_n_u_task_configuration_file_name __pyx_mstate_global->__pyx_n_u_task_configuration_file_name
#define __pyx_n_u_task_data_time __pyx_mstate_global->__pyx_n_u_task_data_time
#define __pyx_n_u_task_description __pyx_mstate_global->__pyx_n_u_task_description
#define __pyx_n_s_task_dsp_info __pyx_mstate_global->__pyx_n_s_task_dsp_info
#define __pyx_n_u_task_dsp_info __pyx_mstate_global->__pyx_n_u_task_dsp_info
#define __pyx_n_u_task_dsp_mode __pyx_mstate_global->__pyx_n_u_task_dsp_mode
#define __pyx_n_s_task_end_info __pyx_mstate_global->__pyx_n_s_task_end_info
#define __pyx_n_u_task_end_info __pyx_mstate_global->__pyx_n_u_task_end_info
#define __pyx_n_u_task_major_number __pyx_mstate_global->__pyx_n_u_task_major_number
#define __pyx_n_u_task_minor_number __pyx_mstate_global->__pyx_n_u_task_minor_number
#define __pyx_n_u_task_misc_info __pyx_mstate_global->__pyx_n_u_task_misc_info
#define __pyx_n_u_task_name __pyx_mstate_global->__pyx_n_u_task_name
#define __pyx_n_u_task_range_info __pyx_mstate_global->__pyx_n_u_task_range_info
#define __pyx_n_u_task_scan_info __pyx_mstate_global->__pyx_n_u_task_scan_info
#define __pyx_n_u_task_scan_type_scan_info __pyx_mstate_global->__pyx_n_u_task_scan_type_scan_info
#define __pyx_n_u_task_sched_info __pyx_mstate_global->__pyx_n_u_task_sched_info
#define __pyx_n_u_task_state __pyx_mstate_global->__pyx_n_u_task_state
#define __pyx_n_u_tcf_cal_flags_1 __pyx_mstate_global->__pyx_n_u_tcf_cal_flags_1
#define __pyx_n_u_tcf_cal_flags_2 __pyx_mstate_global->__pyx_n_u_tcf_cal_flags_2
#define __pyx_n_s_temp __pyx_mstate_global->__pyx_n_s_temp
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_time __pyx_mstate_global->__pyx_n_s_time
#define __pyx_n_u_time __pyx_mstate_global->__pyx_n_u_time
#define __pyx_n_u_time_used_last_run __pyx_mstate_global->__pyx_n_u_time_used_last_run
#define __pyx_n_u_times_run __pyx_mstate_global->__pyx_n_u_times_run
#define __pyx_n_u_total_size __pyx_mstate_global->__pyx_n_u_total_size
#define __pyx_n_u_tr_serial_number __pyx_mstate_global->__pyx_n_u_tr_serial_number
#define __pyx_n_u_tranmitter_phase_sequence __pyx_mstate_global->__pyx_n_u_tranmitter_phase_sequence
#define __pyx_n_u_transmit_power __pyx_mstate_global->__pyx_n_u_transmit_power
#define __pyx_n_u_trigger_rate __pyx_mstate_global->__pyx_n_u_trigger_rate
#define __pyx_n_u_trucation_height __pyx_mstate_global->__pyx_n_u_trucation_height
#define __pyx_n_u_truncation_height __pyx_mstate_global->__pyx_n_u_truncation_height
#define __pyx_n_u_tz_name __pyx_mstate_global->__pyx_n_u_tz_name
#define __pyx_n_u_uint16 __pyx_mstate_global->__pyx_n_u_uint16
#define __pyx_n_s_uint8 __pyx_mstate_global->__pyx_n_s_uint8
#define __pyx_n_u_uncorrected_reflectivity_thresho __pyx_mstate_global->__pyx_n_u_uncorrected_reflectivity_thresho
#define __pyx_n_s_unpack __pyx_mstate_global->__pyx_n_s_unpack
#define __pyx_n_s_unpack_ingest_data_header __pyx_mstate_global->__pyx_n_s_unpack_ingest_data_header
#define __pyx_n_s_unpack_ingest_data_headers __pyx_mstate_global->__pyx_n_s_unpack_ingest_data_headers
#define __pyx_n_s_unpack_ingest_header __pyx_mstate_global->__pyx_n_s_unpack_ingest_header
#define __pyx_n_s_unpack_key __pyx_mstate_global->__pyx_n_s_unpack_key
#define __pyx_n_s_unpack_product_hdr __pyx_mstate_global->__pyx_n_s_unpack_product_hdr
#define __pyx_n_s_unpack_raw_prod_bhdr __pyx_mstate_global->__pyx_n_s_unpack_raw_prod_bhdr
#define __pyx_n_s_unpack_structure __pyx_mstate_global->__pyx_n_s_unpack_structure
#define __pyx_n_u_upper_elevation_limit __pyx_mstate_global->__pyx_n_u_upper_elevation_limit
#define __pyx_n_u_variable_range_bin_flag __pyx_mstate_global->__pyx_n_u_variable_range_bin_flag
#define __pyx_n_u_velocity_east __pyx_mstate_global->__pyx_n_u_velocity_east
#define __pyx_n_u_velocity_north __pyx_mstate_global->__pyx_n_u_velocity_north
#define __pyx_n_u_velocity_threshold_flags __pyx_mstate_global->__pyx_n_u_velocity_threshold_flags
#define __pyx_n_u_velocity_unfolding_threshold __pyx_mstate_global->__pyx_n_u_velocity_unfolding_threshold
#define __pyx_n_u_velocity_up __pyx_mstate_global->__pyx_n_u_velocity_up
#define __pyx_n_u_vertical_beamwidth __pyx_mstate_global->__pyx_n_u_vertical_beamwidth
#define __pyx_n_u_vertical_current_noise __pyx_mstate_global->__pyx_n_u_vertical_current_noise
#define __pyx_n_u_vertical_i0_calibration __pyx_mstate_global->__pyx_n_u_vertical_i0_calibration
#define __pyx_n_u_vertical_noise_calibration __pyx_mstate_global->__pyx_n_u_vertical_noise_calibration
#define __pyx_n_u_vertical_radar_constant __pyx_mstate_global->__pyx_n_u_vertical_radar_constant
#define __pyx_n_s_view __pyx_mstate_global->__pyx_n_s_view
#define __pyx_n_u_volume_scan_start_time __pyx_mstate_global->__pyx_n_u_volume_scan_start_time
#define __pyx_n_s_warn __pyx_mstate_global->__pyx_n_s_warn
#define __pyx_n_s_warnings __pyx_mstate_global->__pyx_n_s_warnings
#define __pyx_n_u_wavelength __pyx_mstate_global->__pyx_n_u_wavelength
#define __pyx_n_s_wavelength_cm __pyx_mstate_global->__pyx_n_s_wavelength_cm
#define __pyx_n_u_width_threshold_flags __pyx_mstate_global->__pyx_n_u_width_threshold_flags
#define __pyx_n_u_width_unfolding_threshold __pyx_mstate_global->__pyx_n_u_width_unfolding_threshold
#define __pyx_n_s_word0 __pyx_mstate_global->__pyx_n_s_word0
#define __pyx_n_s_word1 __pyx_mstate_global->__pyx_n_s_word1
#define __pyx_n_s_word2 __pyx_mstate_global->__pyx_n_s_word2
#define __pyx_n_s_word3 __pyx_mstate_global->__pyx_n_s_word3
#define __pyx_n_u_x_location __pyx_mstate_global->__pyx_n_u_x_location
#define __pyx_n_u_x_scale __pyx_mstate_global->__pyx_n_u_x_scale
#define __pyx_n_u_x_size __pyx_mstate_global->__pyx_n_u_x_size
#define __pyx_n_u_x_smoother __pyx_mstate_global->__pyx_n_u_x_smoother
#define __pyx_n_s_xrange __pyx_mstate_global->__pyx_n_s_xrange
#define __pyx_n_u_y_location __pyx_mstate_global->__pyx_n_u_y_location
#define __pyx_n_u_y_scale __pyx_mstate_global->__pyx_n_u_y_scale
#define __pyx_n_u_y_size __pyx_mstate_global->__pyx_n_u_y_size
#define __pyx_n_u_y_smoother __pyx_mstate_global->__pyx_n_u_y_smoother
#define __pyx_n_u_year __pyx_mstate_global->__pyx_n_u_year
#define __pyx_n_u_z_location __pyx_mstate_global->__pyx_n_u_z_location
#define __pyx_n_u_z_scale __pyx_mstate_global->__pyx_n_u_z_scale
#define __pyx_n_u_z_size __pyx_mstate_global->__pyx_n_u_z_size
#define __pyx_n_u_zdr_bias __pyx_mstate_global->__pyx_n_u_zdr_bias
#define __pyx_n_u_zdr_offset __pyx_mstate_global->__pyx_n_u_zdr_offset
#define __pyx_n_u_zdr_threshold_flags __pyx_mstate_global->__pyx_n_u_zdr_threshold_flags
#define __pyx_n_s_zeros_like __pyx_mstate_global->__pyx_n_s_zeros_like
#define __pyx_n_s_zip __pyx_mstate_global->__pyx_n_s_zip
#define __pyx_n_u_zr_constant __pyx_mstate_global->__pyx_n_u_zr_constant
#define __pyx_n_u_zr_exponent __pyx_mstate_global->__pyx_n_u_zr_exponent
#define __pyx_float_1_ __pyx_mstate_global->__pyx_float_1_
#define __pyx_float_2_ __pyx_mstate_global->__pyx_float_2_
#define __pyx_float_16_ __pyx_mstate_global->__pyx_float_16_
#define __pyx_float_64_ __pyx_mstate_global->__pyx_float_64_
#define __pyx_float_75_ __pyx_mstate_global->__pyx_float_75_
#define __pyx_float_0_25 __pyx_mstate_global->__pyx_float_0_25
#define __pyx_float_100_ __pyx_mstate_global->__pyx_float_100_
#define __pyx_float_126_ __pyx_mstate_global->__pyx_float_126_
#define __pyx_float_127_ __pyx_mstate_global->__pyx_float_127_
#define __pyx_float_128_ __pyx_mstate_global->__pyx_float_128_
#define __pyx_float_129_ __pyx_mstate_global->__pyx_float_129_
#define __pyx_float_180_ __pyx_mstate_global->__pyx_float_180_
#define __pyx_float_253_ __pyx_mstate_global->__pyx_float_253_
#define __pyx_float_254_ __pyx_mstate_global->__pyx_float_254_
#define __pyx_float_256_ __pyx_mstate_global->__pyx_float_256_
#define __pyx_float_360_ __pyx_mstate_global->__pyx_float_360_
#define __pyx_float_600_ __pyx_mstate_global->__pyx_float_600_
#define __pyx_float_100_0 __pyx_mstate_global->__pyx_float_100_0
#define __pyx_float_32768_ __pyx_mstate_global->__pyx_float_32768_
#define __pyx_float_65533_ __pyx_mstate_global->__pyx_float_65533_
#define __pyx_float_65534_ __pyx_mstate_global->__pyx_float_65534_
#define __pyx_float_neg_0_25 __pyx_mstate_global->__pyx_float_neg_0_25
#define __pyx_float_neg_9999_0 __pyx_mstate_global->__pyx_float_neg_9999_0
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_int_2 __pyx_mstate_global->__pyx_int_2
#define __pyx_int_3 __pyx_mstate_global->__pyx_int_3
#define __pyx_int_4 __pyx_mstate_global->__pyx_int_4
#define __pyx_int_5 __pyx_mstate_global->__pyx_int_5
#define __pyx_int_6 __pyx_mstate_global->__pyx_int_6
#define __pyx_int_7 __pyx_mstate_global->__pyx_int_7
#define __pyx_int_8 __pyx_mstate_global->__pyx_int_8
#define __pyx_int_9 __pyx_mstate_global->__pyx_int_9
#define __pyx_int_10 __pyx_mstate_global->__pyx_int_10
#define __pyx_int_11 __pyx_mstate_global->__pyx_int_11
#define __pyx_int_12 __pyx_mstate_global->__pyx_int_12
#define __pyx_int_13 __pyx_mstate_global->__pyx_int_13
#define __pyx_int_14 __pyx_mstate_global->__pyx_int_14
#define __pyx_int_15 __pyx_mstate_global->__pyx_int_15
#define __pyx_int_16 __pyx_mstate_global->__pyx_int_16
#define __pyx_int_17 __pyx_mstate_global->__pyx_int_17
#define __pyx_int_18 __pyx_mstate_global->__pyx_int_18
#define __pyx_int_19 __pyx_mstate_global->__pyx_int_19
#define __pyx_int_20 __pyx_mstate_global->__pyx_int_20
#define __pyx_int_21 __pyx_mstate_global->__pyx_int_21
#define __pyx_int_22 __pyx_mstate_global->__pyx_int_22
#define __pyx_int_23 __pyx_mstate_global->__pyx_int_23
#define __pyx_int_24 __pyx_mstate_global->__pyx_int_24
#define __pyx_int_25 __pyx_mstate_global->__pyx_int_25
#define __pyx_int_26 __pyx_mstate_global->__pyx_int_26
#define __pyx_int_27 __pyx_mstate_global->__pyx_int_27
#define __pyx_int_28 __pyx_mstate_global->__pyx_int_28
#define __pyx_int_29 __pyx_mstate_global->__pyx_int_29
#define __pyx_int_30 __pyx_mstate_global->__pyx_int_30
#define __pyx_int_31 __pyx_mstate_global->__pyx_int_31
#define __pyx_int_32 __pyx_mstate_global->__pyx_int_32
#define __pyx_int_33 __pyx_mstate_global->__pyx_int_33
#define __pyx_int_34 __pyx_mstate_global->__pyx_int_34
#define __pyx_int_35 __pyx_mstate_global->__pyx_int_35
#define __pyx_int_36 __pyx_mstate_global->__pyx_int_36
#define __pyx_int_37 __pyx_mstate_global->__pyx_int_37
#define __pyx_int_38 __pyx_mstate_global->__pyx_int_38
#define __pyx_int_39 __pyx_mstate_global->__pyx_int_39
#define __pyx_int_40 __pyx_mstate_global->__pyx_int_40
#define __pyx_int_41 __pyx_mstate_global->__pyx_int_41
#define __pyx_int_42 __pyx_mstate_global->__pyx_int_42
#define __pyx_int_43 __pyx_mstate_global->__pyx_int_43
#define __pyx_int_44 __pyx_mstate_global->__pyx_int_44
#define __pyx_int_45 __pyx_mstate_global->__pyx_int_45
#define __pyx_int_46 __pyx_mstate_global->__pyx_int_46
#define __pyx_int_47 __pyx_mstate_global->__pyx_int_47
#define __pyx_int_48 __pyx_mstate_global->__pyx_int_48
#define __pyx_int_49 __pyx_mstate_global->__pyx_int_49
#define __pyx_int_50 __pyx_mstate_global->__pyx_int_50
#define __pyx_int_51 __pyx_mstate_global->__pyx_int_51
#define __pyx_int_52 __pyx_mstate_global->__pyx_int_52
#define __pyx_int_53 __pyx_mstate_global->__pyx_int_53
#define __pyx_int_54 __pyx_mstate_global->__pyx_int_54
#define __pyx_int_55 __pyx_mstate_global->__pyx_int_55
#define __pyx_int_56 __pyx_mstate_global->__pyx_int_56
#define __pyx_int_57 __pyx_mstate_global->__pyx_int_57
#define __pyx_int_58 __pyx_mstate_global->__pyx_int_58
#define __pyx_int_59 __pyx_mstate_global->__pyx_int_59
#define __pyx_int_60 __pyx_mstate_global->__pyx_int_60
#define __pyx_int_61 __pyx_mstate_global->__pyx_int_61
#define __pyx_int_62 __pyx_mstate_global->__pyx_int_62
#define __pyx_int_63 __pyx_mstate_global->__pyx_int_63
#define __pyx_int_64 __pyx_mstate_global->__pyx_int_64
#define __pyx_int_65 __pyx_mstate_global->__pyx_int_65
#define __pyx_int_66 __pyx_mstate_global->__pyx_int_66
#define __pyx_int_67 __pyx_mstate_global->__pyx_int_67
#define __pyx_int_68 __pyx_mstate_global->__pyx_int_68
#define __pyx_int_69 __pyx_mstate_global->__pyx_int_69
#define __pyx_int_70 __pyx_mstate_global->__pyx_int_70
#define __pyx_int_71 __pyx_mstate_global->__pyx_int_71
#define __pyx_int_72 __pyx_mstate_global->__pyx_int_72
#define __pyx_int_73 __pyx_mstate_global->__pyx_int_73
#define __pyx_int_74 __pyx_mstate_global->__pyx_int_74
#define __pyx_int_75 __pyx_mstate_global->__pyx_int_75
#define __pyx_int_76 __pyx_mstate_global->__pyx_int_76
#define __pyx_int_77 __pyx_mstate_global->__pyx_int_77
#define __pyx_int_78 __pyx_mstate_global->__pyx_int_78
#define __pyx_int_79 __pyx_mstate_global->__pyx_int_79
#define __pyx_int_80 __pyx_mstate_global->__pyx_int_80
#define __pyx_int_81 __pyx_mstate_global->__pyx_int_81
#define __pyx_int_82 __pyx_mstate_global->__pyx_int_82
#define __pyx_int_83 __pyx_mstate_global->__pyx_int_83
#define __pyx_int_84 __pyx_mstate_global->__pyx_int_84
#define __pyx_int_85 __pyx_mstate_global->__pyx_int_85
#define __pyx_int_86 __pyx_mstate_global->__pyx_int_86
#define __pyx_int_87 __pyx_mstate_global->__pyx_int_87
#define __pyx_int_88 __pyx_mstate_global->__pyx_int_88
#define __pyx_int_89 __pyx_mstate_global->__pyx_int_89
#define __pyx_int_90 __pyx_mstate_global->__pyx_int_90
#define __pyx_int_91 __pyx_mstate_global->__pyx_int_91
#define __pyx_int_92 __pyx_mstate_global->__pyx_int_92
#define __pyx_int_93 __pyx_mstate_global->__pyx_int_93
#define __pyx_int_94 __pyx_mstate_global->__pyx_int_94
#define __pyx_int_95 __pyx_mstate_global->__pyx_int_95
#define __pyx_int_96 __pyx_mstate_global->__pyx_int_96
#define __pyx_int_97 __pyx_mstate_global->__pyx_int_97
#define __pyx_int_98 __pyx_mstate_global->__pyx_int_98
#define __pyx_int_99 __pyx_mstate_global->__pyx_int_99
#define __pyx_int_100 __pyx_mstate_global->__pyx_int_100
#define __pyx_int_101 __pyx_mstate_global->__pyx_int_101
#define __pyx_int_102 __pyx_mstate_global->__pyx_int_102
#define __pyx_int_103 __pyx_mstate_global->__pyx_int_103
#define __pyx_int_104 __pyx_mstate_global->__pyx_int_104
#define __pyx_int_105 __pyx_mstate_global->__pyx_int_105
#define __pyx_int_106 __pyx_mstate_global->__pyx_int_106
#define __pyx_int_107 __pyx_mstate_global->__pyx_int_107
#define __pyx_int_108 __pyx_mstate_global->__pyx_int_108
#define __pyx_int_109 __pyx_mstate_global->__pyx_int_109
#define __pyx_int_110 __pyx_mstate_global->__pyx_int_110
#define __pyx_int_111 __pyx_mstate_global->__pyx_int_111
#define __pyx_int_112 __pyx_mstate_global->__pyx_int_112
#define __pyx_int_113 __pyx_mstate_global->__pyx_int_113
#define __pyx_int_114 __pyx_mstate_global->__pyx_int_114
#define __pyx_int_115 __pyx_mstate_global->__pyx_int_115
#define __pyx_int_116 __pyx_mstate_global->__pyx_int_116
#define __pyx_int_117 __pyx_mstate_global->__pyx_int_117
#define __pyx_int_118 __pyx_mstate_global->__pyx_int_118
#define __pyx_int_119 __pyx_mstate_global->__pyx_int_119
#define __pyx_int_120 __pyx_mstate_global->__pyx_int_120
#define __pyx_int_121 __pyx_mstate_global->__pyx_int_121
#define __pyx_int_122 __pyx_mstate_global->__pyx_int_122
#define __pyx_int_123 __pyx_mstate_global->__pyx_int_123
#define __pyx_int_124 __pyx_mstate_global->__pyx_int_124
#define __pyx_int_125 __pyx_mstate_global->__pyx_int_125
#define __pyx_int_126 __pyx_mstate_global->__pyx_int_126
#define __pyx_int_127 __pyx_mstate_global->__pyx_int_127
#define __pyx_int_128 __pyx_mstate_global->__pyx_int_128
#define __pyx_int_255 __pyx_mstate_global->__pyx_int_255
#define __pyx_int_640 __pyx_mstate_global->__pyx_int_640
#define __pyx_int_4884 __pyx_mstate_global->__pyx_int_4884
#define __pyx_int_6144 __pyx_mstate_global->__pyx_int_6144
#define __pyx_int_65536 __pyx_mstate_global->__pyx_int_65536
#define __pyx_int_4294967296 __pyx_mstate_global->__pyx_int_4294967296
#define __pyx_int_neg_1 __pyx_mstate_global->__pyx_int_neg_1
#define __pyx_tuple_ __pyx_mstate_global->__pyx_tuple_
#define __pyx_slice__4 __pyx_mstate_global->__pyx_slice__4
#define __pyx_slice__5 __pyx_mstate_global->__pyx_slice__5
#define __pyx_tuple__2 __pyx_mstate_global->__pyx_tuple__2
#define __pyx_tuple__3 __pyx_mstate_global->__pyx_tuple__3
#define __pyx_tuple__6 __pyx_mstate_global->__pyx_tuple__6
#define __pyx_tuple__7 __pyx_mstate_global->__pyx_tuple__7
#define __pyx_tuple__8 __pyx_mstate_global->__pyx_tuple__8
#define __pyx_tuple__9 __pyx_mstate_global->__pyx_tuple__9
#define __pyx_slice__12 __pyx_mstate_global->__pyx_slice__12
#define __pyx_slice__13 __pyx_mstate_global->__pyx_slice__13
#define __pyx_slice__16 __pyx_mstate_global->__pyx_slice__16
#define __pyx_slice__17 __pyx_mstate_global->__pyx_slice__17
#define __pyx_slice__18 __pyx_mstate_global->__pyx_slice__18
#define __pyx_tuple__10 __pyx_mstate_global->__pyx_tuple__10
#define __pyx_tuple__11 __pyx_mstate_global->__pyx_tuple__11
#define __pyx_tuple__14 __pyx_mstate_global->__pyx_tuple__14
#define __pyx_tuple__20 __pyx_mstate_global->__pyx_tuple__20
#define __pyx_tuple__22 __pyx_mstate_global->__pyx_tuple__22
#define __pyx_tuple__24 __pyx_mstate_global->__pyx_tuple__24
#define __pyx_tuple__26 __pyx_mstate_global->__pyx_tuple__26
#define __pyx_tuple__27 __pyx_mstate_global->__pyx_tuple__27
#define __pyx_tuple__29 __pyx_mstate_global->__pyx_tuple__29
#define __pyx_tuple__31 __pyx_mstate_global->__pyx_tuple__31
#define __pyx_tuple__33 __pyx_mstate_global->__pyx_tuple__33
#define __pyx_tuple__35 __pyx_mstate_global->__pyx_tuple__35
#define __pyx_tuple__37 __pyx_mstate_global->__pyx_tuple__37
#define __pyx_tuple__39 __pyx_mstate_global->__pyx_tuple__39
#define __pyx_tuple__41 __pyx_mstate_global->__pyx_tuple__41
#define __pyx_tuple__43 __pyx_mstate_global->__pyx_tuple__43
#define __pyx_tuple__45 __pyx_mstate_global->__pyx_tuple__45
#define __pyx_tuple__47 __pyx_mstate_global->__pyx_tuple__47
#define __pyx_tuple__49 __pyx_mstate_global->__pyx_tuple__49
#define __pyx_tuple__51 __pyx_mstate_global->__pyx_tuple__51
#define __pyx_tuple__53 __pyx_mstate_global->__pyx_tuple__53
#define __pyx_tuple__55 __pyx_mstate_global->__pyx_tuple__55
#define __pyx_tuple__57 __pyx_mstate_global->__pyx_tuple__57
#define __pyx_tuple__59 __pyx_mstate_global->__pyx_tuple__59
#define __pyx_tuple__60 __pyx_mstate_global->__pyx_tuple__60
#define __pyx_tuple__61 __pyx_mstate_global->__pyx_tuple__61
#define __pyx_tuple__62 __pyx_mstate_global->__pyx_tuple__62
#define __pyx_tuple__63 __pyx_mstate_global->__pyx_tuple__63
#define __pyx_tuple__64 __pyx_mstate_global->__pyx_tuple__64
#define __pyx_tuple__65 __pyx_mstate_global->__pyx_tuple__65
#define __pyx_tuple__66 __pyx_mstate_global->__pyx_tuple__66
#define __pyx_tuple__67 __pyx_mstate_global->__pyx_tuple__67
#define __pyx_tuple__68 __pyx_mstate_global->__pyx_tuple__68
#define __pyx_tuple__69 __pyx_mstate_global->__pyx_tuple__69
#define __pyx_tuple__70 __pyx_mstate_global->__pyx_tuple__70
#define __pyx_tuple__71 __pyx_mstate_global->__pyx_tuple__71
#define __pyx_tuple__72 __pyx_mstate_global->__pyx_tuple__72
#define __pyx_tuple__73 __pyx_mstate_global->__pyx_tuple__73
#define __pyx_tuple__74 __pyx_mstate_global->__pyx_tuple__74
#define __pyx_tuple__75 __pyx_mstate_global->__pyx_tuple__75
#define __pyx_tuple__76 __pyx_mstate_global->__pyx_tuple__76
#define __pyx_tuple__77 __pyx_mstate_global->__pyx_tuple__77
#define __pyx_tuple__78 __pyx_mstate_global->__pyx_tuple__78
#define __pyx_tuple__79 __pyx_mstate_global->__pyx_tuple__79
#define __pyx_tuple__80 __pyx_mstate_global->__pyx_tuple__80
#define __pyx_tuple__81 __pyx_mstate_global->__pyx_tuple__81
#define __pyx_tuple__82 __pyx_mstate_global->__pyx_tuple__82
#define __pyx_tuple__83 __pyx_mstate_global->__pyx_tuple__83
#define __pyx_tuple__84 __pyx_mstate_global->__pyx_tuple__84
#define __pyx_tuple__85 __pyx_mstate_global->__pyx_tuple__85
#define __pyx_tuple__86 __pyx_mstate_global->__pyx_tuple__86
#define __pyx_tuple__87 __pyx_mstate_global->__pyx_tuple__87
#define __pyx_tuple__88 __pyx_mstate_global->__pyx_tuple__88
#define __pyx_tuple__89 __pyx_mstate_global->__pyx_tuple__89
#define __pyx_tuple__90 __pyx_mstate_global->__pyx_tuple__90
#define __pyx_tuple__91 __pyx_mstate_global->__pyx_tuple__91
#define __pyx_tuple__92 __pyx_mstate_global->__pyx_tuple__92
#define __pyx_tuple__93 __pyx_mstate_global->__pyx_tuple__93
#define __pyx_tuple__94 __pyx_mstate_global->__pyx_tuple__94
#define __pyx_tuple__95 __pyx_mstate_global->__pyx_tuple__95
#define __pyx_tuple__96 __pyx_mstate_global->__pyx_tuple__96
#define __pyx_tuple__97 __pyx_mstate_global->__pyx_tuple__97
#define __pyx_tuple__98 __pyx_mstate_global->__pyx_tuple__98
#define __pyx_tuple__99 __pyx_mstate_global->__pyx_tuple__99
#define __pyx_tuple__100 __pyx_mstate_global->__pyx_tuple__100
#define __pyx_tuple__101 __pyx_mstate_global->__pyx_tuple__101
#define __pyx_tuple__102 __pyx_mstate_global->__pyx_tuple__102
#define __pyx_tuple__103 __pyx_mstate_global->__pyx_tuple__103
#define __pyx_tuple__104 __pyx_mstate_global->__pyx_tuple__104
#define __pyx_tuple__105 __pyx_mstate_global->__pyx_tuple__105
#define __pyx_tuple__106 __pyx_mstate_global->__pyx_tuple__106
#define __pyx_tuple__107 __pyx_mstate_global->__pyx_tuple__107
#define __pyx_tuple__108 __pyx_mstate_global->__pyx_tuple__108
#define __pyx_tuple__109 __pyx_mstate_global->__pyx_tuple__109
#define __pyx_tuple__110 __pyx_mstate_global->__pyx_tuple__110
#define __pyx_tuple__111 __pyx_mstate_global->__pyx_tuple__111
#define __pyx_tuple__112 __pyx_mstate_global->__pyx_tuple__112
#define __pyx_tuple__113 __pyx_mstate_global->__pyx_tuple__113
#define __pyx_tuple__114 __pyx_mstate_global->__pyx_tuple__114
#define __pyx_tuple__115 __pyx_mstate_global->__pyx_tuple__115
#define __pyx_tuple__116 __pyx_mstate_global->__pyx_tuple__116
#define __pyx_tuple__117 __pyx_mstate_global->__pyx_tuple__117
#define __pyx_tuple__118 __pyx_mstate_global->__pyx_tuple__118
#define __pyx_tuple__119 __pyx_mstate_global->__pyx_tuple__119
#define __pyx_tuple__120 __pyx_mstate_global->__pyx_tuple__120
#define __pyx_tuple__121 __pyx_mstate_global->__pyx_tuple__121
#define __pyx_tuple__122 __pyx_mstate_global->__pyx_tuple__122
#define __pyx_tuple__123 __pyx_mstate_global->__pyx_tuple__123
#define __pyx_tuple__124 __pyx_mstate_global->__pyx_tuple__124
#define __pyx_tuple__125 __pyx_mstate_global->__pyx_tuple__125
#define __pyx_tuple__126 __pyx_mstate_global->__pyx_tuple__126
#define __pyx_tuple__127 __pyx_mstate_global->__pyx_tuple__127
#define __pyx_tuple__128 __pyx_mstate_global->__pyx_tuple__128
#define __pyx_tuple__129 __pyx_mstate_global->__pyx_tuple__129
#define __pyx_tuple__130 __pyx_mstate_global->__pyx_tuple__130
#define __pyx_tuple__131 __pyx_mstate_global->__pyx_tuple__131
#define __pyx_tuple__132 __pyx_mstate_global->__pyx_tuple__132
#define __pyx_tuple__133 __pyx_mstate_global->__pyx_tuple__133
#define __pyx_tuple__134 __pyx_mstate_global->__pyx_tuple__134
#define __pyx_tuple__135 __pyx_mstate_global->__pyx_tuple__135
#define __pyx_tuple__136 __pyx_mstate_global->__pyx_tuple__136
#define __pyx_tuple__137 __pyx_mstate_global->__pyx_tuple__137
#define __pyx_tuple__138 __pyx_mstate_global->__pyx_tuple__138
#define __pyx_tuple__139 __pyx_mstate_global->__pyx_tuple__139
#define __pyx_tuple__140 __pyx_mstate_global->__pyx_tuple__140
#define __pyx_tuple__141 __pyx_mstate_global->__pyx_tuple__141
#define __pyx_tuple__142 __pyx_mstate_global->__pyx_tuple__142
#define __pyx_tuple__143 __pyx_mstate_global->__pyx_tuple__143
#define __pyx_tuple__144 __pyx_mstate_global->__pyx_tuple__144
#define __pyx_tuple__145 __pyx_mstate_global->__pyx_tuple__145
#define __pyx_tuple__146 __pyx_mstate_global->__pyx_tuple__146
#define __pyx_tuple__147 __pyx_mstate_global->__pyx_tuple__147
#define __pyx_tuple__148 __pyx_mstate_global->__pyx_tuple__148
#define __pyx_codeobj__21 __pyx_mstate_global->__pyx_codeobj__21
#define __pyx_codeobj__23 __pyx_mstate_global->__pyx_codeobj__23
#define __pyx_codeobj__25 __pyx_mstate_global->__pyx_codeobj__25
#define __pyx_codeobj__28 __pyx_mstate_global->__pyx_codeobj__28
#define __pyx_codeobj__30 __pyx_mstate_global->__pyx_codeobj__30
#define __pyx_codeobj__32 __pyx_mstate_global->__pyx_codeobj__32
#define __pyx_codeobj__34 __pyx_mstate_global->__pyx_codeobj__34
#define __pyx_codeobj__36 __pyx_mstate_global->__pyx_codeobj__36
#define __pyx_codeobj__38 __pyx_mstate_global->__pyx_codeobj__38
#define __pyx_codeobj__40 __pyx_mstate_global->__pyx_codeobj__40
#define __pyx_codeobj__42 __pyx_mstate_global->__pyx_codeobj__42
#define __pyx_codeobj__44 __pyx_mstate_global->__pyx_codeobj__44
#define __pyx_codeobj__46 __pyx_mstate_global->__pyx_codeobj__46
#define __pyx_codeobj__48 __pyx_mstate_global->__pyx_codeobj__48
#define __pyx_codeobj__50 __pyx_mstate_global->__pyx_codeobj__50
#define __pyx_codeobj__52 __pyx_mstate_global->__pyx_codeobj__52
#define __pyx_codeobj__54 __pyx_mstate_global->__pyx_codeobj__54
#define __pyx_codeobj__56 __pyx_mstate_global->__pyx_codeobj__56
#define __pyx_codeobj__58 __pyx_mstate_global->__pyx_codeobj__58
/* #### Code section: module_code ### */

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":287
 * 
 *         @property
 *         cdef inline npy_intp itemsize(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_ELSIZE(self)
 * 
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_8itemsize_itemsize(PyArray_Descr *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":288
 *         @property
 *         cdef inline npy_intp itemsize(self) noexcept nogil:
 *             return PyDataType_ELSIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDataType_ELSIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":287
 * 
 *         @property
 *         cdef inline npy_intp itemsize(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_ELSIZE(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":291
 * 
 *         @property
 *         cdef inline npy_intp alignment(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_ALIGNMENT(self)
 * 
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_9alignment_alignment(PyArray_Descr *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":292
 *         @property
 *         cdef inline npy_intp alignment(self) noexcept nogil:
 *             return PyDataType_ALIGNMENT(self)             # <<<<<<<<<<<<<<
 * 
 *         # Use fields/names with care as they may be NULL.  You must check
 */
  __pyx_r = PyDataType_ALIGNMENT(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":291
 * 
 *         @property
 *         cdef inline npy_intp alignment(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_ALIGNMENT(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":297
 *         # for this using PyDataType_HASFIELDS.
 *         @property
 *         cdef inline object fields(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDataType_FIELDS(self)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_6fields_fields(PyArray_Descr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("fields", 1);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":298
 *         @property
 *         cdef inline object fields(self):
 *             return <object>PyDataType_FIELDS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDataType_FIELDS(__pyx_v_self);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":297
 *         # for this using PyDataType_HASFIELDS.
 *         @property
 *         cdef inline object fields(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDataType_FIELDS(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":301
 * 
 *         @property
 *         cdef inline tuple names(self):             # <<<<<<<<<<<<<<
 *             return <tuple>PyDataType_NAMES(self)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_5names_names(PyArray_Descr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("names", 1);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":302
 *         @property
 *         cdef inline tuple names(self):
 *             return <tuple>PyDataType_NAMES(self)             # <<<<<<<<<<<<<<
 * 
 *         # Use PyDataType_HASSUBARRAY to test whether this field is
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDataType_NAMES(__pyx_v_self);
  __Pyx_INCREF(((PyObject*)__pyx_t_1));
  __pyx_r = ((PyObject*)__pyx_t_1);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":301
 * 
 *         @property
 *         cdef inline tuple names(self):             # <<<<<<<<<<<<<<
 *             return <tuple>PyDataType_NAMES(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":308
 *         # this field via the inline helper method PyDataType_SHAPE.
 *         @property
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_SUBARRAY(self)
 * 
 */

static CYTHON_INLINE PyArray_ArrayDescr *__pyx_f_5numpy_5dtype_8subarray_subarray(PyArray_Descr *__pyx_v_self) {
  PyArray_ArrayDescr *__pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":309
 *         @property
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:
 *             return PyDataType_SUBARRAY(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDataType_SUBARRAY(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":308
 *         # this field via the inline helper method PyDataType_SHAPE.
 *         @property
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             return PyDataType_SUBARRAY(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":312
 * 
 *         @property
 *         cdef inline npy_uint64 flags(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The data types flags."""
 *             return PyDataType_FLAGS(self)
 */

static CYTHON_INLINE npy_uint64 __pyx_f_5numpy_5dtype_5flags_flags(PyArray_Descr *__pyx_v_self) {
  npy_uint64 __pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":314
 *         cdef inline npy_uint64 flags(self) noexcept nogil:
 *             """The data types flags."""
 *             return PyDataType_FLAGS(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyDataType_FLAGS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":312
 * 
 *         @property
 *         cdef inline npy_uint64 flags(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The data types flags."""
 *             return PyDataType_FLAGS(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":324
 * 
 *         @property
 *         cdef inline int numiter(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The number of arrays that need to be broadcast to the same shape."""
 *             return PyArray_MultiIter_NUMITER(self)
 */

static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_7numiter_numiter(PyArrayMultiIterObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":326
 *         cdef inline int numiter(self) noexcept nogil:
 *             """The number of arrays that need to be broadcast to the same shape."""
 *             return PyArray_MultiIter_NUMITER(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_MultiIter_NUMITER(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":324
 * 
 *         @property
 *         cdef inline int numiter(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The number of arrays that need to be broadcast to the same shape."""
 *             return PyArray_MultiIter_NUMITER(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":329
 * 
 *         @property
 *         cdef inline npy_intp size(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The total broadcasted size."""
 *             return PyArray_MultiIter_SIZE(self)
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_4size_size(PyArrayMultiIterObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":331
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """The total broadcasted size."""
 *             return PyArray_MultiIter_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_MultiIter_SIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":329
 * 
 *         @property
 *         cdef inline npy_intp size(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The total broadcasted size."""
 *             return PyArray_MultiIter_SIZE(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":334
 * 
 *         @property
 *         cdef inline npy_intp index(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The current (1-d) index into the broadcasted result."""
 *             return PyArray_MultiIter_INDEX(self)
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_5index_index(PyArrayMultiIterObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":336
 *         cdef inline npy_intp index(self) noexcept nogil:
 *             """The current (1-d) index into the broadcasted result."""
 *             return PyArray_MultiIter_INDEX(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_MultiIter_INDEX(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":334
 * 
 *         @property
 *         cdef inline npy_intp index(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The current (1-d) index into the broadcasted result."""
 *             return PyArray_MultiIter_INDEX(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":339
 * 
 *         @property
 *         cdef inline int nd(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The number of dimensions in the broadcasted result."""
 *             return PyArray_MultiIter_NDIM(self)
 */

static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_2nd_nd(PyArrayMultiIterObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":341
 *         cdef inline int nd(self) noexcept nogil:
 *             """The number of dimensions in the broadcasted result."""
 *             return PyArray_MultiIter_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_MultiIter_NDIM(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":339
 * 
 *         @property
 *         cdef inline int nd(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The number of dimensions in the broadcasted result."""
 *             return PyArray_MultiIter_NDIM(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":344
 * 
 *         @property
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The shape of the broadcasted result."""
 *             return PyArray_MultiIter_DIMS(self)
 */

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_9broadcast_10dimensions_dimensions(PyArrayMultiIterObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":346
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:
 *             """The shape of the broadcasted result."""
 *             return PyArray_MultiIter_DIMS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_MultiIter_DIMS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":344
 * 
 *         @property
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The shape of the broadcasted result."""
 *             return PyArray_MultiIter_DIMS(self)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":349
 * 
 *         @property
 *         cdef inline void** iters(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
 *             On return, the iterators are adjusted for broadcasting."""
 */

static CYTHON_INLINE void **__pyx_f_5numpy_9broadcast_5iters_iters(PyArrayMultiIterObject *__pyx_v_self) {
  void **__pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":352
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
 *             On return, the iterators are adjusted for broadcasting."""
 *             return PyArray_MultiIter_ITERS(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyArray_MultiIter_ITERS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":349
 * 
 *         @property
 *         cdef inline void** iters(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
 *             On return, the iterators are adjusted for broadcasting."""
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":367
 * 
 *         @property
 *         cdef inline PyObject* base(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self) {
  PyObject *__pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":370
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 *             return PyArray_BASE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_BASE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":367
 * 
 *         @property
 *         cdef inline PyObject* base(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":373
 * 
 *         @property
 *         cdef inline dtype descr(self):             # <<<<<<<<<<<<<<
 *             """Returns an owned reference to the dtype of the array.
 *             """
 */

static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self) {
  PyArray_Descr *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyArray_Descr *__pyx_t_1;
  __Pyx_RefNannySetupContext("descr", 1);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":376
 *             """Returns an owned reference to the dtype of the array.
 *             """
 *             return <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyArray_DESCR(__pyx_v_self);
  __Pyx_INCREF((PyObject *)((PyArray_Descr *)__pyx_t_1));
  __pyx_r = ((PyArray_Descr *)__pyx_t_1);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":373
 * 
 *         @property
 *         cdef inline dtype descr(self):             # <<<<<<<<<<<<<<
 *             """Returns an owned reference to the dtype of the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":379
 * 
 *         @property
 *         cdef inline int ndim(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns the number of dimensions in the array.
 *             """
 */

static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":382
 *             """Returns the number of dimensions in the array.
 *             """
 *             return PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_NDIM(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":379
 * 
 *         @property
 *         cdef inline int ndim(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns the number of dimensions in the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":385
 * 
 *         @property
 *         cdef inline npy_intp *shape(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the dimensions/shape of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":390
 *             Can return NULL for 0-dimensional arrays.
 *             """
 *             return PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_DIMS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":385
 * 
 *         @property
 *         cdef inline npy_intp *shape(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the dimensions/shape of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":393
 * 
 *         @property
 *         cdef inline npy_intp *strides(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the strides of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":397
 *             The number of elements matches the number of dimensions of the array (ndim).
 *             """
 *             return PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_STRIDES(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":393
 * 
 *         @property
 *         cdef inline npy_intp *strides(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the strides of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":400
 * 
 *         @property
 *         cdef inline npy_intp size(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns the total size (in number of elements) of the array.
 *             """
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":403
 *             """Returns the total size (in number of elements) of the array.
 *             """
 *             return PyArray_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_SIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":400
 * 
 *         @property
 *         cdef inline npy_intp size(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """Returns the total size (in number of elements) of the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":406
 * 
 *         @property
 *         cdef inline char* data(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The pointer to the data buffer as a char*.
 *             This is provided for legacy reasons to avoid direct struct field access.
 */

static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self) {
  char *__pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":412
 *             of `PyArray_DATA()` instead, which returns a 'void*'.
 *             """
 *             return PyArray_BYTES(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyArray_BYTES(__pyx_v_self);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":406
 * 
 *         @property
 *         cdef inline char* data(self) noexcept nogil:             # <<<<<<<<<<<<<<
 *             """The pointer to the data buffer as a char*.
 *             This is provided for legacy reasons to avoid direct struct field access.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":824
 * ctypedef long double complex clongdouble_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 1);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":825
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":824
 * ctypedef long double complex clongdouble_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":827
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 1);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":828
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":827
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":830
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 1);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":831
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":830
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":833
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 1);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":834
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":833
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":836
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 1);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":837
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":836
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":839
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 1);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":840
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = PyDataType_HASSUBARRAY(__pyx_v_d);
  if (__pyx_t_1) {

    /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":841
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_5numpy_5dtype_8subarray_subarray(__pyx_v_d)->shape;
    __Pyx_INCREF(((PyObject*)__pyx_t_2));
    __pyx_r = ((PyObject*)__pyx_t_2);
    goto __pyx_L0;

    /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":840
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":843
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":839
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1027
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1028
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1029
 * cdef inline void set_array_base(ndarray arr, object base) except *:
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  __pyx_t_1 = PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 1029, __pyx_L1_error)

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1027
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("numpy.set_array_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1031
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 1);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1032
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1033
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = (__pyx_v_base == NULL);
  if (__pyx_t_1) {

    /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1034
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1033
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1035
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1031
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1039
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 1);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1040
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1041
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1041, __pyx_L3_error)

      /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1040
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1042
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1042, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1043
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy._core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1043, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1043, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1040
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1039
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1045
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 1);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1046
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1047
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1047, __pyx_L3_error)

      /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1046
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1048
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1048, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1049
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1049, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1049, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1046
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1045
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1051
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 1);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1052
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1053
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1053, __pyx_L3_error)

      /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1052
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1054
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1054, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1055
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1055, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1055, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1052
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1051
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1058
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1070
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1058
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1073
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1085
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1073
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1088
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1095
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1088
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1098
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1102
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1098
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1105
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1109
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1105
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":58
 *     cdef public int _rbuf_pos, _record_number
 * 
 *     def __init__(self, filename, debug=False):             # <<<<<<<<<<<<<<
 *         """ initalize the object. """
 * 
 */

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile___init__, " initalize the object. ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_5pyart_2io_11_sigmetfile_10SigmetFile___init__;
#endif
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_filename = 0;
  PyObject *__pyx_v_debug = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_filename,&__pyx_n_s_debug,0};
    values[1] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_filename)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 58, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_debug);
          if (value) { values[1] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 58, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 58, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_filename = values[0];
    __pyx_v_debug = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 58, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile___init__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), __pyx_v_filename, __pyx_v_debug);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile___init__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_filename, PyObject *__pyx_v_debug) {
  PyObject *__pyx_v_fh = NULL;
  PyObject *__pyx_7genexpr__pyx_v_i = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "pyart/io/_sigmetfile.pyx":61
 *         """ initalize the object. """
 * 
 *         self.debug = debug             # <<<<<<<<<<<<<<
 * 
 *         # open the file
 */
  __Pyx_INCREF(__pyx_v_debug);
  __Pyx_GIVEREF(__pyx_v_debug);
  __Pyx_GOTREF(__pyx_v_self->debug);
  __Pyx_DECREF(__pyx_v_self->debug);
  __pyx_v_self->debug = __pyx_v_debug;

  /* "pyart/io/_sigmetfile.pyx":64
 * 
 *         # open the file
 *         if hasattr(filename, 'read'):             # <<<<<<<<<<<<<<
 *             fh = filename
 *         else:
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_filename, __pyx_n_u_read); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 64, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "pyart/io/_sigmetfile.pyx":65
 *         # open the file
 *         if hasattr(filename, 'read'):
 *             fh = filename             # <<<<<<<<<<<<<<
 *         else:
 *             fh = open(filename, 'rb')
 */
    __Pyx_INCREF(__pyx_v_filename);
    __pyx_v_fh = __pyx_v_filename;

    /* "pyart/io/_sigmetfile.pyx":64
 * 
 *         # open the file
 *         if hasattr(filename, 'read'):             # <<<<<<<<<<<<<<
 *             fh = filename
 *         else:
 */
    goto __pyx_L3;
  }

  /* "pyart/io/_sigmetfile.pyx":67
 *             fh = filename
 *         else:
 *             fh = open(filename, 'rb')             # <<<<<<<<<<<<<<
 * 
 *         # read the headers from the first 2 records.
 */
  /*else*/ {
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_filename);
    __Pyx_GIVEREF(__pyx_v_filename);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_filename)) __PYX_ERR(0, 67, __pyx_L1_error);
    __Pyx_INCREF(__pyx_n_u_rb);
    __Pyx_GIVEREF(__pyx_n_u_rb);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_u_rb)) __PYX_ERR(0, 67, __pyx_L1_error);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_fh = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "pyart/io/_sigmetfile.pyx":70
 * 
 *         # read the headers from the first 2 records.
 *         self.product_hdr = _unpack_product_hdr(fh.read(RECORD_SIZE))             # <<<<<<<<<<<<<<
 *         self.ingest_header = _unpack_ingest_header(fh.read(RECORD_SIZE))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_product_hdr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_fh, __pyx_n_s_read); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_RECORD_SIZE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_6};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->product_hdr);
  __Pyx_DECREF(__pyx_v_self->product_hdr);
  __pyx_v_self->product_hdr = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":71
 *         # read the headers from the first 2 records.
 *         self.product_hdr = _unpack_product_hdr(fh.read(RECORD_SIZE))
 *         self.ingest_header = _unpack_ingest_header(fh.read(RECORD_SIZE))             # <<<<<<<<<<<<<<
 * 
 *         # determine data types contained in the file
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_ingest_header); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_fh, __pyx_n_s_read); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_RECORD_SIZE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_6};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->ingest_header);
  __Pyx_DECREF(__pyx_v_self->ingest_header);
  __pyx_v_self->ingest_header = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":74
 * 
 *         # determine data types contained in the file
 *         self.data_types = self._determine_data_types()             # <<<<<<<<<<<<<<
 *         self.ndata_types = len(self.data_types)
 *         self.data_type_names = [SIGMET_DATA_TYPES[i] for i in self.data_types]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_determine_data_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->data_types);
  __Pyx_DECREF(__pyx_v_self->data_types);
  __pyx_v_self->data_types = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":75
 *         # determine data types contained in the file
 *         self.data_types = self._determine_data_types()
 *         self.ndata_types = len(self.data_types)             # <<<<<<<<<<<<<<
 *         self.data_type_names = [SIGMET_DATA_TYPES[i] for i in self.data_types]
 * 
 */
  __pyx_t_3 = __pyx_v_self->data_types;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_9 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->ndata_types);
  __Pyx_DECREF(__pyx_v_self->ndata_types);
  __pyx_v_self->ndata_types = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":76
 *         self.data_types = self._determine_data_types()
 *         self.ndata_types = len(self.data_types)
 *         self.data_type_names = [SIGMET_DATA_TYPES[i] for i in self.data_types]             # <<<<<<<<<<<<<<
 * 
 *         # set attributes
 */
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 76, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (likely(PyList_CheckExact(__pyx_v_self->data_types)) || PyTuple_CheckExact(__pyx_v_self->data_types)) {
      __pyx_t_2 = __pyx_v_self->data_types; __Pyx_INCREF(__pyx_t_2);
      __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_self->data_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 76, __pyx_L6_error)
    }
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 76, __pyx_L6_error)
            #endif
            if (__pyx_t_9 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_4); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(0, 76, __pyx_L6_error)
          #else
          __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 76, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 76, __pyx_L6_error)
            #endif
            if (__pyx_t_9 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_4); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(0, 76, __pyx_L6_error)
          #else
          __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 76, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_10(__pyx_t_2);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 76, __pyx_L6_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_i, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SIGMET_DATA_TYPES); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 76, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_t_4, __pyx_7genexpr__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 76, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 76, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_i); __pyx_7genexpr__pyx_v_i = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_i); __pyx_7genexpr__pyx_v_i = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->data_type_names);
  __Pyx_DECREF(__pyx_v_self->data_type_names);
  __pyx_v_self->data_type_names = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":79
 * 
 *         # set attributes
 *         self.ingest_data_headers = None             # <<<<<<<<<<<<<<
 *         self._fh = fh
 *         self._record_number = 2
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->ingest_data_headers);
  __Pyx_DECREF(__pyx_v_self->ingest_data_headers);
  __pyx_v_self->ingest_data_headers = Py_None;

  /* "pyart/io/_sigmetfile.pyx":80
 *         # set attributes
 *         self.ingest_data_headers = None
 *         self._fh = fh             # <<<<<<<<<<<<<<
 *         self._record_number = 2
 *         self._raw_product_bhdrs = []
 */
  __Pyx_INCREF(__pyx_v_fh);
  __Pyx_GIVEREF(__pyx_v_fh);
  __Pyx_GOTREF(__pyx_v_self->_fh);
  __Pyx_DECREF(__pyx_v_self->_fh);
  __pyx_v_self->_fh = __pyx_v_fh;

  /* "pyart/io/_sigmetfile.pyx":81
 *         self.ingest_data_headers = None
 *         self._fh = fh
 *         self._record_number = 2             # <<<<<<<<<<<<<<
 *         self._raw_product_bhdrs = []
 * 
 */
  __pyx_v_self->_record_number = 2;

  /* "pyart/io/_sigmetfile.pyx":82
 *         self._fh = fh
 *         self._record_number = 2
 *         self._raw_product_bhdrs = []             # <<<<<<<<<<<<<<
 * 
 *     def _determine_data_types(self):
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_raw_product_bhdrs);
  __Pyx_DECREF(__pyx_v_self->_raw_product_bhdrs);
  __pyx_v_self->_raw_product_bhdrs = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":58
 *     cdef public int _rbuf_pos, _record_number
 * 
 *     def __init__(self, filename, debug=False):             # <<<<<<<<<<<<<<
 *         """ initalize the object. """
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fh);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_i);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":84
 *         self._raw_product_bhdrs = []
 * 
 *     def _determine_data_types(self):             # <<<<<<<<<<<<<<
 *         """ Determine the available data types in the file. """
 *         # determine the available fields
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_determine_data_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_2_determine_data_types, " Determine the available data types in the file. ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_10SigmetFile_3_determine_data_types = {"_determine_data_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_determine_data_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_2_determine_data_types};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_determine_data_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_determine_data_types (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("_determine_data_types", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "_determine_data_types", 0))) return NULL;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_2_determine_data_types(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_2_determine_data_types(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_v_task_config = NULL;
  PyObject *__pyx_v_task_dsp_info = NULL;
  PyObject *__pyx_v_word0 = NULL;
  PyObject *__pyx_v_word1 = NULL;
  PyObject *__pyx_v_word2 = NULL;
  PyObject *__pyx_v_word3 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_determine_data_types", 1);

  /* "pyart/io/_sigmetfile.pyx":87
 *         """ Determine the available data types in the file. """
 *         # determine the available fields
 *         task_config = self.ingest_header['task_configuration']             # <<<<<<<<<<<<<<
 *         task_dsp_info = task_config['task_dsp_info']
 *         word0 = task_dsp_info['current_data_type_mask']['mask_word_0']
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_self->ingest_header, __pyx_n_u_task_configuration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_task_config = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":88
 *         # determine the available fields
 *         task_config = self.ingest_header['task_configuration']
 *         task_dsp_info = task_config['task_dsp_info']             # <<<<<<<<<<<<<<
 *         word0 = task_dsp_info['current_data_type_mask']['mask_word_0']
 *         word1 = task_dsp_info['current_data_type_mask']['mask_word_1']
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_task_config, __pyx_n_u_task_dsp_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_task_dsp_info = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":89
 *         task_config = self.ingest_header['task_configuration']
 *         task_dsp_info = task_config['task_dsp_info']
 *         word0 = task_dsp_info['current_data_type_mask']['mask_word_0']             # <<<<<<<<<<<<<<
 *         word1 = task_dsp_info['current_data_type_mask']['mask_word_1']
 *         word2 = task_dsp_info['current_data_type_mask']['mask_word_2']
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_task_dsp_info, __pyx_n_u_current_data_type_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_mask_word_0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_word0 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":90
 *         task_dsp_info = task_config['task_dsp_info']
 *         word0 = task_dsp_info['current_data_type_mask']['mask_word_0']
 *         word1 = task_dsp_info['current_data_type_mask']['mask_word_1']             # <<<<<<<<<<<<<<
 *         word2 = task_dsp_info['current_data_type_mask']['mask_word_2']
 *         word3 = task_dsp_info['current_data_type_mask']['mask_word_3']
 */
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_task_dsp_info, __pyx_n_u_current_data_type_mask); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_mask_word_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_word1 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":91
 *         word0 = task_dsp_info['current_data_type_mask']['mask_word_0']
 *         word1 = task_dsp_info['current_data_type_mask']['mask_word_1']
 *         word2 = task_dsp_info['current_data_type_mask']['mask_word_2']             # <<<<<<<<<<<<<<
 *         word3 = task_dsp_info['current_data_type_mask']['mask_word_3']
 *         return _data_types_from_mask(word0, word1, word2, word3)
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_task_dsp_info, __pyx_n_u_current_data_type_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_mask_word_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_word2 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":92
 *         word1 = task_dsp_info['current_data_type_mask']['mask_word_1']
 *         word2 = task_dsp_info['current_data_type_mask']['mask_word_2']
 *         word3 = task_dsp_info['current_data_type_mask']['mask_word_3']             # <<<<<<<<<<<<<<
 *         return _data_types_from_mask(word0, word1, word2, word3)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_task_dsp_info, __pyx_n_u_current_data_type_mask); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_mask_word_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_word3 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":93
 *         word2 = task_dsp_info['current_data_type_mask']['mask_word_2']
 *         word3 = task_dsp_info['current_data_type_mask']['mask_word_3']
 *         return _data_types_from_mask(word0, word1, word2, word3)             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_data_types_from_mask); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[5] = {__pyx_t_3, __pyx_v_word0, __pyx_v_word1, __pyx_v_word2, __pyx_v_word3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 4+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":84
 *         self._raw_product_bhdrs = []
 * 
 *     def _determine_data_types(self):             # <<<<<<<<<<<<<<
 *         """ Determine the available data types in the file. """
 *         # determine the available fields
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._determine_data_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_task_config);
  __Pyx_XDECREF(__pyx_v_task_dsp_info);
  __Pyx_XDECREF(__pyx_v_word0);
  __Pyx_XDECREF(__pyx_v_word1);
  __Pyx_XDECREF(__pyx_v_word2);
  __Pyx_XDECREF(__pyx_v_word3);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":95
 *         return _data_types_from_mask(word0, word1, word2, word3)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """ Close the file. """
 *         self._fh.close()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_4close, " Close the file. ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_10SigmetFile_5close = {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_4close};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "close", 0))) return NULL;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_4close(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_4close(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 1);

  /* "pyart/io/_sigmetfile.pyx":97
 *     def close(self):
 *         """ Close the file. """
 *         self._fh.close()             # <<<<<<<<<<<<<<
 * 
 *     def read_data(self, full_xhdr=False):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_fh, __pyx_n_s_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":95
 *         return _data_types_from_mask(word0, word1, word2, word3)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """ Close the file. """
 *         self._fh.close()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":99
 *         self._fh.close()
 * 
 *     def read_data(self, full_xhdr=False):             # <<<<<<<<<<<<<<
 *         """
 *         Read all data from the file.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_7read_data(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_6read_data, "\n        Read all data from the file.\n\n        Parameters\n        ----------\n        full_xhdr : bool\n            True to return the full extended headers if they exist padded with\n            ones.  False will return a length 1 extended header converted to\n            int32.  This is useful when the file contains a customer specified\n            extended header (for example aircraft radar).\n\n        Returns\n        -------\n        data : dict of ndarrays\n            Data arrays of shape=(nsweeps, nrays, nbins) for each data type.\n            Indexed by data type name (str).\n        metadata : dict of dicts\n            Arrays of 'azimuth_0', 'azimuth_1', 'elevation_0', 'elevation_1',\n            'nbins', and 'time' for each data type.  Indexed by data type name\n            (str).  Rays which were not collected are marked with a value of\n            -1 in the 'nbins' array.\n\n        ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_10SigmetFile_7read_data = {"read_data", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_7read_data, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_6read_data};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_7read_data(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_full_xhdr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_data (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_full_xhdr,0};
    values[0] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_False));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_full_xhdr);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 99, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "read_data") < 0)) __PYX_ERR(0, 99, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_full_xhdr = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("read_data", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 99, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile.read_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_6read_data(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), __pyx_v_full_xhdr);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_6read_data(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_full_xhdr) {
  PyObject *__pyx_v_nsweeps = NULL;
  PyObject *__pyx_v_nbins = NULL;
  PyObject *__pyx_v_nrays = NULL;
  PyObject *__pyx_v_shape = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_metadata = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_header_dic = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_ingest_data_hdrs = NULL;
  PyObject *__pyx_v_sweep_data = NULL;
  PyObject *__pyx_v_sweep_metadata = NULL;
  PyObject *__pyx_v_mess = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_v_temp = NULL;
  PyObject *__pyx_v_az0 = NULL;
  PyObject *__pyx_v_el0 = NULL;
  PyObject *__pyx_v_az1 = NULL;
  PyObject *__pyx_v_el1 = NULL;
  PyObject *__pyx_v_ray_nbins = NULL;
  PyObject *__pyx_v_ray_time = NULL;
  PyObject *__pyx_v_prf_flag = NULL;
  PyObject *__pyx_v_wavelength_cm = NULL;
  PyObject *__pyx_v_prt_value = NULL;
  PyObject *__pyx_v_task_config = NULL;
  PyObject *__pyx_v_multi_prf_flag = NULL;
  PyObject *__pyx_v_multiplier = NULL;
  PyObject *__pyx_v_nyquist = NULL;
  PyObject *__pyx_8genexpr1__pyx_v_name = NULL;
  PyObject *__pyx_8genexpr2__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  unsigned int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  Py_ssize_t __pyx_t_14;
  PyObject *(*__pyx_t_15)(PyObject *);
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_t_21;
  int __pyx_t_22;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_data", 1);

  /* "pyart/io/_sigmetfile.pyx":125
 * 
 *         # determine size of data
 *         nsweeps = self.ingest_header['task_configuration'][             # <<<<<<<<<<<<<<
 *             'task_scan_info']['number_sweeps']
 *         nbins = self.product_hdr['product_end']['number_bins']
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_self->ingest_header, __pyx_n_u_task_configuration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_task_scan_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":126
 *         # determine size of data
 *         nsweeps = self.ingest_header['task_configuration'][
 *             'task_scan_info']['number_sweeps']             # <<<<<<<<<<<<<<
 *         nbins = self.product_hdr['product_end']['number_bins']
 *         nrays = self.ingest_header['ingest_configuration'][
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_number_sweeps); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nsweeps = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":127
 *         nsweeps = self.ingest_header['task_configuration'][
 *             'task_scan_info']['number_sweeps']
 *         nbins = self.product_hdr['product_end']['number_bins']             # <<<<<<<<<<<<<<
 *         nrays = self.ingest_header['ingest_configuration'][
 *             'number_rays_sweep']
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_self->product_hdr, __pyx_n_u_product_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_number_bins); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_nbins = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":128
 *             'task_scan_info']['number_sweeps']
 *         nbins = self.product_hdr['product_end']['number_bins']
 *         nrays = self.ingest_header['ingest_configuration'][             # <<<<<<<<<<<<<<
 *             'number_rays_sweep']
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_self->ingest_header, __pyx_n_u_ingest_configuration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_n_u_number_rays_sweep); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nrays = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":132
 * 
 *         # create empty outputs
 *         shape = (nsweeps, nrays, nbins)             # <<<<<<<<<<<<<<
 *         data = dict([(name, np.ma.empty(shape, dtype='float32'))
 *                     for name in self.data_type_names])
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_nsweeps);
  __Pyx_GIVEREF(__pyx_v_nsweeps);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_nsweeps)) __PYX_ERR(0, 132, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_nrays);
  __Pyx_GIVEREF(__pyx_v_nrays);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_nrays)) __PYX_ERR(0, 132, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_nbins);
  __Pyx_GIVEREF(__pyx_v_nbins);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_nbins)) __PYX_ERR(0, 132, __pyx_L1_error);
  __pyx_v_shape = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":133
 *         # create empty outputs
 *         shape = (nsweeps, nrays, nbins)
 *         data = dict([(name, np.ma.empty(shape, dtype='float32'))             # <<<<<<<<<<<<<<
 *                     for name in self.data_type_names])
 *         if 'XHDR' in self.data_type_names:
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "pyart/io/_sigmetfile.pyx":134
 *         shape = (nsweeps, nrays, nbins)
 *         data = dict([(name, np.ma.empty(shape, dtype='float32'))
 *                     for name in self.data_type_names])             # <<<<<<<<<<<<<<
 *         if 'XHDR' in self.data_type_names:
 *             if full_xhdr:
 */
    if (likely(PyList_CheckExact(__pyx_v_self->data_type_names)) || PyTuple_CheckExact(__pyx_v_self->data_type_names)) {
      __pyx_t_2 = __pyx_v_self->data_type_names; __Pyx_INCREF(__pyx_t_2);
      __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_self->data_type_names); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 134, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 134, __pyx_L5_error)
    }
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 134, __pyx_L5_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 134, __pyx_L5_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 134, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 134, __pyx_L5_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 134, __pyx_L5_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 134, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_4(__pyx_t_2);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 134, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_name, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "pyart/io/_sigmetfile.pyx":133
 *         # create empty outputs
 *         shape = (nsweeps, nrays, nbins)
 *         data = dict([(name, np.ma.empty(shape, dtype='float32'))             # <<<<<<<<<<<<<<
 *                     for name in self.data_type_names])
 *         if 'XHDR' in self.data_type_names:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 133, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ma); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 133, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 133, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 133, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_v_shape);
      __Pyx_GIVEREF(__pyx_v_shape);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_shape)) __PYX_ERR(0, 133, __pyx_L5_error);
      __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 133, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_n_u_float32) < 0) __PYX_ERR(0, 133, __pyx_L5_error)
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 133, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 133, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_8genexpr1__pyx_v_name);
      __Pyx_GIVEREF(__pyx_8genexpr1__pyx_v_name);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_8genexpr1__pyx_v_name)) __PYX_ERR(0, 133, __pyx_L5_error);
      __Pyx_GIVEREF(__pyx_t_8);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_8)) __PYX_ERR(0, 133, __pyx_L5_error);
      __pyx_t_8 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 133, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "pyart/io/_sigmetfile.pyx":134
 *         shape = (nsweeps, nrays, nbins)
 *         data = dict([(name, np.ma.empty(shape, dtype='float32'))
 *                     for name in self.data_type_names])             # <<<<<<<<<<<<<<
 *         if 'XHDR' in self.data_type_names:
 *             if full_xhdr:
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_name); __pyx_8genexpr1__pyx_v_name = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_name); __pyx_8genexpr1__pyx_v_name = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */

  /* "pyart/io/_sigmetfile.pyx":133
 *         # create empty outputs
 *         shape = (nsweeps, nrays, nbins)
 *         data = dict([(name, np.ma.empty(shape, dtype='float32'))             # <<<<<<<<<<<<<<
 *                     for name in self.data_type_names])
 *         if 'XHDR' in self.data_type_names:
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_data = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":135
 *         data = dict([(name, np.ma.empty(shape, dtype='float32'))
 *                     for name in self.data_type_names])
 *         if 'XHDR' in self.data_type_names:             # <<<<<<<<<<<<<<
 *             if full_xhdr:
 *                 data['XHDR'] = np.ones(shape, dtype='int16')
 */
  __pyx_t_9 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_XHDR, __pyx_v_self->data_type_names, Py_EQ)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 135, __pyx_L1_error)
  if (__pyx_t_9) {

    /* "pyart/io/_sigmetfile.pyx":136
 *                     for name in self.data_type_names])
 *         if 'XHDR' in self.data_type_names:
 *             if full_xhdr:             # <<<<<<<<<<<<<<
 *                 data['XHDR'] = np.ones(shape, dtype='int16')
 *             else:
 */
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_full_xhdr); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 136, __pyx_L1_error)
    if (__pyx_t_9) {

      /* "pyart/io/_sigmetfile.pyx":137
 *         if 'XHDR' in self.data_type_names:
 *             if full_xhdr:
 *                 data['XHDR'] = np.ones(shape, dtype='int16')             # <<<<<<<<<<<<<<
 *             else:
 *                 data['XHDR'] = np.ones((nsweeps, nrays, 1), dtype='int32')
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ones); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_shape);
      __Pyx_GIVEREF(__pyx_v_shape);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_shape)) __PYX_ERR(0, 137, __pyx_L1_error);
      __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_n_u_int16) < 0) __PYX_ERR(0, 137, __pyx_L1_error)
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_n_u_XHDR, __pyx_t_8) < 0))) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "pyart/io/_sigmetfile.pyx":136
 *                     for name in self.data_type_names])
 *         if 'XHDR' in self.data_type_names:
 *             if full_xhdr:             # <<<<<<<<<<<<<<
 *                 data['XHDR'] = np.ones(shape, dtype='int16')
 *             else:
 */
      goto __pyx_L11;
    }

    /* "pyart/io/_sigmetfile.pyx":139
 *                 data['XHDR'] = np.ones(shape, dtype='int16')
 *             else:
 *                 data['XHDR'] = np.ones((nsweeps, nrays, 1), dtype='int32')             # <<<<<<<<<<<<<<
 * 
 *         metadata = {}
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_ones); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_v_nsweeps);
      __Pyx_GIVEREF(__pyx_v_nsweeps);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_nsweeps)) __PYX_ERR(0, 139, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_nrays);
      __Pyx_GIVEREF(__pyx_v_nrays);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_nrays)) __PYX_ERR(0, 139, __pyx_L1_error);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_int_1)) __PYX_ERR(0, 139, __pyx_L1_error);
      __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_8);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_8)) __PYX_ERR(0, 139, __pyx_L1_error);
      __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_n_u_int32) < 0) __PYX_ERR(0, 139, __pyx_L1_error)
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_2, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_n_u_XHDR, __pyx_t_1) < 0))) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L11:;

    /* "pyart/io/_sigmetfile.pyx":135
 *         data = dict([(name, np.ma.empty(shape, dtype='float32'))
 *                     for name in self.data_type_names])
 *         if 'XHDR' in self.data_type_names:             # <<<<<<<<<<<<<<
 *             if full_xhdr:
 *                 data['XHDR'] = np.ones(shape, dtype='int16')
 */
  }

  /* "pyart/io/_sigmetfile.pyx":141
 *                 data['XHDR'] = np.ones((nsweeps, nrays, 1), dtype='int32')
 * 
 *         metadata = {}             # <<<<<<<<<<<<<<
 *         for name in self.data_type_names:
 *             header_dic = {
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_metadata = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":142
 * 
 *         metadata = {}
 *         for name in self.data_type_names:             # <<<<<<<<<<<<<<
 *             header_dic = {
 *                 'azimuth_0': np.empty((nsweeps, nrays), dtype='float32'),
 */
  if (likely(PyList_CheckExact(__pyx_v_self->data_type_names)) || PyTuple_CheckExact(__pyx_v_self->data_type_names)) {
    __pyx_t_1 = __pyx_v_self->data_type_names; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_self->data_type_names); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 142, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 142, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_8); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 142, __pyx_L1_error)
        #else
        __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 142, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 142, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_8); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 142, __pyx_L1_error)
        #else
        __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 142, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      }
    } else {
      __pyx_t_8 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_8)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 142, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_8);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "pyart/io/_sigmetfile.pyx":144
 *         for name in self.data_type_names:
 *             header_dic = {
 *                 'azimuth_0': np.empty((nsweeps, nrays), dtype='float32'),             # <<<<<<<<<<<<<<
 *                 'elevation_0': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'azimuth_1': np.empty((nsweeps, nrays), dtype='float32'),
 */
    __pyx_t_8 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_nsweeps);
    __Pyx_GIVEREF(__pyx_v_nsweeps);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_nsweeps)) __PYX_ERR(0, 144, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_nrays);
    __Pyx_GIVEREF(__pyx_v_nrays);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_nrays)) __PYX_ERR(0, 144, __pyx_L1_error);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2)) __PYX_ERR(0, 144, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_n_u_float32) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_u_azimuth_0, __pyx_t_5) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "pyart/io/_sigmetfile.pyx":145
 *             header_dic = {
 *                 'azimuth_0': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'elevation_0': np.empty((nsweeps, nrays), dtype='float32'),             # <<<<<<<<<<<<<<
 *                 'azimuth_1': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'elevation_1': np.empty((nsweeps, nrays), dtype='float32'),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_nsweeps);
    __Pyx_GIVEREF(__pyx_v_nsweeps);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_nsweeps)) __PYX_ERR(0, 145, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_nrays);
    __Pyx_GIVEREF(__pyx_v_nrays);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_nrays)) __PYX_ERR(0, 145, __pyx_L1_error);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5)) __PYX_ERR(0, 145, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_n_u_float32) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_u_elevation_0, __pyx_t_7) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "pyart/io/_sigmetfile.pyx":146
 *                 'azimuth_0': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'elevation_0': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'azimuth_1': np.empty((nsweeps, nrays), dtype='float32'),             # <<<<<<<<<<<<<<
 *                 'elevation_1': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'nbins': np.empty((nsweeps, nrays), dtype='int16'),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_nsweeps);
    __Pyx_GIVEREF(__pyx_v_nsweeps);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_nsweeps)) __PYX_ERR(0, 146, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_nrays);
    __Pyx_GIVEREF(__pyx_v_nrays);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_nrays)) __PYX_ERR(0, 146, __pyx_L1_error);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7)) __PYX_ERR(0, 146, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_n_u_float32) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_u_azimuth_1, __pyx_t_2) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyart/io/_sigmetfile.pyx":147
 *                 'elevation_0': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'azimuth_1': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'elevation_1': np.empty((nsweeps, nrays), dtype='float32'),             # <<<<<<<<<<<<<<
 *                 'nbins': np.empty((nsweeps, nrays), dtype='int16'),
 *                 'time': np.empty((nsweeps, nrays), dtype='uint16'),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_nsweeps);
    __Pyx_GIVEREF(__pyx_v_nsweeps);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_nsweeps)) __PYX_ERR(0, 147, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_nrays);
    __Pyx_GIVEREF(__pyx_v_nrays);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_nrays)) __PYX_ERR(0, 147, __pyx_L1_error);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_n_u_float32) < 0) __PYX_ERR(0, 147, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_u_elevation_1, __pyx_t_5) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "pyart/io/_sigmetfile.pyx":148
 *                 'azimuth_1': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'elevation_1': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'nbins': np.empty((nsweeps, nrays), dtype='int16'),             # <<<<<<<<<<<<<<
 *                 'time': np.empty((nsweeps, nrays), dtype='uint16'),
 *                 'prf_flag': np.empty((nsweeps, nrays), dtype='int16')}
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_nsweeps);
    __Pyx_GIVEREF(__pyx_v_nsweeps);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_nsweeps)) __PYX_ERR(0, 148, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_nrays);
    __Pyx_GIVEREF(__pyx_v_nrays);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_nrays)) __PYX_ERR(0, 148, __pyx_L1_error);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5)) __PYX_ERR(0, 148, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_n_u_int16) < 0) __PYX_ERR(0, 148, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_u_nbins, __pyx_t_7) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "pyart/io/_sigmetfile.pyx":149
 *                 'elevation_1': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'nbins': np.empty((nsweeps, nrays), dtype='int16'),
 *                 'time': np.empty((nsweeps, nrays), dtype='uint16'),             # <<<<<<<<<<<<<<
 *                 'prf_flag': np.empty((nsweeps, nrays), dtype='int16')}
 *             metadata[name] = header_dic
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_nsweeps);
    __Pyx_GIVEREF(__pyx_v_nsweeps);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_nsweeps)) __PYX_ERR(0, 149, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_nrays);
    __Pyx_GIVEREF(__pyx_v_nrays);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_nrays)) __PYX_ERR(0, 149, __pyx_L1_error);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7)) __PYX_ERR(0, 149, __pyx_L1_error);
    __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_n_u_uint16) < 0) __PYX_ERR(0, 149, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_u_time, __pyx_t_2) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyart/io/_sigmetfile.pyx":150
 *                 'nbins': np.empty((nsweeps, nrays), dtype='int16'),
 *                 'time': np.empty((nsweeps, nrays), dtype='uint16'),
 *                 'prf_flag': np.empty((nsweeps, nrays), dtype='int16')}             # <<<<<<<<<<<<<<
 *             metadata[name] = header_dic
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_nsweeps);
    __Pyx_GIVEREF(__pyx_v_nsweeps);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_nsweeps)) __PYX_ERR(0, 150, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_nrays);
    __Pyx_GIVEREF(__pyx_v_nrays);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_nrays)) __PYX_ERR(0, 150, __pyx_L1_error);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2)) __PYX_ERR(0, 150, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_n_u_int16) < 0) __PYX_ERR(0, 150, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_u_prf_flag, __pyx_t_5) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_header_dic, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "pyart/io/_sigmetfile.pyx":151
 *                 'time': np.empty((nsweeps, nrays), dtype='uint16'),
 *                 'prf_flag': np.empty((nsweeps, nrays), dtype='int16')}
 *             metadata[name] = header_dic             # <<<<<<<<<<<<<<
 * 
 *         self.ingest_data_headers = dict([(name, []) for name in
 */
    if (unlikely((PyDict_SetItem(__pyx_v_metadata, __pyx_v_name, __pyx_v_header_dic) < 0))) __PYX_ERR(0, 151, __pyx_L1_error)

    /* "pyart/io/_sigmetfile.pyx":142
 * 
 *         metadata = {}
 *         for name in self.data_type_names:             # <<<<<<<<<<<<<<
 *             header_dic = {
 *                 'azimuth_0': np.empty((nsweeps, nrays), dtype='float32'),
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":153
 *             metadata[name] = header_dic
 * 
 *         self.ingest_data_headers = dict([(name, []) for name in             # <<<<<<<<<<<<<<
 *                                          self.data_type_names])
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L17_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "pyart/io/_sigmetfile.pyx":154
 * 
 *         self.ingest_data_headers = dict([(name, []) for name in
 *                                          self.data_type_names])             # <<<<<<<<<<<<<<
 * 
 *         self._raw_product_bhdrs = []
 */
    if (likely(PyList_CheckExact(__pyx_v_self->data_type_names)) || PyTuple_CheckExact(__pyx_v_self->data_type_names)) {
      __pyx_t_8 = __pyx_v_self->data_type_names; __Pyx_INCREF(__pyx_t_8);
      __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_v_self->data_type_names); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 154, __pyx_L17_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 154, __pyx_L17_error)
    }
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_8))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_8);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 154, __pyx_L17_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 154, __pyx_L17_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 154, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_8);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 154, __pyx_L17_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 154, __pyx_L17_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 154, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_4(__pyx_t_8);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 154, __pyx_L17_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_name, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "pyart/io/_sigmetfile.pyx":153
 *             metadata[name] = header_dic
 * 
 *         self.ingest_data_headers = dict([(name, []) for name in             # <<<<<<<<<<<<<<
 *                                          self.data_type_names])
 * 
 */
      __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 153, __pyx_L17_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L17_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_8genexpr2__pyx_v_name);
      __Pyx_GIVEREF(__pyx_8genexpr2__pyx_v_name);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_8genexpr2__pyx_v_name)) __PYX_ERR(0, 153, __pyx_L17_error);
      __Pyx_GIVEREF(__pyx_t_5);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5)) __PYX_ERR(0, 153, __pyx_L17_error);
      __pyx_t_5 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 153, __pyx_L17_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_name); __pyx_8genexpr2__pyx_v_name = 0;
    goto __pyx_L21_exit_scope;
    __pyx_L17_error:;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_name); __pyx_8genexpr2__pyx_v_name = 0;
    goto __pyx_L1_error;
    __pyx_L21_exit_scope:;
  } /* exit inner scope */
  __pyx_t_8 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_8);
  __Pyx_GOTREF(__pyx_v_self->ingest_data_headers);
  __Pyx_DECREF(__pyx_v_self->ingest_data_headers);
  __pyx_v_self->ingest_data_headers = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "pyart/io/_sigmetfile.pyx":156
 *                                          self.data_type_names])
 * 
 *         self._raw_product_bhdrs = []             # <<<<<<<<<<<<<<
 * 
 *         # read in data sweep by sweep
 */
  __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_8);
  __Pyx_GOTREF(__pyx_v_self->_raw_product_bhdrs);
  __Pyx_DECREF(__pyx_v_self->_raw_product_bhdrs);
  __pyx_v_self->_raw_product_bhdrs = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "pyart/io/_sigmetfile.pyx":159
 * 
 *         # read in data sweep by sweep
 *         for i in xrange(nsweeps):             # <<<<<<<<<<<<<<
 *             ingest_data_hdrs, sweep_data, sweep_metadata = self._get_sweep(
 *                 full_xhdr=full_xhdr)
 */
  __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_xrange, __pyx_v_nsweeps); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
    __pyx_t_1 = __pyx_t_8; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 159, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 159, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_8); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 159, __pyx_L1_error)
        #else
        __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 159, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 159, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_8); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 159, __pyx_L1_error)
        #else
        __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 159, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      }
    } else {
      __pyx_t_8 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_8)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 159, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_8);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "pyart/io/_sigmetfile.pyx":160
 *         # read in data sweep by sweep
 *         for i in xrange(nsweeps):
 *             ingest_data_hdrs, sweep_data, sweep_metadata = self._get_sweep(             # <<<<<<<<<<<<<<
 *                 full_xhdr=full_xhdr)
 * 
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_sweep); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "pyart/io/_sigmetfile.pyx":161
 *         for i in xrange(nsweeps):
 *             ingest_data_hdrs, sweep_data, sweep_metadata = self._get_sweep(
 *                 full_xhdr=full_xhdr)             # <<<<<<<<<<<<<<
 * 
 *             # check for a truncated file, return sweep(s) read up until error
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_full_xhdr, __pyx_v_full_xhdr) < 0) __PYX_ERR(0, 161, __pyx_L1_error)

    /* "pyart/io/_sigmetfile.pyx":160
 *         # read in data sweep by sweep
 *         for i in xrange(nsweeps):
 *             ingest_data_hdrs, sweep_data, sweep_metadata = self._get_sweep(             # <<<<<<<<<<<<<<
 *                 full_xhdr=full_xhdr)
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
      PyObject* sequence = __pyx_t_5;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 160, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 160, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 160, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_6 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 160, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 160, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
      index = 0; __pyx_t_2 = __pyx_t_10(__pyx_t_7); if (unlikely(!__pyx_t_2)) goto __pyx_L24_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_7); if (unlikely(!__pyx_t_8)) goto __pyx_L24_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      index = 2; __pyx_t_6 = __pyx_t_10(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L24_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_7), 3) < 0) __PYX_ERR(0, 160, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L25_unpacking_done;
      __pyx_L24_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 160, __pyx_L1_error)
      __pyx_L25_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_ingest_data_hdrs, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_sweep_data, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_XDECREF_SET(__pyx_v_sweep_metadata, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "pyart/io/_sigmetfile.pyx":164
 * 
 *             # check for a truncated file, return sweep(s) read up until error
 *             if ingest_data_hdrs is None:             # <<<<<<<<<<<<<<
 * 
 *                 mess = ('File truncated or corrupt, %i of %i sweeps read' %
 */
    __pyx_t_9 = (__pyx_v_ingest_data_hdrs == Py_None);
    if (__pyx_t_9) {

      /* "pyart/io/_sigmetfile.pyx":167
 * 
 *                 mess = ('File truncated or corrupt, %i of %i sweeps read' %
 *                         (i, nsweeps))             # <<<<<<<<<<<<<<
 *                 warnings.warn(mess)
 * 
 */
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 167, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_i)) __PYX_ERR(0, 167, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_nsweeps);
      __Pyx_GIVEREF(__pyx_v_nsweeps);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_nsweeps)) __PYX_ERR(0, 167, __pyx_L1_error);

      /* "pyart/io/_sigmetfile.pyx":166
 *             if ingest_data_hdrs is None:
 * 
 *                 mess = ('File truncated or corrupt, %i of %i sweeps read' %             # <<<<<<<<<<<<<<
 *                         (i, nsweeps))
 *                 warnings.warn(mess)
 */
      __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_File_truncated_or_corrupt_i_of_i, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_mess = ((PyObject*)__pyx_t_6);
      __pyx_t_6 = 0;

      /* "pyart/io/_sigmetfile.pyx":168
 *                 mess = ('File truncated or corrupt, %i of %i sweeps read' %
 *                         (i, nsweeps))
 *                 warnings.warn(mess)             # <<<<<<<<<<<<<<
 * 
 *                 for name in self.data_type_names:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_warnings); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_warn); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      __pyx_t_11 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_11 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_mess};
        __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 168, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "pyart/io/_sigmetfile.pyx":170
 *                 warnings.warn(mess)
 * 
 *                 for name in self.data_type_names:             # <<<<<<<<<<<<<<
 *                     data[name] = data[name][:i]
 *                     for k in metadata[name]:
 */
      if (likely(PyList_CheckExact(__pyx_v_self->data_type_names)) || PyTuple_CheckExact(__pyx_v_self->data_type_names)) {
        __pyx_t_6 = __pyx_v_self->data_type_names; __Pyx_INCREF(__pyx_t_6);
        __pyx_t_12 = 0;
        __pyx_t_13 = NULL;
      } else {
        __pyx_t_12 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_self->data_type_names); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 170, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_13 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 170, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_13)) {
          if (likely(PyList_CheckExact(__pyx_t_6))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_6);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 170, __pyx_L1_error)
              #endif
              if (__pyx_t_12 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_8 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_12); __Pyx_INCREF(__pyx_t_8); __pyx_t_12++; if (unlikely((0 < 0))) __PYX_ERR(0, 170, __pyx_L1_error)
            #else
            __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 170, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_6);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 170, __pyx_L1_error)
              #endif
              if (__pyx_t_12 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_12); __Pyx_INCREF(__pyx_t_8); __pyx_t_12++; if (unlikely((0 < 0))) __PYX_ERR(0, 170, __pyx_L1_error)
            #else
            __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 170, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            #endif
          }
        } else {
          __pyx_t_8 = __pyx_t_13(__pyx_t_6);
          if (unlikely(!__pyx_t_8)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 170, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_8);
        __pyx_t_8 = 0;

        /* "pyart/io/_sigmetfile.pyx":171
 * 
 *                 for name in self.data_type_names:
 *                     data[name] = data[name][:i]             # <<<<<<<<<<<<<<
 *                     for k in metadata[name]:
 *                         metadata[name][k] = metadata[name][k][:i]
 */
        __pyx_t_8 = __Pyx_PyDict_GetItem(__pyx_v_data, __pyx_v_name); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 171, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_t_8, 0, 0, NULL, &__pyx_v_i, NULL, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 171, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_v_name, __pyx_t_5) < 0))) __PYX_ERR(0, 171, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "pyart/io/_sigmetfile.pyx":172
 *                 for name in self.data_type_names:
 *                     data[name] = data[name][:i]
 *                     for k in metadata[name]:             # <<<<<<<<<<<<<<
 *                         metadata[name][k] = metadata[name][k][:i]
 *                 return data, metadata
 */
        __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 172, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
          __pyx_t_8 = __pyx_t_5; __Pyx_INCREF(__pyx_t_8);
          __pyx_t_14 = 0;
          __pyx_t_15 = NULL;
        } else {
          __pyx_t_14 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 172, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_15 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 172, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        for (;;) {
          if (likely(!__pyx_t_15)) {
            if (likely(PyList_CheckExact(__pyx_t_8))) {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_8);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 172, __pyx_L1_error)
                #endif
                if (__pyx_t_14 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_14); __Pyx_INCREF(__pyx_t_5); __pyx_t_14++; if (unlikely((0 < 0))) __PYX_ERR(0, 172, __pyx_L1_error)
              #else
              __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 172, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            } else {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_8);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 172, __pyx_L1_error)
                #endif
                if (__pyx_t_14 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_14); __Pyx_INCREF(__pyx_t_5); __pyx_t_14++; if (unlikely((0 < 0))) __PYX_ERR(0, 172, __pyx_L1_error)
              #else
              __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 172, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            }
          } else {
            __pyx_t_5 = __pyx_t_15(__pyx_t_8);
            if (unlikely(!__pyx_t_5)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 172, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_5);
          }
          __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "pyart/io/_sigmetfile.pyx":173
 *                     data[name] = data[name][:i]
 *                     for k in metadata[name]:
 *                         metadata[name][k] = metadata[name][k][:i]             # <<<<<<<<<<<<<<
 *                 return data, metadata
 * 
 */
          __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 173, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_v_k); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, 0, NULL, &__pyx_v_i, NULL, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 173, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (unlikely((PyObject_SetItem(__pyx_t_2, __pyx_v_k, __pyx_t_5) < 0))) __PYX_ERR(0, 173, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "pyart/io/_sigmetfile.pyx":172
 *                 for name in self.data_type_names:
 *                     data[name] = data[name][:i]
 *                     for k in metadata[name]:             # <<<<<<<<<<<<<<
 *                         metadata[name][k] = metadata[name][k][:i]
 *                 return data, metadata
 */
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "pyart/io/_sigmetfile.pyx":170
 *                 warnings.warn(mess)
 * 
 *                 for name in self.data_type_names:             # <<<<<<<<<<<<<<
 *                     data[name] = data[name][:i]
 *                     for k in metadata[name]:
 */
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "pyart/io/_sigmetfile.pyx":174
 *                     for k in metadata[name]:
 *                         metadata[name][k] = metadata[name][k][:i]
 *                 return data, metadata             # <<<<<<<<<<<<<<
 * 
 *             for j, name in enumerate(self.data_type_names):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 174, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_v_data);
      __Pyx_GIVEREF(__pyx_v_data);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_data)) __PYX_ERR(0, 174, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_metadata);
      __Pyx_GIVEREF(__pyx_v_metadata);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_metadata)) __PYX_ERR(0, 174, __pyx_L1_error);
      __pyx_r = __pyx_t_6;
      __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "pyart/io/_sigmetfile.pyx":164
 * 
 *             # check for a truncated file, return sweep(s) read up until error
 *             if ingest_data_hdrs is None:             # <<<<<<<<<<<<<<
 * 
 *                 mess = ('File truncated or corrupt, %i of %i sweeps read' %
 */
    }

    /* "pyart/io/_sigmetfile.pyx":176
 *                 return data, metadata
 * 
 *             for j, name in enumerate(self.data_type_names):             # <<<<<<<<<<<<<<
 *                 temp = sweep_metadata[j]
 *                 (az0, el0, az1, el1, ray_nbins, ray_time, prf_flag) = temp
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_6 = __pyx_int_0;
    if (likely(PyList_CheckExact(__pyx_v_self->data_type_names)) || PyTuple_CheckExact(__pyx_v_self->data_type_names)) {
      __pyx_t_8 = __pyx_v_self->data_type_names; __Pyx_INCREF(__pyx_t_8);
      __pyx_t_12 = 0;
      __pyx_t_13 = NULL;
    } else {
      __pyx_t_12 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_v_self->data_type_names); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_13 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 176, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_13)) {
        if (likely(PyList_CheckExact(__pyx_t_8))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_8);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 176, __pyx_L1_error)
            #endif
            if (__pyx_t_12 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_12); __Pyx_INCREF(__pyx_t_5); __pyx_t_12++; if (unlikely((0 < 0))) __PYX_ERR(0, 176, __pyx_L1_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 176, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_8);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 176, __pyx_L1_error)
            #endif
            if (__pyx_t_12 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_12); __Pyx_INCREF(__pyx_t_5); __pyx_t_12++; if (unlikely((0 < 0))) __PYX_ERR(0, 176, __pyx_L1_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 176, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_13(__pyx_t_8);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 176, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_6);
      __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_t_6, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6);
      __pyx_t_6 = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "pyart/io/_sigmetfile.pyx":177
 * 
 *             for j, name in enumerate(self.data_type_names):
 *                 temp = sweep_metadata[j]             # <<<<<<<<<<<<<<
 *                 (az0, el0, az1, el1, ray_nbins, ray_time, prf_flag) = temp
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_sweep_metadata, __pyx_v_j); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_temp, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "pyart/io/_sigmetfile.pyx":178
 *             for j, name in enumerate(self.data_type_names):
 *                 temp = sweep_metadata[j]
 *                 (az0, el0, az1, el1, ray_nbins, ray_time, prf_flag) = temp             # <<<<<<<<<<<<<<
 * 
 *                 data[name][i] = sweep_data[j]
 */
      if ((likely(PyTuple_CheckExact(__pyx_v_temp))) || (PyList_CheckExact(__pyx_v_temp))) {
        PyObject* sequence = __pyx_v_temp;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 7)) {
          if (size > 7) __Pyx_RaiseTooManyValuesError(7);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 178, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 2); 
          __pyx_t_16 = PyTuple_GET_ITEM(sequence, 3); 
          __pyx_t_17 = PyTuple_GET_ITEM(sequence, 4); 
          __pyx_t_18 = PyTuple_GET_ITEM(sequence, 5); 
          __pyx_t_19 = PyTuple_GET_ITEM(sequence, 6); 
        } else {
          __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
          __pyx_t_7 = PyList_GET_ITEM(sequence, 2); 
          __pyx_t_16 = PyList_GET_ITEM(sequence, 3); 
          __pyx_t_17 = PyList_GET_ITEM(sequence, 4); 
          __pyx_t_18 = PyList_GET_ITEM(sequence, 5); 
          __pyx_t_19 = PyList_GET_ITEM(sequence, 6); 
        }
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_16);
        __Pyx_INCREF(__pyx_t_17);
        __Pyx_INCREF(__pyx_t_18);
        __Pyx_INCREF(__pyx_t_19);
        #else
        {
          Py_ssize_t i;
          PyObject** temps[7] = {&__pyx_t_5,&__pyx_t_2,&__pyx_t_7,&__pyx_t_16,&__pyx_t_17,&__pyx_t_18,&__pyx_t_19};
          for (i=0; i < 7; i++) {
            PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 178, __pyx_L1_error)
            __Pyx_GOTREF(item);
            *(temps[i]) = item;
          }
        }
        #endif
      } else {
        Py_ssize_t index = -1;
        PyObject** temps[7] = {&__pyx_t_5,&__pyx_t_2,&__pyx_t_7,&__pyx_t_16,&__pyx_t_17,&__pyx_t_18,&__pyx_t_19};
        __pyx_t_20 = PyObject_GetIter(__pyx_v_temp); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 178, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_20);
        __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_20);
        for (index=0; index < 7; index++) {
          PyObject* item = __pyx_t_10(__pyx_t_20); if (unlikely(!item)) goto __pyx_L35_unpacking_failed;
          __Pyx_GOTREF(item);
          *(temps[index]) = item;
        }
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_20), 7) < 0) __PYX_ERR(0, 178, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        goto __pyx_L36_unpacking_done;
        __pyx_L35_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 178, __pyx_L1_error)
        __pyx_L36_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_az0, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_el0, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_az1, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_el1, __pyx_t_16);
      __pyx_t_16 = 0;
      __Pyx_XDECREF_SET(__pyx_v_ray_nbins, __pyx_t_17);
      __pyx_t_17 = 0;
      __Pyx_XDECREF_SET(__pyx_v_ray_time, __pyx_t_18);
      __pyx_t_18 = 0;
      __Pyx_XDECREF_SET(__pyx_v_prf_flag, __pyx_t_19);
      __pyx_t_19 = 0;

      /* "pyart/io/_sigmetfile.pyx":180
 *                 (az0, el0, az1, el1, ray_nbins, ray_time, prf_flag) = temp
 * 
 *                 data[name][i] = sweep_data[j]             # <<<<<<<<<<<<<<
 *                 metadata[name]['azimuth_0'][i] = az0
 *                 metadata[name]['azimuth_1'][i] = az1
 */
      __pyx_t_19 = __Pyx_PyObject_GetItem(__pyx_v_sweep_data, __pyx_v_j); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 180, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_18 = __Pyx_PyDict_GetItem(__pyx_v_data, __pyx_v_name); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 180, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      if (unlikely((PyObject_SetItem(__pyx_t_18, __pyx_v_i, __pyx_t_19) < 0))) __PYX_ERR(0, 180, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

      /* "pyart/io/_sigmetfile.pyx":181
 * 
 *                 data[name][i] = sweep_data[j]
 *                 metadata[name]['azimuth_0'][i] = az0             # <<<<<<<<<<<<<<
 *                 metadata[name]['azimuth_1'][i] = az1
 *                 metadata[name]['elevation_0'][i] = el0
 */
      __pyx_t_19 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_18 = __Pyx_PyObject_Dict_GetItem(__pyx_t_19, __pyx_n_u_azimuth_0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      if (unlikely((PyObject_SetItem(__pyx_t_18, __pyx_v_i, __pyx_v_az0) < 0))) __PYX_ERR(0, 181, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

      /* "pyart/io/_sigmetfile.pyx":182
 *                 data[name][i] = sweep_data[j]
 *                 metadata[name]['azimuth_0'][i] = az0
 *                 metadata[name]['azimuth_1'][i] = az1             # <<<<<<<<<<<<<<
 *                 metadata[name]['elevation_0'][i] = el0
 *                 metadata[name]['elevation_1'][i] = el1
 */
      __pyx_t_18 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 182, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_t_19 = __Pyx_PyObject_Dict_GetItem(__pyx_t_18, __pyx_n_u_azimuth_1); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 182, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      if (unlikely((PyObject_SetItem(__pyx_t_19, __pyx_v_i, __pyx_v_az1) < 0))) __PYX_ERR(0, 182, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

      /* "pyart/io/_sigmetfile.pyx":183
 *                 metadata[name]['azimuth_0'][i] = az0
 *                 metadata[name]['azimuth_1'][i] = az1
 *                 metadata[name]['elevation_0'][i] = el0             # <<<<<<<<<<<<<<
 *                 metadata[name]['elevation_1'][i] = el1
 *                 metadata[name]['nbins'][i] = ray_nbins
 */
      __pyx_t_19 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 183, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_18 = __Pyx_PyObject_Dict_GetItem(__pyx_t_19, __pyx_n_u_elevation_0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 183, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      if (unlikely((PyObject_SetItem(__pyx_t_18, __pyx_v_i, __pyx_v_el0) < 0))) __PYX_ERR(0, 183, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

      /* "pyart/io/_sigmetfile.pyx":184
 *                 metadata[name]['azimuth_1'][i] = az1
 *                 metadata[name]['elevation_0'][i] = el0
 *                 metadata[name]['elevation_1'][i] = el1             # <<<<<<<<<<<<<<
 *                 metadata[name]['nbins'][i] = ray_nbins
 *                 metadata[name]['time'][i] = ray_time
 */
      __pyx_t_18 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_t_19 = __Pyx_PyObject_Dict_GetItem(__pyx_t_18, __pyx_n_u_elevation_1); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      if (unlikely((PyObject_SetItem(__pyx_t_19, __pyx_v_i, __pyx_v_el1) < 0))) __PYX_ERR(0, 184, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

      /* "pyart/io/_sigmetfile.pyx":185
 *                 metadata[name]['elevation_0'][i] = el0
 *                 metadata[name]['elevation_1'][i] = el1
 *                 metadata[name]['nbins'][i] = ray_nbins             # <<<<<<<<<<<<<<
 *                 metadata[name]['time'][i] = ray_time
 *                 metadata[name]['prf_flag'][i] = prf_flag
 */
      __pyx_t_19 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_18 = __Pyx_PyObject_Dict_GetItem(__pyx_t_19, __pyx_n_u_nbins); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      if (unlikely((PyObject_SetItem(__pyx_t_18, __pyx_v_i, __pyx_v_ray_nbins) < 0))) __PYX_ERR(0, 185, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

      /* "pyart/io/_sigmetfile.pyx":186
 *                 metadata[name]['elevation_1'][i] = el1
 *                 metadata[name]['nbins'][i] = ray_nbins
 *                 metadata[name]['time'][i] = ray_time             # <<<<<<<<<<<<<<
 *                 metadata[name]['prf_flag'][i] = prf_flag
 *                 self.ingest_data_headers[name].append(ingest_data_hdrs[j])
 */
      __pyx_t_18 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_t_19 = __Pyx_PyObject_Dict_GetItem(__pyx_t_18, __pyx_n_u_time); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      if (unlikely((PyObject_SetItem(__pyx_t_19, __pyx_v_i, __pyx_v_ray_time) < 0))) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

      /* "pyart/io/_sigmetfile.pyx":187
 *                 metadata[name]['nbins'][i] = ray_nbins
 *                 metadata[name]['time'][i] = ray_time
 *                 metadata[name]['prf_flag'][i] = prf_flag             # <<<<<<<<<<<<<<
 *                 self.ingest_data_headers[name].append(ingest_data_hdrs[j])
 * 
 */
      __pyx_t_19 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_18 = __Pyx_PyObject_Dict_GetItem(__pyx_t_19, __pyx_n_u_prf_flag); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      if (unlikely((PyObject_SetItem(__pyx_t_18, __pyx_v_i, __pyx_v_prf_flag) < 0))) __PYX_ERR(0, 187, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

      /* "pyart/io/_sigmetfile.pyx":188
 *                 metadata[name]['time'][i] = ray_time
 *                 metadata[name]['prf_flag'][i] = prf_flag
 *                 self.ingest_data_headers[name].append(ingest_data_hdrs[j])             # <<<<<<<<<<<<<<
 * 
 *         # scale 1-byte velocity by the Nyquist (section 4.3.29)
 */
      __pyx_t_18 = __Pyx_PyObject_GetItem(__pyx_v_self->ingest_data_headers, __pyx_v_name); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 188, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_t_19 = __Pyx_PyObject_GetItem(__pyx_v_ingest_data_hdrs, __pyx_v_j); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 188, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_21 = __Pyx_PyObject_Append(__pyx_t_18, __pyx_t_19); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 188, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

      /* "pyart/io/_sigmetfile.pyx":176
 *                 return data, metadata
 * 
 *             for j, name in enumerate(self.data_type_names):             # <<<<<<<<<<<<<<
 *                 temp = sweep_metadata[j]
 *                 (az0, el0, az1, el1, ray_nbins, ray_time, prf_flag) = temp
 */
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "pyart/io/_sigmetfile.pyx":159
 * 
 *         # read in data sweep by sweep
 *         for i in xrange(nsweeps):             # <<<<<<<<<<<<<<
 *             ingest_data_hdrs, sweep_data, sweep_metadata = self._get_sweep(
 *                 full_xhdr=full_xhdr)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":194
 *         # product_hdr does not need to be accessed at lower abstraction
 *         # layers.
 *         if 'VEL' in self.data_type_names:             # <<<<<<<<<<<<<<
 *             wavelength_cm = self.product_hdr['product_end']['wavelength']
 *             prt_value = 1. / self.product_hdr['product_end']['prf']
 */
  __pyx_t_9 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_VEL, __pyx_v_self->data_type_names, Py_EQ)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 194, __pyx_L1_error)
  if (__pyx_t_9) {

    /* "pyart/io/_sigmetfile.pyx":195
 *         # layers.
 *         if 'VEL' in self.data_type_names:
 *             wavelength_cm = self.product_hdr['product_end']['wavelength']             # <<<<<<<<<<<<<<
 *             prt_value = 1. / self.product_hdr['product_end']['prf']
 *             task_config = self.ingest_header['task_configuration']
 */
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_self->product_hdr, __pyx_n_u_product_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_wavelength); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_wavelength_cm = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "pyart/io/_sigmetfile.pyx":196
 *         if 'VEL' in self.data_type_names:
 *             wavelength_cm = self.product_hdr['product_end']['wavelength']
 *             prt_value = 1. / self.product_hdr['product_end']['prf']             # <<<<<<<<<<<<<<
 *             task_config = self.ingest_header['task_configuration']
 *             multi_prf_flag = task_config['task_dsp_info']['multi_prf_flag']
 */
    __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_self->product_hdr, __pyx_n_u_product_end); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_6, __pyx_n_u_prf); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyFloat_TrueDivideCObj(__pyx_float_1_, __pyx_t_1, 1., 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_prt_value = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "pyart/io/_sigmetfile.pyx":197
 *             wavelength_cm = self.product_hdr['product_end']['wavelength']
 *             prt_value = 1. / self.product_hdr['product_end']['prf']
 *             task_config = self.ingest_header['task_configuration']             # <<<<<<<<<<<<<<
 *             multi_prf_flag = task_config['task_dsp_info']['multi_prf_flag']
 *             if multi_prf_flag > 3 or multi_prf_flag < 0:
 */
    __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_self->ingest_header, __pyx_n_u_task_configuration); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_v_task_config = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "pyart/io/_sigmetfile.pyx":198
 *             prt_value = 1. / self.product_hdr['product_end']['prf']
 *             task_config = self.ingest_header['task_configuration']
 *             multi_prf_flag = task_config['task_dsp_info']['multi_prf_flag']             # <<<<<<<<<<<<<<
 *             if multi_prf_flag > 3 or multi_prf_flag < 0:
 *                 multiplier = 1  # multiplier not defined in IRIS manual
 */
    __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_task_config, __pyx_n_u_task_dsp_info); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_6, __pyx_n_u_multi_prf_flag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_multi_prf_flag = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":199
 *             task_config = self.ingest_header['task_configuration']
 *             multi_prf_flag = task_config['task_dsp_info']['multi_prf_flag']
 *             if multi_prf_flag > 3 or multi_prf_flag < 0:             # <<<<<<<<<<<<<<
 *                 multiplier = 1  # multiplier not defined in IRIS manual
 *             else:
 */
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_multi_prf_flag, __pyx_int_3, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error)
    __pyx_t_22 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_22 < 0))) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_22) {
    } else {
      __pyx_t_9 = __pyx_t_22;
      goto __pyx_L41_bool_binop_done;
    }
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_multi_prf_flag, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error)
    __pyx_t_22 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_22 < 0))) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = __pyx_t_22;
    __pyx_L41_bool_binop_done:;
    if (__pyx_t_9) {

      /* "pyart/io/_sigmetfile.pyx":200
 *             multi_prf_flag = task_config['task_dsp_info']['multi_prf_flag']
 *             if multi_prf_flag > 3 or multi_prf_flag < 0:
 *                 multiplier = 1  # multiplier not defined in IRIS manual             # <<<<<<<<<<<<<<
 *             else:
 *                 multiplier = [1, 2, 3, 4][multi_prf_flag]
 */
      __Pyx_INCREF(__pyx_int_1);
      __pyx_v_multiplier = __pyx_int_1;

      /* "pyart/io/_sigmetfile.pyx":199
 *             task_config = self.ingest_header['task_configuration']
 *             multi_prf_flag = task_config['task_dsp_info']['multi_prf_flag']
 *             if multi_prf_flag > 3 or multi_prf_flag < 0:             # <<<<<<<<<<<<<<
 *                 multiplier = 1  # multiplier not defined in IRIS manual
 *             else:
 */
      goto __pyx_L40;
    }

    /* "pyart/io/_sigmetfile.pyx":202
 *                 multiplier = 1  # multiplier not defined in IRIS manual
 *             else:
 *                 multiplier = [1, 2, 3, 4][multi_prf_flag]             # <<<<<<<<<<<<<<
 *             nyquist = wavelength_cm / (10000.0 * 4.0 * prt_value) * multiplier
 *             data['VEL'] *= nyquist
 */
    /*else*/ {
      __pyx_t_1 = PyList_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_int_1)) __PYX_ERR(0, 202, __pyx_L1_error);
      __Pyx_INCREF(__pyx_int_2);
      __Pyx_GIVEREF(__pyx_int_2);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_int_2)) __PYX_ERR(0, 202, __pyx_L1_error);
      __Pyx_INCREF(__pyx_int_3);
      __Pyx_GIVEREF(__pyx_int_3);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, __pyx_int_3)) __PYX_ERR(0, 202, __pyx_L1_error);
      __Pyx_INCREF(__pyx_int_4);
      __Pyx_GIVEREF(__pyx_int_4);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 3, __pyx_int_4)) __PYX_ERR(0, 202, __pyx_L1_error);
      __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_multi_prf_flag); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_multiplier = __pyx_t_6;
      __pyx_t_6 = 0;
    }
    __pyx_L40:;

    /* "pyart/io/_sigmetfile.pyx":203
 *             else:
 *                 multiplier = [1, 2, 3, 4][multi_prf_flag]
 *             nyquist = wavelength_cm / (10000.0 * 4.0 * prt_value) * multiplier             # <<<<<<<<<<<<<<
 *             data['VEL'] *= nyquist
 *         # scale 1-byte width by the Nyquist
 */
    __pyx_t_6 = PyFloat_FromDouble((10000.0 * 4.0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyNumber_Multiply(__pyx_t_6, __pyx_v_prt_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyNumber_Divide(__pyx_v_wavelength_cm, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Multiply(__pyx_t_6, __pyx_v_multiplier); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_nyquist = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":204
 *                 multiplier = [1, 2, 3, 4][multi_prf_flag]
 *             nyquist = wavelength_cm / (10000.0 * 4.0 * prt_value) * multiplier
 *             data['VEL'] *= nyquist             # <<<<<<<<<<<<<<
 *         # scale 1-byte width by the Nyquist
 *         if 'WIDTH' in self.data_type_names:
 */
    __Pyx_INCREF(__pyx_n_u_VEL);
    __pyx_t_23 = __pyx_n_u_VEL;
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_data, __pyx_t_23); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyNumber_InPlaceMultiply(__pyx_t_1, __pyx_v_nyquist); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_t_23, __pyx_t_6) < 0))) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "pyart/io/_sigmetfile.pyx":194
 *         # product_hdr does not need to be accessed at lower abstraction
 *         # layers.
 *         if 'VEL' in self.data_type_names:             # <<<<<<<<<<<<<<
 *             wavelength_cm = self.product_hdr['product_end']['wavelength']
 *             prt_value = 1. / self.product_hdr['product_end']['prf']
 */
  }

  /* "pyart/io/_sigmetfile.pyx":206
 *             data['VEL'] *= nyquist
 *         # scale 1-byte width by the Nyquist
 *         if 'WIDTH' in self.data_type_names:             # <<<<<<<<<<<<<<
 *             # The IRIS Programmer's Manual indicates 1-byte width format data
 *             # should be scaled by the unambiguous velocity, twice the nyquist,
 */
  __pyx_t_9 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_WIDTH, __pyx_v_self->data_type_names, Py_EQ)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 206, __pyx_L1_error)
  if (__pyx_t_9) {

    /* "pyart/io/_sigmetfile.pyx":212
 *             # nyquist.  Therefore to agree with these two packages the width
 *             # is scaled by the nyquist.
 *             wavelength_cm = self.product_hdr['product_end']['wavelength']             # <<<<<<<<<<<<<<
 *             prt_value = 1. / self.product_hdr['product_end']['prf']
 *             nyquist = wavelength_cm / (10000.0 * 4.0 * prt_value)
 */
    __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_self->product_hdr, __pyx_n_u_product_end); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_6, __pyx_n_u_wavelength); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_wavelength_cm, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":213
 *             # is scaled by the nyquist.
 *             wavelength_cm = self.product_hdr['product_end']['wavelength']
 *             prt_value = 1. / self.product_hdr['product_end']['prf']             # <<<<<<<<<<<<<<
 *             nyquist = wavelength_cm / (10000.0 * 4.0 * prt_value)
 *             data['WIDTH'] *= nyquist
 */
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_self->product_hdr, __pyx_n_u_product_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_prf); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyFloat_TrueDivideCObj(__pyx_float_1_, __pyx_t_6, 1., 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_prt_value, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":214
 *             wavelength_cm = self.product_hdr['product_end']['wavelength']
 *             prt_value = 1. / self.product_hdr['product_end']['prf']
 *             nyquist = wavelength_cm / (10000.0 * 4.0 * prt_value)             # <<<<<<<<<<<<<<
 *             data['WIDTH'] *= nyquist
 *         # scale 1-byte KDP by the wavelength
 */
    __pyx_t_1 = PyFloat_FromDouble((10000.0 * 4.0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyNumber_Multiply(__pyx_t_1, __pyx_v_prt_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_v_wavelength_cm, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_nyquist, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":215
 *             prt_value = 1. / self.product_hdr['product_end']['prf']
 *             nyquist = wavelength_cm / (10000.0 * 4.0 * prt_value)
 *             data['WIDTH'] *= nyquist             # <<<<<<<<<<<<<<
 *         # scale 1-byte KDP by the wavelength
 *         if 'KDP' in self.data_type_names:
 */
    __Pyx_INCREF(__pyx_n_u_WIDTH);
    __pyx_t_23 = __pyx_n_u_WIDTH;
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_data, __pyx_t_23); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyNumber_InPlaceMultiply(__pyx_t_1, __pyx_v_nyquist); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_t_23, __pyx_t_6) < 0))) __PYX_ERR(0, 215, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "pyart/io/_sigmetfile.pyx":206
 *             data['VEL'] *= nyquist
 *         # scale 1-byte width by the Nyquist
 *         if 'WIDTH' in self.data_type_names:             # <<<<<<<<<<<<<<
 *             # The IRIS Programmer's Manual indicates 1-byte width format data
 *             # should be scaled by the unambiguous velocity, twice the nyquist,
 */
  }

  /* "pyart/io/_sigmetfile.pyx":217
 *             data['WIDTH'] *= nyquist
 *         # scale 1-byte KDP by the wavelength
 *         if 'KDP' in self.data_type_names:             # <<<<<<<<<<<<<<
 *             # The IRIS Programmer's Manual indicates 1-byte differential phase format
 *             # data should be divided by the wavelength in cm (section 4.3.12).
 */
  __pyx_t_9 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_KDP, __pyx_v_self->data_type_names, Py_EQ)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 217, __pyx_L1_error)
  if (__pyx_t_9) {

    /* "pyart/io/_sigmetfile.pyx":220
 *             # The IRIS Programmer's Manual indicates 1-byte differential phase format
 *             # data should be divided by the wavelength in cm (section 4.3.12).
 *             wavelength_cm = self.product_hdr['product_end']['wavelength']             # <<<<<<<<<<<<<<
 *             data['KDP'] /= (wavelength_cm / 100.0)
 *         return data, metadata
 */
    __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_self->product_hdr, __pyx_n_u_product_end); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_6, __pyx_n_u_wavelength); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_wavelength_cm, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":221
 *             # data should be divided by the wavelength in cm (section 4.3.12).
 *             wavelength_cm = self.product_hdr['product_end']['wavelength']
 *             data['KDP'] /= (wavelength_cm / 100.0)             # <<<<<<<<<<<<<<
 *         return data, metadata
 * 
 */
    __Pyx_INCREF(__pyx_n_u_KDP);
    __pyx_t_23 = __pyx_n_u_KDP;
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_data, __pyx_t_23); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyFloat_TrueDivideObjC(__pyx_v_wavelength_cm, __pyx_float_100_0, 100.0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyNumber_InPlaceDivide(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_t_23, __pyx_t_8) < 0))) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "pyart/io/_sigmetfile.pyx":217
 *             data['WIDTH'] *= nyquist
 *         # scale 1-byte KDP by the wavelength
 *         if 'KDP' in self.data_type_names:             # <<<<<<<<<<<<<<
 *             # The IRIS Programmer's Manual indicates 1-byte differential phase format
 *             # data should be divided by the wavelength in cm (section 4.3.12).
 */
  }

  /* "pyart/io/_sigmetfile.pyx":222
 *             wavelength_cm = self.product_hdr['product_end']['wavelength']
 *             data['KDP'] /= (wavelength_cm / 100.0)
 *         return data, metadata             # <<<<<<<<<<<<<<
 * 
 *     def _get_sweep(self, full_xhdr=False, raw_data=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_data)) __PYX_ERR(0, 222, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_metadata);
  __Pyx_GIVEREF(__pyx_v_metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_metadata)) __PYX_ERR(0, 222, __pyx_L1_error);
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":99
 *         self._fh.close()
 * 
 *     def read_data(self, full_xhdr=False):             # <<<<<<<<<<<<<<
 *         """
 *         Read all data from the file.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile.read_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nsweeps);
  __Pyx_XDECREF(__pyx_v_nbins);
  __Pyx_XDECREF(__pyx_v_nrays);
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_metadata);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_header_dic);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_ingest_data_hdrs);
  __Pyx_XDECREF(__pyx_v_sweep_data);
  __Pyx_XDECREF(__pyx_v_sweep_metadata);
  __Pyx_XDECREF(__pyx_v_mess);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XDECREF(__pyx_v_temp);
  __Pyx_XDECREF(__pyx_v_az0);
  __Pyx_XDECREF(__pyx_v_el0);
  __Pyx_XDECREF(__pyx_v_az1);
  __Pyx_XDECREF(__pyx_v_el1);
  __Pyx_XDECREF(__pyx_v_ray_nbins);
  __Pyx_XDECREF(__pyx_v_ray_time);
  __Pyx_XDECREF(__pyx_v_prf_flag);
  __Pyx_XDECREF(__pyx_v_wavelength_cm);
  __Pyx_XDECREF(__pyx_v_prt_value);
  __Pyx_XDECREF(__pyx_v_task_config);
  __Pyx_XDECREF(__pyx_v_multi_prf_flag);
  __Pyx_XDECREF(__pyx_v_multiplier);
  __Pyx_XDECREF(__pyx_v_nyquist);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_name);
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":224
 *         return data, metadata
 * 
 *     def _get_sweep(self, full_xhdr=False, raw_data=False):             # <<<<<<<<<<<<<<
 *         """
 *         Get the data and metadata from the next sweep.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_get_sweep(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_8_get_sweep, "\n        Get the data and metadata from the next sweep.\n\n        If the file ends early None is returned for all values.\n\n        Parameters\n        ----------\n        full_xhdr : bool\n            True to return the full extended headers if they exist padded with\n            ones.  False will return a length 1 extended header converted to\n            int32.  This is useful when the file contains a customer specified\n            extended header (for example aircraft radar).\n        raw_data : bool, optional\n            True to return the raw_data for the given sweep, False to\n            convert the data to floating point representation.\n\n        Returns\n        -------\n        ingest_data_headers : list of dict\n            List of ingest_data_header structures for each data type.\n        sweep_data : list of arrays\n            Sweep data for each data types in the order they appear in the\n            file.\n        sweep_metadata : list of tuples\n            Sweep metadata for each data type in the same order as sweep_data.\n\n        ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_10SigmetFile_9_get_sweep = {"_get_sweep", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_get_sweep, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_8_get_sweep};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_get_sweep(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_full_xhdr = 0;
  PyObject *__pyx_v_raw_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_sweep (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_full_xhdr,&__pyx_n_s_raw_data,0};
    values[0] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_False));
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_False));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_full_xhdr);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 224, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_raw_data);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 224, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_sweep") < 0)) __PYX_ERR(0, 224, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_full_xhdr = values[0];
    __pyx_v_raw_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_sweep", 0, 0, 2, __pyx_nargs); __PYX_ERR(0, 224, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._get_sweep", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_8_get_sweep(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), __pyx_v_full_xhdr, __pyx_v_raw_data);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_8_get_sweep(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_full_xhdr, PyObject *__pyx_v_raw_data) {
  PyObject *__pyx_v_lead_record = NULL;
  PyObject *__pyx_v_raw_prod_bhdr = NULL;
  PyObject *__pyx_v_ingest_data_headers = NULL;
  PyObject *__pyx_v_nray_data_types = NULL;
  PyObject *__pyx_v_nrays = NULL;
  PyObject *__pyx_v_nbins = NULL;
  PyObject *__pyx_v_raw_sweep_data = NULL;
  PyObject *__pyx_v_ray_i = NULL;
  PyObject *__pyx_v_sweep_data = NULL;
  PyObject *__pyx_v_sweep_metadata = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_data_type = NULL;
  PyObject *__pyx_8genexpr3__pyx_v_d = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_sweep", 1);

  /* "pyart/io/_sigmetfile.pyx":254
 * 
 *         # get the next record
 *         lead_record = self._fh.read(RECORD_SIZE)             # <<<<<<<<<<<<<<
 *         self._record_number += 1
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_fh, __pyx_n_s_read); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_RECORD_SIZE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_lead_record = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":255
 *         # get the next record
 *         lead_record = self._fh.read(RECORD_SIZE)
 *         self._record_number += 1             # <<<<<<<<<<<<<<
 * 
 *         # check if the file ended early, if so return Nones
 */
  __pyx_v_self->_record_number = (__pyx_v_self->_record_number + 1);

  /* "pyart/io/_sigmetfile.pyx":258
 * 
 *         # check if the file ended early, if so return Nones
 *         if len(lead_record) != RECORD_SIZE:             # <<<<<<<<<<<<<<
 *             return None, None, None
 * 
 */
  __pyx_t_6 = PyObject_Length(__pyx_v_lead_record); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 258, __pyx_L1_error)
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_RECORD_SIZE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_7) {

    /* "pyart/io/_sigmetfile.pyx":259
 *         # check if the file ended early, if so return Nones
 *         if len(lead_record) != RECORD_SIZE:
 *             return None, None, None             # <<<<<<<<<<<<<<
 * 
 *         # unpack structures
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_tuple__3);
    __pyx_r = __pyx_tuple__3;
    goto __pyx_L0;

    /* "pyart/io/_sigmetfile.pyx":258
 * 
 *         # check if the file ended early, if so return Nones
 *         if len(lead_record) != RECORD_SIZE:             # <<<<<<<<<<<<<<
 *             return None, None, None
 * 
 */
  }

  /* "pyart/io/_sigmetfile.pyx":262
 * 
 *         # unpack structures
 *         raw_prod_bhdr = _unpack_raw_prod_bhdr(lead_record)             # <<<<<<<<<<<<<<
 *         self._raw_product_bhdrs.append([raw_prod_bhdr])
 *         ingest_data_headers = _unpack_ingest_data_headers(
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_raw_prod_bhdr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_lead_record};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_raw_prod_bhdr = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":263
 *         # unpack structures
 *         raw_prod_bhdr = _unpack_raw_prod_bhdr(lead_record)
 *         self._raw_product_bhdrs.append([raw_prod_bhdr])             # <<<<<<<<<<<<<<
 *         ingest_data_headers = _unpack_ingest_data_headers(
 *             lead_record, self.ndata_types)
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_raw_prod_bhdr);
  __Pyx_GIVEREF(__pyx_v_raw_prod_bhdr);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_raw_prod_bhdr)) __PYX_ERR(0, 263, __pyx_L1_error);
  __pyx_t_8 = __Pyx_PyObject_Append(__pyx_v_self->_raw_product_bhdrs, __pyx_t_3); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":264
 *         raw_prod_bhdr = _unpack_raw_prod_bhdr(lead_record)
 *         self._raw_product_bhdrs.append([raw_prod_bhdr])
 *         ingest_data_headers = _unpack_ingest_data_headers(             # <<<<<<<<<<<<<<
 *             lead_record, self.ndata_types)
 *         if ingest_data_headers is None:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_ingest_data_headers); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "pyart/io/_sigmetfile.pyx":265
 *         self._raw_product_bhdrs.append([raw_prod_bhdr])
 *         ingest_data_headers = _unpack_ingest_data_headers(
 *             lead_record, self.ndata_types)             # <<<<<<<<<<<<<<
 *         if ingest_data_headers is None:
 *             return None, None, None
 */
  __pyx_t_1 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_lead_record, __pyx_v_self->ndata_types};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_ingest_data_headers = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":266
 *         ingest_data_headers = _unpack_ingest_data_headers(
 *             lead_record, self.ndata_types)
 *         if ingest_data_headers is None:             # <<<<<<<<<<<<<<
 *             return None, None, None
 * 
 */
  __pyx_t_7 = (__pyx_v_ingest_data_headers == Py_None);
  if (__pyx_t_7) {

    /* "pyart/io/_sigmetfile.pyx":267
 *             lead_record, self.ndata_types)
 *         if ingest_data_headers is None:
 *             return None, None, None             # <<<<<<<<<<<<<<
 * 
 *         # determine size of data
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_tuple__3);
    __pyx_r = __pyx_tuple__3;
    goto __pyx_L0;

    /* "pyart/io/_sigmetfile.pyx":266
 *         ingest_data_headers = _unpack_ingest_data_headers(
 *             lead_record, self.ndata_types)
 *         if ingest_data_headers is None:             # <<<<<<<<<<<<<<
 *             return None, None, None
 * 
 */
  }

  /* "pyart/io/_sigmetfile.pyx":270
 * 
 *         # determine size of data
 *         nray_data_types = [d['number_rays_file_expected']             # <<<<<<<<<<<<<<
 *                            for d in ingest_data_headers]
 *         nrays = sum(nray_data_types)    # total rays
 */
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 270, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "pyart/io/_sigmetfile.pyx":271
 *         # determine size of data
 *         nray_data_types = [d['number_rays_file_expected']
 *                            for d in ingest_data_headers]             # <<<<<<<<<<<<<<
 *         nrays = sum(nray_data_types)    # total rays
 *         nbins = self.product_hdr['product_end']['number_bins']
 */
    if (likely(PyList_CheckExact(__pyx_v_ingest_data_headers)) || PyTuple_CheckExact(__pyx_v_ingest_data_headers)) {
      __pyx_t_2 = __pyx_v_ingest_data_headers; __Pyx_INCREF(__pyx_t_2);
      __pyx_t_6 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_ingest_data_headers); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 271, __pyx_L7_error)
    }
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 271, __pyx_L7_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 271, __pyx_L7_error)
          #else
          __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 271, __pyx_L7_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 271, __pyx_L7_error)
          #else
          __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_9(__pyx_t_2);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 271, __pyx_L7_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_d, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "pyart/io/_sigmetfile.pyx":270
 * 
 *         # determine size of data
 *         nray_data_types = [d['number_rays_file_expected']             # <<<<<<<<<<<<<<
 *                            for d in ingest_data_headers]
 *         nrays = sum(nray_data_types)    # total rays
 */
      __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_8genexpr3__pyx_v_d, __pyx_n_u_number_rays_file_expected); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 270, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyart/io/_sigmetfile.pyx":271
 *         # determine size of data
 *         nray_data_types = [d['number_rays_file_expected']
 *                            for d in ingest_data_headers]             # <<<<<<<<<<<<<<
 *         nrays = sum(nray_data_types)    # total rays
 *         nbins = self.product_hdr['product_end']['number_bins']
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_d); __pyx_8genexpr3__pyx_v_d = 0;
    goto __pyx_L11_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_d); __pyx_8genexpr3__pyx_v_d = 0;
    goto __pyx_L1_error;
    __pyx_L11_exit_scope:;
  } /* exit inner scope */
  __pyx_v_nray_data_types = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":272
 *         nray_data_types = [d['number_rays_file_expected']
 *                            for d in ingest_data_headers]
 *         nrays = sum(nray_data_types)    # total rays             # <<<<<<<<<<<<<<
 *         nbins = self.product_hdr['product_end']['number_bins']
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_v_nray_data_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_nrays = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":273
 *                            for d in ingest_data_headers]
 *         nrays = sum(nray_data_types)    # total rays
 *         nbins = self.product_hdr['product_end']['number_bins']             # <<<<<<<<<<<<<<
 * 
 *         # prepare to read rays
 */
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_self->product_hdr, __pyx_n_u_product_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_number_bins); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_nbins = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":276
 * 
 *         # prepare to read rays
 *         self._rbuf = np.frombuffer(lead_record, dtype='int16')             # <<<<<<<<<<<<<<
 *         self._rbuf_p = <np.int16_t*>self._rbuf.data
 *         self._rbuf_pos = int((12 + 76 * self.ndata_types) / 2) - 1
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_frombuffer); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_lead_record);
  __Pyx_GIVEREF(__pyx_v_lead_record);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_lead_record)) __PYX_ERR(0, 276, __pyx_L1_error);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_u_int16) < 0) __PYX_ERR(0, 276, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_rbuf);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_rbuf);
  __pyx_v_self->_rbuf = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "pyart/io/_sigmetfile.pyx":277
 *         # prepare to read rays
 *         self._rbuf = np.frombuffer(lead_record, dtype='int16')
 *         self._rbuf_p = <np.int16_t*>self._rbuf.data             # <<<<<<<<<<<<<<
 *         self._rbuf_pos = int((12 + 76 * self.ndata_types) / 2) - 1
 *         # set data initially to ones so that missing data can be better
 */
  __pyx_t_4 = ((PyObject *)__pyx_v_self->_rbuf);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_self->_rbuf_p = ((__pyx_t_5numpy_int16_t *)__pyx_f_5numpy_7ndarray_4data_data(((PyArrayObject *)__pyx_t_4)));
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyart/io/_sigmetfile.pyx":278
 *         self._rbuf = np.frombuffer(lead_record, dtype='int16')
 *         self._rbuf_p = <np.int16_t*>self._rbuf.data
 *         self._rbuf_pos = int((12 + 76 * self.ndata_types) / 2) - 1             # <<<<<<<<<<<<<<
 *         # set data initially to ones so that missing data can be better
 *         # seen when debugging
 */
  __pyx_t_4 = __Pyx_PyInt_MultiplyCObj(__pyx_int_76, __pyx_v_self->ndata_types, 76, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyInt_AddCObj(__pyx_int_12, __pyx_t_4, 12, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_TrueDivideObjC(__pyx_t_1, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 278, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_self->_rbuf_pos = __pyx_t_10;

  /* "pyart/io/_sigmetfile.pyx":281
 *         # set data initially to ones so that missing data can be better
 *         # seen when debugging
 *         raw_sweep_data = np.ones((nrays, nbins + 6), dtype='int16')             # <<<<<<<<<<<<<<
 * 
 *         # get the raw data ray-by-ray
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ones); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_v_nbins, __pyx_int_6, 6, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_nrays);
  __Pyx_GIVEREF(__pyx_v_nrays);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_nrays)) __PYX_ERR(0, 281, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4)) __PYX_ERR(0, 281, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(0, 281, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_n_u_int16) < 0) __PYX_ERR(0, 281, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_raw_sweep_data = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":284
 * 
 *         # get the raw data ray-by-ray
 *         for ray_i in xrange(nrays):             # <<<<<<<<<<<<<<
 *             if self.debug:
 *                 print("Reading ray: %i of %i" % (ray_i, nrays))
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_xrange, __pyx_v_nrays); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_6 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 284, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 284, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 284, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 284, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 284, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_9(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 284, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_ray_i, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyart/io/_sigmetfile.pyx":285
 *         # get the raw data ray-by-ray
 *         for ray_i in xrange(nrays):
 *             if self.debug:             # <<<<<<<<<<<<<<
 *                 print("Reading ray: %i of %i" % (ray_i, nrays))
 *                 print("self._rbuf_pos is", self._rbuf_pos)
 */
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_self->debug); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 285, __pyx_L1_error)
    if (__pyx_t_7) {

      /* "pyart/io/_sigmetfile.pyx":286
 *         for ray_i in xrange(nrays):
 *             if self.debug:
 *                 print("Reading ray: %i of %i" % (ray_i, nrays))             # <<<<<<<<<<<<<<
 *                 print("self._rbuf_pos is", self._rbuf_pos)
 *             if self._get_ray(nbins, raw_sweep_data[ray_i]):
 */
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_ray_i);
      __Pyx_GIVEREF(__pyx_v_ray_i);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_ray_i)) __PYX_ERR(0, 286, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_nrays);
      __Pyx_GIVEREF(__pyx_v_nrays);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_nrays)) __PYX_ERR(0, 286, __pyx_L1_error);
      __pyx_t_4 = PyUnicode_Format(__pyx_kp_u_Reading_ray_i_of_i, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":287
 *             if self.debug:
 *                 print("Reading ray: %i of %i" % (ray_i, nrays))
 *                 print("self._rbuf_pos is", self._rbuf_pos)             # <<<<<<<<<<<<<<
 *             if self._get_ray(nbins, raw_sweep_data[ray_i]):
 *                 return None, None, None
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->_rbuf_pos); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_kp_u_self__rbuf_pos_is);
      __Pyx_GIVEREF(__pyx_kp_u_self__rbuf_pos_is);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_self__rbuf_pos_is)) __PYX_ERR(0, 287, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_3);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(0, 287, __pyx_L1_error);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":285
 *         # get the raw data ray-by-ray
 *         for ray_i in xrange(nrays):
 *             if self.debug:             # <<<<<<<<<<<<<<
 *                 print("Reading ray: %i of %i" % (ray_i, nrays))
 *                 print("self._rbuf_pos is", self._rbuf_pos)
 */
    }

    /* "pyart/io/_sigmetfile.pyx":288
 *                 print("Reading ray: %i of %i" % (ray_i, nrays))
 *                 print("self._rbuf_pos is", self._rbuf_pos)
 *             if self._get_ray(nbins, raw_sweep_data[ray_i]):             # <<<<<<<<<<<<<<
 *                 return None, None, None
 * 
 */
    __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_v_nbins); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 288, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_raw_sweep_data, __pyx_v_ray_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 288, __pyx_L1_error)
    __pyx_t_11 = ((struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self->__pyx_vtab)->_get_ray(__pyx_v_self, __pyx_t_10, ((PyArrayObject *)__pyx_t_3)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 288, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = (__pyx_t_11 != 0);
    if (__pyx_t_7) {

      /* "pyart/io/_sigmetfile.pyx":289
 *                 print("self._rbuf_pos is", self._rbuf_pos)
 *             if self._get_ray(nbins, raw_sweep_data[ray_i]):
 *                 return None, None, None             # <<<<<<<<<<<<<<
 * 
 *         # return raw data if requested
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_tuple__3);
      __pyx_r = __pyx_tuple__3;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "pyart/io/_sigmetfile.pyx":288
 *                 print("Reading ray: %i of %i" % (ray_i, nrays))
 *                 print("self._rbuf_pos is", self._rbuf_pos)
 *             if self._get_ray(nbins, raw_sweep_data[ray_i]):             # <<<<<<<<<<<<<<
 *                 return None, None, None
 * 
 */
    }

    /* "pyart/io/_sigmetfile.pyx":284
 * 
 *         # get the raw data ray-by-ray
 *         for ray_i in xrange(nrays):             # <<<<<<<<<<<<<<
 *             if self.debug:
 *                 print("Reading ray: %i of %i" % (ray_i, nrays))
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":292
 * 
 *         # return raw data if requested
 *         if raw_data:             # <<<<<<<<<<<<<<
 *             return ingest_data_headers, raw_sweep_data
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_raw_data); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 292, __pyx_L1_error)
  if (__pyx_t_7) {

    /* "pyart/io/_sigmetfile.pyx":293
 *         # return raw data if requested
 *         if raw_data:
 *             return ingest_data_headers, raw_sweep_data             # <<<<<<<<<<<<<<
 * 
 *         # convert the data and parse the metadata
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_ingest_data_headers);
    __Pyx_GIVEREF(__pyx_v_ingest_data_headers);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_ingest_data_headers)) __PYX_ERR(0, 293, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_raw_sweep_data);
    __Pyx_GIVEREF(__pyx_v_raw_sweep_data);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_raw_sweep_data)) __PYX_ERR(0, 293, __pyx_L1_error);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pyart/io/_sigmetfile.pyx":292
 * 
 *         # return raw data if requested
 *         if raw_data:             # <<<<<<<<<<<<<<
 *             return ingest_data_headers, raw_sweep_data
 * 
 */
  }

  /* "pyart/io/_sigmetfile.pyx":296
 * 
 *         # convert the data and parse the metadata
 *         sweep_data = []             # <<<<<<<<<<<<<<
 *         sweep_metadata = []
 *         for i, data_type in enumerate(self.data_types):
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_sweep_data = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":297
 *         # convert the data and parse the metadata
 *         sweep_data = []
 *         sweep_metadata = []             # <<<<<<<<<<<<<<
 *         for i, data_type in enumerate(self.data_types):
 *             if data_type == 0 and full_xhdr:
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_sweep_metadata = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":298
 *         sweep_data = []
 *         sweep_metadata = []
 *         for i, data_type in enumerate(self.data_types):             # <<<<<<<<<<<<<<
 *             if data_type == 0 and full_xhdr:
 *                 sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_2 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_self->data_types)) || PyTuple_CheckExact(__pyx_v_self->data_types)) {
    __pyx_t_3 = __pyx_v_self->data_types; __Pyx_INCREF(__pyx_t_3);
    __pyx_t_6 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_self->data_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 298, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 298, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 298, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 298, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 298, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 298, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 298, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_9(__pyx_t_3);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 298, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_data_type, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "pyart/io/_sigmetfile.pyx":299
 *         sweep_metadata = []
 *         for i, data_type in enumerate(self.data_types):
 *             if data_type == 0 and full_xhdr:             # <<<<<<<<<<<<<<
 *                 sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])
 *             else:
 */
    __pyx_t_12 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_data_type, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 299, __pyx_L1_error)
    if (__pyx_t_12) {
    } else {
      __pyx_t_7 = __pyx_t_12;
      goto __pyx_L21_bool_binop_done;
    }
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_full_xhdr); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 299, __pyx_L1_error)
    __pyx_t_7 = __pyx_t_12;
    __pyx_L21_bool_binop_done:;
    if (__pyx_t_7) {

      /* "pyart/io/_sigmetfile.pyx":300
 *         for i, data_type in enumerate(self.data_types):
 *             if data_type == 0 and full_xhdr:
 *                 sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])             # <<<<<<<<<<<<<<
 *             else:
 *                 sweep_data.append(convert_sigmet_data(
 */
      __pyx_t_4 = PySlice_New(__pyx_v_i, Py_None, __pyx_v_self->ndata_types); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 300, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 300, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_4);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(0, 300, __pyx_L1_error);
      __Pyx_INCREF(__pyx_slice__4);
      __Pyx_GIVEREF(__pyx_slice__4);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_slice__4)) __PYX_ERR(0, 300, __pyx_L1_error);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_raw_sweep_data, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 300, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_sweep_data, __pyx_t_4); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 300, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "pyart/io/_sigmetfile.pyx":299
 *         sweep_metadata = []
 *         for i, data_type in enumerate(self.data_types):
 *             if data_type == 0 and full_xhdr:             # <<<<<<<<<<<<<<
 *                 sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])
 *             else:
 */
      goto __pyx_L20;
    }

    /* "pyart/io/_sigmetfile.pyx":302
 *                 sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])
 *             else:
 *                 sweep_data.append(convert_sigmet_data(             # <<<<<<<<<<<<<<
 *                     data_type, raw_sweep_data[i::self.ndata_types, 6:],
 *                     raw_sweep_data[i::self.ndata_types, 4]))
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_convert_sigmet_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "pyart/io/_sigmetfile.pyx":303
 *             else:
 *                 sweep_data.append(convert_sigmet_data(
 *                     data_type, raw_sweep_data[i::self.ndata_types, 6:],             # <<<<<<<<<<<<<<
 *                     raw_sweep_data[i::self.ndata_types, 4]))
 *             sweep_metadata.append(_parse_ray_headers(
 */
      __pyx_t_13 = PySlice_New(__pyx_v_i, Py_None, __pyx_v_self->ndata_types); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 303, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 303, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_GIVEREF(__pyx_t_13);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13)) __PYX_ERR(0, 303, __pyx_L1_error);
      __Pyx_INCREF(__pyx_slice__4);
      __Pyx_GIVEREF(__pyx_slice__4);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_slice__4)) __PYX_ERR(0, 303, __pyx_L1_error);
      __pyx_t_13 = 0;
      __pyx_t_13 = __Pyx_PyObject_GetItem(__pyx_v_raw_sweep_data, __pyx_t_14); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 303, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

      /* "pyart/io/_sigmetfile.pyx":304
 *                 sweep_data.append(convert_sigmet_data(
 *                     data_type, raw_sweep_data[i::self.ndata_types, 6:],
 *                     raw_sweep_data[i::self.ndata_types, 4]))             # <<<<<<<<<<<<<<
 *             sweep_metadata.append(_parse_ray_headers(
 *                 raw_sweep_data[i::self.ndata_types, :6]))
 */
      __pyx_t_14 = PySlice_New(__pyx_v_i, Py_None, __pyx_v_self->ndata_types); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 304, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 304, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_GIVEREF(__pyx_t_14);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_14)) __PYX_ERR(0, 304, __pyx_L1_error);
      __Pyx_INCREF(__pyx_int_4);
      __Pyx_GIVEREF(__pyx_int_4);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_4)) __PYX_ERR(0, 304, __pyx_L1_error);
      __pyx_t_14 = 0;
      __pyx_t_14 = __Pyx_PyObject_GetItem(__pyx_v_raw_sweep_data, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 304, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_15 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_15)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_15);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[4] = {__pyx_t_15, __pyx_v_data_type, __pyx_t_13, __pyx_t_14};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 302, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }

      /* "pyart/io/_sigmetfile.pyx":302
 *                 sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])
 *             else:
 *                 sweep_data.append(convert_sigmet_data(             # <<<<<<<<<<<<<<
 *                     data_type, raw_sweep_data[i::self.ndata_types, 6:],
 *                     raw_sweep_data[i::self.ndata_types, 4]))
 */
      __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_sweep_data, __pyx_t_4); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_L20:;

    /* "pyart/io/_sigmetfile.pyx":305
 *                     data_type, raw_sweep_data[i::self.ndata_types, 6:],
 *                     raw_sweep_data[i::self.ndata_types, 4]))
 *             sweep_metadata.append(_parse_ray_headers(             # <<<<<<<<<<<<<<
 *                 raw_sweep_data[i::self.ndata_types, :6]))
 *         return ingest_data_headers, sweep_data, sweep_metadata
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_parse_ray_headers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "pyart/io/_sigmetfile.pyx":306
 *                     raw_sweep_data[i::self.ndata_types, 4]))
 *             sweep_metadata.append(_parse_ray_headers(
 *                 raw_sweep_data[i::self.ndata_types, :6]))             # <<<<<<<<<<<<<<
 *         return ingest_data_headers, sweep_data, sweep_metadata
 * 
 */
    __pyx_t_14 = PySlice_New(__pyx_v_i, Py_None, __pyx_v_self->ndata_types); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14)) __PYX_ERR(0, 306, __pyx_L1_error);
    __Pyx_INCREF(__pyx_slice__5);
    __Pyx_GIVEREF(__pyx_slice__5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_slice__5)) __PYX_ERR(0, 306, __pyx_L1_error);
    __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_GetItem(__pyx_v_raw_sweep_data, __pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_t_14};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "pyart/io/_sigmetfile.pyx":305
 *                     data_type, raw_sweep_data[i::self.ndata_types, 6:],
 *                     raw_sweep_data[i::self.ndata_types, 4]))
 *             sweep_metadata.append(_parse_ray_headers(             # <<<<<<<<<<<<<<
 *                 raw_sweep_data[i::self.ndata_types, :6]))
 *         return ingest_data_headers, sweep_data, sweep_metadata
 */
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_sweep_metadata, __pyx_t_4); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 305, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyart/io/_sigmetfile.pyx":298
 *         sweep_data = []
 *         sweep_metadata = []
 *         for i, data_type in enumerate(self.data_types):             # <<<<<<<<<<<<<<
 *             if data_type == 0 and full_xhdr:
 *                 sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":307
 *             sweep_metadata.append(_parse_ray_headers(
 *                 raw_sweep_data[i::self.ndata_types, :6]))
 *         return ingest_data_headers, sweep_data, sweep_metadata             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_ingest_data_headers);
  __Pyx_GIVEREF(__pyx_v_ingest_data_headers);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_ingest_data_headers)) __PYX_ERR(0, 307, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_sweep_data);
  __Pyx_GIVEREF(__pyx_v_sweep_data);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_sweep_data)) __PYX_ERR(0, 307, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_sweep_metadata);
  __Pyx_GIVEREF(__pyx_v_sweep_metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_sweep_metadata)) __PYX_ERR(0, 307, __pyx_L1_error);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":224
 *         return data, metadata
 * 
 *     def _get_sweep(self, full_xhdr=False, raw_data=False):             # <<<<<<<<<<<<<<
 *         """
 *         Get the data and metadata from the next sweep.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._get_sweep", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_lead_record);
  __Pyx_XDECREF(__pyx_v_raw_prod_bhdr);
  __Pyx_XDECREF(__pyx_v_ingest_data_headers);
  __Pyx_XDECREF(__pyx_v_nray_data_types);
  __Pyx_XDECREF(__pyx_v_nrays);
  __Pyx_XDECREF(__pyx_v_nbins);
  __Pyx_XDECREF(__pyx_v_raw_sweep_data);
  __Pyx_XDECREF(__pyx_v_ray_i);
  __Pyx_XDECREF(__pyx_v_sweep_data);
  __Pyx_XDECREF(__pyx_v_sweep_metadata);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_data_type);
  __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_d);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":310
 * 
 *     @cython.wraparound(False)
 *     cdef int _get_ray(self, int nbins, np.ndarray[np.int16_t, ndim=1] out):             # <<<<<<<<<<<<<<
 *         """
 *         Get the next ray, loading new records as needed.
 */

static int __pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__get_ray(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, int __pyx_v_nbins, PyArrayObject *__pyx_v_out) {
  short __pyx_v_compression_code;
  int __pyx_v_words;
  int __pyx_v_remain;
  int __pyx_v_out_pos;
  int __pyx_v_first_end;
  int __pyx_v_i;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_out;
  __Pyx_Buffer __pyx_pybuffer_out;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  struct __pyx_opt_args_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos __pyx_t_7;
  short __pyx_t_8;
  short __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_pybuffer_out.pybuffer.buf = NULL;
  __pyx_pybuffer_out.refcount = 0;
  __pyx_pybuffernd_out.data = NULL;
  __pyx_pybuffernd_out.rcbuffer = &__pyx_pybuffer_out;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out.rcbuffer->pybuffer, (PyObject*)__pyx_v_out, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int16_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 310, __pyx_L1_error)
  }
  __pyx_pybuffernd_out.diminfo[0].strides = __pyx_pybuffernd_out.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_out.diminfo[0].shape = __pyx_pybuffernd_out.rcbuffer->pybuffer.shape[0];

  /* "pyart/io/_sigmetfile.pyx":330
 *         cdef int words, remain, out_pos, first_end, i
 * 
 *         if self._incr_rbuf_pos():             # <<<<<<<<<<<<<<
 *             return -1   # failed read
 *         compression_code = self._rbuf_p[self._rbuf_pos]
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self->__pyx_vtab)->_incr_rbuf_pos(__pyx_v_self, NULL); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 330, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyart/io/_sigmetfile.pyx":331
 * 
 *         if self._incr_rbuf_pos():
 *             return -1   # failed read             # <<<<<<<<<<<<<<
 *         compression_code = self._rbuf_p[self._rbuf_pos]
 *         out_pos = 0
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "pyart/io/_sigmetfile.pyx":330
 *         cdef int words, remain, out_pos, first_end, i
 * 
 *         if self._incr_rbuf_pos():             # <<<<<<<<<<<<<<
 *             return -1   # failed read
 *         compression_code = self._rbuf_p[self._rbuf_pos]
 */
  }

  /* "pyart/io/_sigmetfile.pyx":332
 *         if self._incr_rbuf_pos():
 *             return -1   # failed read
 *         compression_code = self._rbuf_p[self._rbuf_pos]             # <<<<<<<<<<<<<<
 *         out_pos = 0
 * 
 */
  __pyx_v_compression_code = (__pyx_v_self->_rbuf_p[__pyx_v_self->_rbuf_pos]);

  /* "pyart/io/_sigmetfile.pyx":333
 *             return -1   # failed read
 *         compression_code = self._rbuf_p[self._rbuf_pos]
 *         out_pos = 0             # <<<<<<<<<<<<<<
 * 
 *         if compression_code == 1:
 */
  __pyx_v_out_pos = 0;

  /* "pyart/io/_sigmetfile.pyx":335
 *         out_pos = 0
 * 
 *         if compression_code == 1:             # <<<<<<<<<<<<<<
 *             # mark ray as missing by setting numbers of bins to -1
 *             out[4] = -1
 */
  __pyx_t_2 = (__pyx_v_compression_code == 1);
  if (__pyx_t_2) {

    /* "pyart/io/_sigmetfile.pyx":337
 *         if compression_code == 1:
 *             # mark ray as missing by setting numbers of bins to -1
 *             out[4] = -1             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
    __pyx_t_3 = 4;
    __pyx_t_1 = -1;
    if (__pyx_t_3 < 0) {
      __pyx_t_1 = 0;
    } else if (unlikely(__pyx_t_3 >= __pyx_pybuffernd_out.diminfo[0].shape)) __pyx_t_1 = 0;
    if (unlikely(__pyx_t_1 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_1);
      __PYX_ERR(0, 337, __pyx_L1_error)
    }
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int16_t *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_out.diminfo[0].strides) = -1;

    /* "pyart/io/_sigmetfile.pyx":338
 *             # mark ray as missing by setting numbers of bins to -1
 *             out[4] = -1
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         while compression_code != 1:
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "pyart/io/_sigmetfile.pyx":335
 *         out_pos = 0
 * 
 *         if compression_code == 1:             # <<<<<<<<<<<<<<
 *             # mark ray as missing by setting numbers of bins to -1
 *             out[4] = -1
 */
  }

  /* "pyart/io/_sigmetfile.pyx":340
 *             return 0
 * 
 *         while compression_code != 1:             # <<<<<<<<<<<<<<
 * 
 *             if self._incr_rbuf_pos():
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_compression_code != 1);
    if (!__pyx_t_2) break;

    /* "pyart/io/_sigmetfile.pyx":342
 *         while compression_code != 1:
 * 
 *             if self._incr_rbuf_pos():             # <<<<<<<<<<<<<<
 *                 return -1   # failed read
 *             if compression_code < 0:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self->__pyx_vtab)->_incr_rbuf_pos(__pyx_v_self, NULL); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 342, __pyx_L1_error)
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {

      /* "pyart/io/_sigmetfile.pyx":343
 * 
 *             if self._incr_rbuf_pos():
 *                 return -1   # failed read             # <<<<<<<<<<<<<<
 *             if compression_code < 0:
 *                 words = compression_code + 32768    # last 7 bits give size
 */
      __pyx_r = -1;
      goto __pyx_L0;

      /* "pyart/io/_sigmetfile.pyx":342
 *         while compression_code != 1:
 * 
 *             if self._incr_rbuf_pos():             # <<<<<<<<<<<<<<
 *                 return -1   # failed read
 *             if compression_code < 0:
 */
    }

    /* "pyart/io/_sigmetfile.pyx":344
 *             if self._incr_rbuf_pos():
 *                 return -1   # failed read
 *             if compression_code < 0:             # <<<<<<<<<<<<<<
 *                 words = compression_code + 32768    # last 7 bits give size
 *                 if self._rbuf_pos + words <= 3072:
 */
    __pyx_t_2 = (__pyx_v_compression_code < 0);
    if (__pyx_t_2) {

      /* "pyart/io/_sigmetfile.pyx":345
 *                 return -1   # failed read
 *             if compression_code < 0:
 *                 words = compression_code + 32768    # last 7 bits give size             # <<<<<<<<<<<<<<
 *                 if self._rbuf_pos + words <= 3072:
 *                     # all compressed data is in the current record
 */
      __pyx_v_words = (__pyx_v_compression_code + 0x8000);

      /* "pyart/io/_sigmetfile.pyx":346
 *             if compression_code < 0:
 *                 words = compression_code + 32768    # last 7 bits give size
 *                 if self._rbuf_pos + words <= 3072:             # <<<<<<<<<<<<<<
 *                     # all compressed data is in the current record
 *                     for i in range(words):
 */
      __pyx_t_2 = ((__pyx_v_self->_rbuf_pos + __pyx_v_words) <= 0xC00);
      if (__pyx_t_2) {

        /* "pyart/io/_sigmetfile.pyx":348
 *                 if self._rbuf_pos + words <= 3072:
 *                     # all compressed data is in the current record
 *                     for i in range(words):             # <<<<<<<<<<<<<<
 *                         out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]
 *                     if self._incr_rbuf_pos(words):
 */
        __pyx_t_1 = __pyx_v_words;
        __pyx_t_4 = __pyx_t_1;
        for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
          __pyx_v_i = __pyx_t_5;

          /* "pyart/io/_sigmetfile.pyx":349
 *                     # all compressed data is in the current record
 *                     for i in range(words):
 *                         out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]             # <<<<<<<<<<<<<<
 *                     if self._incr_rbuf_pos(words):
 *                         return -1   # failed read
 */
          __pyx_t_3 = (__pyx_v_out_pos + __pyx_v_i);
          __pyx_t_6 = -1;
          if (__pyx_t_3 < 0) {
            __pyx_t_6 = 0;
          } else if (unlikely(__pyx_t_3 >= __pyx_pybuffernd_out.diminfo[0].shape)) __pyx_t_6 = 0;
          if (unlikely(__pyx_t_6 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_6);
            __PYX_ERR(0, 349, __pyx_L1_error)
          }
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int16_t *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_out.diminfo[0].strides) = (__pyx_v_self->_rbuf_p[(__pyx_v_self->_rbuf_pos + __pyx_v_i)]);
        }

        /* "pyart/io/_sigmetfile.pyx":350
 *                     for i in range(words):
 *                         out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]
 *                     if self._incr_rbuf_pos(words):             # <<<<<<<<<<<<<<
 *                         return -1   # failed read
 *                     out_pos += words
 */
        __pyx_t_7.__pyx_n = 1;
        __pyx_t_7.incr = __pyx_v_words;
        __pyx_t_1 = ((struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self->__pyx_vtab)->_incr_rbuf_pos(__pyx_v_self, &__pyx_t_7); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 350, __pyx_L1_error)
        __pyx_t_2 = (__pyx_t_1 != 0);
        if (__pyx_t_2) {

          /* "pyart/io/_sigmetfile.pyx":351
 *                         out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]
 *                     if self._incr_rbuf_pos(words):
 *                         return -1   # failed read             # <<<<<<<<<<<<<<
 *                     out_pos += words
 *                 else:
 */
          __pyx_r = -1;
          goto __pyx_L0;

          /* "pyart/io/_sigmetfile.pyx":350
 *                     for i in range(words):
 *                         out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]
 *                     if self._incr_rbuf_pos(words):             # <<<<<<<<<<<<<<
 *                         return -1   # failed read
 *                     out_pos += words
 */
        }

        /* "pyart/io/_sigmetfile.pyx":352
 *                     if self._incr_rbuf_pos(words):
 *                         return -1   # failed read
 *                     out_pos += words             # <<<<<<<<<<<<<<
 *                 else:
 *                     # data is split between current and next record
 */
        __pyx_v_out_pos = (__pyx_v_out_pos + __pyx_v_words);

        /* "pyart/io/_sigmetfile.pyx":346
 *             if compression_code < 0:
 *                 words = compression_code + 32768    # last 7 bits give size
 *                 if self._rbuf_pos + words <= 3072:             # <<<<<<<<<<<<<<
 *                     # all compressed data is in the current record
 *                     for i in range(words):
 */
        goto __pyx_L9;
      }

      /* "pyart/io/_sigmetfile.pyx":356
 *                     # data is split between current and next record
 *                     # store data from current record
 *                     remain = words - (3072 - self._rbuf_pos)             # <<<<<<<<<<<<<<
 *                     first_end = out_pos + words - remain
 *                     for i in range(first_end - out_pos):
 */
      /*else*/ {
        __pyx_v_remain = (__pyx_v_words - (0xC00 - __pyx_v_self->_rbuf_pos));

        /* "pyart/io/_sigmetfile.pyx":357
 *                     # store data from current record
 *                     remain = words - (3072 - self._rbuf_pos)
 *                     first_end = out_pos + words - remain             # <<<<<<<<<<<<<<
 *                     for i in range(first_end - out_pos):
 *                         out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]
 */
        __pyx_v_first_end = ((__pyx_v_out_pos + __pyx_v_words) - __pyx_v_remain);

        /* "pyart/io/_sigmetfile.pyx":358
 *                     remain = words - (3072 - self._rbuf_pos)
 *                     first_end = out_pos + words - remain
 *                     for i in range(first_end - out_pos):             # <<<<<<<<<<<<<<
 *                         out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]
 * 
 */
        __pyx_t_1 = (__pyx_v_first_end - __pyx_v_out_pos);
        __pyx_t_4 = __pyx_t_1;
        for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
          __pyx_v_i = __pyx_t_5;

          /* "pyart/io/_sigmetfile.pyx":359
 *                     first_end = out_pos + words - remain
 *                     for i in range(first_end - out_pos):
 *                         out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]             # <<<<<<<<<<<<<<
 * 
 *                     # read data from next record and store
 */
          __pyx_t_3 = (__pyx_v_out_pos + __pyx_v_i);
          __pyx_t_6 = -1;
          if (__pyx_t_3 < 0) {
            __pyx_t_6 = 0;
          } else if (unlikely(__pyx_t_3 >= __pyx_pybuffernd_out.diminfo[0].shape)) __pyx_t_6 = 0;
          if (unlikely(__pyx_t_6 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_6);
            __PYX_ERR(0, 359, __pyx_L1_error)
          }
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int16_t *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_out.diminfo[0].strides) = (__pyx_v_self->_rbuf_p[(__pyx_v_self->_rbuf_pos + __pyx_v_i)]);
        }

        /* "pyart/io/_sigmetfile.pyx":362
 * 
 *                     # read data from next record and store
 *                     self._load_record()             # <<<<<<<<<<<<<<
 *                     for i in range(out_pos + words - first_end):
 *                         out[first_end + i] = self._rbuf_p[self._rbuf_pos + i]
 */
        ((struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self->__pyx_vtab)->_load_record(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 362, __pyx_L1_error)

        /* "pyart/io/_sigmetfile.pyx":363
 *                     # read data from next record and store
 *                     self._load_record()
 *                     for i in range(out_pos + words - first_end):             # <<<<<<<<<<<<<<
 *                         out[first_end + i] = self._rbuf_p[self._rbuf_pos + i]
 * 
 */
        __pyx_t_1 = ((__pyx_v_out_pos + __pyx_v_words) - __pyx_v_first_end);
        __pyx_t_4 = __pyx_t_1;
        for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
          __pyx_v_i = __pyx_t_5;

          /* "pyart/io/_sigmetfile.pyx":364
 *                     self._load_record()
 *                     for i in range(out_pos + words - first_end):
 *                         out[first_end + i] = self._rbuf_p[self._rbuf_pos + i]             # <<<<<<<<<<<<<<
 * 
 *                     if self._incr_rbuf_pos(remain):
 */
          __pyx_t_3 = (__pyx_v_first_end + __pyx_v_i);
          __pyx_t_6 = -1;
          if (__pyx_t_3 < 0) {
            __pyx_t_6 = 0;
          } else if (unlikely(__pyx_t_3 >= __pyx_pybuffernd_out.diminfo[0].shape)) __pyx_t_6 = 0;
          if (unlikely(__pyx_t_6 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_6);
            __PYX_ERR(0, 364, __pyx_L1_error)
          }
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int16_t *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_out.diminfo[0].strides) = (__pyx_v_self->_rbuf_p[(__pyx_v_self->_rbuf_pos + __pyx_v_i)]);
        }

        /* "pyart/io/_sigmetfile.pyx":366
 *                         out[first_end + i] = self._rbuf_p[self._rbuf_pos + i]
 * 
 *                     if self._incr_rbuf_pos(remain):             # <<<<<<<<<<<<<<
 *                         return -1   # failed read
 *                     out_pos += words
 */
        __pyx_t_7.__pyx_n = 1;
        __pyx_t_7.incr = __pyx_v_remain;
        __pyx_t_1 = ((struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self->__pyx_vtab)->_incr_rbuf_pos(__pyx_v_self, &__pyx_t_7); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 366, __pyx_L1_error)
        __pyx_t_2 = (__pyx_t_1 != 0);
        if (__pyx_t_2) {

          /* "pyart/io/_sigmetfile.pyx":367
 * 
 *                     if self._incr_rbuf_pos(remain):
 *                         return -1   # failed read             # <<<<<<<<<<<<<<
 *                     out_pos += words
 *             else:
 */
          __pyx_r = -1;
          goto __pyx_L0;

          /* "pyart/io/_sigmetfile.pyx":366
 *                         out[first_end + i] = self._rbuf_p[self._rbuf_pos + i]
 * 
 *                     if self._incr_rbuf_pos(remain):             # <<<<<<<<<<<<<<
 *                         return -1   # failed read
 *                     out_pos += words
 */
        }

        /* "pyart/io/_sigmetfile.pyx":368
 *                     if self._incr_rbuf_pos(remain):
 *                         return -1   # failed read
 *                     out_pos += words             # <<<<<<<<<<<<<<
 *             else:
 *                 # add zeros to out
 */
        __pyx_v_out_pos = (__pyx_v_out_pos + __pyx_v_words);
      }
      __pyx_L9:;

      /* "pyart/io/_sigmetfile.pyx":344
 *             if self._incr_rbuf_pos():
 *                 return -1   # failed read
 *             if compression_code < 0:             # <<<<<<<<<<<<<<
 *                 words = compression_code + 32768    # last 7 bits give size
 *                 if self._rbuf_pos + words <= 3072:
 */
      goto __pyx_L8;
    }

    /* "pyart/io/_sigmetfile.pyx":371
 *             else:
 *                 # add zeros to out
 *                 if compression_code + out_pos > nbins + 6:             # <<<<<<<<<<<<<<
 *                     return -1   # file is corrupt
 *                 for i in range(compression_code):
 */
    /*else*/ {
      __pyx_t_2 = ((__pyx_v_compression_code + __pyx_v_out_pos) > (__pyx_v_nbins + 6));
      if (__pyx_t_2) {

        /* "pyart/io/_sigmetfile.pyx":372
 *                 # add zeros to out
 *                 if compression_code + out_pos > nbins + 6:
 *                     return -1   # file is corrupt             # <<<<<<<<<<<<<<
 *                 for i in range(compression_code):
 *                     out[out_pos + i] = 0
 */
        __pyx_r = -1;
        goto __pyx_L0;

        /* "pyart/io/_sigmetfile.pyx":371
 *             else:
 *                 # add zeros to out
 *                 if compression_code + out_pos > nbins + 6:             # <<<<<<<<<<<<<<
 *                     return -1   # file is corrupt
 *                 for i in range(compression_code):
 */
      }

      /* "pyart/io/_sigmetfile.pyx":373
 *                 if compression_code + out_pos > nbins + 6:
 *                     return -1   # file is corrupt
 *                 for i in range(compression_code):             # <<<<<<<<<<<<<<
 *                     out[out_pos + i] = 0
 *                 out_pos += compression_code
 */
      __pyx_t_8 = __pyx_v_compression_code;
      __pyx_t_9 = __pyx_t_8;
      for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_9; __pyx_t_1+=1) {
        __pyx_v_i = __pyx_t_1;

        /* "pyart/io/_sigmetfile.pyx":374
 *                     return -1   # file is corrupt
 *                 for i in range(compression_code):
 *                     out[out_pos + i] = 0             # <<<<<<<<<<<<<<
 *                 out_pos += compression_code
 *             compression_code = self._rbuf_p[self._rbuf_pos]
 */
        __pyx_t_3 = (__pyx_v_out_pos + __pyx_v_i);
        __pyx_t_4 = -1;
        if (__pyx_t_3 < 0) {
          __pyx_t_4 = 0;
        } else if (unlikely(__pyx_t_3 >= __pyx_pybuffernd_out.diminfo[0].shape)) __pyx_t_4 = 0;
        if (unlikely(__pyx_t_4 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_4);
          __PYX_ERR(0, 374, __pyx_L1_error)
        }
        *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int16_t *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_out.diminfo[0].strides) = 0;
      }

      /* "pyart/io/_sigmetfile.pyx":375
 *                 for i in range(compression_code):
 *                     out[out_pos + i] = 0
 *                 out_pos += compression_code             # <<<<<<<<<<<<<<
 *             compression_code = self._rbuf_p[self._rbuf_pos]
 * 
 */
      __pyx_v_out_pos = (__pyx_v_out_pos + __pyx_v_compression_code);
    }
    __pyx_L8:;

    /* "pyart/io/_sigmetfile.pyx":376
 *                     out[out_pos + i] = 0
 *                 out_pos += compression_code
 *             compression_code = self._rbuf_p[self._rbuf_pos]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */
    __pyx_v_compression_code = (__pyx_v_self->_rbuf_p[__pyx_v_self->_rbuf_pos]);
  }

  /* "pyart/io/_sigmetfile.pyx":378
 *             compression_code = self._rbuf_p[self._rbuf_pos]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _incr_rbuf_pos(self, int incr=1):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":310
 * 
 *     @cython.wraparound(False)
 *     cdef int _get_ray(self, int nbins, np.ndarray[np.int16_t, ndim=1] out):             # <<<<<<<<<<<<<<
 *         """
 *         Get the next ray, loading new records as needed.
 */

  /* function exit code */
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._get_ray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
  __pyx_L2:;
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":380
 *         return 0
 * 
 *     cdef int _incr_rbuf_pos(self, int incr=1):             # <<<<<<<<<<<<<<
 *         """
 *         Increment the record buffer position, load a new record if needed.
 */

static int __pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, struct __pyx_opt_args_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos *__pyx_optional_args) {
  int __pyx_v_incr = ((int)1);
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_incr = __pyx_optional_args->incr;
    }
  }

  /* "pyart/io/_sigmetfile.pyx":384
 *         Increment the record buffer position, load a new record if needed.
 *         """
 *         self._rbuf_pos += incr             # <<<<<<<<<<<<<<
 *         if self._rbuf_pos >= 3072:
 *             if self._load_record():
 */
  __pyx_v_self->_rbuf_pos = (__pyx_v_self->_rbuf_pos + __pyx_v_incr);

  /* "pyart/io/_sigmetfile.pyx":385
 *         """
 *         self._rbuf_pos += incr
 *         if self._rbuf_pos >= 3072:             # <<<<<<<<<<<<<<
 *             if self._load_record():
 *                 return -1   # failed read
 */
  __pyx_t_1 = (__pyx_v_self->_rbuf_pos >= 0xC00);
  if (__pyx_t_1) {

    /* "pyart/io/_sigmetfile.pyx":386
 *         self._rbuf_pos += incr
 *         if self._rbuf_pos >= 3072:
 *             if self._load_record():             # <<<<<<<<<<<<<<
 *                 return -1   # failed read
 *         return 0
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self->__pyx_vtab)->_load_record(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 386, __pyx_L1_error)
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "pyart/io/_sigmetfile.pyx":387
 *         if self._rbuf_pos >= 3072:
 *             if self._load_record():
 *                 return -1   # failed read             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
      __pyx_r = -1;
      goto __pyx_L0;

      /* "pyart/io/_sigmetfile.pyx":386
 *         self._rbuf_pos += incr
 *         if self._rbuf_pos >= 3072:
 *             if self._load_record():             # <<<<<<<<<<<<<<
 *                 return -1   # failed read
 *         return 0
 */
    }

    /* "pyart/io/_sigmetfile.pyx":385
 *         """
 *         self._rbuf_pos += incr
 *         if self._rbuf_pos >= 3072:             # <<<<<<<<<<<<<<
 *             if self._load_record():
 *                 return -1   # failed read
 */
  }

  /* "pyart/io/_sigmetfile.pyx":388
 *             if self._load_record():
 *                 return -1   # failed read
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _load_record(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":380
 *         return 0
 * 
 *     cdef int _incr_rbuf_pos(self, int incr=1):             # <<<<<<<<<<<<<<
 *         """
 *         Increment the record buffer position, load a new record if needed.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._incr_rbuf_pos", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":390
 *         return 0
 * 
 *     cdef int _load_record(self):             # <<<<<<<<<<<<<<
 *         """ Load the next record. returns -1 on fail, 0 if success. """
 *         record = self._fh.read(RECORD_SIZE)
 */

static int __pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__load_record(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_v_record = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_load_record", 1);

  /* "pyart/io/_sigmetfile.pyx":392
 *     cdef int _load_record(self):
 *         """ Load the next record. returns -1 on fail, 0 if success. """
 *         record = self._fh.read(RECORD_SIZE)             # <<<<<<<<<<<<<<
 *         if len(record) != RECORD_SIZE:
 *             return -1   # failed read
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_fh, __pyx_n_s_read); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_RECORD_SIZE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_record = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":393
 *         """ Load the next record. returns -1 on fail, 0 if success. """
 *         record = self._fh.read(RECORD_SIZE)
 *         if len(record) != RECORD_SIZE:             # <<<<<<<<<<<<<<
 *             return -1   # failed read
 *         self._record_number += 1
 */
  __pyx_t_6 = PyObject_Length(__pyx_v_record); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 393, __pyx_L1_error)
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_RECORD_SIZE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_7) {

    /* "pyart/io/_sigmetfile.pyx":394
 *         record = self._fh.read(RECORD_SIZE)
 *         if len(record) != RECORD_SIZE:
 *             return -1   # failed read             # <<<<<<<<<<<<<<
 *         self._record_number += 1
 *         if self.debug:
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "pyart/io/_sigmetfile.pyx":393
 *         """ Load the next record. returns -1 on fail, 0 if success. """
 *         record = self._fh.read(RECORD_SIZE)
 *         if len(record) != RECORD_SIZE:             # <<<<<<<<<<<<<<
 *             return -1   # failed read
 *         self._record_number += 1
 */
  }

  /* "pyart/io/_sigmetfile.pyx":395
 *         if len(record) != RECORD_SIZE:
 *             return -1   # failed read
 *         self._record_number += 1             # <<<<<<<<<<<<<<
 *         if self.debug:
 *             print("Finished loading record:", self._record_number)
 */
  __pyx_v_self->_record_number = (__pyx_v_self->_record_number + 1);

  /* "pyart/io/_sigmetfile.pyx":396
 *             return -1   # failed read
 *         self._record_number += 1
 *         if self.debug:             # <<<<<<<<<<<<<<
 *             print("Finished loading record:", self._record_number)
 *         self._raw_product_bhdrs[-1].append(_unpack_raw_prod_bhdr(record))
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_self->debug); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 396, __pyx_L1_error)
  if (__pyx_t_7) {

    /* "pyart/io/_sigmetfile.pyx":397
 *         self._record_number += 1
 *         if self.debug:
 *             print("Finished loading record:", self._record_number)             # <<<<<<<<<<<<<<
 *         self._raw_product_bhdrs[-1].append(_unpack_raw_prod_bhdr(record))
 *         self._rbuf = np.frombuffer(record, dtype='int16')
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->_record_number); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_kp_u_Finished_loading_record);
    __Pyx_GIVEREF(__pyx_kp_u_Finished_loading_record);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_Finished_loading_record)) __PYX_ERR(0, 397, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3)) __PYX_ERR(0, 397, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyart/io/_sigmetfile.pyx":396
 *             return -1   # failed read
 *         self._record_number += 1
 *         if self.debug:             # <<<<<<<<<<<<<<
 *             print("Finished loading record:", self._record_number)
 *         self._raw_product_bhdrs[-1].append(_unpack_raw_prod_bhdr(record))
 */
  }

  /* "pyart/io/_sigmetfile.pyx":398
 *         if self.debug:
 *             print("Finished loading record:", self._record_number)
 *         self._raw_product_bhdrs[-1].append(_unpack_raw_prod_bhdr(record))             # <<<<<<<<<<<<<<
 *         self._rbuf = np.frombuffer(record, dtype='int16')
 *         self._rbuf_pos = 6
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_self->_raw_product_bhdrs, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_unpack_raw_prod_bhdr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_record};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_t_8 = __Pyx_PyObject_Append(__pyx_t_3, __pyx_t_2); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":399
 *             print("Finished loading record:", self._record_number)
 *         self._raw_product_bhdrs[-1].append(_unpack_raw_prod_bhdr(record))
 *         self._rbuf = np.frombuffer(record, dtype='int16')             # <<<<<<<<<<<<<<
 *         self._rbuf_pos = 6
 *         self._rbuf_p = <np.int16_t*>self._rbuf.data
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_frombuffer); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_record);
  __Pyx_GIVEREF(__pyx_v_record);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_record)) __PYX_ERR(0, 399, __pyx_L1_error);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_u_int16) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_rbuf);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_rbuf);
  __pyx_v_self->_rbuf = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "pyart/io/_sigmetfile.pyx":400
 *         self._raw_product_bhdrs[-1].append(_unpack_raw_prod_bhdr(record))
 *         self._rbuf = np.frombuffer(record, dtype='int16')
 *         self._rbuf_pos = 6             # <<<<<<<<<<<<<<
 *         self._rbuf_p = <np.int16_t*>self._rbuf.data
 *         return 0
 */
  __pyx_v_self->_rbuf_pos = 6;

  /* "pyart/io/_sigmetfile.pyx":401
 *         self._rbuf = np.frombuffer(record, dtype='int16')
 *         self._rbuf_pos = 6
 *         self._rbuf_p = <np.int16_t*>self._rbuf.data             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_4 = ((PyObject *)__pyx_v_self->_rbuf);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_self->_rbuf_p = ((__pyx_t_5numpy_int16_t *)__pyx_f_5numpy_7ndarray_4data_data(((PyArrayObject *)__pyx_t_4)));
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyart/io/_sigmetfile.pyx":402
 *         self._rbuf_pos = 6
 *         self._rbuf_p = <np.int16_t*>self._rbuf.data
 *         return 0             # <<<<<<<<<<<<<<
 * 
 * # functions used by the SigmetFile class
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":390
 *         return 0
 * 
 *     cdef int _load_record(self):             # <<<<<<<<<<<<<<
 *         """ Load the next record. returns -1 on fail, 0 if success. """
 *         record = self._fh.read(RECORD_SIZE)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._load_record", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_record);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":50
 * 
 *     """
 *     cdef public debug, product_hdr, ingest_header, ingest_data_headers, \             # <<<<<<<<<<<<<<
 *         data_types, data_type_names, ndata_types,
 *     cdef public _fh, _raw_product_bhdrs
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->debug);
  __pyx_r = __pyx_v_self->debug;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->debug);
  __Pyx_DECREF(__pyx_v_self->debug);
  __pyx_v_self->debug = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->debug);
  __Pyx_DECREF(__pyx_v_self->debug);
  __pyx_v_self->debug = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->product_hdr);
  __pyx_r = __pyx_v_self->product_hdr;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->product_hdr);
  __Pyx_DECREF(__pyx_v_self->product_hdr);
  __pyx_v_self->product_hdr = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->product_hdr);
  __Pyx_DECREF(__pyx_v_self->product_hdr);
  __pyx_v_self->product_hdr = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->ingest_header);
  __pyx_r = __pyx_v_self->ingest_header;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->ingest_header);
  __Pyx_DECREF(__pyx_v_self->ingest_header);
  __pyx_v_self->ingest_header = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->ingest_header);
  __Pyx_DECREF(__pyx_v_self->ingest_header);
  __pyx_v_self->ingest_header = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->ingest_data_headers);
  __pyx_r = __pyx_v_self->ingest_data_headers;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->ingest_data_headers);
  __Pyx_DECREF(__pyx_v_self->ingest_data_headers);
  __pyx_v_self->ingest_data_headers = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->ingest_data_headers);
  __Pyx_DECREF(__pyx_v_self->ingest_data_headers);
  __pyx_v_self->ingest_data_headers = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":51
 *     """
 *     cdef public debug, product_hdr, ingest_header, ingest_data_headers, \
 *         data_types, data_type_names, ndata_types,             # <<<<<<<<<<<<<<
 *     cdef public _fh, _raw_product_bhdrs
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->data_types);
  __pyx_r = __pyx_v_self->data_types;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->data_types);
  __Pyx_DECREF(__pyx_v_self->data_types);
  __pyx_v_self->data_types = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->data_types);
  __Pyx_DECREF(__pyx_v_self->data_types);
  __pyx_v_self->data_types = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->data_type_names);
  __pyx_r = __pyx_v_self->data_type_names;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->data_type_names);
  __Pyx_DECREF(__pyx_v_self->data_type_names);
  __pyx_v_self->data_type_names = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->data_type_names);
  __Pyx_DECREF(__pyx_v_self->data_type_names);
  __pyx_v_self->data_type_names = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->ndata_types);
  __pyx_r = __pyx_v_self->ndata_types;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->ndata_types);
  __Pyx_DECREF(__pyx_v_self->ndata_types);
  __pyx_v_self->ndata_types = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->ndata_types);
  __Pyx_DECREF(__pyx_v_self->ndata_types);
  __pyx_v_self->ndata_types = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":52
 *     cdef public debug, product_hdr, ingest_header, ingest_data_headers, \
 *         data_types, data_type_names, ndata_types,
 *     cdef public _fh, _raw_product_bhdrs             # <<<<<<<<<<<<<<
 * 
 *     cdef np.ndarray _rbuf
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_fh);
  __pyx_r = __pyx_v_self->_fh;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_fh);
  __Pyx_DECREF(__pyx_v_self->_fh);
  __pyx_v_self->_fh = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_fh);
  __Pyx_DECREF(__pyx_v_self->_fh);
  __pyx_v_self->_fh = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_raw_product_bhdrs);
  __pyx_r = __pyx_v_self->_raw_product_bhdrs;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_raw_product_bhdrs);
  __Pyx_DECREF(__pyx_v_self->_raw_product_bhdrs);
  __pyx_v_self->_raw_product_bhdrs = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_raw_product_bhdrs);
  __Pyx_DECREF(__pyx_v_self->_raw_product_bhdrs);
  __pyx_v_self->_raw_product_bhdrs = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":56
 *     cdef np.ndarray _rbuf
 *     cdef np.int16_t * _rbuf_p   # hack for fast indexing of _rbuf
 *     cdef public int _rbuf_pos, _record_number             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, filename, debug=False):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_rbuf_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._rbuf_pos.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 56, __pyx_L1_error)
  __pyx_v_self->_rbuf_pos = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._rbuf_pos.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_record_number); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._record_number.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 56, __pyx_L1_error)
  __pyx_v_self->_record_number = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._record_number.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_10SigmetFile_11__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10__reduce_cython__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self__rbuf_p_cannot_be_converted, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_10SigmetFile_13__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_12__setstate_cython__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self__rbuf_p_cannot_be_converted, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":407
 * 
 * 
 * def _data_types_from_mask(word0, word1, word2, word3):             # <<<<<<<<<<<<<<
 *     """
 *     Return a list of the data types from the words in the data_type mask.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_1_data_types_from_mask(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile__data_types_from_mask, "\n    Return a list of the data types from the words in the data_type mask.\n    ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_1_data_types_from_mask = {"_data_types_from_mask", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_1_data_types_from_mask, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile__data_types_from_mask};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_1_data_types_from_mask(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_word0 = 0;
  PyObject *__pyx_v_word1 = 0;
  PyObject *__pyx_v_word2 = 0;
  PyObject *__pyx_v_word3 = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_data_types_from_mask (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_word0,&__pyx_n_s_word1,&__pyx_n_s_word2,&__pyx_n_s_word3,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_word0)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 407, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_word1)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 407, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_data_types_from_mask", 1, 4, 4, 1); __PYX_ERR(0, 407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_word2)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 407, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_data_types_from_mask", 1, 4, 4, 2); __PYX_ERR(0, 407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_word3)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 407, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_data_types_from_mask", 1, 4, 4, 3); __PYX_ERR(0, 407, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_data_types_from_mask") < 0)) __PYX_ERR(0, 407, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
      values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
    }
    __pyx_v_word0 = values[0];
    __pyx_v_word1 = values[1];
    __pyx_v_word2 = values[2];
    __pyx_v_word3 = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_data_types_from_mask", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 407, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile._data_types_from_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile__data_types_from_mask(__pyx_self, __pyx_v_word0, __pyx_v_word1, __pyx_v_word2, __pyx_v_word3);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile__data_types_from_mask(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_word0, PyObject *__pyx_v_word1, PyObject *__pyx_v_word2, PyObject *__pyx_v_word3) {
  PyObject *__pyx_v_data_types = NULL;
  long __pyx_8genexpr4__pyx_v_i;
  long __pyx_8genexpr5__pyx_v_i;
  long __pyx_8genexpr6__pyx_v_i;
  long __pyx_8genexpr7__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_data_types_from_mask", 1);

  /* "pyart/io/_sigmetfile.pyx":411
 *     Return a list of the data types from the words in the data_type mask.
 *     """
 *     data_types = [i for i in range(32) if _is_bit_set(word0, i)]             # <<<<<<<<<<<<<<
 *     data_types += [i+32 for i in range(32) if _is_bit_set(word1, i)]
 *     data_types += [i+64 for i in range(32) if _is_bit_set(word2, i)]
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    for (__pyx_t_2 = 0; __pyx_t_2 < 32; __pyx_t_2+=1) {
      __pyx_8genexpr4__pyx_v_i = __pyx_t_2;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_is_bit_set); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 411, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyInt_From_long(__pyx_8genexpr4__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 411, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_word0, __pyx_t_5};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 411, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 411, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_8) {
        __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_8genexpr4__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 411, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 411, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
    }
  } /* exit inner scope */
  __pyx_v_data_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":412
 *     """
 *     data_types = [i for i in range(32) if _is_bit_set(word0, i)]
 *     data_types += [i+32 for i in range(32) if _is_bit_set(word1, i)]             # <<<<<<<<<<<<<<
 *     data_types += [i+64 for i in range(32) if _is_bit_set(word2, i)]
 *     data_types += [i+96 for i in range(32) if _is_bit_set(word3, i)]
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 412, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    for (__pyx_t_2 = 0; __pyx_t_2 < 32; __pyx_t_2+=1) {
      __pyx_8genexpr5__pyx_v_i = __pyx_t_2;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_is_bit_set); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyInt_From_long(__pyx_8genexpr5__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_word1, __pyx_t_5};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 412, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_8) {
        __pyx_t_3 = __Pyx_PyInt_From_long((__pyx_8genexpr5__pyx_v_i + 32)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 412, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 412, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
    }
  } /* exit inner scope */
  __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_data_types, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_data_types, ((PyObject*)__pyx_t_3));
  __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":413
 *     data_types = [i for i in range(32) if _is_bit_set(word0, i)]
 *     data_types += [i+32 for i in range(32) if _is_bit_set(word1, i)]
 *     data_types += [i+64 for i in range(32) if _is_bit_set(word2, i)]             # <<<<<<<<<<<<<<
 *     data_types += [i+96 for i in range(32) if _is_bit_set(word3, i)]
 *     return data_types
 */
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    for (__pyx_t_2 = 0; __pyx_t_2 < 32; __pyx_t_2+=1) {
      __pyx_8genexpr6__pyx_v_i = __pyx_t_2;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_is_bit_set); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyInt_From_long(__pyx_8genexpr6__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_word2, __pyx_t_5};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_8) {
        __pyx_t_1 = __Pyx_PyInt_From_long((__pyx_8genexpr6__pyx_v_i + 64)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 413, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
    }
  } /* exit inner scope */
  __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_data_types, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_data_types, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":414
 *     data_types += [i+32 for i in range(32) if _is_bit_set(word1, i)]
 *     data_types += [i+64 for i in range(32) if _is_bit_set(word2, i)]
 *     data_types += [i+96 for i in range(32) if _is_bit_set(word3, i)]             # <<<<<<<<<<<<<<
 *     return data_types
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    for (__pyx_t_2 = 0; __pyx_t_2 < 32; __pyx_t_2+=1) {
      __pyx_8genexpr7__pyx_v_i = __pyx_t_2;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_is_bit_set); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 414, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyInt_From_long(__pyx_8genexpr7__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 414, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_word3, __pyx_t_5};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 414, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_8) {
        __pyx_t_3 = __Pyx_PyInt_From_long((__pyx_8genexpr7__pyx_v_i + 96)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 414, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
    }
  } /* exit inner scope */
  __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_data_types, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_data_types, ((PyObject*)__pyx_t_3));
  __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":415
 *     data_types += [i+64 for i in range(32) if _is_bit_set(word2, i)]
 *     data_types += [i+96 for i in range(32) if _is_bit_set(word3, i)]
 *     return data_types             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_data_types);
  __pyx_r = __pyx_v_data_types;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":407
 * 
 * 
 * def _data_types_from_mask(word0, word1, word2, word3):             # <<<<<<<<<<<<<<
 *     """
 *     Return a list of the data types from the words in the data_type mask.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyart.io._sigmetfile._data_types_from_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data_types);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":418
 * 
 * 
 * def _is_bit_set(number, bit):             # <<<<<<<<<<<<<<
 *     """ Return True if bit is set in number. """
 *     return number >> bit & 1 == 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_3_is_bit_set(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_2_is_bit_set, " Return True if bit is set in number. ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_3_is_bit_set = {"_is_bit_set", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_3_is_bit_set, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_2_is_bit_set};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_3_is_bit_set(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_number = 0;
  PyObject *__pyx_v_bit = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_is_bit_set (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_number,&__pyx_n_s_bit,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_number)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 418, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_bit)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 418, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_is_bit_set", 1, 2, 2, 1); __PYX_ERR(0, 418, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_is_bit_set") < 0)) __PYX_ERR(0, 418, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_number = values[0];
    __pyx_v_bit = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_is_bit_set", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 418, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile._is_bit_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_2_is_bit_set(__pyx_self, __pyx_v_number, __pyx_v_bit);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_2_is_bit_set(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_number, PyObject *__pyx_v_bit) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_is_bit_set", 1);

  /* "pyart/io/_sigmetfile.pyx":420
 * def _is_bit_set(number, bit):
 *     """ Return True if bit is set in number. """
 *     return number >> bit & 1 == 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Rshift(__pyx_v_number, __pyx_v_bit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_AndObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":418
 * 
 * 
 * def _is_bit_set(number, bit):             # <<<<<<<<<<<<<<
 *     """ Return True if bit is set in number. """
 *     return number >> bit & 1 == 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyart.io._sigmetfile._is_bit_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":423
 * 
 * 
 * def _parse_ray_headers(ray_headers):             # <<<<<<<<<<<<<<
 *     """
 *     Parse the metadata from Sigmet ray headers.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_5_parse_ray_headers(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_4_parse_ray_headers, "\n    Parse the metadata from Sigmet ray headers.\n\n    Parameters\n    ----------\n    ray_headers : array, shape=(..., 6)\n        Ray headers to parse.\n\n    Returns\n    -------\n    az0 : array\n        Azimuth angles (in degrees) at beginning of the rays.\n    el0 : array\n        Elevation angles at the beginning of the rays.\n    az1 : array\n        Azimuth angles at the end of the rays.\n    el1 : array\n        Elevation angles at the end of the rays.\n    nbins : array\n        Number of bins in the rays.\n    time : array\n        Seconds since the start of the sweep for the rays.\n    prf_flag : array\n        Numerical indication of what PRF was used, 0 for high, 1 for low.\n        Not applicable if dual-PRF is not used during collection.\n\n    ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_5_parse_ray_headers = {"_parse_ray_headers", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_5_parse_ray_headers, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_4_parse_ray_headers};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_5_parse_ray_headers(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_ray_headers = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_parse_ray_headers (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ray_headers,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_ray_headers)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 423, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_parse_ray_headers") < 0)) __PYX_ERR(0, 423, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_ray_headers = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_parse_ray_headers", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 423, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile._parse_ray_headers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_4_parse_ray_headers(__pyx_self, __pyx_v_ray_headers);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_4_parse_ray_headers(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ray_headers) {
  PyObject *__pyx_v_az0 = NULL;
  PyObject *__pyx_v_el0 = NULL;
  PyObject *__pyx_v_az1 = NULL;
  PyObject *__pyx_v_el1 = NULL;
  PyObject *__pyx_v_nbins = NULL;
  PyObject *__pyx_v_time = NULL;
  PyObject *__pyx_v_prf_flag = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_parse_ray_headers", 1);

  /* "pyart/io/_sigmetfile.pyx":451
 * 
 *     """
 *     az0 = bin2_to_angle(ray_headers.view('uint16')[..., 0])             # <<<<<<<<<<<<<<
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_bin2_to_angle); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_ray_headers, __pyx_n_s_view); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_n_u_uint16};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_tuple__6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_az0 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":452
 *     """
 *     az0 = bin2_to_angle(ray_headers.view('uint16')[..., 0])
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])             # <<<<<<<<<<<<<<
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_bin2_to_angle); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_ray_headers, __pyx_n_s_view); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_n_u_uint16};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 452, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_4, __pyx_tuple__7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 452, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_el0 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":453
 *     az0 = bin2_to_angle(ray_headers.view('uint16')[..., 0])
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])             # <<<<<<<<<<<<<<
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 *     nbins = ray_headers.view('int16')[..., 4]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_bin2_to_angle); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_ray_headers, __pyx_n_s_view); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_n_u_uint16};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_tuple__8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_az1 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":454
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])             # <<<<<<<<<<<<<<
 *     nbins = ray_headers.view('int16')[..., 4]
 *     time = ray_headers.view('uint16')[..., 5]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_bin2_to_angle); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_ray_headers, __pyx_n_s_view); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_n_u_uint16};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_4, __pyx_tuple__9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_el1 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":455
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 *     nbins = ray_headers.view('int16')[..., 4]             # <<<<<<<<<<<<<<
 *     time = ray_headers.view('uint16')[..., 5]
 *     prf_flag = np.mod(ray_headers.view('int16')[..., 0], 2)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_ray_headers, __pyx_n_s_view); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_n_u_int16};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_tuple__10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_nbins = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":456
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 *     nbins = ray_headers.view('int16')[..., 4]
 *     time = ray_headers.view('uint16')[..., 5]             # <<<<<<<<<<<<<<
 *     prf_flag = np.mod(ray_headers.view('int16')[..., 0], 2)
 *     return (az0, el0, az1, el1, nbins, time, prf_flag)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_ray_headers, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_n_u_uint16};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_tuple__11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_time = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":457
 *     nbins = ray_headers.view('int16')[..., 4]
 *     time = ray_headers.view('uint16')[..., 5]
 *     prf_flag = np.mod(ray_headers.view('int16')[..., 0], 2)             # <<<<<<<<<<<<<<
 *     return (az0, el0, az1, el1, nbins, time, prf_flag)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_mod); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_ray_headers, __pyx_n_s_view); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_n_u_int16};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_tuple__6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_4, __pyx_int_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_prf_flag = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":458
 *     time = ray_headers.view('uint16')[..., 5]
 *     prf_flag = np.mod(ray_headers.view('int16')[..., 0], 2)
 *     return (az0, el0, az1, el1, nbins, time, prf_flag)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_az0);
  __Pyx_GIVEREF(__pyx_v_az0);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_az0)) __PYX_ERR(0, 458, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_el0);
  __Pyx_GIVEREF(__pyx_v_el0);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_el0)) __PYX_ERR(0, 458, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_az1);
  __Pyx_GIVEREF(__pyx_v_az1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_az1)) __PYX_ERR(0, 458, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_el1);
  __Pyx_GIVEREF(__pyx_v_el1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_el1)) __PYX_ERR(0, 458, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_nbins);
  __Pyx_GIVEREF(__pyx_v_nbins);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_nbins)) __PYX_ERR(0, 458, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_time);
  __Pyx_GIVEREF(__pyx_v_time);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_v_time)) __PYX_ERR(0, 458, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_prf_flag);
  __Pyx_GIVEREF(__pyx_v_prf_flag);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_v_prf_flag)) __PYX_ERR(0, 458, __pyx_L1_error);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":423
 * 
 * 
 * def _parse_ray_headers(ray_headers):             # <<<<<<<<<<<<<<
 *     """
 *     Parse the metadata from Sigmet ray headers.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyart.io._sigmetfile._parse_ray_headers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_az0);
  __Pyx_XDECREF(__pyx_v_el0);
  __Pyx_XDECREF(__pyx_v_az1);
  __Pyx_XDECREF(__pyx_v_el1);
  __Pyx_XDECREF(__pyx_v_nbins);
  __Pyx_XDECREF(__pyx_v_time);
  __Pyx_XDECREF(__pyx_v_prf_flag);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":602
 * # file. Rewriting the convertions/masking in Cython does not seem to improved
 * # performance likely since most of the routines are already vectorized.
 * def convert_sigmet_data(data_type, data, nbins):             # <<<<<<<<<<<<<<
 *     """ Convert sigmet data. """
 *     out = np.empty_like(data, dtype='float32')
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_7convert_sigmet_data(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_6convert_sigmet_data, " Convert sigmet data. ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_7convert_sigmet_data = {"convert_sigmet_data", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_7convert_sigmet_data, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_6convert_sigmet_data};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_7convert_sigmet_data(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_data_type = 0;
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_nbins = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("convert_sigmet_data (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data_type,&__pyx_n_s_data,&__pyx_n_s_nbins,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_data_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 602, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_data)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 602, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("convert_sigmet_data", 1, 3, 3, 1); __PYX_ERR(0, 602, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_nbins)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 602, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("convert_sigmet_data", 1, 3, 3, 2); __PYX_ERR(0, 602, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "convert_sigmet_data") < 0)) __PYX_ERR(0, 602, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_data_type = values[0];
    __pyx_v_data = values[1];
    __pyx_v_nbins = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("convert_sigmet_data", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 602, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile.convert_sigmet_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_6convert_sigmet_data(__pyx_self, __pyx_v_data_type, __pyx_v_data, __pyx_v_nbins);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_6convert_sigmet_data(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data_type, PyObject *__pyx_v_data, PyObject *__pyx_v_nbins) {
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_v_mask = NULL;
  PyObject *__pyx_v_data_type_name = NULL;
  PyObject *__pyx_v_like_dbt2 = NULL;
  PyObject *__pyx_v_like_sqi = NULL;
  PyObject *__pyx_v_like_sqi2 = NULL;
  PyObject *__pyx_v_like_dbt = NULL;
  PyObject *__pyx_v_nrays = NULL;
  PyObject *__pyx_v_nbin = NULL;
  PyObject *__pyx_v_ndata = NULL;
  PyObject *__pyx_v_exp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  unsigned int __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_sigmet_data", 1);

  /* "pyart/io/_sigmetfile.pyx":604
 * def convert_sigmet_data(data_type, data, nbins):
 *     """ Convert sigmet data. """
 *     out = np.empty_like(data, dtype='float32')             # <<<<<<<<<<<<<<
 *     mask = np.zeros_like(data, dtype='bool')
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty_like); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_data)) __PYX_ERR(0, 604, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_n_u_float32) < 0) __PYX_ERR(0, 604, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_out = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pyart/io/_sigmetfile.pyx":605
 *     """ Convert sigmet data. """
 *     out = np.empty_like(data, dtype='float32')
 *     mask = np.zeros_like(data, dtype='bool')             # <<<<<<<<<<<<<<
 * 
 *     data_type_name = SIGMET_DATA_TYPES[data_type]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros_like); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_data)) __PYX_ERR(0, 605, __pyx_L1_error);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_u_bool) < 0) __PYX_ERR(0, 605, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mask = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":607
 *     mask = np.zeros_like(data, dtype='bool')
 * 
 *     data_type_name = SIGMET_DATA_TYPES[data_type]             # <<<<<<<<<<<<<<
 * 
 *     like_dbt2 = [
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SIGMET_DATA_TYPES); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_v_data_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_data_type_name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":609
 *     data_type_name = SIGMET_DATA_TYPES[data_type]
 * 
 *     like_dbt2 = [             # <<<<<<<<<<<<<<
 *         'DBT2',     # 2-byte Reflectivity Format, section 4.3.4
 *         'DBZ2',     # " "
 */
  __pyx_t_1 = PyList_New(17); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_DBT2);
  __Pyx_GIVEREF(__pyx_n_u_DBT2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_DBT2)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_DBZ2);
  __Pyx_GIVEREF(__pyx_n_u_DBZ2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_DBZ2)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_KDP2);
  __Pyx_GIVEREF(__pyx_n_u_KDP2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_KDP2)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_LDRH2);
  __Pyx_GIVEREF(__pyx_n_u_LDRH2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_LDRH2)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_LDRV2);
  __Pyx_GIVEREF(__pyx_n_u_LDRV2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_LDRV2)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_VEL2);
  __Pyx_GIVEREF(__pyx_n_u_VEL2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_VEL2)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_VELC2);
  __Pyx_GIVEREF(__pyx_n_u_VELC2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_u_VELC2)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_ZDR2);
  __Pyx_GIVEREF(__pyx_n_u_ZDR2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_u_ZDR2)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_DBZC2);
  __Pyx_GIVEREF(__pyx_n_u_DBZC2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_u_DBZC2)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_ZDRC2);
  __Pyx_GIVEREF(__pyx_n_u_ZDRC2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 9, __pyx_n_u_ZDRC2)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_DBTV16);
  __Pyx_GIVEREF(__pyx_n_u_DBTV16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 10, __pyx_n_u_DBTV16)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_DBZV16);
  __Pyx_GIVEREF(__pyx_n_u_DBZV16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 11, __pyx_n_u_DBZV16)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_SNR16);
  __Pyx_GIVEREF(__pyx_n_u_SNR16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 12, __pyx_n_u_SNR16)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_DBTE16);
  __Pyx_GIVEREF(__pyx_n_u_DBTE16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 13, __pyx_n_u_DBTE16)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_DBZE16);
  __Pyx_GIVEREF(__pyx_n_u_DBZE16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 14, __pyx_n_u_DBZE16)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_LOG16);
  __Pyx_GIVEREF(__pyx_n_u_LOG16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 15, __pyx_n_u_LOG16)) __PYX_ERR(0, 609, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_CSP16);
  __Pyx_GIVEREF(__pyx_n_u_CSP16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 16, __pyx_n_u_CSP16)) __PYX_ERR(0, 609, __pyx_L1_error);
  __pyx_v_like_dbt2 = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":629
 *     ]
 * 
 *     like_sqi = [             # <<<<<<<<<<<<<<
 *         'RHOH',     # 1-byte Rho Format, section 4.3.21
 *         'RHOV',     # " "
 */
  __pyx_t_1 = PyList_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_RHOH);
  __Pyx_GIVEREF(__pyx_n_u_RHOH);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_RHOH)) __PYX_ERR(0, 629, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_RHOV);
  __Pyx_GIVEREF(__pyx_n_u_RHOV);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_RHOV)) __PYX_ERR(0, 629, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_RHOHV);
  __Pyx_GIVEREF(__pyx_n_u_RHOHV);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_RHOHV)) __PYX_ERR(0, 629, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_SQI);
  __Pyx_GIVEREF(__pyx_n_u_SQI);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_SQI)) __PYX_ERR(0, 629, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_PMI8);
  __Pyx_GIVEREF(__pyx_n_u_PMI8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_PMI8)) __PYX_ERR(0, 629, __pyx_L1_error);
  __pyx_v_like_sqi = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":637
 *     ]
 * 
 *     like_sqi2 = [             # <<<<<<<<<<<<<<
 *         'RHOV2',    # 2-byte Rho Format, section 4.3.22
 *         'RHOH2',    # " "
 */
  __pyx_t_1 = PyList_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_RHOV2);
  __Pyx_GIVEREF(__pyx_n_u_RHOV2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_RHOV2)) __PYX_ERR(0, 637, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_RHOH2);
  __Pyx_GIVEREF(__pyx_n_u_RHOH2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_RHOH2)) __PYX_ERR(0, 637, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_RHOHV2);
  __Pyx_GIVEREF(__pyx_n_u_RHOHV2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_RHOHV2)) __PYX_ERR(0, 637, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_SQI2);
  __Pyx_GIVEREF(__pyx_n_u_SQI2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_SQI2)) __PYX_ERR(0, 637, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_PMI16);
  __Pyx_GIVEREF(__pyx_n_u_PMI16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_PMI16)) __PYX_ERR(0, 637, __pyx_L1_error);
  __pyx_v_like_sqi2 = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":645
 *     ]
 * 
 *     like_dbt = [             # <<<<<<<<<<<<<<
 *         'DBT',      # 1-bytes Reflectivity Format, section 4.3.3
 *         'DBZ',      # " "
 */
  __pyx_t_1 = PyList_New(9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_DBT);
  __Pyx_GIVEREF(__pyx_n_u_DBT);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_DBT)) __PYX_ERR(0, 645, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_DBZ);
  __Pyx_GIVEREF(__pyx_n_u_DBZ);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_DBZ)) __PYX_ERR(0, 645, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_DBTV8);
  __Pyx_GIVEREF(__pyx_n_u_DBTV8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_DBTV8)) __PYX_ERR(0, 645, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_DBZV8);
  __Pyx_GIVEREF(__pyx_n_u_DBZV8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_DBZV8)) __PYX_ERR(0, 645, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_SNR8);
  __Pyx_GIVEREF(__pyx_n_u_SNR8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_SNR8)) __PYX_ERR(0, 645, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_DBTE8);
  __Pyx_GIVEREF(__pyx_n_u_DBTE8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_DBTE8)) __PYX_ERR(0, 645, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_DBZE8);
  __Pyx_GIVEREF(__pyx_n_u_DBZE8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_u_DBZE8)) __PYX_ERR(0, 645, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_LOG8);
  __Pyx_GIVEREF(__pyx_n_u_LOG8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_u_LOG8)) __PYX_ERR(0, 645, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_CSP8);
  __Pyx_GIVEREF(__pyx_n_u_CSP8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_u_CSP8)) __PYX_ERR(0, 645, __pyx_L1_error);
  __pyx_v_like_dbt = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":657
 *     ]
 * 
 *     if data_type_name in like_dbt2:             # <<<<<<<<<<<<<<
 *         # value = (N - 32768) / 100.
 *         # 0 : no data available (mask)
 */
  __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_v_data_type_name, __pyx_v_like_dbt2, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 657, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "pyart/io/_sigmetfile.pyx":661
 *         # 0 : no data available (mask)
 *         # 65535 Reserved for area not scanned in product file (nothing)
 *         out[:] = (data.view('uint16') - 32768.) / 100.             # <<<<<<<<<<<<<<
 *         mask[data.view('uint16') == 0] = True
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 661, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_n_u_uint16};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 661, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_2 = __Pyx_PyFloat_SubtractObjC(__pyx_t_1, __pyx_float_32768_, 32768., 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 661, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_2, __pyx_float_100_, 100., 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 661, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1) < 0) __PYX_ERR(0, 661, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":662
 *         # 65535 Reserved for area not scanned in product file (nothing)
 *         out[:] = (data.view('uint16') - 32768.) / 100.
 *         mask[data.view('uint16') == 0] = True             # <<<<<<<<<<<<<<
 * 
 *     elif data_type_name in like_sqi2:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_n_u_uint16};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 662, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_2, Py_True) < 0))) __PYX_ERR(0, 662, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyart/io/_sigmetfile.pyx":657
 *     ]
 * 
 *     if data_type_name in like_dbt2:             # <<<<<<<<<<<<<<
 *         # value = (N - 32768) / 100.
 *         # 0 : no data available (mask)
 */
    goto __pyx_L3;
  }

  /* "pyart/io/_sigmetfile.pyx":664
 *         mask[data.view('uint16') == 0] = True
 * 
 *     elif data_type_name in like_sqi2:             # <<<<<<<<<<<<<<
 *         # value = (N - 1) / 65533
 *         # 0 : no data available (mask)
 */
  __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_v_data_type_name, __pyx_v_like_sqi2, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 664, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "pyart/io/_sigmetfile.pyx":668
 *         # 0 : no data available (mask)
 *         # 65535 Area not scanned
 *         out[:] = (data.view('uint16') - 1.) / 65533.             # <<<<<<<<<<<<<<
 *         mask[data.view('uint16') == 0] = True
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 668, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_n_u_uint16};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 668, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_t_1 = __Pyx_PyFloat_SubtractObjC(__pyx_t_2, __pyx_float_1_, 1., 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 668, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_1, __pyx_float_65533_, 65533., 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 668, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_2, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1) < 0) __PYX_ERR(0, 668, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyart/io/_sigmetfile.pyx":669
 *         # 65535 Area not scanned
 *         out[:] = (data.view('uint16') - 1.) / 65533.
 *         mask[data.view('uint16') == 0] = True             # <<<<<<<<<<<<<<
 * 
 *     elif data_type_name == 'WIDTH2':
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_n_u_uint16};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_1, Py_True) < 0))) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":664
 *         mask[data.view('uint16') == 0] = True
 * 
 *     elif data_type_name in like_sqi2:             # <<<<<<<<<<<<<<
 *         # value = (N - 1) / 65533
 *         # 0 : no data available (mask)
 */
    goto __pyx_L3;
  }

  /* "pyart/io/_sigmetfile.pyx":671
 *         mask[data.view('uint16') == 0] = True
 * 
 *     elif data_type_name == 'WIDTH2':             # <<<<<<<<<<<<<<
 *         # DB_WIDTH2, 11, Width (2 byte)
 *         # 2-byte Width Format, section 4.3.36
 */
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_data_type_name, __pyx_n_u_WIDTH2, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 671, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "pyart/io/_sigmetfile.pyx":674
 *         # DB_WIDTH2, 11, Width (2 byte)
 *         # 2-byte Width Format, section 4.3.36
 *         out[:] = data.view('uint16') / 100.             # <<<<<<<<<<<<<<
 *         mask[data.view('uint16') == 0] = True
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_n_u_uint16};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_2 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_1, __pyx_float_100_, 100., 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_2, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1) < 0) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyart/io/_sigmetfile.pyx":675
 *         # 2-byte Width Format, section 4.3.36
 *         out[:] = data.view('uint16') / 100.
 *         mask[data.view('uint16') == 0] = True             # <<<<<<<<<<<<<<
 * 
 *     elif data_type_name == 'PHIDP2':
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_n_u_uint16};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 675, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_1, Py_True) < 0))) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":671
 *         mask[data.view('uint16') == 0] = True
 * 
 *     elif data_type_name == 'WIDTH2':             # <<<<<<<<<<<<<<
 *         # DB_WIDTH2, 11, Width (2 byte)
 *         # 2-byte Width Format, section 4.3.36
 */
    goto __pyx_L3;
  }

  /* "pyart/io/_sigmetfile.pyx":677
 *         mask[data.view('uint16') == 0] = True
 * 
 *     elif data_type_name == 'PHIDP2':             # <<<<<<<<<<<<<<
 *         # DB_PHIDP2, 24, PhiDP (Differential Phase) (2 byte)
 *         # 2-byte PhiDP format, section 4.3.19
 */
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_data_type_name, __pyx_n_u_PHIDP2, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 677, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "pyart/io/_sigmetfile.pyx":680
 *         # DB_PHIDP2, 24, PhiDP (Differential Phase) (2 byte)
 *         # 2-byte PhiDP format, section 4.3.19
 *         out[:] = 360. * (data.view('uint16') - 1.) / 65534.             # <<<<<<<<<<<<<<
 *         mask[data.view('uint16') == 0] = True
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 680, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_n_u_uint16};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 680, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_2 = __Pyx_PyFloat_SubtractObjC(__pyx_t_1, __pyx_float_1_, 1., 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 680, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Multiply(__pyx_float_360_, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 680, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_1, __pyx_float_65534_, 65534., 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 680, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_2, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1) < 0) __PYX_ERR(0, 680, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyart/io/_sigmetfile.pyx":681
 *         # 2-byte PhiDP format, section 4.3.19
 *         out[:] = 360. * (data.view('uint16') - 1.) / 65534.
 *         mask[data.view('uint16') == 0] = True             # <<<<<<<<<<<<<<
 * 
 *     elif data_type_name == 'HCLASS2':
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_n_u_uint16};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 681, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_1, Py_True) < 0))) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":677
 *         mask[data.view('uint16') == 0] = True
 * 
 *     elif data_type_name == 'PHIDP2':             # <<<<<<<<<<<<<<
 *         # DB_PHIDP2, 24, PhiDP (Differential Phase) (2 byte)
 *         # 2-byte PhiDP format, section 4.3.19
 */
    goto __pyx_L3;
  }

  /* "pyart/io/_sigmetfile.pyx":683
 *         mask[data.view('uint16') == 0] = True
 * 
 *     elif data_type_name == 'HCLASS2':             # <<<<<<<<<<<<<<
 *         # DB_HCLASS2, 56, Hydrometeor class (2 byte)
 *         # 2-byte HydroClass Format, section 4.3.9
 */
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_data_type_name, __pyx_n_u_HCLASS2, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 683, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "pyart/io/_sigmetfile.pyx":686
 *         # DB_HCLASS2, 56, Hydrometeor class (2 byte)
 *         # 2-byte HydroClass Format, section 4.3.9
 *         out[:] = data.view('uint16')             # <<<<<<<<<<<<<<
 * 
 *     elif data_type_name == 'XHDR':
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 686, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_n_u_uint16};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 686, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1) < 0) __PYX_ERR(0, 686, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":683
 *         mask[data.view('uint16') == 0] = True
 * 
 *     elif data_type_name == 'HCLASS2':             # <<<<<<<<<<<<<<
 *         # DB_HCLASS2, 56, Hydrometeor class (2 byte)
 *         # 2-byte HydroClass Format, section 4.3.9
 */
    goto __pyx_L3;
  }

  /* "pyart/io/_sigmetfile.pyx":688
 *         out[:] = data.view('uint16')
 * 
 *     elif data_type_name == 'XHDR':             # <<<<<<<<<<<<<<
 *         # Extended Headers, 0
 *         # extended_header_v0, _v1, _v2, section 4.2.8-4.2.10
 */
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_data_type_name, __pyx_n_u_XHDR, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 688, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "pyart/io/_sigmetfile.pyx":692
 *         # extended_header_v0, _v1, _v2, section 4.2.8-4.2.10
 *         # Here we return an array with the times in milliseconds.
 *         return data[..., :2].copy().view('i4')             # <<<<<<<<<<<<<<
 * 
 *     # one byte data types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_data, __pyx_tuple__14); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_copy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 692, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_view); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_n_u_i4};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 692, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pyart/io/_sigmetfile.pyx":688
 *         out[:] = data.view('uint16')
 * 
 *     elif data_type_name == 'XHDR':             # <<<<<<<<<<<<<<
 *         # Extended Headers, 0
 *         # extended_header_v0, _v1, _v2, section 4.2.8-4.2.10
 */
  }

  /* "pyart/io/_sigmetfile.pyx":695
 * 
 *     # one byte data types
 *     elif data_type_name[-1] != '2':             # <<<<<<<<<<<<<<
 *         # make a view of left half of the data as uint8,
 *         # this is the actual ray data collected, the right half is blank.
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_data_type_name, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_kp_u_2, Py_NE)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "pyart/io/_sigmetfile.pyx":698
 *         # make a view of left half of the data as uint8,
 *         # this is the actual ray data collected, the right half is blank.
 *         nrays, nbin = data.shape             # <<<<<<<<<<<<<<
 *         ndata = data.view('(2,) uint8').reshape(nrays, -1)[:, :nbin]
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 698, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 698, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 698, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 698, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4);
      index = 0; __pyx_t_3 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L4_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_2 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L4_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_4), 2) < 0) __PYX_ERR(0, 698, __pyx_L1_error)
      __pyx_t_7 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L5_unpacking_done;
      __pyx_L4_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_7 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 698, __pyx_L1_error)
      __pyx_L5_unpacking_done:;
    }
    __pyx_v_nrays = __pyx_t_3;
    __pyx_t_3 = 0;
    __pyx_v_nbin = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pyart/io/_sigmetfile.pyx":699
 *         # this is the actual ray data collected, the right half is blank.
 *         nrays, nbin = data.shape
 *         ndata = data.view('(2,) uint8').reshape(nrays, -1)[:, :nbin]             # <<<<<<<<<<<<<<
 * 
 *         if data_type_name in like_dbt:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_2_uint8};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 699, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_reshape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_nrays, __pyx_int_neg_1};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 699, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_3 = PySlice_New(Py_None, __pyx_v_nbin, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_slice__12);
    __Pyx_GIVEREF(__pyx_slice__12);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_slice__12)) __PYX_ERR(0, 699, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3)) __PYX_ERR(0, 699, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_ndata = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pyart/io/_sigmetfile.pyx":701
 *         ndata = data.view('(2,) uint8').reshape(nrays, -1)[:, :nbin]
 * 
 *         if data_type_name in like_dbt:             # <<<<<<<<<<<<<<
 *             # DB_DBT, 1, Total Power (1 byte)
 *             # 1-byte Reflectivity Format, section 4.3.3
 */
    __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_v_data_type_name, __pyx_v_like_dbt, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 701, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "pyart/io/_sigmetfile.pyx":704
 *             # DB_DBT, 1, Total Power (1 byte)
 *             # 1-byte Reflectivity Format, section 4.3.3
 *             out[:] = (ndata - 64.) / 2.             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 * 
 */
      __pyx_t_3 = __Pyx_PyFloat_SubtractObjC(__pyx_v_ndata, __pyx_float_64_, 64., 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 704, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_3, __pyx_float_2_, 2., 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 704, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_2, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1) < 0) __PYX_ERR(0, 704, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":705
 *             # 1-byte Reflectivity Format, section 4.3.3
 *             out[:] = (ndata - 64.) / 2.
 *             mask[ndata == 0] = True             # <<<<<<<<<<<<<<
 * 
 *         elif data_type_name in like_sqi:
 */
      __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_ndata, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 705, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_2, Py_True) < 0))) __PYX_ERR(0, 705, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":701
 *         ndata = data.view('(2,) uint8').reshape(nrays, -1)[:, :nbin]
 * 
 *         if data_type_name in like_dbt:             # <<<<<<<<<<<<<<
 *             # DB_DBT, 1, Total Power (1 byte)
 *             # 1-byte Reflectivity Format, section 4.3.3
 */
      goto __pyx_L6;
    }

    /* "pyart/io/_sigmetfile.pyx":707
 *             mask[ndata == 0] = True
 * 
 *         elif data_type_name in like_sqi:             # <<<<<<<<<<<<<<
 *             # value = sqrt((N - 1) / 253)
 *             # 0 : no data available (mask)
 */
    __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_v_data_type_name, __pyx_v_like_sqi, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 707, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "pyart/io/_sigmetfile.pyx":711
 *             # 0 : no data available (mask)
 *             # 255 Area not scanned
 *             out[:] = np.sqrt((ndata - 1.) / 253.)             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 *             mask[ndata == 255] = True
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_sqrt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyFloat_SubtractObjC(__pyx_v_ndata, __pyx_float_1_, 1., 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_3, __pyx_float_253_, 253., 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 711, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_2, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1) < 0) __PYX_ERR(0, 711, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":712
 *             # 255 Area not scanned
 *             out[:] = np.sqrt((ndata - 1.) / 253.)
 *             mask[ndata == 0] = True             # <<<<<<<<<<<<<<
 *             mask[ndata == 255] = True
 * 
 */
      __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_ndata, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_2, Py_True) < 0))) __PYX_ERR(0, 712, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":713
 *             out[:] = np.sqrt((ndata - 1.) / 253.)
 *             mask[ndata == 0] = True
 *             mask[ndata == 255] = True             # <<<<<<<<<<<<<<
 * 
 *         elif data_type_name == 'VEL':
 */
      __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_ndata, __pyx_int_255, 0xFF, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 713, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_2, Py_True) < 0))) __PYX_ERR(0, 713, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":707
 *             mask[ndata == 0] = True
 * 
 *         elif data_type_name in like_sqi:             # <<<<<<<<<<<<<<
 *             # value = sqrt((N - 1) / 253)
 *             # 0 : no data available (mask)
 */
      goto __pyx_L6;
    }

    /* "pyart/io/_sigmetfile.pyx":715
 *             mask[ndata == 255] = True
 * 
 *         elif data_type_name == 'VEL':             # <<<<<<<<<<<<<<
 *             # VEL, 3, Velocity (1 byte)
 *             # 1-byte Velocity Format, section 4.3.29
 */
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_data_type_name, __pyx_n_u_VEL, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 715, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "pyart/io/_sigmetfile.pyx":720
 *             # Note that this data should be multiplied by Nyquist,
 *             # this is done in the get_data method of the SigmetFile class.
 *             out[:] = (ndata - 128.) / 127.             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 * 
 */
      __pyx_t_2 = __Pyx_PyFloat_SubtractObjC(__pyx_v_ndata, __pyx_float_128_, 128., 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 720, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_2, __pyx_float_127_, 127., 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 720, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1) < 0) __PYX_ERR(0, 720, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyart/io/_sigmetfile.pyx":721
 *             # this is done in the get_data method of the SigmetFile class.
 *             out[:] = (ndata - 128.) / 127.
 *             mask[ndata == 0] = True             # <<<<<<<<<<<<<<
 * 
 *         elif data_type_name == 'VELC':
 */
      __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_ndata, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 721, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_1, Py_True) < 0))) __PYX_ERR(0, 721, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyart/io/_sigmetfile.pyx":715
 *             mask[ndata == 255] = True
 * 
 *         elif data_type_name == 'VEL':             # <<<<<<<<<<<<<<
 *             # VEL, 3, Velocity (1 byte)
 *             # 1-byte Velocity Format, section 4.3.29
 */
      goto __pyx_L6;
    }

    /* "pyart/io/_sigmetfile.pyx":723
 *             mask[ndata == 0] = True
 * 
 *         elif data_type_name == 'VELC':             # <<<<<<<<<<<<<<
 *             # VELC, 3, Velocity (1 byte)
 *             # 1-byte Corrected Velocity Format, section 4.4.42
 */
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_data_type_name, __pyx_n_u_VELC, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 723, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "pyart/io/_sigmetfile.pyx":726
 *             # VELC, 3, Velocity (1 byte)
 *             # 1-byte Corrected Velocity Format, section 4.4.42
 *             out[:] = (ndata - 128.) / 127. *75.             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 *             mask[ndata == 255] = True
 */
      __pyx_t_1 = __Pyx_PyFloat_SubtractObjC(__pyx_v_ndata, __pyx_float_128_, 128., 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_1, __pyx_float_127_, 127., 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyNumber_Multiply(__pyx_t_2, __pyx_float_75_); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1) < 0) __PYX_ERR(0, 726, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyart/io/_sigmetfile.pyx":727
 *             # 1-byte Corrected Velocity Format, section 4.4.42
 *             out[:] = (ndata - 128.) / 127. *75.
 *             mask[ndata == 0] = True             # <<<<<<<<<<<<<<
 *             mask[ndata == 255] = True
 * 
 */
      __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_ndata, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 727, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_1, Py_True) < 0))) __PYX_ERR(0, 727, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyart/io/_sigmetfile.pyx":728
 *             out[:] = (ndata - 128.) / 127. *75.
 *             mask[ndata == 0] = True
 *             mask[ndata == 255] = True             # <<<<<<<<<<<<<<
 * 
 *         elif data_type_name == 'WIDTH':
 */
      __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_ndata, __pyx_int_255, 0xFF, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 728, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_1, Py_True) < 0))) __PYX_ERR(0, 728, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyart/io/_sigmetfile.pyx":723
 *             mask[ndata == 0] = True
 * 
 *         elif data_type_name == 'VELC':             # <<<<<<<<<<<<<<
 *             # VELC, 3, Velocity (1 byte)
 *             # 1-byte Corrected Velocity Format, section 4.4.42
 */
      goto __pyx_L6;
    }

    /* "pyart/io/_sigmetfile.pyx":730
 *             mask[ndata == 255] = True
 * 
 *         elif data_type_name == 'WIDTH':             # <<<<<<<<<<<<<<
 *             # WIDTH, 4, Width (1 byte)
 *             # 1-byte Width format, section 4.3.25
 */
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_data_type_name, __pyx_n_u_WIDTH, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 730, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "pyart/io/_sigmetfile.pyx":735
 *             # Note that this data should be multiplied by the unambiguous
 *             # velocity
 *             out[:] = ndata / 256.             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 * 
 */
      __pyx_t_1 = __Pyx_PyFloat_TrueDivideObjC(__pyx_v_ndata, __pyx_float_256_, 256., 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 735, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1) < 0) __PYX_ERR(0, 735, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyart/io/_sigmetfile.pyx":736
 *             # velocity
 *             out[:] = ndata / 256.
 *             mask[ndata == 0] = True             # <<<<<<<<<<<<<<
 * 
 *         elif data_type_name == 'ZDR':
 */
      __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_ndata, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 736, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_1, Py_True) < 0))) __PYX_ERR(0, 736, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyart/io/_sigmetfile.pyx":730
 *             mask[ndata == 255] = True
 * 
 *         elif data_type_name == 'WIDTH':             # <<<<<<<<<<<<<<
 *             # WIDTH, 4, Width (1 byte)
 *             # 1-byte Width format, section 4.3.25
 */
      goto __pyx_L6;
    }

    /* "pyart/io/_sigmetfile.pyx":738
 *             mask[ndata == 0] = True
 * 
 *         elif data_type_name == 'ZDR':             # <<<<<<<<<<<<<<
 *             # ZDR, 5, Differential reflectivity (1 byte)
 *             # 1-byte ZDR format, section 4.3.37
 */
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_data_type_name, __pyx_n_u_ZDR, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 738, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "pyart/io/_sigmetfile.pyx":741
 *             # ZDR, 5, Differential reflectivity (1 byte)
 *             # 1-byte ZDR format, section 4.3.37
 *             out[:] = (ndata - 128.) / 16.             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 * 
 */
      __pyx_t_1 = __Pyx_PyFloat_SubtractObjC(__pyx_v_ndata, __pyx_float_128_, 128., 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 741, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_1, __pyx_float_16_, 16., 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 741, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_2, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1) < 0) __PYX_ERR(0, 741, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":742
 *             # 1-byte ZDR format, section 4.3.37
 *             out[:] = (ndata - 128.) / 16.
 *             mask[ndata == 0] = True             # <<<<<<<<<<<<<<
 * 
 *         elif data_type_name == 'KDP':
 */
      __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_ndata, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 742, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_2, Py_True) < 0))) __PYX_ERR(0, 742, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":738
 *             mask[ndata == 0] = True
 * 
 *         elif data_type_name == 'ZDR':             # <<<<<<<<<<<<<<
 *             # ZDR, 5, Differential reflectivity (1 byte)
 *             # 1-byte ZDR format, section 4.3.37
 */
      goto __pyx_L6;
    }

    /* "pyart/io/_sigmetfile.pyx":744
 *             mask[ndata == 0] = True
 * 
 *         elif data_type_name == 'KDP':             # <<<<<<<<<<<<<<
 *             # KDP, 14, KDP (Differential phase) (1 byte)
 *             # 1-byte KDP format, section 4.3.12
 */
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_data_type_name, __pyx_n_u_KDP, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 744, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "pyart/io/_sigmetfile.pyx":751
 * 
 *             # above 128 use positive value equation
 *             exp = np.power(600., (ndata[ndata > 128] - 129.) / 126.)             # <<<<<<<<<<<<<<
 *             out[ndata > 128] = 0.25 * exp
 *             # below 128, use negative value equation
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 751, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_power); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 751, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_128, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 751, __pyx_L1_error)
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_ndata, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 751, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyFloat_SubtractObjC(__pyx_t_3, __pyx_float_129_, 129., 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 751, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_1, __pyx_float_126_, 126., 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 751, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_float_600_, __pyx_t_3};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 751, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_v_exp = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":752
 *             # above 128 use positive value equation
 *             exp = np.power(600., (ndata[ndata > 128] - 129.) / 126.)
 *             out[ndata > 128] = 0.25 * exp             # <<<<<<<<<<<<<<
 *             # below 128, use negative value equation
 *             exp = np.power(600., (127. - ndata[ndata < 128]) / 126.)
 */
      __pyx_t_2 = PyNumber_Multiply(__pyx_float_0_25, __pyx_v_exp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 752, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_128, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 752, __pyx_L1_error)
      if (unlikely((PyObject_SetItem(__pyx_v_out, __pyx_t_4, __pyx_t_2) < 0))) __PYX_ERR(0, 752, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":754
 *             out[ndata > 128] = 0.25 * exp
 *             # below 128, use negative value equation
 *             exp = np.power(600., (127. - ndata[ndata < 128]) / 126.)             # <<<<<<<<<<<<<<
 *             out[ndata < 128] = -0.25 * exp
 *             # equal to 128, zero
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 754, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_power); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 754, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_128, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 754, __pyx_L1_error)
      __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_ndata, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 754, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyFloat_SubtractCObj(__pyx_float_127_, __pyx_t_1, 127., 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 754, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_4, __pyx_float_126_, 126., 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 754, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_float_600_, __pyx_t_1};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 754, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF_SET(__pyx_v_exp, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":755
 *             # below 128, use negative value equation
 *             exp = np.power(600., (127. - ndata[ndata < 128]) / 126.)
 *             out[ndata < 128] = -0.25 * exp             # <<<<<<<<<<<<<<
 *             # equal to 128, zero
 *             out[ndata == 128] = 0
 */
      __pyx_t_2 = PyNumber_Multiply(__pyx_float_neg_0_25, __pyx_v_exp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 755, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_128, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 755, __pyx_L1_error)
      if (unlikely((PyObject_SetItem(__pyx_v_out, __pyx_t_3, __pyx_t_2) < 0))) __PYX_ERR(0, 755, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":757
 *             out[ndata < 128] = -0.25 * exp
 *             # equal to 128, zero
 *             out[ndata == 128] = 0             # <<<<<<<<<<<<<<
 * 
 *             mask[ndata == 0] = True
 */
      __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_ndata, __pyx_int_128, 0x80, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 757, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely((PyObject_SetItem(__pyx_v_out, __pyx_t_2, __pyx_int_0) < 0))) __PYX_ERR(0, 757, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":759
 *             out[ndata == 128] = 0
 * 
 *             mask[ndata == 0] = True             # <<<<<<<<<<<<<<
 *             mask[ndata == 255] = True
 * 
 */
      __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_ndata, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 759, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_2, Py_True) < 0))) __PYX_ERR(0, 759, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":760
 * 
 *             mask[ndata == 0] = True
 *             mask[ndata == 255] = True             # <<<<<<<<<<<<<<
 * 
 *         elif data_type_name == 'PHIDP':
 */
      __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_ndata, __pyx_int_255, 0xFF, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 760, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_2, Py_True) < 0))) __PYX_ERR(0, 760, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":744
 *             mask[ndata == 0] = True
 * 
 *         elif data_type_name == 'KDP':             # <<<<<<<<<<<<<<
 *             # KDP, 14, KDP (Differential phase) (1 byte)
 *             # 1-byte KDP format, section 4.3.12
 */
      goto __pyx_L6;
    }

    /* "pyart/io/_sigmetfile.pyx":762
 *             mask[ndata == 255] = True
 * 
 *         elif data_type_name == 'PHIDP':             # <<<<<<<<<<<<<<
 *             # PHIDP, 16, PhiDP(Differential phase) (1 byte)
 *             # 1-byte PhiDP format, section 4.3.18
 */
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_data_type_name, __pyx_n_u_PHIDP, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 762, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "pyart/io/_sigmetfile.pyx":765
 *             # PHIDP, 16, PhiDP(Differential phase) (1 byte)
 *             # 1-byte PhiDP format, section 4.3.18
 *             out[:] = 180. * ((ndata - 1.) / 254.)             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 *             mask[ndata == 255] = True
 */
      __pyx_t_2 = __Pyx_PyFloat_SubtractObjC(__pyx_v_ndata, __pyx_float_1_, 1., 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 765, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyFloat_TrueDivideObjC(__pyx_t_2, __pyx_float_254_, 254., 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 765, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Multiply(__pyx_float_180_, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 765, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_2, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1) < 0) __PYX_ERR(0, 765, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":766
 *             # 1-byte PhiDP format, section 4.3.18
 *             out[:] = 180. * ((ndata - 1.) / 254.)
 *             mask[ndata == 0] = True             # <<<<<<<<<<<<<<
 *             mask[ndata == 255] = True
 * 
 */
      __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_ndata, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_2, Py_True) < 0))) __PYX_ERR(0, 766, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":767
 *             out[:] = 180. * ((ndata - 1.) / 254.)
 *             mask[ndata == 0] = True
 *             mask[ndata == 255] = True             # <<<<<<<<<<<<<<
 * 
 *         elif data_type_name == "HCLASS":
 */
      __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_ndata, __pyx_int_255, 0xFF, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 767, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_2, Py_True) < 0))) __PYX_ERR(0, 767, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":762
 *             mask[ndata == 255] = True
 * 
 *         elif data_type_name == 'PHIDP':             # <<<<<<<<<<<<<<
 *             # PHIDP, 16, PhiDP(Differential phase) (1 byte)
 *             # 1-byte PhiDP format, section 4.3.18
 */
      goto __pyx_L6;
    }

    /* "pyart/io/_sigmetfile.pyx":769
 *             mask[ndata == 255] = True
 * 
 *         elif data_type_name == "HCLASS":             # <<<<<<<<<<<<<<
 *             # HCLASS, 55, Hydrometeor class (1 byte)
 *             # 1-byte HydroClass format, section 4.3.8
 */
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_data_type_name, __pyx_n_u_HCLASS, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 769, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "pyart/io/_sigmetfile.pyx":772
 *             # HCLASS, 55, Hydrometeor class (1 byte)
 *             # 1-byte HydroClass format, section 4.3.8
 *             out[:] = ndata[:]             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True     # No data available
 *             mask[ndata == 255] = True   # Area not scanned
 */
      __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_ndata, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 772, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_2, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1) < 0) __PYX_ERR(0, 772, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":773
 *             # 1-byte HydroClass format, section 4.3.8
 *             out[:] = ndata[:]
 *             mask[ndata == 0] = True     # No data available             # <<<<<<<<<<<<<<
 *             mask[ndata == 255] = True   # Area not scanned
 * 
 */
      __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_ndata, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_2, Py_True) < 0))) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":774
 *             out[:] = ndata[:]
 *             mask[ndata == 0] = True     # No data available
 *             mask[ndata == 255] = True   # Area not scanned             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
      __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_ndata, __pyx_int_255, 0xFF, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 774, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely((PyObject_SetItem(__pyx_v_mask, __pyx_t_2, Py_True) < 0))) __PYX_ERR(0, 774, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":769
 *             mask[ndata == 255] = True
 * 
 *         elif data_type_name == "HCLASS":             # <<<<<<<<<<<<<<
 *             # HCLASS, 55, Hydrometeor class (1 byte)
 *             # 1-byte HydroClass format, section 4.3.8
 */
      goto __pyx_L6;
    }

    /* "pyart/io/_sigmetfile.pyx":778
 *         else:
 *             # TODO implement conversions for addition 1-byte formats
 *             warnings.warn('Unknown type: %s, returning raw data' % data_type)             # <<<<<<<<<<<<<<
 *             out[:] = np.ma.masked_array(data)
 *             return out
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_warnings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 778, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 778, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Unknown_type_s_returning_raw_dat, __pyx_v_data_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 778, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 778, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":779
 *             # TODO implement conversions for addition 1-byte formats
 *             warnings.warn('Unknown type: %s, returning raw data' % data_type)
 *             out[:] = np.ma.masked_array(data)             # <<<<<<<<<<<<<<
 *             return out
 *     else:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 779, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ma); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 779, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_masked_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 779, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_data};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 779, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_2, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1) < 0) __PYX_ERR(0, 779, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyart/io/_sigmetfile.pyx":780
 *             warnings.warn('Unknown type: %s, returning raw data' % data_type)
 *             out[:] = np.ma.masked_array(data)
 *             return out             # <<<<<<<<<<<<<<
 *     else:
 *         # TODO implement conversions for additional formats.
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_out);
      __pyx_r = __pyx_v_out;
      goto __pyx_L0;
    }
    __pyx_L6:;

    /* "pyart/io/_sigmetfile.pyx":695
 * 
 *     # one byte data types
 *     elif data_type_name[-1] != '2':             # <<<<<<<<<<<<<<
 *         # make a view of left half of the data as uint8,
 *         # this is the actual ray data collected, the right half is blank.
 */
    goto __pyx_L3;
  }

  /* "pyart/io/_sigmetfile.pyx":783
 *     else:
 *         # TODO implement conversions for additional formats.
 *         warnings.warn('Unknown type: %s, returning raw data' % data_type)             # <<<<<<<<<<<<<<
 *         out[:] = data
 *         return np.ma.masked_array(out)
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Unknown_type_s_returning_raw_dat, __pyx_v_data_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_1};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 783, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyart/io/_sigmetfile.pyx":784
 *         # TODO implement conversions for additional formats.
 *         warnings.warn('Unknown type: %s, returning raw data' % data_type)
 *         out[:] = data             # <<<<<<<<<<<<<<
 *         return np.ma.masked_array(out)
 * 
 */
    if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_v_data, 0, 0, NULL, NULL, &__pyx_slice__12, 0, 0, 1) < 0) __PYX_ERR(0, 784, __pyx_L1_error)

    /* "pyart/io/_sigmetfile.pyx":785
 *         warnings.warn('Unknown type: %s, returning raw data' % data_type)
 *         out[:] = data
 *         return np.ma.masked_array(out)             # <<<<<<<<<<<<<<
 * 
 *     # mask any gates which are beyond the number of gates in that ray.
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 785, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ma); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 785, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_masked_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 785, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_out};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 785, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  __pyx_L3:;

  /* "pyart/io/_sigmetfile.pyx":788
 * 
 *     # mask any gates which are beyond the number of gates in that ray.
 *     _mask_gates_not_collected(mask.view(np.uint8), nbins)             # <<<<<<<<<<<<<<
 * 
 *     return np.ma.masked_array(out, mask=mask, fill_value=-9999.0,
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_mask, __pyx_n_s_view); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_uint8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_t_4};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 788, __pyx_L1_error)
  if (!(likely(((__pyx_v_nbins) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_nbins, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 788, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5pyart_2io_11_sigmetfile__mask_gates_not_collected(((PyArrayObject *)__pyx_t_2), ((PyArrayObject *)__pyx_v_nbins)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":790
 *     _mask_gates_not_collected(mask.view(np.uint8), nbins)
 * 
 *     return np.ma.masked_array(out, mask=mask, fill_value=-9999.0,             # <<<<<<<<<<<<<<
 *                               shrink=False)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ma); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_masked_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_out);
  __Pyx_GIVEREF(__pyx_v_out);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_out)) __PYX_ERR(0, 790, __pyx_L1_error);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_mask, __pyx_v_mask) < 0) __PYX_ERR(0, 790, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_fill_value, __pyx_float_neg_9999_0) < 0) __PYX_ERR(0, 790, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":791
 * 
 *     return np.ma.masked_array(out, mask=mask, fill_value=-9999.0,
 *                               shrink=False)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_shrink, Py_False) < 0) __PYX_ERR(0, 790, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":790
 *     _mask_gates_not_collected(mask.view(np.uint8), nbins)
 * 
 *     return np.ma.masked_array(out, mask=mask, fill_value=-9999.0,             # <<<<<<<<<<<<<<
 *                               shrink=False)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":602
 * # file. Rewriting the convertions/masking in Cython does not seem to improved
 * # performance likely since most of the routines are already vectorized.
 * def convert_sigmet_data(data_type, data, nbins):             # <<<<<<<<<<<<<<
 *     """ Convert sigmet data. """
 *     out = np.empty_like(data, dtype='float32')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyart.io._sigmetfile.convert_sigmet_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XDECREF(__pyx_v_data_type_name);
  __Pyx_XDECREF(__pyx_v_like_dbt2);
  __Pyx_XDECREF(__pyx_v_like_sqi);
  __Pyx_XDECREF(__pyx_v_like_sqi2);
  __Pyx_XDECREF(__pyx_v_like_dbt);
  __Pyx_XDECREF(__pyx_v_nrays);
  __Pyx_XDECREF(__pyx_v_nbin);
  __Pyx_XDECREF(__pyx_v_ndata);
  __Pyx_XDECREF(__pyx_v_exp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":795
 * 
 * @cython.boundscheck(False)
 * cdef _mask_gates_not_collected(             # <<<<<<<<<<<<<<
 *         np.ndarray[np.uint8_t, ndim=2] mask,
 *         np.ndarray[np.int16_t, ndim=1] nbins):
 */

static PyObject *__pyx_f_5pyart_2io_11_sigmetfile__mask_gates_not_collected(PyArrayObject *__pyx_v_mask, PyArrayObject *__pyx_v_nbins) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_nrays;
  int __pyx_v_nbin;
  int __pyx_v_full_nbins;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_nbins;
  __Pyx_Buffer __pyx_pybuffer_nbins;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_mask_gates_not_collected", 1);
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  __pyx_pybuffer_nbins.pybuffer.buf = NULL;
  __pyx_pybuffer_nbins.refcount = 0;
  __pyx_pybuffernd_nbins.data = NULL;
  __pyx_pybuffernd_nbins.rcbuffer = &__pyx_pybuffer_nbins;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 795, __pyx_L1_error)
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask.diminfo[1].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask.diminfo[1].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_nbins.rcbuffer->pybuffer, (PyObject*)__pyx_v_nbins, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int16_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 795, __pyx_L1_error)
  }
  __pyx_pybuffernd_nbins.diminfo[0].strides = __pyx_pybuffernd_nbins.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_nbins.diminfo[0].shape = __pyx_pybuffernd_nbins.rcbuffer->pybuffer.shape[0];

  /* "pyart/io/_sigmetfile.pyx":800
 *     """ Add gates not collected (beyond nbin) to the mask. """
 *     cdef int i, j, nrays, nbin, full_nbins
 *     nrays = mask.shape[0]             # <<<<<<<<<<<<<<
 *     full_nbins = mask.shape[1]
 *     for i in range(nrays):
 */
  __pyx_v_nrays = (__pyx_f_5numpy_7ndarray_5shape_shape(((PyArrayObject *)__pyx_v_mask))[0]);

  /* "pyart/io/_sigmetfile.pyx":801
 *     cdef int i, j, nrays, nbin, full_nbins
 *     nrays = mask.shape[0]
 *     full_nbins = mask.shape[1]             # <<<<<<<<<<<<<<
 *     for i in range(nrays):
 *         nbin = nbins[i]
 */
  __pyx_v_full_nbins = (__pyx_f_5numpy_7ndarray_5shape_shape(((PyArrayObject *)__pyx_v_mask))[1]);

  /* "pyart/io/_sigmetfile.pyx":802
 *     nrays = mask.shape[0]
 *     full_nbins = mask.shape[1]
 *     for i in range(nrays):             # <<<<<<<<<<<<<<
 *         nbin = nbins[i]
 *         for j in range(nbin, full_nbins):
 */
  __pyx_t_1 = __pyx_v_nrays;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "pyart/io/_sigmetfile.pyx":803
 *     full_nbins = mask.shape[1]
 *     for i in range(nrays):
 *         nbin = nbins[i]             # <<<<<<<<<<<<<<
 *         for j in range(nbin, full_nbins):
 *             mask[i, j] = 1
 */
    __pyx_t_4 = __pyx_v_i;
    if (__pyx_t_4 < 0) __pyx_t_4 += __pyx_pybuffernd_nbins.diminfo[0].shape;
    __pyx_v_nbin = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int16_t *, __pyx_pybuffernd_nbins.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_nbins.diminfo[0].strides));

    /* "pyart/io/_sigmetfile.pyx":804
 *     for i in range(nrays):
 *         nbin = nbins[i]
 *         for j in range(nbin, full_nbins):             # <<<<<<<<<<<<<<
 *             mask[i, j] = 1
 *     return
 */
    __pyx_t_5 = __pyx_v_full_nbins;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = __pyx_v_nbin; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_j = __pyx_t_7;

      /* "pyart/io/_sigmetfile.pyx":805
 *         nbin = nbins[i]
 *         for j in range(nbin, full_nbins):
 *             mask[i, j] = 1             # <<<<<<<<<<<<<<
 *     return
 * 
 */
      __pyx_t_4 = __pyx_v_i;
      __pyx_t_8 = __pyx_v_j;
      if (__pyx_t_4 < 0) __pyx_t_4 += __pyx_pybuffernd_mask.diminfo[0].shape;
      if (__pyx_t_8 < 0) __pyx_t_8 += __pyx_pybuffernd_mask.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_mask.diminfo[0].strides, __pyx_t_8, __pyx_pybuffernd_mask.diminfo[1].strides) = 1;
    }
  }

  /* "pyart/io/_sigmetfile.pyx":806
 *         for j in range(nbin, full_nbins):
 *             mask[i, j] = 1
 *     return             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":795
 * 
 * @cython.boundscheck(False)
 * cdef _mask_gates_not_collected(             # <<<<<<<<<<<<<<
 *         np.ndarray[np.uint8_t, ndim=2] mask,
 *         np.ndarray[np.int16_t, ndim=1] nbins):
 */

  /* function exit code */
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_nbins.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pyart.io._sigmetfile._mask_gates_not_collected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_nbins.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":809
 * 
 * 
 * def bin2_to_angle(bin2):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin2 encoded value (or array). """
 *     return 360. * bin2 / 65536
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_9bin2_to_angle(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_8bin2_to_angle, " Return an angle from Sigmet bin2 encoded value (or array). ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_9bin2_to_angle = {"bin2_to_angle", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_9bin2_to_angle, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_8bin2_to_angle};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_9bin2_to_angle(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_bin2 = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bin2_to_angle (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_bin2,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_bin2)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 809, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "bin2_to_angle") < 0)) __PYX_ERR(0, 809, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_bin2 = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bin2_to_angle", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 809, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile.bin2_to_angle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_8bin2_to_angle(__pyx_self, __pyx_v_bin2);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_8bin2_to_angle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bin2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bin2_to_angle", 1);

  /* "pyart/io/_sigmetfile.pyx":811
 * def bin2_to_angle(bin2):
 *     """ Return an angle from Sigmet bin2 encoded value (or array). """
 *     return 360. * bin2 / 65536             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_360_, __pyx_v_bin2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_TrueDivideObjC(__pyx_t_1, __pyx_int_65536, 0x10000, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":809
 * 
 * 
 * def bin2_to_angle(bin2):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin2 encoded value (or array). """
 *     return 360. * bin2 / 65536
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyart.io._sigmetfile.bin2_to_angle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":814
 * 
 * 
 * def bin4_to_angle(bin4):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin4 encoded value (or array). """
 *     return 360. * bin4 / 4294967296
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_11bin4_to_angle(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_10bin4_to_angle, " Return an angle from Sigmet bin4 encoded value (or array). ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_11bin4_to_angle = {"bin4_to_angle", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_11bin4_to_angle, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_10bin4_to_angle};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_11bin4_to_angle(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_bin4 = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bin4_to_angle (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_bin4,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_bin4)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 814, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "bin4_to_angle") < 0)) __PYX_ERR(0, 814, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_bin4 = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bin4_to_angle", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 814, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile.bin4_to_angle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10bin4_to_angle(__pyx_self, __pyx_v_bin4);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10bin4_to_angle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bin4) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bin4_to_angle", 1);

  /* "pyart/io/_sigmetfile.pyx":816
 * def bin4_to_angle(bin4):
 *     """ Return an angle from Sigmet bin4 encoded value (or array). """
 *     return 360. * bin4 / 4294967296             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_360_, __pyx_v_bin4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_int_4294967296); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":814
 * 
 * 
 * def bin4_to_angle(bin4):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin4 encoded value (or array). """
 *     return 360. * bin4 / 4294967296
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyart.io._sigmetfile.bin4_to_angle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":824
 * 
 * 
 * def _unpack_structure(string, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a structure """
 *     fmt = ''.join([i[1] for i in structure])
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_13_unpack_structure(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_12_unpack_structure, " Unpack a structure ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_13_unpack_structure = {"_unpack_structure", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_13_unpack_structure, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_12_unpack_structure};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_13_unpack_structure(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_string = 0;
  PyObject *__pyx_v_structure = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_unpack_structure (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_string,&__pyx_n_s_structure,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_string)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 824, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_structure)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 824, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_unpack_structure", 1, 2, 2, 1); __PYX_ERR(0, 824, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_unpack_structure") < 0)) __PYX_ERR(0, 824, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_string = values[0];
    __pyx_v_structure = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_unpack_structure", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 824, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_structure", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_12_unpack_structure(__pyx_self, __pyx_v_string, __pyx_v_structure);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_12_unpack_structure(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_string, PyObject *__pyx_v_structure) {
  PyObject *__pyx_v_fmt = NULL;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_8genexpr8__pyx_v_i = NULL;
  PyObject *__pyx_8genexpr9__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unpack_structure", 1);

  /* "pyart/io/_sigmetfile.pyx":826
 * def _unpack_structure(string, structure):
 *     """ Unpack a structure """
 *     fmt = ''.join([i[1] for i in structure])             # <<<<<<<<<<<<<<
 *     l = struct.unpack(fmt, string)
 *     return dict(zip([i[0] for i in structure], l))
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 826, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_v_structure)) || PyTuple_CheckExact(__pyx_v_structure)) {
      __pyx_t_2 = __pyx_v_structure; __Pyx_INCREF(__pyx_t_2);
      __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_structure); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 826, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 826, __pyx_L5_error)
    }
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 826, __pyx_L5_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 826, __pyx_L5_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 826, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 826, __pyx_L5_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 826, __pyx_L5_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 826, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_4(__pyx_t_2);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 826, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr8__pyx_v_i, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_8genexpr8__pyx_v_i, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 826, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 826, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_i); __pyx_8genexpr8__pyx_v_i = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_i); __pyx_8genexpr8__pyx_v_i = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_t_2 = PyUnicode_Join(__pyx_kp_u__15, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_fmt = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":827
 *     """ Unpack a structure """
 *     fmt = ''.join([i[1] for i in structure])
 *     l = struct.unpack(fmt, string)             # <<<<<<<<<<<<<<
 *     return dict(zip([i[0] for i in structure], l))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_struct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_fmt, __pyx_v_string};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_v_l = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":828
 *     fmt = ''.join([i[1] for i in structure])
 *     l = struct.unpack(fmt, string)
 *     return dict(zip([i[0] for i in structure], l))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L12_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_v_structure)) || PyTuple_CheckExact(__pyx_v_structure)) {
      __pyx_t_5 = __pyx_v_structure; __Pyx_INCREF(__pyx_t_5);
      __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_structure); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 828, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 828, __pyx_L12_error)
    }
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 828, __pyx_L12_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 828, __pyx_L12_error)
          #else
          __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 828, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 828, __pyx_L12_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 828, __pyx_L12_error)
          #else
          __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 828, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_4(__pyx_t_5);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 828, __pyx_L12_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr9__pyx_v_i, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_8genexpr9__pyx_v_i, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 828, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 828, __pyx_L12_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_i); __pyx_8genexpr9__pyx_v_i = 0;
    goto __pyx_L16_exit_scope;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_i); __pyx_8genexpr9__pyx_v_i = 0;
    goto __pyx_L1_error;
    __pyx_L16_exit_scope:;
  } /* exit inner scope */
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2)) __PYX_ERR(0, 828, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_l);
  __Pyx_GIVEREF(__pyx_v_l);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_l)) __PYX_ERR(0, 828, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":824
 * 
 * 
 * def _unpack_structure(string, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a structure """
 *     fmt = ''.join([i[1] for i in structure])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_structure", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fmt);
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_i);
  __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":831
 * 
 * 
 * def _unpack_key(dic, key, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a key. """
 *     dic[key] = _unpack_structure(dic[key], structure)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_15_unpack_key(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_14_unpack_key, " Unpack a key. ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_15_unpack_key = {"_unpack_key", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_15_unpack_key, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_14_unpack_key};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_15_unpack_key(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_dic = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_structure = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_unpack_key (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dic,&__pyx_n_s_key,&__pyx_n_s_structure,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dic)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 831, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_key)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 831, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_unpack_key", 1, 3, 3, 1); __PYX_ERR(0, 831, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_structure)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 831, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_unpack_key", 1, 3, 3, 2); __PYX_ERR(0, 831, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_unpack_key") < 0)) __PYX_ERR(0, 831, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_dic = values[0];
    __pyx_v_key = values[1];
    __pyx_v_structure = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_unpack_key", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 831, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_14_unpack_key(__pyx_self, __pyx_v_dic, __pyx_v_key, __pyx_v_structure);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_14_unpack_key(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_dic, PyObject *__pyx_v_key, PyObject *__pyx_v_structure) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unpack_key", 1);

  /* "pyart/io/_sigmetfile.pyx":833
 * def _unpack_key(dic, key, structure):
 *     """ Unpack a key. """
 *     dic[key] = _unpack_structure(dic[key], structure)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_structure); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_dic, __pyx_v_key); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_3, __pyx_v_structure};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 833, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (unlikely((PyObject_SetItem(__pyx_v_dic, __pyx_v_key, __pyx_t_1) < 0))) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":831
 * 
 * 
 * def _unpack_key(dic, key, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a key. """
 *     dic[key] = _unpack_structure(dic[key], structure)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":836
 * 
 * 
 * def _unpack_ingest_data_headers(record, ndata_types):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack one or more ingest_data_header from a record.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_17_unpack_ingest_data_headers(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_16_unpack_ingest_data_headers, "\n    Unpack one or more ingest_data_header from a record.\n\n    Returns a list of dictionaries or None when an error occurs.\n\n    ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_17_unpack_ingest_data_headers = {"_unpack_ingest_data_headers", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_17_unpack_ingest_data_headers, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_16_unpack_ingest_data_headers};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_17_unpack_ingest_data_headers(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_record = 0;
  PyObject *__pyx_v_ndata_types = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_unpack_ingest_data_headers (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_record,&__pyx_n_s_ndata_types,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_record)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 836, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_ndata_types)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 836, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_unpack_ingest_data_headers", 1, 2, 2, 1); __PYX_ERR(0, 836, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_unpack_ingest_data_headers") < 0)) __PYX_ERR(0, 836, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_record = values[0];
    __pyx_v_ndata_types = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_unpack_ingest_data_headers", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 836, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_ingest_data_headers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_16_unpack_ingest_data_headers(__pyx_self, __pyx_v_record, __pyx_v_ndata_types);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_16_unpack_ingest_data_headers(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record, PyObject *__pyx_v_ndata_types) {
  PyObject *__pyx_v_idh = NULL;
  PyObject *__pyx_9genexpr10__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unpack_ingest_data_headers", 1);

  /* "pyart/io/_sigmetfile.pyx":843
 * 
 *     """
 *     idh = [_unpack_ingest_data_header(record, i) for i in range(ndata_types)]             # <<<<<<<<<<<<<<
 *     if None in idh:
 *         return None
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 843, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_ndata_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 843, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 843, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 843, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 843, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 843, __pyx_L5_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 843, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 843, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 843, __pyx_L5_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 843, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 843, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr10__pyx_v_i, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_unpack_ingest_data_header); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 843, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_v_record, __pyx_9genexpr10__pyx_v_i};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 843, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 843, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_i); __pyx_9genexpr10__pyx_v_i = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_i); __pyx_9genexpr10__pyx_v_i = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_v_idh = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":844
 *     """
 *     idh = [_unpack_ingest_data_header(record, i) for i in range(ndata_types)]
 *     if None in idh:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
  __pyx_t_9 = (__Pyx_PySequence_ContainsTF(Py_None, __pyx_v_idh, Py_EQ)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 844, __pyx_L1_error)
  if (__pyx_t_9) {

    /* "pyart/io/_sigmetfile.pyx":845
 *     idh = [_unpack_ingest_data_header(record, i) for i in range(ndata_types)]
 *     if None in idh:
 *         return None             # <<<<<<<<<<<<<<
 *     else:
 *         return idh
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "pyart/io/_sigmetfile.pyx":844
 *     """
 *     idh = [_unpack_ingest_data_header(record, i) for i in range(ndata_types)]
 *     if None in idh:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
  }

  /* "pyart/io/_sigmetfile.pyx":847
 *         return None
 *     else:
 *         return idh             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_idh);
    __pyx_r = __pyx_v_idh;
    goto __pyx_L0;
  }

  /* "pyart/io/_sigmetfile.pyx":836
 * 
 * 
 * def _unpack_ingest_data_headers(record, ndata_types):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack one or more ingest_data_header from a record.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_ingest_data_headers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idh);
  __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":850
 * 
 * 
 * def _unpack_ingest_data_header(record, number):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack a single ingest_data_header from record.  Return None on error.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_19_unpack_ingest_data_header(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_18_unpack_ingest_data_header, "\n    Unpack a single ingest_data_header from record.  Return None on error.\n    ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_19_unpack_ingest_data_header = {"_unpack_ingest_data_header", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_19_unpack_ingest_data_header, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_18_unpack_ingest_data_header};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_19_unpack_ingest_data_header(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_record = 0;
  PyObject *__pyx_v_number = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_unpack_ingest_data_header (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_record,&__pyx_n_s_number,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_record)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 850, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_number)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 850, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_unpack_ingest_data_header", 1, 2, 2, 1); __PYX_ERR(0, 850, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_unpack_ingest_data_header") < 0)) __PYX_ERR(0, 850, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_record = values[0];
    __pyx_v_number = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_unpack_ingest_data_header", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 850, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_ingest_data_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_18_unpack_ingest_data_header(__pyx_self, __pyx_v_record, __pyx_v_number);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_18_unpack_ingest_data_header(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record, PyObject *__pyx_v_number) {
  PyObject *__pyx_v_offset = NULL;
  PyObject *__pyx_v_string = NULL;
  PyObject *__pyx_v_idh = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unpack_ingest_data_header", 1);

  /* "pyart/io/_sigmetfile.pyx":854
 *     Unpack a single ingest_data_header from record.  Return None on error.
 *     """
 *     offset = 12 + 76 * number             # <<<<<<<<<<<<<<
 *     string = record[offset:offset + 76]
 *     idh = _unpack_structure(string, INGEST_DATA_HEADER)
 */
  __pyx_t_1 = __Pyx_PyInt_MultiplyCObj(__pyx_int_76, __pyx_v_number, 76, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_AddCObj(__pyx_int_12, __pyx_t_1, 12, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_offset = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":855
 *     """
 *     offset = 12 + 76 * number
 *     string = record[offset:offset + 76]             # <<<<<<<<<<<<<<
 *     idh = _unpack_structure(string, INGEST_DATA_HEADER)
 *     _unpack_key(idh, 'structure_header', STRUCTURE_HEADER)
 */
  __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_offset, __pyx_int_76, 76, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_record, 0, 0, &__pyx_v_offset, &__pyx_t_2, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_string = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":856
 *     offset = 12 + 76 * number
 *     string = record[offset:offset + 76]
 *     idh = _unpack_structure(string, INGEST_DATA_HEADER)             # <<<<<<<<<<<<<<
 *     _unpack_key(idh, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(idh, 'sweep_start_time', YMDS_TIME)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_structure); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_INGEST_DATA_HEADER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_string, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 856, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_idh = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":857
 *     string = record[offset:offset + 76]
 *     idh = _unpack_structure(string, INGEST_DATA_HEADER)
 *     _unpack_key(idh, 'structure_header', STRUCTURE_HEADER)             # <<<<<<<<<<<<<<
 *     _unpack_key(idh, 'sweep_start_time', YMDS_TIME)
 *     if idh['structure_header']['structure_identifier'] != 24:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_STRUCTURE_HEADER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_v_idh, __pyx_n_u_structure_header, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 857, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":858
 *     idh = _unpack_structure(string, INGEST_DATA_HEADER)
 *     _unpack_key(idh, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(idh, 'sweep_start_time', YMDS_TIME)             # <<<<<<<<<<<<<<
 *     if idh['structure_header']['structure_identifier'] != 24:
 *         return None
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_YMDS_TIME); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_v_idh, __pyx_n_u_sweep_start_time, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 858, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":859
 *     _unpack_key(idh, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(idh, 'sweep_start_time', YMDS_TIME)
 *     if idh['structure_header']['structure_identifier'] != 24:             # <<<<<<<<<<<<<<
 *         return None
 *     return idh
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_idh, __pyx_n_u_structure_header); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_structure_identifier); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (__Pyx_PyInt_BoolNeObjC(__pyx_t_2, __pyx_int_24, 24, 0)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {

    /* "pyart/io/_sigmetfile.pyx":860
 *     _unpack_key(idh, 'sweep_start_time', YMDS_TIME)
 *     if idh['structure_header']['structure_identifier'] != 24:
 *         return None             # <<<<<<<<<<<<<<
 *     return idh
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "pyart/io/_sigmetfile.pyx":859
 *     _unpack_key(idh, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(idh, 'sweep_start_time', YMDS_TIME)
 *     if idh['structure_header']['structure_identifier'] != 24:             # <<<<<<<<<<<<<<
 *         return None
 *     return idh
 */
  }

  /* "pyart/io/_sigmetfile.pyx":861
 *     if idh['structure_header']['structure_identifier'] != 24:
 *         return None
 *     return idh             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_idh);
  __pyx_r = __pyx_v_idh;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":850
 * 
 * 
 * def _unpack_ingest_data_header(record, number):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack a single ingest_data_header from record.  Return None on error.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_ingest_data_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_offset);
  __Pyx_XDECREF(__pyx_v_string);
  __Pyx_XDECREF(__pyx_v_idh);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":864
 * 
 * 
 * def _unpack_raw_prod_bhdr(record):             # <<<<<<<<<<<<<<
 *     """ Return a dict with the unpacked raw_prod_bhdr from a record. """
 *     return _unpack_structure(record[:12], RAW_PROD_BHDR)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_21_unpack_raw_prod_bhdr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_20_unpack_raw_prod_bhdr, " Return a dict with the unpacked raw_prod_bhdr from a record. ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_21_unpack_raw_prod_bhdr = {"_unpack_raw_prod_bhdr", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_21_unpack_raw_prod_bhdr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_20_unpack_raw_prod_bhdr};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_21_unpack_raw_prod_bhdr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_record = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_unpack_raw_prod_bhdr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_record,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_record)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 864, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_unpack_raw_prod_bhdr") < 0)) __PYX_ERR(0, 864, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_record = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_unpack_raw_prod_bhdr", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 864, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_raw_prod_bhdr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_20_unpack_raw_prod_bhdr(__pyx_self, __pyx_v_record);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_20_unpack_raw_prod_bhdr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unpack_raw_prod_bhdr", 1);

  /* "pyart/io/_sigmetfile.pyx":866
 * def _unpack_raw_prod_bhdr(record):
 *     """ Return a dict with the unpacked raw_prod_bhdr from a record. """
 *     return _unpack_structure(record[:12], RAW_PROD_BHDR)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_structure); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_record, 0, 12, NULL, NULL, &__pyx_slice__16, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_RAW_PROD_BHDR); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":864
 * 
 * 
 * def _unpack_raw_prod_bhdr(record):             # <<<<<<<<<<<<<<
 *     """ Return a dict with the unpacked raw_prod_bhdr from a record. """
 *     return _unpack_structure(record[:12], RAW_PROD_BHDR)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_raw_prod_bhdr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":869
 * 
 * 
 * def _unpack_product_hdr(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked product_hdr from the first record.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_23_unpack_product_hdr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_22_unpack_product_hdr, "\n    Return a dict with the unpacked product_hdr from the first record.\n    ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_23_unpack_product_hdr = {"_unpack_product_hdr", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_23_unpack_product_hdr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_22_unpack_product_hdr};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_23_unpack_product_hdr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_record = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_unpack_product_hdr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_record,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_record)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 869, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_unpack_product_hdr") < 0)) __PYX_ERR(0, 869, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_record = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_unpack_product_hdr", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 869, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_product_hdr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_22_unpack_product_hdr(__pyx_self, __pyx_v_record);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_22_unpack_product_hdr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record) {
  PyObject *__pyx_v_product_hdr = NULL;
  PyObject *__pyx_v_product_config = NULL;
  PyObject *__pyx_v_product_end = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unpack_product_hdr", 1);

  /* "pyart/io/_sigmetfile.pyx":875
 * 
 *     # unpack the product_hdr structure from the first record
 *     product_hdr = _unpack_structure(record[:640], PRODUCT_HDR)             # <<<<<<<<<<<<<<
 * 
 *     # product_hdr substructure
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_structure); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_record, 0, 0x280, NULL, NULL, &__pyx_slice__17, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_PRODUCT_HDR); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 875, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_product_hdr = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":878
 * 
 *     # product_hdr substructure
 *     _unpack_key(product_hdr, 'structure_header', STRUCTURE_HEADER)             # <<<<<<<<<<<<<<
 *     _unpack_key(product_hdr, 'product_configuration',
 *                 PRODUCT_CONFIGURATION)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_STRUCTURE_HEADER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_product_hdr, __pyx_n_u_structure_header, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 878, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":879
 *     # product_hdr substructure
 *     _unpack_key(product_hdr, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(product_hdr, 'product_configuration',             # <<<<<<<<<<<<<<
 *                 PRODUCT_CONFIGURATION)
 *     _unpack_key(product_hdr, 'product_end', PRODUCT_END)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "pyart/io/_sigmetfile.pyx":880
 *     _unpack_key(product_hdr, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(product_hdr, 'product_configuration',
 *                 PRODUCT_CONFIGURATION)             # <<<<<<<<<<<<<<
 *     _unpack_key(product_hdr, 'product_end', PRODUCT_END)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_PRODUCT_CONFIGURATION); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_product_hdr, __pyx_n_u_product_configuration, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":881
 *     _unpack_key(product_hdr, 'product_configuration',
 *                 PRODUCT_CONFIGURATION)
 *     _unpack_key(product_hdr, 'product_end', PRODUCT_END)             # <<<<<<<<<<<<<<
 * 
 *     # product_config substructure
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_PRODUCT_END); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_product_hdr, __pyx_n_u_product_end, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 881, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":884
 * 
 *     # product_config substructure
 *     product_config = product_hdr['product_configuration']             # <<<<<<<<<<<<<<
 *     _unpack_key(product_config, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(product_config, 'generation_time', YMDS_TIME)
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_product_hdr, __pyx_n_u_product_configuration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_product_config = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":885
 *     # product_config substructure
 *     product_config = product_hdr['product_configuration']
 *     _unpack_key(product_config, 'structure_header', STRUCTURE_HEADER)             # <<<<<<<<<<<<<<
 *     _unpack_key(product_config, 'generation_time', YMDS_TIME)
 *     _unpack_key(product_config, 'sweep_ingest_time', YMDS_TIME)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_STRUCTURE_HEADER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_product_config, __pyx_n_u_structure_header, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":886
 *     product_config = product_hdr['product_configuration']
 *     _unpack_key(product_config, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(product_config, 'generation_time', YMDS_TIME)             # <<<<<<<<<<<<<<
 *     _unpack_key(product_config, 'sweep_ingest_time', YMDS_TIME)
 *     _unpack_key(product_config, 'file_ingest_time', YMDS_TIME)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_YMDS_TIME); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_product_config, __pyx_n_u_generation_time, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 886, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":887
 *     _unpack_key(product_config, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(product_config, 'generation_time', YMDS_TIME)
 *     _unpack_key(product_config, 'sweep_ingest_time', YMDS_TIME)             # <<<<<<<<<<<<<<
 *     _unpack_key(product_config, 'file_ingest_time', YMDS_TIME)
 *     _unpack_key(product_config, 'color_scale_def', COLOR_SCALE_DEF)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_YMDS_TIME); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_product_config, __pyx_n_u_sweep_ingest_time, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 887, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":888
 *     _unpack_key(product_config, 'generation_time', YMDS_TIME)
 *     _unpack_key(product_config, 'sweep_ingest_time', YMDS_TIME)
 *     _unpack_key(product_config, 'file_ingest_time', YMDS_TIME)             # <<<<<<<<<<<<<<
 *     _unpack_key(product_config, 'color_scale_def', COLOR_SCALE_DEF)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_YMDS_TIME); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_product_config, __pyx_n_u_file_ingest_time, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 888, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":889
 *     _unpack_key(product_config, 'sweep_ingest_time', YMDS_TIME)
 *     _unpack_key(product_config, 'file_ingest_time', YMDS_TIME)
 *     _unpack_key(product_config, 'color_scale_def', COLOR_SCALE_DEF)             # <<<<<<<<<<<<<<
 * 
 *     # product_end substructure
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_COLOR_SCALE_DEF); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_product_config, __pyx_n_u_color_scale_def, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":892
 * 
 *     # product_end substructure
 *     product_end = product_hdr['product_end']             # <<<<<<<<<<<<<<
 *     _unpack_key(product_end, 'ingest_time', YMDS_TIME)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_product_hdr, __pyx_n_u_product_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_product_end = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":893
 *     # product_end substructure
 *     product_end = product_hdr['product_end']
 *     _unpack_key(product_end, 'ingest_time', YMDS_TIME)             # <<<<<<<<<<<<<<
 * 
 *     return product_hdr
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_YMDS_TIME); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_product_end, __pyx_n_u_ingest_time, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 893, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":895
 *     _unpack_key(product_end, 'ingest_time', YMDS_TIME)
 * 
 *     return product_hdr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_product_hdr);
  __pyx_r = __pyx_v_product_hdr;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":869
 * 
 * 
 * def _unpack_product_hdr(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked product_hdr from the first record.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_product_hdr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_product_hdr);
  __Pyx_XDECREF(__pyx_v_product_config);
  __Pyx_XDECREF(__pyx_v_product_end);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":898
 * 
 * 
 * def _unpack_ingest_header(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked ingest_header from the second record.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_25_unpack_ingest_header(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5pyart_2io_11_sigmetfile_24_unpack_ingest_header, "\n    Return a dict with the unpacked ingest_header from the second record.\n    ");
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_25_unpack_ingest_header = {"_unpack_ingest_header", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_25_unpack_ingest_header, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_24_unpack_ingest_header};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_25_unpack_ingest_header(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_record = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_unpack_ingest_header (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_record,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_record)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 898, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_unpack_ingest_header") < 0)) __PYX_ERR(0, 898, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_record = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_unpack_ingest_header", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 898, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_ingest_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_24_unpack_ingest_header(__pyx_self, __pyx_v_record);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_24_unpack_ingest_header(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record) {
  PyObject *__pyx_v_ingest_header = NULL;
  PyObject *__pyx_v_ingest_configuration = NULL;
  PyObject *__pyx_v_task_configuration = NULL;
  PyObject *__pyx_v_task_dsp_info = NULL;
  PyObject *__pyx_v_task_end_info = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unpack_ingest_header", 1);

  /* "pyart/io/_sigmetfile.pyx":904
 * 
 *     # unpack the ingest_header structure from the second_record
 *     ingest_header = _unpack_structure(record[:4884], INGEST_HEADER)             # <<<<<<<<<<<<<<
 * 
 *     # ingest_header substructure
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_structure); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_record, 0, 0x1314, NULL, NULL, &__pyx_slice__18, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_INGEST_HEADER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 904, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_ingest_header = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":907
 * 
 *     # ingest_header substructure
 *     _unpack_key(ingest_header, 'structure_header', STRUCTURE_HEADER)             # <<<<<<<<<<<<<<
 *     _unpack_key(ingest_header, 'ingest_configuration',
 *                 INGEST_CONFIGURATION)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_STRUCTURE_HEADER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_ingest_header, __pyx_n_u_structure_header, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 907, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":908
 *     # ingest_header substructure
 *     _unpack_key(ingest_header, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(ingest_header, 'ingest_configuration',             # <<<<<<<<<<<<<<
 *                 INGEST_CONFIGURATION)
 *     _unpack_key(ingest_header, 'task_configuration', TASK_CONFIGURATION)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "pyart/io/_sigmetfile.pyx":909
 *     _unpack_key(ingest_header, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(ingest_header, 'ingest_configuration',
 *                 INGEST_CONFIGURATION)             # <<<<<<<<<<<<<<
 *     _unpack_key(ingest_header, 'task_configuration', TASK_CONFIGURATION)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_INGEST_CONFIGURATION); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_ingest_header, __pyx_n_u_ingest_configuration, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 908, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":910
 *     _unpack_key(ingest_header, 'ingest_configuration',
 *                 INGEST_CONFIGURATION)
 *     _unpack_key(ingest_header, 'task_configuration', TASK_CONFIGURATION)             # <<<<<<<<<<<<<<
 * 
 *     # ingest_configuration substructure
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_TASK_CONFIGURATION); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_ingest_header, __pyx_n_u_task_configuration, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":913
 * 
 *     # ingest_configuration substructure
 *     ingest_configuration = ingest_header['ingest_configuration']             # <<<<<<<<<<<<<<
 *     _unpack_key(ingest_configuration, 'volume_scan_start_time', YMDS_TIME)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_ingest_header, __pyx_n_u_ingest_configuration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ingest_configuration = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":914
 *     # ingest_configuration substructure
 *     ingest_configuration = ingest_header['ingest_configuration']
 *     _unpack_key(ingest_configuration, 'volume_scan_start_time', YMDS_TIME)             # <<<<<<<<<<<<<<
 * 
 *     # task_configuration substructure
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_YMDS_TIME); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_ingest_configuration, __pyx_n_u_volume_scan_start_time, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 914, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":917
 * 
 *     # task_configuration substructure
 *     task_configuration = ingest_header['task_configuration']             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(task_configuration, 'task_sched_info', TASK_SCHED_INFO)
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_ingest_header, __pyx_n_u_task_configuration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_task_configuration = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":918
 *     # task_configuration substructure
 *     task_configuration = ingest_header['task_configuration']
 *     _unpack_key(task_configuration, 'structure_header', STRUCTURE_HEADER)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'task_sched_info', TASK_SCHED_INFO)
 *     _unpack_key(task_configuration, 'task_dsp_info', TASK_DSP_INFO)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_STRUCTURE_HEADER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_task_configuration, __pyx_n_u_structure_header, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 918, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":919
 *     task_configuration = ingest_header['task_configuration']
 *     _unpack_key(task_configuration, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(task_configuration, 'task_sched_info', TASK_SCHED_INFO)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'task_dsp_info', TASK_DSP_INFO)
 *     _unpack_key(task_configuration, 'task_calib_info', TASK_CALIB_INFO)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_TASK_SCHED_INFO); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_task_configuration, __pyx_n_u_task_sched_info, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 919, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":920
 *     _unpack_key(task_configuration, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(task_configuration, 'task_sched_info', TASK_SCHED_INFO)
 *     _unpack_key(task_configuration, 'task_dsp_info', TASK_DSP_INFO)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'task_calib_info', TASK_CALIB_INFO)
 *     _unpack_key(task_configuration, 'task_range_info', TASK_RANGE_INFO)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_TASK_DSP_INFO); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_task_configuration, __pyx_n_u_task_dsp_info, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 920, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":921
 *     _unpack_key(task_configuration, 'task_sched_info', TASK_SCHED_INFO)
 *     _unpack_key(task_configuration, 'task_dsp_info', TASK_DSP_INFO)
 *     _unpack_key(task_configuration, 'task_calib_info', TASK_CALIB_INFO)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'task_range_info', TASK_RANGE_INFO)
 *     _unpack_key(task_configuration, 'task_scan_info', TASK_SCAN_INFO)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_TASK_CALIB_INFO); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_task_configuration, __pyx_n_u_task_calib_info, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 921, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":922
 *     _unpack_key(task_configuration, 'task_dsp_info', TASK_DSP_INFO)
 *     _unpack_key(task_configuration, 'task_calib_info', TASK_CALIB_INFO)
 *     _unpack_key(task_configuration, 'task_range_info', TASK_RANGE_INFO)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'task_scan_info', TASK_SCAN_INFO)
 *     _unpack_key(task_configuration, 'task_misc_info', TASK_MISC_INFO)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_TASK_RANGE_INFO); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_task_configuration, __pyx_n_u_task_range_info, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":923
 *     _unpack_key(task_configuration, 'task_calib_info', TASK_CALIB_INFO)
 *     _unpack_key(task_configuration, 'task_range_info', TASK_RANGE_INFO)
 *     _unpack_key(task_configuration, 'task_scan_info', TASK_SCAN_INFO)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'task_misc_info', TASK_MISC_INFO)
 *     _unpack_key(task_configuration, 'task_end_info', TASK_END_INFO)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 923, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_TASK_SCAN_INFO); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 923, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_task_configuration, __pyx_n_u_task_scan_info, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":924
 *     _unpack_key(task_configuration, 'task_range_info', TASK_RANGE_INFO)
 *     _unpack_key(task_configuration, 'task_scan_info', TASK_SCAN_INFO)
 *     _unpack_key(task_configuration, 'task_misc_info', TASK_MISC_INFO)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'task_end_info', TASK_END_INFO)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_TASK_MISC_INFO); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_task_configuration, __pyx_n_u_task_misc_info, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":925
 *     _unpack_key(task_configuration, 'task_scan_info', TASK_SCAN_INFO)
 *     _unpack_key(task_configuration, 'task_misc_info', TASK_MISC_INFO)
 *     _unpack_key(task_configuration, 'task_end_info', TASK_END_INFO)             # <<<<<<<<<<<<<<
 * 
 *     # task_dsp_info substructure
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_TASK_END_INFO); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_task_configuration, __pyx_n_u_task_end_info, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 925, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":928
 * 
 *     # task_dsp_info substructure
 *     task_dsp_info = task_configuration['task_dsp_info']             # <<<<<<<<<<<<<<
 *     _unpack_key(task_dsp_info, 'current_data_type_mask', DSP_DATA_MASK)
 *     _unpack_key(task_dsp_info, 'original_data_type_mask', DSP_DATA_MASK)
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_task_configuration, __pyx_n_u_task_dsp_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_task_dsp_info = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":929
 *     # task_dsp_info substructure
 *     task_dsp_info = task_configuration['task_dsp_info']
 *     _unpack_key(task_dsp_info, 'current_data_type_mask', DSP_DATA_MASK)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_dsp_info, 'original_data_type_mask', DSP_DATA_MASK)
 *     _unpack_key(task_dsp_info, 'task_dsp_mode', TASK_DSP_MODE_BATCH)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_DSP_DATA_MASK); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_task_dsp_info, __pyx_n_u_current_data_type_mask, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 929, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":930
 *     task_dsp_info = task_configuration['task_dsp_info']
 *     _unpack_key(task_dsp_info, 'current_data_type_mask', DSP_DATA_MASK)
 *     _unpack_key(task_dsp_info, 'original_data_type_mask', DSP_DATA_MASK)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_dsp_info, 'task_dsp_mode', TASK_DSP_MODE_BATCH)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_DSP_DATA_MASK); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_task_dsp_info, __pyx_n_u_original_data_type_mask, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 930, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":931
 *     _unpack_key(task_dsp_info, 'current_data_type_mask', DSP_DATA_MASK)
 *     _unpack_key(task_dsp_info, 'original_data_type_mask', DSP_DATA_MASK)
 *     _unpack_key(task_dsp_info, 'task_dsp_mode', TASK_DSP_MODE_BATCH)             # <<<<<<<<<<<<<<
 * 
 *     # task_scan_info substructure
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_TASK_DSP_MODE_BATCH); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_task_dsp_info, __pyx_n_u_task_dsp_mode, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":941
 * 
 *     # task_end_info substructure
 *     task_end_info = task_configuration['task_end_info']             # <<<<<<<<<<<<<<
 *     _unpack_key(task_end_info, 'task_data_time', YMDS_TIME)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_task_configuration, __pyx_n_u_task_end_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_task_end_info = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":942
 *     # task_end_info substructure
 *     task_end_info = task_configuration['task_end_info']
 *     _unpack_key(task_end_info, 'task_data_time', YMDS_TIME)             # <<<<<<<<<<<<<<
 * 
 *     return ingest_header
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_YMDS_TIME); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_task_end_info, __pyx_n_u_task_data_time, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":944
 *     _unpack_key(task_end_info, 'task_data_time', YMDS_TIME)
 * 
 *     return ingest_header             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ingest_header);
  __pyx_r = __pyx_v_ingest_header;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":898
 * 
 * 
 * def _unpack_ingest_header(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked ingest_header from the second record.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_ingest_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ingest_header);
  __Pyx_XDECREF(__pyx_v_ingest_configuration);
  __Pyx_XDECREF(__pyx_v_task_configuration);
  __Pyx_XDECREF(__pyx_v_task_dsp_info);
  __Pyx_XDECREF(__pyx_v_task_end_info);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile __pyx_vtable_5pyart_2io_11_sigmetfile_SigmetFile;

static PyObject *__pyx_tp_new_5pyart_2io_11_sigmetfile_SigmetFile(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)o);
  p->__pyx_vtab = __pyx_vtabptr_5pyart_2io_11_sigmetfile_SigmetFile;
  p->debug = Py_None; Py_INCREF(Py_None);
  p->product_hdr = Py_None; Py_INCREF(Py_None);
  p->ingest_header = Py_None; Py_INCREF(Py_None);
  p->ingest_data_headers = Py_None; Py_INCREF(Py_None);
  p->data_types = Py_None; Py_INCREF(Py_None);
  p->data_type_names = Py_None; Py_INCREF(Py_None);
  p->ndata_types = Py_None; Py_INCREF(Py_None);
  p->_fh = Py_None; Py_INCREF(Py_None);
  p->_raw_product_bhdrs = Py_None; Py_INCREF(Py_None);
  p->_rbuf = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5pyart_2io_11_sigmetfile_SigmetFile(PyObject *o) {
  struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *p = (struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5pyart_2io_11_sigmetfile_SigmetFile) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->debug);
  Py_CLEAR(p->product_hdr);
  Py_CLEAR(p->ingest_header);
  Py_CLEAR(p->ingest_data_headers);
  Py_CLEAR(p->data_types);
  Py_CLEAR(p->data_type_names);
  Py_CLEAR(p->ndata_types);
  Py_CLEAR(p->_fh);
  Py_CLEAR(p->_raw_product_bhdrs);
  Py_CLEAR(p->_rbuf);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_5pyart_2io_11_sigmetfile_SigmetFile(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *p = (struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)o;
  if (p->debug) {
    e = (*v)(p->debug, a); if (e) return e;
  }
  if (p->product_hdr) {
    e = (*v)(p->product_hdr, a); if (e) return e;
  }
  if (p->ingest_header) {
    e = (*v)(p->ingest_header, a); if (e) return e;
  }
  if (p->ingest_data_headers) {
    e = (*v)(p->ingest_data_headers, a); if (e) return e;
  }
  if (p->data_types) {
    e = (*v)(p->data_types, a); if (e) return e;
  }
  if (p->data_type_names) {
    e = (*v)(p->data_type_names, a); if (e) return e;
  }
  if (p->ndata_types) {
    e = (*v)(p->ndata_types, a); if (e) return e;
  }
  if (p->_fh) {
    e = (*v)(p->_fh, a); if (e) return e;
  }
  if (p->_raw_product_bhdrs) {
    e = (*v)(p->_raw_product_bhdrs, a); if (e) return e;
  }
  if (p->_rbuf) {
    e = (*v)(((PyObject *)p->_rbuf), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5pyart_2io_11_sigmetfile_SigmetFile(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *p = (struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)o;
  tmp = ((PyObject*)p->debug);
  p->debug = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->product_hdr);
  p->product_hdr = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->ingest_header);
  p->ingest_header = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->ingest_data_headers);
  p->ingest_data_headers = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->data_types);
  p->data_types = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->data_type_names);
  p->data_type_names = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->ndata_types);
  p->ndata_types = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_fh);
  p->_fh = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_raw_product_bhdrs);
  p->_raw_product_bhdrs = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_rbuf);
  p->_rbuf = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_debug(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_debug(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_product_hdr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_product_hdr(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_header(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_header(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_data_headers(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_data_headers(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_types(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_types(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_type_names(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_type_names(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_ndata_types(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_ndata_types(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__fh(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__fh(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__raw_product_bhdrs(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__raw_product_bhdrs(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__rbuf_pos(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__rbuf_pos(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__record_number(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__record_number(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_5pyart_2io_11_sigmetfile_SigmetFile[] = {
  {"_determine_data_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_determine_data_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_2_determine_data_types},
  {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_4close},
  {"read_data", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_7read_data, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_6read_data},
  {"_get_sweep", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_get_sweep, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_8_get_sweep},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5pyart_2io_11_sigmetfile_SigmetFile[] = {
  {(char *)"debug", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_debug, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_debug, (char *)0, 0},
  {(char *)"product_hdr", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_product_hdr, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_product_hdr, (char *)0, 0},
  {(char *)"ingest_header", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_header, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_header, (char *)0, 0},
  {(char *)"ingest_data_headers", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_data_headers, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_data_headers, (char *)0, 0},
  {(char *)"data_types", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_types, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_types, (char *)0, 0},
  {(char *)"data_type_names", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_type_names, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_type_names, (char *)0, 0},
  {(char *)"ndata_types", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_ndata_types, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_ndata_types, (char *)0, 0},
  {(char *)"_fh", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__fh, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__fh, (char *)0, 0},
  {(char *)"_raw_product_bhdrs", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__raw_product_bhdrs, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__raw_product_bhdrs, (char *)0, 0},
  {(char *)"_rbuf_pos", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__rbuf_pos, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__rbuf_pos, (char *)0, 0},
  {(char *)"_record_number", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__record_number, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__record_number, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5pyart_2io_11_sigmetfile_SigmetFile_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5pyart_2io_11_sigmetfile_SigmetFile},
  {Py_tp_doc, (void *)PyDoc_STR("\n    A class for accessing data from Sigmet (IRIS) product files.\n\n    Parameters\n    ----------\n    filename : str\n        Filename or file-like object.\n\n    Attributes\n    ----------\n    debug : bool\n        Set True to print out debugging information, False otherwise.\n    product_hdr : dict\n        Product_hdr structure.\n    ingest_header : dict\n        Ingest_header structure.\n    ingest_data_headers : list of dict\n        Ingest_data_header structures for each data type.  Indexed by the\n        data type name (str).  None when data has not yet been read.\n    data_types : list\n        List of data types (int) in the file.\n    data_type_names : list\n        List of data type names (stR) in the file.\n    ndata_types : int\n        Number of data types in the file.\n    _fh : file\n        Open file being read.\n    _raw_product_bhdrs : list\n        List of raw_product_bhdr structure dictionaries seperated by sweep.\n        None when data has not yet been read.\n\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5pyart_2io_11_sigmetfile_SigmetFile},
  {Py_tp_clear, (void *)__pyx_tp_clear_5pyart_2io_11_sigmetfile_SigmetFile},
  {Py_tp_methods, (void *)__pyx_methods_5pyart_2io_11_sigmetfile_SigmetFile},
  {Py_tp_getset, (void *)__pyx_getsets_5pyart_2io_11_sigmetfile_SigmetFile},
  {Py_tp_init, (void *)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5pyart_2io_11_sigmetfile_SigmetFile},
  {0, 0},
};
static PyType_Spec __pyx_type_5pyart_2io_11_sigmetfile_SigmetFile_spec = {
  "pyart.io._sigmetfile.SigmetFile",
  sizeof(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5pyart_2io_11_sigmetfile_SigmetFile_slots,
};
#else

static PyTypeObject __pyx_type_5pyart_2io_11_sigmetfile_SigmetFile = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyart.io._sigmetfile.""SigmetFile", /*tp_name*/
  sizeof(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5pyart_2io_11_sigmetfile_SigmetFile, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    A class for accessing data from Sigmet (IRIS) product files.\n\n    Parameters\n    ----------\n    filename : str\n        Filename or file-like object.\n\n    Attributes\n    ----------\n    debug : bool\n        Set True to print out debugging information, False otherwise.\n    product_hdr : dict\n        Product_hdr structure.\n    ingest_header : dict\n        Ingest_header structure.\n    ingest_data_headers : list of dict\n        Ingest_data_header structures for each data type.  Indexed by the\n        data type name (str).  None when data has not yet been read.\n    data_types : list\n        List of data types (int) in the file.\n    data_type_names : list\n        List of data type names (stR) in the file.\n    ndata_types : int\n        Number of data types in the file.\n    _fh : file\n        Open file being read.\n    _raw_product_bhdrs : list\n        List of raw_product_bhdr structure dictionaries seperated by sweep.\n        None when data has not yet been read.\n\n    "), /*tp_doc*/
  __pyx_tp_traverse_5pyart_2io_11_sigmetfile_SigmetFile, /*tp_traverse*/
  __pyx_tp_clear_5pyart_2io_11_sigmetfile_SigmetFile, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5pyart_2io_11_sigmetfile_SigmetFile, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5pyart_2io_11_sigmetfile_SigmetFile, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5pyart_2io_11_sigmetfile_SigmetFile, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_kp_u_112s, __pyx_k_112s, sizeof(__pyx_k_112s), 0, 1, 0, 0},
    {&__pyx_kp_u_115s, __pyx_k_115s, sizeof(__pyx_k_115s), 0, 1, 0, 0},
    {&__pyx_kp_u_120s, __pyx_k_120s, sizeof(__pyx_k_120s), 0, 1, 0, 0},
    {&__pyx_kp_u_128s, __pyx_k_128s, sizeof(__pyx_k_128s), 0, 1, 0, 0},
    {&__pyx_kp_u_12s, __pyx_k_12s, sizeof(__pyx_k_12s), 0, 1, 0, 0},
    {&__pyx_kp_u_136s, __pyx_k_136s, sizeof(__pyx_k_136s), 0, 1, 0, 0},
    {&__pyx_kp_u_160s, __pyx_k_160s, sizeof(__pyx_k_160s), 0, 1, 0, 0},
    {&__pyx_kp_u_16s, __pyx_k_16s, sizeof(__pyx_k_16s), 0, 1, 0, 0},
    {&__pyx_kp_u_184s, __pyx_k_184s, sizeof(__pyx_k_184s), 0, 1, 0, 0},
    {&__pyx_kp_u_18s, __pyx_k_18s, sizeof(__pyx_k_18s), 0, 1, 0, 0},
    {&__pyx_kp_u_198s, __pyx_k_198s, sizeof(__pyx_k_198s), 0, 1, 0, 0},
    {&__pyx_kp_u_1s, __pyx_k_1s, sizeof(__pyx_k_1s), 0, 1, 0, 0},
    {&__pyx_kp_u_2, __pyx_k_2, sizeof(__pyx_k_2), 0, 1, 0, 0},
    {&__pyx_kp_u_200s, __pyx_k_200s, sizeof(__pyx_k_200s), 0, 1, 0, 0},
    {&__pyx_kp_u_204s, __pyx_k_204s, sizeof(__pyx_k_204s), 0, 1, 0, 0},
    {&__pyx_kp_u_208s, __pyx_k_208s, sizeof(__pyx_k_208s), 0, 1, 0, 0},
    {&__pyx_kp_u_228s, __pyx_k_228s, sizeof(__pyx_k_228s), 0, 1, 0, 0},
    {&__pyx_kp_u_24s, __pyx_k_24s, sizeof(__pyx_k_24s), 0, 1, 0, 0},
    {&__pyx_kp_u_256s, __pyx_k_256s, sizeof(__pyx_k_256s), 0, 1, 0, 0},
    {&__pyx_kp_u_2612s, __pyx_k_2612s, sizeof(__pyx_k_2612s), 0, 1, 0, 0},
    {&__pyx_kp_u_28s, __pyx_k_28s, sizeof(__pyx_k_28s), 0, 1, 0, 0},
    {&__pyx_kp_u_2_uint8, __pyx_k_2_uint8, sizeof(__pyx_k_2_uint8), 0, 1, 0, 0},
    {&__pyx_kp_u_2s, __pyx_k_2s, sizeof(__pyx_k_2s), 0, 1, 0, 0},
    {&__pyx_kp_u_308s, __pyx_k_308s, sizeof(__pyx_k_308s), 0, 1, 0, 0},
    {&__pyx_kp_u_320s, __pyx_k_320s, sizeof(__pyx_k_320s), 0, 1, 0, 0},
    {&__pyx_kp_u_32s, __pyx_k_32s, sizeof(__pyx_k_32s), 0, 1, 0, 0},
    {&__pyx_kp_u_36s, __pyx_k_36s, sizeof(__pyx_k_36s), 0, 1, 0, 0},
    {&__pyx_kp_u_40s, __pyx_k_40s, sizeof(__pyx_k_40s), 0, 1, 0, 0},
    {&__pyx_kp_u_480s, __pyx_k_480s, sizeof(__pyx_k_480s), 0, 1, 0, 0},
    {&__pyx_kp_u_48s, __pyx_k_48s, sizeof(__pyx_k_48s), 0, 1, 0, 0},
    {&__pyx_kp_u_4s, __pyx_k_4s, sizeof(__pyx_k_4s), 0, 1, 0, 0},
    {&__pyx_kp_u_52s, __pyx_k_52s, sizeof(__pyx_k_52s), 0, 1, 0, 0},
    {&__pyx_kp_u_6s, __pyx_k_6s, sizeof(__pyx_k_6s), 0, 1, 0, 0},
    {&__pyx_kp_u_720s, __pyx_k_720s, sizeof(__pyx_k_720s), 0, 1, 0, 0},
    {&__pyx_kp_u_732s, __pyx_k_732s, sizeof(__pyx_k_732s), 0, 1, 0, 0},
    {&__pyx_kp_u_80s, __pyx_k_80s, sizeof(__pyx_k_80s), 0, 1, 0, 0},
    {&__pyx_kp_u_8s, __pyx_k_8s, sizeof(__pyx_k_8s), 0, 1, 0, 0},
    {&__pyx_kp_u_920s, __pyx_k_920s, sizeof(__pyx_k_920s), 0, 1, 0, 0},
    {&__pyx_kp_u_94s, __pyx_k_94s, sizeof(__pyx_k_94s), 0, 1, 0, 0},
    {&__pyx_n_u_ALBEDO16, __pyx_k_ALBEDO16, sizeof(__pyx_k_ALBEDO16), 0, 1, 0, 1},
    {&__pyx_n_u_ALBEDO8, __pyx_k_ALBEDO8, sizeof(__pyx_k_ALBEDO8), 0, 1, 0, 1},
    {&__pyx_n_u_AXDIL2, __pyx_k_AXDIL2, sizeof(__pyx_k_AXDIL2), 0, 1, 0, 1},
    {&__pyx_n_u_B, __pyx_k_B, sizeof(__pyx_k_B), 0, 1, 0, 1},
    {&__pyx_n_s_BIN1, __pyx_k_BIN1, sizeof(__pyx_k_BIN1), 0, 0, 1, 1},
    {&__pyx_n_s_BIN2, __pyx_k_BIN2, sizeof(__pyx_k_BIN2), 0, 0, 1, 1},
    {&__pyx_n_s_BIN4, __pyx_k_BIN4, sizeof(__pyx_k_BIN4), 0, 0, 1, 1},
    {&__pyx_n_s_COLOR_SCALE_DEF, __pyx_k_COLOR_SCALE_DEF, sizeof(__pyx_k_COLOR_SCALE_DEF), 0, 0, 1, 1},
    {&__pyx_n_u_CSP16, __pyx_k_CSP16, sizeof(__pyx_k_CSP16), 0, 1, 0, 1},
    {&__pyx_n_u_CSP8, __pyx_k_CSP8, sizeof(__pyx_k_CSP8), 0, 1, 0, 1},
    {&__pyx_n_u_DBT, __pyx_k_DBT, sizeof(__pyx_k_DBT), 0, 1, 0, 1},
    {&__pyx_n_u_DBT2, __pyx_k_DBT2, sizeof(__pyx_k_DBT2), 0, 1, 0, 1},
    {&__pyx_n_u_DBTE16, __pyx_k_DBTE16, sizeof(__pyx_k_DBTE16), 0, 1, 0, 1},
    {&__pyx_n_u_DBTE8, __pyx_k_DBTE8, sizeof(__pyx_k_DBTE8), 0, 1, 0, 1},
    {&__pyx_n_u_DBTV16, __pyx_k_DBTV16, sizeof(__pyx_k_DBTV16), 0, 1, 0, 1},
    {&__pyx_n_u_DBTV8, __pyx_k_DBTV8, sizeof(__pyx_k_DBTV8), 0, 1, 0, 1},
    {&__pyx_n_u_DBZ, __pyx_k_DBZ, sizeof(__pyx_k_DBZ), 0, 1, 0, 1},
    {&__pyx_n_u_DBZ2, __pyx_k_DBZ2, sizeof(__pyx_k_DBZ2), 0, 1, 0, 1},
    {&__pyx_n_u_DBZC, __pyx_k_DBZC, sizeof(__pyx_k_DBZC), 0, 1, 0, 1},
    {&__pyx_n_u_DBZC2, __pyx_k_DBZC2, sizeof(__pyx_k_DBZC2), 0, 1, 0, 1},
    {&__pyx_n_u_DBZE16, __pyx_k_DBZE16, sizeof(__pyx_k_DBZE16), 0, 1, 0, 1},
    {&__pyx_n_u_DBZE8, __pyx_k_DBZE8, sizeof(__pyx_k_DBZE8), 0, 1, 0, 1},
    {&__pyx_n_u_DBZV16, __pyx_k_DBZV16, sizeof(__pyx_k_DBZV16), 0, 1, 0, 1},
    {&__pyx_n_u_DBZV8, __pyx_k_DBZV8, sizeof(__pyx_k_DBZV8), 0, 1, 0, 1},
    {&__pyx_n_u_DEFORM2, __pyx_k_DEFORM2, sizeof(__pyx_k_DEFORM2), 0, 1, 0, 1},
    {&__pyx_n_u_DIVERGE2, __pyx_k_DIVERGE2, sizeof(__pyx_k_DIVERGE2), 0, 1, 0, 1},
    {&__pyx_n_s_DSP_DATA_MASK, __pyx_k_DSP_DATA_MASK, sizeof(__pyx_k_DSP_DATA_MASK), 0, 0, 1, 1},
    {&__pyx_n_u_FLIQUID2, __pyx_k_FLIQUID2, sizeof(__pyx_k_FLIQUID2), 0, 1, 0, 1},
    {&__pyx_n_s_FLT4, __pyx_k_FLT4, sizeof(__pyx_k_FLT4), 0, 0, 1, 1},
    {&__pyx_n_s_FLT8, __pyx_k_FLT8, sizeof(__pyx_k_FLT8), 0, 0, 1, 1},
    {&__pyx_kp_u_File_truncated_or_corrupt_i_of_i, __pyx_k_File_truncated_or_corrupt_i_of_i, sizeof(__pyx_k_File_truncated_or_corrupt_i_of_i), 0, 1, 0, 0},
    {&__pyx_kp_u_Finished_loading_record, __pyx_k_Finished_loading_record, sizeof(__pyx_k_Finished_loading_record), 0, 1, 0, 0},
    {&__pyx_n_u_GMT_minute_offset_local, __pyx_k_GMT_minute_offset_local, sizeof(__pyx_k_GMT_minute_offset_local), 0, 1, 0, 1},
    {&__pyx_n_u_GMT_minute_offset_standard, __pyx_k_GMT_minute_offset_standard, sizeof(__pyx_k_GMT_minute_offset_standard), 0, 1, 0, 1},
    {&__pyx_n_u_H, __pyx_k_H, sizeof(__pyx_k_H), 0, 1, 0, 1},
    {&__pyx_n_u_HCLASS, __pyx_k_HCLASS, sizeof(__pyx_k_HCLASS), 0, 1, 0, 1},
    {&__pyx_n_u_HCLASS2, __pyx_k_HCLASS2, sizeof(__pyx_k_HCLASS2), 0, 1, 0, 1},
    {&__pyx_n_u_HDIR2, __pyx_k_HDIR2, sizeof(__pyx_k_HDIR2), 0, 1, 0, 1},
    {&__pyx_n_u_HEIGHT, __pyx_k_HEIGHT, sizeof(__pyx_k_HEIGHT), 0, 1, 0, 1},
    {&__pyx_n_u_HVEL2, __pyx_k_HVEL2, sizeof(__pyx_k_HVEL2), 0, 1, 0, 1},
    {&__pyx_n_u_I, __pyx_k_I, sizeof(__pyx_k_I), 0, 1, 0, 1},
    {&__pyx_n_s_INGEST_CONFIGURATION, __pyx_k_INGEST_CONFIGURATION, sizeof(__pyx_k_INGEST_CONFIGURATION), 0, 0, 1, 1},
    {&__pyx_n_s_INGEST_DATA_HEADER, __pyx_k_INGEST_DATA_HEADER, sizeof(__pyx_k_INGEST_DATA_HEADER), 0, 0, 1, 1},
    {&__pyx_n_s_INGEST_HEADER, __pyx_k_INGEST_HEADER, sizeof(__pyx_k_INGEST_HEADER), 0, 0, 1, 1},
    {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
    {&__pyx_n_u_KDP, __pyx_k_KDP, sizeof(__pyx_k_KDP), 0, 1, 0, 1},
    {&__pyx_n_u_KDP2, __pyx_k_KDP2, sizeof(__pyx_k_KDP2), 0, 1, 0, 1},
    {&__pyx_n_u_LDRH, __pyx_k_LDRH, sizeof(__pyx_k_LDRH), 0, 1, 0, 1},
    {&__pyx_n_u_LDRH2, __pyx_k_LDRH2, sizeof(__pyx_k_LDRH2), 0, 1, 0, 1},
    {&__pyx_n_u_LDRV, __pyx_k_LDRV, sizeof(__pyx_k_LDRV), 0, 1, 0, 1},
    {&__pyx_n_u_LDRV2, __pyx_k_LDRV2, sizeof(__pyx_k_LDRV2), 0, 1, 0, 1},
    {&__pyx_n_u_LOG16, __pyx_k_LOG16, sizeof(__pyx_k_LOG16), 0, 1, 0, 1},
    {&__pyx_n_u_LOG8, __pyx_k_LOG8, sizeof(__pyx_k_LOG8), 0, 1, 0, 1},
    {&__pyx_n_s_MESSAGE, __pyx_k_MESSAGE, sizeof(__pyx_k_MESSAGE), 0, 0, 1, 1},
    {&__pyx_n_u_OTHER, __pyx_k_OTHER, sizeof(__pyx_k_OTHER), 0, 1, 0, 1},
    {&__pyx_n_u_PHIDP, __pyx_k_PHIDP, sizeof(__pyx_k_PHIDP), 0, 1, 0, 1},
    {&__pyx_n_u_PHIDP2, __pyx_k_PHIDP2, sizeof(__pyx_k_PHIDP2), 0, 1, 0, 1},
    {&__pyx_n_u_PHIH, __pyx_k_PHIH, sizeof(__pyx_k_PHIH), 0, 1, 0, 1},
    {&__pyx_n_u_PHIH2, __pyx_k_PHIH2, sizeof(__pyx_k_PHIH2), 0, 1, 0, 1},
    {&__pyx_n_u_PHIV, __pyx_k_PHIV, sizeof(__pyx_k_PHIV), 0, 1, 0, 1},
    {&__pyx_n_u_PHIV2, __pyx_k_PHIV2, sizeof(__pyx_k_PHIV2), 0, 1, 0, 1},
    {&__pyx_n_u_PMI16, __pyx_k_PMI16, sizeof(__pyx_k_PMI16), 0, 1, 0, 1},
    {&__pyx_n_u_PMI8, __pyx_k_PMI8, sizeof(__pyx_k_PMI8), 0, 1, 0, 1},
    {&__pyx_n_s_PRODUCT_CONFIGURATION, __pyx_k_PRODUCT_CONFIGURATION, sizeof(__pyx_k_PRODUCT_CONFIGURATION), 0, 0, 1, 1},
    {&__pyx_n_s_PRODUCT_END, __pyx_k_PRODUCT_END, sizeof(__pyx_k_PRODUCT_END), 0, 0, 1, 1},
    {&__pyx_n_s_PRODUCT_HDR, __pyx_k_PRODUCT_HDR, sizeof(__pyx_k_PRODUCT_HDR), 0, 0, 1, 1},
    {&__pyx_n_u_RAINRATE2, __pyx_k_RAINRATE2, sizeof(__pyx_k_RAINRATE2), 0, 1, 0, 1},
    {&__pyx_n_u_RAW, __pyx_k_RAW, sizeof(__pyx_k_RAW), 0, 1, 0, 1},
    {&__pyx_n_s_RAW_PROD_BHDR, __pyx_k_RAW_PROD_BHDR, sizeof(__pyx_k_RAW_PROD_BHDR), 0, 0, 1, 1},
    {&__pyx_n_s_RECORD_SIZE, __pyx_k_RECORD_SIZE, sizeof(__pyx_k_RECORD_SIZE), 0, 0, 1, 1},
    {&__pyx_n_u_RHOH, __pyx_k_RHOH, sizeof(__pyx_k_RHOH), 0, 1, 0, 1},
    {&__pyx_n_u_RHOH2, __pyx_k_RHOH2, sizeof(__pyx_k_RHOH2), 0, 1, 0, 1},
    {&__pyx_n_u_RHOHV, __pyx_k_RHOHV, sizeof(__pyx_k_RHOHV), 0, 1, 0, 1},
    {&__pyx_n_u_RHOHV2, __pyx_k_RHOHV2, sizeof(__pyx_k_RHOHV2), 0, 1, 0, 1},
    {&__pyx_n_u_RHOV, __pyx_k_RHOV, sizeof(__pyx_k_RHOV), 0, 1, 0, 1},
    {&__pyx_n_u_RHOV2, __pyx_k_RHOV2, sizeof(__pyx_k_RHOV2), 0, 1, 0, 1},
    {&__pyx_kp_u_Reading_ray_i_of_i, __pyx_k_Reading_ray_i_of_i, sizeof(__pyx_k_Reading_ray_i_of_i), 0, 1, 0, 0},
    {&__pyx_n_u_SHEAR, __pyx_k_SHEAR, sizeof(__pyx_k_SHEAR), 0, 1, 0, 1},
    {&__pyx_n_s_SIGMET_DATA_TYPES, __pyx_k_SIGMET_DATA_TYPES, sizeof(__pyx_k_SIGMET_DATA_TYPES), 0, 0, 1, 1},
    {&__pyx_n_s_SINT1, __pyx_k_SINT1, sizeof(__pyx_k_SINT1), 0, 0, 1, 1},
    {&__pyx_n_s_SINT2, __pyx_k_SINT2, sizeof(__pyx_k_SINT2), 0, 0, 1, 1},
    {&__pyx_n_s_SINT4, __pyx_k_SINT4, sizeof(__pyx_k_SINT4), 0, 0, 1, 1},
    {&__pyx_n_u_SNR16, __pyx_k_SNR16, sizeof(__pyx_k_SNR16), 0, 1, 0, 1},
    {&__pyx_n_u_SNR8, __pyx_k_SNR8, sizeof(__pyx_k_SNR8), 0, 1, 0, 1},
    {&__pyx_n_u_SQI, __pyx_k_SQI, sizeof(__pyx_k_SQI), 0, 1, 0, 1},
    {&__pyx_n_u_SQI2, __pyx_k_SQI2, sizeof(__pyx_k_SQI2), 0, 1, 0, 1},
    {&__pyx_n_s_STRUCTURE_HEADER, __pyx_k_STRUCTURE_HEADER, sizeof(__pyx_k_STRUCTURE_HEADER), 0, 0, 1, 1},
    {&__pyx_n_s_SigmetFile, __pyx_k_SigmetFile, sizeof(__pyx_k_SigmetFile), 0, 0, 1, 1},
    {&__pyx_n_s_SigmetFile___reduce_cython, __pyx_k_SigmetFile___reduce_cython, sizeof(__pyx_k_SigmetFile___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_SigmetFile___setstate_cython, __pyx_k_SigmetFile___setstate_cython, sizeof(__pyx_k_SigmetFile___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_SigmetFile__determine_data_types, __pyx_k_SigmetFile__determine_data_types, sizeof(__pyx_k_SigmetFile__determine_data_types), 0, 0, 1, 1},
    {&__pyx_n_s_SigmetFile__get_sweep, __pyx_k_SigmetFile__get_sweep, sizeof(__pyx_k_SigmetFile__get_sweep), 0, 0, 1, 1},
    {&__pyx_n_s_SigmetFile_close, __pyx_k_SigmetFile_close, sizeof(__pyx_k_SigmetFile_close), 0, 0, 1, 1},
    {&__pyx_n_s_SigmetFile_read_data, __pyx_k_SigmetFile_read_data, sizeof(__pyx_k_SigmetFile_read_data), 0, 0, 1, 1},
    {&__pyx_n_s_TASK_CALIB_INFO, __pyx_k_TASK_CALIB_INFO, sizeof(__pyx_k_TASK_CALIB_INFO), 0, 0, 1, 1},
    {&__pyx_n_s_TASK_CONFIGURATION, __pyx_k_TASK_CONFIGURATION, sizeof(__pyx_k_TASK_CONFIGURATION), 0, 0, 1, 1},
    {&__pyx_n_s_TASK_DSP_INFO, __pyx_k_TASK_DSP_INFO, sizeof(__pyx_k_TASK_DSP_INFO), 0, 0, 1, 1},
    {&__pyx_n_s_TASK_DSP_MODE_BATCH, __pyx_k_TASK_DSP_MODE_BATCH, sizeof(__pyx_k_TASK_DSP_MODE_BATCH), 0, 0, 1, 1},
    {&__pyx_n_s_TASK_END_INFO, __pyx_k_TASK_END_INFO, sizeof(__pyx_k_TASK_END_INFO), 0, 0, 1, 1},
    {&__pyx_n_s_TASK_FILE_SCAN_INFO, __pyx_k_TASK_FILE_SCAN_INFO, sizeof(__pyx_k_TASK_FILE_SCAN_INFO), 0, 0, 1, 1},
    {&__pyx_n_s_TASK_MANUAL_SCAN_INFO, __pyx_k_TASK_MANUAL_SCAN_INFO, sizeof(__pyx_k_TASK_MANUAL_SCAN_INFO), 0, 0, 1, 1},
    {&__pyx_n_s_TASK_MISC_INFO, __pyx_k_TASK_MISC_INFO, sizeof(__pyx_k_TASK_MISC_INFO), 0, 0, 1, 1},
    {&__pyx_n_s_TASK_PPI_SCAN_INFO, __pyx_k_TASK_PPI_SCAN_INFO, sizeof(__pyx_k_TASK_PPI_SCAN_INFO), 0, 0, 1, 1},
    {&__pyx_n_s_TASK_RANGE_INFO, __pyx_k_TASK_RANGE_INFO, sizeof(__pyx_k_TASK_RANGE_INFO), 0, 0, 1, 1},
    {&__pyx_n_s_TASK_RHI_SCAN_INFO, __pyx_k_TASK_RHI_SCAN_INFO, sizeof(__pyx_k_TASK_RHI_SCAN_INFO), 0, 0, 1, 1},
    {&__pyx_n_s_TASK_SCAN_INFO, __pyx_k_TASK_SCAN_INFO, sizeof(__pyx_k_TASK_SCAN_INFO), 0, 0, 1, 1},
    {&__pyx_n_s_TASK_SCHED_INFO, __pyx_k_TASK_SCHED_INFO, sizeof(__pyx_k_TASK_SCHED_INFO), 0, 0, 1, 1},
    {&__pyx_n_u_TEMPERATURE16, __pyx_k_TEMPERATURE16, sizeof(__pyx_k_TEMPERATURE16), 0, 1, 0, 1},
    {&__pyx_n_u_TIME2, __pyx_k_TIME2, sizeof(__pyx_k_TIME2), 0, 1, 0, 1},
    {&__pyx_n_u_TURB16, __pyx_k_TURB16, sizeof(__pyx_k_TURB16), 0, 1, 0, 1},
    {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
    {&__pyx_n_s_UINT1, __pyx_k_UINT1, sizeof(__pyx_k_UINT1), 0, 0, 1, 1},
    {&__pyx_n_s_UINT16_T, __pyx_k_UINT16_T, sizeof(__pyx_k_UINT16_T), 0, 0, 1, 1},
    {&__pyx_n_s_UINT2, __pyx_k_UINT2, sizeof(__pyx_k_UINT2), 0, 0, 1, 1},
    {&__pyx_n_s_UINT4, __pyx_k_UINT4, sizeof(__pyx_k_UINT4), 0, 0, 1, 1},
    {&__pyx_n_u_UNKNOWN_100, __pyx_k_UNKNOWN_100, sizeof(__pyx_k_UNKNOWN_100), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_101, __pyx_k_UNKNOWN_101, sizeof(__pyx_k_UNKNOWN_101), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_102, __pyx_k_UNKNOWN_102, sizeof(__pyx_k_UNKNOWN_102), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_103, __pyx_k_UNKNOWN_103, sizeof(__pyx_k_UNKNOWN_103), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_104, __pyx_k_UNKNOWN_104, sizeof(__pyx_k_UNKNOWN_104), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_105, __pyx_k_UNKNOWN_105, sizeof(__pyx_k_UNKNOWN_105), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_106, __pyx_k_UNKNOWN_106, sizeof(__pyx_k_UNKNOWN_106), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_107, __pyx_k_UNKNOWN_107, sizeof(__pyx_k_UNKNOWN_107), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_108, __pyx_k_UNKNOWN_108, sizeof(__pyx_k_UNKNOWN_108), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_109, __pyx_k_UNKNOWN_109, sizeof(__pyx_k_UNKNOWN_109), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_110, __pyx_k_UNKNOWN_110, sizeof(__pyx_k_UNKNOWN_110), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_112, __pyx_k_UNKNOWN_112, sizeof(__pyx_k_UNKNOWN_112), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_113, __pyx_k_UNKNOWN_113, sizeof(__pyx_k_UNKNOWN_113), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_114, __pyx_k_UNKNOWN_114, sizeof(__pyx_k_UNKNOWN_114), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_115, __pyx_k_UNKNOWN_115, sizeof(__pyx_k_UNKNOWN_115), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_116, __pyx_k_UNKNOWN_116, sizeof(__pyx_k_UNKNOWN_116), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_117, __pyx_k_UNKNOWN_117, sizeof(__pyx_k_UNKNOWN_117), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_118, __pyx_k_UNKNOWN_118, sizeof(__pyx_k_UNKNOWN_118), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_119, __pyx_k_UNKNOWN_119, sizeof(__pyx_k_UNKNOWN_119), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_120, __pyx_k_UNKNOWN_120, sizeof(__pyx_k_UNKNOWN_120), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_121, __pyx_k_UNKNOWN_121, sizeof(__pyx_k_UNKNOWN_121), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_122, __pyx_k_UNKNOWN_122, sizeof(__pyx_k_UNKNOWN_122), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_123, __pyx_k_UNKNOWN_123, sizeof(__pyx_k_UNKNOWN_123), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_124, __pyx_k_UNKNOWN_124, sizeof(__pyx_k_UNKNOWN_124), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_125, __pyx_k_UNKNOWN_125, sizeof(__pyx_k_UNKNOWN_125), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_126, __pyx_k_UNKNOWN_126, sizeof(__pyx_k_UNKNOWN_126), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_127, __pyx_k_UNKNOWN_127, sizeof(__pyx_k_UNKNOWN_127), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_29, __pyx_k_UNKNOWN_29, sizeof(__pyx_k_UNKNOWN_29), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_30, __pyx_k_UNKNOWN_30, sizeof(__pyx_k_UNKNOWN_30), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_31, __pyx_k_UNKNOWN_31, sizeof(__pyx_k_UNKNOWN_31), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_6, __pyx_k_UNKNOWN_6, sizeof(__pyx_k_UNKNOWN_6), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_81, __pyx_k_UNKNOWN_81, sizeof(__pyx_k_UNKNOWN_81), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_82, __pyx_k_UNKNOWN_82, sizeof(__pyx_k_UNKNOWN_82), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_83, __pyx_k_UNKNOWN_83, sizeof(__pyx_k_UNKNOWN_83), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_84, __pyx_k_UNKNOWN_84, sizeof(__pyx_k_UNKNOWN_84), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_85, __pyx_k_UNKNOWN_85, sizeof(__pyx_k_UNKNOWN_85), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_86, __pyx_k_UNKNOWN_86, sizeof(__pyx_k_UNKNOWN_86), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_87, __pyx_k_UNKNOWN_87, sizeof(__pyx_k_UNKNOWN_87), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_88, __pyx_k_UNKNOWN_88, sizeof(__pyx_k_UNKNOWN_88), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_89, __pyx_k_UNKNOWN_89, sizeof(__pyx_k_UNKNOWN_89), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_90, __pyx_k_UNKNOWN_90, sizeof(__pyx_k_UNKNOWN_90), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_91, __pyx_k_UNKNOWN_91, sizeof(__pyx_k_UNKNOWN_91), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_92, __pyx_k_UNKNOWN_92, sizeof(__pyx_k_UNKNOWN_92), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_93, __pyx_k_UNKNOWN_93, sizeof(__pyx_k_UNKNOWN_93), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_94, __pyx_k_UNKNOWN_94, sizeof(__pyx_k_UNKNOWN_94), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_95, __pyx_k_UNKNOWN_95, sizeof(__pyx_k_UNKNOWN_95), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_96, __pyx_k_UNKNOWN_96, sizeof(__pyx_k_UNKNOWN_96), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_97, __pyx_k_UNKNOWN_97, sizeof(__pyx_k_UNKNOWN_97), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_98, __pyx_k_UNKNOWN_98, sizeof(__pyx_k_UNKNOWN_98), 0, 1, 0, 1},
    {&__pyx_n_u_UNKNOWN_99, __pyx_k_UNKNOWN_99, sizeof(__pyx_k_UNKNOWN_99), 0, 1, 0, 1},
    {&__pyx_n_u_USER, __pyx_k_USER, sizeof(__pyx_k_USER), 0, 1, 0, 1},
    {&__pyx_n_u_USER2, __pyx_k_USER2, sizeof(__pyx_k_USER2), 0, 1, 0, 1},
    {&__pyx_kp_u_Unknown_type_s_returning_raw_dat, __pyx_k_Unknown_type_s_returning_raw_dat, sizeof(__pyx_k_Unknown_type_s_returning_raw_dat), 0, 1, 0, 0},
    {&__pyx_n_u_VEL, __pyx_k_VEL, sizeof(__pyx_k_VEL), 0, 1, 0, 1},
    {&__pyx_n_u_VEL2, __pyx_k_VEL2, sizeof(__pyx_k_VEL2), 0, 1, 0, 1},
    {&__pyx_n_u_VELC, __pyx_k_VELC, sizeof(__pyx_k_VELC), 0, 1, 0, 1},
    {&__pyx_n_u_VELC2, __pyx_k_VELC2, sizeof(__pyx_k_VELC2), 0, 1, 0, 1},
    {&__pyx_n_u_VIL2, __pyx_k_VIL2, sizeof(__pyx_k_VIL2), 0, 1, 0, 1},
    {&__pyx_n_u_VILD16, __pyx_k_VILD16, sizeof(__pyx_k_VILD16), 0, 1, 0, 1},
    {&__pyx_n_u_VIR16, __pyx_k_VIR16, sizeof(__pyx_k_VIR16), 0, 1, 0, 1},
    {&__pyx_n_u_VVEL2, __pyx_k_VVEL2, sizeof(__pyx_k_VVEL2), 0, 1, 0, 1},
    {&__pyx_n_u_WIDTH, __pyx_k_WIDTH, sizeof(__pyx_k_WIDTH), 0, 1, 0, 1},
    {&__pyx_n_u_WIDTH2, __pyx_k_WIDTH2, sizeof(__pyx_k_WIDTH2), 0, 1, 0, 1},
    {&__pyx_n_u_XHDR, __pyx_k_XHDR, sizeof(__pyx_k_XHDR), 0, 1, 0, 1},
    {&__pyx_n_s_YMDS_TIME, __pyx_k_YMDS_TIME, sizeof(__pyx_k_YMDS_TIME), 0, 0, 1, 1},
    {&__pyx_n_u_ZDR, __pyx_k_ZDR, sizeof(__pyx_k_ZDR), 0, 1, 0, 1},
    {&__pyx_n_u_ZDR2, __pyx_k_ZDR2, sizeof(__pyx_k_ZDR2), 0, 1, 0, 1},
    {&__pyx_n_u_ZDRC, __pyx_k_ZDRC, sizeof(__pyx_k_ZDRC), 0, 1, 0, 1},
    {&__pyx_n_u_ZDRC2, __pyx_k_ZDRC2, sizeof(__pyx_k_ZDRC2), 0, 1, 0, 1},
    {&__pyx_n_s__149, __pyx_k__149, sizeof(__pyx_k__149), 0, 0, 1, 1},
    {&__pyx_kp_u__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 1, 0, 0},
    {&__pyx_n_s__19, __pyx_k__19, sizeof(__pyx_k__19), 0, 0, 1, 1},
    {&__pyx_n_u_agc_feedback_code, __pyx_k_agc_feedback_code, sizeof(__pyx_k_agc_feedback_code), 0, 1, 0, 1},
    {&__pyx_n_u_altitude_radar, __pyx_k_altitude_radar, sizeof(__pyx_k_altitude_radar), 0, 1, 0, 1},
    {&__pyx_n_u_angular_resolution_desired, __pyx_k_angular_resolution_desired, sizeof(__pyx_k_angular_resolution_desired), 0, 1, 0, 1},
    {&__pyx_n_u_antenna_offset_bow, __pyx_k_antenna_offset_bow, sizeof(__pyx_k_antenna_offset_bow), 0, 1, 0, 1},
    {&__pyx_n_u_antenna_offset_starboard, __pyx_k_antenna_offset_starboard, sizeof(__pyx_k_antenna_offset_starboard), 0, 1, 0, 1},
    {&__pyx_n_u_antenna_offset_up, __pyx_k_antenna_offset_up, sizeof(__pyx_k_antenna_offset_up), 0, 1, 0, 1},
    {&__pyx_n_u_antenna_scan_mode, __pyx_k_antenna_scan_mode, sizeof(__pyx_k_antenna_scan_mode), 0, 1, 0, 1},
    {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_n_u_attenuation, __pyx_k_attenuation, sizeof(__pyx_k_attenuation), 0, 1, 0, 1},
    {&__pyx_n_s_az0, __pyx_k_az0, sizeof(__pyx_k_az0), 0, 0, 1, 1},
    {&__pyx_n_s_az1, __pyx_k_az1, sizeof(__pyx_k_az1), 0, 0, 1, 1},
    {&__pyx_n_u_azimuth_0, __pyx_k_azimuth_0, sizeof(__pyx_k_azimuth_0), 0, 1, 0, 1},
    {&__pyx_n_u_azimuth_1, __pyx_k_azimuth_1, sizeof(__pyx_k_azimuth_1), 0, 1, 0, 1},
    {&__pyx_n_u_azimuth_list, __pyx_k_azimuth_list, sizeof(__pyx_k_azimuth_list), 0, 1, 0, 1},
    {&__pyx_n_u_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 1, 0, 1},
    {&__pyx_n_s_bin2, __pyx_k_bin2, sizeof(__pyx_k_bin2), 0, 0, 1, 1},
    {&__pyx_n_s_bin2_to_angle, __pyx_k_bin2_to_angle, sizeof(__pyx_k_bin2_to_angle), 0, 0, 1, 1},
    {&__pyx_n_s_bin4, __pyx_k_bin4, sizeof(__pyx_k_bin4), 0, 0, 1, 1},
    {&__pyx_n_s_bin4_to_angle, __pyx_k_bin4_to_angle, sizeof(__pyx_k_bin4_to_angle), 0, 0, 1, 1},
    {&__pyx_n_s_bit, __pyx_k_bit, sizeof(__pyx_k_bit), 0, 0, 1, 1},
    {&__pyx_n_u_bit_per_bin, __pyx_k_bit_per_bin, sizeof(__pyx_k_bit_per_bin), 0, 1, 0, 1},
    {&__pyx_n_u_bool, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 1, 0, 1},
    {&__pyx_n_u_bytes_in_structure, __pyx_k_bytes_in_structure, sizeof(__pyx_k_bytes_in_structure), 0, 1, 0, 1},
    {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
    {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
    {&__pyx_n_u_clutter_correction_threshold, __pyx_k_clutter_correction_threshold, sizeof(__pyx_k_clutter_correction_threshold), 0, 1, 0, 1},
    {&__pyx_n_u_clutter_filter, __pyx_k_clutter_filter, sizeof(__pyx_k_clutter_filter), 0, 1, 0, 1},
    {&__pyx_n_u_clutter_filter_name, __pyx_k_clutter_filter_name, sizeof(__pyx_k_clutter_filter_name), 0, 1, 0, 1},
    {&__pyx_n_u_cluttermap, __pyx_k_cluttermap, sizeof(__pyx_k_cluttermap), 0, 1, 0, 1},
    {&__pyx_n_u_cluttermap_flag, __pyx_k_cluttermap_flag, sizeof(__pyx_k_cluttermap_flag), 0, 1, 0, 1},
    {&__pyx_n_u_color_scale_def, __pyx_k_color_scale_def, sizeof(__pyx_k_color_scale_def), 0, 1, 0, 1},
    {&__pyx_n_u_comment_bytes, __pyx_k_comment_bytes, sizeof(__pyx_k_comment_bytes), 0, 1, 0, 1},
    {&__pyx_n_u_comments, __pyx_k_comments, sizeof(__pyx_k_comments), 0, 1, 0, 1},
    {&__pyx_n_u_configuration_name, __pyx_k_configuration_name, sizeof(__pyx_k_configuration_name), 0, 1, 0, 1},
    {&__pyx_n_s_convert_sigmet_data, __pyx_k_convert_sigmet_data, sizeof(__pyx_k_convert_sigmet_data), 0, 0, 1, 1},
    {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
    {&__pyx_n_u_corrected_reflectivity_threshold, __pyx_k_corrected_reflectivity_threshold, sizeof(__pyx_k_corrected_reflectivity_threshold), 0, 1, 0, 1},
    {&__pyx_n_u_current_data_type_mask, __pyx_k_current_data_type_mask, sizeof(__pyx_k_current_data_type_mask), 0, 1, 0, 1},
    {&__pyx_n_u_custom_ray_header_name, __pyx_k_custom_ray_header_name, sizeof(__pyx_k_custom_ray_header_name), 0, 1, 0, 1},
    {&__pyx_n_u_customer_storage, __pyx_k_customer_storage, sizeof(__pyx_k_customer_storage), 0, 1, 0, 1},
    {&__pyx_n_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 1},
    {&__pyx_n_u_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 1, 0, 1},
    {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
    {&__pyx_n_s_data_type, __pyx_k_data_type, sizeof(__pyx_k_data_type), 0, 0, 1, 1},
    {&__pyx_n_u_data_type, __pyx_k_data_type, sizeof(__pyx_k_data_type), 0, 1, 0, 1},
    {&__pyx_n_s_data_type_name, __pyx_k_data_type_name, sizeof(__pyx_k_data_type_name), 0, 0, 1, 1},
    {&__pyx_n_s_data_types, __pyx_k_data_types, sizeof(__pyx_k_data_types), 0, 0, 1, 1},
    {&__pyx_n_s_data_types_from_mask, __pyx_k_data_types_from_mask, sizeof(__pyx_k_data_types_from_mask), 0, 0, 1, 1},
    {&__pyx_n_s_datetime, __pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 0, 1, 1},
    {&__pyx_n_u_day, __pyx_k_day, sizeof(__pyx_k_day), 0, 1, 0, 1},
    {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
    {&__pyx_n_s_determine_data_types, __pyx_k_determine_data_types, sizeof(__pyx_k_determine_data_types), 0, 0, 1, 1},
    {&__pyx_n_s_dic, __pyx_k_dic, sizeof(__pyx_k_dic), 0, 0, 1, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_n_u_dsp_type, __pyx_k_dsp_type, sizeof(__pyx_k_dsp_type), 0, 1, 0, 1},
    {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
    {&__pyx_n_u_dual_prf_delay, __pyx_k_dual_prf_delay, sizeof(__pyx_k_dual_prf_delay), 0, 1, 0, 1},
    {&__pyx_n_u_earth_radius, __pyx_k_earth_radius, sizeof(__pyx_k_earth_radius), 0, 1, 0, 1},
    {&__pyx_n_s_el0, __pyx_k_el0, sizeof(__pyx_k_el0), 0, 0, 1, 1},
    {&__pyx_n_s_el1, __pyx_k_el1, sizeof(__pyx_k_el1), 0, 0, 1, 1},
    {&__pyx_n_u_elevation_0, __pyx_k_elevation_0, sizeof(__pyx_k_elevation_0), 0, 1, 0, 1},
    {&__pyx_n_u_elevation_1, __pyx_k_elevation_1, sizeof(__pyx_k_elevation_1), 0, 1, 0, 1},
    {&__pyx_n_u_elevation_list, __pyx_k_elevation_list, sizeof(__pyx_k_elevation_list), 0, 1, 0, 1},
    {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
    {&__pyx_n_s_empty_like, __pyx_k_empty_like, sizeof(__pyx_k_empty_like), 0, 0, 1, 1},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
    {&__pyx_n_s_exp, __pyx_k_exp, sizeof(__pyx_k_exp), 0, 0, 1, 1},
    {&__pyx_n_u_extended_header_type, __pyx_k_extended_header_type, sizeof(__pyx_k_extended_header_type), 0, 1, 0, 1},
    {&__pyx_n_u_extended_product_header_offset, __pyx_k_extended_product_header_offset, sizeof(__pyx_k_extended_product_header_offset), 0, 1, 0, 1},
    {&__pyx_n_u_extended_ray_header_bytes, __pyx_k_extended_ray_header_bytes, sizeof(__pyx_k_extended_ray_header_bytes), 0, 1, 0, 1},
    {&__pyx_n_u_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 1, 0, 1},
    {&__pyx_n_u_fault_status, __pyx_k_fault_status, sizeof(__pyx_k_fault_status), 0, 1, 0, 1},
    {&__pyx_n_u_file_ingest_time, __pyx_k_file_ingest_time, sizeof(__pyx_k_file_ingest_time), 0, 1, 0, 1},
    {&__pyx_n_s_filename, __pyx_k_filename, sizeof(__pyx_k_filename), 0, 0, 1, 1},
    {&__pyx_n_u_filename, __pyx_k_filename, sizeof(__pyx_k_filename), 0, 1, 0, 1},
    {&__pyx_n_s_fill_value, __pyx_k_fill_value, sizeof(__pyx_k_fill_value), 0, 0, 1, 1},
    {&__pyx_n_u_first_azimuth, __pyx_k_first_azimuth, sizeof(__pyx_k_first_azimuth), 0, 1, 0, 1},
    {&__pyx_n_u_first_bin_range, __pyx_k_first_bin_range, sizeof(__pyx_k_first_bin_range), 0, 1, 0, 1},
    {&__pyx_n_u_first_elevation, __pyx_k_first_elevation, sizeof(__pyx_k_first_elevation), 0, 1, 0, 1},
    {&__pyx_n_u_first_ray_index, __pyx_k_first_ray_index, sizeof(__pyx_k_first_ray_index), 0, 1, 0, 1},
    {&__pyx_n_u_first_ray_offset, __pyx_k_first_ray_offset, sizeof(__pyx_k_first_ray_offset), 0, 1, 0, 1},
    {&__pyx_n_u_fixed_angle, __pyx_k_fixed_angle, sizeof(__pyx_k_fixed_angle), 0, 1, 0, 1},
    {&__pyx_n_u_flag, __pyx_k_flag, sizeof(__pyx_k_flag), 0, 1, 0, 1},
    {&__pyx_n_u_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 1, 0, 1},
    {&__pyx_n_u_flags2, __pyx_k_flags2, sizeof(__pyx_k_flags2), 0, 1, 0, 1},
    {&__pyx_n_u_float32, __pyx_k_float32, sizeof(__pyx_k_float32), 0, 1, 0, 1},
    {&__pyx_n_s_fmt, __pyx_k_fmt, sizeof(__pyx_k_fmt), 0, 0, 1, 1},
    {&__pyx_n_u_format_version, __pyx_k_format_version, sizeof(__pyx_k_format_version), 0, 1, 0, 1},
    {&__pyx_n_s_frombuffer, __pyx_k_frombuffer, sizeof(__pyx_k_frombuffer), 0, 0, 1, 1},
    {&__pyx_n_s_full_xhdr, __pyx_k_full_xhdr, sizeof(__pyx_k_full_xhdr), 0, 0, 1, 1},
    {&__pyx_n_u_gain_control_flag, __pyx_k_gain_control_flag, sizeof(__pyx_k_gain_control_flag), 0, 1, 0, 1},
    {&__pyx_n_u_gas_attenuation, __pyx_k_gas_attenuation, sizeof(__pyx_k_gas_attenuation), 0, 1, 0, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_u_generation_time, __pyx_k_generation_time, sizeof(__pyx_k_generation_time), 0, 1, 0, 1},
    {&__pyx_n_s_get_sweep, __pyx_k_get_sweep, sizeof(__pyx_k_get_sweep), 0, 0, 1, 1},
    {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
    {&__pyx_n_u_gmt_offset_minutes_local, __pyx_k_gmt_offset_minutes_local, sizeof(__pyx_k_gmt_offset_minutes_local), 0, 1, 0, 1},
    {&__pyx_n_u_gmt_offset_minutes_standard, __pyx_k_gmt_offset_minutes_standard, sizeof(__pyx_k_gmt_offset_minutes_standard), 0, 1, 0, 1},
    {&__pyx_n_u_gparam_bytes, __pyx_k_gparam_bytes, sizeof(__pyx_k_gparam_bytes), 0, 1, 0, 1},
    {&__pyx_n_u_gparm, __pyx_k_gparm, sizeof(__pyx_k_gparm), 0, 1, 0, 1},
    {&__pyx_n_u_ground_height, __pyx_k_ground_height, sizeof(__pyx_k_ground_height), 0, 1, 0, 1},
    {&__pyx_n_u_h, __pyx_k_h, sizeof(__pyx_k_h), 0, 1, 0, 1},
    {&__pyx_n_u_hardware_site, __pyx_k_hardware_site, sizeof(__pyx_k_hardware_site), 0, 1, 0, 1},
    {&__pyx_n_s_header_dic, __pyx_k_header_dic, sizeof(__pyx_k_header_dic), 0, 0, 1, 1},
    {&__pyx_n_u_height_radar, __pyx_k_height_radar, sizeof(__pyx_k_height_radar), 0, 1, 0, 1},
    {&__pyx_n_u_height_site, __pyx_k_height_site, sizeof(__pyx_k_height_site), 0, 1, 0, 1},
    {&__pyx_n_u_horizontal_beamwidth, __pyx_k_horizontal_beamwidth, sizeof(__pyx_k_horizontal_beamwidth), 0, 1, 0, 1},
    {&__pyx_n_u_horizontal_calibration_i0, __pyx_k_horizontal_calibration_i0, sizeof(__pyx_k_horizontal_calibration_i0), 0, 1, 0, 1},
    {&__pyx_n_u_horizontal_calibration_noise, __pyx_k_horizontal_calibration_noise, sizeof(__pyx_k_horizontal_calibration_noise), 0, 1, 0, 1},
    {&__pyx_n_u_horizontal_current_noise, __pyx_k_horizontal_current_noise, sizeof(__pyx_k_horizontal_current_noise), 0, 1, 0, 1},
    {&__pyx_n_u_horizontal_i0_calibration, __pyx_k_horizontal_i0_calibration, sizeof(__pyx_k_horizontal_i0_calibration), 0, 1, 0, 1},
    {&__pyx_n_u_horizontal_noise_calibration, __pyx_k_horizontal_noise_calibration, sizeof(__pyx_k_horizontal_noise_calibration), 0, 1, 0, 1},
    {&__pyx_n_u_horizontal_radar_constant, __pyx_k_horizontal_radar_constant, sizeof(__pyx_k_horizontal_radar_constant), 0, 1, 0, 1},
    {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
    {&__pyx_n_u_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 1, 0, 1},
    {&__pyx_n_u_i4, __pyx_k_i4, sizeof(__pyx_k_i4), 0, 1, 0, 1},
    {&__pyx_n_u_icolcnt, __pyx_k_icolcnt, sizeof(__pyx_k_icolcnt), 0, 1, 0, 1},
    {&__pyx_n_s_idh, __pyx_k_idh, sizeof(__pyx_k_idh), 0, 0, 1, 1},
    {&__pyx_n_u_iflags, __pyx_k_iflags, sizeof(__pyx_k_iflags), 0, 1, 0, 1},
    {&__pyx_n_u_ilevel_seams, __pyx_k_ilevel_seams, sizeof(__pyx_k_ilevel_seams), 0, 1, 0, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_s_ingest_configuration, __pyx_k_ingest_configuration, sizeof(__pyx_k_ingest_configuration), 0, 0, 1, 1},
    {&__pyx_n_u_ingest_configuration, __pyx_k_ingest_configuration, sizeof(__pyx_k_ingest_configuration), 0, 1, 0, 1},
    {&__pyx_n_s_ingest_data_hdrs, __pyx_k_ingest_data_hdrs, sizeof(__pyx_k_ingest_data_hdrs), 0, 0, 1, 1},
    {&__pyx_n_s_ingest_data_headers, __pyx_k_ingest_data_headers, sizeof(__pyx_k_ingest_data_headers), 0, 0, 1, 1},
    {&__pyx_n_u_ingest_hardware_name, __pyx_k_ingest_hardware_name, sizeof(__pyx_k_ingest_hardware_name), 0, 1, 0, 1},
    {&__pyx_n_s_ingest_header, __pyx_k_ingest_header, sizeof(__pyx_k_ingest_header), 0, 0, 1, 1},
    {&__pyx_n_u_ingest_iris_version, __pyx_k_ingest_iris_version, sizeof(__pyx_k_ingest_iris_version), 0, 1, 0, 1},
    {&__pyx_n_u_ingest_site_name, __pyx_k_ingest_site_name, sizeof(__pyx_k_ingest_site_name), 0, 1, 0, 1},
    {&__pyx_n_u_ingest_time, __pyx_k_ingest_time, sizeof(__pyx_k_ingest_time), 0, 1, 0, 1},
    {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
    {&__pyx_n_u_input_data_type, __pyx_k_input_data_type, sizeof(__pyx_k_input_data_type), 0, 1, 0, 1},
    {&__pyx_n_u_input_mask, __pyx_k_input_mask, sizeof(__pyx_k_input_mask), 0, 1, 0, 1},
    {&__pyx_n_u_int16, __pyx_k_int16, sizeof(__pyx_k_int16), 0, 1, 0, 1},
    {&__pyx_n_u_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 1, 0, 1},
    {&__pyx_n_u_inverse_flatting, __pyx_k_inverse_flatting, sizeof(__pyx_k_inverse_flatting), 0, 1, 0, 1},
    {&__pyx_n_u_iris_version, __pyx_k_iris_version, sizeof(__pyx_k_iris_version), 0, 1, 0, 1},
    {&__pyx_n_u_iris_version_created, __pyx_k_iris_version_created, sizeof(__pyx_k_iris_version_created), 0, 1, 0, 1},
    {&__pyx_n_s_is_bit_set, __pyx_k_is_bit_set, sizeof(__pyx_k_is_bit_set), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_u_iset_and_scale, __pyx_k_iset_and_scale, sizeof(__pyx_k_iset_and_scale), 0, 1, 0, 1},
    {&__pyx_n_u_istart, __pyx_k_istart, sizeof(__pyx_k_istart), 0, 1, 0, 1},
    {&__pyx_n_u_istep, __pyx_k_istep, sizeof(__pyx_k_istep), 0, 1, 0, 1},
    {&__pyx_n_s_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 0, 1, 1},
    {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
    {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
    {&__pyx_n_s_l, __pyx_k_l, sizeof(__pyx_k_l), 0, 0, 1, 1},
    {&__pyx_n_u_last_bin_range, __pyx_k_last_bin_range, sizeof(__pyx_k_last_bin_range), 0, 1, 0, 1},
    {&__pyx_n_u_last_run_day, __pyx_k_last_run_day, sizeof(__pyx_k_last_run_day), 0, 1, 0, 1},
    {&__pyx_n_u_last_run_time, __pyx_k_last_run_time, sizeof(__pyx_k_last_run_time), 0, 1, 0, 1},
    {&__pyx_n_u_latitude, __pyx_k_latitude, sizeof(__pyx_k_latitude), 0, 1, 0, 1},
    {&__pyx_n_u_latitude_projection, __pyx_k_latitude_projection, sizeof(__pyx_k_latitude_projection), 0, 1, 0, 1},
    {&__pyx_n_u_latitude_radar, __pyx_k_latitude_radar, sizeof(__pyx_k_latitude_radar), 0, 1, 0, 1},
    {&__pyx_n_u_ldr_bias, __pyx_k_ldr_bias, sizeof(__pyx_k_ldr_bias), 0, 1, 0, 1},
    {&__pyx_n_u_ldr_offset, __pyx_k_ldr_offset, sizeof(__pyx_k_ldr_offset), 0, 1, 0, 1},
    {&__pyx_n_s_lead_record, __pyx_k_lead_record, sizeof(__pyx_k_lead_record), 0, 0, 1, 1},
    {&__pyx_n_u_left_azimuth_limit, __pyx_k_left_azimuth_limit, sizeof(__pyx_k_left_azimuth_limit), 0, 1, 0, 1},
    {&__pyx_n_s_like_dbt, __pyx_k_like_dbt, sizeof(__pyx_k_like_dbt), 0, 0, 1, 1},
    {&__pyx_n_s_like_dbt2, __pyx_k_like_dbt2, sizeof(__pyx_k_like_dbt2), 0, 0, 1, 1},
    {&__pyx_n_s_like_sqi, __pyx_k_like_sqi, sizeof(__pyx_k_like_sqi), 0, 0, 1, 1},
    {&__pyx_n_s_like_sqi2, __pyx_k_like_sqi2, sizeof(__pyx_k_like_sqi2), 0, 0, 1, 1},
    {&__pyx_n_u_linear_filter_first_bin, __pyx_k_linear_filter_first_bin, sizeof(__pyx_k_linear_filter_first_bin), 0, 1, 0, 1},
    {&__pyx_n_u_local_timezone, __pyx_k_local_timezone, sizeof(__pyx_k_local_timezone), 0, 1, 0, 1},
    {&__pyx_n_u_log_filter_first_bin, __pyx_k_log_filter_first_bin, sizeof(__pyx_k_log_filter_first_bin), 0, 1, 0, 1},
    {&__pyx_n_u_longitude, __pyx_k_longitude, sizeof(__pyx_k_longitude), 0, 1, 0, 1},
    {&__pyx_n_u_longitude_projection, __pyx_k_longitude_projection, sizeof(__pyx_k_longitude_projection), 0, 1, 0, 1},
    {&__pyx_n_u_longitude_radar, __pyx_k_longitude_radar, sizeof(__pyx_k_longitude_radar), 0, 1, 0, 1},
    {&__pyx_n_u_low_prf_factional, __pyx_k_low_prf_factional, sizeof(__pyx_k_low_prf_factional), 0, 1, 0, 1},
    {&__pyx_n_u_low_prf_hz, __pyx_k_low_prf_hz, sizeof(__pyx_k_low_prf_hz), 0, 1, 0, 1},
    {&__pyx_n_u_low_prf_range_averaging, __pyx_k_low_prf_range_averaging, sizeof(__pyx_k_low_prf_range_averaging), 0, 1, 0, 1},
    {&__pyx_n_u_low_prf_sample_size, __pyx_k_low_prf_sample_size, sizeof(__pyx_k_low_prf_sample_size), 0, 1, 0, 1},
    {&__pyx_n_u_lower_elevation_limit, __pyx_k_lower_elevation_limit, sizeof(__pyx_k_lower_elevation_limit), 0, 1, 0, 1},
    {&__pyx_n_s_ma, __pyx_k_ma, sizeof(__pyx_k_ma), 0, 0, 1, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_u_major_mode, __pyx_k_major_mode, sizeof(__pyx_k_major_mode), 0, 1, 0, 1},
    {&__pyx_n_s_mask, __pyx_k_mask, sizeof(__pyx_k_mask), 0, 0, 1, 1},
    {&__pyx_n_u_mask_word_0, __pyx_k_mask_word_0, sizeof(__pyx_k_mask_word_0), 0, 1, 0, 1},
    {&__pyx_n_u_mask_word_1, __pyx_k_mask_word_1, sizeof(__pyx_k_mask_word_1), 0, 1, 0, 1},
    {&__pyx_n_u_mask_word_2, __pyx_k_mask_word_2, sizeof(__pyx_k_mask_word_2), 0, 1, 0, 1},
    {&__pyx_n_u_mask_word_3, __pyx_k_mask_word_3, sizeof(__pyx_k_mask_word_3), 0, 1, 0, 1},
    {&__pyx_n_u_mask_word_4, __pyx_k_mask_word_4, sizeof(__pyx_k_mask_word_4), 0, 1, 0, 1},
    {&__pyx_n_s_masked_array, __pyx_k_masked_array, sizeof(__pyx_k_masked_array), 0, 0, 1, 1},
    {&__pyx_n_u_maximum_range, __pyx_k_maximum_range, sizeof(__pyx_k_maximum_range), 0, 1, 0, 1},
    {&__pyx_n_u_mean_wind_direction, __pyx_k_mean_wind_direction, sizeof(__pyx_k_mean_wind_direction), 0, 1, 0, 1},
    {&__pyx_n_u_mean_wind_speed, __pyx_k_mean_wind_speed, sizeof(__pyx_k_mean_wind_speed), 0, 1, 0, 1},
    {&__pyx_n_u_melting_layer, __pyx_k_melting_layer, sizeof(__pyx_k_melting_layer), 0, 1, 0, 1},
    {&__pyx_n_u_melting_level, __pyx_k_melting_level, sizeof(__pyx_k_melting_level), 0, 1, 0, 1},
    {&__pyx_n_s_mess, __pyx_k_mess, sizeof(__pyx_k_mess), 0, 0, 1, 1},
    {&__pyx_n_s_metadata, __pyx_k_metadata, sizeof(__pyx_k_metadata), 0, 0, 1, 1},
    {&__pyx_n_u_milliseconds, __pyx_k_milliseconds, sizeof(__pyx_k_milliseconds), 0, 1, 0, 1},
    {&__pyx_n_u_minor_task_suffix, __pyx_k_minor_task_suffix, sizeof(__pyx_k_minor_task_suffix), 0, 1, 0, 1},
    {&__pyx_n_s_mod, __pyx_k_mod, sizeof(__pyx_k_mod), 0, 0, 1, 1},
    {&__pyx_n_u_month, __pyx_k_month, sizeof(__pyx_k_month), 0, 1, 0, 1},
    {&__pyx_n_s_multi_prf_flag, __pyx_k_multi_prf_flag, sizeof(__pyx_k_multi_prf_flag), 0, 0, 1, 1},
    {&__pyx_n_u_multi_prf_flag, __pyx_k_multi_prf_flag, sizeof(__pyx_k_multi_prf_flag), 0, 1, 0, 1},
    {&__pyx_n_s_multiplier, __pyx_k_multiplier, sizeof(__pyx_k_multiplier), 0, 0, 1, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
    {&__pyx_n_s_nbin, __pyx_k_nbin, sizeof(__pyx_k_nbin), 0, 0, 1, 1},
    {&__pyx_n_s_nbins, __pyx_k_nbins, sizeof(__pyx_k_nbins), 0, 0, 1, 1},
    {&__pyx_n_u_nbins, __pyx_k_nbins, sizeof(__pyx_k_nbins), 0, 1, 0, 1},
    {&__pyx_n_s_ndata, __pyx_k_ndata, sizeof(__pyx_k_ndata), 0, 0, 1, 1},
    {&__pyx_n_s_ndata_types, __pyx_k_ndata_types, sizeof(__pyx_k_ndata_types), 0, 0, 1, 1},
    {&__pyx_n_u_nexrad_clutter_skip, __pyx_k_nexrad_clutter_skip, sizeof(__pyx_k_nexrad_clutter_skip), 0, 1, 0, 1},
    {&__pyx_n_u_nexrad_clutter_threshold, __pyx_k_nexrad_clutter_threshold, sizeof(__pyx_k_nexrad_clutter_threshold), 0, 1, 0, 1},
    {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
    {&__pyx_n_s_nray_data_types, __pyx_k_nray_data_types, sizeof(__pyx_k_nray_data_types), 0, 0, 1, 1},
    {&__pyx_n_s_nrays, __pyx_k_nrays, sizeof(__pyx_k_nrays), 0, 0, 1, 1},
    {&__pyx_n_s_nsweeps, __pyx_k_nsweeps, sizeof(__pyx_k_nsweeps), 0, 0, 1, 1},
    {&__pyx_n_s_number, __pyx_k_number, sizeof(__pyx_k_number), 0, 0, 1, 1},
    {&__pyx_n_u_number_bins, __pyx_k_number_bins, sizeof(__pyx_k_number_bins), 0, 1, 0, 1},
    {&__pyx_n_u_number_elements, __pyx_k_number_elements, sizeof(__pyx_k_number_elements), 0, 1, 0, 1},
    {&__pyx_n_u_number_files, __pyx_k_number_files, sizeof(__pyx_k_number_files), 0, 1, 0, 1},
    {&__pyx_n_u_number_ingest, __pyx_k_number_ingest, sizeof(__pyx_k_number_ingest), 0, 1, 0, 1},
    {&__pyx_n_u_number_input_bins, __pyx_k_number_input_bins, sizeof(__pyx_k_number_input_bins), 0, 1, 0, 1},
    {&__pyx_n_u_number_linear_filter, __pyx_k_number_linear_filter, sizeof(__pyx_k_number_linear_filter), 0, 1, 0, 1},
    {&__pyx_n_u_number_log_filter, __pyx_k_number_log_filter, sizeof(__pyx_k_number_log_filter), 0, 1, 0, 1},
    {&__pyx_n_u_number_output_bins, __pyx_k_number_output_bins, sizeof(__pyx_k_number_output_bins), 0, 1, 0, 1},
    {&__pyx_n_u_number_rays_file_actual, __pyx_k_number_rays_file_actual, sizeof(__pyx_k_number_rays_file_actual), 0, 1, 0, 1},
    {&__pyx_n_u_number_rays_file_expected, __pyx_k_number_rays_file_expected, sizeof(__pyx_k_number_rays_file_expected), 0, 1, 0, 1},
    {&__pyx_n_u_number_rays_sweep, __pyx_k_number_rays_sweep, sizeof(__pyx_k_number_rays_sweep), 0, 1, 0, 1},
    {&__pyx_n_u_number_sweeps, __pyx_k_number_sweeps, sizeof(__pyx_k_number_sweeps), 0, 1, 0, 1},
    {&__pyx_n_u_number_sweeps_completed, __pyx_k_number_sweeps_completed, sizeof(__pyx_k_number_sweeps_completed), 0, 1, 0, 1},
    {&__pyx_n_u_number_task_config_table, __pyx_k_number_task_config_table, sizeof(__pyx_k_number_task_config_table), 0, 1, 0, 1},
    {&__pyx_n_u_number_tasks, __pyx_k_number_tasks, sizeof(__pyx_k_number_tasks), 0, 1, 0, 1},
    {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
    {&__pyx_kp_u_numpy__core_multiarray_failed_to, __pyx_k_numpy__core_multiarray_failed_to, sizeof(__pyx_k_numpy__core_multiarray_failed_to), 0, 1, 0, 0},
    {&__pyx_kp_u_numpy__core_umath_failed_to_impo, __pyx_k_numpy__core_umath_failed_to_impo, sizeof(__pyx_k_numpy__core_umath_failed_to_impo), 0, 1, 0, 0},
    {&__pyx_n_s_nyquist, __pyx_k_nyquist, sizeof(__pyx_k_nyquist), 0, 0, 1, 1},
    {&__pyx_n_s_offset, __pyx_k_offset, sizeof(__pyx_k_offset), 0, 0, 1, 1},
    {&__pyx_n_s_ones, __pyx_k_ones, sizeof(__pyx_k_ones), 0, 0, 1, 1},
    {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
    {&__pyx_n_u_original_data_type_mask, __pyx_k_original_data_type_mask, sizeof(__pyx_k_original_data_type_mask), 0, 1, 0, 1},
    {&__pyx_n_s_out, __pyx_k_out, sizeof(__pyx_k_out), 0, 0, 1, 1},
    {&__pyx_n_s_parse_ray_headers, __pyx_k_parse_ray_headers, sizeof(__pyx_k_parse_ray_headers), 0, 0, 1, 1},
    {&__pyx_n_u_playback_flag, __pyx_k_playback_flag, sizeof(__pyx_k_playback_flag), 0, 1, 0, 1},
    {&__pyx_n_u_playback_version, __pyx_k_playback_version, sizeof(__pyx_k_playback_version), 0, 1, 0, 1},
    {&__pyx_n_u_polarization, __pyx_k_polarization, sizeof(__pyx_k_polarization), 0, 1, 0, 1},
    {&__pyx_n_u_polarization_type, __pyx_k_polarization_type, sizeof(__pyx_k_polarization_type), 0, 1, 0, 1},
    {&__pyx_n_s_power, __pyx_k_power, sizeof(__pyx_k_power), 0, 0, 1, 1},
    {&__pyx_n_u_power_threshold, __pyx_k_power_threshold, sizeof(__pyx_k_power_threshold), 0, 1, 0, 1},
    {&__pyx_n_u_prf, __pyx_k_prf, sizeof(__pyx_k_prf), 0, 1, 0, 1},
    {&__pyx_n_s_prf_flag, __pyx_k_prf_flag, sizeof(__pyx_k_prf_flag), 0, 0, 1, 1},
    {&__pyx_n_u_prf_flag, __pyx_k_prf_flag, sizeof(__pyx_k_prf_flag), 0, 1, 0, 1},
    {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
    {&__pyx_n_s_product_config, __pyx_k_product_config, sizeof(__pyx_k_product_config), 0, 0, 1, 1},
    {&__pyx_n_u_product_configuration, __pyx_k_product_configuration, sizeof(__pyx_k_product_configuration), 0, 1, 0, 1},
    {&__pyx_n_s_product_end, __pyx_k_product_end, sizeof(__pyx_k_product_end), 0, 0, 1, 1},
    {&__pyx_n_u_product_end, __pyx_k_product_end, sizeof(__pyx_k_product_end), 0, 1, 0, 1},
    {&__pyx_n_s_product_hdr, __pyx_k_product_hdr, sizeof(__pyx_k_product_hdr), 0, 0, 1, 1},
    {&__pyx_n_u_product_name, __pyx_k_product_name, sizeof(__pyx_k_product_name), 0, 1, 0, 1},
    {&__pyx_n_u_product_sequence_number, __pyx_k_product_sequence_number, sizeof(__pyx_k_product_sequence_number), 0, 1, 0, 1},
    {&__pyx_n_u_product_specific_bytes, __pyx_k_product_specific_bytes, sizeof(__pyx_k_product_specific_bytes), 0, 1, 0, 1},
    {&__pyx_n_u_product_type_code, __pyx_k_product_type_code, sizeof(__pyx_k_product_type_code), 0, 1, 0, 1},
    {&__pyx_n_u_projection_name, __pyx_k_projection_name, sizeof(__pyx_k_projection_name), 0, 1, 0, 1},
    {&__pyx_n_u_projection_type, __pyx_k_projection_type, sizeof(__pyx_k_projection_type), 0, 1, 0, 1},
    {&__pyx_n_s_prt_value, __pyx_k_prt_value, sizeof(__pyx_k_prt_value), 0, 0, 1, 1},
    {&__pyx_n_u_pulse_width, __pyx_k_pulse_width, sizeof(__pyx_k_pulse_width), 0, 1, 0, 1},
    {&__pyx_n_s_pyart_io__sigmetfile, __pyx_k_pyart_io__sigmetfile, sizeof(__pyx_k_pyart_io__sigmetfile), 0, 0, 1, 1},
    {&__pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_k_pyart_io__sigmetfile_pyx, sizeof(__pyx_k_pyart_io__sigmetfile_pyx), 0, 0, 1, 0},
    {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
    {&__pyx_n_u_radar_height, __pyx_k_radar_height, sizeof(__pyx_k_radar_height), 0, 1, 0, 1},
    {&__pyx_n_u_radar_height_above_reference, __pyx_k_radar_height_above_reference, sizeof(__pyx_k_radar_height_above_reference), 0, 1, 0, 1},
    {&__pyx_n_u_radial_smoother, __pyx_k_radial_smoother, sizeof(__pyx_k_radial_smoother), 0, 1, 0, 1},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_u_range_bin_averaging_flag, __pyx_k_range_bin_averaging_flag, sizeof(__pyx_k_range_bin_averaging_flag), 0, 1, 0, 1},
    {&__pyx_n_s_raw_data, __pyx_k_raw_data, sizeof(__pyx_k_raw_data), 0, 0, 1, 1},
    {&__pyx_n_s_raw_prod_bhdr, __pyx_k_raw_prod_bhdr, sizeof(__pyx_k_raw_prod_bhdr), 0, 0, 1, 1},
    {&__pyx_n_s_raw_sweep_data, __pyx_k_raw_sweep_data, sizeof(__pyx_k_raw_sweep_data), 0, 0, 1, 1},
    {&__pyx_n_u_ray_header_bytes, __pyx_k_ray_header_bytes, sizeof(__pyx_k_ray_header_bytes), 0, 1, 0, 1},
    {&__pyx_n_u_ray_header_mask, __pyx_k_ray_header_mask, sizeof(__pyx_k_ray_header_mask), 0, 1, 0, 1},
    {&__pyx_n_s_ray_headers, __pyx_k_ray_headers, sizeof(__pyx_k_ray_headers), 0, 0, 1, 1},
    {&__pyx_n_s_ray_i, __pyx_k_ray_i, sizeof(__pyx_k_ray_i), 0, 0, 1, 1},
    {&__pyx_n_s_ray_nbins, __pyx_k_ray_nbins, sizeof(__pyx_k_ray_nbins), 0, 0, 1, 1},
    {&__pyx_n_u_ray_number, __pyx_k_ray_number, sizeof(__pyx_k_ray_number), 0, 1, 0, 1},
    {&__pyx_n_s_ray_time, __pyx_k_ray_time, sizeof(__pyx_k_ray_time), 0, 0, 1, 1},
    {&__pyx_n_u_rb, __pyx_k_rb, sizeof(__pyx_k_rb), 0, 1, 0, 1},
    {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
    {&__pyx_n_u_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 1, 0, 1},
    {&__pyx_n_s_read_data, __pyx_k_read_data, sizeof(__pyx_k_read_data), 0, 0, 1, 1},
    {&__pyx_n_u_reciever_bandwidth, __pyx_k_reciever_bandwidth, sizeof(__pyx_k_reciever_bandwidth), 0, 1, 0, 1},
    {&__pyx_n_s_record, __pyx_k_record, sizeof(__pyx_k_record), 0, 0, 1, 1},
    {&__pyx_n_u_record_number, __pyx_k_record_number, sizeof(__pyx_k_record_number), 0, 1, 0, 1},
    {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
    {&__pyx_n_u_reflectivity_calibration, __pyx_k_reflectivity_calibration, sizeof(__pyx_k_reflectivity_calibration), 0, 1, 0, 1},
    {&__pyx_n_u_reflectivity_noise_threshold, __pyx_k_reflectivity_noise_threshold, sizeof(__pyx_k_reflectivity_noise_threshold), 0, 1, 0, 1},
    {&__pyx_n_u_reflectivity_slope, __pyx_k_reflectivity_slope, sizeof(__pyx_k_reflectivity_slope), 0, 1, 0, 1},
    {&__pyx_n_u_reflectivity_unfolding_threshold, __pyx_k_reflectivity_unfolding_threshold, sizeof(__pyx_k_reflectivity_unfolding_threshold), 0, 1, 0, 1},
    {&__pyx_n_u_reserved, __pyx_k_reserved, sizeof(__pyx_k_reserved), 0, 1, 0, 1},
    {&__pyx_n_s_reshape, __pyx_k_reshape, sizeof(__pyx_k_reshape), 0, 0, 1, 1},
    {&__pyx_n_u_resolution_rays, __pyx_k_resolution_rays, sizeof(__pyx_k_resolution_rays), 0, 1, 0, 1},
    {&__pyx_n_u_right_azimuth_limit, __pyx_k_right_azimuth_limit, sizeof(__pyx_k_right_azimuth_limit), 0, 1, 0, 1},
    {&__pyx_n_u_sample_size, __pyx_k_sample_size, sizeof(__pyx_k_sample_size), 0, 1, 0, 1},
    {&__pyx_n_u_samples_used, __pyx_k_samples_used, sizeof(__pyx_k_samples_used), 0, 1, 0, 1},
    {&__pyx_n_u_scheduling_code, __pyx_k_scheduling_code, sizeof(__pyx_k_scheduling_code), 0, 1, 0, 1},
    {&__pyx_n_u_seconds, __pyx_k_seconds, sizeof(__pyx_k_seconds), 0, 1, 0, 1},
    {&__pyx_n_u_seconds_between_runs, __pyx_k_seconds_between_runs, sizeof(__pyx_k_seconds_between_runs), 0, 1, 0, 1},
    {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
    {&__pyx_kp_s_self__rbuf_p_cannot_be_converted, __pyx_k_self__rbuf_p_cannot_be_converted, sizeof(__pyx_k_self__rbuf_p_cannot_be_converted), 0, 0, 1, 0},
    {&__pyx_kp_u_self__rbuf_pos_is, __pyx_k_self__rbuf_pos_is, sizeof(__pyx_k_self__rbuf_pos_is), 0, 1, 0, 0},
    {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
    {&__pyx_n_s_shrink, __pyx_k_shrink, sizeof(__pyx_k_shrink), 0, 0, 1, 1},
    {&__pyx_n_u_signal_processor_type, __pyx_k_signal_processor_type, sizeof(__pyx_k_signal_processor_type), 0, 1, 0, 1},
    {&__pyx_n_u_site_name, __pyx_k_site_name, sizeof(__pyx_k_site_name), 0, 1, 0, 1},
    {&__pyx_n_u_skip_time, __pyx_k_skip_time, sizeof(__pyx_k_skip_time), 0, 1, 0, 1},
    {&__pyx_n_u_spare_0, __pyx_k_spare_0, sizeof(__pyx_k_spare_0), 0, 1, 0, 1},
    {&__pyx_n_u_spare_1, __pyx_k_spare_1, sizeof(__pyx_k_spare_1), 0, 1, 0, 1},
    {&__pyx_n_u_spare_2, __pyx_k_spare_2, sizeof(__pyx_k_spare_2), 0, 1, 0, 1},
    {&__pyx_n_u_spare_3, __pyx_k_spare_3, sizeof(__pyx_k_spare_3), 0, 1, 0, 1},
    {&__pyx_n_u_spare_4, __pyx_k_spare_4, sizeof(__pyx_k_spare_4), 0, 1, 0, 1},
    {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
    {&__pyx_n_u_sqi_threshold, __pyx_k_sqi_threshold, sizeof(__pyx_k_sqi_threshold), 0, 1, 0, 1},
    {&__pyx_n_s_sqrt, __pyx_k_sqrt, sizeof(__pyx_k_sqrt), 0, 0, 1, 1},
    {&__pyx_n_u_standard_parallel_1, __pyx_k_standard_parallel_1, sizeof(__pyx_k_standard_parallel_1), 0, 1, 0, 1},
    {&__pyx_n_u_standard_parallel_2, __pyx_k_standard_parallel_2, sizeof(__pyx_k_standard_parallel_2), 0, 1, 0, 1},
    {&__pyx_n_u_start_first_section_flag, __pyx_k_start_first_section_flag, sizeof(__pyx_k_start_first_section_flag), 0, 1, 0, 1},
    {&__pyx_n_u_start_first_sector_flag, __pyx_k_start_first_sector_flag, sizeof(__pyx_k_start_first_sector_flag), 0, 1, 0, 1},
    {&__pyx_n_u_start_time, __pyx_k_start_time, sizeof(__pyx_k_start_time), 0, 1, 0, 1},
    {&__pyx_n_u_step_input_bins, __pyx_k_step_input_bins, sizeof(__pyx_k_step_input_bins), 0, 1, 0, 1},
    {&__pyx_n_u_step_output_bins, __pyx_k_step_output_bins, sizeof(__pyx_k_step_output_bins), 0, 1, 0, 1},
    {&__pyx_n_u_stop_time, __pyx_k_stop_time, sizeof(__pyx_k_stop_time), 0, 1, 0, 1},
    {&__pyx_n_s_string, __pyx_k_string, sizeof(__pyx_k_string), 0, 0, 1, 1},
    {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
    {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
    {&__pyx_n_s_structure, __pyx_k_structure, sizeof(__pyx_k_structure), 0, 0, 1, 1},
    {&__pyx_n_u_structure_header, __pyx_k_structure_header, sizeof(__pyx_k_structure_header), 0, 1, 0, 1},
    {&__pyx_n_u_structure_identifier, __pyx_k_structure_identifier, sizeof(__pyx_k_structure_identifier), 0, 1, 0, 1},
    {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
    {&__pyx_n_s_sweep_data, __pyx_k_sweep_data, sizeof(__pyx_k_sweep_data), 0, 0, 1, 1},
    {&__pyx_n_u_sweep_ingest_time, __pyx_k_sweep_ingest_time, sizeof(__pyx_k_sweep_ingest_time), 0, 1, 0, 1},
    {&__pyx_n_s_sweep_metadata, __pyx_k_sweep_metadata, sizeof(__pyx_k_sweep_metadata), 0, 0, 1, 1},
    {&__pyx_n_u_sweep_number, __pyx_k_sweep_number, sizeof(__pyx_k_sweep_number), 0, 1, 0, 1},
    {&__pyx_n_u_sweep_start_time, __pyx_k_sweep_start_time, sizeof(__pyx_k_sweep_start_time), 0, 1, 0, 1},
    {&__pyx_n_u_task_calib_info, __pyx_k_task_calib_info, sizeof(__pyx_k_task_calib_info), 0, 1, 0, 1},
    {&__pyx_n_s_task_config, __pyx_k_task_config, sizeof(__pyx_k_task_config), 0, 0, 1, 1},
    {&__pyx_n_s_task_configuration, __pyx_k_task_configuration, sizeof(__pyx_k_task_configuration), 0, 0, 1, 1},
    {&__pyx_n_u_task_configuration, __pyx_k_task_configuration, sizeof(__pyx_k_task_configuration), 0, 1, 0, 1},
    {&__pyx_n_u_task_configuration_file_name, __pyx_k_task_configuration_file_name, sizeof(__pyx_k_task_configuration_file_name), 0, 1, 0, 1},
    {&__pyx_n_u_task_data_time, __pyx_k_task_data_time, sizeof(__pyx_k_task_data_time), 0, 1, 0, 1},
    {&__pyx_n_u_task_description, __pyx_k_task_description, sizeof(__pyx_k_task_description), 0, 1, 0, 1},
    {&__pyx_n_s_task_dsp_info, __pyx_k_task_dsp_info, sizeof(__pyx_k_task_dsp_info), 0, 0, 1, 1},
    {&__pyx_n_u_task_dsp_info, __pyx_k_task_dsp_info, sizeof(__pyx_k_task_dsp_info), 0, 1, 0, 1},
    {&__pyx_n_u_task_dsp_mode, __pyx_k_task_dsp_mode, sizeof(__pyx_k_task_dsp_mode), 0, 1, 0, 1},
    {&__pyx_n_s_task_end_info, __pyx_k_task_end_info, sizeof(__pyx_k_task_end_info), 0, 0, 1, 1},
    {&__pyx_n_u_task_end_info, __pyx_k_task_end_info, sizeof(__pyx_k_task_end_info), 0, 1, 0, 1},
    {&__pyx_n_u_task_major_number, __pyx_k_task_major_number, sizeof(__pyx_k_task_major_number), 0, 1, 0, 1},
    {&__pyx_n_u_task_minor_number, __pyx_k_task_minor_number, sizeof(__pyx_k_task_minor_number), 0, 1, 0, 1},
    {&__pyx_n_u_task_misc_info, __pyx_k_task_misc_info, sizeof(__pyx_k_task_misc_info), 0, 1, 0, 1},
    {&__pyx_n_u_task_name, __pyx_k_task_name, sizeof(__pyx_k_task_name), 0, 1, 0, 1},
    {&__pyx_n_u_task_range_info, __pyx_k_task_range_info, sizeof(__pyx_k_task_range_info), 0, 1, 0, 1},
    {&__pyx_n_u_task_scan_info, __pyx_k_task_scan_info, sizeof(__pyx_k_task_scan_info), 0, 1, 0, 1},
    {&__pyx_n_u_task_scan_type_scan_info, __pyx_k_task_scan_type_scan_info, sizeof(__pyx_k_task_scan_type_scan_info), 0, 1, 0, 1},
    {&__pyx_n_u_task_sched_info, __pyx_k_task_sched_info, sizeof(__pyx_k_task_sched_info), 0, 1, 0, 1},
    {&__pyx_n_u_task_state, __pyx_k_task_state, sizeof(__pyx_k_task_state), 0, 1, 0, 1},
    {&__pyx_n_u_tcf_cal_flags_1, __pyx_k_tcf_cal_flags_1, sizeof(__pyx_k_tcf_cal_flags_1), 0, 1, 0, 1},
    {&__pyx_n_u_tcf_cal_flags_2, __pyx_k_tcf_cal_flags_2, sizeof(__pyx_k_tcf_cal_flags_2), 0, 1, 0, 1},
    {&__pyx_n_s_temp, __pyx_k_temp, sizeof(__pyx_k_temp), 0, 0, 1, 1},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
    {&__pyx_n_u_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 1, 0, 1},
    {&__pyx_n_u_time_used_last_run, __pyx_k_time_used_last_run, sizeof(__pyx_k_time_used_last_run), 0, 1, 0, 1},
    {&__pyx_n_u_times_run, __pyx_k_times_run, sizeof(__pyx_k_times_run), 0, 1, 0, 1},
    {&__pyx_n_u_total_size, __pyx_k_total_size, sizeof(__pyx_k_total_size), 0, 1, 0, 1},
    {&__pyx_n_u_tr_serial_number, __pyx_k_tr_serial_number, sizeof(__pyx_k_tr_serial_number), 0, 1, 0, 1},
    {&__pyx_n_u_tranmitter_phase_sequence, __pyx_k_tranmitter_phase_sequence, sizeof(__pyx_k_tranmitter_phase_sequence), 0, 1, 0, 1},
    {&__pyx_n_u_transmit_power, __pyx_k_transmit_power, sizeof(__pyx_k_transmit_power), 0, 1, 0, 1},
    {&__pyx_n_u_trigger_rate, __pyx_k_trigger_rate, sizeof(__pyx_k_trigger_rate), 0, 1, 0, 1},
    {&__pyx_n_u_trucation_height, __pyx_k_trucation_height, sizeof(__pyx_k_trucation_height), 0, 1, 0, 1},
    {&__pyx_n_u_truncation_height, __pyx_k_truncation_height, sizeof(__pyx_k_truncation_height), 0, 1, 0, 1},
    {&__pyx_n_u_tz_name, __pyx_k_tz_name, sizeof(__pyx_k_tz_name), 0, 1, 0, 1},
    {&__pyx_n_u_uint16, __pyx_k_uint16, sizeof(__pyx_k_uint16), 0, 1, 0, 1},
    {&__pyx_n_s_uint8, __pyx_k_uint8, sizeof(__pyx_k_uint8), 0, 0, 1, 1},
    {&__pyx_n_u_uncorrected_reflectivity_thresho, __pyx_k_uncorrected_reflectivity_thresho, sizeof(__pyx_k_uncorrected_reflectivity_thresho), 0, 1, 0, 1},
    {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
    {&__pyx_n_s_unpack_ingest_data_header, __pyx_k_unpack_ingest_data_header, sizeof(__pyx_k_unpack_ingest_data_header), 0, 0, 1, 1},
    {&__pyx_n_s_unpack_ingest_data_headers, __pyx_k_unpack_ingest_data_headers, sizeof(__pyx_k_unpack_ingest_data_headers), 0, 0, 1, 1},
    {&__pyx_n_s_unpack_ingest_header, __pyx_k_unpack_ingest_header, sizeof(__pyx_k_unpack_ingest_header), 0, 0, 1, 1},
    {&__pyx_n_s_unpack_key, __pyx_k_unpack_key, sizeof(__pyx_k_unpack_key), 0, 0, 1, 1},
    {&__pyx_n_s_unpack_product_hdr, __pyx_k_unpack_product_hdr, sizeof(__pyx_k_unpack_product_hdr), 0, 0, 1, 1},
    {&__pyx_n_s_unpack_raw_prod_bhdr, __pyx_k_unpack_raw_prod_bhdr, sizeof(__pyx_k_unpack_raw_prod_bhdr), 0, 0, 1, 1},
    {&__pyx_n_s_unpack_structure, __pyx_k_unpack_structure, sizeof(__pyx_k_unpack_structure), 0, 0, 1, 1},
    {&__pyx_n_u_upper_elevation_limit, __pyx_k_upper_elevation_limit, sizeof(__pyx_k_upper_elevation_limit), 0, 1, 0, 1},
    {&__pyx_n_u_variable_range_bin_flag, __pyx_k_variable_range_bin_flag, sizeof(__pyx_k_variable_range_bin_flag), 0, 1, 0, 1},
    {&__pyx_n_u_velocity_east, __pyx_k_velocity_east, sizeof(__pyx_k_velocity_east), 0, 1, 0, 1},
    {&__pyx_n_u_velocity_north, __pyx_k_velocity_north, sizeof(__pyx_k_velocity_north), 0, 1, 0, 1},
    {&__pyx_n_u_velocity_threshold_flags, __pyx_k_velocity_threshold_flags, sizeof(__pyx_k_velocity_threshold_flags), 0, 1, 0, 1},
    {&__pyx_n_u_velocity_unfolding_threshold, __pyx_k_velocity_unfolding_threshold, sizeof(__pyx_k_velocity_unfolding_threshold), 0, 1, 0, 1},
    {&__pyx_n_u_velocity_up, __pyx_k_velocity_up, sizeof(__pyx_k_velocity_up), 0, 1, 0, 1},
    {&__pyx_n_u_vertical_beamwidth, __pyx_k_vertical_beamwidth, sizeof(__pyx_k_vertical_beamwidth), 0, 1, 0, 1},
    {&__pyx_n_u_vertical_current_noise, __pyx_k_vertical_current_noise, sizeof(__pyx_k_vertical_current_noise), 0, 1, 0, 1},
    {&__pyx_n_u_vertical_i0_calibration, __pyx_k_vertical_i0_calibration, sizeof(__pyx_k_vertical_i0_calibration), 0, 1, 0, 1},
    {&__pyx_n_u_vertical_noise_calibration, __pyx_k_vertical_noise_calibration, sizeof(__pyx_k_vertical_noise_calibration), 0, 1, 0, 1},
    {&__pyx_n_u_vertical_radar_constant, __pyx_k_vertical_radar_constant, sizeof(__pyx_k_vertical_radar_constant), 0, 1, 0, 1},
    {&__pyx_n_s_view, __pyx_k_view, sizeof(__pyx_k_view), 0, 0, 1, 1},
    {&__pyx_n_u_volume_scan_start_time, __pyx_k_volume_scan_start_time, sizeof(__pyx_k_volume_scan_start_time), 0, 1, 0, 1},
    {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
    {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
    {&__pyx_n_u_wavelength, __pyx_k_wavelength, sizeof(__pyx_k_wavelength), 0, 1, 0, 1},
    {&__pyx_n_s_wavelength_cm, __pyx_k_wavelength_cm, sizeof(__pyx_k_wavelength_cm), 0, 0, 1, 1},
    {&__pyx_n_u_width_threshold_flags, __pyx_k_width_threshold_flags, sizeof(__pyx_k_width_threshold_flags), 0, 1, 0, 1},
    {&__pyx_n_u_width_unfolding_threshold, __pyx_k_width_unfolding_threshold, sizeof(__pyx_k_width_unfolding_threshold), 0, 1, 0, 1},
    {&__pyx_n_s_word0, __pyx_k_word0, sizeof(__pyx_k_word0), 0, 0, 1, 1},
    {&__pyx_n_s_word1, __pyx_k_word1, sizeof(__pyx_k_word1), 0, 0, 1, 1},
    {&__pyx_n_s_word2, __pyx_k_word2, sizeof(__pyx_k_word2), 0, 0, 1, 1},
    {&__pyx_n_s_word3, __pyx_k_word3, sizeof(__pyx_k_word3), 0, 0, 1, 1},
    {&__pyx_n_u_x_location, __pyx_k_x_location, sizeof(__pyx_k_x_location), 0, 1, 0, 1},
    {&__pyx_n_u_x_scale, __pyx_k_x_scale, sizeof(__pyx_k_x_scale), 0, 1, 0, 1},
    {&__pyx_n_u_x_size, __pyx_k_x_size, sizeof(__pyx_k_x_size), 0, 1, 0, 1},
    {&__pyx_n_u_x_smoother, __pyx_k_x_smoother, sizeof(__pyx_k_x_smoother), 0, 1, 0, 1},
    {&__pyx_n_s_xrange, __pyx_k_xrange, sizeof(__pyx_k_xrange), 0, 0, 1, 1},
    {&__pyx_n_u_y_location, __pyx_k_y_location, sizeof(__pyx_k_y_location), 0, 1, 0, 1},
    {&__pyx_n_u_y_scale, __pyx_k_y_scale, sizeof(__pyx_k_y_scale), 0, 1, 0, 1},
    {&__pyx_n_u_y_size, __pyx_k_y_size, sizeof(__pyx_k_y_size), 0, 1, 0, 1},
    {&__pyx_n_u_y_smoother, __pyx_k_y_smoother, sizeof(__pyx_k_y_smoother), 0, 1, 0, 1},
    {&__pyx_n_u_year, __pyx_k_year, sizeof(__pyx_k_year), 0, 1, 0, 1},
    {&__pyx_n_u_z_location, __pyx_k_z_location, sizeof(__pyx_k_z_location), 0, 1, 0, 1},
    {&__pyx_n_u_z_scale, __pyx_k_z_scale, sizeof(__pyx_k_z_scale), 0, 1, 0, 1},
    {&__pyx_n_u_z_size, __pyx_k_z_size, sizeof(__pyx_k_z_size), 0, 1, 0, 1},
    {&__pyx_n_u_zdr_bias, __pyx_k_zdr_bias, sizeof(__pyx_k_zdr_bias), 0, 1, 0, 1},
    {&__pyx_n_u_zdr_offset, __pyx_k_zdr_offset, sizeof(__pyx_k_zdr_offset), 0, 1, 0, 1},
    {&__pyx_n_u_zdr_threshold_flags, __pyx_k_zdr_threshold_flags, sizeof(__pyx_k_zdr_threshold_flags), 0, 1, 0, 1},
    {&__pyx_n_s_zeros_like, __pyx_k_zeros_like, sizeof(__pyx_k_zeros_like), 0, 0, 1, 1},
    {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
    {&__pyx_n_u_zr_constant, __pyx_k_zr_constant, sizeof(__pyx_k_zr_constant), 0, 1, 0, 1},
    {&__pyx_n_u_zr_exponent, __pyx_k_zr_exponent, sizeof(__pyx_k_zr_exponent), 0, 1, 0, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_open = __Pyx_GetBuiltinName(__pyx_n_s_open); if (!__pyx_builtin_open) __PYX_ERR(0, 67, __pyx_L1_error)
  #if PY_MAJOR_VERSION >= 3
  __pyx_builtin_xrange = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_xrange) __PYX_ERR(0, 159, __pyx_L1_error)
  #else
  __pyx_builtin_xrange = __Pyx_GetBuiltinName(__pyx_n_s_xrange); if (!__pyx_builtin_xrange) __PYX_ERR(0, 159, __pyx_L1_error)
  #endif
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 176, __pyx_L1_error)
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(0, 272, __pyx_L1_error)
  __pyx_builtin_print = __Pyx_GetBuiltinName(__pyx_n_s_print); if (!__pyx_builtin_print) __PYX_ERR(0, 286, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 348, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 828, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(2, 1043, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1043
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy._core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_numpy__core_multiarray_failed_to); if (unlikely(!__pyx_tuple_)) __PYX_ERR(2, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "../../../../../tmp/build-env-di4p05fc/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1049
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_u_numpy__core_umath_failed_to_impo); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(2, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "pyart/io/_sigmetfile.pyx":259
 *         # check if the file ended early, if so return Nones
 *         if len(lead_record) != RECORD_SIZE:
 *             return None, None, None             # <<<<<<<<<<<<<<
 * 
 *         # unpack structures
 */
  __pyx_tuple__3 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "pyart/io/_sigmetfile.pyx":300
 *         for i, data_type in enumerate(self.data_types):
 *             if data_type == 0 and full_xhdr:
 *                 sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])             # <<<<<<<<<<<<<<
 *             else:
 *                 sweep_data.append(convert_sigmet_data(
 */
  __pyx_slice__4 = PySlice_New(__pyx_int_6, Py_None, Py_None); if (unlikely(!__pyx_slice__4)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__4);
  __Pyx_GIVEREF(__pyx_slice__4);

  /* "pyart/io/_sigmetfile.pyx":306
 *                     raw_sweep_data[i::self.ndata_types, 4]))
 *             sweep_metadata.append(_parse_ray_headers(
 *                 raw_sweep_data[i::self.ndata_types, :6]))             # <<<<<<<<<<<<<<
 *         return ingest_data_headers, sweep_data, sweep_metadata
 * 
 */
  __pyx_slice__5 = PySlice_New(Py_None, __pyx_int_6, Py_None); if (unlikely(!__pyx_slice__5)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__5);
  __Pyx_GIVEREF(__pyx_slice__5);

  /* "pyart/io/_sigmetfile.pyx":451
 * 
 *     """
 *     az0 = bin2_to_angle(ray_headers.view('uint16')[..., 0])             # <<<<<<<<<<<<<<
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 */
  __pyx_tuple__6 = PyTuple_Pack(2, Py_Ellipsis, __pyx_int_0); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "pyart/io/_sigmetfile.pyx":452
 *     """
 *     az0 = bin2_to_angle(ray_headers.view('uint16')[..., 0])
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])             # <<<<<<<<<<<<<<
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 */
  __pyx_tuple__7 = PyTuple_Pack(2, Py_Ellipsis, __pyx_int_1); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "pyart/io/_sigmetfile.pyx":453
 *     az0 = bin2_to_angle(ray_headers.view('uint16')[..., 0])
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])             # <<<<<<<<<<<<<<
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 *     nbins = ray_headers.view('int16')[..., 4]
 */
  __pyx_tuple__8 = PyTuple_Pack(2, Py_Ellipsis, __pyx_int_2); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "pyart/io/_sigmetfile.pyx":454
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])             # <<<<<<<<<<<<<<
 *     nbins = ray_headers.view('int16')[..., 4]
 *     time = ray_headers.view('uint16')[..., 5]
 */
  __pyx_tuple__9 = PyTuple_Pack(2, Py_Ellipsis, __pyx_int_3); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "pyart/io/_sigmetfile.pyx":455
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 *     nbins = ray_headers.view('int16')[..., 4]             # <<<<<<<<<<<<<<
 *     time = ray_headers.view('uint16')[..., 5]
 *     prf_flag = np.mod(ray_headers.view('int16')[..., 0], 2)
 */
  __pyx_tuple__10 = PyTuple_Pack(2, Py_Ellipsis, __pyx_int_4); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "pyart/io/_sigmetfile.pyx":456
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 *     nbins = ray_headers.view('int16')[..., 4]
 *     time = ray_headers.view('uint16')[..., 5]             # <<<<<<<<<<<<<<
 *     prf_flag = np.mod(ray_headers.view('int16')[..., 0], 2)
 *     return (az0, el0, az1, el1, nbins, time, prf_flag)
 */
  __pyx_tuple__11 = PyTuple_Pack(2, Py_Ellipsis, __pyx_int_5); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "pyart/io/_sigmetfile.pyx":661
 *         # 0 : no data available (mask)
 *         # 65535 Reserved for area not scanned in product file (nothing)
 *         out[:] = (data.view('uint16') - 32768.) / 100.             # <<<<<<<<<<<<<<
 *         mask[data.view('uint16') == 0] = True
 * 
 */
  __pyx_slice__12 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__12)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__12);
  __Pyx_GIVEREF(__pyx_slice__12);

  /* "pyart/io/_sigmetfile.pyx":692
 *         # extended_header_v0, _v1, _v2, section 4.2.8-4.2.10
 *         # Here we return an array with the times in milliseconds.
 *         return data[..., :2].copy().view('i4')             # <<<<<<<<<<<<<<
 * 
 *     # one byte data types
 */
  __pyx_slice__13 = PySlice_New(Py_None, __pyx_int_2, Py_None); if (unlikely(!__pyx_slice__13)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__13);
  __Pyx_GIVEREF(__pyx_slice__13);
  __pyx_tuple__14 = PyTuple_Pack(2, Py_Ellipsis, __pyx_slice__13); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "pyart/io/_sigmetfile.pyx":866
 * def _unpack_raw_prod_bhdr(record):
 *     """ Return a dict with the unpacked raw_prod_bhdr from a record. """
 *     return _unpack_structure(record[:12], RAW_PROD_BHDR)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_slice__16 = PySlice_New(Py_None, __pyx_int_12, Py_None); if (unlikely(!__pyx_slice__16)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__16);
  __Pyx_GIVEREF(__pyx_slice__16);

  /* "pyart/io/_sigmetfile.pyx":875
 * 
 *     # unpack the product_hdr structure from the first record
 *     product_hdr = _unpack_structure(record[:640], PRODUCT_HDR)             # <<<<<<<<<<<<<<
 * 
 *     # product_hdr substructure
 */
  __pyx_slice__17 = PySlice_New(Py_None, __pyx_int_640, Py_None); if (unlikely(!__pyx_slice__17)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__17);
  __Pyx_GIVEREF(__pyx_slice__17);

  /* "pyart/io/_sigmetfile.pyx":904
 * 
 *     # unpack the ingest_header structure from the second_record
 *     ingest_header = _unpack_structure(record[:4884], INGEST_HEADER)             # <<<<<<<<<<<<<<
 * 
 *     # ingest_header substructure
 */
  __pyx_slice__18 = PySlice_New(Py_None, __pyx_int_4884, Py_None); if (unlikely(!__pyx_slice__18)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__18);
  __Pyx_GIVEREF(__pyx_slice__18);

  /* "pyart/io/_sigmetfile.pyx":84
 *         self._raw_product_bhdrs = []
 * 
 *     def _determine_data_types(self):             # <<<<<<<<<<<<<<
 *         """ Determine the available data types in the file. """
 *         # determine the available fields
 */
  __pyx_tuple__20 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_task_config, __pyx_n_s_task_dsp_info, __pyx_n_s_word0, __pyx_n_s_word1, __pyx_n_s_word2, __pyx_n_s_word3); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_determine_data_types, 84, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 84, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":95
 *         return _data_types_from_mask(word0, word1, word2, word3)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """ Close the file. """
 *         self._fh.close()
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_close, 95, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 95, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":99
 *         self._fh.close()
 * 
 *     def read_data(self, full_xhdr=False):             # <<<<<<<<<<<<<<
 *         """
 *         Read all data from the file.
 */
  __pyx_tuple__24 = PyTuple_Pack(33, __pyx_n_s_self, __pyx_n_s_full_xhdr, __pyx_n_s_nsweeps, __pyx_n_s_nbins, __pyx_n_s_nrays, __pyx_n_s_shape, __pyx_n_s_data, __pyx_n_s_metadata, __pyx_n_s_name_2, __pyx_n_s_header_dic, __pyx_n_s_i, __pyx_n_s_ingest_data_hdrs, __pyx_n_s_sweep_data, __pyx_n_s_sweep_metadata, __pyx_n_s_mess, __pyx_n_s_k, __pyx_n_s_j, __pyx_n_s_temp, __pyx_n_s_az0, __pyx_n_s_el0, __pyx_n_s_az1, __pyx_n_s_el1, __pyx_n_s_ray_nbins, __pyx_n_s_ray_time, __pyx_n_s_prf_flag, __pyx_n_s_wavelength_cm, __pyx_n_s_prt_value, __pyx_n_s_task_config, __pyx_n_s_multi_prf_flag, __pyx_n_s_multiplier, __pyx_n_s_nyquist, __pyx_n_s_name_2, __pyx_n_s_name_2); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 33, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_read_data, 99, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 99, __pyx_L1_error)
  __pyx_tuple__26 = PyTuple_Pack(1, Py_False); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "pyart/io/_sigmetfile.pyx":224
 *         return data, metadata
 * 
 *     def _get_sweep(self, full_xhdr=False, raw_data=False):             # <<<<<<<<<<<<<<
 *         """
 *         Get the data and metadata from the next sweep.
 */
  __pyx_tuple__27 = PyTuple_Pack(16, __pyx_n_s_self, __pyx_n_s_full_xhdr, __pyx_n_s_raw_data, __pyx_n_s_lead_record, __pyx_n_s_raw_prod_bhdr, __pyx_n_s_ingest_data_headers, __pyx_n_s_nray_data_types, __pyx_n_s_nrays, __pyx_n_s_nbins, __pyx_n_s_raw_sweep_data, __pyx_n_s_ray_i, __pyx_n_s_sweep_data, __pyx_n_s_sweep_metadata, __pyx_n_s_i, __pyx_n_s_data_type, __pyx_n_s_d); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 16, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_get_sweep, 224, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(0, 224, __pyx_L1_error)
  __pyx_tuple__29 = PyTuple_Pack(2, Py_False, Py_False); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"
 */
  __pyx_tuple__31 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":407
 * 
 * 
 * def _data_types_from_mask(word0, word1, word2, word3):             # <<<<<<<<<<<<<<
 *     """
 *     Return a list of the data types from the words in the data_type mask.
 */
  __pyx_tuple__33 = PyTuple_Pack(9, __pyx_n_s_word0, __pyx_n_s_word1, __pyx_n_s_word2, __pyx_n_s_word3, __pyx_n_s_data_types, __pyx_n_s_i, __pyx_n_s_i, __pyx_n_s_i, __pyx_n_s_i); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_data_types_from_mask, 407, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 407, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":418
 * 
 * 
 * def _is_bit_set(number, bit):             # <<<<<<<<<<<<<<
 *     """ Return True if bit is set in number. """
 *     return number >> bit & 1 == 1
 */
  __pyx_tuple__35 = PyTuple_Pack(2, __pyx_n_s_number, __pyx_n_s_bit); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_is_bit_set, 418, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 418, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":423
 * 
 * 
 * def _parse_ray_headers(ray_headers):             # <<<<<<<<<<<<<<
 *     """
 *     Parse the metadata from Sigmet ray headers.
 */
  __pyx_tuple__37 = PyTuple_Pack(8, __pyx_n_s_ray_headers, __pyx_n_s_az0, __pyx_n_s_el0, __pyx_n_s_az1, __pyx_n_s_el1, __pyx_n_s_nbins, __pyx_n_s_time, __pyx_n_s_prf_flag); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_parse_ray_headers, 423, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 423, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":602
 * # file. Rewriting the convertions/masking in Cython does not seem to improved
 * # performance likely since most of the routines are already vectorized.
 * def convert_sigmet_data(data_type, data, nbins):             # <<<<<<<<<<<<<<
 *     """ Convert sigmet data. """
 *     out = np.empty_like(data, dtype='float32')
 */
  __pyx_tuple__39 = PyTuple_Pack(14, __pyx_n_s_data_type, __pyx_n_s_data, __pyx_n_s_nbins, __pyx_n_s_out, __pyx_n_s_mask, __pyx_n_s_data_type_name, __pyx_n_s_like_dbt2, __pyx_n_s_like_sqi, __pyx_n_s_like_sqi2, __pyx_n_s_like_dbt, __pyx_n_s_nrays, __pyx_n_s_nbin, __pyx_n_s_ndata, __pyx_n_s_exp); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_convert_sigmet_data, 602, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 602, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":809
 * 
 * 
 * def bin2_to_angle(bin2):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin2 encoded value (or array). """
 *     return 360. * bin2 / 65536
 */
  __pyx_tuple__41 = PyTuple_Pack(1, __pyx_n_s_bin2); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_bin2_to_angle, 809, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 809, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":814
 * 
 * 
 * def bin4_to_angle(bin4):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin4 encoded value (or array). """
 *     return 360. * bin4 / 4294967296
 */
  __pyx_tuple__43 = PyTuple_Pack(1, __pyx_n_s_bin4); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_bin4_to_angle, 814, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 814, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":824
 * 
 * 
 * def _unpack_structure(string, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a structure """
 *     fmt = ''.join([i[1] for i in structure])
 */
  __pyx_tuple__45 = PyTuple_Pack(6, __pyx_n_s_string, __pyx_n_s_structure, __pyx_n_s_fmt, __pyx_n_s_l, __pyx_n_s_i, __pyx_n_s_i); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__45, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_unpack_structure, 824, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 824, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":831
 * 
 * 
 * def _unpack_key(dic, key, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a key. """
 *     dic[key] = _unpack_structure(dic[key], structure)
 */
  __pyx_tuple__47 = PyTuple_Pack(3, __pyx_n_s_dic, __pyx_n_s_key, __pyx_n_s_structure); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_unpack_key, 831, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 831, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":836
 * 
 * 
 * def _unpack_ingest_data_headers(record, ndata_types):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack one or more ingest_data_header from a record.
 */
  __pyx_tuple__49 = PyTuple_Pack(4, __pyx_n_s_record, __pyx_n_s_ndata_types, __pyx_n_s_idh, __pyx_n_s_i); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__49, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_unpack_ingest_data_headers, 836, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 836, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":850
 * 
 * 
 * def _unpack_ingest_data_header(record, number):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack a single ingest_data_header from record.  Return None on error.
 */
  __pyx_tuple__51 = PyTuple_Pack(5, __pyx_n_s_record, __pyx_n_s_number, __pyx_n_s_offset, __pyx_n_s_string, __pyx_n_s_idh); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__51, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_unpack_ingest_data_header, 850, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(0, 850, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":864
 * 
 * 
 * def _unpack_raw_prod_bhdr(record):             # <<<<<<<<<<<<<<
 *     """ Return a dict with the unpacked raw_prod_bhdr from a record. """
 *     return _unpack_structure(record[:12], RAW_PROD_BHDR)
 */
  __pyx_tuple__53 = PyTuple_Pack(1, __pyx_n_s_record); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__53, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_unpack_raw_prod_bhdr, 864, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(0, 864, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":869
 * 
 * 
 * def _unpack_product_hdr(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked product_hdr from the first record.
 */
  __pyx_tuple__55 = PyTuple_Pack(4, __pyx_n_s_record, __pyx_n_s_product_hdr, __pyx_n_s_product_config, __pyx_n_s_product_end); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__55, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_unpack_product_hdr, 869, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(0, 869, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":898
 * 
 * 
 * def _unpack_ingest_header(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked ingest_header from the second record.
 */
  __pyx_tuple__57 = PyTuple_Pack(6, __pyx_n_s_record, __pyx_n_s_ingest_header, __pyx_n_s_ingest_configuration, __pyx_n_s_task_configuration, __pyx_n_s_task_dsp_info, __pyx_n_s_task_end_info); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__57, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyart_io__sigmetfile_pyx, __pyx_n_s_unpack_ingest_header, 898, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(0, 898, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":974
 * # 640 bytes: product_hdr (section 4.2.25, page 47)
 * PRODUCT_HDR = (
 *     ('structure_header', '12s'),        # 12 bytes             # <<<<<<<<<<<<<<
 *     ('product_configuration', '320s'),  # 320 bytes
 *     ('product_end', '308s'),            # 308 bytes
 */
  __pyx_tuple__59 = PyTuple_Pack(2, __pyx_n_u_structure_header, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);

  /* "pyart/io/_sigmetfile.pyx":975
 * PRODUCT_HDR = (
 *     ('structure_header', '12s'),        # 12 bytes
 *     ('product_configuration', '320s'),  # 320 bytes             # <<<<<<<<<<<<<<
 *     ('product_end', '308s'),            # 308 bytes
 * )
 */
  __pyx_tuple__60 = PyTuple_Pack(2, __pyx_n_u_product_configuration, __pyx_kp_u_320s); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);

  /* "pyart/io/_sigmetfile.pyx":976
 *     ('structure_header', '12s'),        # 12 bytes
 *     ('product_configuration', '320s'),  # 320 bytes
 *     ('product_end', '308s'),            # 308 bytes             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__61 = PyTuple_Pack(2, __pyx_n_u_product_end, __pyx_kp_u_308s); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);

  /* "pyart/io/_sigmetfile.pyx":974
 * # 640 bytes: product_hdr (section 4.2.25, page 47)
 * PRODUCT_HDR = (
 *     ('structure_header', '12s'),        # 12 bytes             # <<<<<<<<<<<<<<
 *     ('product_configuration', '320s'),  # 320 bytes
 *     ('product_end', '308s'),            # 308 bytes
 */
  __pyx_tuple__62 = PyTuple_Pack(3, __pyx_tuple__59, __pyx_tuple__60, __pyx_tuple__61); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);

  /* "pyart/io/_sigmetfile.pyx":994
 *     ('scheduling_code', UINT2),
 *     ('seconds_between_runs', SINT4),
 *     ('generation_time', '12s'),     # 12 bytes: ymds_time             # <<<<<<<<<<<<<<
 *     ('sweep_ingest_time', '12s'),   # 12 bytes: ymds_time
 *     ('file_ingest_time', '12s'),    # 12 bytes: ymds_time
 */
  __pyx_tuple__63 = PyTuple_Pack(2, __pyx_n_u_generation_time, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(0, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);

  /* "pyart/io/_sigmetfile.pyx":995
 *     ('seconds_between_runs', SINT4),
 *     ('generation_time', '12s'),     # 12 bytes: ymds_time
 *     ('sweep_ingest_time', '12s'),   # 12 bytes: ymds_time             # <<<<<<<<<<<<<<
 *     ('file_ingest_time', '12s'),    # 12 bytes: ymds_time
 *     ('spare_0', '6s'),              # 6 bytes
 */
  __pyx_tuple__64 = PyTuple_Pack(2, __pyx_n_u_sweep_ingest_time, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);

  /* "pyart/io/_sigmetfile.pyx":996
 *     ('generation_time', '12s'),     # 12 bytes: ymds_time
 *     ('sweep_ingest_time', '12s'),   # 12 bytes: ymds_time
 *     ('file_ingest_time', '12s'),    # 12 bytes: ymds_time             # <<<<<<<<<<<<<<
 *     ('spare_0', '6s'),              # 6 bytes
 *     ('product_name', '12s'),
 */
  __pyx_tuple__65 = PyTuple_Pack(2, __pyx_n_u_file_ingest_time, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);

  /* "pyart/io/_sigmetfile.pyx":997
 *     ('sweep_ingest_time', '12s'),   # 12 bytes: ymds_time
 *     ('file_ingest_time', '12s'),    # 12 bytes: ymds_time
 *     ('spare_0', '6s'),              # 6 bytes             # <<<<<<<<<<<<<<
 *     ('product_name', '12s'),
 *     ('task_name', '12s'),
 */
  __pyx_tuple__66 = PyTuple_Pack(2, __pyx_n_u_spare_0, __pyx_kp_u_6s); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);

  /* "pyart/io/_sigmetfile.pyx":998
 *     ('file_ingest_time', '12s'),    # 12 bytes: ymds_time
 *     ('spare_0', '6s'),              # 6 bytes
 *     ('product_name', '12s'),             # <<<<<<<<<<<<<<
 *     ('task_name', '12s'),
 *     ('flag', UINT2),
 */
  __pyx_tuple__67 = PyTuple_Pack(2, __pyx_n_u_product_name, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);

  /* "pyart/io/_sigmetfile.pyx":999
 *     ('spare_0', '6s'),              # 6 bytes
 *     ('product_name', '12s'),
 *     ('task_name', '12s'),             # <<<<<<<<<<<<<<
 *     ('flag', UINT2),
 *     ('x_scale', SINT4),
 */
  __pyx_tuple__68 = PyTuple_Pack(2, __pyx_n_u_task_name, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);

  /* "pyart/io/_sigmetfile.pyx":1012
 *     ('maximum_range', SINT4),
 *     ('data_type', UINT2),
 *     ('projection_name', '12s'),             # <<<<<<<<<<<<<<
 *     ('input_data_type', UINT2),
 *     ('projection_type', UINT1),
 */
  __pyx_tuple__69 = PyTuple_Pack(2, __pyx_n_u_projection_name, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(0, 1012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);

  /* "pyart/io/_sigmetfile.pyx":1015
 *     ('input_data_type', UINT2),
 *     ('projection_type', UINT1),
 *     ('spare_1', '1s'),              # 1 bytes             # <<<<<<<<<<<<<<
 *     ('radial_smoother', SINT2),
 *     ('times_run', SINT2),
 */
  __pyx_tuple__70 = PyTuple_Pack(2, __pyx_n_u_spare_1, __pyx_kp_u_1s); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);

  /* "pyart/io/_sigmetfile.pyx":1022
 *     ('x_smoother', SINT2),
 *     ('y_smoother', SINT2),
 *     ('product_specific_bytes', '80s'),  # 80 bytes:             # <<<<<<<<<<<<<<
 *     ('minor_task_suffix', '16s'),
 *     ('spare_2', '12s'),             # 12 bytes
 */
  __pyx_tuple__71 = PyTuple_Pack(2, __pyx_n_u_product_specific_bytes, __pyx_kp_u_80s); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);

  /* "pyart/io/_sigmetfile.pyx":1023
 *     ('y_smoother', SINT2),
 *     ('product_specific_bytes', '80s'),  # 80 bytes:
 *     ('minor_task_suffix', '16s'),             # <<<<<<<<<<<<<<
 *     ('spare_2', '12s'),             # 12 bytes
 *     ('color_scale_def', '48s')      # 48 bytes: color_scale_def
 */
  __pyx_tuple__72 = PyTuple_Pack(2, __pyx_n_u_minor_task_suffix, __pyx_kp_u_16s); if (unlikely(!__pyx_tuple__72)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);

  /* "pyart/io/_sigmetfile.pyx":1024
 *     ('product_specific_bytes', '80s'),  # 80 bytes:
 *     ('minor_task_suffix', '16s'),
 *     ('spare_2', '12s'),             # 12 bytes             # <<<<<<<<<<<<<<
 *     ('color_scale_def', '48s')      # 48 bytes: color_scale_def
 * )
 */
  __pyx_tuple__73 = PyTuple_Pack(2, __pyx_n_u_spare_2, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 1024, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);

  /* "pyart/io/_sigmetfile.pyx":1025
 *     ('minor_task_suffix', '16s'),
 *     ('spare_2', '12s'),             # 12 bytes
 *     ('color_scale_def', '48s')      # 48 bytes: color_scale_def             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__74 = PyTuple_Pack(2, __pyx_n_u_color_scale_def, __pyx_kp_u_48s); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);

  /* "pyart/io/_sigmetfile.pyx":1044
 *     ('icolcnt', SINT2),
 *     ('iset_and_scale', UINT2),
 *     ('ilevel_seams', '32s')     # 32 bytes: UINT2[16]             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__75 = PyTuple_Pack(2, __pyx_n_u_ilevel_seams, __pyx_kp_u_32s); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);

  /* "pyart/io/_sigmetfile.pyx":1049
 * # 308 bytes : product_end (section 4.2.24)
 * PRODUCT_END = (
 *     ('site_name', '16s'),             # <<<<<<<<<<<<<<
 *     ('iris_version_created', '8s'),
 *     ('ingest_iris_version', '8s'),
 */
  __pyx_tuple__76 = PyTuple_Pack(2, __pyx_n_u_site_name, __pyx_kp_u_16s); if (unlikely(!__pyx_tuple__76)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__76);
  __Pyx_GIVEREF(__pyx_tuple__76);

  /* "pyart/io/_sigmetfile.pyx":1050
 * PRODUCT_END = (
 *     ('site_name', '16s'),
 *     ('iris_version_created', '8s'),             # <<<<<<<<<<<<<<
 *     ('ingest_iris_version', '8s'),
 *     ('ingest_time', '12s'),         # 12 bytes: ymds_time
 */
  __pyx_tuple__77 = PyTuple_Pack(2, __pyx_n_u_iris_version_created, __pyx_kp_u_8s); if (unlikely(!__pyx_tuple__77)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__77);
  __Pyx_GIVEREF(__pyx_tuple__77);

  /* "pyart/io/_sigmetfile.pyx":1051
 *     ('site_name', '16s'),
 *     ('iris_version_created', '8s'),
 *     ('ingest_iris_version', '8s'),             # <<<<<<<<<<<<<<
 *     ('ingest_time', '12s'),         # 12 bytes: ymds_time
 *     ('spare_0', '28s'),             # 28 bytes
 */
  __pyx_tuple__78 = PyTuple_Pack(2, __pyx_n_u_ingest_iris_version, __pyx_kp_u_8s); if (unlikely(!__pyx_tuple__78)) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_tuple__78);

  /* "pyart/io/_sigmetfile.pyx":1052
 *     ('iris_version_created', '8s'),
 *     ('ingest_iris_version', '8s'),
 *     ('ingest_time', '12s'),         # 12 bytes: ymds_time             # <<<<<<<<<<<<<<
 *     ('spare_0', '28s'),             # 28 bytes
 *     ('GMT_minute_offset_local', SINT2),
 */
  __pyx_tuple__79 = PyTuple_Pack(2, __pyx_n_u_ingest_time, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__79)) __PYX_ERR(0, 1052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__79);
  __Pyx_GIVEREF(__pyx_tuple__79);

  /* "pyart/io/_sigmetfile.pyx":1053
 *     ('ingest_iris_version', '8s'),
 *     ('ingest_time', '12s'),         # 12 bytes: ymds_time
 *     ('spare_0', '28s'),             # 28 bytes             # <<<<<<<<<<<<<<
 *     ('GMT_minute_offset_local', SINT2),
 *     ('ingest_hardware_name_', '16s'),
 */
  __pyx_tuple__80 = PyTuple_Pack(2, __pyx_n_u_spare_0, __pyx_kp_u_28s); if (unlikely(!__pyx_tuple__80)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__80);
  __Pyx_GIVEREF(__pyx_tuple__80);

  /* "pyart/io/_sigmetfile.pyx":1055
 *     ('spare_0', '28s'),             # 28 bytes
 *     ('GMT_minute_offset_local', SINT2),
 *     ('ingest_hardware_name_', '16s'),             # <<<<<<<<<<<<<<
 *     ('ingest_site_name_', '16s'),
 *     ('GMT_minute_offset_standard', SINT2),
 */
  __pyx_tuple__81 = PyTuple_Pack(2, __pyx_n_u_ingest_hardware_name, __pyx_kp_u_16s); if (unlikely(!__pyx_tuple__81)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);

  /* "pyart/io/_sigmetfile.pyx":1056
 *     ('GMT_minute_offset_local', SINT2),
 *     ('ingest_hardware_name_', '16s'),
 *     ('ingest_site_name_', '16s'),             # <<<<<<<<<<<<<<
 *     ('GMT_minute_offset_standard', SINT2),
 *     ('latitude', BIN4),
 */
  __pyx_tuple__82 = PyTuple_Pack(2, __pyx_n_u_ingest_site_name, __pyx_kp_u_16s); if (unlikely(!__pyx_tuple__82)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__82);
  __Pyx_GIVEREF(__pyx_tuple__82);

  /* "pyart/io/_sigmetfile.pyx":1067
 *     ('trigger_rate', UINT2),
 *     ('samples_used', SINT2),
 *     ('clutter_filter', '12s'),             # <<<<<<<<<<<<<<
 *     ('number_linear_filter', UINT2),
 *     ('wavelength', SINT4),
 */
  __pyx_tuple__83 = PyTuple_Pack(2, __pyx_n_u_clutter_filter, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__83)) __PYX_ERR(0, 1067, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);

  /* "pyart/io/_sigmetfile.pyx":1087
 *     ('tcf_cal_flags_1', UINT16_T),
 *     ('tcf_cal_flags_2', UINT16_T),
 *     ('spare_1', '18s'),             # 18 bytes             # <<<<<<<<<<<<<<
 *     ('standard_parallel_1', BIN4),
 *     ('standard_parallel_2', BIN4),
 */
  __pyx_tuple__84 = PyTuple_Pack(2, __pyx_n_u_spare_1, __pyx_kp_u_18s); if (unlikely(!__pyx_tuple__84)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__84);
  __Pyx_GIVEREF(__pyx_tuple__84);

  /* "pyart/io/_sigmetfile.pyx":1099
 *     ('longitude_projection', BIN4),
 *     ('product_sequence_number', SINT2),
 *     ('spare_2', '32s'),             # 32 bytes             # <<<<<<<<<<<<<<
 *     ('melting_level', SINT2),
 *     ('radar_height_above_reference', SINT2),
 */
  __pyx_tuple__85 = PyTuple_Pack(2, __pyx_n_u_spare_2, __pyx_kp_u_32s); if (unlikely(!__pyx_tuple__85)) __PYX_ERR(0, 1099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__85);
  __Pyx_GIVEREF(__pyx_tuple__85);

  /* "pyart/io/_sigmetfile.pyx":1105
 *     ('mean_wind_speed', UINT1),
 *     ('mean_wind_direction', BIN1),
 *     ('spare_3', '2s'),              # 2 bytes             # <<<<<<<<<<<<<<
 *     ('tz_name', '8s'),
 *     ('extended_product_header_offset', UINT4),
 */
  __pyx_tuple__86 = PyTuple_Pack(2, __pyx_n_u_spare_3, __pyx_kp_u_2s); if (unlikely(!__pyx_tuple__86)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);

  /* "pyart/io/_sigmetfile.pyx":1106
 *     ('mean_wind_direction', BIN1),
 *     ('spare_3', '2s'),              # 2 bytes
 *     ('tz_name', '8s'),             # <<<<<<<<<<<<<<
 *     ('extended_product_header_offset', UINT4),
 *     ('spare_4', '4s'),              # 4 bytes
 */
  __pyx_tuple__87 = PyTuple_Pack(2, __pyx_n_u_tz_name, __pyx_kp_u_8s); if (unlikely(!__pyx_tuple__87)) __PYX_ERR(0, 1106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__87);
  __Pyx_GIVEREF(__pyx_tuple__87);

  /* "pyart/io/_sigmetfile.pyx":1108
 *     ('tz_name', '8s'),
 *     ('extended_product_header_offset', UINT4),
 *     ('spare_4', '4s'),              # 4 bytes             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__88 = PyTuple_Pack(2, __pyx_n_u_spare_4, __pyx_kp_u_4s); if (unlikely(!__pyx_tuple__88)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__88);
  __Pyx_GIVEREF(__pyx_tuple__88);

  /* "pyart/io/_sigmetfile.pyx":1114
 * INGEST_HEADER = (
 *     ('structure_header', '12s'),        # 12 bytes: structure_header
 *     ('ingest_configuration', '480s'),   # 480 bytes: ingest_configuration             # <<<<<<<<<<<<<<
 *     ('task_configuration', '2612s'),    # 2612 bytes: task_configuration
 *     ('spare_0', '732s'),                # 732 bytes
 */
  __pyx_tuple__89 = PyTuple_Pack(2, __pyx_n_u_ingest_configuration, __pyx_kp_u_480s); if (unlikely(!__pyx_tuple__89)) __PYX_ERR(0, 1114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__89);
  __Pyx_GIVEREF(__pyx_tuple__89);

  /* "pyart/io/_sigmetfile.pyx":1115
 *     ('structure_header', '12s'),        # 12 bytes: structure_header
 *     ('ingest_configuration', '480s'),   # 480 bytes: ingest_configuration
 *     ('task_configuration', '2612s'),    # 2612 bytes: task_configuration             # <<<<<<<<<<<<<<
 *     ('spare_0', '732s'),                # 732 bytes
 *     ('gparm', '128s'),                  # 128 bytes
 */
  __pyx_tuple__90 = PyTuple_Pack(2, __pyx_n_u_task_configuration, __pyx_kp_u_2612s); if (unlikely(!__pyx_tuple__90)) __PYX_ERR(0, 1115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__90);
  __Pyx_GIVEREF(__pyx_tuple__90);

  /* "pyart/io/_sigmetfile.pyx":1116
 *     ('ingest_configuration', '480s'),   # 480 bytes: ingest_configuration
 *     ('task_configuration', '2612s'),    # 2612 bytes: task_configuration
 *     ('spare_0', '732s'),                # 732 bytes             # <<<<<<<<<<<<<<
 *     ('gparm', '128s'),                  # 128 bytes
 *     ('reserved', '920s'),               # 920 bytes
 */
  __pyx_tuple__91 = PyTuple_Pack(2, __pyx_n_u_spare_0, __pyx_kp_u_732s); if (unlikely(!__pyx_tuple__91)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);

  /* "pyart/io/_sigmetfile.pyx":1117
 *     ('task_configuration', '2612s'),    # 2612 bytes: task_configuration
 *     ('spare_0', '732s'),                # 732 bytes
 *     ('gparm', '128s'),                  # 128 bytes             # <<<<<<<<<<<<<<
 *     ('reserved', '920s'),               # 920 bytes
 * )
 */
  __pyx_tuple__92 = PyTuple_Pack(2, __pyx_n_u_gparm, __pyx_kp_u_128s); if (unlikely(!__pyx_tuple__92)) __PYX_ERR(0, 1117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__92);
  __Pyx_GIVEREF(__pyx_tuple__92);

  /* "pyart/io/_sigmetfile.pyx":1118
 *     ('spare_0', '732s'),                # 732 bytes
 *     ('gparm', '128s'),                  # 128 bytes
 *     ('reserved', '920s'),               # 920 bytes             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__93 = PyTuple_Pack(2, __pyx_n_u_reserved, __pyx_kp_u_920s); if (unlikely(!__pyx_tuple__93)) __PYX_ERR(0, 1118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);

  /* "pyart/io/_sigmetfile.pyx":1113
 * # 4884 bytes ingest_header Structure (section 4.2.16, page 40)
 * INGEST_HEADER = (
 *     ('structure_header', '12s'),        # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('ingest_configuration', '480s'),   # 480 bytes: ingest_configuration
 *     ('task_configuration', '2612s'),    # 2612 bytes: task_configuration
 */
  __pyx_tuple__94 = PyTuple_Pack(6, __pyx_tuple__59, __pyx_tuple__89, __pyx_tuple__90, __pyx_tuple__91, __pyx_tuple__92, __pyx_tuple__93); if (unlikely(!__pyx_tuple__94)) __PYX_ERR(0, 1113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__94);
  __Pyx_GIVEREF(__pyx_tuple__94);

  /* "pyart/io/_sigmetfile.pyx":1123
 * # 480 bytes ingest_configuration Structure (section 4.2.14, page 38)
 * INGEST_CONFIGURATION = (
 *     ('filename', '80s'),             # <<<<<<<<<<<<<<
 *     ('number_files', SINT2),
 *     ('number_sweeps_completed', SINT2),
 */
  __pyx_tuple__95 = PyTuple_Pack(2, __pyx_n_u_filename, __pyx_kp_u_80s); if (unlikely(!__pyx_tuple__95)) __PYX_ERR(0, 1123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__95);
  __Pyx_GIVEREF(__pyx_tuple__95);

  /* "pyart/io/_sigmetfile.pyx":1127
 *     ('number_sweeps_completed', SINT2),
 *     ('total_size', SINT4),
 *     ('volume_scan_start_time', '12s'),  # 12 bytes: ymds_time             # <<<<<<<<<<<<<<
 *     ('spare_0', '12s'),                 # 12 bytes
 *     ('ray_header_bytes', SINT2),
 */
  __pyx_tuple__96 = PyTuple_Pack(2, __pyx_n_u_volume_scan_start_time, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__96)) __PYX_ERR(0, 1127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);

  /* "pyart/io/_sigmetfile.pyx":1128
 *     ('total_size', SINT4),
 *     ('volume_scan_start_time', '12s'),  # 12 bytes: ymds_time
 *     ('spare_0', '12s'),                 # 12 bytes             # <<<<<<<<<<<<<<
 *     ('ray_header_bytes', SINT2),
 *     ('extended_ray_header_bytes', SINT2),
 */
  __pyx_tuple__97 = PyTuple_Pack(2, __pyx_n_u_spare_0, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__97)) __PYX_ERR(0, 1128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__97);
  __Pyx_GIVEREF(__pyx_tuple__97);

  /* "pyart/io/_sigmetfile.pyx":1133
 *     ('number_task_config_table', SINT2),
 *     ('playback_version', SINT2),
 *     ('spare_1', '4s'),                  # 4 bytes             # <<<<<<<<<<<<<<
 *     ('iris_version', '8s'),
 *     ('hardware_site', '16s'),
 */
  __pyx_tuple__98 = PyTuple_Pack(2, __pyx_n_u_spare_1, __pyx_kp_u_4s); if (unlikely(!__pyx_tuple__98)) __PYX_ERR(0, 1133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);

  /* "pyart/io/_sigmetfile.pyx":1134
 *     ('playback_version', SINT2),
 *     ('spare_1', '4s'),                  # 4 bytes
 *     ('iris_version', '8s'),             # <<<<<<<<<<<<<<
 *     ('hardware_site', '16s'),
 *     ('gmt_offset_minutes_local', SINT2),
 */
  __pyx_tuple__99 = PyTuple_Pack(2, __pyx_n_u_iris_version, __pyx_kp_u_8s); if (unlikely(!__pyx_tuple__99)) __PYX_ERR(0, 1134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__99);
  __Pyx_GIVEREF(__pyx_tuple__99);

  /* "pyart/io/_sigmetfile.pyx":1135
 *     ('spare_1', '4s'),                  # 4 bytes
 *     ('iris_version', '8s'),
 *     ('hardware_site', '16s'),             # <<<<<<<<<<<<<<
 *     ('gmt_offset_minutes_local', SINT2),
 *     ('site_name', '16s'),
 */
  __pyx_tuple__100 = PyTuple_Pack(2, __pyx_n_u_hardware_site, __pyx_kp_u_16s); if (unlikely(!__pyx_tuple__100)) __PYX_ERR(0, 1135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__100);
  __Pyx_GIVEREF(__pyx_tuple__100);

  /* "pyart/io/_sigmetfile.pyx":1156
 *     ('fault_status', UINT4),
 *     ('melting_layer', SINT2),
 *     ('spare_2', '2s'),              # 2 bytes             # <<<<<<<<<<<<<<
 *     ('local_timezone', '8s'),
 *     ('flags', UINT4),
 */
  __pyx_tuple__101 = PyTuple_Pack(2, __pyx_n_u_spare_2, __pyx_kp_u_2s); if (unlikely(!__pyx_tuple__101)) __PYX_ERR(0, 1156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);

  /* "pyart/io/_sigmetfile.pyx":1157
 *     ('melting_layer', SINT2),
 *     ('spare_2', '2s'),              # 2 bytes
 *     ('local_timezone', '8s'),             # <<<<<<<<<<<<<<
 *     ('flags', UINT4),
 *     ('configuration_name', '16s'),
 */
  __pyx_tuple__102 = PyTuple_Pack(2, __pyx_n_u_local_timezone, __pyx_kp_u_8s); if (unlikely(!__pyx_tuple__102)) __PYX_ERR(0, 1157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__102);
  __Pyx_GIVEREF(__pyx_tuple__102);

  /* "pyart/io/_sigmetfile.pyx":1159
 *     ('local_timezone', '8s'),
 *     ('flags', UINT4),
 *     ('configuration_name', '16s'),             # <<<<<<<<<<<<<<
 *     ('spare_3', '228s')
 * )
 */
  __pyx_tuple__103 = PyTuple_Pack(2, __pyx_n_u_configuration_name, __pyx_kp_u_16s); if (unlikely(!__pyx_tuple__103)) __PYX_ERR(0, 1159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__103);
  __Pyx_GIVEREF(__pyx_tuple__103);

  /* "pyart/io/_sigmetfile.pyx":1160
 *     ('flags', UINT4),
 *     ('configuration_name', '16s'),
 *     ('spare_3', '228s')             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__104 = PyTuple_Pack(2, __pyx_n_u_spare_3, __pyx_kp_u_228s); if (unlikely(!__pyx_tuple__104)) __PYX_ERR(0, 1160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);

  /* "pyart/io/_sigmetfile.pyx":1166
 * TASK_CONFIGURATION = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header
 *     ('task_sched_info', '120s'),    # 120 bytes: task_sched_info             # <<<<<<<<<<<<<<
 *     ('task_dsp_info', '320s'),      # 320 bytes: task_dsp_info
 *     ('task_calib_info', '320s'),    # 320 bytes: task_calib_info
 */
  __pyx_tuple__105 = PyTuple_Pack(2, __pyx_n_u_task_sched_info, __pyx_kp_u_120s); if (unlikely(!__pyx_tuple__105)) __PYX_ERR(0, 1166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__105);
  __Pyx_GIVEREF(__pyx_tuple__105);

  /* "pyart/io/_sigmetfile.pyx":1167
 *     ('structure_header', '12s'),    # 12 bytes: structure_header
 *     ('task_sched_info', '120s'),    # 120 bytes: task_sched_info
 *     ('task_dsp_info', '320s'),      # 320 bytes: task_dsp_info             # <<<<<<<<<<<<<<
 *     ('task_calib_info', '320s'),    # 320 bytes: task_calib_info
 *     ('task_range_info', '160s'),    # 160 bytes: task_range_info
 */
  __pyx_tuple__106 = PyTuple_Pack(2, __pyx_n_u_task_dsp_info, __pyx_kp_u_320s); if (unlikely(!__pyx_tuple__106)) __PYX_ERR(0, 1167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__106);
  __Pyx_GIVEREF(__pyx_tuple__106);

  /* "pyart/io/_sigmetfile.pyx":1168
 *     ('task_sched_info', '120s'),    # 120 bytes: task_sched_info
 *     ('task_dsp_info', '320s'),      # 320 bytes: task_dsp_info
 *     ('task_calib_info', '320s'),    # 320 bytes: task_calib_info             # <<<<<<<<<<<<<<
 *     ('task_range_info', '160s'),    # 160 bytes: task_range_info
 *     ('task_scan_info', '320s'),     # 320 bytes: task_scan_info
 */
  __pyx_tuple__107 = PyTuple_Pack(2, __pyx_n_u_task_calib_info, __pyx_kp_u_320s); if (unlikely(!__pyx_tuple__107)) __PYX_ERR(0, 1168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__107);
  __Pyx_GIVEREF(__pyx_tuple__107);

  /* "pyart/io/_sigmetfile.pyx":1169
 *     ('task_dsp_info', '320s'),      # 320 bytes: task_dsp_info
 *     ('task_calib_info', '320s'),    # 320 bytes: task_calib_info
 *     ('task_range_info', '160s'),    # 160 bytes: task_range_info             # <<<<<<<<<<<<<<
 *     ('task_scan_info', '320s'),     # 320 bytes: task_scan_info
 *     ('task_misc_info', '320s'),     # 320 bytes: task_misc_info
 */
  __pyx_tuple__108 = PyTuple_Pack(2, __pyx_n_u_task_range_info, __pyx_kp_u_160s); if (unlikely(!__pyx_tuple__108)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);

  /* "pyart/io/_sigmetfile.pyx":1170
 *     ('task_calib_info', '320s'),    # 320 bytes: task_calib_info
 *     ('task_range_info', '160s'),    # 160 bytes: task_range_info
 *     ('task_scan_info', '320s'),     # 320 bytes: task_scan_info             # <<<<<<<<<<<<<<
 *     ('task_misc_info', '320s'),     # 320 bytes: task_misc_info
 *     ('task_end_info', '320s'),      # 320 bytes: task_end_info
 */
  __pyx_tuple__109 = PyTuple_Pack(2, __pyx_n_u_task_scan_info, __pyx_kp_u_320s); if (unlikely(!__pyx_tuple__109)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__109);
  __Pyx_GIVEREF(__pyx_tuple__109);

  /* "pyart/io/_sigmetfile.pyx":1171
 *     ('task_range_info', '160s'),    # 160 bytes: task_range_info
 *     ('task_scan_info', '320s'),     # 320 bytes: task_scan_info
 *     ('task_misc_info', '320s'),     # 320 bytes: task_misc_info             # <<<<<<<<<<<<<<
 *     ('task_end_info', '320s'),      # 320 bytes: task_end_info
 *     ('comments', '720s'),
 */
  __pyx_tuple__110 = PyTuple_Pack(2, __pyx_n_u_task_misc_info, __pyx_kp_u_320s); if (unlikely(!__pyx_tuple__110)) __PYX_ERR(0, 1171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__110);
  __Pyx_GIVEREF(__pyx_tuple__110);

  /* "pyart/io/_sigmetfile.pyx":1172
 *     ('task_scan_info', '320s'),     # 320 bytes: task_scan_info
 *     ('task_misc_info', '320s'),     # 320 bytes: task_misc_info
 *     ('task_end_info', '320s'),      # 320 bytes: task_end_info             # <<<<<<<<<<<<<<
 *     ('comments', '720s'),
 * )
 */
  __pyx_tuple__111 = PyTuple_Pack(2, __pyx_n_u_task_end_info, __pyx_kp_u_320s); if (unlikely(!__pyx_tuple__111)) __PYX_ERR(0, 1172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__111);
  __Pyx_GIVEREF(__pyx_tuple__111);

  /* "pyart/io/_sigmetfile.pyx":1173
 *     ('task_misc_info', '320s'),     # 320 bytes: task_misc_info
 *     ('task_end_info', '320s'),      # 320 bytes: task_end_info
 *     ('comments', '720s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__112 = PyTuple_Pack(2, __pyx_n_u_comments, __pyx_kp_u_720s); if (unlikely(!__pyx_tuple__112)) __PYX_ERR(0, 1173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__112);
  __Pyx_GIVEREF(__pyx_tuple__112);

  /* "pyart/io/_sigmetfile.pyx":1165
 * # 2612 bytes: task_configuration Structure (section 4.2.50, page 61)
 * TASK_CONFIGURATION = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('task_sched_info', '120s'),    # 120 bytes: task_sched_info
 *     ('task_dsp_info', '320s'),      # 320 bytes: task_dsp_info
 */
  __pyx_tuple__113 = PyTuple_Pack(9, __pyx_tuple__59, __pyx_tuple__105, __pyx_tuple__106, __pyx_tuple__107, __pyx_tuple__108, __pyx_tuple__109, __pyx_tuple__110, __pyx_tuple__111, __pyx_tuple__112); if (unlikely(!__pyx_tuple__113)) __PYX_ERR(0, 1165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__113);
  __Pyx_GIVEREF(__pyx_tuple__113);

  /* "pyart/io/_sigmetfile.pyx":1185
 *     ('last_run_day', SINT4),
 *     ('flag', UINT2),
 *     ('spare_0', '94s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__114 = PyTuple_Pack(2, __pyx_n_u_spare_0, __pyx_kp_u_94s); if (unlikely(!__pyx_tuple__114)) __PYX_ERR(0, 1185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__114);
  __Pyx_GIVEREF(__pyx_tuple__114);

  /* "pyart/io/_sigmetfile.pyx":1192
 *     ('major_mode', UINT2),
 *     ('dsp_type', UINT2),
 *     ('current_data_type_mask', '24s'),      # 24 bytes: dsp_data_mask             # <<<<<<<<<<<<<<
 *     ('original_data_type_mask', '24s'),     # 24 bytes: dsp_data_mask
 *     ('task_dsp_mode', '32s'),               # 32 bytes: task_dsp_mode_batch
 */
  __pyx_tuple__115 = PyTuple_Pack(2, __pyx_n_u_current_data_type_mask, __pyx_kp_u_24s); if (unlikely(!__pyx_tuple__115)) __PYX_ERR(0, 1192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__115);
  __Pyx_GIVEREF(__pyx_tuple__115);

  /* "pyart/io/_sigmetfile.pyx":1193
 *     ('dsp_type', UINT2),
 *     ('current_data_type_mask', '24s'),      # 24 bytes: dsp_data_mask
 *     ('original_data_type_mask', '24s'),     # 24 bytes: dsp_data_mask             # <<<<<<<<<<<<<<
 *     ('task_dsp_mode', '32s'),               # 32 bytes: task_dsp_mode_batch
 *     ('spare_0', '52s'),
 */
  __pyx_tuple__116 = PyTuple_Pack(2, __pyx_n_u_original_data_type_mask, __pyx_kp_u_24s); if (unlikely(!__pyx_tuple__116)) __PYX_ERR(0, 1193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__116);
  __Pyx_GIVEREF(__pyx_tuple__116);

  /* "pyart/io/_sigmetfile.pyx":1194
 *     ('current_data_type_mask', '24s'),      # 24 bytes: dsp_data_mask
 *     ('original_data_type_mask', '24s'),     # 24 bytes: dsp_data_mask
 *     ('task_dsp_mode', '32s'),               # 32 bytes: task_dsp_mode_batch             # <<<<<<<<<<<<<<
 *     ('spare_0', '52s'),
 *     ('prf', SINT4),
 */
  __pyx_tuple__117 = PyTuple_Pack(2, __pyx_n_u_task_dsp_mode, __pyx_kp_u_32s); if (unlikely(!__pyx_tuple__117)) __PYX_ERR(0, 1194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__117);
  __Pyx_GIVEREF(__pyx_tuple__117);

  /* "pyart/io/_sigmetfile.pyx":1195
 *     ('original_data_type_mask', '24s'),     # 24 bytes: dsp_data_mask
 *     ('task_dsp_mode', '32s'),               # 32 bytes: task_dsp_mode_batch
 *     ('spare_0', '52s'),             # <<<<<<<<<<<<<<
 *     ('prf', SINT4),
 *     ('pulse_width', SINT4),
 */
  __pyx_tuple__118 = PyTuple_Pack(2, __pyx_n_u_spare_0, __pyx_kp_u_52s); if (unlikely(!__pyx_tuple__118)) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__118);
  __Pyx_GIVEREF(__pyx_tuple__118);

  /* "pyart/io/_sigmetfile.pyx":1203
 *     ('sample_size', SINT2),
 *     ('gain_control_flag', UINT2),
 *     ('clutter_filter_name', '12s'),             # <<<<<<<<<<<<<<
 *     ('linear_filter_first_bin', UINT1),
 *     ('log_filter_first_bin', UINT1),
 */
  __pyx_tuple__119 = PyTuple_Pack(2, __pyx_n_u_clutter_filter_name, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__119)) __PYX_ERR(0, 1203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__119);
  __Pyx_GIVEREF(__pyx_tuple__119);

  /* "pyart/io/_sigmetfile.pyx":1212
 *     ('ray_header_mask', UINT4),
 *     ('playback_flag', UINT2),
 *     ('spare_1', '2s'),             # <<<<<<<<<<<<<<
 *     ('custom_ray_header_name', '16s'),
 *     ('spare_2', '120s')
 */
  __pyx_tuple__120 = PyTuple_Pack(2, __pyx_n_u_spare_1, __pyx_kp_u_2s); if (unlikely(!__pyx_tuple__120)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__120);
  __Pyx_GIVEREF(__pyx_tuple__120);

  /* "pyart/io/_sigmetfile.pyx":1213
 *     ('playback_flag', UINT2),
 *     ('spare_1', '2s'),
 *     ('custom_ray_header_name', '16s'),             # <<<<<<<<<<<<<<
 *     ('spare_2', '120s')
 * )
 */
  __pyx_tuple__121 = PyTuple_Pack(2, __pyx_n_u_custom_ray_header_name, __pyx_kp_u_16s); if (unlikely(!__pyx_tuple__121)) __PYX_ERR(0, 1213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__121);
  __Pyx_GIVEREF(__pyx_tuple__121);

  /* "pyart/io/_sigmetfile.pyx":1214
 *     ('spare_1', '2s'),
 *     ('custom_ray_header_name', '16s'),
 *     ('spare_2', '120s')             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__122 = PyTuple_Pack(2, __pyx_n_u_spare_2, __pyx_kp_u_120s); if (unlikely(!__pyx_tuple__122)) __PYX_ERR(0, 1214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__122);
  __Pyx_GIVEREF(__pyx_tuple__122);

  /* "pyart/io/_sigmetfile.pyx":1236
 *     ('velocity_unfolding_threshold', SINT2),
 *     ('width_unfolding_threshold', SINT2),
 *     ('spare_0', '18s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__123 = PyTuple_Pack(2, __pyx_n_u_spare_0, __pyx_kp_u_18s); if (unlikely(!__pyx_tuple__123)) __PYX_ERR(0, 1236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__123);
  __Pyx_GIVEREF(__pyx_tuple__123);

  /* "pyart/io/_sigmetfile.pyx":1246
 *     ('sqi_threshold', SINT2),
 *     ('power_threshold', SINT2),
 *     ('spare_0', '8s'),             # <<<<<<<<<<<<<<
 *     ('reflectivity_calibration', SINT2),
 *     ('uncorrected_reflectivity_threshold_flags', UINT2),
 */
  __pyx_tuple__124 = PyTuple_Pack(2, __pyx_n_u_spare_0, __pyx_kp_u_8s); if (unlikely(!__pyx_tuple__124)) __PYX_ERR(0, 1246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__124);
  __Pyx_GIVEREF(__pyx_tuple__124);

  /* "pyart/io/_sigmetfile.pyx":1253
 *     ('width_threshold_flags', UINT2),
 *     ('zdr_threshold_flags', UINT2),
 *     ('spare_1', '6s'),             # <<<<<<<<<<<<<<
 *     ('flags', UINT2),
 *     ('spare_2', '2s'),
 */
  __pyx_tuple__125 = PyTuple_Pack(2, __pyx_n_u_spare_1, __pyx_kp_u_6s); if (unlikely(!__pyx_tuple__125)) __PYX_ERR(0, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__125);
  __Pyx_GIVEREF(__pyx_tuple__125);

  /* "pyart/io/_sigmetfile.pyx":1268
 *     ('reciever_bandwidth', UINT2),
 *     ('flags2', UINT16_T),
 *     ('spare_3', '256s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__126 = PyTuple_Pack(2, __pyx_n_u_spare_3, __pyx_kp_u_256s); if (unlikely(!__pyx_tuple__126)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);

  /* "pyart/io/_sigmetfile.pyx":1281
 *     ('variable_range_bin_flag', UINT2),
 *     ('range_bin_averaging_flag', SINT2),
 *     ('spare_0', '136s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__127 = PyTuple_Pack(2, __pyx_n_u_spare_0, __pyx_kp_u_136s); if (unlikely(!__pyx_tuple__127)) __PYX_ERR(0, 1281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__127);
  __Pyx_GIVEREF(__pyx_tuple__127);

  /* "pyart/io/_sigmetfile.pyx":1288
 *     ('antenna_scan_mode', UINT2),
 *     ('angular_resolution_desired', SINT2),
 *     ('spare_0', '2s'),             # <<<<<<<<<<<<<<
 *     ('number_sweeps', SINT2),
 *     ('task_scan_type_scan_info', '200s'),   # 200 bytes: task_foo_scan_info
 */
  __pyx_tuple__128 = PyTuple_Pack(2, __pyx_n_u_spare_0, __pyx_kp_u_2s); if (unlikely(!__pyx_tuple__128)) __PYX_ERR(0, 1288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);

  /* "pyart/io/_sigmetfile.pyx":1290
 *     ('spare_0', '2s'),
 *     ('number_sweeps', SINT2),
 *     ('task_scan_type_scan_info', '200s'),   # 200 bytes: task_foo_scan_info             # <<<<<<<<<<<<<<
 *     ('spare_1', '112s'),
 * )
 */
  __pyx_tuple__129 = PyTuple_Pack(2, __pyx_n_u_task_scan_type_scan_info, __pyx_kp_u_200s); if (unlikely(!__pyx_tuple__129)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);

  /* "pyart/io/_sigmetfile.pyx":1291
 *     ('number_sweeps', SINT2),
 *     ('task_scan_type_scan_info', '200s'),   # 200 bytes: task_foo_scan_info
 *     ('spare_1', '112s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__130 = PyTuple_Pack(2, __pyx_n_u_spare_1, __pyx_kp_u_112s); if (unlikely(!__pyx_tuple__130)) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__130);
  __Pyx_GIVEREF(__pyx_tuple__130);

  /* "pyart/io/_sigmetfile.pyx":1298
 *     ('lower_elevation_limit', UINT2),
 *     ('upper_elevation_limit', UINT2),
 *     ('azimuth_list', '80s'),            # UINT2[40]             # <<<<<<<<<<<<<<
 *     ('spare_0', '115s'),
 *     ('start_first_sector_flag', 'c'),   # unknown type
 */
  __pyx_tuple__131 = PyTuple_Pack(2, __pyx_n_u_azimuth_list, __pyx_kp_u_80s); if (unlikely(!__pyx_tuple__131)) __PYX_ERR(0, 1298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__131);
  __Pyx_GIVEREF(__pyx_tuple__131);

  /* "pyart/io/_sigmetfile.pyx":1299
 *     ('upper_elevation_limit', UINT2),
 *     ('azimuth_list', '80s'),            # UINT2[40]
 *     ('spare_0', '115s'),             # <<<<<<<<<<<<<<
 *     ('start_first_sector_flag', 'c'),   # unknown type
 * )
 */
  __pyx_tuple__132 = PyTuple_Pack(2, __pyx_n_u_spare_0, __pyx_kp_u_115s); if (unlikely(!__pyx_tuple__132)) __PYX_ERR(0, 1299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);

  /* "pyart/io/_sigmetfile.pyx":1300
 *     ('azimuth_list', '80s'),            # UINT2[40]
 *     ('spare_0', '115s'),
 *     ('start_first_sector_flag', 'c'),   # unknown type             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__133 = PyTuple_Pack(2, __pyx_n_u_start_first_sector_flag, __pyx_n_u_c); if (unlikely(!__pyx_tuple__133)) __PYX_ERR(0, 1300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__133);
  __Pyx_GIVEREF(__pyx_tuple__133);

  /* "pyart/io/_sigmetfile.pyx":1307
 *     ('left_azimuth_limit', BIN2),
 *     ('right_azimuth_limit', BIN2),
 *     ('elevation_list', '80s'),          # UINT2[40]             # <<<<<<<<<<<<<<
 *     ('spare_0', '115s'),
 *     ('start_first_section_flag', 'c'),  # unknown type
 */
  __pyx_tuple__134 = PyTuple_Pack(2, __pyx_n_u_elevation_list, __pyx_kp_u_80s); if (unlikely(!__pyx_tuple__134)) __PYX_ERR(0, 1307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__134);
  __Pyx_GIVEREF(__pyx_tuple__134);

  /* "pyart/io/_sigmetfile.pyx":1309
 *     ('elevation_list', '80s'),          # UINT2[40]
 *     ('spare_0', '115s'),
 *     ('start_first_section_flag', 'c'),  # unknown type             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__135 = PyTuple_Pack(2, __pyx_n_u_start_first_section_flag, __pyx_n_u_c); if (unlikely(!__pyx_tuple__135)) __PYX_ERR(0, 1309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__135);
  __Pyx_GIVEREF(__pyx_tuple__135);

  /* "pyart/io/_sigmetfile.pyx":1316
 *     ('first_azimuth', UINT2),
 *     ('first_elevation', UINT2),
 *     ('filename', '12s'),             # <<<<<<<<<<<<<<
 *     ('spare_0', '184s'),
 * )
 */
  __pyx_tuple__136 = PyTuple_Pack(2, __pyx_n_u_filename, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__136)) __PYX_ERR(0, 1316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__136);
  __Pyx_GIVEREF(__pyx_tuple__136);

  /* "pyart/io/_sigmetfile.pyx":1317
 *     ('first_elevation', UINT2),
 *     ('filename', '12s'),
 *     ('spare_0', '184s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__137 = PyTuple_Pack(2, __pyx_n_u_spare_0, __pyx_kp_u_184s); if (unlikely(!__pyx_tuple__137)) __PYX_ERR(0, 1317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__137);
  __Pyx_GIVEREF(__pyx_tuple__137);

  /* "pyart/io/_sigmetfile.pyx":1323
 * TASK_MANUAL_SCAN_INFO = (
 *     ('flags', UINT2),
 *     ('spare_0', '198s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__138 = PyTuple_Pack(2, __pyx_n_u_spare_0, __pyx_kp_u_198s); if (unlikely(!__pyx_tuple__138)) __PYX_ERR(0, 1323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__138);
  __Pyx_GIVEREF(__pyx_tuple__138);

  /* "pyart/io/_sigmetfile.pyx":1329
 * TASK_MISC_INFO = (
 *     ('wavelength', SINT4),
 *     ('tr_serial_number', '16s'),             # <<<<<<<<<<<<<<
 *     ('transmit_power', SINT4),
 *     ('flags', UINT2),
 */
  __pyx_tuple__139 = PyTuple_Pack(2, __pyx_n_u_tr_serial_number, __pyx_kp_u_16s); if (unlikely(!__pyx_tuple__139)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__139);
  __Pyx_GIVEREF(__pyx_tuple__139);

  /* "pyart/io/_sigmetfile.pyx":1335
 *     ('trucation_height', SINT4),
 *     ('spare_0', '18s'),             # 18 bytes
 *     ('spare_1', '12s'),             # 12 bytes             # <<<<<<<<<<<<<<
 *     ('comment_bytes', SINT2),
 *     ('horizontal_beamwidth', BIN4),
 */
  __pyx_tuple__140 = PyTuple_Pack(2, __pyx_n_u_spare_1, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__140)) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__140);
  __Pyx_GIVEREF(__pyx_tuple__140);

  /* "pyart/io/_sigmetfile.pyx":1339
 *     ('horizontal_beamwidth', BIN4),
 *     ('vertical_beamwidth', BIN4),
 *     ('customer_storage', '40s'),    # 40 bytes, uint4[10]             # <<<<<<<<<<<<<<
 *     ('spare_2', '208s'),
 * )
 */
  __pyx_tuple__141 = PyTuple_Pack(2, __pyx_n_u_customer_storage, __pyx_kp_u_40s); if (unlikely(!__pyx_tuple__141)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__141);
  __Pyx_GIVEREF(__pyx_tuple__141);

  /* "pyart/io/_sigmetfile.pyx":1340
 *     ('vertical_beamwidth', BIN4),
 *     ('customer_storage', '40s'),    # 40 bytes, uint4[10]
 *     ('spare_2', '208s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__142 = PyTuple_Pack(2, __pyx_n_u_spare_2, __pyx_kp_u_208s); if (unlikely(!__pyx_tuple__142)) __PYX_ERR(0, 1340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__142);
  __Pyx_GIVEREF(__pyx_tuple__142);

  /* "pyart/io/_sigmetfile.pyx":1347
 *     ('task_major_number', SINT2),
 *     ('task_minor_number', SINT2),
 *     ('task_configuration_file_name', '12s'),             # <<<<<<<<<<<<<<
 *     ('task_description', '80s'),
 *     ('number_tasks', SINT4),
 */
  __pyx_tuple__143 = PyTuple_Pack(2, __pyx_n_u_task_configuration_file_name, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__143)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);

  /* "pyart/io/_sigmetfile.pyx":1348
 *     ('task_minor_number', SINT2),
 *     ('task_configuration_file_name', '12s'),
 *     ('task_description', '80s'),             # <<<<<<<<<<<<<<
 *     ('number_tasks', SINT4),
 *     ('task_state', UINT2),
 */
  __pyx_tuple__144 = PyTuple_Pack(2, __pyx_n_u_task_description, __pyx_kp_u_80s); if (unlikely(!__pyx_tuple__144)) __PYX_ERR(0, 1348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__144);
  __Pyx_GIVEREF(__pyx_tuple__144);

  /* "pyart/io/_sigmetfile.pyx":1352
 *     ('task_state', UINT2),
 *     ('spare_0', '2s'),
 *     ('task_data_time', '12s'),      # 12 bytes: ymds_time             # <<<<<<<<<<<<<<
 *     ('spare_1', '204s'),
 * )
 */
  __pyx_tuple__145 = PyTuple_Pack(2, __pyx_n_u_task_data_time, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__145)) __PYX_ERR(0, 1352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);

  /* "pyart/io/_sigmetfile.pyx":1353
 *     ('spare_0', '2s'),
 *     ('task_data_time', '12s'),      # 12 bytes: ymds_time
 *     ('spare_1', '204s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__146 = PyTuple_Pack(2, __pyx_n_u_spare_1, __pyx_kp_u_204s); if (unlikely(!__pyx_tuple__146)) __PYX_ERR(0, 1353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__146);
  __Pyx_GIVEREF(__pyx_tuple__146);

  /* "pyart/io/_sigmetfile.pyx":1370
 * INGEST_DATA_HEADER = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header
 *     ('sweep_start_time', '12s'),    # 12 bytes: ymds_time             # <<<<<<<<<<<<<<
 *     ('sweep_number', SINT2),
 *     ('number_rays_sweep', SINT2),
 */
  __pyx_tuple__147 = PyTuple_Pack(2, __pyx_n_u_sweep_start_time, __pyx_kp_u_12s); if (unlikely(!__pyx_tuple__147)) __PYX_ERR(0, 1370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);

  /* "pyart/io/_sigmetfile.pyx":1379
 *     ('bit_per_bin', SINT2),
 *     ('data_type', UINT2),
 *     ('spare_0', '36s')      # 36 bytes             # <<<<<<<<<<<<<<
 * )
 */
  __pyx_tuple__148 = PyTuple_Pack(2, __pyx_n_u_spare_0, __pyx_kp_u_36s); if (unlikely(!__pyx_tuple__148)) __PYX_ERR(0, 1379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__148);
  __Pyx_GIVEREF(__pyx_tuple__148);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_1_ = PyFloat_FromDouble(1.); if (unlikely(!__pyx_float_1_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_2_ = PyFloat_FromDouble(2.); if (unlikely(!__pyx_float_2_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_16_ = PyFloat_FromDouble(16.); if (unlikely(!__pyx_float_16_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_64_ = PyFloat_FromDouble(64.); if (unlikely(!__pyx_float_64_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_75_ = PyFloat_FromDouble(75.); if (unlikely(!__pyx_float_75_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_25 = PyFloat_FromDouble(0.25); if (unlikely(!__pyx_float_0_25)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_100_ = PyFloat_FromDouble(100.); if (unlikely(!__pyx_float_100_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_126_ = PyFloat_FromDouble(126.); if (unlikely(!__pyx_float_126_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_127_ = PyFloat_FromDouble(127.); if (unlikely(!__pyx_float_127_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_128_ = PyFloat_FromDouble(128.); if (unlikely(!__pyx_float_128_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_129_ = PyFloat_FromDouble(129.); if (unlikely(!__pyx_float_129_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_180_ = PyFloat_FromDouble(180.); if (unlikely(!__pyx_float_180_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_253_ = PyFloat_FromDouble(253.); if (unlikely(!__pyx_float_253_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_254_ = PyFloat_FromDouble(254.); if (unlikely(!__pyx_float_254_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_256_ = PyFloat_FromDouble(256.); if (unlikely(!__pyx_float_256_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_360_ = PyFloat_FromDouble(360.); if (unlikely(!__pyx_float_360_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_600_ = PyFloat_FromDouble(600.); if (unlikely(!__pyx_float_600_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_100_0 = PyFloat_FromDouble(100.0); if (unlikely(!__pyx_float_100_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_32768_ = PyFloat_FromDouble(32768.); if (unlikely(!__pyx_float_32768_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_65533_ = PyFloat_FromDouble(65533.); if (unlikely(!__pyx_float_65533_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_65534_ = PyFloat_FromDouble(65534.); if (unlikely(!__pyx_float_65534_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_neg_0_25 = PyFloat_FromDouble(-0.25); if (unlikely(!__pyx_float_neg_0_25)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_neg_9999_0 = PyFloat_FromDouble(-9999.0); if (unlikely(!__pyx_float_neg_9999_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_9 = PyInt_FromLong(9); if (unlikely(!__pyx_int_9)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_11 = PyInt_FromLong(11); if (unlikely(!__pyx_int_11)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_12 = PyInt_FromLong(12); if (unlikely(!__pyx_int_12)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_13 = PyInt_FromLong(13); if (unlikely(!__pyx_int_13)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_14 = PyInt_FromLong(14); if (unlikely(!__pyx_int_14)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_15 = PyInt_FromLong(15); if (unlikely(!__pyx_int_15)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_16 = PyInt_FromLong(16); if (unlikely(!__pyx_int_16)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_17 = PyInt_FromLong(17); if (unlikely(!__pyx_int_17)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_18 = PyInt_FromLong(18); if (unlikely(!__pyx_int_18)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_19 = PyInt_FromLong(19); if (unlikely(!__pyx_int_19)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_20 = PyInt_FromLong(20); if (unlikely(!__pyx_int_20)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_21 = PyInt_FromLong(21); if (unlikely(!__pyx_int_21)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_22 = PyInt_FromLong(22); if (unlikely(!__pyx_int_22)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_23 = PyInt_FromLong(23); if (unlikely(!__pyx_int_23)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_24 = PyInt_FromLong(24); if (unlikely(!__pyx_int_24)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_25 = PyInt_FromLong(25); if (unlikely(!__pyx_int_25)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_26 = PyInt_FromLong(26); if (unlikely(!__pyx_int_26)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_27 = PyInt_FromLong(27); if (unlikely(!__pyx_int_27)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_28 = PyInt_FromLong(28); if (unlikely(!__pyx_int_28)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_29 = PyInt_FromLong(29); if (unlikely(!__pyx_int_29)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_30 = PyInt_FromLong(30); if (unlikely(!__pyx_int_30)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_31 = PyInt_FromLong(31); if (unlikely(!__pyx_int_31)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_32 = PyInt_FromLong(32); if (unlikely(!__pyx_int_32)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_33 = PyInt_FromLong(33); if (unlikely(!__pyx_int_33)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_34 = PyInt_FromLong(34); if (unlikely(!__pyx_int_34)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_35 = PyInt_FromLong(35); if (unlikely(!__pyx_int_35)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_36 = PyInt_FromLong(36); if (unlikely(!__pyx_int_36)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_37 = PyInt_FromLong(37); if (unlikely(!__pyx_int_37)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_38 = PyInt_FromLong(38); if (unlikely(!__pyx_int_38)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_39 = PyInt_FromLong(39); if (unlikely(!__pyx_int_39)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_40 = PyInt_FromLong(40); if (unlikely(!__pyx_int_40)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_41 = PyInt_FromLong(41); if (unlikely(!__pyx_int_41)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_42 = PyInt_FromLong(42); if (unlikely(!__pyx_int_42)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_43 = PyInt_FromLong(43); if (unlikely(!__pyx_int_43)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_44 = PyInt_FromLong(44); if (unlikely(!__pyx_int_44)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_45 = PyInt_FromLong(45); if (unlikely(!__pyx_int_45)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_46 = PyInt_FromLong(46); if (unlikely(!__pyx_int_46)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_47 = PyInt_FromLong(47); if (unlikely(!__pyx_int_47)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_48 = PyInt_FromLong(48); if (unlikely(!__pyx_int_48)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_49 = PyInt_FromLong(49); if (unlikely(!__pyx_int_49)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_50 = PyInt_FromLong(50); if (unlikely(!__pyx_int_50)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_51 = PyInt_FromLong(51); if (unlikely(!__pyx_int_51)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_52 = PyInt_FromLong(52); if (unlikely(!__pyx_int_52)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_53 = PyInt_FromLong(53); if (unlikely(!__pyx_int_53)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_54 = PyInt_FromLong(54); if (unlikely(!__pyx_int_54)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_55 = PyInt_FromLong(55); if (unlikely(!__pyx_int_55)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_56 = PyInt_FromLong(56); if (unlikely(!__pyx_int_56)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_57 = PyInt_FromLong(57); if (unlikely(!__pyx_int_57)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_58 = PyInt_FromLong(58); if (unlikely(!__pyx_int_58)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_59 = PyInt_FromLong(59); if (unlikely(!__pyx_int_59)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_60 = PyInt_FromLong(60); if (unlikely(!__pyx_int_60)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_61 = PyInt_FromLong(61); if (unlikely(!__pyx_int_61)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_62 = PyInt_FromLong(62); if (unlikely(!__pyx_int_62)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_63 = PyInt_FromLong(63); if (unlikely(!__pyx_int_63)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_64 = PyInt_FromLong(64); if (unlikely(!__pyx_int_64)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_65 = PyInt_FromLong(65); if (unlikely(!__pyx_int_65)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_66 = PyInt_FromLong(66); if (unlikely(!__pyx_int_66)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_67 = PyInt_FromLong(67); if (unlikely(!__pyx_int_67)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_68 = PyInt_FromLong(68); if (unlikely(!__pyx_int_68)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_69 = PyInt_FromLong(69); if (unlikely(!__pyx_int_69)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_70 = PyInt_FromLong(70); if (unlikely(!__pyx_int_70)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_71 = PyInt_FromLong(71); if (unlikely(!__pyx_int_71)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_72 = PyInt_FromLong(72); if (unlikely(!__pyx_int_72)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_73 = PyInt_FromLong(73); if (unlikely(!__pyx_int_73)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_74 = PyInt_FromLong(74); if (unlikely(!__pyx_int_74)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_75 = PyInt_FromLong(75); if (unlikely(!__pyx_int_75)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_76 = PyInt_FromLong(76); if (unlikely(!__pyx_int_76)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_77 = PyInt_FromLong(77); if (unlikely(!__pyx_int_77)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_78 = PyInt_FromLong(78); if (unlikely(!__pyx_int_78)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_79 = PyInt_FromLong(79); if (unlikely(!__pyx_int_79)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_80 = PyInt_FromLong(80); if (unlikely(!__pyx_int_80)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_81 = PyInt_FromLong(81); if (unlikely(!__pyx_int_81)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_82 = PyInt_FromLong(82); if (unlikely(!__pyx_int_82)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_83 = PyInt_FromLong(83); if (unlikely(!__pyx_int_83)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_84 = PyInt_FromLong(84); if (unlikely(!__pyx_int_84)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_85 = PyInt_FromLong(85); if (unlikely(!__pyx_int_85)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_86 = PyInt_FromLong(86); if (unlikely(!__pyx_int_86)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_87 = PyInt_FromLong(87); if (unlikely(!__pyx_int_87)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_88 = PyInt_FromLong(88); if (unlikely(!__pyx_int_88)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_89 = PyInt_FromLong(89); if (unlikely(!__pyx_int_89)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_90 = PyInt_FromLong(90); if (unlikely(!__pyx_int_90)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_91 = PyInt_FromLong(91); if (unlikely(!__pyx_int_91)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_92 = PyInt_FromLong(92); if (unlikely(!__pyx_int_92)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_93 = PyInt_FromLong(93); if (unlikely(!__pyx_int_93)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_94 = PyInt_FromLong(94); if (unlikely(!__pyx_int_94)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_95 = PyInt_FromLong(95); if (unlikely(!__pyx_int_95)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_96 = PyInt_FromLong(96); if (unlikely(!__pyx_int_96)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_97 = PyInt_FromLong(97); if (unlikely(!__pyx_int_97)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_98 = PyInt_FromLong(98); if (unlikely(!__pyx_int_98)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_99 = PyInt_FromLong(99); if (unlikely(!__pyx_int_99)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100 = PyInt_FromLong(100); if (unlikely(!__pyx_int_100)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_101 = PyInt_FromLong(101); if (unlikely(!__pyx_int_101)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_102 = PyInt_FromLong(102); if (unlikely(!__pyx_int_102)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_103 = PyInt_FromLong(103); if (unlikely(!__pyx_int_103)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_104 = PyInt_FromLong(104); if (unlikely(!__pyx_int_104)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_105 = PyInt_FromLong(105); if (unlikely(!__pyx_int_105)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_106 = PyInt_FromLong(106); if (unlikely(!__pyx_int_106)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_107 = PyInt_FromLong(107); if (unlikely(!__pyx_int_107)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_108 = PyInt_FromLong(108); if (unlikely(!__pyx_int_108)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_109 = PyInt_FromLong(109); if (unlikely(!__pyx_int_109)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_110 = PyInt_FromLong(110); if (unlikely(!__pyx_int_110)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_111 = PyInt_FromLong(111); if (unlikely(!__pyx_int_111)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_112 = PyInt_FromLong(112); if (unlikely(!__pyx_int_112)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_113 = PyInt_FromLong(113); if (unlikely(!__pyx_int_113)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_114 = PyInt_FromLong(114); if (unlikely(!__pyx_int_114)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_115 = PyInt_FromLong(115); if (unlikely(!__pyx_int_115)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_116 = PyInt_FromLong(116); if (unlikely(!__pyx_int_116)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_117 = PyInt_FromLong(117); if (unlikely(!__pyx_int_117)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_118 = PyInt_FromLong(118); if (unlikely(!__pyx_int_118)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_119 = PyInt_FromLong(119); if (unlikely(!__pyx_int_119)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_120 = PyInt_FromLong(120); if (unlikely(!__pyx_int_120)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_121 = PyInt_FromLong(121); if (unlikely(!__pyx_int_121)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_122 = PyInt_FromLong(122); if (unlikely(!__pyx_int_122)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_123 = PyInt_FromLong(123); if (unlikely(!__pyx_int_123)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_124 = PyInt_FromLong(124); if (unlikely(!__pyx_int_124)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_125 = PyInt_FromLong(125); if (unlikely(!__pyx_int_125)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_126 = PyInt_FromLong(126); if (unlikely(!__pyx_int_126)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_127 = PyInt_FromLong(127); if (unlikely(!__pyx_int_127)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_128 = PyInt_FromLong(128); if (unlikely(!__pyx_int_128)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_255 = PyInt_FromLong(255); if (unlikely(!__pyx_int_255)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_640 = PyInt_FromLong(640); if (unlikely(!__pyx_int_640)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4884 = PyInt_FromLong(4884); if (unlikely(!__pyx_int_4884)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_6144 = PyInt_FromLong(6144); if (unlikely(!__pyx_int_6144)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_65536 = PyInt_FromLong(65536L); if (unlikely(!__pyx_int_65536)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4294967296 = PyInt_FromString((char *)"4294967296", 0, 0); if (unlikely(!__pyx_int_4294967296)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* NumpyImportArray.init */
  /*
 * Cython has automatically inserted a call to _import_array since
 * you didn't include one when you cimported numpy. To disable this
 * add the line
 *   <void>numpy._import_array
 */
#ifdef NPY_FEATURE_VERSION
#ifndef NO_IMPORT_ARRAY
if (unlikely(_import_array() == -1)) {
    PyErr_SetString(PyExc_ImportError, "numpy.core.multiarray failed to import "
    "(auto-generated because you didn't call 'numpy.import_array()' after cimporting numpy; "
    "use '<void>numpy._import_array' to disable if you are certain you don't need it).");
}
#endif
#endif

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_5pyart_2io_11_sigmetfile_SigmetFile = &__pyx_vtable_5pyart_2io_11_sigmetfile_SigmetFile;
  __pyx_vtable_5pyart_2io_11_sigmetfile_SigmetFile._get_ray = (int (*)(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *, int, PyArrayObject *))__pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__get_ray;
  __pyx_vtable_5pyart_2io_11_sigmetfile_SigmetFile._incr_rbuf_pos = (int (*)(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *, struct __pyx_opt_args_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos *__pyx_optional_args))__pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos;
  __pyx_vtable_5pyart_2io_11_sigmetfile_SigmetFile._load_record = (int (*)(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *))__pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__load_record;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5pyart_2io_11_sigmetfile_SigmetFile_spec, NULL); if (unlikely(!__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile)) __PYX_ERR(0, 17, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5pyart_2io_11_sigmetfile_SigmetFile_spec, __pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  #else
  __pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile = &__pyx_type_5pyart_2io_11_sigmetfile_SigmetFile;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile->tp_dictoffset && __pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 17, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_5pyart_2io_11_sigmetfile_10SigmetFile___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_5pyart_2io_11_sigmetfile_10SigmetFile___init__.doc = __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_5pyart_2io_11_sigmetfile_10SigmetFile___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile, __pyx_vtabptr_5pyart_2io_11_sigmetfile_SigmetFile) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SigmetFile, (PyObject *) __pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_0_12(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArray_Descr),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 272, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArrayIterObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 317, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArrayMultiIterObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 321, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArrayObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 360, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_generic) __PYX_ERR(2, 865, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "number", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_number) __PYX_ERR(2, 867, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_integer) __PYX_ERR(2, 869, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(2, 871, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(2, 873, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(2, 875, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_floating) __PYX_ERR(2, 877, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(2, 879, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(2, 881, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "character", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_character) __PYX_ERR(2, 883, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyUFuncObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 947, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__sigmetfile(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__sigmetfile},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "_sigmetfile",
      __pyx_k_A_class_and_supporting_function, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_sigmetfile(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_sigmetfile(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__sigmetfile(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__sigmetfile(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__sigmetfile(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  PyObject *__pyx_t_33 = NULL;
  PyObject *__pyx_t_34 = NULL;
  PyObject *__pyx_t_35 = NULL;
  PyObject *__pyx_t_36 = NULL;
  PyObject *__pyx_t_37 = NULL;
  PyObject *__pyx_t_38 = NULL;
  PyObject *__pyx_t_39 = NULL;
  PyObject *__pyx_t_40 = NULL;
  PyObject *__pyx_t_41 = NULL;
  PyObject *__pyx_t_42 = NULL;
  PyObject *__pyx_t_43 = NULL;
  PyObject *__pyx_t_44 = NULL;
  PyObject *__pyx_t_45 = NULL;
  PyObject *__pyx_t_46 = NULL;
  PyObject *__pyx_t_47 = NULL;
  PyObject *__pyx_t_48 = NULL;
  PyObject *__pyx_t_49 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_sigmetfile' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_sigmetfile", __pyx_methods, __pyx_k_A_class_and_supporting_function, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "_sigmetfile" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = __Pyx_PyImport_AddModuleRef((const char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__sigmetfile(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_pyart__io___sigmetfile) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pyart.io._sigmetfile")) {
      if (unlikely((PyDict_SetItemString(modules, "pyart.io._sigmetfile", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "pyart/io/_sigmetfile.pyx":5
 * 
 * """
 * import datetime             # <<<<<<<<<<<<<<
 * import struct
 * import warnings
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_datetime, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_datetime, __pyx_t_2) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":6
 * """
 * import datetime
 * import struct             # <<<<<<<<<<<<<<
 * import warnings
 * 
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_struct, __pyx_t_2) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":7
 * import datetime
 * import struct
 * import warnings             # <<<<<<<<<<<<<<
 * 
 * import numpy as np
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_warnings, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_2) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":9
 * import warnings
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * cimport cython
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_numpy, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_2) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":14
 * cimport numpy as np
 * 
 * RECORD_SIZE = 6144      # Raw product file blocked into 6144 byte records             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RECORD_SIZE, __pyx_int_6144) < 0) __PYX_ERR(0, 14, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":84
 *         self._raw_product_bhdrs = []
 * 
 *     def _determine_data_types(self):             # <<<<<<<<<<<<<<
 *         """ Determine the available data types in the file. """
 *         # determine the available fields
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_10SigmetFile_3_determine_data_types, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SigmetFile__determine_data_types, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__21)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile, __pyx_n_s_determine_data_types, __pyx_t_2) < 0) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile);

  /* "pyart/io/_sigmetfile.pyx":95
 *         return _data_types_from_mask(word0, word1, word2, word3)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """ Close the file. """
 *         self._fh.close()
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_10SigmetFile_5close, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SigmetFile_close, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__23)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile, __pyx_n_s_close, __pyx_t_2) < 0) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile);

  /* "pyart/io/_sigmetfile.pyx":99
 *         self._fh.close()
 * 
 *     def read_data(self, full_xhdr=False):             # <<<<<<<<<<<<<<
 *         """
 *         Read all data from the file.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_10SigmetFile_7read_data, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SigmetFile_read_data, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__26);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile, __pyx_n_s_read_data, __pyx_t_2) < 0) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile);

  /* "pyart/io/_sigmetfile.pyx":224
 *         return data, metadata
 * 
 *     def _get_sweep(self, full_xhdr=False, raw_data=False):             # <<<<<<<<<<<<<<
 *         """
 *         Get the data and metadata from the next sweep.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_10SigmetFile_9_get_sweep, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SigmetFile__get_sweep, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__29);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile, __pyx_n_s_get_sweep, __pyx_t_2) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_10SigmetFile_11__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SigmetFile___reduce_cython, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._rbuf_p cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_10SigmetFile_13__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SigmetFile___setstate_cython, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":407
 * 
 * 
 * def _data_types_from_mask(word0, word1, word2, word3):             # <<<<<<<<<<<<<<
 *     """
 *     Return a list of the data types from the words in the data_type mask.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_1_data_types_from_mask, 0, __pyx_n_s_data_types_from_mask, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__34)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_data_types_from_mask, __pyx_t_2) < 0) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":418
 * 
 * 
 * def _is_bit_set(number, bit):             # <<<<<<<<<<<<<<
 *     """ Return True if bit is set in number. """
 *     return number >> bit & 1 == 1
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_3_is_bit_set, 0, __pyx_n_s_is_bit_set, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__36)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_is_bit_set, __pyx_t_2) < 0) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":423
 * 
 * 
 * def _parse_ray_headers(ray_headers):             # <<<<<<<<<<<<<<
 *     """
 *     Parse the metadata from Sigmet ray headers.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_5_parse_ray_headers, 0, __pyx_n_s_parse_ray_headers, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__38)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_parse_ray_headers, __pyx_t_2) < 0) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":467
 * # Data type constants, table 13, section 4.8
 * SIGMET_DATA_TYPES = {
 *     0: 'XHDR',             # <<<<<<<<<<<<<<
 *     1: 'DBT',
 *     2: 'DBZ',
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(128); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_0, __pyx_n_u_XHDR) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_1, __pyx_n_u_DBT) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_2, __pyx_n_u_DBZ) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_3, __pyx_n_u_VEL) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_4, __pyx_n_u_WIDTH) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_5, __pyx_n_u_ZDR) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_6, __pyx_n_u_UNKNOWN_6) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_7, __pyx_n_u_DBZC) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_8, __pyx_n_u_DBT2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_9, __pyx_n_u_DBZ2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_10, __pyx_n_u_VEL2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_11, __pyx_n_u_WIDTH2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_12, __pyx_n_u_ZDR2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_13, __pyx_n_u_RAINRATE2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_14, __pyx_n_u_KDP) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_15, __pyx_n_u_KDP2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_16, __pyx_n_u_PHIDP) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_17, __pyx_n_u_VELC) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_18, __pyx_n_u_SQI) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_19, __pyx_n_u_RHOHV) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_20, __pyx_n_u_RHOHV2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_21, __pyx_n_u_DBZC2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_22, __pyx_n_u_VELC2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_23, __pyx_n_u_SQI2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_24, __pyx_n_u_PHIDP2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_25, __pyx_n_u_LDRH) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_26, __pyx_n_u_LDRH2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_27, __pyx_n_u_LDRV) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_28, __pyx_n_u_LDRV2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_29, __pyx_n_u_UNKNOWN_29) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_30, __pyx_n_u_UNKNOWN_30) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_31, __pyx_n_u_UNKNOWN_31) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_32, __pyx_n_u_HEIGHT) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_33, __pyx_n_u_VIL2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_34, __pyx_n_u_RAW) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_35, __pyx_n_u_SHEAR) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_36, __pyx_n_u_DIVERGE2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_37, __pyx_n_u_FLIQUID2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_38, __pyx_n_u_USER) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_39, __pyx_n_u_OTHER) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_40, __pyx_n_u_DEFORM2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_41, __pyx_n_u_VVEL2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_42, __pyx_n_u_HVEL2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_43, __pyx_n_u_HDIR2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_44, __pyx_n_u_AXDIL2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_45, __pyx_n_u_TIME2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_46, __pyx_n_u_RHOH) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_47, __pyx_n_u_RHOH2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_48, __pyx_n_u_RHOV) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_49, __pyx_n_u_RHOV2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_50, __pyx_n_u_PHIH) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_51, __pyx_n_u_PHIH2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_52, __pyx_n_u_PHIV) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_53, __pyx_n_u_PHIV2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_54, __pyx_n_u_USER2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_55, __pyx_n_u_HCLASS) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_56, __pyx_n_u_HCLASS2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_57, __pyx_n_u_ZDRC) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_58, __pyx_n_u_ZDRC2) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_59, __pyx_n_u_TEMPERATURE16) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_60, __pyx_n_u_VIR16) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_61, __pyx_n_u_DBTV8) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_62, __pyx_n_u_DBTV16) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_63, __pyx_n_u_DBZV8) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_64, __pyx_n_u_DBZV16) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_65, __pyx_n_u_SNR8) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_66, __pyx_n_u_SNR16) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_67, __pyx_n_u_ALBEDO8) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_68, __pyx_n_u_ALBEDO16) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_69, __pyx_n_u_VILD16) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_70, __pyx_n_u_TURB16) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_71, __pyx_n_u_DBTE8) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_72, __pyx_n_u_DBTE16) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_73, __pyx_n_u_DBZE8) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_74, __pyx_n_u_DBZE16) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_75, __pyx_n_u_PMI8) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_76, __pyx_n_u_PMI16) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_77, __pyx_n_u_LOG8) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_78, __pyx_n_u_LOG16) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_79, __pyx_n_u_CSP8) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_80, __pyx_n_u_CSP16) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_81, __pyx_n_u_UNKNOWN_81) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_82, __pyx_n_u_UNKNOWN_82) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_83, __pyx_n_u_UNKNOWN_83) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_84, __pyx_n_u_UNKNOWN_84) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_85, __pyx_n_u_UNKNOWN_85) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_86, __pyx_n_u_UNKNOWN_86) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_87, __pyx_n_u_UNKNOWN_87) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_88, __pyx_n_u_UNKNOWN_88) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_89, __pyx_n_u_UNKNOWN_89) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_90, __pyx_n_u_UNKNOWN_90) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_91, __pyx_n_u_UNKNOWN_91) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_92, __pyx_n_u_UNKNOWN_92) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_93, __pyx_n_u_UNKNOWN_93) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_94, __pyx_n_u_UNKNOWN_94) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_95, __pyx_n_u_UNKNOWN_95) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_96, __pyx_n_u_UNKNOWN_96) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_97, __pyx_n_u_UNKNOWN_97) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_98, __pyx_n_u_UNKNOWN_98) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_99, __pyx_n_u_UNKNOWN_99) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_100, __pyx_n_u_UNKNOWN_100) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_101, __pyx_n_u_UNKNOWN_101) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_102, __pyx_n_u_UNKNOWN_102) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_103, __pyx_n_u_UNKNOWN_103) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_104, __pyx_n_u_UNKNOWN_104) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_105, __pyx_n_u_UNKNOWN_105) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_106, __pyx_n_u_UNKNOWN_106) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_107, __pyx_n_u_UNKNOWN_107) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_108, __pyx_n_u_UNKNOWN_108) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_109, __pyx_n_u_UNKNOWN_109) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_110, __pyx_n_u_UNKNOWN_110) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_111, __pyx_n_u_UNKNOWN_110) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_112, __pyx_n_u_UNKNOWN_112) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_113, __pyx_n_u_UNKNOWN_113) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_114, __pyx_n_u_UNKNOWN_114) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_115, __pyx_n_u_UNKNOWN_115) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_116, __pyx_n_u_UNKNOWN_116) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_117, __pyx_n_u_UNKNOWN_117) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_118, __pyx_n_u_UNKNOWN_118) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_119, __pyx_n_u_UNKNOWN_119) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_120, __pyx_n_u_UNKNOWN_120) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_121, __pyx_n_u_UNKNOWN_121) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_122, __pyx_n_u_UNKNOWN_122) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_123, __pyx_n_u_UNKNOWN_123) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_124, __pyx_n_u_UNKNOWN_124) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_125, __pyx_n_u_UNKNOWN_125) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_126, __pyx_n_u_UNKNOWN_126) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_127, __pyx_n_u_UNKNOWN_127) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SIGMET_DATA_TYPES, __pyx_t_2) < 0) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":602
 * # file. Rewriting the convertions/masking in Cython does not seem to improved
 * # performance likely since most of the routines are already vectorized.
 * def convert_sigmet_data(data_type, data, nbins):             # <<<<<<<<<<<<<<
 *     """ Convert sigmet data. """
 *     out = np.empty_like(data, dtype='float32')
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_7convert_sigmet_data, 0, __pyx_n_s_convert_sigmet_data, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_convert_sigmet_data, __pyx_t_2) < 0) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":809
 * 
 * 
 * def bin2_to_angle(bin2):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin2 encoded value (or array). """
 *     return 360. * bin2 / 65536
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_9bin2_to_angle, 0, __pyx_n_s_bin2_to_angle, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bin2_to_angle, __pyx_t_2) < 0) __PYX_ERR(0, 809, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":814
 * 
 * 
 * def bin4_to_angle(bin4):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin4 encoded value (or array). """
 *     return 360. * bin4 / 4294967296
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_11bin4_to_angle, 0, __pyx_n_s_bin4_to_angle, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bin4_to_angle, __pyx_t_2) < 0) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":824
 * 
 * 
 * def _unpack_structure(string, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a structure """
 *     fmt = ''.join([i[1] for i in structure])
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_13_unpack_structure, 0, __pyx_n_s_unpack_structure, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__46)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unpack_structure, __pyx_t_2) < 0) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":831
 * 
 * 
 * def _unpack_key(dic, key, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a key. """
 *     dic[key] = _unpack_structure(dic[key], structure)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_15_unpack_key, 0, __pyx_n_s_unpack_key, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unpack_key, __pyx_t_2) < 0) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":836
 * 
 * 
 * def _unpack_ingest_data_headers(record, ndata_types):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack one or more ingest_data_header from a record.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_17_unpack_ingest_data_headers, 0, __pyx_n_s_unpack_ingest_data_headers, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unpack_ingest_data_headers, __pyx_t_2) < 0) __PYX_ERR(0, 836, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":850
 * 
 * 
 * def _unpack_ingest_data_header(record, number):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack a single ingest_data_header from record.  Return None on error.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_19_unpack_ingest_data_header, 0, __pyx_n_s_unpack_ingest_data_header, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unpack_ingest_data_header, __pyx_t_2) < 0) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":864
 * 
 * 
 * def _unpack_raw_prod_bhdr(record):             # <<<<<<<<<<<<<<
 *     """ Return a dict with the unpacked raw_prod_bhdr from a record. """
 *     return _unpack_structure(record[:12], RAW_PROD_BHDR)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_21_unpack_raw_prod_bhdr, 0, __pyx_n_s_unpack_raw_prod_bhdr, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unpack_raw_prod_bhdr, __pyx_t_2) < 0) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":869
 * 
 * 
 * def _unpack_product_hdr(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked product_hdr from the first record.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_23_unpack_product_hdr, 0, __pyx_n_s_unpack_product_hdr, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__56)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unpack_product_hdr, __pyx_t_2) < 0) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":898
 * 
 * 
 * def _unpack_ingest_header(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked ingest_header from the second record.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5pyart_2io_11_sigmetfile_25_unpack_ingest_header, 0, __pyx_n_s_unpack_ingest_header, NULL, __pyx_n_s_pyart_io__sigmetfile, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unpack_ingest_header, __pyx_t_2) < 0) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":952
 * 
 * # scalar defitions, section 4.1, table 7, and corresponding
 * SINT1 = 'b'             # <<<<<<<<<<<<<<
 * UINT1 = 'B'
 * SINT2 = 'h'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SINT1, __pyx_n_u_b) < 0) __PYX_ERR(0, 952, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":953
 * # scalar defitions, section 4.1, table 7, and corresponding
 * SINT1 = 'b'
 * UINT1 = 'B'             # <<<<<<<<<<<<<<
 * SINT2 = 'h'
 * UINT2 = 'H'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UINT1, __pyx_n_u_B) < 0) __PYX_ERR(0, 953, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":954
 * SINT1 = 'b'
 * UINT1 = 'B'
 * SINT2 = 'h'             # <<<<<<<<<<<<<<
 * UINT2 = 'H'
 * SINT4 = 'i'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SINT2, __pyx_n_u_h) < 0) __PYX_ERR(0, 954, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":955
 * UINT1 = 'B'
 * SINT2 = 'h'
 * UINT2 = 'H'             # <<<<<<<<<<<<<<
 * SINT4 = 'i'
 * UINT4 = 'I'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UINT2, __pyx_n_u_H) < 0) __PYX_ERR(0, 955, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":956
 * SINT2 = 'h'
 * UINT2 = 'H'
 * SINT4 = 'i'             # <<<<<<<<<<<<<<
 * UINT4 = 'I'
 * FLT4 = 'f'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SINT4, __pyx_n_u_i) < 0) __PYX_ERR(0, 956, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":957
 * UINT2 = 'H'
 * SINT4 = 'i'
 * UINT4 = 'I'             # <<<<<<<<<<<<<<
 * FLT4 = 'f'
 * FLT8 = 'd'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UINT4, __pyx_n_u_I) < 0) __PYX_ERR(0, 957, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":958
 * SINT4 = 'i'
 * UINT4 = 'I'
 * FLT4 = 'f'             # <<<<<<<<<<<<<<
 * FLT8 = 'd'
 * BIN1 = 'B'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FLT4, __pyx_n_u_f) < 0) __PYX_ERR(0, 958, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":959
 * UINT4 = 'I'
 * FLT4 = 'f'
 * FLT8 = 'd'             # <<<<<<<<<<<<<<
 * BIN1 = 'B'
 * BIN2 = 'H'      # these values need to be decoded with _bin2_to_angle
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FLT8, __pyx_n_u_d) < 0) __PYX_ERR(0, 959, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":960
 * FLT4 = 'f'
 * FLT8 = 'd'
 * BIN1 = 'B'             # <<<<<<<<<<<<<<
 * BIN2 = 'H'      # these values need to be decoded with _bin2_to_angle
 * BIN4 = 'I'      # these values need to be decoded with _bin4_to_angle
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BIN1, __pyx_n_u_B) < 0) __PYX_ERR(0, 960, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":961
 * FLT8 = 'd'
 * BIN1 = 'B'
 * BIN2 = 'H'      # these values need to be decoded with _bin2_to_angle             # <<<<<<<<<<<<<<
 * BIN4 = 'I'      # these values need to be decoded with _bin4_to_angle
 * MESSAGE = 'I'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BIN2, __pyx_n_u_H) < 0) __PYX_ERR(0, 961, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":962
 * BIN1 = 'B'
 * BIN2 = 'H'      # these values need to be decoded with _bin2_to_angle
 * BIN4 = 'I'      # these values need to be decoded with _bin4_to_angle             # <<<<<<<<<<<<<<
 * MESSAGE = 'I'
 * UINT16_T = 'H'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BIN4, __pyx_n_u_I) < 0) __PYX_ERR(0, 962, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":963
 * BIN2 = 'H'      # these values need to be decoded with _bin2_to_angle
 * BIN4 = 'I'      # these values need to be decoded with _bin4_to_angle
 * MESSAGE = 'I'             # <<<<<<<<<<<<<<
 * UINT16_T = 'H'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MESSAGE, __pyx_n_u_I) < 0) __PYX_ERR(0, 963, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":964
 * BIN4 = 'I'      # these values need to be decoded with _bin4_to_angle
 * MESSAGE = 'I'
 * UINT16_T = 'H'             # <<<<<<<<<<<<<<
 * 
 * # structures are taken from Vaisala PROGRAMMER'S MANUAL IRIS
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UINT16_T, __pyx_n_u_H) < 0) __PYX_ERR(0, 964, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":974
 * # 640 bytes: product_hdr (section 4.2.25, page 47)
 * PRODUCT_HDR = (
 *     ('structure_header', '12s'),        # 12 bytes             # <<<<<<<<<<<<<<
 *     ('product_configuration', '320s'),  # 320 bytes
 *     ('product_end', '308s'),            # 308 bytes
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PRODUCT_HDR, __pyx_tuple__62) < 0) __PYX_ERR(0, 973, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":981
 * # 12 bytes : structure_header (section 4.2.47)
 * STRUCTURE_HEADER = (
 *     ('structure_identifier', SINT2),             # <<<<<<<<<<<<<<
 *     ('format_version', SINT2),
 *     ('bytes_in_structure', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_u_structure_identifier);
  __Pyx_GIVEREF(__pyx_n_u_structure_identifier);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_u_structure_identifier)) __PYX_ERR(0, 981, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(0, 981, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":982
 * STRUCTURE_HEADER = (
 *     ('structure_identifier', SINT2),
 *     ('format_version', SINT2),             # <<<<<<<<<<<<<<
 *     ('bytes_in_structure', SINT4),
 *     ('reserved', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_u_format_version);
  __Pyx_GIVEREF(__pyx_n_u_format_version);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_u_format_version)) __PYX_ERR(0, 982, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 982, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":983
 *     ('structure_identifier', SINT2),
 *     ('format_version', SINT2),
 *     ('bytes_in_structure', SINT4),             # <<<<<<<<<<<<<<
 *     ('reserved', SINT2),
 *     ('flag', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 983, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 983, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_u_bytes_in_structure);
  __Pyx_GIVEREF(__pyx_n_u_bytes_in_structure);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_u_bytes_in_structure)) __PYX_ERR(0, 983, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(0, 983, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":984
 *     ('format_version', SINT2),
 *     ('bytes_in_structure', SINT4),
 *     ('reserved', SINT2),             # <<<<<<<<<<<<<<
 *     ('flag', SINT2),
 * )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_u_reserved);
  __Pyx_GIVEREF(__pyx_n_u_reserved);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_u_reserved)) __PYX_ERR(0, 984, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2)) __PYX_ERR(0, 984, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":985
 *     ('bytes_in_structure', SINT4),
 *     ('reserved', SINT2),
 *     ('flag', SINT2),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_u_flag);
  __Pyx_GIVEREF(__pyx_n_u_flag);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_u_flag)) __PYX_ERR(0, 985, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2)) __PYX_ERR(0, 985, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":981
 * # 12 bytes : structure_header (section 4.2.47)
 * STRUCTURE_HEADER = (
 *     ('structure_identifier', SINT2),             # <<<<<<<<<<<<<<
 *     ('format_version', SINT2),
 *     ('bytes_in_structure', SINT4),
 */
  __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3)) __PYX_ERR(0, 981, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4)) __PYX_ERR(0, 981, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5)) __PYX_ERR(0, 981, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_6)) __PYX_ERR(0, 981, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_7)) __PYX_ERR(0, 981, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_STRUCTURE_HEADER, __pyx_t_2) < 0) __PYX_ERR(0, 980, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":991
 * PRODUCT_CONFIGURATION = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header
 *     ('product_type_code', UINT2),             # <<<<<<<<<<<<<<
 *     ('scheduling_code', UINT2),
 *     ('seconds_between_runs', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_u_product_type_code);
  __Pyx_GIVEREF(__pyx_n_u_product_type_code);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_u_product_type_code)) __PYX_ERR(0, 991, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2)) __PYX_ERR(0, 991, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":992
 *     ('structure_header', '12s'),    # 12 bytes: structure_header
 *     ('product_type_code', UINT2),
 *     ('scheduling_code', UINT2),             # <<<<<<<<<<<<<<
 *     ('seconds_between_runs', SINT4),
 *     ('generation_time', '12s'),     # 12 bytes: ymds_time
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_u_scheduling_code);
  __Pyx_GIVEREF(__pyx_n_u_scheduling_code);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_u_scheduling_code)) __PYX_ERR(0, 992, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2)) __PYX_ERR(0, 992, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":993
 *     ('product_type_code', UINT2),
 *     ('scheduling_code', UINT2),
 *     ('seconds_between_runs', SINT4),             # <<<<<<<<<<<<<<
 *     ('generation_time', '12s'),     # 12 bytes: ymds_time
 *     ('sweep_ingest_time', '12s'),   # 12 bytes: ymds_time
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_u_seconds_between_runs);
  __Pyx_GIVEREF(__pyx_n_u_seconds_between_runs);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_u_seconds_between_runs)) __PYX_ERR(0, 993, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1000
 *     ('product_name', '12s'),
 *     ('task_name', '12s'),
 *     ('flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('x_scale', SINT4),
 *     ('y_scale', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_u_flag);
  __Pyx_GIVEREF(__pyx_n_u_flag);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_u_flag)) __PYX_ERR(0, 1000, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 1000, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1001
 *     ('task_name', '12s'),
 *     ('flag', UINT2),
 *     ('x_scale', SINT4),             # <<<<<<<<<<<<<<
 *     ('y_scale', SINT4),
 *     ('z_scale', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_u_x_scale);
  __Pyx_GIVEREF(__pyx_n_u_x_scale);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_u_x_scale)) __PYX_ERR(0, 1001, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(0, 1001, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1002
 *     ('flag', UINT2),
 *     ('x_scale', SINT4),
 *     ('y_scale', SINT4),             # <<<<<<<<<<<<<<
 *     ('z_scale', SINT4),
 *     ('x_size', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_n_u_y_scale);
  __Pyx_GIVEREF(__pyx_n_u_y_scale);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_n_u_y_scale)) __PYX_ERR(0, 1002, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2)) __PYX_ERR(0, 1002, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1003
 *     ('x_scale', SINT4),
 *     ('y_scale', SINT4),
 *     ('z_scale', SINT4),             # <<<<<<<<<<<<<<
 *     ('x_size', SINT4),
 *     ('y_size', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_n_u_z_scale);
  __Pyx_GIVEREF(__pyx_n_u_z_scale);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_u_z_scale)) __PYX_ERR(0, 1003, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2)) __PYX_ERR(0, 1003, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1004
 *     ('y_scale', SINT4),
 *     ('z_scale', SINT4),
 *     ('x_size', SINT4),             # <<<<<<<<<<<<<<
 *     ('y_size', SINT4),
 *     ('z_size', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_n_u_x_size);
  __Pyx_GIVEREF(__pyx_n_u_x_size);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_n_u_x_size)) __PYX_ERR(0, 1004, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_2)) __PYX_ERR(0, 1004, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1005
 *     ('z_scale', SINT4),
 *     ('x_size', SINT4),
 *     ('y_size', SINT4),             # <<<<<<<<<<<<<<
 *     ('z_size', SINT4),
 *     ('x_location', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_INCREF(__pyx_n_u_y_size);
  __Pyx_GIVEREF(__pyx_n_u_y_size);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_n_u_y_size)) __PYX_ERR(0, 1005, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_2)) __PYX_ERR(0, 1005, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1006
 *     ('x_size', SINT4),
 *     ('y_size', SINT4),
 *     ('z_size', SINT4),             # <<<<<<<<<<<<<<
 *     ('x_location', SINT4),
 *     ('y_location', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_INCREF(__pyx_n_u_z_size);
  __Pyx_GIVEREF(__pyx_n_u_z_size);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_n_u_z_size)) __PYX_ERR(0, 1006, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_2)) __PYX_ERR(0, 1006, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1007
 *     ('y_size', SINT4),
 *     ('z_size', SINT4),
 *     ('x_location', SINT4),             # <<<<<<<<<<<<<<
 *     ('y_location', SINT4),
 *     ('z_location', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_INCREF(__pyx_n_u_x_location);
  __Pyx_GIVEREF(__pyx_n_u_x_location);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_n_u_x_location)) __PYX_ERR(0, 1007, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_2)) __PYX_ERR(0, 1007, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1008
 *     ('z_size', SINT4),
 *     ('x_location', SINT4),
 *     ('y_location', SINT4),             # <<<<<<<<<<<<<<
 *     ('z_location', SINT4),
 *     ('maximum_range', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_INCREF(__pyx_n_u_y_location);
  __Pyx_GIVEREF(__pyx_n_u_y_location);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_n_u_y_location)) __PYX_ERR(0, 1008, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_2)) __PYX_ERR(0, 1008, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1009
 *     ('x_location', SINT4),
 *     ('y_location', SINT4),
 *     ('z_location', SINT4),             # <<<<<<<<<<<<<<
 *     ('maximum_range', SINT4),
 *     ('data_type', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_INCREF(__pyx_n_u_z_location);
  __Pyx_GIVEREF(__pyx_n_u_z_location);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_n_u_z_location)) __PYX_ERR(0, 1009, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2)) __PYX_ERR(0, 1009, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1010
 *     ('y_location', SINT4),
 *     ('z_location', SINT4),
 *     ('maximum_range', SINT4),             # <<<<<<<<<<<<<<
 *     ('data_type', UINT2),
 *     ('projection_name', '12s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_INCREF(__pyx_n_u_maximum_range);
  __Pyx_GIVEREF(__pyx_n_u_maximum_range);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_n_u_maximum_range)) __PYX_ERR(0, 1010, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_2)) __PYX_ERR(0, 1010, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1011
 *     ('z_location', SINT4),
 *     ('maximum_range', SINT4),
 *     ('data_type', UINT2),             # <<<<<<<<<<<<<<
 *     ('projection_name', '12s'),
 *     ('input_data_type', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_INCREF(__pyx_n_u_data_type);
  __Pyx_GIVEREF(__pyx_n_u_data_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_n_u_data_type)) __PYX_ERR(0, 1011, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_2)) __PYX_ERR(0, 1011, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1013
 *     ('data_type', UINT2),
 *     ('projection_name', '12s'),
 *     ('input_data_type', UINT2),             # <<<<<<<<<<<<<<
 *     ('projection_type', UINT1),
 *     ('spare_1', '1s'),              # 1 bytes
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_INCREF(__pyx_n_u_input_data_type);
  __Pyx_GIVEREF(__pyx_n_u_input_data_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_n_u_input_data_type)) __PYX_ERR(0, 1013, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_2)) __PYX_ERR(0, 1013, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1014
 *     ('projection_name', '12s'),
 *     ('input_data_type', UINT2),
 *     ('projection_type', UINT1),             # <<<<<<<<<<<<<<
 *     ('spare_1', '1s'),              # 1 bytes
 *     ('radial_smoother', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_19 = PyTuple_New(2); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_INCREF(__pyx_n_u_projection_type);
  __Pyx_GIVEREF(__pyx_n_u_projection_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_n_u_projection_type)) __PYX_ERR(0, 1014, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_2)) __PYX_ERR(0, 1014, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1016
 *     ('projection_type', UINT1),
 *     ('spare_1', '1s'),              # 1 bytes
 *     ('radial_smoother', SINT2),             # <<<<<<<<<<<<<<
 *     ('times_run', SINT2),
 *     ('zr_constant', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_INCREF(__pyx_n_u_radial_smoother);
  __Pyx_GIVEREF(__pyx_n_u_radial_smoother);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_n_u_radial_smoother)) __PYX_ERR(0, 1016, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_2)) __PYX_ERR(0, 1016, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1017
 *     ('spare_1', '1s'),              # 1 bytes
 *     ('radial_smoother', SINT2),
 *     ('times_run', SINT2),             # <<<<<<<<<<<<<<
 *     ('zr_constant', SINT4),
 *     ('zr_exponent', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_INCREF(__pyx_n_u_times_run);
  __Pyx_GIVEREF(__pyx_n_u_times_run);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_n_u_times_run)) __PYX_ERR(0, 1017, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_2)) __PYX_ERR(0, 1017, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1018
 *     ('radial_smoother', SINT2),
 *     ('times_run', SINT2),
 *     ('zr_constant', SINT4),             # <<<<<<<<<<<<<<
 *     ('zr_exponent', SINT4),
 *     ('x_smoother', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_INCREF(__pyx_n_u_zr_constant);
  __Pyx_GIVEREF(__pyx_n_u_zr_constant);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_n_u_zr_constant)) __PYX_ERR(0, 1018, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_2)) __PYX_ERR(0, 1018, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1019
 *     ('times_run', SINT2),
 *     ('zr_constant', SINT4),
 *     ('zr_exponent', SINT4),             # <<<<<<<<<<<<<<
 *     ('x_smoother', SINT2),
 *     ('y_smoother', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_INCREF(__pyx_n_u_zr_exponent);
  __Pyx_GIVEREF(__pyx_n_u_zr_exponent);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_n_u_zr_exponent)) __PYX_ERR(0, 1019, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_2)) __PYX_ERR(0, 1019, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1020
 *     ('zr_constant', SINT4),
 *     ('zr_exponent', SINT4),
 *     ('x_smoother', SINT2),             # <<<<<<<<<<<<<<
 *     ('y_smoother', SINT2),
 *     ('product_specific_bytes', '80s'),  # 80 bytes:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_INCREF(__pyx_n_u_x_smoother);
  __Pyx_GIVEREF(__pyx_n_u_x_smoother);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_u_x_smoother)) __PYX_ERR(0, 1020, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_2)) __PYX_ERR(0, 1020, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1021
 *     ('zr_exponent', SINT4),
 *     ('x_smoother', SINT2),
 *     ('y_smoother', SINT2),             # <<<<<<<<<<<<<<
 *     ('product_specific_bytes', '80s'),  # 80 bytes:
 *     ('minor_task_suffix', '16s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_INCREF(__pyx_n_u_y_smoother);
  __Pyx_GIVEREF(__pyx_n_u_y_smoother);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_u_y_smoother)) __PYX_ERR(0, 1021, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_2)) __PYX_ERR(0, 1021, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":990
 * # 320 bytes: product_configuration (section 4.2.23, page 43) 320 bytes
 * PRODUCT_CONFIGURATION = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('product_type_code', UINT2),
 *     ('scheduling_code', UINT2),
 */
  __pyx_t_2 = PyTuple_New(36); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_tuple__59)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_7)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_6)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_5)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_tuple__63)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_tuple__64)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_tuple__65)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_tuple__66)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_tuple__67)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 9, __pyx_tuple__68)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 10, __pyx_t_4)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 11, __pyx_t_3)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 12, __pyx_t_8)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_9);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 13, __pyx_t_9)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_10);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 14, __pyx_t_10)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_11);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 15, __pyx_t_11)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_12);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 16, __pyx_t_12)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_13);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 17, __pyx_t_13)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_14);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 18, __pyx_t_14)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_15);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 19, __pyx_t_15)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_16);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 20, __pyx_t_16)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_17);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 21, __pyx_t_17)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 22, __pyx_tuple__69)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_18);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 23, __pyx_t_18)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_19);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 24, __pyx_t_19)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 25, __pyx_tuple__70)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_20);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 26, __pyx_t_20)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_21);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 27, __pyx_t_21)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_22);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 28, __pyx_t_22)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_23);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 29, __pyx_t_23)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_24);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 30, __pyx_t_24)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_25);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 31, __pyx_t_25)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 32, __pyx_tuple__71)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 33, __pyx_tuple__72)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 34, __pyx_tuple__73)) __PYX_ERR(0, 990, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 35, __pyx_tuple__74)) __PYX_ERR(0, 990, __pyx_L1_error);
  __pyx_t_7 = 0;
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_t_15 = 0;
  __pyx_t_16 = 0;
  __pyx_t_17 = 0;
  __pyx_t_18 = 0;
  __pyx_t_19 = 0;
  __pyx_t_20 = 0;
  __pyx_t_21 = 0;
  __pyx_t_22 = 0;
  __pyx_t_23 = 0;
  __pyx_t_24 = 0;
  __pyx_t_25 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PRODUCT_CONFIGURATION, __pyx_t_2) < 0) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1030
 * # 12 bytes: ymds_time Structure (section 4.2.76, page 72)
 * YMDS_TIME = (
 *     ('seconds', SINT4),             # <<<<<<<<<<<<<<
 *     ('milliseconds', UINT2),    # milliseconds in lowest 10 bits,
 *     ('year', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_INCREF(__pyx_n_u_seconds);
  __Pyx_GIVEREF(__pyx_n_u_seconds);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_u_seconds)) __PYX_ERR(0, 1030, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_2)) __PYX_ERR(0, 1030, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1031
 * YMDS_TIME = (
 *     ('seconds', SINT4),
 *     ('milliseconds', UINT2),    # milliseconds in lowest 10 bits,             # <<<<<<<<<<<<<<
 *     ('year', SINT2),
 *     ('month', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_INCREF(__pyx_n_u_milliseconds);
  __Pyx_GIVEREF(__pyx_n_u_milliseconds);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_u_milliseconds)) __PYX_ERR(0, 1031, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_2)) __PYX_ERR(0, 1031, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1032
 *     ('seconds', SINT4),
 *     ('milliseconds', UINT2),    # milliseconds in lowest 10 bits,
 *     ('year', SINT2),             # <<<<<<<<<<<<<<
 *     ('month', SINT2),
 *     ('day', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_INCREF(__pyx_n_u_year);
  __Pyx_GIVEREF(__pyx_n_u_year);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_n_u_year)) __PYX_ERR(0, 1032, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_2)) __PYX_ERR(0, 1032, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1033
 *     ('milliseconds', UINT2),    # milliseconds in lowest 10 bits,
 *     ('year', SINT2),
 *     ('month', SINT2),             # <<<<<<<<<<<<<<
 *     ('day', SINT2),
 * )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_INCREF(__pyx_n_u_month);
  __Pyx_GIVEREF(__pyx_n_u_month);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_n_u_month)) __PYX_ERR(0, 1033, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_2)) __PYX_ERR(0, 1033, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1034
 *     ('year', SINT2),
 *     ('month', SINT2),
 *     ('day', SINT2),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_INCREF(__pyx_n_u_day);
  __Pyx_GIVEREF(__pyx_n_u_day);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_n_u_day)) __PYX_ERR(0, 1034, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_2)) __PYX_ERR(0, 1034, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1030
 * # 12 bytes: ymds_time Structure (section 4.2.76, page 72)
 * YMDS_TIME = (
 *     ('seconds', SINT4),             # <<<<<<<<<<<<<<
 *     ('milliseconds', UINT2),    # milliseconds in lowest 10 bits,
 *     ('year', SINT2),
 */
  __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_25);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_25)) __PYX_ERR(0, 1030, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_24);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_24)) __PYX_ERR(0, 1030, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_23);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_23)) __PYX_ERR(0, 1030, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_22);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_22)) __PYX_ERR(0, 1030, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_21);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_21)) __PYX_ERR(0, 1030, __pyx_L1_error);
  __pyx_t_25 = 0;
  __pyx_t_24 = 0;
  __pyx_t_23 = 0;
  __pyx_t_22 = 0;
  __pyx_t_21 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_YMDS_TIME, __pyx_t_2) < 0) __PYX_ERR(0, 1029, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1039
 * # 48 bytes: color_scale_def (section 4.2.5, page 34)
 * COLOR_SCALE_DEF = (
 *     ('iflags', UINT4),             # <<<<<<<<<<<<<<
 *     ('istart', SINT4),
 *     ('istep', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_INCREF(__pyx_n_u_iflags);
  __Pyx_GIVEREF(__pyx_n_u_iflags);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_n_u_iflags)) __PYX_ERR(0, 1039, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_2)) __PYX_ERR(0, 1039, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1040
 * COLOR_SCALE_DEF = (
 *     ('iflags', UINT4),
 *     ('istart', SINT4),             # <<<<<<<<<<<<<<
 *     ('istep', SINT4),
 *     ('icolcnt', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_INCREF(__pyx_n_u_istart);
  __Pyx_GIVEREF(__pyx_n_u_istart);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_n_u_istart)) __PYX_ERR(0, 1040, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_2)) __PYX_ERR(0, 1040, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1041
 *     ('iflags', UINT4),
 *     ('istart', SINT4),
 *     ('istep', SINT4),             # <<<<<<<<<<<<<<
 *     ('icolcnt', SINT2),
 *     ('iset_and_scale', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_INCREF(__pyx_n_u_istep);
  __Pyx_GIVEREF(__pyx_n_u_istep);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_n_u_istep)) __PYX_ERR(0, 1041, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_2)) __PYX_ERR(0, 1041, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1042
 *     ('istart', SINT4),
 *     ('istep', SINT4),
 *     ('icolcnt', SINT2),             # <<<<<<<<<<<<<<
 *     ('iset_and_scale', UINT2),
 *     ('ilevel_seams', '32s')     # 32 bytes: UINT2[16]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_INCREF(__pyx_n_u_icolcnt);
  __Pyx_GIVEREF(__pyx_n_u_icolcnt);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_u_icolcnt)) __PYX_ERR(0, 1042, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_2)) __PYX_ERR(0, 1042, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1043
 *     ('istep', SINT4),
 *     ('icolcnt', SINT2),
 *     ('iset_and_scale', UINT2),             # <<<<<<<<<<<<<<
 *     ('ilevel_seams', '32s')     # 32 bytes: UINT2[16]
 * )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_INCREF(__pyx_n_u_iset_and_scale);
  __Pyx_GIVEREF(__pyx_n_u_iset_and_scale);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_u_iset_and_scale)) __PYX_ERR(0, 1043, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_2)) __PYX_ERR(0, 1043, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1039
 * # 48 bytes: color_scale_def (section 4.2.5, page 34)
 * COLOR_SCALE_DEF = (
 *     ('iflags', UINT4),             # <<<<<<<<<<<<<<
 *     ('istart', SINT4),
 *     ('istep', SINT4),
 */
  __pyx_t_2 = PyTuple_New(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_21);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_21)) __PYX_ERR(0, 1039, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_22);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_22)) __PYX_ERR(0, 1039, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_23);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_23)) __PYX_ERR(0, 1039, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_24);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_24)) __PYX_ERR(0, 1039, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_25);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_25)) __PYX_ERR(0, 1039, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_tuple__75)) __PYX_ERR(0, 1039, __pyx_L1_error);
  __pyx_t_21 = 0;
  __pyx_t_22 = 0;
  __pyx_t_23 = 0;
  __pyx_t_24 = 0;
  __pyx_t_25 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_COLOR_SCALE_DEF, __pyx_t_2) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1054
 *     ('ingest_time', '12s'),         # 12 bytes: ymds_time
 *     ('spare_0', '28s'),             # 28 bytes
 *     ('GMT_minute_offset_local', SINT2),             # <<<<<<<<<<<<<<
 *     ('ingest_hardware_name_', '16s'),
 *     ('ingest_site_name_', '16s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_INCREF(__pyx_n_u_GMT_minute_offset_local);
  __Pyx_GIVEREF(__pyx_n_u_GMT_minute_offset_local);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_u_GMT_minute_offset_local)) __PYX_ERR(0, 1054, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_2)) __PYX_ERR(0, 1054, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1057
 *     ('ingest_hardware_name_', '16s'),
 *     ('ingest_site_name_', '16s'),
 *     ('GMT_minute_offset_standard', SINT2),             # <<<<<<<<<<<<<<
 *     ('latitude', BIN4),
 *     ('longitude', BIN4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_INCREF(__pyx_n_u_GMT_minute_offset_standard);
  __Pyx_GIVEREF(__pyx_n_u_GMT_minute_offset_standard);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_u_GMT_minute_offset_standard)) __PYX_ERR(0, 1057, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_2)) __PYX_ERR(0, 1057, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1058
 *     ('ingest_site_name_', '16s'),
 *     ('GMT_minute_offset_standard', SINT2),
 *     ('latitude', BIN4),             # <<<<<<<<<<<<<<
 *     ('longitude', BIN4),
 *     ('ground_height', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BIN4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_INCREF(__pyx_n_u_latitude);
  __Pyx_GIVEREF(__pyx_n_u_latitude);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_n_u_latitude)) __PYX_ERR(0, 1058, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_2)) __PYX_ERR(0, 1058, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1059
 *     ('GMT_minute_offset_standard', SINT2),
 *     ('latitude', BIN4),
 *     ('longitude', BIN4),             # <<<<<<<<<<<<<<
 *     ('ground_height', SINT2),
 *     ('radar_height', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BIN4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_INCREF(__pyx_n_u_longitude);
  __Pyx_GIVEREF(__pyx_n_u_longitude);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_n_u_longitude)) __PYX_ERR(0, 1059, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_2)) __PYX_ERR(0, 1059, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1060
 *     ('latitude', BIN4),
 *     ('longitude', BIN4),
 *     ('ground_height', SINT2),             # <<<<<<<<<<<<<<
 *     ('radar_height', SINT2),
 *     ('prf', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_INCREF(__pyx_n_u_ground_height);
  __Pyx_GIVEREF(__pyx_n_u_ground_height);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_n_u_ground_height)) __PYX_ERR(0, 1060, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_2)) __PYX_ERR(0, 1060, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1061
 *     ('longitude', BIN4),
 *     ('ground_height', SINT2),
 *     ('radar_height', SINT2),             # <<<<<<<<<<<<<<
 *     ('prf', SINT4),
 *     ('pulse_width', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_INCREF(__pyx_n_u_radar_height);
  __Pyx_GIVEREF(__pyx_n_u_radar_height);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_n_u_radar_height)) __PYX_ERR(0, 1061, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_2)) __PYX_ERR(0, 1061, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1062
 *     ('ground_height', SINT2),
 *     ('radar_height', SINT2),
 *     ('prf', SINT4),             # <<<<<<<<<<<<<<
 *     ('pulse_width', SINT4),
 *     ('signal_processor_type', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_19 = PyTuple_New(2); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_INCREF(__pyx_n_u_prf);
  __Pyx_GIVEREF(__pyx_n_u_prf);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_n_u_prf)) __PYX_ERR(0, 1062, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_2)) __PYX_ERR(0, 1062, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1063
 *     ('radar_height', SINT2),
 *     ('prf', SINT4),
 *     ('pulse_width', SINT4),             # <<<<<<<<<<<<<<
 *     ('signal_processor_type', UINT2),
 *     ('trigger_rate', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_INCREF(__pyx_n_u_pulse_width);
  __Pyx_GIVEREF(__pyx_n_u_pulse_width);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_n_u_pulse_width)) __PYX_ERR(0, 1063, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_2)) __PYX_ERR(0, 1063, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1064
 *     ('prf', SINT4),
 *     ('pulse_width', SINT4),
 *     ('signal_processor_type', UINT2),             # <<<<<<<<<<<<<<
 *     ('trigger_rate', UINT2),
 *     ('samples_used', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_INCREF(__pyx_n_u_signal_processor_type);
  __Pyx_GIVEREF(__pyx_n_u_signal_processor_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_n_u_signal_processor_type)) __PYX_ERR(0, 1064, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_2)) __PYX_ERR(0, 1064, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1065
 *     ('pulse_width', SINT4),
 *     ('signal_processor_type', UINT2),
 *     ('trigger_rate', UINT2),             # <<<<<<<<<<<<<<
 *     ('samples_used', SINT2),
 *     ('clutter_filter', '12s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_INCREF(__pyx_n_u_trigger_rate);
  __Pyx_GIVEREF(__pyx_n_u_trigger_rate);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_n_u_trigger_rate)) __PYX_ERR(0, 1065, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_2)) __PYX_ERR(0, 1065, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1066
 *     ('signal_processor_type', UINT2),
 *     ('trigger_rate', UINT2),
 *     ('samples_used', SINT2),             # <<<<<<<<<<<<<<
 *     ('clutter_filter', '12s'),
 *     ('number_linear_filter', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_INCREF(__pyx_n_u_samples_used);
  __Pyx_GIVEREF(__pyx_n_u_samples_used);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_n_u_samples_used)) __PYX_ERR(0, 1066, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2)) __PYX_ERR(0, 1066, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1068
 *     ('samples_used', SINT2),
 *     ('clutter_filter', '12s'),
 *     ('number_linear_filter', UINT2),             # <<<<<<<<<<<<<<
 *     ('wavelength', SINT4),
 *     ('truncation_height', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_INCREF(__pyx_n_u_number_linear_filter);
  __Pyx_GIVEREF(__pyx_n_u_number_linear_filter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_n_u_number_linear_filter)) __PYX_ERR(0, 1068, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_2)) __PYX_ERR(0, 1068, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1069
 *     ('clutter_filter', '12s'),
 *     ('number_linear_filter', UINT2),
 *     ('wavelength', SINT4),             # <<<<<<<<<<<<<<
 *     ('truncation_height', SINT4),
 *     ('first_bin_range', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_INCREF(__pyx_n_u_wavelength);
  __Pyx_GIVEREF(__pyx_n_u_wavelength);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_n_u_wavelength)) __PYX_ERR(0, 1069, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_2)) __PYX_ERR(0, 1069, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1070
 *     ('number_linear_filter', UINT2),
 *     ('wavelength', SINT4),
 *     ('truncation_height', SINT4),             # <<<<<<<<<<<<<<
 *     ('first_bin_range', SINT4),
 *     ('last_bin_range', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_INCREF(__pyx_n_u_truncation_height);
  __Pyx_GIVEREF(__pyx_n_u_truncation_height);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_n_u_truncation_height)) __PYX_ERR(0, 1070, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_2)) __PYX_ERR(0, 1070, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1071
 *     ('wavelength', SINT4),
 *     ('truncation_height', SINT4),
 *     ('first_bin_range', SINT4),             # <<<<<<<<<<<<<<
 *     ('last_bin_range', SINT4),
 *     ('number_bins', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_INCREF(__pyx_n_u_first_bin_range);
  __Pyx_GIVEREF(__pyx_n_u_first_bin_range);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_n_u_first_bin_range)) __PYX_ERR(0, 1071, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_2)) __PYX_ERR(0, 1071, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1072
 *     ('truncation_height', SINT4),
 *     ('first_bin_range', SINT4),
 *     ('last_bin_range', SINT4),             # <<<<<<<<<<<<<<
 *     ('number_bins', SINT4),
 *     ('flag', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_n_u_last_bin_range);
  __Pyx_GIVEREF(__pyx_n_u_last_bin_range);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_n_u_last_bin_range)) __PYX_ERR(0, 1072, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_2)) __PYX_ERR(0, 1072, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1073
 *     ('first_bin_range', SINT4),
 *     ('last_bin_range', SINT4),
 *     ('number_bins', SINT4),             # <<<<<<<<<<<<<<
 *     ('flag', UINT2),
 *     ('number_ingest', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_n_u_number_bins);
  __Pyx_GIVEREF(__pyx_n_u_number_bins);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_u_number_bins)) __PYX_ERR(0, 1073, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2)) __PYX_ERR(0, 1073, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1074
 *     ('last_bin_range', SINT4),
 *     ('number_bins', SINT4),
 *     ('flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('number_ingest', SINT2),
 *     ('polarization', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_n_u_flag);
  __Pyx_GIVEREF(__pyx_n_u_flag);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_n_u_flag)) __PYX_ERR(0, 1074, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2)) __PYX_ERR(0, 1074, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1075
 *     ('number_bins', SINT4),
 *     ('flag', UINT2),
 *     ('number_ingest', SINT2),             # <<<<<<<<<<<<<<
 *     ('polarization', UINT2),
 *     ('horizontal_calibration_i0', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_u_number_ingest);
  __Pyx_GIVEREF(__pyx_n_u_number_ingest);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_u_number_ingest)) __PYX_ERR(0, 1075, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(0, 1075, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1076
 *     ('flag', UINT2),
 *     ('number_ingest', SINT2),
 *     ('polarization', UINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_calibration_i0', SINT2),
 *     ('horizontal_calibration_noise', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_u_polarization);
  __Pyx_GIVEREF(__pyx_n_u_polarization);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_u_polarization)) __PYX_ERR(0, 1076, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 1076, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1077
 *     ('number_ingest', SINT2),
 *     ('polarization', UINT2),
 *     ('horizontal_calibration_i0', SINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_calibration_noise', SINT2),
 *     ('horizontal_radar_constant', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_u_horizontal_calibration_i0);
  __Pyx_GIVEREF(__pyx_n_u_horizontal_calibration_i0);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_u_horizontal_calibration_i0)) __PYX_ERR(0, 1077, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(0, 1077, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1078
 *     ('polarization', UINT2),
 *     ('horizontal_calibration_i0', SINT2),
 *     ('horizontal_calibration_noise', SINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_radar_constant', SINT2),
 *     ('reciever_bandwidth', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_u_horizontal_calibration_noise);
  __Pyx_GIVEREF(__pyx_n_u_horizontal_calibration_noise);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_u_horizontal_calibration_noise)) __PYX_ERR(0, 1078, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2)) __PYX_ERR(0, 1078, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1079
 *     ('horizontal_calibration_i0', SINT2),
 *     ('horizontal_calibration_noise', SINT2),
 *     ('horizontal_radar_constant', SINT2),             # <<<<<<<<<<<<<<
 *     ('reciever_bandwidth', UINT2),
 *     ('horizontal_current_noise', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1079, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1079, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_u_horizontal_radar_constant);
  __Pyx_GIVEREF(__pyx_n_u_horizontal_radar_constant);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_u_horizontal_radar_constant)) __PYX_ERR(0, 1079, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2)) __PYX_ERR(0, 1079, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1080
 *     ('horizontal_calibration_noise', SINT2),
 *     ('horizontal_radar_constant', SINT2),
 *     ('reciever_bandwidth', UINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_current_noise', SINT2),
 *     ('vertical_current_noise', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_INCREF(__pyx_n_u_reciever_bandwidth);
  __Pyx_GIVEREF(__pyx_n_u_reciever_bandwidth);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_u_reciever_bandwidth)) __PYX_ERR(0, 1080, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_2)) __PYX_ERR(0, 1080, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1081
 *     ('horizontal_radar_constant', SINT2),
 *     ('reciever_bandwidth', UINT2),
 *     ('horizontal_current_noise', SINT2),             # <<<<<<<<<<<<<<
 *     ('vertical_current_noise', SINT2),
 *     ('ldr_offset', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_INCREF(__pyx_n_u_horizontal_current_noise);
  __Pyx_GIVEREF(__pyx_n_u_horizontal_current_noise);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_n_u_horizontal_current_noise)) __PYX_ERR(0, 1081, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_2)) __PYX_ERR(0, 1081, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1082
 *     ('reciever_bandwidth', UINT2),
 *     ('horizontal_current_noise', SINT2),
 *     ('vertical_current_noise', SINT2),             # <<<<<<<<<<<<<<
 *     ('ldr_offset', SINT2),
 *     ('zdr_offset', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1082, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1082, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_n_u_vertical_current_noise);
  __Pyx_GIVEREF(__pyx_n_u_vertical_current_noise);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_u_vertical_current_noise)) __PYX_ERR(0, 1082, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_2)) __PYX_ERR(0, 1082, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1083
 *     ('horizontal_current_noise', SINT2),
 *     ('vertical_current_noise', SINT2),
 *     ('ldr_offset', SINT2),             # <<<<<<<<<<<<<<
 *     ('zdr_offset', SINT2),
 *     ('tcf_cal_flags_1', UINT16_T),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_INCREF(__pyx_n_u_ldr_offset);
  __Pyx_GIVEREF(__pyx_n_u_ldr_offset);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_n_u_ldr_offset)) __PYX_ERR(0, 1083, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_2)) __PYX_ERR(0, 1083, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1084
 *     ('vertical_current_noise', SINT2),
 *     ('ldr_offset', SINT2),
 *     ('zdr_offset', SINT2),             # <<<<<<<<<<<<<<
 *     ('tcf_cal_flags_1', UINT16_T),
 *     ('tcf_cal_flags_2', UINT16_T),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_30 = PyTuple_New(2); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_INCREF(__pyx_n_u_zdr_offset);
  __Pyx_GIVEREF(__pyx_n_u_zdr_offset);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_30, 0, __pyx_n_u_zdr_offset)) __PYX_ERR(0, 1084, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_30, 1, __pyx_t_2)) __PYX_ERR(0, 1084, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1085
 *     ('ldr_offset', SINT2),
 *     ('zdr_offset', SINT2),
 *     ('tcf_cal_flags_1', UINT16_T),             # <<<<<<<<<<<<<<
 *     ('tcf_cal_flags_2', UINT16_T),
 *     ('spare_1', '18s'),             # 18 bytes
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT16_T); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_31 = PyTuple_New(2); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_INCREF(__pyx_n_u_tcf_cal_flags_1);
  __Pyx_GIVEREF(__pyx_n_u_tcf_cal_flags_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_31, 0, __pyx_n_u_tcf_cal_flags_1)) __PYX_ERR(0, 1085, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_31, 1, __pyx_t_2)) __PYX_ERR(0, 1085, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1086
 *     ('zdr_offset', SINT2),
 *     ('tcf_cal_flags_1', UINT16_T),
 *     ('tcf_cal_flags_2', UINT16_T),             # <<<<<<<<<<<<<<
 *     ('spare_1', '18s'),             # 18 bytes
 *     ('standard_parallel_1', BIN4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT16_T); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_32 = PyTuple_New(2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_INCREF(__pyx_n_u_tcf_cal_flags_2);
  __Pyx_GIVEREF(__pyx_n_u_tcf_cal_flags_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_32, 0, __pyx_n_u_tcf_cal_flags_2)) __PYX_ERR(0, 1086, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_32, 1, __pyx_t_2)) __PYX_ERR(0, 1086, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1088
 *     ('tcf_cal_flags_2', UINT16_T),
 *     ('spare_1', '18s'),             # 18 bytes
 *     ('standard_parallel_1', BIN4),             # <<<<<<<<<<<<<<
 *     ('standard_parallel_2', BIN4),
 *     ('earth_radius', UINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BIN4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_33 = PyTuple_New(2); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 1088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_INCREF(__pyx_n_u_standard_parallel_1);
  __Pyx_GIVEREF(__pyx_n_u_standard_parallel_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_n_u_standard_parallel_1)) __PYX_ERR(0, 1088, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_2)) __PYX_ERR(0, 1088, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1089
 *     ('spare_1', '18s'),             # 18 bytes
 *     ('standard_parallel_1', BIN4),
 *     ('standard_parallel_2', BIN4),             # <<<<<<<<<<<<<<
 *     ('earth_radius', UINT4),
 *     ('inverse_flatting', UINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BIN4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_34 = PyTuple_New(2); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 1089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_INCREF(__pyx_n_u_standard_parallel_2);
  __Pyx_GIVEREF(__pyx_n_u_standard_parallel_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_34, 0, __pyx_n_u_standard_parallel_2)) __PYX_ERR(0, 1089, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_34, 1, __pyx_t_2)) __PYX_ERR(0, 1089, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1090
 *     ('standard_parallel_1', BIN4),
 *     ('standard_parallel_2', BIN4),
 *     ('earth_radius', UINT4),             # <<<<<<<<<<<<<<
 *     ('inverse_flatting', UINT4),
 *     ('fault_status', UINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_35 = PyTuple_New(2); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_INCREF(__pyx_n_u_earth_radius);
  __Pyx_GIVEREF(__pyx_n_u_earth_radius);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_n_u_earth_radius)) __PYX_ERR(0, 1090, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_35, 1, __pyx_t_2)) __PYX_ERR(0, 1090, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1091
 *     ('standard_parallel_2', BIN4),
 *     ('earth_radius', UINT4),
 *     ('inverse_flatting', UINT4),             # <<<<<<<<<<<<<<
 *     ('fault_status', UINT4),
 *     ('input_mask', UINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_36 = PyTuple_New(2); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 1091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_INCREF(__pyx_n_u_inverse_flatting);
  __Pyx_GIVEREF(__pyx_n_u_inverse_flatting);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_n_u_inverse_flatting)) __PYX_ERR(0, 1091, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_2)) __PYX_ERR(0, 1091, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1092
 *     ('earth_radius', UINT4),
 *     ('inverse_flatting', UINT4),
 *     ('fault_status', UINT4),             # <<<<<<<<<<<<<<
 *     ('input_mask', UINT4),
 *     ('number_log_filter', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_37 = PyTuple_New(2); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 1092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_INCREF(__pyx_n_u_fault_status);
  __Pyx_GIVEREF(__pyx_n_u_fault_status);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_n_u_fault_status)) __PYX_ERR(0, 1092, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_2)) __PYX_ERR(0, 1092, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1093
 *     ('inverse_flatting', UINT4),
 *     ('fault_status', UINT4),
 *     ('input_mask', UINT4),             # <<<<<<<<<<<<<<
 *     ('number_log_filter', UINT2),
 *     ('cluttermap', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_38 = PyTuple_New(2); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 1093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_INCREF(__pyx_n_u_input_mask);
  __Pyx_GIVEREF(__pyx_n_u_input_mask);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_n_u_input_mask)) __PYX_ERR(0, 1093, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_38, 1, __pyx_t_2)) __PYX_ERR(0, 1093, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1094
 *     ('fault_status', UINT4),
 *     ('input_mask', UINT4),
 *     ('number_log_filter', UINT2),             # <<<<<<<<<<<<<<
 *     ('cluttermap', UINT2),
 *     ('latitude_projection', BIN4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_u_number_log_filter);
  __Pyx_GIVEREF(__pyx_n_u_number_log_filter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_u_number_log_filter)) __PYX_ERR(0, 1094, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_2)) __PYX_ERR(0, 1094, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1095
 *     ('input_mask', UINT4),
 *     ('number_log_filter', UINT2),
 *     ('cluttermap', UINT2),             # <<<<<<<<<<<<<<
 *     ('latitude_projection', BIN4),
 *     ('longitude_projection', BIN4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 1095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_u_cluttermap);
  __Pyx_GIVEREF(__pyx_n_u_cluttermap);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_u_cluttermap)) __PYX_ERR(0, 1095, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_2)) __PYX_ERR(0, 1095, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1096
 *     ('number_log_filter', UINT2),
 *     ('cluttermap', UINT2),
 *     ('latitude_projection', BIN4),             # <<<<<<<<<<<<<<
 *     ('longitude_projection', BIN4),
 *     ('product_sequence_number', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BIN4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 1096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_u_latitude_projection);
  __Pyx_GIVEREF(__pyx_n_u_latitude_projection);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_u_latitude_projection)) __PYX_ERR(0, 1096, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_2)) __PYX_ERR(0, 1096, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1097
 *     ('cluttermap', UINT2),
 *     ('latitude_projection', BIN4),
 *     ('longitude_projection', BIN4),             # <<<<<<<<<<<<<<
 *     ('product_sequence_number', SINT2),
 *     ('spare_2', '32s'),             # 32 bytes
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BIN4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1097, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 1097, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_n_u_longitude_projection);
  __Pyx_GIVEREF(__pyx_n_u_longitude_projection);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_u_longitude_projection)) __PYX_ERR(0, 1097, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_2)) __PYX_ERR(0, 1097, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1098
 *     ('latitude_projection', BIN4),
 *     ('longitude_projection', BIN4),
 *     ('product_sequence_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('spare_2', '32s'),             # 32 bytes
 *     ('melting_level', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 1098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_u_product_sequence_number);
  __Pyx_GIVEREF(__pyx_n_u_product_sequence_number);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_u_product_sequence_number)) __PYX_ERR(0, 1098, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_2)) __PYX_ERR(0, 1098, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1100
 *     ('product_sequence_number', SINT2),
 *     ('spare_2', '32s'),             # 32 bytes
 *     ('melting_level', SINT2),             # <<<<<<<<<<<<<<
 *     ('radar_height_above_reference', SINT2),
 *     ('number_elements', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 1100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_u_melting_level);
  __Pyx_GIVEREF(__pyx_n_u_melting_level);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_u_melting_level)) __PYX_ERR(0, 1100, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_2)) __PYX_ERR(0, 1100, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1101
 *     ('spare_2', '32s'),             # 32 bytes
 *     ('melting_level', SINT2),
 *     ('radar_height_above_reference', SINT2),             # <<<<<<<<<<<<<<
 *     ('number_elements', SINT2),
 *     ('mean_wind_speed', UINT1),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_u_radar_height_above_reference);
  __Pyx_GIVEREF(__pyx_n_u_radar_height_above_reference);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_u_radar_height_above_reference)) __PYX_ERR(0, 1101, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_2)) __PYX_ERR(0, 1101, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1102
 *     ('melting_level', SINT2),
 *     ('radar_height_above_reference', SINT2),
 *     ('number_elements', SINT2),             # <<<<<<<<<<<<<<
 *     ('mean_wind_speed', UINT1),
 *     ('mean_wind_direction', BIN1),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 1102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __Pyx_INCREF(__pyx_n_u_number_elements);
  __Pyx_GIVEREF(__pyx_n_u_number_elements);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_u_number_elements)) __PYX_ERR(0, 1102, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_2)) __PYX_ERR(0, 1102, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1103
 *     ('radar_height_above_reference', SINT2),
 *     ('number_elements', SINT2),
 *     ('mean_wind_speed', UINT1),             # <<<<<<<<<<<<<<
 *     ('mean_wind_direction', BIN1),
 *     ('spare_3', '2s'),              # 2 bytes
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 1103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __Pyx_INCREF(__pyx_n_u_mean_wind_speed);
  __Pyx_GIVEREF(__pyx_n_u_mean_wind_speed);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_n_u_mean_wind_speed)) __PYX_ERR(0, 1103, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_t_2)) __PYX_ERR(0, 1103, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1104
 *     ('number_elements', SINT2),
 *     ('mean_wind_speed', UINT1),
 *     ('mean_wind_direction', BIN1),             # <<<<<<<<<<<<<<
 *     ('spare_3', '2s'),              # 2 bytes
 *     ('tz_name', '8s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BIN1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_48 = PyTuple_New(2); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __Pyx_INCREF(__pyx_n_u_mean_wind_direction);
  __Pyx_GIVEREF(__pyx_n_u_mean_wind_direction);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_48, 0, __pyx_n_u_mean_wind_direction)) __PYX_ERR(0, 1104, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_48, 1, __pyx_t_2)) __PYX_ERR(0, 1104, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1107
 *     ('spare_3', '2s'),              # 2 bytes
 *     ('tz_name', '8s'),
 *     ('extended_product_header_offset', UINT4),             # <<<<<<<<<<<<<<
 *     ('spare_4', '4s'),              # 4 bytes
 * )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_49 = PyTuple_New(2); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 1107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __Pyx_INCREF(__pyx_n_u_extended_product_header_offset);
  __Pyx_GIVEREF(__pyx_n_u_extended_product_header_offset);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_49, 0, __pyx_n_u_extended_product_header_offset)) __PYX_ERR(0, 1107, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_49, 1, __pyx_t_2)) __PYX_ERR(0, 1107, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1049
 * # 308 bytes : product_end (section 4.2.24)
 * PRODUCT_END = (
 *     ('site_name', '16s'),             # <<<<<<<<<<<<<<
 *     ('iris_version_created', '8s'),
 *     ('ingest_iris_version', '8s'),
 */
  __pyx_t_2 = PyTuple_New(60); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_tuple__76);
  __Pyx_GIVEREF(__pyx_tuple__76);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_tuple__76)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__77);
  __Pyx_GIVEREF(__pyx_tuple__77);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_tuple__77)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_tuple__78);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_tuple__78)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__79);
  __Pyx_GIVEREF(__pyx_tuple__79);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_tuple__79)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__80);
  __Pyx_GIVEREF(__pyx_tuple__80);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_tuple__80)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_25);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_25)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_tuple__81)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__82);
  __Pyx_GIVEREF(__pyx_tuple__82);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_tuple__82)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_24);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_t_24)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_23);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 9, __pyx_t_23)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_22);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 10, __pyx_t_22)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_21);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 11, __pyx_t_21)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_20);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 12, __pyx_t_20)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_19);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 13, __pyx_t_19)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_18);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 14, __pyx_t_18)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_17);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 15, __pyx_t_17)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_16);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 16, __pyx_t_16)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_15);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 17, __pyx_t_15)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 18, __pyx_tuple__83)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_14);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 19, __pyx_t_14)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_13);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 20, __pyx_t_13)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_12);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 21, __pyx_t_12)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_11);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 22, __pyx_t_11)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_10);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 23, __pyx_t_10)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_9);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 24, __pyx_t_9)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 25, __pyx_t_8)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 26, __pyx_t_3)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 27, __pyx_t_4)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 28, __pyx_t_5)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 29, __pyx_t_6)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 30, __pyx_t_7)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_26);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 31, __pyx_t_26)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_27);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 32, __pyx_t_27)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_28);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 33, __pyx_t_28)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_29);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 34, __pyx_t_29)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_30);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 35, __pyx_t_30)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_31);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 36, __pyx_t_31)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_32);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 37, __pyx_t_32)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__84);
  __Pyx_GIVEREF(__pyx_tuple__84);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 38, __pyx_tuple__84)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_33);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 39, __pyx_t_33)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_34);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 40, __pyx_t_34)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_35);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 41, __pyx_t_35)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_36);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 42, __pyx_t_36)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_37);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 43, __pyx_t_37)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_38);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 44, __pyx_t_38)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_39);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 45, __pyx_t_39)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_40);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 46, __pyx_t_40)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_41);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 47, __pyx_t_41)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_42);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 48, __pyx_t_42)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_43);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 49, __pyx_t_43)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__85);
  __Pyx_GIVEREF(__pyx_tuple__85);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 50, __pyx_tuple__85)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_44);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 51, __pyx_t_44)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_45);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 52, __pyx_t_45)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_46);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 53, __pyx_t_46)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_47);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 54, __pyx_t_47)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_48);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 55, __pyx_t_48)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 56, __pyx_tuple__86)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__87);
  __Pyx_GIVEREF(__pyx_tuple__87);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 57, __pyx_tuple__87)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_49);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 58, __pyx_t_49)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__88);
  __Pyx_GIVEREF(__pyx_tuple__88);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 59, __pyx_tuple__88)) __PYX_ERR(0, 1049, __pyx_L1_error);
  __pyx_t_25 = 0;
  __pyx_t_24 = 0;
  __pyx_t_23 = 0;
  __pyx_t_22 = 0;
  __pyx_t_21 = 0;
  __pyx_t_20 = 0;
  __pyx_t_19 = 0;
  __pyx_t_18 = 0;
  __pyx_t_17 = 0;
  __pyx_t_16 = 0;
  __pyx_t_15 = 0;
  __pyx_t_14 = 0;
  __pyx_t_13 = 0;
  __pyx_t_12 = 0;
  __pyx_t_11 = 0;
  __pyx_t_10 = 0;
  __pyx_t_9 = 0;
  __pyx_t_8 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_26 = 0;
  __pyx_t_27 = 0;
  __pyx_t_28 = 0;
  __pyx_t_29 = 0;
  __pyx_t_30 = 0;
  __pyx_t_31 = 0;
  __pyx_t_32 = 0;
  __pyx_t_33 = 0;
  __pyx_t_34 = 0;
  __pyx_t_35 = 0;
  __pyx_t_36 = 0;
  __pyx_t_37 = 0;
  __pyx_t_38 = 0;
  __pyx_t_39 = 0;
  __pyx_t_40 = 0;
  __pyx_t_41 = 0;
  __pyx_t_42 = 0;
  __pyx_t_43 = 0;
  __pyx_t_44 = 0;
  __pyx_t_45 = 0;
  __pyx_t_46 = 0;
  __pyx_t_47 = 0;
  __pyx_t_48 = 0;
  __pyx_t_49 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PRODUCT_END, __pyx_t_2) < 0) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1113
 * # 4884 bytes ingest_header Structure (section 4.2.16, page 40)
 * INGEST_HEADER = (
 *     ('structure_header', '12s'),        # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('ingest_configuration', '480s'),   # 480 bytes: ingest_configuration
 *     ('task_configuration', '2612s'),    # 2612 bytes: task_configuration
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_INGEST_HEADER, __pyx_tuple__94) < 0) __PYX_ERR(0, 1112, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":1124
 * INGEST_CONFIGURATION = (
 *     ('filename', '80s'),
 *     ('number_files', SINT2),             # <<<<<<<<<<<<<<
 *     ('number_sweeps_completed', SINT2),
 *     ('total_size', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_49 = PyTuple_New(2); if (unlikely(!__pyx_t_49)) __PYX_ERR(0, 1124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __Pyx_INCREF(__pyx_n_u_number_files);
  __Pyx_GIVEREF(__pyx_n_u_number_files);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_49, 0, __pyx_n_u_number_files)) __PYX_ERR(0, 1124, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_49, 1, __pyx_t_2)) __PYX_ERR(0, 1124, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1125
 *     ('filename', '80s'),
 *     ('number_files', SINT2),
 *     ('number_sweeps_completed', SINT2),             # <<<<<<<<<<<<<<
 *     ('total_size', SINT4),
 *     ('volume_scan_start_time', '12s'),  # 12 bytes: ymds_time
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_48 = PyTuple_New(2); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 1125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __Pyx_INCREF(__pyx_n_u_number_sweeps_completed);
  __Pyx_GIVEREF(__pyx_n_u_number_sweeps_completed);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_48, 0, __pyx_n_u_number_sweeps_completed)) __PYX_ERR(0, 1125, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_48, 1, __pyx_t_2)) __PYX_ERR(0, 1125, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1126
 *     ('number_files', SINT2),
 *     ('number_sweeps_completed', SINT2),
 *     ('total_size', SINT4),             # <<<<<<<<<<<<<<
 *     ('volume_scan_start_time', '12s'),  # 12 bytes: ymds_time
 *     ('spare_0', '12s'),                 # 12 bytes
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 1126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __Pyx_INCREF(__pyx_n_u_total_size);
  __Pyx_GIVEREF(__pyx_n_u_total_size);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_n_u_total_size)) __PYX_ERR(0, 1126, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_t_2)) __PYX_ERR(0, 1126, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1129
 *     ('volume_scan_start_time', '12s'),  # 12 bytes: ymds_time
 *     ('spare_0', '12s'),                 # 12 bytes
 *     ('ray_header_bytes', SINT2),             # <<<<<<<<<<<<<<
 *     ('extended_ray_header_bytes', SINT2),
 *     ('number_task_config_table', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 1129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __Pyx_INCREF(__pyx_n_u_ray_header_bytes);
  __Pyx_GIVEREF(__pyx_n_u_ray_header_bytes);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_u_ray_header_bytes)) __PYX_ERR(0, 1129, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_2)) __PYX_ERR(0, 1129, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1130
 *     ('spare_0', '12s'),                 # 12 bytes
 *     ('ray_header_bytes', SINT2),
 *     ('extended_ray_header_bytes', SINT2),             # <<<<<<<<<<<<<<
 *     ('number_task_config_table', SINT2),
 *     ('playback_version', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 1130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_u_extended_ray_header_bytes);
  __Pyx_GIVEREF(__pyx_n_u_extended_ray_header_bytes);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_u_extended_ray_header_bytes)) __PYX_ERR(0, 1130, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_2)) __PYX_ERR(0, 1130, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1131
 *     ('ray_header_bytes', SINT2),
 *     ('extended_ray_header_bytes', SINT2),
 *     ('number_task_config_table', SINT2),             # <<<<<<<<<<<<<<
 *     ('playback_version', SINT2),
 *     ('spare_1', '4s'),                  # 4 bytes
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_u_number_task_config_table);
  __Pyx_GIVEREF(__pyx_n_u_number_task_config_table);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_u_number_task_config_table)) __PYX_ERR(0, 1131, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_2)) __PYX_ERR(0, 1131, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1132
 *     ('extended_ray_header_bytes', SINT2),
 *     ('number_task_config_table', SINT2),
 *     ('playback_version', SINT2),             # <<<<<<<<<<<<<<
 *     ('spare_1', '4s'),                  # 4 bytes
 *     ('iris_version', '8s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_u_playback_version);
  __Pyx_GIVEREF(__pyx_n_u_playback_version);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_u_playback_version)) __PYX_ERR(0, 1132, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_2)) __PYX_ERR(0, 1132, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1136
 *     ('iris_version', '8s'),
 *     ('hardware_site', '16s'),
 *     ('gmt_offset_minutes_local', SINT2),             # <<<<<<<<<<<<<<
 *     ('site_name', '16s'),
 *     ('gmt_offset_minutes_standard', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 1136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_n_u_gmt_offset_minutes_local);
  __Pyx_GIVEREF(__pyx_n_u_gmt_offset_minutes_local);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_u_gmt_offset_minutes_local)) __PYX_ERR(0, 1136, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_2)) __PYX_ERR(0, 1136, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1138
 *     ('gmt_offset_minutes_local', SINT2),
 *     ('site_name', '16s'),
 *     ('gmt_offset_minutes_standard', SINT2),             # <<<<<<<<<<<<<<
 *     ('latitude_radar', BIN4),
 *     ('longitude_radar', BIN4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_u_gmt_offset_minutes_standard);
  __Pyx_GIVEREF(__pyx_n_u_gmt_offset_minutes_standard);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_u_gmt_offset_minutes_standard)) __PYX_ERR(0, 1138, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_2)) __PYX_ERR(0, 1138, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1139
 *     ('site_name', '16s'),
 *     ('gmt_offset_minutes_standard', SINT2),
 *     ('latitude_radar', BIN4),             # <<<<<<<<<<<<<<
 *     ('longitude_radar', BIN4),
 *     ('height_site', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BIN4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_u_latitude_radar);
  __Pyx_GIVEREF(__pyx_n_u_latitude_radar);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_u_latitude_radar)) __PYX_ERR(0, 1139, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_2)) __PYX_ERR(0, 1139, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1140
 *     ('gmt_offset_minutes_standard', SINT2),
 *     ('latitude_radar', BIN4),
 *     ('longitude_radar', BIN4),             # <<<<<<<<<<<<<<
 *     ('height_site', SINT2),
 *     ('height_radar', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BIN4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_u_longitude_radar);
  __Pyx_GIVEREF(__pyx_n_u_longitude_radar);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_u_longitude_radar)) __PYX_ERR(0, 1140, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_2)) __PYX_ERR(0, 1140, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1141
 *     ('latitude_radar', BIN4),
 *     ('longitude_radar', BIN4),
 *     ('height_site', SINT2),             # <<<<<<<<<<<<<<
 *     ('height_radar', SINT2),
 *     ('resolution_rays', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_38 = PyTuple_New(2); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 1141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_INCREF(__pyx_n_u_height_site);
  __Pyx_GIVEREF(__pyx_n_u_height_site);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_n_u_height_site)) __PYX_ERR(0, 1141, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_38, 1, __pyx_t_2)) __PYX_ERR(0, 1141, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1142
 *     ('longitude_radar', BIN4),
 *     ('height_site', SINT2),
 *     ('height_radar', SINT2),             # <<<<<<<<<<<<<<
 *     ('resolution_rays', UINT2),
 *     ('first_ray_index', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_37 = PyTuple_New(2); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 1142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_INCREF(__pyx_n_u_height_radar);
  __Pyx_GIVEREF(__pyx_n_u_height_radar);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_n_u_height_radar)) __PYX_ERR(0, 1142, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_2)) __PYX_ERR(0, 1142, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1143
 *     ('height_site', SINT2),
 *     ('height_radar', SINT2),
 *     ('resolution_rays', UINT2),             # <<<<<<<<<<<<<<
 *     ('first_ray_index', UINT2),
 *     ('number_rays_sweep', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_36 = PyTuple_New(2); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 1143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_INCREF(__pyx_n_u_resolution_rays);
  __Pyx_GIVEREF(__pyx_n_u_resolution_rays);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_n_u_resolution_rays)) __PYX_ERR(0, 1143, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_2)) __PYX_ERR(0, 1143, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1144
 *     ('height_radar', SINT2),
 *     ('resolution_rays', UINT2),
 *     ('first_ray_index', UINT2),             # <<<<<<<<<<<<<<
 *     ('number_rays_sweep', UINT2),
 *     ('gparam_bytes', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_35 = PyTuple_New(2); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_INCREF(__pyx_n_u_first_ray_index);
  __Pyx_GIVEREF(__pyx_n_u_first_ray_index);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_n_u_first_ray_index)) __PYX_ERR(0, 1144, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_35, 1, __pyx_t_2)) __PYX_ERR(0, 1144, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1145
 *     ('resolution_rays', UINT2),
 *     ('first_ray_index', UINT2),
 *     ('number_rays_sweep', UINT2),             # <<<<<<<<<<<<<<
 *     ('gparam_bytes', SINT2),
 *     ('altitude_radar', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_34 = PyTuple_New(2); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 1145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_INCREF(__pyx_n_u_number_rays_sweep);
  __Pyx_GIVEREF(__pyx_n_u_number_rays_sweep);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_34, 0, __pyx_n_u_number_rays_sweep)) __PYX_ERR(0, 1145, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_34, 1, __pyx_t_2)) __PYX_ERR(0, 1145, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1146
 *     ('first_ray_index', UINT2),
 *     ('number_rays_sweep', UINT2),
 *     ('gparam_bytes', SINT2),             # <<<<<<<<<<<<<<
 *     ('altitude_radar', SINT4),
 *     ('velocity_east', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_33 = PyTuple_New(2); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 1146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_INCREF(__pyx_n_u_gparam_bytes);
  __Pyx_GIVEREF(__pyx_n_u_gparam_bytes);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_n_u_gparam_bytes)) __PYX_ERR(0, 1146, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_2)) __PYX_ERR(0, 1146, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1147
 *     ('number_rays_sweep', UINT2),
 *     ('gparam_bytes', SINT2),
 *     ('altitude_radar', SINT4),             # <<<<<<<<<<<<<<
 *     ('velocity_east', SINT4),
 *     ('velocity_north', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_32 = PyTuple_New(2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_INCREF(__pyx_n_u_altitude_radar);
  __Pyx_GIVEREF(__pyx_n_u_altitude_radar);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_32, 0, __pyx_n_u_altitude_radar)) __PYX_ERR(0, 1147, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_32, 1, __pyx_t_2)) __PYX_ERR(0, 1147, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1148
 *     ('gparam_bytes', SINT2),
 *     ('altitude_radar', SINT4),
 *     ('velocity_east', SINT4),             # <<<<<<<<<<<<<<
 *     ('velocity_north', SINT4),
 *     ('velocity_up', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_31 = PyTuple_New(2); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_INCREF(__pyx_n_u_velocity_east);
  __Pyx_GIVEREF(__pyx_n_u_velocity_east);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_31, 0, __pyx_n_u_velocity_east)) __PYX_ERR(0, 1148, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_31, 1, __pyx_t_2)) __PYX_ERR(0, 1148, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1149
 *     ('altitude_radar', SINT4),
 *     ('velocity_east', SINT4),
 *     ('velocity_north', SINT4),             # <<<<<<<<<<<<<<
 *     ('velocity_up', SINT4),
 *     ('antenna_offset_starboard', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_30 = PyTuple_New(2); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 1149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_INCREF(__pyx_n_u_velocity_north);
  __Pyx_GIVEREF(__pyx_n_u_velocity_north);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_30, 0, __pyx_n_u_velocity_north)) __PYX_ERR(0, 1149, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_30, 1, __pyx_t_2)) __PYX_ERR(0, 1149, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1150
 *     ('velocity_east', SINT4),
 *     ('velocity_north', SINT4),
 *     ('velocity_up', SINT4),             # <<<<<<<<<<<<<<
 *     ('antenna_offset_starboard', SINT4),
 *     ('antenna_offset_bow', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_INCREF(__pyx_n_u_velocity_up);
  __Pyx_GIVEREF(__pyx_n_u_velocity_up);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_n_u_velocity_up)) __PYX_ERR(0, 1150, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_2)) __PYX_ERR(0, 1150, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1151
 *     ('velocity_north', SINT4),
 *     ('velocity_up', SINT4),
 *     ('antenna_offset_starboard', SINT4),             # <<<<<<<<<<<<<<
 *     ('antenna_offset_bow', SINT4),
 *     ('antenna_offset_up', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_n_u_antenna_offset_starboard);
  __Pyx_GIVEREF(__pyx_n_u_antenna_offset_starboard);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_u_antenna_offset_starboard)) __PYX_ERR(0, 1151, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_2)) __PYX_ERR(0, 1151, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1152
 *     ('velocity_up', SINT4),
 *     ('antenna_offset_starboard', SINT4),
 *     ('antenna_offset_bow', SINT4),             # <<<<<<<<<<<<<<
 *     ('antenna_offset_up', SINT4),
 *     ('fault_status', UINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_INCREF(__pyx_n_u_antenna_offset_bow);
  __Pyx_GIVEREF(__pyx_n_u_antenna_offset_bow);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_n_u_antenna_offset_bow)) __PYX_ERR(0, 1152, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_2)) __PYX_ERR(0, 1152, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1153
 *     ('antenna_offset_starboard', SINT4),
 *     ('antenna_offset_bow', SINT4),
 *     ('antenna_offset_up', SINT4),             # <<<<<<<<<<<<<<
 *     ('fault_status', UINT4),
 *     ('melting_layer', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_INCREF(__pyx_n_u_antenna_offset_up);
  __Pyx_GIVEREF(__pyx_n_u_antenna_offset_up);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_u_antenna_offset_up)) __PYX_ERR(0, 1153, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_2)) __PYX_ERR(0, 1153, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1154
 *     ('antenna_offset_bow', SINT4),
 *     ('antenna_offset_up', SINT4),
 *     ('fault_status', UINT4),             # <<<<<<<<<<<<<<
 *     ('melting_layer', SINT2),
 *     ('spare_2', '2s'),              # 2 bytes
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_u_fault_status);
  __Pyx_GIVEREF(__pyx_n_u_fault_status);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_u_fault_status)) __PYX_ERR(0, 1154, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2)) __PYX_ERR(0, 1154, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1155
 *     ('antenna_offset_up', SINT4),
 *     ('fault_status', UINT4),
 *     ('melting_layer', SINT2),             # <<<<<<<<<<<<<<
 *     ('spare_2', '2s'),              # 2 bytes
 *     ('local_timezone', '8s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_u_melting_layer);
  __Pyx_GIVEREF(__pyx_n_u_melting_layer);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_u_melting_layer)) __PYX_ERR(0, 1155, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2)) __PYX_ERR(0, 1155, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1158
 *     ('spare_2', '2s'),              # 2 bytes
 *     ('local_timezone', '8s'),
 *     ('flags', UINT4),             # <<<<<<<<<<<<<<
 *     ('configuration_name', '16s'),
 *     ('spare_3', '228s')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_u_flags);
  __Pyx_GIVEREF(__pyx_n_u_flags);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_u_flags)) __PYX_ERR(0, 1158, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(0, 1158, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1123
 * # 480 bytes ingest_configuration Structure (section 4.2.14, page 38)
 * INGEST_CONFIGURATION = (
 *     ('filename', '80s'),             # <<<<<<<<<<<<<<
 *     ('number_files', SINT2),
 *     ('number_sweeps_completed', SINT2),
 */
  __pyx_t_2 = PyTuple_New(38); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_tuple__95);
  __Pyx_GIVEREF(__pyx_tuple__95);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_tuple__95)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_49);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_49)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_48);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_48)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_47);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_47)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_tuple__96)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__97);
  __Pyx_GIVEREF(__pyx_tuple__97);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_tuple__97)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_46);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_t_46)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_45);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_t_45)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_44);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_t_44)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_43);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 9, __pyx_t_43)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 10, __pyx_tuple__98)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__99);
  __Pyx_GIVEREF(__pyx_tuple__99);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 11, __pyx_tuple__99)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__100);
  __Pyx_GIVEREF(__pyx_tuple__100);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 12, __pyx_tuple__100)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_42);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 13, __pyx_t_42)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__76);
  __Pyx_GIVEREF(__pyx_tuple__76);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 14, __pyx_tuple__76)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_41);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 15, __pyx_t_41)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_40);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 16, __pyx_t_40)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_39);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 17, __pyx_t_39)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_38);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 18, __pyx_t_38)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_37);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 19, __pyx_t_37)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_36);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 20, __pyx_t_36)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_35);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 21, __pyx_t_35)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_34);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 22, __pyx_t_34)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_33);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 23, __pyx_t_33)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_32);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 24, __pyx_t_32)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_31);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 25, __pyx_t_31)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_30);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 26, __pyx_t_30)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_29);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 27, __pyx_t_29)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_28);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 28, __pyx_t_28)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_27);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 29, __pyx_t_27)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_26);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 30, __pyx_t_26)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 31, __pyx_t_7)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 32, __pyx_t_6)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 33, __pyx_tuple__101)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__102);
  __Pyx_GIVEREF(__pyx_tuple__102);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 34, __pyx_tuple__102)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 35, __pyx_t_5)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__103);
  __Pyx_GIVEREF(__pyx_tuple__103);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 36, __pyx_tuple__103)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 37, __pyx_tuple__104)) __PYX_ERR(0, 1123, __pyx_L1_error);
  __pyx_t_49 = 0;
  __pyx_t_48 = 0;
  __pyx_t_47 = 0;
  __pyx_t_46 = 0;
  __pyx_t_45 = 0;
  __pyx_t_44 = 0;
  __pyx_t_43 = 0;
  __pyx_t_42 = 0;
  __pyx_t_41 = 0;
  __pyx_t_40 = 0;
  __pyx_t_39 = 0;
  __pyx_t_38 = 0;
  __pyx_t_37 = 0;
  __pyx_t_36 = 0;
  __pyx_t_35 = 0;
  __pyx_t_34 = 0;
  __pyx_t_33 = 0;
  __pyx_t_32 = 0;
  __pyx_t_31 = 0;
  __pyx_t_30 = 0;
  __pyx_t_29 = 0;
  __pyx_t_28 = 0;
  __pyx_t_27 = 0;
  __pyx_t_26 = 0;
  __pyx_t_7 = 0;
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_INGEST_CONFIGURATION, __pyx_t_2) < 0) __PYX_ERR(0, 1122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1165
 * # 2612 bytes: task_configuration Structure (section 4.2.50, page 61)
 * TASK_CONFIGURATION = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('task_sched_info', '120s'),    # 120 bytes: task_sched_info
 *     ('task_dsp_info', '320s'),      # 320 bytes: task_dsp_info
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_CONFIGURATION, __pyx_tuple__113) < 0) __PYX_ERR(0, 1164, __pyx_L1_error)

  /* "pyart/io/_sigmetfile.pyx":1178
 * # 120 bytes: task_sched_info Structure (section 4.2.61, page 65)
 * TASK_SCHED_INFO = (
 *     ('start_time', SINT4),             # <<<<<<<<<<<<<<
 *     ('stop_time', SINT4),
 *     ('skip_time', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_u_start_time);
  __Pyx_GIVEREF(__pyx_n_u_start_time);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_u_start_time)) __PYX_ERR(0, 1178, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(0, 1178, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1179
 * TASK_SCHED_INFO = (
 *     ('start_time', SINT4),
 *     ('stop_time', SINT4),             # <<<<<<<<<<<<<<
 *     ('skip_time', SINT4),
 *     ('last_run_time', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_u_stop_time);
  __Pyx_GIVEREF(__pyx_n_u_stop_time);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_u_stop_time)) __PYX_ERR(0, 1179, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2)) __PYX_ERR(0, 1179, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1180
 *     ('start_time', SINT4),
 *     ('stop_time', SINT4),
 *     ('skip_time', SINT4),             # <<<<<<<<<<<<<<
 *     ('last_run_time', SINT4),
 *     ('time_used_last_run', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_u_skip_time);
  __Pyx_GIVEREF(__pyx_n_u_skip_time);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_u_skip_time)) __PYX_ERR(0, 1180, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2)) __PYX_ERR(0, 1180, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1181
 *     ('stop_time', SINT4),
 *     ('skip_time', SINT4),
 *     ('last_run_time', SINT4),             # <<<<<<<<<<<<<<
 *     ('time_used_last_run', SINT4),
 *     ('last_run_day', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_INCREF(__pyx_n_u_last_run_time);
  __Pyx_GIVEREF(__pyx_n_u_last_run_time);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_u_last_run_time)) __PYX_ERR(0, 1181, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_2)) __PYX_ERR(0, 1181, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1182
 *     ('skip_time', SINT4),
 *     ('last_run_time', SINT4),
 *     ('time_used_last_run', SINT4),             # <<<<<<<<<<<<<<
 *     ('last_run_day', SINT4),
 *     ('flag', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_INCREF(__pyx_n_u_time_used_last_run);
  __Pyx_GIVEREF(__pyx_n_u_time_used_last_run);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_n_u_time_used_last_run)) __PYX_ERR(0, 1182, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_2)) __PYX_ERR(0, 1182, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1183
 *     ('last_run_time', SINT4),
 *     ('time_used_last_run', SINT4),
 *     ('last_run_day', SINT4),             # <<<<<<<<<<<<<<
 *     ('flag', UINT2),
 *     ('spare_0', '94s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_n_u_last_run_day);
  __Pyx_GIVEREF(__pyx_n_u_last_run_day);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_u_last_run_day)) __PYX_ERR(0, 1183, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_2)) __PYX_ERR(0, 1183, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1184
 *     ('time_used_last_run', SINT4),
 *     ('last_run_day', SINT4),
 *     ('flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '94s'),
 * )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_INCREF(__pyx_n_u_flag);
  __Pyx_GIVEREF(__pyx_n_u_flag);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_n_u_flag)) __PYX_ERR(0, 1184, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_2)) __PYX_ERR(0, 1184, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1178
 * # 120 bytes: task_sched_info Structure (section 4.2.61, page 65)
 * TASK_SCHED_INFO = (
 *     ('start_time', SINT4),             # <<<<<<<<<<<<<<
 *     ('stop_time', SINT4),
 *     ('skip_time', SINT4),
 */
  __pyx_t_2 = PyTuple_New(8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5)) __PYX_ERR(0, 1178, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_6)) __PYX_ERR(0, 1178, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_7)) __PYX_ERR(0, 1178, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_26);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_26)) __PYX_ERR(0, 1178, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_27);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_27)) __PYX_ERR(0, 1178, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_28);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_28)) __PYX_ERR(0, 1178, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_29);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_t_29)) __PYX_ERR(0, 1178, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__114);
  __Pyx_GIVEREF(__pyx_tuple__114);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_tuple__114)) __PYX_ERR(0, 1178, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_26 = 0;
  __pyx_t_27 = 0;
  __pyx_t_28 = 0;
  __pyx_t_29 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_SCHED_INFO, __pyx_t_2) < 0) __PYX_ERR(0, 1177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1190
 * # 320 bytes: task_dsp_info Structure (section 4.2.51, page 61)
 * TASK_DSP_INFO = (
 *     ('major_mode', UINT2),             # <<<<<<<<<<<<<<
 *     ('dsp_type', UINT2),
 *     ('current_data_type_mask', '24s'),      # 24 bytes: dsp_data_mask
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_INCREF(__pyx_n_u_major_mode);
  __Pyx_GIVEREF(__pyx_n_u_major_mode);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_n_u_major_mode)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_2)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1191
 * TASK_DSP_INFO = (
 *     ('major_mode', UINT2),
 *     ('dsp_type', UINT2),             # <<<<<<<<<<<<<<
 *     ('current_data_type_mask', '24s'),      # 24 bytes: dsp_data_mask
 *     ('original_data_type_mask', '24s'),     # 24 bytes: dsp_data_mask
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_n_u_dsp_type);
  __Pyx_GIVEREF(__pyx_n_u_dsp_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_u_dsp_type)) __PYX_ERR(0, 1191, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_2)) __PYX_ERR(0, 1191, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1196
 *     ('task_dsp_mode', '32s'),               # 32 bytes: task_dsp_mode_batch
 *     ('spare_0', '52s'),
 *     ('prf', SINT4),             # <<<<<<<<<<<<<<
 *     ('pulse_width', SINT4),
 *     ('multi_prf_flag', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_INCREF(__pyx_n_u_prf);
  __Pyx_GIVEREF(__pyx_n_u_prf);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_n_u_prf)) __PYX_ERR(0, 1196, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_2)) __PYX_ERR(0, 1196, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1197
 *     ('spare_0', '52s'),
 *     ('prf', SINT4),
 *     ('pulse_width', SINT4),             # <<<<<<<<<<<<<<
 *     ('multi_prf_flag', UINT2),
 *     ('dual_prf_delay', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_INCREF(__pyx_n_u_pulse_width);
  __Pyx_GIVEREF(__pyx_n_u_pulse_width);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_u_pulse_width)) __PYX_ERR(0, 1197, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_2)) __PYX_ERR(0, 1197, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1198
 *     ('prf', SINT4),
 *     ('pulse_width', SINT4),
 *     ('multi_prf_flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('dual_prf_delay', SINT2),
 *     ('agc_feedback_code', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_u_multi_prf_flag);
  __Pyx_GIVEREF(__pyx_n_u_multi_prf_flag);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_u_multi_prf_flag)) __PYX_ERR(0, 1198, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2)) __PYX_ERR(0, 1198, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1199
 *     ('pulse_width', SINT4),
 *     ('multi_prf_flag', UINT2),
 *     ('dual_prf_delay', SINT2),             # <<<<<<<<<<<<<<
 *     ('agc_feedback_code', UINT2),
 *     ('sample_size', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_u_dual_prf_delay);
  __Pyx_GIVEREF(__pyx_n_u_dual_prf_delay);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_u_dual_prf_delay)) __PYX_ERR(0, 1199, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2)) __PYX_ERR(0, 1199, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1200
 *     ('multi_prf_flag', UINT2),
 *     ('dual_prf_delay', SINT2),
 *     ('agc_feedback_code', UINT2),             # <<<<<<<<<<<<<<
 *     ('sample_size', SINT2),
 *     ('gain_control_flag', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_u_agc_feedback_code);
  __Pyx_GIVEREF(__pyx_n_u_agc_feedback_code);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_u_agc_feedback_code)) __PYX_ERR(0, 1200, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(0, 1200, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1201
 *     ('dual_prf_delay', SINT2),
 *     ('agc_feedback_code', UINT2),
 *     ('sample_size', SINT2),             # <<<<<<<<<<<<<<
 *     ('gain_control_flag', UINT2),
 *     ('clutter_filter_name', '12s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_30 = PyTuple_New(2); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_INCREF(__pyx_n_u_sample_size);
  __Pyx_GIVEREF(__pyx_n_u_sample_size);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_30, 0, __pyx_n_u_sample_size)) __PYX_ERR(0, 1201, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_30, 1, __pyx_t_2)) __PYX_ERR(0, 1201, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1202
 *     ('agc_feedback_code', UINT2),
 *     ('sample_size', SINT2),
 *     ('gain_control_flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('clutter_filter_name', '12s'),
 *     ('linear_filter_first_bin', UINT1),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_31 = PyTuple_New(2); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_INCREF(__pyx_n_u_gain_control_flag);
  __Pyx_GIVEREF(__pyx_n_u_gain_control_flag);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_31, 0, __pyx_n_u_gain_control_flag)) __PYX_ERR(0, 1202, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_31, 1, __pyx_t_2)) __PYX_ERR(0, 1202, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1204
 *     ('gain_control_flag', UINT2),
 *     ('clutter_filter_name', '12s'),
 *     ('linear_filter_first_bin', UINT1),             # <<<<<<<<<<<<<<
 *     ('log_filter_first_bin', UINT1),
 *     ('attenuation', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_32 = PyTuple_New(2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_INCREF(__pyx_n_u_linear_filter_first_bin);
  __Pyx_GIVEREF(__pyx_n_u_linear_filter_first_bin);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_32, 0, __pyx_n_u_linear_filter_first_bin)) __PYX_ERR(0, 1204, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_32, 1, __pyx_t_2)) __PYX_ERR(0, 1204, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1205
 *     ('clutter_filter_name', '12s'),
 *     ('linear_filter_first_bin', UINT1),
 *     ('log_filter_first_bin', UINT1),             # <<<<<<<<<<<<<<
 *     ('attenuation', SINT2),
 *     ('gas_attenuation', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_33 = PyTuple_New(2); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 1205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_INCREF(__pyx_n_u_log_filter_first_bin);
  __Pyx_GIVEREF(__pyx_n_u_log_filter_first_bin);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_n_u_log_filter_first_bin)) __PYX_ERR(0, 1205, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_2)) __PYX_ERR(0, 1205, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1206
 *     ('linear_filter_first_bin', UINT1),
 *     ('log_filter_first_bin', UINT1),
 *     ('attenuation', SINT2),             # <<<<<<<<<<<<<<
 *     ('gas_attenuation', UINT2),
 *     ('cluttermap_flag', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_34 = PyTuple_New(2); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_INCREF(__pyx_n_u_attenuation);
  __Pyx_GIVEREF(__pyx_n_u_attenuation);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_34, 0, __pyx_n_u_attenuation)) __PYX_ERR(0, 1206, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_34, 1, __pyx_t_2)) __PYX_ERR(0, 1206, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1207
 *     ('log_filter_first_bin', UINT1),
 *     ('attenuation', SINT2),
 *     ('gas_attenuation', UINT2),             # <<<<<<<<<<<<<<
 *     ('cluttermap_flag', UINT2),
 *     ('tranmitter_phase_sequence', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_35 = PyTuple_New(2); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 1207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_INCREF(__pyx_n_u_gas_attenuation);
  __Pyx_GIVEREF(__pyx_n_u_gas_attenuation);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_n_u_gas_attenuation)) __PYX_ERR(0, 1207, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_35, 1, __pyx_t_2)) __PYX_ERR(0, 1207, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1208
 *     ('attenuation', SINT2),
 *     ('gas_attenuation', UINT2),
 *     ('cluttermap_flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('tranmitter_phase_sequence', UINT2),
 *     ('ray_header_mask', UINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_36 = PyTuple_New(2); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 1208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_INCREF(__pyx_n_u_cluttermap_flag);
  __Pyx_GIVEREF(__pyx_n_u_cluttermap_flag);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_n_u_cluttermap_flag)) __PYX_ERR(0, 1208, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_2)) __PYX_ERR(0, 1208, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1209
 *     ('gas_attenuation', UINT2),
 *     ('cluttermap_flag', UINT2),
 *     ('tranmitter_phase_sequence', UINT2),             # <<<<<<<<<<<<<<
 *     ('ray_header_mask', UINT4),
 *     ('playback_flag', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_37 = PyTuple_New(2); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 1209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_INCREF(__pyx_n_u_tranmitter_phase_sequence);
  __Pyx_GIVEREF(__pyx_n_u_tranmitter_phase_sequence);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_n_u_tranmitter_phase_sequence)) __PYX_ERR(0, 1209, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_2)) __PYX_ERR(0, 1209, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1210
 *     ('cluttermap_flag', UINT2),
 *     ('tranmitter_phase_sequence', UINT2),
 *     ('ray_header_mask', UINT4),             # <<<<<<<<<<<<<<
 *     ('playback_flag', UINT2),
 *     ('spare_1', '2s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_38 = PyTuple_New(2); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 1210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_INCREF(__pyx_n_u_ray_header_mask);
  __Pyx_GIVEREF(__pyx_n_u_ray_header_mask);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_n_u_ray_header_mask)) __PYX_ERR(0, 1210, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_38, 1, __pyx_t_2)) __PYX_ERR(0, 1210, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1211
 *     ('tranmitter_phase_sequence', UINT2),
 *     ('ray_header_mask', UINT4),
 *     ('playback_flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_1', '2s'),
 *     ('custom_ray_header_name', '16s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 1211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_u_playback_flag);
  __Pyx_GIVEREF(__pyx_n_u_playback_flag);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_u_playback_flag)) __PYX_ERR(0, 1211, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_2)) __PYX_ERR(0, 1211, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1190
 * # 320 bytes: task_dsp_info Structure (section 4.2.51, page 61)
 * TASK_DSP_INFO = (
 *     ('major_mode', UINT2),             # <<<<<<<<<<<<<<
 *     ('dsp_type', UINT2),
 *     ('current_data_type_mask', '24s'),      # 24 bytes: dsp_data_mask
 */
  __pyx_t_2 = PyTuple_New(25); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_29);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_29)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_28);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_28)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__115);
  __Pyx_GIVEREF(__pyx_tuple__115);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_tuple__115)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__116);
  __Pyx_GIVEREF(__pyx_tuple__116);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_tuple__116)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__117);
  __Pyx_GIVEREF(__pyx_tuple__117);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_tuple__117)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__118);
  __Pyx_GIVEREF(__pyx_tuple__118);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_tuple__118)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_27);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_t_27)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_26);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_t_26)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_t_7)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 9, __pyx_t_6)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 10, __pyx_t_5)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_30);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 11, __pyx_t_30)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_31);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 12, __pyx_t_31)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__119);
  __Pyx_GIVEREF(__pyx_tuple__119);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 13, __pyx_tuple__119)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_32);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 14, __pyx_t_32)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_33);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 15, __pyx_t_33)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_34);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 16, __pyx_t_34)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_35);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 17, __pyx_t_35)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_36);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 18, __pyx_t_36)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_37);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 19, __pyx_t_37)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_38);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 20, __pyx_t_38)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_39);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 21, __pyx_t_39)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__120);
  __Pyx_GIVEREF(__pyx_tuple__120);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 22, __pyx_tuple__120)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__121);
  __Pyx_GIVEREF(__pyx_tuple__121);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 23, __pyx_tuple__121)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__122);
  __Pyx_GIVEREF(__pyx_tuple__122);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 24, __pyx_tuple__122)) __PYX_ERR(0, 1190, __pyx_L1_error);
  __pyx_t_29 = 0;
  __pyx_t_28 = 0;
  __pyx_t_27 = 0;
  __pyx_t_26 = 0;
  __pyx_t_7 = 0;
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  __pyx_t_30 = 0;
  __pyx_t_31 = 0;
  __pyx_t_32 = 0;
  __pyx_t_33 = 0;
  __pyx_t_34 = 0;
  __pyx_t_35 = 0;
  __pyx_t_36 = 0;
  __pyx_t_37 = 0;
  __pyx_t_38 = 0;
  __pyx_t_39 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_DSP_INFO, __pyx_t_2) < 0) __PYX_ERR(0, 1189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1219
 * # 24 bytes: dsp_data_mask Structure (section 4.2.7, page 36)
 * DSP_DATA_MASK = (
 *     ('mask_word_0', UINT4),             # <<<<<<<<<<<<<<
 *     ('extended_header_type', UINT4),
 *     ('mask_word_1', UINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 1219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_u_mask_word_0);
  __Pyx_GIVEREF(__pyx_n_u_mask_word_0);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_u_mask_word_0)) __PYX_ERR(0, 1219, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_2)) __PYX_ERR(0, 1219, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1220
 * DSP_DATA_MASK = (
 *     ('mask_word_0', UINT4),
 *     ('extended_header_type', UINT4),             # <<<<<<<<<<<<<<
 *     ('mask_word_1', UINT4),
 *     ('mask_word_2', UINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_38 = PyTuple_New(2); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 1220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_INCREF(__pyx_n_u_extended_header_type);
  __Pyx_GIVEREF(__pyx_n_u_extended_header_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_n_u_extended_header_type)) __PYX_ERR(0, 1220, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_38, 1, __pyx_t_2)) __PYX_ERR(0, 1220, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1221
 *     ('mask_word_0', UINT4),
 *     ('extended_header_type', UINT4),
 *     ('mask_word_1', UINT4),             # <<<<<<<<<<<<<<
 *     ('mask_word_2', UINT4),
 *     ('mask_word_3', UINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_37 = PyTuple_New(2); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 1221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_INCREF(__pyx_n_u_mask_word_1);
  __Pyx_GIVEREF(__pyx_n_u_mask_word_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_n_u_mask_word_1)) __PYX_ERR(0, 1221, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_2)) __PYX_ERR(0, 1221, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1222
 *     ('extended_header_type', UINT4),
 *     ('mask_word_1', UINT4),
 *     ('mask_word_2', UINT4),             # <<<<<<<<<<<<<<
 *     ('mask_word_3', UINT4),
 *     ('mask_word_4', UINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_36 = PyTuple_New(2); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_INCREF(__pyx_n_u_mask_word_2);
  __Pyx_GIVEREF(__pyx_n_u_mask_word_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_n_u_mask_word_2)) __PYX_ERR(0, 1222, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_2)) __PYX_ERR(0, 1222, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1223
 *     ('mask_word_1', UINT4),
 *     ('mask_word_2', UINT4),
 *     ('mask_word_3', UINT4),             # <<<<<<<<<<<<<<
 *     ('mask_word_4', UINT4),
 * )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_35 = PyTuple_New(2); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 1223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_INCREF(__pyx_n_u_mask_word_3);
  __Pyx_GIVEREF(__pyx_n_u_mask_word_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_n_u_mask_word_3)) __PYX_ERR(0, 1223, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_35, 1, __pyx_t_2)) __PYX_ERR(0, 1223, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1224
 *     ('mask_word_2', UINT4),
 *     ('mask_word_3', UINT4),
 *     ('mask_word_4', UINT4),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_34 = PyTuple_New(2); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_INCREF(__pyx_n_u_mask_word_4);
  __Pyx_GIVEREF(__pyx_n_u_mask_word_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_34, 0, __pyx_n_u_mask_word_4)) __PYX_ERR(0, 1224, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_34, 1, __pyx_t_2)) __PYX_ERR(0, 1224, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1219
 * # 24 bytes: dsp_data_mask Structure (section 4.2.7, page 36)
 * DSP_DATA_MASK = (
 *     ('mask_word_0', UINT4),             # <<<<<<<<<<<<<<
 *     ('extended_header_type', UINT4),
 *     ('mask_word_1', UINT4),
 */
  __pyx_t_2 = PyTuple_New(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_39);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_39)) __PYX_ERR(0, 1219, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_38);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_38)) __PYX_ERR(0, 1219, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_37);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_37)) __PYX_ERR(0, 1219, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_36);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_36)) __PYX_ERR(0, 1219, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_35);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_35)) __PYX_ERR(0, 1219, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_34);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_34)) __PYX_ERR(0, 1219, __pyx_L1_error);
  __pyx_t_39 = 0;
  __pyx_t_38 = 0;
  __pyx_t_37 = 0;
  __pyx_t_36 = 0;
  __pyx_t_35 = 0;
  __pyx_t_34 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DSP_DATA_MASK, __pyx_t_2) < 0) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1229
 * # 32 bytes: task_dsp_mode_batch (section 4.2.52, page 62)
 * TASK_DSP_MODE_BATCH = (
 *     ('low_prf_hz', UINT2),             # <<<<<<<<<<<<<<
 *     ('low_prf_factional', UINT2),
 *     ('low_prf_sample_size', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_34 = PyTuple_New(2); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_INCREF(__pyx_n_u_low_prf_hz);
  __Pyx_GIVEREF(__pyx_n_u_low_prf_hz);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_34, 0, __pyx_n_u_low_prf_hz)) __PYX_ERR(0, 1229, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_34, 1, __pyx_t_2)) __PYX_ERR(0, 1229, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1230
 * TASK_DSP_MODE_BATCH = (
 *     ('low_prf_hz', UINT2),
 *     ('low_prf_factional', UINT2),             # <<<<<<<<<<<<<<
 *     ('low_prf_sample_size', SINT2),
 *     ('low_prf_range_averaging', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_35 = PyTuple_New(2); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_INCREF(__pyx_n_u_low_prf_factional);
  __Pyx_GIVEREF(__pyx_n_u_low_prf_factional);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_n_u_low_prf_factional)) __PYX_ERR(0, 1230, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_35, 1, __pyx_t_2)) __PYX_ERR(0, 1230, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1231
 *     ('low_prf_hz', UINT2),
 *     ('low_prf_factional', UINT2),
 *     ('low_prf_sample_size', SINT2),             # <<<<<<<<<<<<<<
 *     ('low_prf_range_averaging', SINT2),
 *     ('reflectivity_unfolding_threshold', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_36 = PyTuple_New(2); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_INCREF(__pyx_n_u_low_prf_sample_size);
  __Pyx_GIVEREF(__pyx_n_u_low_prf_sample_size);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_n_u_low_prf_sample_size)) __PYX_ERR(0, 1231, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_2)) __PYX_ERR(0, 1231, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1232
 *     ('low_prf_factional', UINT2),
 *     ('low_prf_sample_size', SINT2),
 *     ('low_prf_range_averaging', SINT2),             # <<<<<<<<<<<<<<
 *     ('reflectivity_unfolding_threshold', SINT2),
 *     ('velocity_unfolding_threshold', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_37 = PyTuple_New(2); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 1232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_INCREF(__pyx_n_u_low_prf_range_averaging);
  __Pyx_GIVEREF(__pyx_n_u_low_prf_range_averaging);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_n_u_low_prf_range_averaging)) __PYX_ERR(0, 1232, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_2)) __PYX_ERR(0, 1232, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1233
 *     ('low_prf_sample_size', SINT2),
 *     ('low_prf_range_averaging', SINT2),
 *     ('reflectivity_unfolding_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('velocity_unfolding_threshold', SINT2),
 *     ('width_unfolding_threshold', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_38 = PyTuple_New(2); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_INCREF(__pyx_n_u_reflectivity_unfolding_threshold);
  __Pyx_GIVEREF(__pyx_n_u_reflectivity_unfolding_threshold);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_n_u_reflectivity_unfolding_threshold)) __PYX_ERR(0, 1233, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_38, 1, __pyx_t_2)) __PYX_ERR(0, 1233, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1234
 *     ('low_prf_range_averaging', SINT2),
 *     ('reflectivity_unfolding_threshold', SINT2),
 *     ('velocity_unfolding_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('width_unfolding_threshold', SINT2),
 *     ('spare_0', '18s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 1234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_u_velocity_unfolding_threshold);
  __Pyx_GIVEREF(__pyx_n_u_velocity_unfolding_threshold);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_u_velocity_unfolding_threshold)) __PYX_ERR(0, 1234, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_2)) __PYX_ERR(0, 1234, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1235
 *     ('reflectivity_unfolding_threshold', SINT2),
 *     ('velocity_unfolding_threshold', SINT2),
 *     ('width_unfolding_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '18s'),
 * )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_33 = PyTuple_New(2); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 1235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_INCREF(__pyx_n_u_width_unfolding_threshold);
  __Pyx_GIVEREF(__pyx_n_u_width_unfolding_threshold);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_n_u_width_unfolding_threshold)) __PYX_ERR(0, 1235, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_2)) __PYX_ERR(0, 1235, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1229
 * # 32 bytes: task_dsp_mode_batch (section 4.2.52, page 62)
 * TASK_DSP_MODE_BATCH = (
 *     ('low_prf_hz', UINT2),             # <<<<<<<<<<<<<<
 *     ('low_prf_factional', UINT2),
 *     ('low_prf_sample_size', SINT2),
 */
  __pyx_t_2 = PyTuple_New(8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_34);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_34)) __PYX_ERR(0, 1229, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_35);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_35)) __PYX_ERR(0, 1229, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_36);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_36)) __PYX_ERR(0, 1229, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_37);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_37)) __PYX_ERR(0, 1229, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_38);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_38)) __PYX_ERR(0, 1229, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_39);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_39)) __PYX_ERR(0, 1229, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_33);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_t_33)) __PYX_ERR(0, 1229, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__123);
  __Pyx_GIVEREF(__pyx_tuple__123);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_tuple__123)) __PYX_ERR(0, 1229, __pyx_L1_error);
  __pyx_t_34 = 0;
  __pyx_t_35 = 0;
  __pyx_t_36 = 0;
  __pyx_t_37 = 0;
  __pyx_t_38 = 0;
  __pyx_t_39 = 0;
  __pyx_t_33 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_DSP_MODE_BATCH, __pyx_t_2) < 0) __PYX_ERR(0, 1228, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1241
 * # 320 bytes: task_calib_info Structure (section 4.2.49, page 59)
 * TASK_CALIB_INFO = (
 *     ('reflectivity_slope', SINT2),             # <<<<<<<<<<<<<<
 *     ('reflectivity_noise_threshold', SINT2),
 *     ('clutter_correction_threshold', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_33 = PyTuple_New(2); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_INCREF(__pyx_n_u_reflectivity_slope);
  __Pyx_GIVEREF(__pyx_n_u_reflectivity_slope);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_n_u_reflectivity_slope)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_2)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1242
 * TASK_CALIB_INFO = (
 *     ('reflectivity_slope', SINT2),
 *     ('reflectivity_noise_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('clutter_correction_threshold', SINT2),
 *     ('sqi_threshold', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 1242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_u_reflectivity_noise_threshold);
  __Pyx_GIVEREF(__pyx_n_u_reflectivity_noise_threshold);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_u_reflectivity_noise_threshold)) __PYX_ERR(0, 1242, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_2)) __PYX_ERR(0, 1242, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1243
 *     ('reflectivity_slope', SINT2),
 *     ('reflectivity_noise_threshold', SINT2),
 *     ('clutter_correction_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('sqi_threshold', SINT2),
 *     ('power_threshold', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_38 = PyTuple_New(2); if (unlikely(!__pyx_t_38)) __PYX_ERR(0, 1243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_INCREF(__pyx_n_u_clutter_correction_threshold);
  __Pyx_GIVEREF(__pyx_n_u_clutter_correction_threshold);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_n_u_clutter_correction_threshold)) __PYX_ERR(0, 1243, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_38, 1, __pyx_t_2)) __PYX_ERR(0, 1243, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1244
 *     ('reflectivity_noise_threshold', SINT2),
 *     ('clutter_correction_threshold', SINT2),
 *     ('sqi_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('power_threshold', SINT2),
 *     ('spare_0', '8s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_37 = PyTuple_New(2); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 1244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_INCREF(__pyx_n_u_sqi_threshold);
  __Pyx_GIVEREF(__pyx_n_u_sqi_threshold);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_n_u_sqi_threshold)) __PYX_ERR(0, 1244, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_2)) __PYX_ERR(0, 1244, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1245
 *     ('clutter_correction_threshold', SINT2),
 *     ('sqi_threshold', SINT2),
 *     ('power_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '8s'),
 *     ('reflectivity_calibration', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_36 = PyTuple_New(2); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 1245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_INCREF(__pyx_n_u_power_threshold);
  __Pyx_GIVEREF(__pyx_n_u_power_threshold);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_n_u_power_threshold)) __PYX_ERR(0, 1245, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_2)) __PYX_ERR(0, 1245, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1247
 *     ('power_threshold', SINT2),
 *     ('spare_0', '8s'),
 *     ('reflectivity_calibration', SINT2),             # <<<<<<<<<<<<<<
 *     ('uncorrected_reflectivity_threshold_flags', UINT2),
 *     ('corrected_reflectivity_threshold_flags', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_35 = PyTuple_New(2); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 1247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_INCREF(__pyx_n_u_reflectivity_calibration);
  __Pyx_GIVEREF(__pyx_n_u_reflectivity_calibration);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_n_u_reflectivity_calibration)) __PYX_ERR(0, 1247, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_35, 1, __pyx_t_2)) __PYX_ERR(0, 1247, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1248
 *     ('spare_0', '8s'),
 *     ('reflectivity_calibration', SINT2),
 *     ('uncorrected_reflectivity_threshold_flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('corrected_reflectivity_threshold_flags', UINT2),
 *     ('velocity_threshold_flags', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_34 = PyTuple_New(2); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_INCREF(__pyx_n_u_uncorrected_reflectivity_thresho);
  __Pyx_GIVEREF(__pyx_n_u_uncorrected_reflectivity_thresho);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_34, 0, __pyx_n_u_uncorrected_reflectivity_thresho)) __PYX_ERR(0, 1248, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_34, 1, __pyx_t_2)) __PYX_ERR(0, 1248, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1249
 *     ('reflectivity_calibration', SINT2),
 *     ('uncorrected_reflectivity_threshold_flags', UINT2),
 *     ('corrected_reflectivity_threshold_flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('velocity_threshold_flags', UINT2),
 *     ('width_threshold_flags', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_32 = PyTuple_New(2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_INCREF(__pyx_n_u_corrected_reflectivity_threshold);
  __Pyx_GIVEREF(__pyx_n_u_corrected_reflectivity_threshold);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_32, 0, __pyx_n_u_corrected_reflectivity_threshold)) __PYX_ERR(0, 1249, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_32, 1, __pyx_t_2)) __PYX_ERR(0, 1249, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1250
 *     ('uncorrected_reflectivity_threshold_flags', UINT2),
 *     ('corrected_reflectivity_threshold_flags', UINT2),
 *     ('velocity_threshold_flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('width_threshold_flags', UINT2),
 *     ('zdr_threshold_flags', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_31 = PyTuple_New(2); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 1250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_INCREF(__pyx_n_u_velocity_threshold_flags);
  __Pyx_GIVEREF(__pyx_n_u_velocity_threshold_flags);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_31, 0, __pyx_n_u_velocity_threshold_flags)) __PYX_ERR(0, 1250, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_31, 1, __pyx_t_2)) __PYX_ERR(0, 1250, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1251
 *     ('corrected_reflectivity_threshold_flags', UINT2),
 *     ('velocity_threshold_flags', UINT2),
 *     ('width_threshold_flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('zdr_threshold_flags', UINT2),
 *     ('spare_1', '6s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_30 = PyTuple_New(2); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 1251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_INCREF(__pyx_n_u_width_threshold_flags);
  __Pyx_GIVEREF(__pyx_n_u_width_threshold_flags);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_30, 0, __pyx_n_u_width_threshold_flags)) __PYX_ERR(0, 1251, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_30, 1, __pyx_t_2)) __PYX_ERR(0, 1251, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1252
 *     ('velocity_threshold_flags', UINT2),
 *     ('width_threshold_flags', UINT2),
 *     ('zdr_threshold_flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_1', '6s'),
 *     ('flags', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_u_zdr_threshold_flags);
  __Pyx_GIVEREF(__pyx_n_u_zdr_threshold_flags);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_u_zdr_threshold_flags)) __PYX_ERR(0, 1252, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(0, 1252, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1254
 *     ('zdr_threshold_flags', UINT2),
 *     ('spare_1', '6s'),
 *     ('flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_2', '2s'),
 *     ('ldr_bias', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_u_flags);
  __Pyx_GIVEREF(__pyx_n_u_flags);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_u_flags)) __PYX_ERR(0, 1254, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2)) __PYX_ERR(0, 1254, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1256
 *     ('flags', UINT2),
 *     ('spare_2', '2s'),
 *     ('ldr_bias', SINT2),             # <<<<<<<<<<<<<<
 *     ('zdr_bias', SINT2),
 *     ('nexrad_clutter_threshold', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_u_ldr_bias);
  __Pyx_GIVEREF(__pyx_n_u_ldr_bias);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_u_ldr_bias)) __PYX_ERR(0, 1256, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2)) __PYX_ERR(0, 1256, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1257
 *     ('spare_2', '2s'),
 *     ('ldr_bias', SINT2),
 *     ('zdr_bias', SINT2),             # <<<<<<<<<<<<<<
 *     ('nexrad_clutter_threshold', SINT2),
 *     ('nexrad_clutter_skip', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 1257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_INCREF(__pyx_n_u_zdr_bias);
  __Pyx_GIVEREF(__pyx_n_u_zdr_bias);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_u_zdr_bias)) __PYX_ERR(0, 1257, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_2)) __PYX_ERR(0, 1257, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1258
 *     ('ldr_bias', SINT2),
 *     ('zdr_bias', SINT2),
 *     ('nexrad_clutter_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('nexrad_clutter_skip', UINT2),
 *     ('horizontal_i0_calibration', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_INCREF(__pyx_n_u_nexrad_clutter_threshold);
  __Pyx_GIVEREF(__pyx_n_u_nexrad_clutter_threshold);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_n_u_nexrad_clutter_threshold)) __PYX_ERR(0, 1258, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_2)) __PYX_ERR(0, 1258, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1259
 *     ('zdr_bias', SINT2),
 *     ('nexrad_clutter_threshold', SINT2),
 *     ('nexrad_clutter_skip', UINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_i0_calibration', SINT2),
 *     ('vertical_i0_calibration', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_n_u_nexrad_clutter_skip);
  __Pyx_GIVEREF(__pyx_n_u_nexrad_clutter_skip);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_u_nexrad_clutter_skip)) __PYX_ERR(0, 1259, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_2)) __PYX_ERR(0, 1259, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1260
 *     ('nexrad_clutter_threshold', SINT2),
 *     ('nexrad_clutter_skip', UINT2),
 *     ('horizontal_i0_calibration', SINT2),             # <<<<<<<<<<<<<<
 *     ('vertical_i0_calibration', SINT2),
 *     ('horizontal_noise_calibration', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_INCREF(__pyx_n_u_horizontal_i0_calibration);
  __Pyx_GIVEREF(__pyx_n_u_horizontal_i0_calibration);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_n_u_horizontal_i0_calibration)) __PYX_ERR(0, 1260, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_2)) __PYX_ERR(0, 1260, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1261
 *     ('nexrad_clutter_skip', UINT2),
 *     ('horizontal_i0_calibration', SINT2),
 *     ('vertical_i0_calibration', SINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_noise_calibration', SINT2),
 *     ('vertical_noise_calibration', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 1261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_u_vertical_i0_calibration);
  __Pyx_GIVEREF(__pyx_n_u_vertical_i0_calibration);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_u_vertical_i0_calibration)) __PYX_ERR(0, 1261, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_2)) __PYX_ERR(0, 1261, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1262
 *     ('horizontal_i0_calibration', SINT2),
 *     ('vertical_i0_calibration', SINT2),
 *     ('horizontal_noise_calibration', SINT2),             # <<<<<<<<<<<<<<
 *     ('vertical_noise_calibration', SINT2),
 *     ('horizontal_radar_constant', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 1262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_u_horizontal_noise_calibration);
  __Pyx_GIVEREF(__pyx_n_u_horizontal_noise_calibration);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_u_horizontal_noise_calibration)) __PYX_ERR(0, 1262, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_2)) __PYX_ERR(0, 1262, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1263
 *     ('vertical_i0_calibration', SINT2),
 *     ('horizontal_noise_calibration', SINT2),
 *     ('vertical_noise_calibration', SINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_radar_constant', SINT2),
 *     ('vertical_radar_constant', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_n_u_vertical_noise_calibration);
  __Pyx_GIVEREF(__pyx_n_u_vertical_noise_calibration);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_u_vertical_noise_calibration)) __PYX_ERR(0, 1263, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_2)) __PYX_ERR(0, 1263, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1264
 *     ('horizontal_noise_calibration', SINT2),
 *     ('vertical_noise_calibration', SINT2),
 *     ('horizontal_radar_constant', SINT2),             # <<<<<<<<<<<<<<
 *     ('vertical_radar_constant', SINT2),
 *     ('reciever_bandwidth', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 1264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_u_horizontal_radar_constant);
  __Pyx_GIVEREF(__pyx_n_u_horizontal_radar_constant);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_u_horizontal_radar_constant)) __PYX_ERR(0, 1264, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_2)) __PYX_ERR(0, 1264, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1265
 *     ('vertical_noise_calibration', SINT2),
 *     ('horizontal_radar_constant', SINT2),
 *     ('vertical_radar_constant', SINT2),             # <<<<<<<<<<<<<<
 *     ('reciever_bandwidth', UINT2),
 *     ('flags2', UINT16_T),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_u_vertical_radar_constant);
  __Pyx_GIVEREF(__pyx_n_u_vertical_radar_constant);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_u_vertical_radar_constant)) __PYX_ERR(0, 1265, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_2)) __PYX_ERR(0, 1265, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1266
 *     ('horizontal_radar_constant', SINT2),
 *     ('vertical_radar_constant', SINT2),
 *     ('reciever_bandwidth', UINT2),             # <<<<<<<<<<<<<<
 *     ('flags2', UINT16_T),
 *     ('spare_3', '256s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_u_reciever_bandwidth);
  __Pyx_GIVEREF(__pyx_n_u_reciever_bandwidth);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_u_reciever_bandwidth)) __PYX_ERR(0, 1266, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_2)) __PYX_ERR(0, 1266, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1267
 *     ('vertical_radar_constant', SINT2),
 *     ('reciever_bandwidth', UINT2),
 *     ('flags2', UINT16_T),             # <<<<<<<<<<<<<<
 *     ('spare_3', '256s'),
 * )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT16_T); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 1267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __Pyx_INCREF(__pyx_n_u_flags2);
  __Pyx_GIVEREF(__pyx_n_u_flags2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_u_flags2)) __PYX_ERR(0, 1267, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_2)) __PYX_ERR(0, 1267, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1241
 * # 320 bytes: task_calib_info Structure (section 4.2.49, page 59)
 * TASK_CALIB_INFO = (
 *     ('reflectivity_slope', SINT2),             # <<<<<<<<<<<<<<
 *     ('reflectivity_noise_threshold', SINT2),
 *     ('clutter_correction_threshold', SINT2),
 */
  __pyx_t_2 = PyTuple_New(28); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_33);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_33)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_39);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_39)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_38);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_38)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_37);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_37)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_36);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_36)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__124);
  __Pyx_GIVEREF(__pyx_tuple__124);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_tuple__124)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_35);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_t_35)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_34);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_t_34)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_32);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_t_32)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_31);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 9, __pyx_t_31)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_30);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 10, __pyx_t_30)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 11, __pyx_t_5)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__125);
  __Pyx_GIVEREF(__pyx_tuple__125);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 12, __pyx_tuple__125)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 13, __pyx_t_6)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 14, __pyx_tuple__101)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 15, __pyx_t_7)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_26);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 16, __pyx_t_26)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_27);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 17, __pyx_t_27)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_28);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 18, __pyx_t_28)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_29);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 19, __pyx_t_29)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_40);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 20, __pyx_t_40)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_41);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 21, __pyx_t_41)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_42);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 22, __pyx_t_42)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_43);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 23, __pyx_t_43)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_44);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 24, __pyx_t_44)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_45);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 25, __pyx_t_45)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_46);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 26, __pyx_t_46)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 27, __pyx_tuple__126)) __PYX_ERR(0, 1241, __pyx_L1_error);
  __pyx_t_33 = 0;
  __pyx_t_39 = 0;
  __pyx_t_38 = 0;
  __pyx_t_37 = 0;
  __pyx_t_36 = 0;
  __pyx_t_35 = 0;
  __pyx_t_34 = 0;
  __pyx_t_32 = 0;
  __pyx_t_31 = 0;
  __pyx_t_30 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_26 = 0;
  __pyx_t_27 = 0;
  __pyx_t_28 = 0;
  __pyx_t_29 = 0;
  __pyx_t_40 = 0;
  __pyx_t_41 = 0;
  __pyx_t_42 = 0;
  __pyx_t_43 = 0;
  __pyx_t_44 = 0;
  __pyx_t_45 = 0;
  __pyx_t_46 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_CALIB_INFO, __pyx_t_2) < 0) __PYX_ERR(0, 1240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1273
 * # 160 bytes: task_range_info Structure (section 4.2.58, page 64)
 * TASK_RANGE_INFO = (
 *     ('first_bin_range', SINT4),             # <<<<<<<<<<<<<<
 *     ('last_bin_range', SINT4),
 *     ('number_input_bins', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __Pyx_INCREF(__pyx_n_u_first_bin_range);
  __Pyx_GIVEREF(__pyx_n_u_first_bin_range);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_u_first_bin_range)) __PYX_ERR(0, 1273, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_2)) __PYX_ERR(0, 1273, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1274
 * TASK_RANGE_INFO = (
 *     ('first_bin_range', SINT4),
 *     ('last_bin_range', SINT4),             # <<<<<<<<<<<<<<
 *     ('number_input_bins', SINT2),
 *     ('number_output_bins', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_u_last_bin_range);
  __Pyx_GIVEREF(__pyx_n_u_last_bin_range);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_u_last_bin_range)) __PYX_ERR(0, 1274, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_2)) __PYX_ERR(0, 1274, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1275
 *     ('first_bin_range', SINT4),
 *     ('last_bin_range', SINT4),
 *     ('number_input_bins', SINT2),             # <<<<<<<<<<<<<<
 *     ('number_output_bins', SINT2),
 *     ('step_input_bins', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_u_number_input_bins);
  __Pyx_GIVEREF(__pyx_n_u_number_input_bins);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_u_number_input_bins)) __PYX_ERR(0, 1275, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_2)) __PYX_ERR(0, 1275, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1276
 *     ('last_bin_range', SINT4),
 *     ('number_input_bins', SINT2),
 *     ('number_output_bins', SINT2),             # <<<<<<<<<<<<<<
 *     ('step_input_bins', SINT4),
 *     ('step_output_bins', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_u_number_output_bins);
  __Pyx_GIVEREF(__pyx_n_u_number_output_bins);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_u_number_output_bins)) __PYX_ERR(0, 1276, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_2)) __PYX_ERR(0, 1276, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1277
 *     ('number_input_bins', SINT2),
 *     ('number_output_bins', SINT2),
 *     ('step_input_bins', SINT4),             # <<<<<<<<<<<<<<
 *     ('step_output_bins', SINT4),
 *     ('variable_range_bin_flag', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_n_u_step_input_bins);
  __Pyx_GIVEREF(__pyx_n_u_step_input_bins);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_u_step_input_bins)) __PYX_ERR(0, 1277, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_2)) __PYX_ERR(0, 1277, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1278
 *     ('number_output_bins', SINT2),
 *     ('step_input_bins', SINT4),
 *     ('step_output_bins', SINT4),             # <<<<<<<<<<<<<<
 *     ('variable_range_bin_flag', UINT2),
 *     ('range_bin_averaging_flag', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_u_step_output_bins);
  __Pyx_GIVEREF(__pyx_n_u_step_output_bins);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_u_step_output_bins)) __PYX_ERR(0, 1278, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_2)) __PYX_ERR(0, 1278, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1279
 *     ('step_input_bins', SINT4),
 *     ('step_output_bins', SINT4),
 *     ('variable_range_bin_flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('range_bin_averaging_flag', SINT2),
 *     ('spare_0', '136s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_u_variable_range_bin_flag);
  __Pyx_GIVEREF(__pyx_n_u_variable_range_bin_flag);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_u_variable_range_bin_flag)) __PYX_ERR(0, 1279, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_2)) __PYX_ERR(0, 1279, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1280
 *     ('step_output_bins', SINT4),
 *     ('variable_range_bin_flag', UINT2),
 *     ('range_bin_averaging_flag', SINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '136s'),
 * )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_INCREF(__pyx_n_u_range_bin_averaging_flag);
  __Pyx_GIVEREF(__pyx_n_u_range_bin_averaging_flag);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_n_u_range_bin_averaging_flag)) __PYX_ERR(0, 1280, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_2)) __PYX_ERR(0, 1280, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1273
 * # 160 bytes: task_range_info Structure (section 4.2.58, page 64)
 * TASK_RANGE_INFO = (
 *     ('first_bin_range', SINT4),             # <<<<<<<<<<<<<<
 *     ('last_bin_range', SINT4),
 *     ('number_input_bins', SINT2),
 */
  __pyx_t_2 = PyTuple_New(9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_46);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_46)) __PYX_ERR(0, 1273, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_45);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_45)) __PYX_ERR(0, 1273, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_44);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_44)) __PYX_ERR(0, 1273, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_43);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_43)) __PYX_ERR(0, 1273, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_42);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_42)) __PYX_ERR(0, 1273, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_41);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_41)) __PYX_ERR(0, 1273, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_40);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_t_40)) __PYX_ERR(0, 1273, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_29);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_t_29)) __PYX_ERR(0, 1273, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__127);
  __Pyx_GIVEREF(__pyx_tuple__127);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_tuple__127)) __PYX_ERR(0, 1273, __pyx_L1_error);
  __pyx_t_46 = 0;
  __pyx_t_45 = 0;
  __pyx_t_44 = 0;
  __pyx_t_43 = 0;
  __pyx_t_42 = 0;
  __pyx_t_41 = 0;
  __pyx_t_40 = 0;
  __pyx_t_29 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_RANGE_INFO, __pyx_t_2) < 0) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1286
 * # 320 bytes: task_scan_info Structure (section 4.2.60, page 65)
 * TASK_SCAN_INFO = (
 *     ('antenna_scan_mode', UINT2),             # <<<<<<<<<<<<<<
 *     ('angular_resolution_desired', SINT2),
 *     ('spare_0', '2s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_INCREF(__pyx_n_u_antenna_scan_mode);
  __Pyx_GIVEREF(__pyx_n_u_antenna_scan_mode);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_n_u_antenna_scan_mode)) __PYX_ERR(0, 1286, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_2)) __PYX_ERR(0, 1286, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1287
 * TASK_SCAN_INFO = (
 *     ('antenna_scan_mode', UINT2),
 *     ('angular_resolution_desired', SINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '2s'),
 *     ('number_sweeps', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_u_angular_resolution_desired);
  __Pyx_GIVEREF(__pyx_n_u_angular_resolution_desired);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_u_angular_resolution_desired)) __PYX_ERR(0, 1287, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_2)) __PYX_ERR(0, 1287, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1289
 *     ('angular_resolution_desired', SINT2),
 *     ('spare_0', '2s'),
 *     ('number_sweeps', SINT2),             # <<<<<<<<<<<<<<
 *     ('task_scan_type_scan_info', '200s'),   # 200 bytes: task_foo_scan_info
 *     ('spare_1', '112s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 1289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_u_number_sweeps);
  __Pyx_GIVEREF(__pyx_n_u_number_sweeps);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_u_number_sweeps)) __PYX_ERR(0, 1289, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_2)) __PYX_ERR(0, 1289, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1286
 * # 320 bytes: task_scan_info Structure (section 4.2.60, page 65)
 * TASK_SCAN_INFO = (
 *     ('antenna_scan_mode', UINT2),             # <<<<<<<<<<<<<<
 *     ('angular_resolution_desired', SINT2),
 *     ('spare_0', '2s'),
 */
  __pyx_t_2 = PyTuple_New(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_29);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_29)) __PYX_ERR(0, 1286, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_40);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_40)) __PYX_ERR(0, 1286, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_tuple__128)) __PYX_ERR(0, 1286, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_41);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_41)) __PYX_ERR(0, 1286, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_tuple__129)) __PYX_ERR(0, 1286, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__130);
  __Pyx_GIVEREF(__pyx_tuple__130);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_tuple__130)) __PYX_ERR(0, 1286, __pyx_L1_error);
  __pyx_t_29 = 0;
  __pyx_t_40 = 0;
  __pyx_t_41 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_SCAN_INFO, __pyx_t_2) < 0) __PYX_ERR(0, 1285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1296
 * # 200 bytes: task_rhi_scan_info Structure (section 4.2.59, page 64)
 * TASK_RHI_SCAN_INFO = (
 *     ('lower_elevation_limit', UINT2),             # <<<<<<<<<<<<<<
 *     ('upper_elevation_limit', UINT2),
 *     ('azimuth_list', '80s'),            # UINT2[40]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_u_lower_elevation_limit);
  __Pyx_GIVEREF(__pyx_n_u_lower_elevation_limit);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_u_lower_elevation_limit)) __PYX_ERR(0, 1296, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_2)) __PYX_ERR(0, 1296, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1297
 * TASK_RHI_SCAN_INFO = (
 *     ('lower_elevation_limit', UINT2),
 *     ('upper_elevation_limit', UINT2),             # <<<<<<<<<<<<<<
 *     ('azimuth_list', '80s'),            # UINT2[40]
 *     ('spare_0', '115s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 1297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_u_upper_elevation_limit);
  __Pyx_GIVEREF(__pyx_n_u_upper_elevation_limit);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_u_upper_elevation_limit)) __PYX_ERR(0, 1297, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_2)) __PYX_ERR(0, 1297, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1296
 * # 200 bytes: task_rhi_scan_info Structure (section 4.2.59, page 64)
 * TASK_RHI_SCAN_INFO = (
 *     ('lower_elevation_limit', UINT2),             # <<<<<<<<<<<<<<
 *     ('upper_elevation_limit', UINT2),
 *     ('azimuth_list', '80s'),            # UINT2[40]
 */
  __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_41);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_41)) __PYX_ERR(0, 1296, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_40);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_40)) __PYX_ERR(0, 1296, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__131);
  __Pyx_GIVEREF(__pyx_tuple__131);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_tuple__131)) __PYX_ERR(0, 1296, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_tuple__132)) __PYX_ERR(0, 1296, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__133);
  __Pyx_GIVEREF(__pyx_tuple__133);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_tuple__133)) __PYX_ERR(0, 1296, __pyx_L1_error);
  __pyx_t_41 = 0;
  __pyx_t_40 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_RHI_SCAN_INFO, __pyx_t_2) < 0) __PYX_ERR(0, 1295, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1305
 * # 200 bytes: task_ppi_scan_info (section 4.2.57, page 64)
 * TASK_PPI_SCAN_INFO = (
 *     ('left_azimuth_limit', BIN2),             # <<<<<<<<<<<<<<
 *     ('right_azimuth_limit', BIN2),
 *     ('elevation_list', '80s'),          # UINT2[40]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BIN2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_u_left_azimuth_limit);
  __Pyx_GIVEREF(__pyx_n_u_left_azimuth_limit);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_u_left_azimuth_limit)) __PYX_ERR(0, 1305, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_2)) __PYX_ERR(0, 1305, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1306
 * TASK_PPI_SCAN_INFO = (
 *     ('left_azimuth_limit', BIN2),
 *     ('right_azimuth_limit', BIN2),             # <<<<<<<<<<<<<<
 *     ('elevation_list', '80s'),          # UINT2[40]
 *     ('spare_0', '115s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BIN2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 1306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_u_right_azimuth_limit);
  __Pyx_GIVEREF(__pyx_n_u_right_azimuth_limit);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_u_right_azimuth_limit)) __PYX_ERR(0, 1306, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_2)) __PYX_ERR(0, 1306, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1305
 * # 200 bytes: task_ppi_scan_info (section 4.2.57, page 64)
 * TASK_PPI_SCAN_INFO = (
 *     ('left_azimuth_limit', BIN2),             # <<<<<<<<<<<<<<
 *     ('right_azimuth_limit', BIN2),
 *     ('elevation_list', '80s'),          # UINT2[40]
 */
  __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_40);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_40)) __PYX_ERR(0, 1305, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_41);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_41)) __PYX_ERR(0, 1305, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__134);
  __Pyx_GIVEREF(__pyx_tuple__134);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_tuple__134)) __PYX_ERR(0, 1305, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_tuple__132)) __PYX_ERR(0, 1305, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__135);
  __Pyx_GIVEREF(__pyx_tuple__135);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_tuple__135)) __PYX_ERR(0, 1305, __pyx_L1_error);
  __pyx_t_40 = 0;
  __pyx_t_41 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_PPI_SCAN_INFO, __pyx_t_2) < 0) __PYX_ERR(0, 1304, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1314
 * # 200 bytes: task_file_scan_info (section 4.2.54, page 63)
 * TASK_FILE_SCAN_INFO = (
 *     ('first_azimuth', UINT2),             # <<<<<<<<<<<<<<
 *     ('first_elevation', UINT2),
 *     ('filename', '12s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_u_first_azimuth);
  __Pyx_GIVEREF(__pyx_n_u_first_azimuth);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_u_first_azimuth)) __PYX_ERR(0, 1314, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_2)) __PYX_ERR(0, 1314, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1315
 * TASK_FILE_SCAN_INFO = (
 *     ('first_azimuth', UINT2),
 *     ('first_elevation', UINT2),             # <<<<<<<<<<<<<<
 *     ('filename', '12s'),
 *     ('spare_0', '184s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 1315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_u_first_elevation);
  __Pyx_GIVEREF(__pyx_n_u_first_elevation);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_u_first_elevation)) __PYX_ERR(0, 1315, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_2)) __PYX_ERR(0, 1315, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1314
 * # 200 bytes: task_file_scan_info (section 4.2.54, page 63)
 * TASK_FILE_SCAN_INFO = (
 *     ('first_azimuth', UINT2),             # <<<<<<<<<<<<<<
 *     ('first_elevation', UINT2),
 *     ('filename', '12s'),
 */
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_41);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_41)) __PYX_ERR(0, 1314, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_40);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_40)) __PYX_ERR(0, 1314, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__136);
  __Pyx_GIVEREF(__pyx_tuple__136);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_tuple__136)) __PYX_ERR(0, 1314, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__137);
  __Pyx_GIVEREF(__pyx_tuple__137);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_tuple__137)) __PYX_ERR(0, 1314, __pyx_L1_error);
  __pyx_t_41 = 0;
  __pyx_t_40 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_FILE_SCAN_INFO, __pyx_t_2) < 0) __PYX_ERR(0, 1313, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1322
 * # 200 bytes: task_manual_scan_info (section 4.2.55, page 63)
 * TASK_MANUAL_SCAN_INFO = (
 *     ('flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '198s'),
 * )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 1322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_u_flags);
  __Pyx_GIVEREF(__pyx_n_u_flags);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_u_flags)) __PYX_ERR(0, 1322, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_2)) __PYX_ERR(0, 1322, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1323
 * TASK_MANUAL_SCAN_INFO = (
 *     ('flags', UINT2),
 *     ('spare_0', '198s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_40);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_40)) __PYX_ERR(0, 1322, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__138);
  __Pyx_GIVEREF(__pyx_tuple__138);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_tuple__138)) __PYX_ERR(0, 1322, __pyx_L1_error);
  __pyx_t_40 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_MANUAL_SCAN_INFO, __pyx_t_2) < 0) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1328
 * # 320 bytes: task_misc_info Structure (section 4.2.55, page 63)
 * TASK_MISC_INFO = (
 *     ('wavelength', SINT4),             # <<<<<<<<<<<<<<
 *     ('tr_serial_number', '16s'),
 *     ('transmit_power', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_u_wavelength);
  __Pyx_GIVEREF(__pyx_n_u_wavelength);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_u_wavelength)) __PYX_ERR(0, 1328, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_2)) __PYX_ERR(0, 1328, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1330
 *     ('wavelength', SINT4),
 *     ('tr_serial_number', '16s'),
 *     ('transmit_power', SINT4),             # <<<<<<<<<<<<<<
 *     ('flags', UINT2),
 *     ('polarization_type', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 1330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_u_transmit_power);
  __Pyx_GIVEREF(__pyx_n_u_transmit_power);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_u_transmit_power)) __PYX_ERR(0, 1330, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_2)) __PYX_ERR(0, 1330, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1331
 *     ('tr_serial_number', '16s'),
 *     ('transmit_power', SINT4),
 *     ('flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('polarization_type', UINT2),
 *     ('trucation_height', SINT4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_INCREF(__pyx_n_u_flags);
  __Pyx_GIVEREF(__pyx_n_u_flags);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_n_u_flags)) __PYX_ERR(0, 1331, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_2)) __PYX_ERR(0, 1331, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1332
 *     ('transmit_power', SINT4),
 *     ('flags', UINT2),
 *     ('polarization_type', UINT2),             # <<<<<<<<<<<<<<
 *     ('trucation_height', SINT4),
 *     ('spare_0', '18s'),             # 18 bytes
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_n_u_polarization_type);
  __Pyx_GIVEREF(__pyx_n_u_polarization_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_u_polarization_type)) __PYX_ERR(0, 1332, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_2)) __PYX_ERR(0, 1332, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1333
 *     ('flags', UINT2),
 *     ('polarization_type', UINT2),
 *     ('trucation_height', SINT4),             # <<<<<<<<<<<<<<
 *     ('spare_0', '18s'),             # 18 bytes
 *     ('spare_1', '12s'),             # 12 bytes
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 1333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_u_trucation_height);
  __Pyx_GIVEREF(__pyx_n_u_trucation_height);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_u_trucation_height)) __PYX_ERR(0, 1333, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_2)) __PYX_ERR(0, 1333, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1336
 *     ('spare_0', '18s'),             # 18 bytes
 *     ('spare_1', '12s'),             # 12 bytes
 *     ('comment_bytes', SINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_beamwidth', BIN4),
 *     ('vertical_beamwidth', BIN4),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_u_comment_bytes);
  __Pyx_GIVEREF(__pyx_n_u_comment_bytes);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_u_comment_bytes)) __PYX_ERR(0, 1336, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_2)) __PYX_ERR(0, 1336, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1337
 *     ('spare_1', '12s'),             # 12 bytes
 *     ('comment_bytes', SINT2),
 *     ('horizontal_beamwidth', BIN4),             # <<<<<<<<<<<<<<
 *     ('vertical_beamwidth', BIN4),
 *     ('customer_storage', '40s'),    # 40 bytes, uint4[10]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BIN4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_u_horizontal_beamwidth);
  __Pyx_GIVEREF(__pyx_n_u_horizontal_beamwidth);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_u_horizontal_beamwidth)) __PYX_ERR(0, 1337, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_2)) __PYX_ERR(0, 1337, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1338
 *     ('comment_bytes', SINT2),
 *     ('horizontal_beamwidth', BIN4),
 *     ('vertical_beamwidth', BIN4),             # <<<<<<<<<<<<<<
 *     ('customer_storage', '40s'),    # 40 bytes, uint4[10]
 *     ('spare_2', '208s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BIN4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 1338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __Pyx_INCREF(__pyx_n_u_vertical_beamwidth);
  __Pyx_GIVEREF(__pyx_n_u_vertical_beamwidth);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_u_vertical_beamwidth)) __PYX_ERR(0, 1338, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_2)) __PYX_ERR(0, 1338, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1328
 * # 320 bytes: task_misc_info Structure (section 4.2.55, page 63)
 * TASK_MISC_INFO = (
 *     ('wavelength', SINT4),             # <<<<<<<<<<<<<<
 *     ('tr_serial_number', '16s'),
 *     ('transmit_power', SINT4),
 */
  __pyx_t_2 = PyTuple_New(13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_40);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_40)) __PYX_ERR(0, 1328, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__139);
  __Pyx_GIVEREF(__pyx_tuple__139);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_tuple__139)) __PYX_ERR(0, 1328, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_41);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_41)) __PYX_ERR(0, 1328, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_29);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_29)) __PYX_ERR(0, 1328, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_42);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_42)) __PYX_ERR(0, 1328, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_43);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_43)) __PYX_ERR(0, 1328, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__123);
  __Pyx_GIVEREF(__pyx_tuple__123);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_tuple__123)) __PYX_ERR(0, 1328, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__140);
  __Pyx_GIVEREF(__pyx_tuple__140);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_tuple__140)) __PYX_ERR(0, 1328, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_44);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_t_44)) __PYX_ERR(0, 1328, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_45);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 9, __pyx_t_45)) __PYX_ERR(0, 1328, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_46);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 10, __pyx_t_46)) __PYX_ERR(0, 1328, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__141);
  __Pyx_GIVEREF(__pyx_tuple__141);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 11, __pyx_tuple__141)) __PYX_ERR(0, 1328, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__142);
  __Pyx_GIVEREF(__pyx_tuple__142);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 12, __pyx_tuple__142)) __PYX_ERR(0, 1328, __pyx_L1_error);
  __pyx_t_40 = 0;
  __pyx_t_41 = 0;
  __pyx_t_29 = 0;
  __pyx_t_42 = 0;
  __pyx_t_43 = 0;
  __pyx_t_44 = 0;
  __pyx_t_45 = 0;
  __pyx_t_46 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_MISC_INFO, __pyx_t_2) < 0) __PYX_ERR(0, 1327, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1345
 * # 320 bytes: task_end_info Structure (section 4.2.53, page 62)
 * TASK_END_INFO = (
 *     ('task_major_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('task_minor_number', SINT2),
 *     ('task_configuration_file_name', '12s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __Pyx_INCREF(__pyx_n_u_task_major_number);
  __Pyx_GIVEREF(__pyx_n_u_task_major_number);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_u_task_major_number)) __PYX_ERR(0, 1345, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_2)) __PYX_ERR(0, 1345, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1346
 * TASK_END_INFO = (
 *     ('task_major_number', SINT2),
 *     ('task_minor_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('task_configuration_file_name', '12s'),
 *     ('task_description', '80s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_u_task_minor_number);
  __Pyx_GIVEREF(__pyx_n_u_task_minor_number);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_u_task_minor_number)) __PYX_ERR(0, 1346, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_2)) __PYX_ERR(0, 1346, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1349
 *     ('task_configuration_file_name', '12s'),
 *     ('task_description', '80s'),
 *     ('number_tasks', SINT4),             # <<<<<<<<<<<<<<
 *     ('task_state', UINT2),
 *     ('spare_0', '2s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_u_number_tasks);
  __Pyx_GIVEREF(__pyx_n_u_number_tasks);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_u_number_tasks)) __PYX_ERR(0, 1349, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_2)) __PYX_ERR(0, 1349, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1350
 *     ('task_description', '80s'),
 *     ('number_tasks', SINT4),
 *     ('task_state', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '2s'),
 *     ('task_data_time', '12s'),      # 12 bytes: ymds_time
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_u_task_state);
  __Pyx_GIVEREF(__pyx_n_u_task_state);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_u_task_state)) __PYX_ERR(0, 1350, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_2)) __PYX_ERR(0, 1350, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1345
 * # 320 bytes: task_end_info Structure (section 4.2.53, page 62)
 * TASK_END_INFO = (
 *     ('task_major_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('task_minor_number', SINT2),
 *     ('task_configuration_file_name', '12s'),
 */
  __pyx_t_2 = PyTuple_New(9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_46);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_46)) __PYX_ERR(0, 1345, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_45);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_45)) __PYX_ERR(0, 1345, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_tuple__143)) __PYX_ERR(0, 1345, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__144);
  __Pyx_GIVEREF(__pyx_tuple__144);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_tuple__144)) __PYX_ERR(0, 1345, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_44);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_44)) __PYX_ERR(0, 1345, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_43);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_43)) __PYX_ERR(0, 1345, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_tuple__128)) __PYX_ERR(0, 1345, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_tuple__145)) __PYX_ERR(0, 1345, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__146);
  __Pyx_GIVEREF(__pyx_tuple__146);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_tuple__146)) __PYX_ERR(0, 1345, __pyx_L1_error);
  __pyx_t_46 = 0;
  __pyx_t_45 = 0;
  __pyx_t_44 = 0;
  __pyx_t_43 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_END_INFO, __pyx_t_2) < 0) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1359
 * # 12 bytes raw_prod_bhdr structure (section 4.2.30, page 50)
 * RAW_PROD_BHDR = (
 *     ('record_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('sweep_number', SINT2),
 *     ('first_ray_offset', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 1359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_u_record_number);
  __Pyx_GIVEREF(__pyx_n_u_record_number);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_u_record_number)) __PYX_ERR(0, 1359, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_2)) __PYX_ERR(0, 1359, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1360
 * RAW_PROD_BHDR = (
 *     ('record_number', SINT2),
 *     ('sweep_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('first_ray_offset', SINT2),
 *     ('ray_number', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_u_sweep_number);
  __Pyx_GIVEREF(__pyx_n_u_sweep_number);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_u_sweep_number)) __PYX_ERR(0, 1360, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_2)) __PYX_ERR(0, 1360, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1361
 *     ('record_number', SINT2),
 *     ('sweep_number', SINT2),
 *     ('first_ray_offset', SINT2),             # <<<<<<<<<<<<<<
 *     ('ray_number', SINT2),
 *     ('flags', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 1361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_u_first_ray_offset);
  __Pyx_GIVEREF(__pyx_n_u_first_ray_offset);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_u_first_ray_offset)) __PYX_ERR(0, 1361, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_2)) __PYX_ERR(0, 1361, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1362
 *     ('sweep_number', SINT2),
 *     ('first_ray_offset', SINT2),
 *     ('ray_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('flags', UINT2),
 *     ('spare_0', '2s'),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 1362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __Pyx_INCREF(__pyx_n_u_ray_number);
  __Pyx_GIVEREF(__pyx_n_u_ray_number);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_u_ray_number)) __PYX_ERR(0, 1362, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_2)) __PYX_ERR(0, 1362, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1363
 *     ('first_ray_offset', SINT2),
 *     ('ray_number', SINT2),
 *     ('flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '2s'),
 * )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 1363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_n_u_flags);
  __Pyx_GIVEREF(__pyx_n_u_flags);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_u_flags)) __PYX_ERR(0, 1363, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_2)) __PYX_ERR(0, 1363, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1359
 * # 12 bytes raw_prod_bhdr structure (section 4.2.30, page 50)
 * RAW_PROD_BHDR = (
 *     ('record_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('sweep_number', SINT2),
 *     ('first_ray_offset', SINT2),
 */
  __pyx_t_2 = PyTuple_New(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_43);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_43)) __PYX_ERR(0, 1359, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_44);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_44)) __PYX_ERR(0, 1359, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_45);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_45)) __PYX_ERR(0, 1359, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_46);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_46)) __PYX_ERR(0, 1359, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_42);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_42)) __PYX_ERR(0, 1359, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_tuple__128)) __PYX_ERR(0, 1359, __pyx_L1_error);
  __pyx_t_43 = 0;
  __pyx_t_44 = 0;
  __pyx_t_45 = 0;
  __pyx_t_46 = 0;
  __pyx_t_42 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RAW_PROD_BHDR, __pyx_t_2) < 0) __PYX_ERR(0, 1358, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1371
 *     ('structure_header', '12s'),    # 12 bytes: structure_header
 *     ('sweep_start_time', '12s'),    # 12 bytes: ymds_time
 *     ('sweep_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('number_rays_sweep', SINT2),
 *     ('first_ray_index', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 1371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_n_u_sweep_number);
  __Pyx_GIVEREF(__pyx_n_u_sweep_number);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_u_sweep_number)) __PYX_ERR(0, 1371, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_2)) __PYX_ERR(0, 1371, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1372
 *     ('sweep_start_time', '12s'),    # 12 bytes: ymds_time
 *     ('sweep_number', SINT2),
 *     ('number_rays_sweep', SINT2),             # <<<<<<<<<<<<<<
 *     ('first_ray_index', SINT2),
 *     ('number_rays_file_expected', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 1372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __Pyx_INCREF(__pyx_n_u_number_rays_sweep);
  __Pyx_GIVEREF(__pyx_n_u_number_rays_sweep);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_u_number_rays_sweep)) __PYX_ERR(0, 1372, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_2)) __PYX_ERR(0, 1372, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1373
 *     ('sweep_number', SINT2),
 *     ('number_rays_sweep', SINT2),
 *     ('first_ray_index', SINT2),             # <<<<<<<<<<<<<<
 *     ('number_rays_file_expected', SINT2),
 *     ('number_rays_file_actual', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) __PYX_ERR(0, 1373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_u_first_ray_index);
  __Pyx_GIVEREF(__pyx_n_u_first_ray_index);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_u_first_ray_index)) __PYX_ERR(0, 1373, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_2)) __PYX_ERR(0, 1373, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1374
 *     ('number_rays_sweep', SINT2),
 *     ('first_ray_index', SINT2),
 *     ('number_rays_file_expected', SINT2),             # <<<<<<<<<<<<<<
 *     ('number_rays_file_actual', SINT2),
 *     ('fixed_angle', BIN2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 1374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_u_number_rays_file_expected);
  __Pyx_GIVEREF(__pyx_n_u_number_rays_file_expected);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_u_number_rays_file_expected)) __PYX_ERR(0, 1374, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_2)) __PYX_ERR(0, 1374, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1375
 *     ('first_ray_index', SINT2),
 *     ('number_rays_file_expected', SINT2),
 *     ('number_rays_file_actual', SINT2),             # <<<<<<<<<<<<<<
 *     ('fixed_angle', BIN2),
 *     ('bit_per_bin', SINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 1375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_u_number_rays_file_actual);
  __Pyx_GIVEREF(__pyx_n_u_number_rays_file_actual);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_u_number_rays_file_actual)) __PYX_ERR(0, 1375, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_2)) __PYX_ERR(0, 1375, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1376
 *     ('number_rays_file_expected', SINT2),
 *     ('number_rays_file_actual', SINT2),
 *     ('fixed_angle', BIN2),             # <<<<<<<<<<<<<<
 *     ('bit_per_bin', SINT2),
 *     ('data_type', UINT2),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BIN2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_INCREF(__pyx_n_u_fixed_angle);
  __Pyx_GIVEREF(__pyx_n_u_fixed_angle);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_n_u_fixed_angle)) __PYX_ERR(0, 1376, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_2)) __PYX_ERR(0, 1376, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1377
 *     ('number_rays_file_actual', SINT2),
 *     ('fixed_angle', BIN2),
 *     ('bit_per_bin', SINT2),             # <<<<<<<<<<<<<<
 *     ('data_type', UINT2),
 *     ('spare_0', '36s')      # 36 bytes
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 1377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_u_bit_per_bin);
  __Pyx_GIVEREF(__pyx_n_u_bit_per_bin);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_u_bit_per_bin)) __PYX_ERR(0, 1377, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_2)) __PYX_ERR(0, 1377, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1378
 *     ('fixed_angle', BIN2),
 *     ('bit_per_bin', SINT2),
 *     ('data_type', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '36s')      # 36 bytes
 * )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UINT2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) __PYX_ERR(0, 1378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_u_data_type);
  __Pyx_GIVEREF(__pyx_n_u_data_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_u_data_type)) __PYX_ERR(0, 1378, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_2)) __PYX_ERR(0, 1378, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1369
 * # 76 bytes ingest_data_header (section 4.2.15, pages 40)
 * INGEST_DATA_HEADER = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('sweep_start_time', '12s'),    # 12 bytes: ymds_time
 *     ('sweep_number', SINT2),
 */
  __pyx_t_2 = PyTuple_New(11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_tuple__59)) __PYX_ERR(0, 1369, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_tuple__147)) __PYX_ERR(0, 1369, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_42);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_42)) __PYX_ERR(0, 1369, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_46);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_46)) __PYX_ERR(0, 1369, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_45);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_45)) __PYX_ERR(0, 1369, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_44);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_44)) __PYX_ERR(0, 1369, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_43);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_t_43)) __PYX_ERR(0, 1369, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_29);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_t_29)) __PYX_ERR(0, 1369, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_41);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_t_41)) __PYX_ERR(0, 1369, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_40);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 9, __pyx_t_40)) __PYX_ERR(0, 1369, __pyx_L1_error);
  __Pyx_INCREF(__pyx_tuple__148);
  __Pyx_GIVEREF(__pyx_tuple__148);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 10, __pyx_tuple__148)) __PYX_ERR(0, 1369, __pyx_L1_error);
  __pyx_t_42 = 0;
  __pyx_t_46 = 0;
  __pyx_t_45 = 0;
  __pyx_t_44 = 0;
  __pyx_t_43 = 0;
  __pyx_t_29 = 0;
  __pyx_t_41 = 0;
  __pyx_t_40 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_INGEST_DATA_HEADER, __pyx_t_2) < 0) __PYX_ERR(0, 1368, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":1
 * """             # <<<<<<<<<<<<<<
 * A class and supporting functions for reading Sigmet (raw format) files.
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_33);
  __Pyx_XDECREF(__pyx_t_34);
  __Pyx_XDECREF(__pyx_t_35);
  __Pyx_XDECREF(__pyx_t_36);
  __Pyx_XDECREF(__pyx_t_37);
  __Pyx_XDECREF(__pyx_t_38);
  __Pyx_XDECREF(__pyx_t_39);
  __Pyx_XDECREF(__pyx_t_40);
  __Pyx_XDECREF(__pyx_t_41);
  __Pyx_XDECREF(__pyx_t_42);
  __Pyx_XDECREF(__pyx_t_43);
  __Pyx_XDECREF(__pyx_t_44);
  __Pyx_XDECREF(__pyx_t_45);
  __Pyx_XDECREF(__pyx_t_46);
  __Pyx_XDECREF(__pyx_t_47);
  __Pyx_XDECREF(__pyx_t_48);
  __Pyx_XDECREF(__pyx_t_49);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init pyart.io._sigmetfile", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyart.io._sigmetfile");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs = PyTuple_GET_SIZE(kwnames);
    PyObject *dict;
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value);
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (!r) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    #endif
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if PY_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        Py_ssize_t kwsize;
#if CYTHON_ASSUME_SAFE_MACROS
        kwsize = PyTuple_GET_SIZE(kw);
#else
        kwsize = PyTuple_Size(kw);
        if (kwsize < 0) return 0;
#endif
        if (unlikely(kwsize == 0))
            return 1;
        if (!kw_allowed) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, 0);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < kwsize; pos++) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, int wraparound) {
    __Pyx_TypeName obj_type_name;
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is unsliceable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
bad:
    return NULL;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
        
            x = (long)((unsigned long)a + (unsigned long)b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op2);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2)
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_GetMethod;
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* PyFloatBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_TrueDivideCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double a = floatval;
    double b, result;
    CYTHON_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    if (likely(PyFloat_CheckExact(op2))) {
#if CYTHON_COMPILING_IN_LIMITED_API
        b = __pyx_PyFloat_AsDouble(op2);
#else
        b = PyFloat_AS_DOUBLE(op2);
#endif
        if (unlikely(zerodivision_check && ((b) == 0.0))) { PyErr_SetString(PyExc_ZeroDivisionError, "float division by zero"); return NULL;}
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        b = (double) PyInt_AS_LONG(op2);
        if (unlikely(zerodivision_check && ((b) == 0.0))) { PyErr_SetString(PyExc_ZeroDivisionError, "float division by zero"); return NULL;}
    } else
    #endif
    if (likely(PyLong_CheckExact(op2))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsZero(op2)) {
            b = 0.0;
            if (unlikely(zerodivision_check && ((b) == 0.0))) { PyErr_SetString(PyExc_ZeroDivisionError, "float division by zero"); return NULL;}
        } else if (__Pyx_PyLong_IsCompact(op2)) {
            b = (double) __Pyx_PyLong_CompactValue(op2);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op2);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op2);
            switch (size) {
                case -2:
                case 2:
                    if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                        b = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                            if (size == -2)
                                b = -b;
                            break;
                        }
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                case 3:
                    if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                        b = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                            if (size == -3)
                                b = -b;
                            break;
                        }
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                case 4:
                    if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                        b = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                            if (size == -4)
                                b = -b;
                            break;
                        }
                    }
                    CYTHON_FALLTHROUGH;
                default:
        #endif
                    b = PyLong_AsDouble(op2);
                    if (unlikely(b == -1.0 && PyErr_Occurred())) return NULL;
                    #if !CYTHON_USE_PYLONG_INTERNALS
                    if (unlikely(zerodivision_check && ((b) == 0.0))) { PyErr_SetString(PyExc_ZeroDivisionError, "float division by zero"); return NULL;}
                    #endif
        #if CYTHON_USE_PYLONG_INTERNALS
            }
        }
        #endif
    } else {
        return (inplace ? PyNumber_InPlaceTrueDivide : PyNumber_TrueDivide)(op1, op2);
    }
        PyFPE_START_PROTECT("divide", return NULL)
        result = a / b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* PyFloatBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_TrueDivideObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double b = floatval;
    double a, result;
    CYTHON_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    if (likely(PyFloat_CheckExact(op1))) {
#if CYTHON_COMPILING_IN_LIMITED_API
        a = __pyx_PyFloat_AsDouble(op1);
#else
        a = PyFloat_AS_DOUBLE(op1);
#endif
        
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        a = (double) PyInt_AS_LONG(op1);
        
    } else
    #endif
    if (likely(PyLong_CheckExact(op1))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsZero(op1)) {
            a = 0.0;
            
        } else if (__Pyx_PyLong_IsCompact(op1)) {
            a = (double) __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                case 2:
                    if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                        a = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                            if (size == -2)
                                a = -a;
                            break;
                        }
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                case 3:
                    if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                        a = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                            if (size == -3)
                                a = -a;
                            break;
                        }
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                case 4:
                    if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                        a = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                            if (size == -4)
                                a = -a;
                            break;
                        }
                    }
                    CYTHON_FALLTHROUGH;
                default:
        #endif
                    a = PyLong_AsDouble(op1);
                    if (unlikely(a == -1.0 && PyErr_Occurred())) return NULL;
        #if CYTHON_USE_PYLONG_INTERNALS
            }
        }
        #endif
    } else {
        return (inplace ? PyNumber_InPlaceTrueDivide : PyNumber_TrueDivide)(op1, op2);
    }
        PyFPE_START_PROTECT("divide", return NULL)
        result = a / b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
        
            x = (long)((unsigned long)a - (unsigned long)b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return PyLong_FromLong(-intval);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_TrueDivideObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        
            if (8 * sizeof(long) <= 53 || likely(labs(a) <= ((PY_LONG_LONG)1 << 53))) {
                return PyFloat_FromDouble((double)a / (double)b);
            }
            return PyInt_Type.tp_as_number->nb_true_divide(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT && 1 * PyLong_SHIFT < 53) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT && 1 * PyLong_SHIFT < 53) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT && 2 * PyLong_SHIFT < 53) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT && 2 * PyLong_SHIFT < 53) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT && 3 * PyLong_SHIFT < 53) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT && 3 * PyLong_SHIFT < 53) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_true_divide(op1, op2);
            }
        }
                if ((8 * sizeof(long) <= 53 || likely(labs(a) <= ((PY_LONG_LONG)1 << 53)))
                        || __Pyx_PyLong_DigitCount(op1) <= 52 / PyLong_SHIFT) {
                    return PyFloat_FromDouble((double)a / (double)b);
                }
                return PyLong_Type.tp_as_number->nb_true_divide(op1, op2);
            return PyLong_FromLong(x);
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("divide", return NULL)
            result = ((double)a) / (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceTrueDivide : PyNumber_TrueDivide)(op1, op2);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        const long a = intval;
        long x;
        long b = PyInt_AS_LONG(op2);
        
            x = (long)((unsigned long)a + (unsigned long)b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op2))) {
        const long a = intval;
        long b, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG lla = intval;
        PY_LONG_LONG llb, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op2))) {
            return __Pyx_NewRef(op1);
        }
        if (likely(__Pyx_PyLong_IsCompact(op2))) {
            b = __Pyx_PyLong_CompactValue(op2);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op2);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op2);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        b = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        llb = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        b = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        llb = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        b = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        llb = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        b = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        llb = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        b = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        llb = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        b = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        llb = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op2)) {
        const long a = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double b = __pyx_PyFloat_AsDouble(op2);
#else
        double b = PyFloat_AS_DOUBLE(op2);
#endif
            double result;
            
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_MultiplyCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        const long a = intval;
        long b = PyInt_AS_LONG(op2);
        
#ifdef HAVE_LONG_LONG
            if (sizeof(PY_LONG_LONG) > sizeof(long)) {
                PY_LONG_LONG result = (PY_LONG_LONG)a * (PY_LONG_LONG)b;
                return (result >= LONG_MIN && result <= LONG_MAX) ?
                    PyInt_FromLong((long)result) : PyLong_FromLongLong(result);
            }
#endif
#if CYTHON_USE_TYPE_SLOTS
            return PyInt_Type.tp_as_number->nb_multiply(op1, op2);
#else
            return PyNumber_Multiply(op1, op2);
#endif
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op2))) {
        const long a = intval;
        long b, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG lla = intval;
        PY_LONG_LONG llb, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op2))) {
            return __Pyx_NewRef(op2);
        }
        if (likely(__Pyx_PyLong_IsCompact(op2))) {
            b = __Pyx_PyLong_CompactValue(op2);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op2);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op2);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                        b = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                        llb = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                        b = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                        llb = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                        b = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                        llb = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                        b = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                        llb = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                        b = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                        llb = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                        b = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                        llb = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
            }
        }
                CYTHON_UNUSED_VAR(a);
                CYTHON_UNUSED_VAR(b);
                #ifdef HAVE_LONG_LONG
                llb = b;
                goto long_long;
                #else
                return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
                #endif
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla * llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op2)) {
        const long a = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double b = __pyx_PyFloat_AsDouble(op2);
#else
        double b = PyFloat_AS_DOUBLE(op2);
#endif
            double result;
            
            PyFPE_START_PROTECT("multiply", return NULL)
            result = ((double)a) * (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceMultiply : PyNumber_Multiply)(op1, op2);
}
#endif

/* PyIntCompare */
static CYTHON_INLINE int __Pyx_PyInt_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 1;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        return (a == b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) == 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 0;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 0;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal == 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
        return ((double)a == (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparsable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, int is_complex) {
  CYTHON_UNUSED_VAR(is_complex);
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, int is_complex) {
  CYTHON_UNUSED_VAR(is_complex);
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static int
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return -1;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return -1;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return -1;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i]) {
            PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
            return -1;
        }
        if (*ts != ',' && *ts != ')') {
            PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
            return -1;
        }
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim) {
        PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
        return -1;
    }
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return -1;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return 0;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (__pyx_buffmt_parse_array(ctx, &ts) < 0) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
  static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (unlikely(info->buf == NULL)) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}
static void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static int __Pyx__GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  buf->buf = NULL;
  if (unlikely(__Pyx_GetBuffer(obj, buf, flags) == -1)) {
    __Pyx_ZeroBuffer(buf);
    return -1;
  }
  if (unlikely(buf->ndim != nd)) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if (unlikely((size_t)buf->itemsize != dtype->size)) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_SafeReleaseBuffer(buf);
  return -1;
}

/* BufferIndexError */
  static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

/* PyIntCompare */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            if (__Pyx_PyLong_IsZero(op1) == 1) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                Py_RETURN_FALSE;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AndObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        
            return PyInt_FromLong(a & b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op1);
        }
        if ((intval & PyLong_MASK) == intval) {
            long last_digit = (long) __Pyx_PyLong_Digits(op1)[0];
            long result = intval & (likely(__Pyx_PyLong_IsPos(op1)) ? last_digit : (PyLong_MASK - last_digit + 1));
            return PyLong_FromLong(result);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_and(op1, op2);
            }
        }
                x = a & b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla & llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceAnd : PyNumber_And)(op1, op2);
}
#endif

/* PyFloatBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_SubtractObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double b = floatval;
    double a, result;
    CYTHON_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    if (likely(PyFloat_CheckExact(op1))) {
#if CYTHON_COMPILING_IN_LIMITED_API
        a = __pyx_PyFloat_AsDouble(op1);
#else
        a = PyFloat_AS_DOUBLE(op1);
#endif
        
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        a = (double) PyInt_AS_LONG(op1);
        
    } else
    #endif
    if (likely(PyLong_CheckExact(op1))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsZero(op1)) {
            a = 0.0;
            
        } else if (__Pyx_PyLong_IsCompact(op1)) {
            a = (double) __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                case 2:
                    if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                        a = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                            if (size == -2)
                                a = -a;
                            break;
                        }
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                case 3:
                    if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                        a = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                            if (size == -3)
                                a = -a;
                            break;
                        }
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                case 4:
                    if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                        a = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                            if (size == -4)
                                a = -a;
                            break;
                        }
                    }
                    CYTHON_FALLTHROUGH;
                default:
        #endif
                    a = PyLong_AsDouble(op1);
                    if (unlikely(a == -1.0 && PyErr_Occurred())) return NULL;
        #if CYTHON_USE_PYLONG_INTERNALS
            }
        }
        #endif
    } else {
        return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
    }
        PyFPE_START_PROTECT("subtract", return NULL)
        result = a - b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* SliceObject */
  static CYTHON_INLINE int __Pyx_PyObject_SetSlice(PyObject* obj, PyObject* value,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, int wraparound) {
    __Pyx_TypeName obj_type_name;
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_ass_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_ass_slice(obj, cstart, cstop, value);
    }
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_ass_subscript))
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    {
        int result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_ass_subscript(obj, py_slice, value);
#else
        result = value ? PyObject_SetItem(obj, py_slice, value) : PyObject_DelItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object does not support slice %.10s",
        obj_type_name, value ? "assignment" : "deletion");
    __Pyx_DECREF_TypeName(obj_type_name);
bad:
    return -1;
}

/* PyFloatBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_SubtractCObj(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double a = floatval;
    double b, result;
    CYTHON_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    if (likely(PyFloat_CheckExact(op2))) {
#if CYTHON_COMPILING_IN_LIMITED_API
        b = __pyx_PyFloat_AsDouble(op2);
#else
        b = PyFloat_AS_DOUBLE(op2);
#endif
        
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        b = (double) PyInt_AS_LONG(op2);
        
    } else
    #endif
    if (likely(PyLong_CheckExact(op2))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsZero(op2)) {
            b = 0.0;
            
        } else if (__Pyx_PyLong_IsCompact(op2)) {
            b = (double) __Pyx_PyLong_CompactValue(op2);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op2);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op2);
            switch (size) {
                case -2:
                case 2:
                    if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                        b = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                            if (size == -2)
                                b = -b;
                            break;
                        }
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                case 3:
                    if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                        b = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                            if (size == -3)
                                b = -b;
                            break;
                        }
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                case 4:
                    if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                        b = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (b < (double) ((PY_LONG_LONG)1 << 53))) {
                            if (size == -4)
                                b = -b;
                            break;
                        }
                    }
                    CYTHON_FALLTHROUGH;
                default:
        #endif
                    b = PyLong_AsDouble(op2);
                    if (unlikely(b == -1.0 && PyErr_Occurred())) return NULL;
        #if CYTHON_USE_PYLONG_INTERNALS
            }
        }
        #endif
    } else {
        return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
    }
        PyFPE_START_PROTECT("subtract", return NULL)
        result = a - b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* PyIntCompare */
  static CYTHON_INLINE int __Pyx_PyInt_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 0;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        return (a != b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) != 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 1;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 1;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal != 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
        return ((double)a != (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* FixUpExtensionType */
  #if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* PyObjectCallNoArg */
  static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod0 */
  static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* ValidateBasesTuple */
  #if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_MACROS
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (n < 0) return -1;
#endif
    for (i = 1; i < n; i++)
    {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !(CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY)
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
  static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* PyObject_GenericGetAttrNoDict */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
  static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
  static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* SetupReduce */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType_3_0_12
#define __PYX_HAVE_RT_ImportType_3_0_12
static PyTypeObject *__Pyx_ImportType_3_0_12(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_12 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_12 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_12 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
  static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.') != NULL) {
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportDottedModule */
  #if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__19;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* FetchSharedCythonModule */
  static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef((char*) __PYX_ABI_MODULE_NAME);
}

/* FetchCommonType */
  static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
  #if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
  #if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#elif PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyObject *py_name = NULL;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
    if (!py_name) return NULL;
    PyErr_Format(PyExc_TypeError, "%.200S() takes no keyword arguments",
                 py_name);
    Py_DECREF(py_name);
#else
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
#endif
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_MACROS
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(!argc) < 0) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
  static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #else
    return NULL;
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    __Pyx_TypeName obj_type_name;
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
                 "'" __Pyx_FMT_TYPENAME "' does not have the buffer interface",
                 obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


  /* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
      return ::std::complex< long double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
      return x + y*(__pyx_t_long_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
      __pyx_t_long_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_sum_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        __pyx_t_long_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_diff_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        __pyx_t_long_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_prod_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        __pyx_t_long_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsl(b.real) >= fabsl(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                long double r = b.imag / b.real;
                long double s = (long double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_long_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            long double r = b.real / b.imag;
            long double s = (long double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_long_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            long double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_long_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_neg_long__double(__pyx_t_long_double_complex a) {
        __pyx_t_long_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_long__double(__pyx_t_long_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_conj_long__double(__pyx_t_long_double_complex a) {
        __pyx_t_long_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE long double __Pyx_c_abs_long__double(__pyx_t_long_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtl(z.real*z.real + z.imag*z.imag);
          #else
            return hypotl(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_pow_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
            __pyx_t_long_double_complex z;
            long double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    long double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_long__double(a, a);
                    case 3:
                        z = __Pyx_c_prod_long__double(a, a);
                        return __Pyx_c_prod_long__double(z, a);
                    case 4:
                        z = __Pyx_c_prod_long__double(a, a);
                        return __Pyx_c_prod_long__double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powl(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2l(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_long__double(a);
                theta = atan2l(a.imag, a.real);
            }
            lnr = logl(r);
            z_r = expl(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosl(z_theta);
            z.imag = z_r * sinl(z_theta);
            return z;
        }
    #endif
#endif

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_short(short value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const short neg_one = (short) -1, const_zero = (short) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(short) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(short) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(short) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(short) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(short) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(short),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(short));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FormatTypeName */
  #if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__149);
    }
    return name;
}
#endif

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
  static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* InitStrings */
  #if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
